// Copyright 2003 Regents of the University of California

// Astropulse is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// Astropulse is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with Astropulse; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/* ap_fileio.C */
/* Functions for reading data, etc for AstroPulse client */
#include "ap_config.h"

#ifdef _WIN32
#include "windows.h"
#endif

#include <cstdio>
#include <cstdlib>
#include <vector>
#include <string>
#include <iostream>
#include <sstream>
#include <algorithm>

#include "parse.h"
#include "xml_util.h"
#include "db_table.h"
#include "ap_schema.h"
#include "astropulse.h"
#include "util.h"
#undef read

#include <cstring>
#include <cassert>
#include <cstdlib>
#include <cstdio>
#include <cerrno>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif


namespace Astropulse {
const int AP_OUTPUT_PROTOCOL_VERSION = 3;

// Filenames
const char* WU_FNAME    = "in.dat";
const char* OUT_FNAME   = "pulse.out";
const char* STATE_FNAME = "ap_state.dat";
const char* FOLD_FNAME  = "fold.dat";

inline size_t get_file_size(const char* filename) {
  FILE* f = fopen(filename, "r");
  if (!f) return 0;
  fseek(f, 0, SEEK_END);
  size_t size = ftell(f);
  fclose(f);
  return size;
}

inline bool file_exists_p(const char* filename) {
  FILE* f = fopen(filename, "r");
  if (f) {
    fclose(f);
    return true;
  } else {
    return false;
  }
}

void read_signal_vector(std::vector<ap_signal> &signal_vector) {
  client.status = "Reading signal vector.";

  string resolved_filename;
  /* See Outfile::init for discussion */
  boinc_resolve_filename_s(OUT_FNAME, resolved_filename);

  bool signal_file_exists = file_exists_p(resolved_filename.c_str());
  std::ifstream signal_file(resolved_filename.c_str(), std::ifstream::in);
  if(!signal_file) {
    printf("No signal file %s, signal vector will remain empty\n", resolved_filename.c_str());
    // fprintf(stderr, "Error opening signal file %s: %s\n", resolved_filename.c_str(), strerror(errno));
    // exit(FOPEN_ERR);
    return;
  }

  ap_signal s;

  while(!signal_file.eof()) {
    signal_file >> s;
    if(is_blank(s)) break;
    signal_vector.push_back(s);
  }

  signal_file.close();
}

bool is_blank(ap_signal s) {
  if(s.dm == 0 && s.peak_bin == 0 && s.scale == 0)
    return 1;
  else
    return 0;
}

/* Add signal s to vector signal_vector.
 * If another signal in signal_vector already has the same
 * peak_bin as the signal s, then we choose whichever of the two
 * signals has the larger peak_power.
 */
bool add_signal(std::vector<ap_signal> &signal_vector, ap_signal s) {
  std::vector<ap_signal>::iterator iter;
  iter = signal_vector.begin();
  while(iter != signal_vector.end()) {
    if(similar_ap_signals(*iter, s)) {
      // Choose between the two signals, based on power level.
      if(s.peak_power > (*iter).peak_power)
        *iter = s;
      return false;
    }
    iter++;
  }
  signal_vector.push_back(s);
  return true;
}

bool similar_ap_signals(ap_signal s1, ap_signal s2) {
  if(s1.peak_bin == s2.peak_bin) return true;
  if(s1.period != 0 && s2.period != 0 && 
    fabs(1 - s2.period / s1.period) < 0.01) 
     return true;
  return false;
}

void Outfile::init() {
  client.status = "Opening output file.";

  string resolved_filename;
  /* boinc API.  finds the resolved filename (physical filename)
   * corresponding to the logical filename OUT_FNAME.
   * The association between the two is established in pulse_split.C,
   * by the boinc API function create_work.  (I think.)
   */
  boinc_resolve_filename_s(OUT_FNAME, resolved_filename);

  // size_t size = get_file_size(resolved_filename.c_str());
  bool outfile_exists = file_exists_p(resolved_filename.c_str());

  if (outfile.open(resolved_filename.c_str(), "w")) {
    fprintf(stderr, "Error opening outfile %s: %s\n", resolved_filename.c_str(), strerror(errno));
    exit(FOPEN_ERR);
  }

  // consistency checking
  if (client.statefile.resumed) {
    /* This is okay now, because Outfile::init gets called inside of
     * Outfile::write_main, so there may be no outfile yet.
    if (!outfile_exists) {
      fprintf(stderr,
              "Error resuming: statefile exists, but output file doesn't exist.\n");
      exit(FREAD_ERR);
    }
    */
  } else {
    /* I don't see how this is inconsistent.  The output file gets
       created first, and there is a pause of 1-2 seconds before the
       statefile gets created.  So it's entirely possible for the
       program to crash or quit during that 1-2 seconds.
    if (outfile_exists) {
      fprintf(stderr,
              "Error initializing: statefile doesn't exist, but output file exists.\n");
      exit(FREAD_ERR);
    }
    */

    outfile.flush();
  }
}

// Read in state file to resume computation
void Statefile::Read() {
  client.status = "Reading state file.";

  boinc_time_to_checkpoint();

  FILE* statefile = fopen(STATE_FNAME, "rb");
  if (statefile == NULL) return;
  int nread = fread(&client.state, 1, sizeof(state_t), statefile);
  // fclose(statefile);
  // exit(0);
  if (nread != sizeof(state_t)) {
    fprintf(stderr,
            "Error reading from statefile: wanted %ld bytes, got %ld\n",
            sizeof(state_t), nread
           );
    exit(FREAD_ERR);
  }
  fclose(statefile);
  resumed = true;
  if (boinc_is_standalone()) {
    fprintf(stderr, "## Warning: resuming in standalone mode\n");
  }

  return;
}

// Write state file so we can resume later
void Statefile::write() {
  client.status = "Saving computation state.";

  boinc_time_to_checkpoint();

  FILE* statefile = fopen(STATE_FNAME, "wb");
  if (statefile == NULL) {
    fprintf(stderr, "Error opening statefile for write: %s.\n", strerror(errno));
    exit(FOPEN_ERR);
  }

  int nwrite = fwrite(&client.state, sizeof(state_t), 1, statefile);
  if (nwrite != 1) {
    fprintf(stderr, "Error writing to statefile: %s.\n", strerror(errno));
    exit(FWRITE_ERR);
  }

  fclose(statefile);
}

void Foldfile::write() {
  client.status = "Saving fold buffer.";

  boinc_time_to_checkpoint();

  FILE* foldfile = fopen(FOLD_FNAME, "wb");
  if (foldfile == NULL) {
    fprintf(stderr, "Error opening foldfile for write: %s.\n", strerror(errno));
    exit(FOPEN_ERR);
  }

  int nwrite = fwrite(&(client.fold_buf[0]), sizeof(float)*client.state.fold_buf_size, 1, foldfile);
  if (nwrite != 1) {
    fprintf(stderr, "Error writing to foldfile: %s.\n", strerror(errno));
    exit(FWRITE_ERR);
  }

  fclose(foldfile);
}

// Read saved folding array
int Foldfile::Read() {
  client.status = "Reading fold buffer.";
  boinc_time_to_checkpoint();

  FILE* foldfile = fopen(FOLD_FNAME, "rb");
  if (foldfile == NULL) {
    return(1);
  }

  int nread = fread(&(client.fold_buf[0]), sizeof(float), client.state.fold_buf_size, foldfile);
  if (nread != client.state.fold_buf_size) {
    fprintf(stderr, "Error reading from foldfile: wanted %ld bytes, but read only %ld bytes.\n", sizeof(float)*client.state.fold_buf_size, nread);
    return(1);
  }

  fclose(foldfile);
  return(0);
}

void Wufile::init() {
  client.status = "Opening WU file.";
  string resolved_filename;
  /* boinc API.  See discussion in Outfile::init */
  boinc_resolve_filename_s(WU_FNAME, resolved_filename);

  wufile.open(resolved_filename.c_str(),std::ios_base::binary);
  if (!wufile) {
    fprintf(stderr, "Error opening wufile %s: %s\n", resolved_filename.c_str(), strerror(errno));
    exit(FOPEN_ERR);
  }
}

// This parses the wuheader into client.wuheader, which
// is a temporary storage location.  It is used
// 
void Wufile::parse_header_for_writeout() {
  workunit_header &wuheader(client.wuheader);

  wufile.seekg(0, std::ios::beg);
  wufile >> wuheader;

  if (!wufile) {
    fprintf(stderr, "Error parsing WU header.\n");
    exit(PARSE_ERR);
  }
}

// Parse header from wufile
void Wufile::parse_header_to_state() {
  workunit_header &wuheader(client.wuheader);

  state_t& state = client.state;

  static const char *locations[]= {"Arecibo 403MHz Line Feed",
                                   "Arecibo 1.42GHz Flat Feed",
                                   "Arecibo 1.42GHz Gregorian Feed",
                                   "Unknown"
                                  };
  int s4_id=3;
  s4_id=wuheader.receiver_cfg->s4_id;
  if (s4_id>2) {
    locations[0]=client.wuheader.receiver_cfg->name;
    s4_id=0;
  }

  strncpy(state.recvname,locations[s4_id],sizeof(state.recvname));
  strncpy(state.wuname,wuheader.name,sizeof(state.wuname));
  state.datasize=wuheader.data_desc.nsamples*wuheader.recorder_cfg->bits_per_sample/8;
  state.max_coadd=wuheader.analysis_cfg->max_coadd;
  state.dm_low=wuheader.analysis_cfg->dm_low;
  state.dm_hi=wuheader.analysis_cfg->dm_high;
  state.dm_chunk=wuheader.analysis_cfg->dm_chunk;
  state.fft_len=wuheader.analysis_cfg->fft_len;
  state.high_pass=wuheader.analysis_cfg->high_pass;
  state.ffa_threshold=wuheader.analysis_cfg->ffa_threshold;
  state.ra0=wuheader.data_desc.start.ra;
  state.dec0=wuheader.data_desc.start.dec;
  state.jd0=wuheader.data_desc.start.time;
  state.ra1=wuheader.data_desc.end.ra;
  state.dec1=wuheader.data_desc.end.dec;
  state.jd1=wuheader.data_desc.end.time;
  for (unsigned int i=0; i<wuheader.analysis_cfg->thresholds.size(); i++) {
    state.thresh[wuheader.analysis_cfg->thresholds[i].coadds]=
      wuheader.analysis_cfg->thresholds[i].threshold;
  }

  // Check for inconsistent/bad values
  if (state.fft_len > state.datasize*4) {
    fprintf(stderr, "Error parsing WU header: fft_len=%d > datasize=%d * 4\n",
            state.fft_len, state.datasize);
    exit(PARSE_ERR);
  }
  if (state.dm_hi <= state.dm_low) {
    fprintf(stderr, "Error parsing WU header: dm_hi=%d <= dm_low=%d\n",
            state.dm_hi, state.dm_low);
    exit(PARSE_ERR);
  }
  if (state.max_coadd > MAX_COADD) {
    fprintf(stderr, "Error parsing WU header: max_coadd=%d > %d\n",
            state.max_coadd, MAX_COADD);
    exit(PARSE_ERR);
  }
  if ((state.dm_chunk>>state.max_coadd) > 1) {
    fprintf(stderr, "Error parsing WU header: (dm_chunk=%d) >> (max_coadd=%d) > 1\n",
            state.dm_chunk, state.max_coadd);
    exit(PARSE_ERR);
  }
  if (state.fft_len < 0) {
    fprintf(stderr, "Error parsing WU header: fft_len=%d < 0\n",
            state.fft_len);
    exit(PARSE_ERR);
  }
  if (state.datasize < 0) {
    fprintf(stderr, "Error parsing WU header: datasize=%d < 0\n",
            state.datasize);
    exit(PARSE_ERR);
  }
  if (state.max_coadd < 0) {
    fprintf(stderr, "Error parsing WU header: max_coadd=%d < 0\n",
            state.max_coadd);
    exit(PARSE_ERR);
  }
  if (state.dm_hi < 0) {
    fprintf(stderr, "Error parsing WU header: dm_hi=%d < 0\n",
            state.dm_hi);
    exit(PARSE_ERR);
  }
  if (state.dm_low < 0) {
    fprintf(stderr, "Error parsing WU header: dm_low=%d < 0\n",
            state.dm_low);
    exit(PARSE_ERR);
  }

  // client.outfile.write_header();
}

// returns true on success
inline bool find_start_tag(FILE* file, const char* tag) {
  rewind(file);
  char buf[256];
  while (fgets(buf, 256, file)) {
    if (match_tag(buf, tag)) return true;
  }
  return false;
}

void Wufile::read_raw_data() {
  char buf[4096];
  client.status = "Reading raw data.";
  boinc_time_to_checkpoint();
  std::string data_chunk;
  data_chunk.reserve(client.state.datasize+1024);
  wufile.unsetf(std::ios_base::skipws);
  while (!wufile.eof()) {
    wufile.read(buf,4096);
    data_chunk+=std::string(buf,wufile.gcount());
  }
  client.rawdata=xml_decode_field<unsigned char>(data_chunk,(const char *)"data");
  if (client.rawdata.size() != client.state.datasize) {
    fprintf(stderr, "Error reading raw data: Expected %ld bytes, read %ld.\n", client.state.datasize, client.rawdata.size());
    exit(FREAD_ERR);
  }
}

void Wufile::finish() {
  wufile.close();
}

// encode the time series as a ASCII integers

void Outfile::write_header() {
  outfile.printf("<astropulse_output>\n");
  std::string s(client.wuheader.print_xml());
  outfile.printf("%s",s.c_str());
  outfile.printf("<pulses>\n");
}

void Outfile::output_result(const ap_signal &sig) {
  std::string s(sig.print_xml(0,0,1));
  outfile.printf("%s",s.c_str());
  // REMINDER: do not flush here; only do so at checkpoint.  All state
  // files have to be consistent in case of crash.
}

void Outfile::write_main(std::vector<ap_signal> signal_vector) {
  if(outfile.tell() != -1) 
    outfile.close();
  init();
  write_header();
  std::vector<ap_signal>::iterator iter1;
  iter1 = signal_vector.begin();
  while(iter1 != signal_vector.end()) {
    output_result(*iter1);
    iter1++;
  }
  outfile.flush();
  outfile.close();

} 

// write best results for each coadd level (we do this for the sake of
// the validator).
void Outfile::finish_results() {

  if(outfile.tell() != -1) 
    outfile.close();

  string resolved_filename;
  /* See Outfile::init for discussion */
  boinc_resolve_filename_s(OUT_FNAME, resolved_filename);
  
  if (outfile.open(resolved_filename.c_str(), "a")) {
    fprintf(stderr, "Error opening outfile %s: %s\n", resolved_filename.c_str(), strerror(errno));
    exit(FOPEN_ERR);
  }

  outfile.printf("</pulses>\n");
  outfile.printf("<best_pulses>\n");
  state_signal *best=client.state.best;
  for (int l = 0; l < client.state.max_coadd; ++l) {
    assert(best[l].peak_power > -INFINITY);
    std::string s((ap_signal(best[l])).print_xml(0,0,1));
    outfile.printf("%s",s.c_str());
  }
  outfile.printf("</best_pulses>\n");
  outfile.printf("</astropulse_output>\n");
  outfile.flush();
  outfile.close();
}

void Outfile::finish() {
  outfile.flush();
  outfile.close();
}

void Statefile::finish() {
  unlink(STATE_FNAME);
}

void Foldfile::finish() {
  unlink(FOLD_FNAME);
}
}

