// Copyright 2003 Regents of the University of California

// Astropulse is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// Astropulse is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with Astropulse; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef H_ASTROPULSE
#define H_ASTROPULSE
#define DEBUGGING_0 // Tells the client to regularly inform stdout about its progress

#include "ctype.h"
#include "stdio.h"
#include <string>
#include "fftw3.h"
#include "boinc_api.h"
#include "mfile.h"
#include "db_table.h"
#include "ap_schema.h"
#include "time.h"

namespace Astropulse {
using std::string;

static const int MAX_COADD = 15;
static const int MAXLENGTH_OF_TS = 64;


enum OnOff {
  on, off
};

const double frac_in_main_ffa[10] = {0.895, 0.89, 0.87, 0.85, 0.8, 
                                   // 0      1     2     3     4 
                                     0.7, 0.6, 0.4, 0.2, 0.05};
                                   // 5    6    7    8    9

const double frac_in_btt[10] = {0.1,   0.1,  0.1,  0.1,  0.1,
							   // 0      1     2     3     4
							   0.1, 0.1, 0.1, 0.1, 0.05};
							   // 5    6    7    8    9

inline int int_log2(int input) {
   int i = 0;
   while(input >>= 1) i++;
   return i;
}

class state_signal {
    public:
        sqlint8_t  id;
        db_reference<result,sqlint8_t> resultid;
        long  fft_num;
        long  peak_bin;
        double  peak_power;
        long  scale;
        long  dm;
        double  period;
        double  ra;
        double  decl;
        double  time;
        long  rfi_found;
        long  reserved;
        long  time_series_len;
        // sqlblob<unsigned char>  time_series;
        unsigned char time_series[MAXLENGTH_OF_TS];

        state_signal() {}
        state_signal(const ap_signal &sig) : 
            id(sig.id),
            fft_num(sig.fft_num), peak_bin(sig.peak_bin),
            peak_power(sig.peak_power), scale(sig.scale), dm(sig.dm),
            period(sig.period), ra(sig.ra), decl(sig.decl), time(sig.time),
            rfi_found(sig.rfi_found), reserved(sig.reserved),
            time_series_len(sig.time_series_len) {

            memcpy(time_series, &(sig.time_series[0]), sig.time_series.size());
        }
        operator ap_signal() {
            ap_signal sig;
            sig.id = id;
            // sig.resultid = resultid;
            sig.fft_num = fft_num;
            sig.peak_bin = peak_bin;
            sig.peak_power = peak_power;
            sig.scale = scale;
            sig.dm = dm;
            sig.period = period;
            sig.ra = ra;
            sig.decl = decl;
            sig.time = time;
            sig.rfi_found = rfi_found;
            sig.reserved = reserved;
            sig.time_series_len = time_series_len;
            sig.time_series = sqlblob<unsigned char>(time_series, 64);
            return sig;
        }
};

class timer {
  private:
    clock_t ticks;
    clock_t last_clock;
    clock_t start_clock;

    long seconds;
    long last_time;
    long start_seconds;

    OnOff state;

  public:
    timer();
    void start();
    void stop();
    void reset();
    clock_t get_ticks();
    clock_t get_total_ticks();
    long get_seconds();
    long get_total_seconds();
    void print_total_time();
};


// Structure which includes everything that might be in a wu header.  Also
// contains info for state file.
struct state_t {
  char recvname[64];      // receiver name
  char wuname[64];        // workunit name
  int datasize;           // Size of raw data in bytes
  int dm_low;             // Lowest dm to calculate
  int dm_hi;              // Higest dm
  int dm_chunk;           // dm resolution for folding - also dm table size
  int fft_len;            // fft length >= 2*dm_hi
  float high_pass;       // Hertz.  See ap_science.cpp, high_pass_filter()
  float ffa_threshold;    // lcgf threshold, should be a negative number
  int max_coadd;          // max level to collapse down to ( >= log(dm_chunk) )
  int fold_buf_size;      // size of array to fold
  int fold_level;         // should be log2(dm_chunk)
  double thresh[MAX_COADD];// array of thresholds (max_collapse total)
  double ra0;             // RA at start of wu, in hours
  double ra1;             // RA at end of wu
  double dec0;            // Dec at start of wu, in degrees
  double dec1;            // Dec at end of wu
  double jd0;             // JD at start
  double jd1;             // JD at end
  int data_chunk_now;     // Current location being worked on
  int dm_chunk_now;       // Current dm chunk
  int dm_now;             // Current dm within chunk
  int dm_sign;            // positive (+1) or negative (-1) dispersion
  int num_sub_buffers;    // for ffa number of sub buffers
  int sub_buffer;         // for ffa, current sub buffer
  float freq;             // for ffa, = number of periods
  float min_freq;         // for ffa, minimum freq to consider
  int nfb;				  // for ffa build threshold table
  int fold_buf_loc;       // location to write in fold buffer
  double frac_done;       // fraction of wu completed so far
  double pulse_jd;        // dummy to calculate the julian date of pulse
  int result_count;       // number of results found so far
  state_signal best[MAX_COADD];

  state_t();
  void compute_fraction_done(double l);
};

struct time_series_f_t {
  enum { MAXLENGTH = 256 };
  std::vector<float> data;
  int length;

  time_series_f_t() : length(0) {
    data.reserve(MAXLENGTH);
  }
  void build(float array[], int index, int range);
};

struct time_series_b_t {
  enum { MAXLENGTH = 64 };
  std::vector<unsigned char> data;
  int length;

  time_series_b_t() : length(0) {
    data.reserve(MAXLENGTH);
  }
  void build(float array[], int index, int range, float n_bins);
};

enum ERROR_CODES {
  FWRITE_ERR = -1,                      // Error writing a file.
  SBTF_ERR  = -2,                       // Error from splitter_bits_to_float
  FOPEN_ERR = -3,
  PARSE_ERR = -4,
  FREAD_ERR = -5
};

class Wufile {
    std::ifstream wufile;
  public:
    void init();
    void parse_header_to_state();
    void parse_header_for_writeout();
    void read_raw_data();
    void finish();
};

void read_signal_vector(std::vector<ap_signal> &signal_vector);
bool add_signal(std::vector<ap_signal> &signal_vector, ap_signal s);
bool is_blank(ap_signal s);
bool similar_ap_signals(ap_signal s1, ap_signal s2);


class Outfile {
    /* MFILE is a boinc API construct that can be used to
     * write to output and state files.  It "facilitates atomic
     * checkpoint" meaning it allows output to be written as
     * a single operation (instead of separate character writes?)
     */
    MFILE outfile;
  public:
    void init();
    void finish_results();
    void finish();

    void output_result(const ap_signal &s);
    void write_header();
    void write_main(std::vector<ap_signal> signal_vector);
    void flush() {
      outfile.flush();
    }
};

class Statefile {
  public:
    bool resumed;
    Statefile() : resumed(0) {}
    void Read();
    void write();
    void finish();
};

class Foldfile {
  public:
    int Read();
    void write();
    void finish();
};

class Science {
    fftwf_complex* data;
    fftwf_complex* temp;
    fftwf_complex* chirp;
    fftwf_plan fp, ip;

    class Wisdom {
      public:
        void load();
        void save();
    };
    Wisdom wisdom;
  public:
    void init();
    void mainloop();
    void finish();
  protected:
    void build_chirp_table(int dm_start);
    void convert_bits_to_float();
    void compute_forward_fft();
    void high_pass_filter();
    void dechirp(int dm, float *power, int sign);
    void coadd_in_place(float *power, int size);
    int ffa(float *data, int num_sub_buffers, bool write_lcgf, bool debug_msg, int pulse_limit);
    void build_threshold_table(int*& table, float thresh_lcgf, bool debug_msg);
    int lcgf_binary_search(int min, int max, int ntc, float thresh_lcgf);
};

class Client {
  public:
    workunit_header wuheader;
    state_t state;
    const char* status;         // Spot to put program status for display.
    // NOTE: any data assigned to `status' must be static data -- if this
    // constraint is not acceptable, change it back to `string' instead
    // of `const char*'.

    Outfile   outfile;
    Wufile    wufile;
    Foldfile  foldfile;
    Statefile statefile;

    Science science;

    std::vector<ap_signal> signal_vector;

    std::vector<unsigned char> rawdata;
    std::vector<float> power;
    std::vector<float> fold_buf;

    void init();
    void finish();

    inline void do_checkpoint() {
      // outfile.flush();
      outfile.write_main(signal_vector);
      foldfile.write();
      statefile.write();
    };
    inline void checkpoint() {
      if (boinc_time_to_checkpoint()) {
	do_checkpoint();
	boinc_checkpoint_completed();
      }
    };
};

extern Client client;
extern timer main_timer;
};

#ifndef INFINITY
#define INFINITY (-log(0.0))
#endif

#endif


