#ifndef INCapShmComh
#define INCapShmComh

/*
 *	data structure and commands for shm communication between
 *	vxWorks apProg and skybolt programs.
 *
 *	communication sequence is:
 *
 *       apProg 			skyprog
 *
 *      start skyprog   -->
 *         wait...                      clear data struct shm. 
 *					uncache pages
 *       	        <--		Signal startup done.
 *	wakeUp
 *      signal setup info -->                   waiting
 *         wait		                do setup, load memoffset in shm
 *                        <--               signal setup done
 *         wakeup...                           waiting for buf or next Setup
 *	Start sending bufs		spin inpBufFull or Another cmd
 *	  spin inpBuf available 
 *	  mark busy..fill
 *	  mark full
 *
 *	1.Startup,setup, and exit commands are sent to the bolt via 
 *	  signals. The bolt will signal cmd complete with a signal for
 *	  setup and startup (exit doesn't send back a signal).
 *	2.Buffer i/o is multi buffered (inp and output). There are N
 *	  input bufs and M output bufs (user settable). Each buffer has a 
 *	  bufNode in shared memory that holds the buf status, words, cmd
 *	  and a pointer to the actual buffer in ap memory.
 *	  Buffers are cacheable and the get/put routines keep track of the
 *	  cache.
 *	3. Buffer passing:
 *	   input buffers are passed in by setting the bufSTatus to Full in the
 *	   bufNode.
 *	   Buffers are passed back by setting the bufFull flag in the output
 *	   node and by then sending a bufFull signal back to the ap.
 *      4. Currently the host spins on an input buf free and the ap spins
 *	   on inputBuf full. The ap spins on output buf free and the host
 *	   receives an msg (via the signal ) when an output buffer is ready.
 *	5. Commands are passed to the ap in  the sigToAp structure. 
 *	   Commands are passed back to the host in the sigToHost structure.
 *	   Included in up to 50 words of data that can be passed.
*/
/* 
 *	commands and status for communications
 *
 *	HC = HostCommand to ap
 *	BC = BoltCommand to host
 *
 *        cmd        
 *	BC_STARTUP_DONE	     numargs=0     status=ok/error
 *
 *      HC_SETUP             numargs=(depends on program)   status=ok/error=6	
 *
 *	BC_SETUP_DONE        numargs=0 	   status=ok/error   
 *
 *	(not implemented)HC_READBUF 	numargs=1 
 *                   0	  interrupt # received.
 *
 *	BC_OUTBUF_RDY        numargs=2
 *  	  	             0  node index 
 *                           1  bytes toread
 *	BC_ERROR             numargs=0    status=error #
 *
 *	HC_EXIT		     numargs=0	 
 * 	
*/
/*
 *	try and parameterize memory available.
 *	hold out 100 pages for code/system..
 *	currently setup for 1 partition max
*/
#define AP_WORDS_PAGE	1024
#define AP_MAX_PAGES    (2*1024)
#define AP_PAGES_TO_SAVE 100
#define	AP_MAXWORDS_BUF_1PROG ((AP_MAX_PAGES - AP_PAGES_TO_SAVE) * \
			(AP_WORDS_PAGE ) )
/*
 *	defines for communications commands/status
*/
#define  BS_TOO_MANY_SETUP_PARAMS -4
#define  BS_VALLOC_ERR_IOBUFS  -3
#define  BS_SHM_TOO_SMALL      -2
#define  BS_ERROR	       -1
#define  BS_OK			0
#define	 BC_STARTUP_DONE      	100
#define  HC_SETUP		101
#define  BC_SETUP_DONE		102
#define	 HC_READBUF		103
#define  BC_ERROR		104
#define	 HC_EXIT		105
#define  BC_OUTBUF_RDY		106
/*
 *	status values for input/output nodes
*/
#define NODE_FREE             1
#define NODE_FULL             2
#define NODE_BUSY_AP          3
#define NODE_BUSY_HOST        4
/*
 *   node special command values
 *  sync.. we are syncing with program, no data to xfer.
*/
#define NODE_CMD_NORMAL       0
#define NODE_CMD_SYNC	     99

/*	number dbg.times[]	*/
#define APDEBUG_TIMES	      20
/*
 *      ap shared memory usage:
 *	command structure
 *	input bufnodes   
 *	output bufnode  
 *
 *	the input and output buffers are allocated from normal ap memory
 *	space.
*/
/*	assume 1 page reserved	*/
#define SHM_WDS_RESERVED	1024	
#define	SHM_CMD_ARGS_MAX 	50
#define SHM_WDS_NEEDED(numInodes,numOnodes) ( (sizeof(APSHMCOM)/sizeof(int) + \
		6*((numInodes)+)numOnodes))
/*
 *	this is the command structure
*/
typedef struct { 
        int     cmd;             
	int	stat;			/* status returning...*/
	int	numArgs;		/* passing to other side*/
        int     arg[SHM_CMD_ARGS_MAX];
        } SHMCMD;
/*
 *	and a node structure for the data bufs 
*/
typedef	struct {
	int	stat;		/* free, full, proc, etc*/
	int	wdsThisBuf;	/* to use */
	int	bufCmd;		/* 0,99-->sync buf, not data to xfer*/
	unsigned long	*pbuf_860;	/* 860 buf address*/
	unsigned long	*pbuf_960;	/* 960 phy address*/
	char		*pbuf_vme;	/* vme address for data*/
	}  APNODE;
/*
 * shared memory partitioning
 *
 *  command struct to   ap	  3+maxArgsIn	(200)  	 	203
 *  command struct from ap	  3+maxArgsOut	(10)		 13
 *  monitoring info 		  3				 3
 *  bufInfo			  7				 7
 *  debugInfo			  1 + maxTimes	(20)		21
 *  input Nodes			  6*NumInpNodes (10)		60
 *  output Nodes		  6*NumOutNodes (10)		60
 * total (should fit in 1024 words)
*/
/*
 *	we keep a list of nodes in ap shm memory. 1 list for input bufs and
 *	1 list for output bufs. The below structure tells us how the list
 *	and the bufs are configured.
*/
typedef struct {
	   int  shmTotW;	/* total number words in shm*/
	   int  numInpNodes;	/* how many input (to ap) nodes we want */
	   int  numOutNodes;	/* how many output (from ap)  nodes we want*/
	   int  inpBufLenW;	/* number words (int)  1 input  buf*/
	   int  outBufLenW;	/* number words (int) 1 output buf*/
	   int  inpNodesOffW;	/* from start of shared memory (ints)*/
	   int  outNodesOffW;	/* from start of shared memory (ints)*/
	} APBUF_INFO;
typedef struct {
	unsigned long   skyProgState;	   /* for debugging*/
	float		times[APDEBUG_TIMES];	
	}  APDEBUG_INFO;
/*
 *      here is the communication structure
 *	note that all ptrs to addresses are 860 adresses...
*/
typedef struct {
        SHMCMD sigToAp;                     /* from host to ap*/
        SHMCMD sigToHost;                   /* from ap to host*/
	unsigned long   host_int_send_cnt;  /* interrupts sent from host*/
	unsigned long   host_int_rcv_flag;  /* for dbg. Trig of logAna*/
	unsigned long   exitFlag;	   /* not 0 --> ap exit*/
	APBUF_INFO      bufI;		   /* hold shared buffer memory info*/
	APDEBUG_INFO	dbg;
        }  APSHMCOM;            

#endif
