!======================================================================
program HotSpots
  !======================================================================
  !     Finds extrema in the 2-D field on the sphere.
  !     Uses file "param.dat" for inputs
  !
  ! Benjamin D. Wandelt October 1997
  !======================================================================
  use healpix_types      
  use maxima_tools, ONLY : find_maxima
  use fitstools, ONLY : getsize_fits, input_map, write_bintab
  use pix_tools, ONLY : convert_inplace_real, convert_inplace_int, ring2nest, nest2ring
  USE head_fits, ONLY : add_card
  use utilities, ONLY : die_alloc
  USE extension, ONLY : getArgument, nArguments
  use paramfile_io

  implicit none
  integer(i4b) :: nside,npix,nmin,nmax
  integer(i4b) :: i,face_num

  real(sp), pointer ::  map(:)
  real(sp), allocatable :: inmap(:,:)
  real(sp), allocatable :: realpeak(:,:)
  integer(i4b),pointer :: peak(:)

  integer(i4b) :: ordering,order_type
  integer(i4b) :: nlheader
  integer(i4b) :: nmaps = 1

  character(len=80), dimension(1:120) :: header
  character(len=filenamelen) :: parafile,infile,outfile_extrema, &
                                outfile_max,outfile_min
  integer(i4b) :: status

  type(paramfile_handle) :: handle

  CHARACTER(LEN=*), PARAMETER :: code    = 'HOTSPOTS'
  CHARACTER(LEN=*), PARAMETER :: version = '0.8.5'
  !--------------------------------------------------------------------------
100 FORMAT(A)

  !-----------------------------------------------------------------------
  !                    get input parameters and create arrays
  !-----------------------------------------------------------------------

  !     --- read parameters interactively if no command-line arguments
  !     --- are given, otherwise interpret first command-line argument
  !     --- as parameter file name and read parameters from it:
  if (nArguments() == 0) then
    parafile=''
  else
   if (nArguments() /= 1) then
     print '(" Usage: hotspots [parameter file name]")'
     stop 1
   end if
   call getArgument(1,parafile)
  end if

  handle=parse_init (parafile)

  !     --- announces program, default banner ***
  PRINT *,'                    '//code//' '//version
  PRINT *,' ***  Finding Maxima and Minima of a HEALPIX Map  ***'
  PRINT *,' *** (Application of the Neighbour-finding Tools) ***'
  PRINT *,''

  !     --- gets the file name for the map ---
  infile=parse_string(handle,'infile',default='map.fits', &
    descr='Input file name (Map FITS file, e.g. map_CDM.fits):', &
    filestatus='old')
  PRINT *,''

  !     --- finds out the pixel number of the map and its ordering ---
  npix = getsize_fits(infile ,ordering=ordering, nside=nside)
  
  if (nside.eq.0) then 
     print*,'Keyword NSIDE not found in FITS header!'
     stop 1
  endif
  if (nside.ne.NINT(SQRT( npix / 12.) )) then
     print*,'FITS header keyword NSIDE does not correspond'
     print*,'to the size of the map!'
     stop 1
  endif
  
  if(nside<=1) then
     PRINT*,' '//code//'> Nside needs to be at least 2 for extrema calculations.'
     stop 1
  end if

  !     --- check ordering scheme ---

if ((ordering.ne.1).and.(ordering.ne.2)) then
PRINT*,'The ordering scheme of the map must be RING or NESTED.'
PRINT*,'No ordering specification is given in the FITS-header!'
stop 1
endif

order_type = ordering

!     --- gets the output extrema map filename ---
  outfile_extrema=parse_string(handle,'extrema_outfile', &
    default='pixlminmax.fits', &
    descr='Extrema map file name (eg, pixlminmax.fits) :', &
    filestatus='new')
  PRINT *,''

!     --- gets the output filename for ascii file with MAXIMA ---
  outfile_max=parse_string(handle,'maxima_outfile', &
    default='maxima.dat', &
    descr='Maxima file name (eg, maxima.dat) :', &
    filestatus='new')
  PRINT *,''

!     --- gets the output filename for ascii file with MINIMA ---
  outfile_min=parse_string(handle,'minima_outfile', &
    default='minima.dat', &
    descr='Maxima file name (eg, minima.dat) :', &
    filestatus='new')
  PRINT *,''

  allocate(map(0:npix-1),stat=status)
  if (status/=0) call die_alloc(code,'map')

  allocate(inmap(0:npix-1,1),stat=status)
  if (status/=0) call die_alloc(code,'inmap')

  PRINT*, ' '//code//'> Reading '//TRIM(infile)//'...'
  
  call input_map(infile,inmap,npix,nmaps)

  map(0:npix-1) = inmap(0:npix-1,1)
  deallocate(inmap)

  if(order_type==1) then
     write(*,*) code//'> Converting to NESTED numbering scheme.'       
     call convert_inplace_real(ring2nest,map)
  endif

  allocate(peak(0:npix-1))
  if (status/=0) call die_alloc(code,'peak')
  peak=0

  write(*,*) code//'> Starting extrema search in pixel space.'

  ! This is fully parallelisable. All 12 faces can be searched synchronously
  !*$*assert concurrent call
  ! peak = +1 for maximum, -1 for minimum, 0 otherwise
  do face_num=0,11
     call find_maxima(map,face_num,peak)
  end do

  write(*,*) 'Done.'

  if(order_type==1) then
     write(*,*) code//'> Converting map back to RING numbering.'
     call convert_inplace_real(nest2ring,map)
     write(*,*) code//'> Converting peaks to RING numbering.'
     call convert_inplace_int(nest2ring,peak)
  endif

  write (*,*) code//' Writing '//trim(outfile_extrema)//' FITS file...'
  header = ''
  nlheader = SIZE(header)
  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'COMMENT','     Sky Map Pixelisation Specific Keywords    ')
  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'PIXTYPE','HEALPIX','HEALPIX Pixelisation')
  call add_card(header,'ORDERING','RING',  'Pixel ordering scheme, either RING or NESTED')
  call add_card(header,'NSIDE'   ,nside,   'Resolution parameter for HEALPIX')
  call add_card(header,'FIRSTPIX',0,'First pixel # (0 based)')
  call add_card(header,'LASTPIX',npix-1,'Last pixel # (0 based)')
  call add_card(header) ! blank line

  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'COMMENT','     Planck Analysis Specific Keywords      ')
  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'EXTNAME','ANALYSIS')
  call add_card(header,'CREATOR',code,        'Software creating the FITS file')
  call add_card(header,'VERSION',version,     'Version of the simulation software')
  call add_card(header) ! blank line
  call add_card(header,'COMMENT', 'Input Map ='//TRIM(infile))
  call add_card(header) ! blank line
  call add_card(header,'TTYPE1', 'PEAKS','Maxima and Minima')
  call add_card(header,'COMMENT', 'Local Min and Max keep their original value, the rest is set to 0')
  call add_card(header) ! blank line
  ! ------
  ! outputs a FITS map set to 0 but for extrema that keep their original value
  allocate(realpeak(0:npix-1,1))
  realpeak=0
  where(peak/=0) realpeak(:,1)=map

  call write_bintab(realpeak, npix, 1, header, nlheader, outfile_extrema)

  deallocate(realpeak)

  ! outputs the set of maxima (pixel number and value)
  ! and the set of minima (pixel number and value) in 2 ascii files
  open(14,file=outfile_max,status="unknown")
  open(15,file=outfile_min,status="unknown")
  write(*,*) 'Output: Writing '//trim(outfile_max)//' and '//trim(outfile_min)//'.'

  nmax=count(peak>0)
  write(*,*) "Found ",nmax," Maxima"
  nmin=count(peak<0)
  write(*,*) "Found ",nmin," Minima"

  do i=0,npix-1
     if(peak(i)>0) then     !Maximum
        write(14,'(i10,f12.6)') i,map(i)
        elseif(peak(i)<0) then !Minimum
        write(15,'(i10,f12.6)') i,map(i)
     endif
  enddo

  close(14)
  close(15)

  deallocate(peak)
  deallocate(map)
  stop
end program HotSpots
