SUBROUTINE PROJ_GNOMIC(&
     map,      &
     nside,    &
     ordering, &
     xsize,    &
     lon0,     &
     lat0,     &
     resgrid,  &
     image,    &
     mask      &
     )

   ! --------------------------------------------------------------
   ! This subroutine takes an input HEALPIX sky map array and
   ! generates the gnomic (tangential) projection on a rectangular grid.
   !
   ! INPUTS:
   ! 	   map     = array containing pixel temperatures
   !       nside   = defines number of pixels on the sky
   !       ordering= whether input map is in RING or NESTED format
   !       xsize   = x-dimension of output image array
   !       lon0    = longitude (in degrees) of the center of the plot
   !       lat0    = latitude (in degrees) of the center of the plot
   !       gridresn= resolution of the grid elements in ARCMINUTES
   !       image   = output gnomic projection image
   !       mask    = logical array defining usable pixels in output
   !                 image array
   !
   ! PROCEDURES USED:
   !     	ang_pix, ang_nest_pix, euler_matrix
   !
   !  RELATED LITERATURE:
   !  see the web site http://www.tac.dk/~healpix
   !
   ! MODIFICATION HISTORY:
   ! 	October 1997, Eric Hivon, TAC (original IDL code)
   !    January 1999, A.J. Banday MPA (conversion to F90)
   !    March   1999, EH Caltech (consistent with new IDL version, 
   !                       fixed a bug on rotation)
   !
   ! --------------------------------------------------------------

   USE healpix_types
   USE pix_tools

   IMPLICIT NONE

   INTEGER(I4B) :: I,J
   INTEGER(I4B) :: id_pix
   INTEGER(I4B) :: ordering
   INTEGER(I4B) :: nside

   INTEGER(I4B) :: xsize

   REAL(SP), DIMENSION(0:12*nside**2-1) :: map

   REAL(SP),     DIMENSION(0:xsize-1,0:xsize-1) :: image
   LOGICAL(LGT), DIMENSION(0:xsize-1,0:xsize-1) :: mask

   REAL(SP) :: lon0, lat0, resgrid
   REAL(DP) :: lon0rad, lat0rad, resgrid_rad
   REAL(DP) :: DtoR = PI/180.0_dp

   REAL(DP), DIMENSION(:,:), ALLOCATABLE :: xg, yg, zg, rg
   REAL(DP), DIMENSION(3,3)              :: mm
   REAL(DP), DIMENSION(3)                :: nvs

   INTEGER(I4B) :: status

   ! --------------------------------------------------------------

   !-- some basic variable manipulations --
   resgrid_rad = (real(resgrid,dp)/60.0_dp) * DtoR !/ real(xsize-1,dp)
   lon0rad = real(lon0,dp) * DtoR
   lat0rad = real(lat0,dp) * DtoR
!   theta0  = halfpi - lat0rad
!   phi0    = lon0rad

   !-- generates the projection around the chosen contact point
   !   on the planar grid  (1,Y,Z) --
   image(:,:) = 1.0
   mask(:,:)  = .true.

   !--- allocate memory for arrays ---
   ALLOCATE(xg(0:xsize-1,0:xsize-1),stat = status)
   if (status .NE. 0) then
      print *,' cannot allocate space for array'
      stop
   end if
   xg = 1.0_dp
   ALLOCATE(yg(0:xsize-1,0:xsize-1),stat = status)
   if (status .NE. 0) then
      print *,' cannot allocate space for array'
      stop
   end if
   do i = 0,xsize-1
      yg(i,:) = -(real(i,dp) - 0.5_dp*real(xsize,dp) + 0.5_dp) *&
           & resgrid_rad  ! minus sign = astro convention
   end do
   ALLOCATE(zg(0:xsize-1,0:xsize-1),stat = status)
   if (status .NE. 0) then
      print *,' cannot allocate space for array'
      stop
   end if
   do i = 0,xsize-1
      zg(:,i) = (real(i,dp) - 0.5_dp*real(xsize,dp) + 0.5_dp) *&
           & resgrid_rad
   end do
   ALLOCATE(rg(0:xsize-1,0:xsize-1),stat = status)
   if (status .NE. 0) then
      print *,' cannot allocate space for array'
      stop
   end if
   rg = sqrt(1.0_dp + yg**2 + zg**2)

   !-- the contact point is rotated by theta0 around y (=colatitude) 
   !   and phi0 around z (=azimut) --
!   call euler_matrix(0.0_dp,theta0,phi0,mm) ! /y option used in IDL here
   call euler_matrix(lon0rad,lat0rad,0.0_dp,mm) ! /z option used in IDL here

   xg = 1.0_dp / rg
   yg = yg / rg
   zg = zg / rg


   !--- converts the position on the sphere into pixel number and project the 
   !    corresponding data value on the map ---
   DO I = 0,xsize-1
      DO J = 0,xsize-1
         !-- rotated positions --
         nvs = matmul(mm,(/ xg(i,j), yg(i,j), zg(i,j) /))
         if (ordering .eq. 2) then
            call vec2pix_nest( nside, nvs, id_pix ) 
            ! id_pix in [0,12*nside^2-1]
         else ! default
            call vec2pix_ring( nside, nvs, id_pix ) 
            ! id_pix in [0,12*nside^2-1]
         end if
         image(I,J) = map(id_pix)
      END DO
   END DO
   
!   print*,mm
!   print*,matmul(mm,(/ 1, 0, 0 /))
!   print*,matmul(mm,(/ 0, 1, 0 /))
!   print*,matmul(mm,(/ 0, 0, 1 /))

!   call euler_matrix(1.0_dp,0.0_dp,0.0_dp,mm) ! /z option used in IDL here
!   print*,mm
!   print*,matmul(mm,(/ 1, 0, 0 /))
!   print*,matmul(mm,(/ 0, 1, 0 /))
!   print*,matmul(mm,(/ 0, 0, 1 /))


   !--- deallocate memory for arrays ---
   deallocate( xg )
   deallocate( yg )
   deallocate( zg )
   
END SUBROUTINE PROJ_GNOMIC
