PROGRAM MAP2GIF
   !-------------------------------------------------------------------
   ! This program is a simplified kluge of mollview.pro, a basic tool to 
   ! view a Mollweide projection of maps made in the Healpix pixelisation
   ! scheme. A map in FITS format is read and converted to a GIF image.
   !
   ! PURPOSE:
   ! 	basic tool to view a Mollweide projection of maps made 
   !	in Healpix pixelisation
   !
   ! CALLING SEQUENCE:
   ! 	MAP2GIF [-inp input_file] [-out output_file]
   !            [-col color_table] [-sig signal]
   !            [-xsz xsize] [-log logflg] [-hlp]
   !                        
   ! INPUTS:
   ! 	input_file = string containing the name of a FITS file of 
   !                 the healpix map in an extension or in the image field
   !
   !	output_file= string containing the name of the output GIF file(s)
   !
   ! 	color_table= color table number, in [1,5]
   !        	     if unset the color table will be 5 (Blue-Red)
   !
   !    signal     = integer defining whether I,Q,U map should be plotted
   !                 1 = I; 2 = Q; 3 = U
   !
   ! 	xsize      = x-dimension of image array to be stored in GIF file
   !                 if unset the x-size will be ysize = xsize/2
   !
   !    logflg     = display the log of map
   !                 if unset logflg = .false. (ie. use actual temperatures)
   !
   ! PROCEDURES USED:
   !	parseOptions, getsize_fits, input_map, proj_mollw, proj_gnomic
   !
   ! EXTERNAL LIBRARY:
   !    this code uses the CFITSIO library that can be found at
   !    http://heasarc.gsfc.nasa.gov/docs/software/fitsio/
   !
   ! RELATED LITERATURE:
   !    see the web site http://www.tac.dk/~healpix
   !
   ! MODIFICATION HISTORY:
   ! 	October 1997,  Eric Hivon,      TAC (original IDL code)
   !    December 1998, A.J. Banday,     MPA (simple conversion to F90)
   !    December 1998, M.S. Bartelmann, MPA (cleanup, color tables, ...)
   !    January 1999,  A.J. Banday,     MPA (added gnomic routine)
   !
   !-------------------------------------------------------------------

   USE healpix_types
   USE fitstools
   USE options
   USE gifmod

   IMPLICIT NONE

   INTEGER(I4B) :: nsmax
   INTEGER(I4B) :: npixtot, nmaps
   INTEGER(I4B) :: ordering
   INTEGER(I4B) :: i

   REAL(SP),     DIMENSION(:,:),   ALLOCATABLE :: map_IQU
   REAL(SP),     DIMENSION(:),     ALLOCATABLE :: map
   REAL(SP),     DIMENSION(:,:),   ALLOCATABLE :: image

   INTEGER(I4B), ALLOCATABLE, DIMENSION(:,:) :: imgint

   INTEGER(I4B), POINTER, DIMENSION(:,:) :: imgbar
   INTEGER(I4B), POINTER, DIMENSION(:,:) :: imgttl

   LOGICAL(LGT), DIMENSION(:,:),   ALLOCATABLE :: mask
   LOGICAL(LGT), DIMENSION(:),     ALLOCATABLE :: observed

   REAL(SP)     :: mindata, maxdata, Tmin, Tmax

   INTEGER(I4B) :: status

   CHARACTER(LEN=7), PARAMETER :: code = 'MAP2GIF'
   CHARACTER(LEN=5), PARAMETER :: version = '1.0.0'

   !-------------------------------------------------------------------

   !--- acquire input parameters ---
   call parseOptions()

   !--- read in the FITS file header, find out the pixel number of the map 
   !    and its ordering ---
   npixtot = getsize_fits(input_file, nmaps = nmaps, ordering=ordering)
   nsmax = NINT(SQRT( npixtot / 12.) )

   !--- are requested signal and map compatible? ---
   if (nmaps == 1 .and. signal > 1) THEN
      print *,' requested signal incompatible with map'
      stop
   end if

   !--- allocate space for arrays ---
   ALLOCATE(map_IQU(0:12*nsmax**2-1,1:nmaps),stat = status)
   if (status /= 0)THEN
      print *,' cannot allocate space for map_IQU array'
      stop
   end if
   ALLOCATE(map(0:12*nsmax**2-1),stat = status)
   if (status /= 0)THEN
      print *,' cannot allocate space for map array'
      stop
   end if
   ALLOCATE(observed(0:size(map)-1),stat = status)
   if (status /= 0)THEN
      print *,' cannot allocate space for observed array'
      stop
   end if

   !--- read in the FITS file ---
   call input_map(input_file, map_IQU, npixtot, nmaps, fmissval=-1.6375e30_sp)

   !--- select correct signal to plot ---
   map(:) = map_IQU(:,signal)            

   !--- Test for unobserved pixels ---
   observed = map > -1.6375e30
   where (observed) map = (map+offset)*factor
   if (ashflag) then
      do i=0,size(map)-1
        if (observed(i)) then
          if (map(i)>=0) then
            map(i) = log(map(i)+sqrt(map(i)**2+1))
          else
            map(i) = -log(-map(i)+sqrt(map(i)**2+1))
          endif
        endif
      end do      
   endif
   IF( logflg )THEN
      observed = map > 0.
      where(observed)
        map=log10(map)
      elsewhere
        map=-1.6375e30
      endwhere
   END IF
   mindata = minval(map, mask = observed)
   maxdata = maxval(map, mask = observed)

   if (mindata==maxdata) then
     print *,"Warning: map contains completely uniform values!"
     maxdata=mindata+1e-5
   endif

   !--- sets MIN and MAX using linear scaling ---
   if (ABS((maxdata+mindata)/(maxdata-mindata)) < 5.e-2) then
      !--- center of color scale is 0
      Tmax = MAXVAL(ABS((/mindata,maxdata/)))
      Tmin = -Tmax
   else
      Tmax = maxdata 
      Tmin = mindata
   end if

   !--- allow for user min/max settings ---
   if (usermax <  1.1e30) Tmax = usermax
   if (usermin > -1.1e30) Tmin = usermin

   !--- adjust 'offscale' temperatures ---
   WHERE(observed) map = max(min(map,Tmax),Tmin)

   !--- create the projection ---
   select case (projection)
   case ('MOL', 'mol')
      !--- allocate space for image and mask ---
      ALLOCATE(image(0:xsize-1,0:xsize/2-1),stat = status)
      if (status /= 0)THEN
         print *,' cannot allocate space for image array'
         stop
      end if
      ALLOCATE(mask(0:xsize-1,0:xsize/2-1),stat = status)
      if (status /= 0)THEN
         print *,' cannot allocate space for mask array'
         stop
      end if

      !--- create the mollweide projection ---
      call proj_mollw(&
           map,      &
           nsmax,    &
           ordering, &
           xsize,    &
           lon0,     &
           image,    &                
           mask      &                
           )
   case ('GNO', 'gno')
      !--- allocate space for image and mask ---
      ALLOCATE(image(0:xsize-1,0:xsize-1),stat = status)
      if (status /= 0)THEN
         print *,' cannot allocate space for image array'
         stop
      end if
      ALLOCATE(mask(0:xsize-1,0:xsize-1),stat = status)
      if (status /= 0)THEN
         print *,' cannot allocate space for mask array'
         stop
      end if

      !--- create the gnomic projection ---
      call proj_gnomic(&
           map,        &
           nsmax,      &
           ordering,   &
           xsize,      &
           lon0,       &
           lat0,       &
           gridresn,   &
           image,      &                  
           mask        &                
           )
   case default
      print '(" unknown projection type: ",a3)', projection
      stop 1
   end select 

   !--- load the color table ---
   call setcol( color_table )

   !--- scale image ---
   allocate(imgint(size(image,1),size(image,2)))
   call imgscl(&
        image,  &
        imgint, &
        mask    &
        )
   deallocate(image)

   !--- add color bar ---
   if (bar) then
      call addbar(&
           imgint, &
           imgbar  &
           )
      deallocate(imgint)
      allocate(imgint(size(imgbar,1),size(imgbar,2)))
      imgint = imgbar
      deallocate(imgbar)
   end if

   !--- add title ---
   if (title /= '') then
      call addttl(&
           imgint, &
           imgttl, &
           title   &
           )
      deallocate(imgint)
      allocate(imgint(size(imgttl,1),size(imgttl,2)))
      imgint = imgttl
      deallocate(imgttl)
   end if

   !--- create the GIF image ---
   call gifmap(&
        imgint,           &
        trim(output_file) &
        )
   
   !--- deallocate memory for arrays ---
   DEALLOCATE(map_IQU,stat = status)
   DEALLOCATE(map,stat = status)
   DEALLOCATE(image,stat = status)
   DEALLOCATE(mask,stat = status)

   !--- Exit, stage left ...
   
END PROGRAM MAP2GIF
