SUBROUTINE PROJ_MOLLW(&
     map,      &
     nside,    &
     ordering, &
     xsize,    &
     lon0,     &
     image,    &
     mask      &
     )
   ! --------------------------------------------------------------
   ! This subroutine takes an input HEALPIX sky map array and
   ! generates the (u,v) position on a mollweide projection.
   !
   ! INPUTS:
   ! 	   map     = array containing pixel temperatures
   !       nside   = defines number of pixels on the sky
   !       ordering= whether input map is in RING or NESTED format
   !       xsize   = x-dimension of output image array
   !       lon0    = longitude (in degrees) of the center of the plot
   !       image   = output mollweide projection image
   !       mask    = logical array defining usable pixels in output image array
   !
   ! PROCEDURES USED:
   !     	ang_pix, ang_nest_pix
   !
   !  RELATED LITERATURE:
   !  see the web site http://www.tac.dk/~healpix
   !
   ! MODIFICATION HISTORY:
   ! 	October 1997, Eric Hivon, TAC (original IDL code)
   !       December 1998, A.J. Banday MPA (conversion to F90)
   !
   ! --------------------------------------------------------------

   USE healpix_types
   USE pix_tools
   
   IMPLICIT NONE

   INTEGER(I4B) :: I,J
   INTEGER(I4B) :: id_pix
   INTEGER(I4B) :: ordering
   INTEGER(I4B) :: nside

   INTEGER(I4B) :: xsize, ysize

   REAL(SP), DIMENSION(0:12*nside**2-1) :: map

   REAL(SP),     DIMENSION(0:xsize-1,0:xsize/2-1) :: image
   LOGICAL(LGT), DIMENSION(0:xsize-1,0:xsize/2-1) :: mask

   REAL(SP) :: lon0
   REAL(DP) :: lon0rad
   REAL(DP) :: DtoR = PI/180.0_dp

   INTEGER(I4B) :: xc, dx, yc, dy

   REAL(DP), DIMENSION(:,:), ALLOCATABLE :: u, v
   REAL(DP), DIMENSION(:,:), ALLOCATABLE :: out_lat, out_lon

   INTEGER(I4B) :: status

   ! --------------------------------------------------------------

   ysize = xsize/2

   !--- generates the (u,v) position on the mollweide map ---
   image(:,:) = 1.                         ! white background ???

   xc = (xsize-1)/2
   dx = xc
   yc = (ysize-1)/2
   dy = yc
   
   !--- allocate memory for arrays ---
   ALLOCATE(u(0:xsize-1,0:ysize-1),stat = status)
   if (status .NE. 0) then
      print *,' cannot allocate space for array'
      stop
   end if
   do i = 0,xsize-1
      u(i,:) = real(i,dp)
   end do
   ALLOCATE(v(0:xsize-1,0:ysize-1),stat = status)
   if (status .NE. 0) then
      print *,' cannot allocate space for array'
      stop
   end if
   do i = 0,ysize-1
      v(:,i) = real(i,dp)
   end do
   ALLOCATE(out_lat(0:xsize-1,0:ysize-1),stat = status)
   if (status .NE. 0) then
      print *,' cannot allocate space for array'
      stop
   end if
   ALLOCATE(out_lon(0:xsize-1,0:ysize-1),stat = status)
   if (status .NE. 0) then
      print *,' cannot allocate space for array'
      stop
   end if
   
   u =  2._dp*(u - real(xc,dp))/(real(dx,dp)/1.02_dp)   ! 1.02 : fudge factor in [-2,2]*1.02
   v =    (v - real(yc,dp))/(real(dy,dp)/1.02_dp)   ! in [-1,1] * 1.02
   
   !--- for each point on the mollweide map looks for the corresponding 
   !    position (lon, lat) on the sphere ---
   mask = (u**2/4._dp + v**2) <= 1._dp
   out_lat = 1.e12_dp ! values for points out of the sphere
   out_lon = 1.e12_dp
   lon0rad = real(lon0,dp) * DtoR
   WHERE( mask )
      out_lat = PI/2._dp - (ASIN ( 2._dp/PI * (&
           & ASIN(v) + v*SQRT( (1._dp-v)*(1._dp+v) )) )) 
      ! colat in [0,pi]
      out_lon = -lon0rad - PI/2._dp * u/MAX(SQRT( (1._dp-v)*(1._dp+v) ),1.e-6_dp)
      ! lon in [-pi,pi], the minus sign for astro convention
   END WHERE
   
   WHERE(out_lon < 0._dp)
      out_lon = out_lon  + 2._dp*PI
      ! lon in RAD in [0,2pi]
   END WHERE
   
   !--- converts the position on the sphere into pixel number and project the 
   !    corresponding data value on the map ---
   DO I = 0,xsize-1
      DO J = 0,ysize-1
         IF(ABS(out_lat(I,J)) .LE. 1.e5_dp)THEN ! keep only meaningful points
            if (ordering .eq. 1) then
               call ang2pix_ring( nside, out_lat(I,J), out_lon(I,J), id_pix ) 
               ! id_pix in [0,12*nside^2-1]
            else
               call ang2pix_nest( nside, out_lat(I,J), out_lon(I,J), id_pix ) 
               ! id_pix in [0,12*nside^2-1]
            end if
            image(I,J) = map(id_pix)
         END IF
      END DO
   END DO
   
   !--- deallocate memory for arrays ---
   deallocate( u )
   deallocate( v )
   deallocate( out_lat )
   deallocate( out_lon )
   
END SUBROUTINE PROJ_MOLLW
