module alm_tools
  !   Parallel (Open MP) implementation
  !
  !   subroutine alm2map_sc       ! modified for Parallelisation
  !   subroutine alm2map_sc_pre   ! modified
  !   subroutine init_trig        ! Parallel specific
  !   subroutine ring_synthesis
  !   subroutine ring_synthesis2  ! Parallel specific
  !   subroutine alm2map_pol      ! modified
  !   subroutine alm2map_pol_pre1 ! modified
  !   subroutine alm2map_pol_pre2 ! modified
  !   subroutine map2alm_sc       ! modified
  !   subroutine map2alm_sc_pre   ! modified
  !   subroutine ring_analysis
  !   subroutine ring_analysis2   ! Parallel specific
  !   subroutine map2alm_pol      ! modified
  !   subroutine map2alm_pol_pre1 ! modified
  !   subroutine map2alm_pol_pre2 ! modified
  !
  !   subroutine pow2alm_units
  !   subroutine generate_beam
  !   subroutine gaussbeam
  !   subroutine pixel_window
  !   subroutine alter_alm
  !   subroutine create_alm
  !
  use utilities, only: die_alloc
  use healpix_types
  use healpix_fft, only : real_fft

  IMPLICIT none

  ! keep everything private unless stated otherwise
  private
  ! define large and small numbers used to renormalise the recursion on the Legendre Polynomials
  real(KIND=DP), private, PARAMETER :: FL_LARGE = 1.0e30_dp
  real(KIND=DP), private, PARAMETER :: FL_SMALL = 1.0e-30_dp
  ! make (front end) routines public
  public :: alm2map, map2alm, &
       &    pow2alm_units, generate_beam, gaussbeam, alter_alm, &
       &    create_alm, ring_synthesis, ring_analysis

  interface alm2map
     ! Call the correct alm2map routine, depending on whether polarisation is included
     ! (determined from the rank of the map_TQU array) is included or precomputed plms
     ! (scalar or tensor, determined from the rank of the plm array) is included.
     module procedure alm2map_sc,alm2map_sc_pre,alm2map_pol_pre1,alm2map_pol,alm2map_pol_pre2
  end interface

  interface map2alm
     ! Call the correct map2alm routine, depending on whether polarisation is included
     ! (determined from the rank of the map_TQU array) is included or precomputed plms
     ! (scalar or tensor, determined from the rank of the plm array) is included.
     module procedure map2alm_sc,map2alm_sc_pre,map2alm_pol_pre1,map2alm_pol,map2alm_pol_pre2
  end interface

  !
  ! Aug 14, 2000, EH, Caltech, switch to CMBFAST normalisation of polarisation power spectra
  !  (see Zaldarriaga, astro-ph/9709271; ApJ, 503, 1 (1998))
  !   the factor normal_l (used for synthesis and analysis) is reduced by sqrt(2)
  !
  ! Oct 17, 2001
  !   polarised alm expression is larger by a factor 2
  !   introduced FL_LARGE and FL_SMALL parameters
  !
  ! Sept 2001, EH, Caltech: map2alm* : skip calculations for cut out pixels
  !                                  : use explicit loops for multiplication of alm by omega_pix
  !
  ! Nov 2001, EH, Caltech : added KIND=DP in all CMPLX commands that missed it
  !                         turned 1.d0 in 1.0_dp and so on
  !                         moved 'use healpix_types' and 'implicit none' to module top level
  !
  ! Dec 2001, EH, Caltech : corrected error on inner loop upper bound of lam_fact initialisation
  !   in alm2map_pol_pre1, alm2map_pol, map2alm_pol_pre1, map2alm_pol
  !   pointed out by M. Ashdown to Level S
  !   
  !   added pow2alm_units, gaussbeam, generate_beam and alter_alm (moved for smoothing/alter_alm.f90)
  ! Nov 2002, SC, IAP : parallelization with OpenMP [as part as Planck-HFI L2 task (GPH428)]. 
  !                     The parallel version is more costly in memory, so the serial version 
  !                     should be kept separately.
  !
  !  TODO : check sign of lambda_w
  !
  ! Nov 2002, EH, reordered routines to simplify maintenance of scalar vs parallel routines
  !
contains

  !=======================================================================
  subroutine alm2map_sc(nsmax, nlmax, nmmax, alm, map)
    !=======================================================================
    !     computes a map form its alm for the HEALPIX pixelisation
    !      for the Temperature field
    !     map(theta,phi) = sum_l_m a_lm Y_lm(theta,phi)
    !                    = sum_m {e^(i*m*phi) sum_l a_lm*lambda_lm(theta)}
    !
    !     where Y_lm(theta,phi) = lambda(theta) * e^(i*m*phi)
    !
    !     * the recurrence of Ylm is the standard one (cf Num Rec)
    !     * the sum over m is done by FFT
    !
    !              -------------------------------
    !          precomputes the Lambda_lm recurrence factor 
    !      and is therefore ~50% faster than previous versions
    !              -------------------------------
    !
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    COMPLEX(SPC), INTENT(IN),  DIMENSION(1:1,0:nlmax,0:nmmax) :: alm
    REAL(SP),     INTENT(OUT), DIMENSION(0:12*nsmax**2-1) :: map

    INTEGER(I4B) :: l, m, ith, scalem, scalel          ! alm related
    INTEGER(I4B) :: nph, kphi0                                             ! map related
    INTEGER(I4B), ALLOCATABLE, DIMENSION(:) :: istart_south, istart_north  ! map related

    REAL(DP) :: cth, sth, dth1, dth2, dst1, OVFLOW, UNFLOW, logOVFLOW
    REAL(DP) :: a_rec, lam_mm, lam_lm, lam_0, lam_1, lam_2, par_lm
    REAL(DP) :: f2m, fm2, fl2, corfac
    COMPLEX(DPC) :: b_n, b_s, factor

    CHARACTER(LEN=*), PARAMETER :: code = 'ALM2MAP'
    COMPLEX(DPC), DIMENSION(0:nmmax) :: b_north,b_south
    INTEGER(I4B) :: status

    REAL(DP), DIMENSION(:,:), ALLOCATABLE :: recfac
    REAL(SP), DIMENSION(0:4*nsmax-1) :: ring
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig   
    !=======================================================================

    !     --- allocates space for arrays ---
    ALLOCATE(recfac(0:nlmax,0:nmmax),stat = status)    
    if (status /= 0) call die_alloc(code,'recfac')




    ALLOCATE(istart_north(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_north')

    ALLOCATE(istart_south(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_south')


    call init_trig(nsmax,nmmax,trig)
    !     ------------ initiate arrays ----------------

!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,fm2,l,fl2)
    do m = 0, nmmax
       fm2 = DBLE(m) **2
       do l = m, nlmax
          fl2 = DBLE(l+1) **2
          recfac(l,m) = DSQRT( (4.0_dp * fl2 - 1.0_dp) / (fl2-fm2) )
       enddo
    enddo
!OMP END PARALLEL DO
    !     --------------------------------------------

    !     ----- set the whole map to zero ------
    map = 0.0
    !     --------------------------------------


    istart_north(0)=0
    istart_south(0)=12*nsmax**2
    do ith=1,2*nsmax
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
       else                   ! tropical band (north) + equator
          nph = 4*nsmax
       endif
       istart_north(ith)=istart_north(ith-1)+nph
       istart_south(ith)=istart_south(ith-1)-nph
    enddo

    OVFLOW=exp(log(FL_LARGE))
    UNFLOW=exp(log(FL_SMALL))
    logOVFLOW=log(FL_LARGE)

    dth1 = 1.0_dp / (3.0_dp*DBLE(nsmax)**2)
    dth2 = 2.0_dp / (3.0_dp*DBLE(nsmax))
    dst1 = 1.0_dp / (SQRT(6.0_dp) * DBLE(nsmax) )

!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(ith,cth,nph,kphi0,sth,lam_mm,scalem,m,f2m,par_lm,corfac,b_n,b_s, &
!$OMP   lam_0,lam_1,scalel,a_rec,lam_2,lam_lm,factor,l,b_north,b_south,ring)
    do ith = 1,2*nsmax      ! 0 <= cos theta < 1
       !        cos(theta) in the pixelisation scheme
       if (ith.lt.nsmax) then  ! polar cap (north)
          cth = 1.0_dp  - DBLE(ith)**2 * dth1
          nph = 4*ith
          kphi0 = 1
          sth = SIN( 2.0_dp * ASIN( ith * dst1 ) ) ! sin(theta)
       else                   ! tropical band (north) + equator
          cth = DBLE(2*nsmax-ith) * dth2
          nph = 4*nsmax
          kphi0 = MOD(ith+1-nsmax,2)
          sth = DSQRT((1.0_dp-cth)*(1.0_dp+cth)) ! sin(theta)
       endif
       !        -----------------------------------------------------
       !        for each theta, and each m, computes
       !        b(m,theta) = sum_over_l>m (lambda_l_m(theta) * a_l_m) 
       !        ------------------------------------------------------
       !        lambda_mm tends to go down when m increases (risk of underflow)
       !        lambda_lm tends to go up   when l increases (risk of overflow)

       lam_mm = sq4pi_inv ! lambda_00
       scalem=1
       do m = 0, nmmax
          f2m = 2.0_dp * m

          !           ---------- l = m ----------
          par_lm = 1.0_dp  ! = (-1)^(l+m)
          if (m .ge. 1) then ! lambda_0_0 for m>0
             lam_mm = -lam_mm*sth*dsqrt((f2m+1.0_dp)/f2m)
          endif
          if (abs(lam_mm).lt.UNFLOW) then
             lam_mm=lam_mm*OVFLOW
             scalem=scalem-1
          endif
          corfac=exp(scalem*logOVFLOW) !Actual lam_mm
          lam_lm = lam_mm*corfac/OVFLOW
          b_n = lam_lm * alm(1,m,m)
          b_s = lam_lm * alm(1,m,m)

          !           ---------- l > m ----------
          lam_0 = 0.0_dp
          lam_1 = 1.0_dp 
          scalel=0
          a_rec = recfac(m,m)
          lam_2 = cth * lam_1 * a_rec
          do l = m+1, nlmax
             par_lm = - par_lm  ! = (-1)^(l+m)

             lam_lm = lam_2*corfac*lam_mm/OVFLOW ! Remove OVFLOW-factors 
             factor = lam_lm * alm(1,l,m)
             b_n = b_n +          factor
             b_s = b_s + par_lm * factor

             lam_0 = lam_1 / a_rec
             lam_1 = lam_2
             a_rec = recfac(l,m)
             lam_2 = (cth * lam_1 - lam_0) * a_rec

             if (abs(lam_2) .gt. OVFLOW) then
                lam_0=lam_0/OVFLOW
                lam_1=lam_1/OVFLOW
                lam_2 = (cth * lam_1 - lam_0) * a_rec
                scalel=scalel+1
                corfac=exp((scalel+scalem)*logOVFLOW)
             elseif (abs(lam_2) .lt. UNFLOW) then
                lam_0=lam_0*OVFLOW
                lam_1=lam_1*OVFLOW
                lam_2 = (cth * lam_1 - lam_0) * a_rec 
                scalel=scalel-1
                corfac=exp((scalel+scalem)*logOVFLOW)
             endif
          enddo

          b_north(m) = b_n
          b_south(m) = b_s

       enddo


       !        ---------------------------------------------------------------

       !        obtains the negative m of b(m,theta) (= complex conjugate)


       !        ---------------------------------------------------------------
       !        sum_m  b(m,theta)*exp(i*m*phi)   -> f(phi,theta)
       !        ---------------------------------------------------------------
       call ring_synthesis2(nsmax,nlmax,nmmax,b_north,nph,ring,kphi0,trig)   ! north hemisph. + equator
       map(istart_north(ith-1):istart_north(ith-1)+nph-1) = ring(0:nph-1)


       if (ith .lt. 2*nsmax) then
          call ring_synthesis2(nsmax,nlmax,nmmax,b_south,nph,ring,kphi0,trig) ! south hemisph. w/o equat
          map(istart_south(ith):istart_south(ith)+nph-1) = ring(0:nph-1)
       endif

    enddo    ! loop on cos(theta)
!$OMP END PARALLEL DO
    !     --------------------
    !     free memory and exit
    !     --------------------
    DEALLOCATE(recfac)
    DEALLOCATE(istart_north,istart_south)
    DEALLOCATE(trig)
    return
  end subroutine alm2map_sc

  !=======================================================================
  subroutine alm2map_sc_pre(nsmax, nlmax, nmmax, alm, map, plm)
    !=======================================================================
    !     computes a map form its alm for the HEALPIX pixelisation
    !      for the Temperature field
    !     map(theta,phi) = sum_l_m a_lm Y_lm(theta,phi)
    !                    = sum_m {e^(i*m*phi) sum_l a_lm*lambda_lm(theta)}
    !
    !     where Y_lm(theta,phi) = lambda(theta) * e^(i*m*phi)
    !
    !     * Uses precomputed P_lms
    !     * the sum over m is done by FFT
    !
    !   
    !
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    COMPLEX(SPC), INTENT(IN),  DIMENSION(1:1,0:nlmax,0:nmmax) :: alm
    REAL(SP),     INTENT(OUT), DIMENSION(0:12*nsmax**2-1) :: map
    REAL(DP), INTENT(IN), DIMENSION(0:(nlmax+1)*(nlmax+2)*nsmax-1) :: plm
    INTEGER(I4B) :: l, m, ith        ! alm related
    INTEGER(I4B) :: nph, kphi0                                             ! map related
    INTEGER(I4B), ALLOCATABLE, DIMENSION(:) :: istart_south, istart_north  ! map related

    REAL(DP) :: par_lm
    COMPLEX(DPC) :: b_n, b_s, factor

    CHARACTER(LEN=*), PARAMETER :: code = 'ALM2MAP'
    COMPLEX(DPC), DIMENSION(0:nmmax) :: b_north,b_south
    INTEGER(I4B) :: status, lm, lmstep
    REAL(SP), DIMENSION(0:4*nsmax-1) :: ring
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig   
    !=======================================================================

    !     --- allocates space for arrays ---



    ALLOCATE(istart_north(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_north')

    ALLOCATE(istart_south(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_south')


    call init_trig(nsmax,nmmax,trig)

    !     --------------------------------------------

    !     ----- set the whole map to zero ------
    map = 0.0


    !     --------------------------------------


    istart_north(0)=0
    istart_south(0)=12*nsmax**2
    do ith=1,2*nsmax
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
       else                   ! tropical band (north) + equator
          nph = 4*nsmax
       endif
       istart_north(ith)=istart_north(ith-1)+nph
       istart_south(ith)=istart_south(ith-1)-nph
    enddo

    lmstep=(nlmax+1)*(nlmax+2)/2

!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(ith,nph,kphi0,m,par_lm,b_n,b_s,lm,l,factor,b_north,b_south,ring)
    do ith = 1,2*nsmax      ! 0 <= cos theta < 1
       !        cos(theta) in the pixelisation scheme
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
          kphi0 = 1
       else                   ! tropical band (north) + equator
          nph = 4*nsmax
          kphi0 = MOD(ith+1-nsmax,2)
       endif

       lm=(ith-1)*lmstep

       do m = 0, nmmax
          par_lm=1.0_dp
          b_n = plm(lm) * alm(1,m,m)
          b_s = plm(lm) * alm(1,m,m)
          lm=lm+1
          do l = m+1, nlmax
             par_lm=-par_lm
             factor = plm(lm) * alm(1,l,m)
             b_n = b_n +          factor
             b_s = b_s + par_lm * factor
             lm=lm+1
          enddo

          b_north(m) = b_n
          b_south(m) = b_s

       enddo

       !        ---------------------------------------------------------------

       !        obtains the negative m of b(m,theta) (= complex conjugate)


       !        ---------------------------------------------------------------
       !        sum_m  b(m,theta)*exp(i*m*phi)   -> f(phi,theta)
       !        ---------------------------------------------------------------
       call ring_synthesis2(nsmax,nlmax,nmmax,b_north,nph,ring,kphi0,trig)   ! north hemisph. + equator
       map(istart_north(ith-1):istart_north(ith-1)+nph-1) = ring(0:nph-1)


       if (ith .lt. 2*nsmax) then
          call ring_synthesis2(nsmax,nlmax,nmmax,b_south,nph,ring,kphi0,trig) ! south hemisph. w/o equat
          map(istart_south(ith):istart_south(ith)+nph-1) = ring(0:nph-1)
       endif

    enddo    ! loop on cos(theta)
!$OMP END PARALLEL DO

    !     --------------------
    !     free memory and exit
    !     --------------------
    DEALLOCATE(istart_north,istart_south)
    DEALLOCATE(trig)
    return
  end subroutine alm2map_sc_pre

  !=======================================================================
  subroutine init_trig(nsmax,nmmax,trig)
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: nsmax,nmmax
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig   
    

    REAL(DP) :: phi0
    INTEGER(I4B) :: i,m,status

       ALLOCATE(trig(0:max(2*nsmax,nmmax),0:nsmax),stat = status) 
       if (status /= 0) call die_alloc('init_trig','trig')
       trig(:,:)=CMPLX(1.0_dp,0.0_dp, KIND=DP)
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(i,phi0,m)
       do i=1,nsmax
          phi0=PI/DBLE(i*4)
          do m=0,max(2*nsmax,nmmax)
             trig(m,i)= CMPLX( DCOS(m*phi0), DSIN(m*phi0), kind=DP)
          enddo
       enddo
!$OMP END PARALLEL DO
  end subroutine init_trig

  !=======================================================================
  subroutine ring_synthesis(nsmax,nlmax,nmmax,datain,nph,dataout,kphi0)
    !=======================================================================
    !     RING_SYNTHESIS
    !       called by alm2map
    !       calls     fft_gpd
    !
    !     dataout(j) = sum_m datain(m) * exp(i*m*phi(j)) 
    !     with phi(j) = j*2pi/nph + kphi0*pi/nph and kphi0 =0 or 1
    !
    !     as the set of frequencies {m} is larger than nph, 
    !     we wrap frequencies within {0..nph-1}
    !     ie  m = k*nph + m' with m' in {0..nph-1}
    !     then
    !     noting bw(m') = exp(i*m'*phi0) 
    !                   * sum_k (datain(k*nph+m') exp(i*k*pi*kphi0))
    !        with bw(nph-m') = CONJ(bw(m')) (if datain(-m) = CONJ(datain(m)))
    !     dataout(j) = sum_m' [ bw(m') exp (i*j*m'*2pi/nph) ]
    !                = Fourier Transform of bw
    !        is real
    !
    !         NB nph is not necessarily a power of 2
    !
    !=======================================================================


    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    INTEGER(I4B), INTENT(IN) :: nph, kphi0

    COMPLEX(DPC), DIMENSION(0:nmmax), INTENT(IN) :: datain
    REAL(SP),     DIMENSION(0:nph-1), INTENT(OUT):: dataout
    REAL(DP),     DIMENSION(0:nph-1)             :: data
    COMPLEX(DPC),     DIMENSION(:,:), ALLOCATABLE, SAVE :: trig
    REAL(DP) :: phi0
    INTEGER(I4B) :: i,iw,ksign,m,k,kshift
    COMPLEX(DPC), DIMENSION(0:nph-1) :: bw
    INTEGER(I4B), SAVE :: trg=-1,oldnmmax=-1
    COMPLEX(DPC) :: dat
    INTEGER(I4B) :: status

    !=======================================================================

    if ((trg/=nsmax).or.(oldnmmax/=nmmax)) then
       if(allocated(trig)) deallocate(trig)

       ALLOCATE(trig(0:max(2*nsmax,nmmax),0:nsmax),stat = status) 
       if (status /= 0) call die_alloc('ring_synthesis','trig')


      trig(:,:)=CMPLX(1.0_dp,0.0_dp, KIND=DP)
       do i=1,nsmax
          phi0=PI/DBLE(i*4)
          do m=0,max(2*nsmax,nmmax)
             trig(m,i)= CMPLX( DCOS(m*phi0), DSIN(m*phi0), kind=DP)
          enddo
      enddo
       trg=nsmax
       oldnmmax=nmmax

    endif

    !-----------------------------------------------------------------------

    ksign = + 1
    kshift = (-1)**kphi0  ! either 1 or -1
    bw(0:nph-1) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)

    !     all frequencies [-m,m] are wrapped in [0,nph-1]
    bw(0)=datain(0)
    do m  = 1, nmmax                        ! in -nmmax, nmmax
       iw = MODULO(m, nph)  ! between 0 and nph-1  = m', F90 intrisic
       k  = (m - iw) / nph                ! number of 'turns'
       bw(iw) = bw(iw) + datain(m)*(kshift**k)  ! complex number
       iw = MODULO(-m, nph)  ! between 0 and nph-1  = m', F90 intrisic
       k  = (-m - iw) / nph                ! number of 'turns'
       bw(iw) = bw(iw) + CONJG(datain(m))*(kshift**k)  ! complex number
    enddo
    !     kshift**k = 1       for even turn numbers
    !               = 1 or -1 for odd  turn numbers : results from the shift in space

    !     applies the shift in position <-> phase factor in Fourier space
    data(0)=REAL(bw(0))
    do iw = 1, nph/2-1
       m = ksign*(iw)
       if(kphi0==1) then
          dat =bw(iw) * trig(m,nph/4)
       else
          dat =bw(iw)
       endif
       data(iw*2-1) = REAL(dat)
       data(iw*2  ) = AIMAG(dat)
    enddo
    iw=nph/2
    m = ksign*(iw)
    if(kphi0==1) then
       dat =bw(iw) * trig(m,nph/4)
    else
       dat =bw(iw)
    endif
    data(iw*2-1) = REAL(dat)

    call real_fft (data, backward=.true.)
    !     ^^^^^^^^^^^^
    dataout=REAL(data)

    RETURN
  END subroutine ring_synthesis

  !=======================================================================
  subroutine ring_synthesis2(nsmax,nlmax,nmmax,datain,nph,dataout,kphi0,trig)
    !=======================================================================
    !     RING_SYNTHESIS
    !       called by alm2map
    !       calls     fft_gpd
    !
    !     dataout(j) = sum_m datain(m) * exp(i*m*phi(j)) 
    !     with phi(j) = j*2pi/nph + kphi0*pi/nph and kphi0 =0 or 1
    !
    !     as the set of frequencies {m} is larger than nph, 
    !     we wrap frequencies within {0..nph-1}
    !     ie  m = k*nph + m' with m' in {0..nph-1}
    !     then
    !     noting bw(m') = exp(i*m'*phi0) 
    !                   * sum_k (datain(k*nph+m') exp(i*k*pi*kphi0))
    !        with bw(nph-m') = CONJ(bw(m')) (if datain(-m) = CONJ(datain(m)))
    !     dataout(j) = sum_m' [ bw(m') exp (i*j*m'*2pi/nph) ]
    !                = Fourier Transform of bw
    !        is real
    !
    !         NB nph is not necessarily a power of 2
    !
    !=======================================================================


    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    INTEGER(I4B), INTENT(IN) :: nph, kphi0
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig

    COMPLEX(DPC), DIMENSION(0:nmmax), INTENT(IN) :: datain
    REAL(SP),     DIMENSION(0:nph-1), INTENT(OUT):: dataout
    REAL(DP),     DIMENSION(0:nph-1)             :: data
    INTEGER(I4B) :: i,iw,ksign,m,k,kshift
    COMPLEX(DPC), DIMENSION(0:nph-1) :: bw
    COMPLEX(DPC) :: dat
    INTEGER(I4B) :: status

    !=======================================================================


    !-----------------------------------------------------------------------

    ksign = + 1
    kshift = (-1)**kphi0  ! either 1 or -1
    bw(0:nph-1) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)

    !     all frequencies [-m,m] are wrapped in [0,nph-1]
    bw(0)=datain(0)
    do m  = 1, nmmax                        ! in -nmmax, nmmax
       iw = MODULO(m, nph)  ! between 0 and nph-1  = m', F90 intrisic
       k  = (m - iw) / nph                ! number of 'turns'
       bw(iw) = bw(iw) + datain(m)*(kshift**k)  ! complex number
       iw = MODULO(-m, nph)  ! between 0 and nph-1  = m', F90 intrisic
       k  = (-m - iw) / nph                ! number of 'turns'
       bw(iw) = bw(iw) + CONJG(datain(m))*(kshift**k)  ! complex number
    enddo
    !     kshift**k = 1       for even turn numbers
    !               = 1 or -1 for odd  turn numbers : results from the shift in space

    !     applies the shift in position <-> phase factor in Fourier space
    data(0)=REAL(bw(0))
    do iw = 1, nph/2-1
       m = ksign*(iw)
       if(kphi0==1) then
          dat =bw(iw) * trig(m,nph/4)
       else
          dat =bw(iw)
       endif
       data(iw*2-1) = REAL(dat)
       data(iw*2  ) = AIMAG(dat)
    enddo
    iw=nph/2
    m = ksign*(iw)
    if(kphi0==1) then
       dat =bw(iw) * trig(m,nph/4)
    else
       dat =bw(iw)
    endif
    data(iw*2-1) = REAL(dat)

    call real_fft (data, backward=.true.)
    !     ^^^^^^^^^^^^
    dataout=REAL(data)

    RETURN
  END subroutine ring_synthesis2

  !=======================================================================
  subroutine alm2map_pol(nsmax, nlmax, nmmax, alm_TGC, map_TQU)
    !=======================================================================
    !     computes a map form its alm for the HEALPIX pixelisation
    !     map(theta,phi) = sum_l_m a_lm Y_lm(theta,phi)
    !                    = sum_m {e^(i*m*phi) sum_l a_lm*lambda_lm(theta)}
    !
    !     where Y_lm(theta,phi) = lambda(theta) * e^(i*m*phi)
    !
    !     * the recurrence of Ylm is the standard one (cf Num Rec)
    !     * the sum over m is done by FFT
    !
    !              -------------------------------
    !          precomputes the Lambda_lm recurrence factor 
    !      and is therefore ~50% faster than previous versions
    !              -------------------------------
    !      This routine also include polarisation, map_TQU and alm_TGC
    !      are excpected to have an extra dimension for tensors.
    !
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    COMPLEX(SPC), INTENT(IN),  DIMENSION(1:3,0:nlmax,0:nmmax) :: alm_TGC
    REAL(SP),     INTENT(OUT), DIMENSION(0:12*nsmax**2-1,1:3) :: map_TQU

    INTEGER(I4B) :: l, m, ith, scalem, scalel          ! alm related
    INTEGER(I4B) :: nph, kphi0                                             ! map related
    INTEGER(I4B), ALLOCATABLE, DIMENSION(:) :: istart_south, istart_north  ! map related

    REAL(DP) :: cth, sth, dth1, dth2, dst1, OVFLOW, UNFLOW, logOVFLOW
    REAL(DP) :: a_rec, lam_mm, lam_lm, lam_lm1m, lam_0, lam_1, lam_2, par_lm
    REAL(DP) :: fm, f2m, fm2, fl, fl2, corfac
    REAL(DP) :: c_on_s2, fm_on_s2, one_on_s2
    REAL(DP) :: lambda_w, lambda_x, a_w, b_w, a_x
    COMPLEX(DPC) :: zi_lam_x
    COMPLEX(DPC) :: factor, factor_1, factor_2
    COMPLEX(DPC) :: b_n_T, b_s_T, b_n_Q, b_s_Q, b_n_U, b_s_U

    CHARACTER(LEN=*), PARAMETER :: code = 'ALM2MAP_POL'
    COMPLEX(DPC), DIMENSION(0:nmmax) :: b_north_T, b_north_Q, b_north_U
    COMPLEX(DPC), DIMENSION(0:nmmax) :: b_south_T, b_south_Q, b_south_U
    INTEGER(I4B) :: status

    REAL(DP), DIMENSION(:,:), ALLOCATABLE :: recfac
    REAL(DP), DIMENSION(:,:), ALLOCATABLE :: lam_fact
    REAL(SP), DIMENSION(0:4*nsmax-1) :: ring
    REAL(DP), DIMENSION(:),   ALLOCATABLE :: normal_l
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig   
   !=======================================================================

    !     --- allocates space for arrays ---
    ALLOCATE(recfac(0:nlmax,0:nmmax),stat = status)    
    if (status /= 0) call die_alloc(code,'recfac')

    ALLOCATE(lam_fact(0:nlmax,0:nmmax),stat = status)    
    if (status /= 0) call die_alloc(code,'lam_fact')

    ALLOCATE(normal_l(0:nlmax),stat = status)    
    if (status /= 0) call die_alloc(code,'normal_l')



    ALLOCATE(istart_north(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_north')

    ALLOCATE(istart_south(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_south')


    call init_trig(nsmax,nmmax,trig)

    !     ------------ initiate arrays ----------------

    recfac(:,:) = 0.0_dp
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,fm2,l,fl2)
    do m = 0, nmmax
       fm2 = DBLE(m) **2
       do l = m, nlmax
          fl2 = DBLE(l+1) **2
          recfac(l,m) = DSQRT( (4.0_dp * fl2 - 1.0_dp) / (fl2-fm2) )
       enddo
    enddo
!OMP END PARALLEL DO

    lam_fact(:,:) = 0.0_dp
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,fm2,l,fl,fl2)
    do l = 2, nlmax
       fl  = DBLE(l)
       fl2 = fl * fl
       do m = 0, min(l,nmmax) ! corrected 2001-12
          fm2 = DBLE(m) **2
          lam_fact(l,m) = 2.0_dp * SQRT( (2.0_dp * fl + 1.0_dp) / (2.0_dp * fl - 1.0_dp) * (fl2-fm2) )
       enddo
    enddo
!OMP END PARALLEL DO

    normal_l(:) = 0.0_dp
    do l = 2, nlmax
       fl = DBLE(l)
       normal_l(l) = SQRT( KvS / ((fl+2.0_dp)*(fl+1.0_dp)*fl*(fl-1.0_dp)) ) 
       ! either CMBFAST (KvS=1) or Kamionkowski et al. (KvS=2) definition
    enddo

    !     ----- set the whole map to zero ------
    map_TQU = 0.0
    !     --------------------------------------

    istart_north(0)=0
    istart_south(0)=12*nsmax**2
    do ith=1,2*nsmax
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
       else                    ! tropical band (north) + equator
          nph = 4*nsmax
       endif
       istart_north(ith)=istart_north(ith-1)+nph
       istart_south(ith)=istart_south(ith-1)-nph
    enddo

    OVFLOW=exp(log(FL_LARGE))
    UNFLOW=exp(log(FL_SMALL))
    logOVFLOW=log(FL_LARGE)

    dth1 = 1.0_dp / (3.0_dp*DBLE(nsmax)**2)
    dth2 = 2.0_dp / (3.0_dp*DBLE(nsmax))
    dst1 = 1.0_dp / (SQRT(6.0_dp) * DBLE(nsmax) )

    !     --------------------------------------------

!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(ith,cth,nph,kphi0,sth,one_on_s2,c_on_s2,lam_mm,scalem, &
!$OMP         m,fm,f2m,fm2,fm_on_s2,par_lm,corfac,lam_lm, &
!$OMP         b_n_T,b_s_T,lambda_w,lambda_x,zi_lam_x,b_n_Q,b_s_Q,b_n_U, &
!$OMP         b_s_U,lam_0,lam_1,scalel,a_rec,lam_2,l,lam_lm1m,fl,fl2,factor, &
!$OMP         a_w,b_w,a_x,factor_1,factor_2,b_north_T,b_south_T,b_north_Q, &
!$OMP         b_south_Q,b_north_U,b_south_U,ring)
    do ith = 1,2*nsmax      ! 0 <= cos theta < 1
       !        cos(theta) in the pixelisation scheme
       if (ith.lt.nsmax) then  ! polar cap (north)
          cth = 1.0_dp  - DBLE(ith)**2 * dth1
          nph = 4*ith
          kphi0 = 1
          sth = SIN( 2.0_dp * ASIN( ith * dst1 ) ) ! sin(theta)
       else                   ! tropical band (north) + equator
          cth = DBLE(2*nsmax-ith) * dth2
          nph = 4*nsmax
          kphi0 = MOD(ith+1-nsmax,2)
          sth = DSQRT((1.0_dp-cth)*(1.0_dp+cth)) ! sin(theta)
       endif
       one_on_s2 = 1.0_dp / sth**2 ! 1/sin^2
       c_on_s2 = cth * one_on_s2
       !        -----------------------------------------------------
       !        for each theta, and each m, computes
       !        b(m,theta) = sum_over_l>m (lambda_l_m(theta) * a_l_m) 
       !        ------------------------------------------------------
       !        lambda_mm tends to go down when m increases (risk of underflow)
       !        lambda_lm tends to go up   when l increases (risk of overflow)
       lam_mm = sq4pi_inv ! lamda_00
       scalem=1
       do m = 0, nmmax
          fm  = DBLE(m)
          f2m = 2.0_dp * fm
          fm2 = fm * fm
          fm_on_s2 = fm * one_on_s2

          !           ---------- l = m ----------
          par_lm = 1.0_dp  ! = (-1)^(l+m)
          if (m .ge. 1) then ! lambda_0_0 for m>0
             lam_mm = -lam_mm*sth*dsqrt((f2m+1.0_dp)/f2m)
          endif

          if (abs(lam_mm).lt.UNFLOW) then
             lam_mm=lam_mm*OVFLOW
             scalem=scalem-1
          endif
          corfac=exp(scalem*logOVFLOW)  

          ! alm_T * Ylm : Temperature
          lam_lm = corfac*lam_mm/OVFLOW     !  actual lambda_mm      
          b_n_T = lam_lm * alm_TGC(1,m,m)
          b_s_T = lam_lm * alm_TGC(1,m,m)

          lambda_w = ( normal_l(m) * lam_lm * (fm - fm2) ) * ( 2.0_dp * one_on_s2 - 1.0_dp ) * SW1
          lambda_x = ( normal_l(m) * lam_lm * (fm - fm2) ) *   2.0_dp *   c_on_s2
          zi_lam_x = CMPLX(0.0_dp, lambda_x, KIND=DP)

          ! alm_G * Ylm_W - alm_C * Ylm_X : Polarisation Q
          b_n_Q =  lambda_w * alm_TGC(2,m,m) + zi_lam_x * alm_TGC(3,m,m)
          b_s_Q =  lambda_w * alm_TGC(2,m,m) - zi_lam_x * alm_TGC(3,m,m)

          ! - alm_G * Ylm_X - alm_C * Ylm_W : Polarisation U
          b_n_U = -lambda_w * alm_TGC(3,m,m) + zi_lam_x * alm_TGC(2,m,m)
          b_s_U = -lambda_w * alm_TGC(3,m,m) - zi_lam_x * alm_TGC(2,m,m)

          !           ---------- l > m ----------
          lam_0 = 0.0_dp
          lam_1 = 1.0_dp
          scalel=0
          a_rec = recfac(m,m)
          lam_2 = cth * lam_1 * a_rec
          do l = m+1, nlmax
             par_lm = - par_lm  ! = (-1)^(l+m)
             lam_lm1m=lam_lm  ! actual lambda_l-1,m (useful for polarisation)
             lam_lm = lam_2 * corfac*lam_mm/OVFLOW ! actual lambda_lm, OVFLOW factors removed
             fl  = DBLE(l)
             fl2 = fl * fl

             ! alm_T * Ylm : Temperature
             factor = lam_lm * alm_TGC(1,l,m)
             b_n_T = b_n_T +          factor
             b_s_T = b_s_T + par_lm * factor

             a_w =  2.0_dp * (fl - fm2) * one_on_s2 + (fl2 - fl)
             b_w =  c_on_s2 * lam_fact(l,m)
             a_x =  2.0_dp * cth * (fl-1.0_dp) * lam_lm
             lambda_w =   normal_l(l) * ( a_w * lam_lm - b_w * lam_lm1m ) * SW1
             lambda_x = - normal_l(l) * fm_on_s2 * ( a_x - lam_fact(l,m) * lam_lm1m )
             zi_lam_x = CMPLX(0.0_dp, lambda_x, KIND=DP)

             ! alm_G * Ylm_W - alm_C * Ylm_X : Polarisation Q
             factor_1 =  lambda_w * alm_TGC(2,l,m)
             factor_2 =  zi_lam_x * alm_TGC(3,l,m) ! X is imaginary
             b_n_Q = b_n_Q +           factor_1 + factor_2
             b_s_Q = b_s_Q + par_lm * (factor_1 - factor_2)! X has a diff. parity

             !- alm_G * Ylm_X - alm_C * Ylm_W : Polarisation U
             factor_1 = - lambda_w * alm_TGC(3,l,m)
             factor_2 =   zi_lam_x * alm_TGC(2,l,m) ! X is imaginary
             b_n_U = b_n_U +           factor_1 + factor_2
             b_s_U = b_s_U + par_lm * (factor_1 - factor_2)! X has a diff. parity

             lam_0 = lam_1 / a_rec
             lam_1 = lam_2
             a_rec = recfac(l,m)
             lam_2 = (cth * lam_1 - lam_0) * a_rec

             if (abs(lam_2) .gt. OVFLOW) then
                lam_0=lam_0/OVFLOW
                lam_1=lam_1/OVFLOW
                lam_2 = (cth * lam_1 - lam_0) * a_rec
                scalel=scalel+1
                corfac=exp((scalem+scalel)*logOVFLOW)
             elseif (abs(lam_2) .lt. UNFLOW) then
                lam_0=lam_0*OVFLOW
                lam_1=lam_1*OVFLOW
                lam_2 = (cth * lam_1 - lam_0) * a_rec 
                scalel=scalel-1
                corfac=exp((scalem+scalel)*logOVFLOW)
             endif

          enddo

          b_north_T(m) = b_n_T
          b_south_T(m) = b_s_T
          b_north_Q(m) = b_n_Q * sgQ
          b_south_Q(m) = b_s_Q * sgQ
          b_north_U(m) = b_n_U
          b_south_U(m) = b_s_U

       enddo
       !        ---------------------------------------------------------------

       !        obtains the negative m of b(m,theta) (= complex conjugate)


       !        ---------------------------------------------------------------
       !        sum_m  b(m,theta)*exp(i*m*phi)   -> f(phi,theta)
       !        ---------------------------------------------------------------
       call ring_synthesis2(nsmax, nlmax, nmmax, b_north_T, nph, ring, kphi0, trig) ! N hemisph. + equator
       map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,1) = ring(0:nph-1)
       call ring_synthesis2(nsmax, nlmax, nmmax, b_north_Q, nph, ring, kphi0, trig)
       map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,2) = ring(0:nph-1)
       call ring_synthesis2(nsmax, nlmax, nmmax, b_north_U, nph, ring, kphi0, trig)
       map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,3) = ring(0:nph-1)

       if (ith .lt. 2*nsmax) then
          call ring_synthesis2(nsmax, nlmax, nmmax, b_south_T, nph, ring, kphi0, trig) ! S hemisph. w/o equat
          map_TQU(istart_south(ith):istart_south(ith)+nph-1,1) = ring(0:nph-1)
          call ring_synthesis2(nsmax, nlmax, nmmax, b_south_Q, nph, ring, kphi0, trig)
          map_TQU(istart_south(ith):istart_south(ith)+nph-1,2) = ring(0:nph-1)
          call ring_synthesis2(nsmax, nlmax, nmmax, b_south_U, nph, ring, kphi0, trig)
          map_TQU(istart_south(ith):istart_south(ith)+nph-1,3) = ring(0:nph-1)
       endif

    enddo    ! loop on cos(theta)
!$OMP END PARALLEL DO

    !     --------------------
    !     free memory and exit
    !     --------------------
    DEALLOCATE(recfac)
    DEALLOCATE(lam_fact)
    DEALLOCATE(normal_l)
    DEALLOCATE(trig)
    return
  end subroutine alm2map_pol

  !=======================================================================
  subroutine alm2map_pol_pre1(nsmax, nlmax, nmmax, alm_TGC, map_TQU,plm)
    !=======================================================================
    !     computes a map form its alm for the HEALPIX pixelisation
    !     map(theta,phi) = sum_l_m a_lm Y_lm(theta,phi)
    !                    = sum_m {e^(i*m*phi) sum_l a_lm*lambda_lm(theta)}
    !
    !     where Y_lm(theta,phi) = lambda(theta) * e^(i*m*phi)
    !
    !     * Uses precomputed P_lms
    !     * the sum over m is done by FFT
    !
    !              -------------------------------
    !          precomputes the Lambda_lm recurrence factor 
    !      and is therefore ~50% faster than previous versions
    !              -------------------------------
    !
    !      This routine also include polarisation, map_TQU and alm_TGC
    !      are excpected to have an extra dimension for tensors.
    !
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    COMPLEX(SPC), INTENT(IN),  DIMENSION(1:3,0:nlmax,0:nmmax) :: alm_TGC
    REAL(SP),     INTENT(OUT), DIMENSION(0:12*nsmax**2-1,1:3) :: map_TQU
    REAL(DP), INTENT(IN), DIMENSION(0:(nlmax+1)*(nlmax+2)*nsmax-1) :: plm

    INTEGER(I4B) :: l, m, ith         ! alm related
    INTEGER(I4B) :: nph, kphi0                                             ! map related
    INTEGER(I4B), ALLOCATABLE, DIMENSION(:) :: istart_south, istart_north  ! map related

    REAL(DP) :: cth, sth, dth1, dth2, dst1
    REAL(DP) :: par_lm
    REAL(DP) :: fm, f2m, fm2, fl, fl2
    REAL(DP) :: c_on_s2, fm_on_s2, one_on_s2
    REAL(DP) :: lambda_w, lambda_x, a_w, b_w, a_x
    COMPLEX(DPC) :: zi_lam_x
    COMPLEX(DPC) :: factor, factor_1, factor_2
    COMPLEX(DPC) :: b_n_T, b_s_T, b_n_Q, b_s_Q, b_n_U, b_s_U

    CHARACTER(LEN=*), PARAMETER :: code = 'ALM2MAP_POL'
    COMPLEX(DPC), DIMENSION(0:nmmax) :: b_north_T, b_north_Q, b_north_U
    COMPLEX(DPC), DIMENSION(0:nmmax) :: b_south_T, b_south_Q, b_south_U
    INTEGER(I4B) :: status, lm, lmstep

    REAL(DP), DIMENSION(:,:), ALLOCATABLE :: lam_fact
    REAL(SP), DIMENSION(0:4*nsmax-1) :: ring
    REAL(DP), DIMENSION(:),   ALLOCATABLE :: normal_l
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig   
    !=======================================================================

    !     --- allocates space for arrays ---

    ALLOCATE(lam_fact(0:nlmax,0:nmmax),stat = status)    
    if (status /= 0) call die_alloc(code,'lam_fact')

    ALLOCATE(normal_l(0:nlmax),stat = status)    
    if (status /= 0) call die_alloc(code,'normal_l')



    ALLOCATE(istart_north(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_north')

    ALLOCATE(istart_south(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_south')


    call init_trig(nsmax,nmmax,trig)
    !     ------------ initiate arrays ----------------

    lam_fact(:,:) = 0.0_dp
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,fm2,l,fl,fl2)
    do l = 2, nlmax
       fl  = DBLE(l)
       fl2 = fl * fl
       do m = 0, min(l, nmmax) ! corrected 2001-12
          fm2 = DBLE(m) **2
          lam_fact(l,m) = 2.0_dp * SQRT( (2.0_dp * fl + 1.0_dp) / (2.0_dp * fl - 1.0_dp) * (fl2-fm2) )
       enddo
    enddo
!$OMP END PARALLEL DO

    normal_l(:) = 0.0_dp
    do l = 2, nlmax
       fl = DBLE(l)
       normal_l(l) = SQRT( KvS / ((fl+2.0_dp)*(fl+1.0_dp)*fl*(fl-1.0_dp)) ) 
       ! either CMBFAST (KvS=1) or Kamionkowski et al. (KvS=2) definition
    enddo

    !     ----- set the whole map to zero ------
    map_TQU = 0.0
    !     --------------------------------------

    istart_north(0)=0
    istart_south(0)=12*nsmax**2
    do ith=1,2*nsmax
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
       else                    ! tropical band (north) + equator
          nph = 4*nsmax
       endif
       istart_north(ith)=istart_north(ith-1)+nph
       istart_south(ith)=istart_south(ith-1)-nph
    enddo

    lmstep=(nlmax+1)*(nlmax+2)/2


    dth1 = 1.0_dp / (3.0_dp*DBLE(nsmax)**2)
    dth2 = 2.0_dp / (3.0_dp*DBLE(nsmax))
    dst1 = 1.0_dp / (SQRT(6.0_dp) * DBLE(nsmax) )


    !     --------------------------------------------

!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(ith,cth,nph,kphi0,sth,one_on_s2,c_on_s2,m,fm,f2m, &
!$OMP         fm2,fm_on_s2,par_lm,b_n_T,b_s_T,lambda_w,lambda_x, &
!$OMP         zi_lam_x,b_n_Q,b_s_Q,b_n_U,b_s_U,lm,l,fl,fl2,factor, &
!$OMP         a_w,b_w,a_x,factor_1,factor_2,b_north_T,b_south_T, &
!$OMP         b_north_Q,b_south_Q,b_north_U,b_south_U,ring)
    do ith = 1,2*nsmax      ! 0 <= cos theta < 1
       !        cos(theta) in the pixelisation scheme
       if (ith.lt.nsmax) then  ! polar cap (north)
          cth = 1.0_dp  - DBLE(ith)**2 * dth1
          nph = 4*ith
          kphi0 = 1
          sth = SIN( 2.0_dp * ASIN( ith * dst1 ) ) ! sin(theta)
       else                   ! tropical band (north) + equator
          cth = DBLE(2*nsmax-ith) * dth2
          nph = 4*nsmax
          kphi0 = MOD(ith+1-nsmax,2)
          sth = DSQRT((1.0_dp-cth)*(1.0_dp+cth)) ! sin(theta)
       endif
       one_on_s2 = 1.0_dp / sth**2 ! 1/sin^2
       c_on_s2 = cth * one_on_s2
       lm=(ith-1)*lmstep
       !        -----------------------------------------------------
       !        for each theta, and each m, computes
       !        b(m,theta) = sum_over_l>m (lambda_l_m(theta) * a_l_m) 
       !        ------------------------------------------------------

       do m = 0, nmmax
          fm  = DBLE(m)
          f2m = 2.0_dp * fm
          fm2 = fm * fm
          fm_on_s2 = fm * one_on_s2

          !           ---------- l = m ----------
          par_lm = 1.0_dp  ! = (-1)^(l+m)

          ! alm_T * Ylm : Temperature
          b_n_T = plm(lm) * alm_TGC(1,m,m)
          b_s_T = plm(lm) * alm_TGC(1,m,m)

          lambda_w = ( normal_l(m) * plm(lm) * (fm - fm2) ) * ( 2.0_dp * one_on_s2 - 1.0_dp ) * SW1
          lambda_x = ( normal_l(m) * plm(lm) * (fm - fm2) ) *   2.0_dp *   c_on_s2
          zi_lam_x = CMPLX(0.0_dp, lambda_x, KIND=DP)

          ! alm_G * Ylm_W - alm_C * Ylm_X : Polarisation Q
          b_n_Q =  lambda_w * alm_TGC(2,m,m) + zi_lam_x * alm_TGC(3,m,m)
          b_s_Q =  lambda_w * alm_TGC(2,m,m) - zi_lam_x * alm_TGC(3,m,m)

          ! - alm_G * Ylm_X - alm_C * Ylm_W : Polarisation U
          b_n_U = -lambda_w * alm_TGC(3,m,m) + zi_lam_x * alm_TGC(2,m,m)
          b_s_U = -lambda_w * alm_TGC(3,m,m) - zi_lam_x * alm_TGC(2,m,m)

          lm=lm+1

          !           ---------- l > m ----------

          do l = m+1, nlmax
             par_lm = - par_lm  ! = (-1)^(l+m)
             fl  = DBLE(l)
             fl2 = fl * fl
             ! alm_T * Ylm : Temperature
             factor = plm(lm) * alm_TGC(1,l,m)
             b_n_T = b_n_T +          factor
             b_s_T = b_s_T + par_lm * factor

             a_w =  2.0_dp * (fl - fm2) * one_on_s2 + (fl2 - fl)
             b_w =  c_on_s2 * lam_fact(l,m)
             a_x =  2.0_dp * cth * (fl-1.0_dp) * plm(lm)
             lambda_w =   normal_l(l) * ( a_w * plm(lm) - b_w * plm(lm-1) ) * SW1
             lambda_x = - normal_l(l) * fm_on_s2 * ( a_x - lam_fact(l,m) * plm(lm-1) )
             zi_lam_x = CMPLX(0.0_dp, lambda_x, KIND=DP)

             ! alm_G * Ylm_W - alm_C * Ylm_X : Polarisation Q
             factor_1 =  lambda_w * alm_TGC(2,l,m)
             factor_2 =  zi_lam_x * alm_TGC(3,l,m) ! X is imaginary
             b_n_Q = b_n_Q +           factor_1 + factor_2
             b_s_Q = b_s_Q + par_lm * (factor_1 - factor_2)! X has a diff. parity

             !- alm_G * Ylm_X - alm_C * Ylm_W : Polarisation U
             factor_1 = - lambda_w * alm_TGC(3,l,m)
             factor_2 =   zi_lam_x * alm_TGC(2,l,m) ! X is imaginary
             b_n_U = b_n_U +           factor_1 + factor_2
             b_s_U = b_s_U + par_lm * (factor_1 - factor_2)! X has a diff. parity
             lm=lm+1

          enddo

          b_north_T(m) = b_n_T
          b_south_T(m) = b_s_T
          b_north_Q(m) = b_n_Q * sgQ
          b_south_Q(m) = b_s_Q * sgQ
          b_north_U(m) = b_n_U
          b_south_U(m) = b_s_U

       enddo
       !        ---------------------------------------------------------------

       !        obtains the negative m of b(m,theta) (= complex conjugate)


       !        ---------------------------------------------------------------
       !        sum_m  b(m,theta)*exp(i*m*phi)   -> f(phi,theta)
       !        ---------------------------------------------------------------
       call ring_synthesis2(nsmax, nlmax, nmmax, b_north_T, nph, ring, kphi0, trig) ! N hemisph. + equator
       map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,1) = ring(0:nph-1)
       call ring_synthesis2(nsmax, nlmax, nmmax, b_north_Q, nph, ring, kphi0, trig)
       map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,2) = ring(0:nph-1)
       call ring_synthesis2(nsmax, nlmax, nmmax, b_north_U, nph, ring, kphi0, trig)
       map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,3) = ring(0:nph-1)

       if (ith .lt. 2*nsmax) then
          call ring_synthesis2(nsmax, nlmax, nmmax, b_south_T, nph, ring, kphi0, trig) ! S hemisph. w/o equat
          map_TQU(istart_south(ith):istart_south(ith)+nph-1,1) = ring(0:nph-1)
          call ring_synthesis2(nsmax, nlmax, nmmax, b_south_Q, nph, ring, kphi0, trig)
          map_TQU(istart_south(ith):istart_south(ith)+nph-1,2) = ring(0:nph-1)
          call ring_synthesis2(nsmax, nlmax, nmmax, b_south_U, nph, ring, kphi0, trig)
          map_TQU(istart_south(ith):istart_south(ith)+nph-1,3) = ring(0:nph-1)
       endif

    enddo    ! loop on cos(theta)
!$OMP END PARALLEL DO

    !     --------------------
    !     free memory and exit
    !     --------------------
    DEALLOCATE(lam_fact)
    DEALLOCATE(normal_l)
 !PPP   DEALLOCATE(b_north_T,b_north_Q,b_north_U)
 !PPP   DEALLOCATE(b_south_T,b_south_Q,b_south_U)
 !PPP   DEALLOCATE(ring)
 !PPP BEGIN
    DEALLOCATE(istart_north,istart_south)
    DEALLOCATE(trig)
 !PPP END
    return
  end subroutine alm2map_pol_pre1

  !=======================================================================
  subroutine alm2map_pol_pre2(nsmax, nlmax, nmmax, alm_TGC, map_TQU,plm)
    !=======================================================================
    !     computes a map form its alm for the HEALPIX pixelisation
    !     map(theta,phi) = sum_l_m a_lm Y_lm(theta,phi)
    !                    = sum_m {e^(i*m*phi) sum_l a_lm*lambda_lm(theta)}
    !
    !     where Y_lm(theta,phi) = lambda(theta) * e^(i*m*phi)
    !
    !     * Uses precomputed P_lms AND tensor harmonics.
    !     * the sum over m is done by FFT
    !
    !              -------------------------------
    !          precomputes the Lambda_lm recurrence factor 
    !      and is therefore ~50% faster than previous versions
    !              -------------------------------
    !
    !      This routine also include polarisation, map_TQU and alm_TGC
    !      are excpected to have an extra dimension for tensors.
    !
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    COMPLEX(SPC), INTENT(IN),  DIMENSION(1:3,0:nlmax,0:nmmax) :: alm_TGC
    REAL(SP),     INTENT(OUT), DIMENSION(0:12*nsmax**2-1,1:3) :: map_TQU
    REAL(DP), INTENT(IN), DIMENSION(0:(nlmax+1)*(nlmax+2)*nsmax-1,1:3) :: plm
    INTEGER(I4B) :: l, m, ith        ! alm related
    INTEGER(I4B) :: nph, kphi0                                             ! map related
    INTEGER(I4B), ALLOCATABLE, DIMENSION(:) :: istart_south, istart_north  ! map related

    REAL(DP) :: par_lm
    COMPLEX(DPC) :: zi_lam_x
    COMPLEX(DPC) :: factor, factor_1, factor_2
    COMPLEX(DPC) :: b_n_T, b_s_T, b_n_Q, b_s_Q, b_n_U, b_s_U

    CHARACTER(LEN=*), PARAMETER :: code = 'ALM2MAP_POL'
    COMPLEX(DPC), DIMENSION(0:nmmax) :: b_north_T, b_north_Q, b_north_U
    COMPLEX(DPC), DIMENSION(0:nmmax) :: b_south_T, b_south_Q, b_south_U
    INTEGER(I4B) :: status, lm, lmstep
    REAL(SP), DIMENSION(0:4*nsmax-1) :: ring
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig   
    !=======================================================================

    !     --- allocates space for arrays ---



    ALLOCATE(istart_north(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_north')

    ALLOCATE(istart_south(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_south')


    call init_trig(nsmax,nmmax,trig)
    !     ----- set the whole map to zero ------
    map_TQU = 0.0

    !     --------------------------------------

    istart_north(0)=0
    istart_south(0)=12*nsmax**2
    do ith=1,2*nsmax
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
       else                    ! tropical band (north) + equator
          nph = 4*nsmax
       endif
       istart_north(ith)=istart_north(ith-1)+nph
       istart_south(ith)=istart_south(ith-1)-nph
    enddo

    lmstep=(nlmax+1)*(nlmax+2)/2



    !     --------------------------------------------

!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(ith,nph,kphi0,m,b_n_T,b_s_T,zi_lam_x,b_n_Q,b_s_Q, &
!$OMP         b_n_U,b_s_U,lm,l,par_lm,factor,factor_1,factor_2, &
!$OMP         b_north_T,b_south_T,b_north_Q,b_south_Q,b_north_U,b_south_U,ring)
    do ith = 1,2*nsmax      ! 0 <= cos theta < 1
       !        cos(theta) in the pixelisation scheme
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
          kphi0 = 1
       else                   ! tropical band (north) + equator
          nph = 4*nsmax
          kphi0 = MOD(ith+1-nsmax,2)
       endif
       lm=(ith-1)*lmstep
       !        -----------------------------------------------------
       !        for each theta, and each m, computes
       !        b(m,theta) = sum_over_l>m (lambda_l_m(theta) * a_l_m) 
       !        ------------------------------------------------------


       do m = 0, nmmax

          !           ---------- l = m ----------
          par_lm = 1.0_dp  ! = (-1)^(l+m)

          ! alm_T * Ylm : Temperature
          b_n_T = plm(lm,1) * alm_TGC(1,m,m)
          b_s_T = plm(lm,1) * alm_TGC(1,m,m)

          zi_lam_x = CMPLX(0.0_dp, plm(lm,3), KIND=DP)

          ! alm_G * Ylm_W - alm_C * Ylm_X : Polarisation Q
          b_n_Q =  plm(lm,2) * alm_TGC(2,m,m) + zi_lam_x * alm_TGC(3,m,m)
          b_s_Q =  plm(lm,2) * alm_TGC(2,m,m) - zi_lam_x * alm_TGC(3,m,m)

          ! - alm_G * Ylm_X - alm_C * Ylm_W : Polarisation U
          b_n_U = -plm(lm,2) * alm_TGC(3,m,m) + zi_lam_x * alm_TGC(2,m,m)
          b_s_U = -plm(lm,2) * alm_TGC(3,m,m) - zi_lam_x * alm_TGC(2,m,m)

          lm=lm+1

          !           ---------- l > m ----------

          do l = m+1, nlmax
             par_lm = - par_lm  ! = (-1)^(l+m)

             ! alm_T * Ylm : Temperature
             factor = plm(lm,1) * alm_TGC(1,l,m)
             b_n_T = b_n_T +          factor
             b_s_T = b_s_T + par_lm * factor

             zi_lam_x = CMPLX(0.0_dp, plm(lm,3), KIND=DP)

             ! alm_G * Ylm_W - alm_C * Ylm_X : Polarisation Q
             factor_1 =  plm(lm,2) * alm_TGC(2,l,m)
             factor_2 =  zi_lam_x * alm_TGC(3,l,m) ! X is imaginary
             b_n_Q = b_n_Q +           factor_1 + factor_2
             b_s_Q = b_s_Q + par_lm * (factor_1 - factor_2)! X has a diff. parity

             !- alm_G * Ylm_X - alm_C * Ylm_W : Polarisation U
             factor_1 = - plm(lm,2) * alm_TGC(3,l,m)
             factor_2 =   zi_lam_x * alm_TGC(2,l,m) ! X is imaginary
             b_n_U = b_n_U +           factor_1 + factor_2
             b_s_U = b_s_U + par_lm * (factor_1 - factor_2)! X has a diff. parity

             lm=lm+1

          enddo

          b_north_T(m) = b_n_T
          b_south_T(m) = b_s_T
          b_north_Q(m) = b_n_Q * sgQ
          b_south_Q(m) = b_s_Q * sgQ
          b_north_U(m) = b_n_U
          b_south_U(m) = b_s_U

       enddo
       !        ---------------------------------------------------------------

       !        obtains the negative m of b(m,theta) (= complex conjugate)


       !        ---------------------------------------------------------------
       !        sum_m  b(m,theta)*exp(i*m*phi)   -> f(phi,theta)
       !        ---------------------------------------------------------------
       call ring_synthesis2(nsmax, nlmax, nmmax, b_north_T, nph, ring, kphi0, trig) ! N hemisph. + equator
       map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,1) = ring(0:nph-1)
       call ring_synthesis2(nsmax, nlmax, nmmax, b_north_Q, nph, ring, kphi0, trig)
       map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,2) = ring(0:nph-1)
       call ring_synthesis2(nsmax, nlmax, nmmax, b_north_U, nph, ring, kphi0, trig)
       map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,3) = ring(0:nph-1)

       if (ith .lt. 2*nsmax) then
          call ring_synthesis2(nsmax, nlmax, nmmax, b_south_T, nph, ring, kphi0, trig) ! S hemisph. w/o equat
          map_TQU(istart_south(ith):istart_south(ith)+nph-1,1) = ring(0:nph-1)
          call ring_synthesis2(nsmax, nlmax, nmmax, b_south_Q, nph, ring, kphi0, trig)
          map_TQU(istart_south(ith):istart_south(ith)+nph-1,2) = ring(0:nph-1)
          call ring_synthesis2(nsmax, nlmax, nmmax, b_south_U, nph, ring, kphi0, trig)
          map_TQU(istart_south(ith):istart_south(ith)+nph-1,3) = ring(0:nph-1)
       endif

    enddo    ! loop on cos(theta)
!$OMP END PARALLEL DO

    !     --------------------
    !     free memory and exit
    !     --------------------
    DEALLOCATE(istart_north,istart_south)
    DEALLOCATE(trig)
    return
  end subroutine alm2map_pol_pre2
  



  !=======================================================================
  subroutine map2alm_sc(nsmax, nlmax, nmmax, map, alm, cos_theta_cut, w8ring)
    !=======================================================================
    !     computes the a(l,m) from a map for the HEALPIX pixelisation
    !      for the Temperature field
    !     a(l,m) = int T(theta,phi) Y_lm(theta,phi)^* dtheta*dphi
    !            = int dtheta lambda_lm(theta) 
    !                  * int dphi T(theta,phi) e^(-i*m*phi)
    !
    !     where Y_lm(theta,phi) = lambda(theta) * e^(i*m*phi)
    !
    !     * the recurrence of Ylm is the standard one (cf Num Rec)
    !     * the integral over phi is done by FFT
    !
    !     cos_theta_cut (>0) is the cosine of the 
    !     symmetric cut applied to the sky
    !     if it is <= 0 no cut is applied
    !
    !     NB : these a(l,m) have to be multiplied by the pixel solid angle
    !      to give the correct coefficients
    !
    !             -------------------------------
    !         precomputes the Lambda_lm recurrence factor
    !      and is therefore ~50% faster than previous versions
    !     the multiplication by omega_pix is done in the routine
    !             -------------------------------
    !
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    REAL(SP),     INTENT(IN),  DIMENSION(0:12*nsmax**2-1) :: map
    COMPLEX(SPC), INTENT(OUT), DIMENSION(1:1,0:nlmax,0:nmmax) :: alm
    REAL(DP),     INTENT(IN) :: cos_theta_cut
    REAL(DP),     INTENT(IN), DIMENSION(1:2*nsmax,1) :: w8ring

    INTEGER(I4B) :: l, m, ith, im_max, scalem, scalel       ! alm related
    INTEGER(I4B) :: nph, kphi0                                             ! map related
    INTEGER(I4B), ALLOCATABLE, DIMENSION(:) :: istart_south, istart_north  ! map related

    REAL(DP) :: omega_pix
    REAL(DP) :: dth1, dth2, dst1, OVFLOW, UNFLOW, logOVFLOW
    REAL(DP) :: a_rec, lam_mm, lam_lm, lam_0, lam_1, lam_2, par_lm
    REAL(DP) :: f2m, fm2, fl2, corfac
    COMPLEX(DPC) :: p_n,p_s


    CHARACTER(LEN=*), PARAMETER :: code = 'MAP2ALM'
    COMPLEX(DPC), DIMENSION(:,:), ALLOCATABLE :: phas_n
    
    COMPLEX(DPC), DIMENSION(:,:), ALLOCATABLE :: phas_s
    INTEGER(I4B) :: status


    REAL(DP), DIMENSION(:,:), ALLOCATABLE :: recfac
    REAL(DP), DIMENSION(0:4*nsmax-1) :: ring
    LOGICAL   :: keep_it

    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig
    INTEGER(I4B), DIMENSION(:,:), ALLOCATABLE :: scalem_
    REAL(DP), DIMENSION(:,:), ALLOCATABLE :: lam_mm_
    COMPLEX(SPC), DIMENSION(0:nlmax) :: alm_tmp
    REAL(DP), DIMENSION(:), ALLOCATABLE :: cth,sth
    !=======================================================================


    !     --- allocates space for arrays ---
    ALLOCATE(recfac(0:nlmax,0:nmmax),stat = status)       
    if (status /= 0) call die_alloc(code,'recfac')


    ALLOCATE(cth(1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'cth')

    ALLOCATE(sth(1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'sth')

    ALLOCATE(phas_n(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_n')

    ALLOCATE(phas_s(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_s')

    ALLOCATE(lam_mm_(1:2*nsmax,0:nmmax),stat = status)
    if (status /= 0) call die_alloc(code,'lam_mm_')

    ALLOCATE(scalem_(1:2*nsmax,0:nmmax),stat = status)
    if (status /= 0) call die_alloc(code,'scalem_')

    ALLOCATE(istart_north(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_north')

    ALLOCATE(istart_south(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_south')

    call init_trig(nsmax,nmmax,trig)
    !     ------------ initiate arrays ----------------

!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,fm2,l,fl2)
    do m = 0, nmmax
       fm2 = DBLE(m) **2
       do l = m, nlmax
          fl2 = DBLE(l+1) **2
          recfac(l,m) = DSQRT( (4.0_dp * fl2 - 1.0_dp) / (fl2-fm2) )
       enddo
    enddo
!OMP END PARALLEL DO
    !     --------------------------------------------

    !     ----- set the whole alm array to zero ------
    alm = 0.0
    !     -------------------------------------------

    istart_north(0)=0
    istart_south(0)=12*nsmax**2
    do ith=1,2*nsmax
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
       else                   ! tropical band (north) + equator
          nph = 4*nsmax
       endif
       istart_north(ith)=istart_north(ith-1)+nph
       istart_south(ith)=istart_south(ith-1)-nph
    enddo

    omega_pix = pi / (3.0_dp * nsmax * nsmax)
    OVFLOW=exp(log(FL_LARGE))
    UNFLOW=exp(log(FL_SMALL))
    logOVFLOW=log(FL_LARGE)


    dth1 = 1.0_dp / (3.0_dp*DBLE(nsmax)**2)
    dth2 = 2.0_dp / (3.0_dp*DBLE(nsmax))
    dst1 = 1.0_dp / (SQRT(6.0_dp) * DBLE(nsmax) )

    !     --------------------------------------------

    !-----------------------------------------------------------------------
    !           computes the integral in phi : phas_m(theta)
    !           for each parallele from north to south pole
    !-----------------------------------------------------------------------
    lam_mm_(1:2*nsmax,0) = sq4pi_inv ! lambda_00
    scalem_(1:2*nsmax,0)=1

!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(ith,nph,kphi0,keep_it,ring,f2m,m,im_max)
    do ith = 1, 2*nsmax

       phas_n(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! North    m >= 0
       phas_s(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! South    m >= 0

       if (ith .le. nsmax-1) then      ! north polar cap
          nph = 4*ith
          kphi0 = 1 
          cth(ith) = 1.0_dp  - DBLE(ith)**2 * dth1
          sth(ith) = SIN( 2.0_dp * ASIN( ith * dst1 ) ) ! sin(theta)
       else                            ! tropical band + equat.
          nph = 4*nsmax
          kphi0 = MOD(ith+1-nsmax,2)
          cth(ith) = DBLE(2*nsmax-ith) * dth2
          sth(ith) = DSQRT((1.0_dp-cth(ith))*(1.0_dp+cth(ith))) ! sin(theta)
       endif
 
       keep_it = (ABS(cth(ith)).gt.cos_theta_cut) ! part of the sky out of the symmetric cut

       if (keep_it) then
          ring(0:nph-1) = map(istart_north(ith-1):istart_north(ith-1)+nph-1) * w8ring(ith,1)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_n(0,ith), kphi0, trig)
       endif

       if (ith .lt. 2*nsmax .and. keep_it) then
          ring(0:nph-1) = map(istart_south(ith):istart_south(ith)+nph-1) * w8ring(ith,1)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_s(0,ith), kphi0, trig)
       endif

!PPP
!PPP Large modification of the original algorithm. To obtain full speed up, it is necessary
!PPP to do first a loop on ith as above (for parallel FFTs) and then a loop on m,
!PPP which is prepared below and then computed next. For this reason, it is necessary
!PPP because of the inversion of the loops, to store all the informations in arrays
!PPP phas_n and phas_s. It seems also necessary to have temporary storage in arrays
!PPP lam_mm_ and scalem_ although some savings might be achieved in this part that I
!PPP did not fully understood (i.e. the subtle recursion on (l,m). As a result, the
!PPP parallel version as implemented is more costly in memory, factor > 2.
!PPP
       if (keep_it) then
          im_max = nmmax !16/05/1999 B. D. Wandelt -> proper delta function sampling
          do  m = 0, im_max
             f2m = 2.0_dp * m
             if (m .ge. 1) then ! lambda_0_0 for m>0
                lam_mm_(ith,m) = -lam_mm_(ith,m-1)*sth(ith)*dsqrt((f2m+1.0_dp)/f2m)
                scalem_(ith,m) = scalem_(ith,m-1)
             endif
          
             if (abs(lam_mm_(ith,m)).lt.UNFLOW) then
                lam_mm_(ith,m)=lam_mm_(ith,m)*OVFLOW
                scalem_(ith,m)=scalem_(ith,m)-1
             endif
          enddo
       endif
    enddo
!$OMP END PARALLEL DO
    
    !-----------------------------------------------------------------------
    !              computes the a_lm by integrating over theta
    !                  lambda_lm(theta) * phas_m(theta)
    !                         for each m and l
    !-----------------------------------------------------------------------
    im_max = nmmax !16/05/1999 B. D. Wandelt -> proper delta function sampling
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(lam_lm,m,par_lm,scalel,corfac,p_n,p_s, &
!$OMP         lam_0,lam_1,lam_2,a_rec,l,alm_tmp,keep_it,ith,lam_mm,scalem)
    do m = 0, im_max

       alm_tmp(m:nlmax)=0.

       do ith= 1, 2*nsmax
          keep_it = (ABS(cth(ith)).gt.cos_theta_cut) ! part of the sky out of the symmetric cut

          if (keep_it) then

          
             !           ---------- l = m ----------
             par_lm = 1.0_dp  ! = (-1)^(l+m)
             lam_mm=lam_mm_(ith,m)
             scalem=scalem_(ith,m)

             corfac = exp(scalem*logOVFLOW)   ! Actual lambda_mm 
             lam_lm = corfac*lam_mm/OVFLOW
             p_n=phas_n(m,ith)
             p_s=phas_s(m,ith)
             alm_tmp(m)=alm_tmp(m)+lam_lm * (p_n + p_s)
             
             !           ---------- l > m ----------
             lam_0 = 0.0_dp
             lam_1 = 1.0_dp
             scalel=0
             a_rec = recfac(m,m)
             lam_2 = cth(ith) * lam_1 * a_rec
             
             
             do l = m+1, nlmax
                par_lm = - par_lm  ! = (-1)^(l+m)
                
                lam_lm = lam_2*corfac*lam_mm/OVFLOW ! Remove OVFLOW-factors
                
                alm_tmp(l)=alm_tmp(l) + lam_lm*(p_n + par_lm*p_s)
                
                lam_0 = lam_1 / a_rec
                lam_1 = lam_2
                a_rec = recfac(l,m)
                lam_2 = (cth(ith) * lam_1 - lam_0) * a_rec
                
                if (abs(lam_2) .gt. OVFLOW) then
                   lam_0=lam_0/OVFLOW
                   lam_1=lam_1/OVFLOW
                   lam_2 = (cth(ith) * lam_1 - lam_0) * a_rec
                   scalel=scalel+1
                   corfac=exp((scalem+scalel)*logOVFLOW)
                elseif (abs(lam_2) .lt. UNFLOW) then
                   lam_0=lam_0*OVFLOW
                   lam_1=lam_1*OVFLOW
                   lam_2 = (cth(ith) * lam_1 - lam_0) * a_rec 
                   scalel=scalel-1
                   corfac=exp((scalem+scalel)*logOVFLOW)
                endif
             enddo !loop on l
          endif ! test on cut sky
       enddo ! loop on theta

       do l = m,nlmax
          alm(1,l,m)=alm_tmp(l)
       enddo
    enddo ! loop on m
!$OMP END PARALLEL DO

 
    !     --------------------------------------------------------------
    !     normalisation of the alm (multiplication by pixel solid angle)
    !     --------------------------------------------------------------
    !      alm = alm * omega_pix
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(l,m)
    do m = 0, nmmax       ! some compilers prefer explicit loops (EH, 09-2001)
       do l = m, nlmax
          alm(1,l,m) = alm(1,l,m) * omega_pix
       enddo
    enddo
!$OMP END PARALLEL DO

    !     --------------------
    !     free memory and exit
    !     --------------------
    DEALLOCATE(phas_n)
    DEALLOCATE(phas_s)
    DEALLOCATE(recfac)
    DEALLOCATE(istart_north,istart_south)
    DEALLOCATE(cth,sth)
    DEALLOCATE(lam_mm_,scalem_)
    DEALLOCATE(trig)
    RETURN
  END subroutine map2alm_sc

  !=======================================================================
  subroutine map2alm_sc_pre(nsmax, nlmax, nmmax, map, alm, cos_theta_cut, w8ring ,plm)
    !=======================================================================
    !     computes the a(l,m) from a map for the HEALPIX pixelisation
    !      for the Temperature field
    !     a(l,m) = int T(theta,phi) Y_lm(theta,phi)^* dtheta*dphi
    !            = int dtheta lambda_lm(theta) 
    !                  * int dphi T(theta,phi) e^(-i*m*phi)
    !
    !     where Y_lm(theta,phi) = lambda(theta) * e^(i*m*phi)
    !
    !     * Uses precomputed P_lms
    !     * the integral over phi is done by FFT
    !
    !     cos_theta_cut (>0) is the cosine of the 
    !     symmetric cut applied to the sky
    !     if it is <= 0 no cut is applied
    !
    !     NB : these a(l,m) have to be multiplied by the pixel solid angle
    !      to give the correct coefficients
    !
    !             -------------------------------
    !         precomputes the Lambda_lm recurrence factor
    !      and is therefore ~50% faster than previous versions
    !     the multiplication by omega_pix is done in the routine
    !             -------------------------------
    !
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    REAL(SP),     INTENT(IN),  DIMENSION(0:12*nsmax**2-1) :: map
    COMPLEX(SPC), INTENT(OUT), DIMENSION(1:1,0:nlmax,0:nmmax) :: alm
    REAL(DP),     INTENT(IN) :: cos_theta_cut
    REAL(DP),     INTENT(IN), DIMENSION(1:2*nsmax,1) :: w8ring
    REAL(DP), INTENT(IN), DIMENSION(0:(nlmax+1)*(nlmax+2)*nsmax-1) :: plm

    INTEGER(I4B) :: l, m, ith, im_max       ! alm related
    INTEGER(I4B) :: nph, kphi0                                             ! map related
    INTEGER(I4B), ALLOCATABLE, DIMENSION(:) :: istart_south, istart_north  ! map related
    REAL(DP) :: omega_pix
    REAL(DP) :: par_lm, dth1, dth2
    COMPLEX(DPC) :: p_n,p_s

    CHARACTER(LEN=*), PARAMETER :: code = 'MAP2ALM'
    COMPLEX(DPC), DIMENSION(:,:), ALLOCATABLE :: phas_n
    COMPLEX(DPC), DIMENSION(:,:), ALLOCATABLE :: phas_s
    INTEGER(I4B) :: status, lm, lmstep, lmstep2
    REAL(DP), DIMENSION(0:4*nsmax-1) :: ring
    LOGICAL   :: keep_it
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig   
    COMPLEX(SPC), DIMENSION(0:nlmax) :: alm_tmp
    REAL(DP), DIMENSION(:), ALLOCATABLE :: cth
    !=======================================================================

    !     --- allocates space for arrays ---


    ALLOCATE(cth(1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'cth')

    ALLOCATE(phas_n(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_n')

    ALLOCATE(phas_s(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_s')

    ALLOCATE(istart_north(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_north')

    ALLOCATE(istart_south(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_south')


    call init_trig(nsmax,nmmax,trig)

    !     ----- set the whole alm array to zero ------
    alm = 0.0
    lmstep=(nlmax+1)*(nlmax+2)/2


    !     -------------------------------------------

    istart_north(0)=0
    istart_south(0)=12*nsmax**2
    do ith=1,2*nsmax
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
       else                   ! tropical band (north) + equator
          nph = 4*nsmax
       endif
       istart_north(ith)=istart_north(ith-1)+nph
       istart_south(ith)=istart_south(ith-1)-nph
    enddo

    omega_pix = pi / (3.0_dp * nsmax * nsmax)
    dth1 = 1.0_dp / (3.0_dp*DBLE(nsmax)**2)
    dth2 = 2.0_dp / (3.0_dp*DBLE(nsmax))
    !-----------------------------------------------------------------------
    !           computes the integral in phi : phas_m(theta)
    !           for each parallele from north to south pole
    !-----------------------------------------------------------------------
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(ith,nph,kphi0,keep_it,ring, &
!$OMP         lm,im_max,m,par_lm,l)
    do ith = 1, 2*nsmax

       phas_n(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! North    m >= 0
       phas_s(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! South    m >= 0

       if (ith .le. nsmax-1) then      ! north polar cap
          nph = 4*ith
          kphi0 = 1 
          cth(ith) = 1.0_dp  - DBLE(ith)**2 * dth1
       else                            ! tropical band + equat.
          nph = 4*nsmax
          kphi0 = MOD(ith+1-nsmax,2)
          cth(ith) = DBLE(2*nsmax-ith) * dth2
       endif

       keep_it = (ABS(cth(ith)).gt.cos_theta_cut) ! part of the sky out of the symmetric cut


       if (keep_it) then ! avoid un-necessary calculations (EH, 09-2001)
          ring(0:nph-1) = map(istart_north(ith-1):istart_north(ith-1)+nph-1) * w8ring(ith,1)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_n(0,ith), kphi0, trig)
       endif

       if (ith .lt. 2*nsmax .and. keep_it) then
          ring(0:nph-1) = map(istart_south(ith):istart_south(ith)+nph-1) * w8ring(ith,1)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_s(0,ith), kphi0, trig)
       endif


    enddo
!$OMP END PARALLEL DO

    !-----------------------------------------------------------------------
    !              computes the a_lm by integrating over theta
    !                  lambda_lm(theta) * phas_m(theta)
    !                         for each m and l
    !-----------------------------------------------------------------------
    im_max = nmmax !16/05/1999 B. D. Wandelt -> proper delta function sampling
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,alm_tmp,ith,keep_it,lm,p_n,p_s,par_lm,l,lmstep2)
    do m = 0, im_max !16/05/1999 B. D. Wandelt -> proper delta function sampling
       alm_tmp(m:nlmax)=0.

       lmstep2=((2*(nlmax+1)-m+1)*m)/2

       do ith = 1, 2*nsmax
          keep_it = (ABS(cth(ith)).gt.cos_theta_cut) ! part of the sky out of the symmetric cut

          if (keep_it) then
             lm=lmstep*(ith-1)+lmstep2
             p_n=phas_n(m,ith)
             p_s=phas_s(m,ith)
             par_lm=1.0_dp
             alm_tmp(m)= alm_tmp(m) + plm(lm) * (p_n + p_s)
             lm=lm+1
             do l = m+1, nlmax
                par_lm=-par_lm
                alm_tmp(l)= alm_tmp(l) + plm(lm) * (p_n + par_lm*p_s)
                lm=lm+1
             enddo
          endif
       enddo
       do l = m,nlmax
          alm(1,l,m)=alm_tmp(l)
       enddo
    enddo
!$OMP END PARALLEL DO    

    !     --------------------------------------------------------------
    !     normalisation of the alm (multiplication by pixel solid angle)
    !     --------------------------------------------------------------
    !      alm = alm * omega_pix
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(l,m)
     do m = 0, nmmax       ! some compilers prefer explicit loops (EH, 09-2001)
       do l = m, nlmax
          alm(1,l,m) = alm(1,l,m) * omega_pix
       enddo
    enddo
!$OMP END PARALLEL DO

    !     --------------------
    !     free memory and exit
    !     --------------------
    DEALLOCATE(phas_n)
    DEALLOCATE(phas_s)
    DEALLOCATE(istart_north,istart_south)
    DEALLOCATE(cth)
    DEALLOCATE(trig)
    RETURN
  END subroutine map2alm_sc_pre

  !=======================================================================
  subroutine ring_analysis(nsmax,nlmax,nmmax,datain,nph,dataout,kphi0)
    !=======================================================================
    !     ring_analysis
    !       called by map2alm
    !       calls     fft_gpd
    !
    !     integrates (data * phi-dependence-of-Ylm) over phi
    !     --> function of m can be computed by FFT
    !     with  0<= m <= npoints/2 (: Nyquist)
    !     because the data is real the negative m are the conjugate of the 
    !     positive ones
    !=======================================================================

    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    INTEGER(I4B), INTENT(IN) :: nph, kphi0

    REAL(DP),     DIMENSION(0:nph-1), INTENT(IN)  :: datain
    COMPLEX(DPC), DIMENSION(0:nmmax), INTENT(OUT) :: dataout
    COMPLEX(DPC), DIMENSION(:,:), ALLOCATABLE , SAVE :: trig
    INTEGER(I4B) :: i,m,im_max,ksign
    REAL(DP) :: phi0
    REAL(DP), DIMENSION(0:nph-1) :: data
    INTEGER(I4B), SAVE :: trg=-1,oldnmmax=-1
    INTEGER(I4B) :: status

    !-----------------------------------------------------------------------


    if ((trg/=nsmax).or.(oldnmmax/=nmmax)) then
       if(allocated(trig)) deallocate(trig)

       ALLOCATE(trig(0:max(2*nsmax,nmmax),0:nsmax),stat = status) 
       if (status /= 0) call die_alloc('ring_analysis','trig')


       trig(:,:)=CMPLX(1.0_dp,0.0_dp, KIND=DP)
       do i=1,nsmax
          phi0=PI/DBLE(i*4)
          do m=0,max(2*nsmax,nmmax)
             trig(m,i)= CMPLX( DCOS(m*phi0), DSIN(m*phi0), kind=DP)
          enddo
       enddo
       trg=nsmax
       oldnmmax=nmmax

    endif

    !-----------------------------------------------------------------------


    ksign = - 1
    data=0.
    data(0:nph-1)=datain(0:nph-1)

    call real_fft(data, backward=.false.)

    im_max = MIN(nph/2,nmmax)
    dataout(0)=CMPLX(data(0),0.0_dp,kind=DP)

    do i = 1, im_max*2-3, 2
       dataout((i+1)/2) = CMPLX( data(i), data(i+1),kind= DP) 
    enddo

    if(im_max==nph/2) then
       dataout(im_max)= CMPLX( data(nph-1),0.0_dp,kind=DP)
    else
       dataout(im_max)= CMPLX( data(2*im_max-1),data(2*im_max),kind=DP)
    endif

    if(im_max==nmmax) goto 1000

    do i =  im_max+1,min(nph,nmmax)
       dataout(i) = conjg(dataout(2*im_max-i) )
    end do

    if(min(nph,nmmax)==nmmax) goto 1000

    do i =  2*im_max+1,nmmax
       dataout(i) = dataout(mod(i,2*im_max)) 
    end do

1000 continue

    if(kphi0==1)then
       do i =  0,nmmax
          m = ksign*i
          dataout(i)=dataout(i)* CONJG(trig(-m,nph/4))
       enddo
    end if

    RETURN
  END subroutine ring_analysis

  !=======================================================================
  subroutine ring_analysis2(nsmax,nlmax,nmmax,datain,nph,dataout,kphi0,trig)
    !=======================================================================
    !     ring_analysis
    !       called by map2alm
    !       calls     fft_gpd
    !
    !     integrates (data * phi-dependence-of-Ylm) over phi
    !     --> function of m can be computed by FFT
    !     with  0<= m <= npoints/2 (: Nyquist)
    !     because the data is real the negative m are the conjugate of the 
    !     positive ones
    !=======================================================================

    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    INTEGER(I4B), INTENT(IN) :: nph, kphi0
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig

    REAL(DP),     DIMENSION(0:nph-1), INTENT(IN)  :: datain
    COMPLEX(DPC), DIMENSION(0:nmmax), INTENT(OUT) :: dataout
    INTEGER(I4B) :: i,m,im_max,ksign
    REAL(DP), DIMENSION(0:nph-1) :: data
    INTEGER(I4B) :: status

    !-----------------------------------------------------------------------



    !-----------------------------------------------------------------------


    ksign = - 1
    data=0.
    data(0:nph-1)=datain(0:nph-1)

    call real_fft(data, backward=.false.)

    im_max = MIN(nph/2,nmmax)
    dataout(0)=CMPLX(data(0),0.0_dp,kind=DP)

    do i = 1, im_max*2-3, 2
       dataout((i+1)/2) = CMPLX( data(i), data(i+1),kind= DP) 
    enddo

    if(im_max==nph/2) then
       dataout(im_max)= CMPLX( data(nph-1),0.0_dp,kind=DP)
    else
       dataout(im_max)= CMPLX( data(2*im_max-1),data(2*im_max),kind=DP)
    endif

    if(im_max==nmmax) goto 1000

    do i =  im_max+1,min(nph,nmmax)
       dataout(i) = conjg(dataout(2*im_max-i) )
    end do

    if(min(nph,nmmax)==nmmax) goto 1000

    do i =  2*im_max+1,nmmax
       dataout(i) = dataout(mod(i,2*im_max)) 
    end do

1000 continue

    if(kphi0==1)then
       do i =  0,nmmax
          m = ksign*i
          dataout(i)=dataout(i)* CONJG(trig(-m,nph/4))
       enddo
    end if

    RETURN
  END subroutine ring_analysis2

  !=======================================================================
  subroutine map2alm_pol(nsmax, nlmax, nmmax, map_TQU,&
       &                     alm_TGC, cos_theta_cut, w8ring_TQU)
    !=======================================================================
    !     computes the a(l,m) from a map for the HEALPIX pixelisation
    !     a(l,m) = int T(theta,phi) Y_lm(theta,phi)^* dtheta*dphi
    !            = int dtheta lambda_lm(theta) 
    !                  * int dphi T(theta,phi) e^(-i*m*phi)
    !
    !     where Y_lm(theta,phi) = lambda(theta) * e^(i*m*phi)
    !
    !     * the recurrence of Ylm is the standard one (cf Num Rec)
    !     * the integral over phi is done by FFT
    !
    !     cos_theta_cut (>0) is the cosine of the 
    !     symmetric cut applied to the sky
    !     if it is <= 0 no cut is applied
    !
    !     NB : these a(l,m) have to be multiplied by the pixel solid angle
    !      to give the correct coefficients
    !
    !           -------------------------------
    !     precomputes the Lambda_lm recurrence factor
    !
    !      the multiplication by omega_pix is done in the routine
    !           -------------------------------
    !
    !      This routine also include polarisation, map_TQU and alm_TGC
    !      are excpected to have an extra dimension for tensors.
    !
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    REAL(SP),     INTENT(IN),  DIMENSION(0:12*nsmax**2-1,1:3) :: map_TQU
    COMPLEX(SPC), INTENT(OUT), DIMENSION(1:3,0:nlmax,0:nmmax) :: alm_TGC
    REAL(DP),     INTENT(IN) :: cos_theta_cut
    REAL(DP),     INTENT(IN), DIMENSION(1:2*nsmax,1:3) :: w8ring_TQU

    INTEGER(I4B) :: l, m, ith, im_max, scalem, scalel       ! alm related
    INTEGER(I4B) :: nph, kphi0                                             ! map related
    INTEGER(I4B), ALLOCATABLE, DIMENSION(:) :: istart_south, istart_north  ! map related

    REAL(DP) :: omega_pix
    REAL(DP) :: dth1, dth2, dst1, OVFLOW, UNFLOW, logOVFLOW
    REAL(DP) :: a_rec, lam_mm, lam_lm, lam_lm1m, lam_0, lam_1, lam_2, par_lm
    REAL(DP) :: fm, f2m, fm2, fl, fl2, corfac
    REAL(DP) :: c_on_s2, fm_on_s2, one_on_s2
    REAL(DP) :: lambda_w, lambda_x, a_w, b_w, a_x
    COMPLEX(DPC) :: zi_lam_x
    COMPLEX(DPC) :: p_nT,p_sT,p_nQ,p_sQ,p_nU,p_sU

    CHARACTER(LEN=*), PARAMETER :: code = 'MAP2ALM_POL'
    COMPLEX(DPC), DIMENSION(:,:), ALLOCATABLE :: phas_nT, phas_nQ, phas_nU
    COMPLEX(DPC), DIMENSION(:,:), ALLOCATABLE :: phas_sT, phas_sQ, phas_sU
    INTEGER(I4B) status


    REAL(DP), DIMENSION(:,:), ALLOCATABLE :: recfac
    REAL(DP), DIMENSION(:,:), ALLOCATABLE :: lam_fact
    REAL(DP), DIMENSION(0:4*nsmax-1) :: ring
    REAL(DP), DIMENSION(:),   ALLOCATABLE :: normal_l
    LOGICAL   :: keep_it
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig   
    INTEGER(I4B), DIMENSION(:,:), ALLOCATABLE :: scalem_
    REAL(DP), DIMENSION(:,:), ALLOCATABLE :: lam_mm_
    COMPLEX(SPC), DIMENSION(3,0:nlmax) :: alm_tmp
    REAL(DP), DIMENSION(:), ALLOCATABLE :: cth,sth

    !=======================================================================

    !     --- allocates space for arrays ---
    ALLOCATE(recfac(0:nlmax,0:nmmax),stat = status)    
    if (status /= 0) call die_alloc(code,'recfac')

    ALLOCATE(lam_fact(0:nlmax,0:nmmax),stat = status)    
    if (status /= 0) call die_alloc(code,'lam_fact')

    ALLOCATE(normal_l(0:nlmax),stat = status)    
    if (status /= 0) call die_alloc(code,'normal_l')


    ALLOCATE(cth(1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'cth')

    ALLOCATE(sth(1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'sth')

    ALLOCATE(phas_nT(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_n')

    ALLOCATE(phas_sT(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_s')

    ALLOCATE(phas_nQ(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_n')

    ALLOCATE(phas_sQ(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_s')

    ALLOCATE(phas_nU(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_n')

    ALLOCATE(phas_sU(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_s')

    ALLOCATE(lam_mm_(1:2*nsmax,0:nmmax),stat = status)
    if (status /= 0) call die_alloc(code,'lam_mm_')

    ALLOCATE(scalem_(1:2*nsmax,0:nmmax),stat = status)
    if (status /= 0) call die_alloc(code,'scalem_')

    ALLOCATE(istart_north(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_north')

    ALLOCATE(istart_south(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_south')


    call init_trig(nsmax,nmmax,trig)
 
    !     ------------ initiate arrays ----------------

!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,fm2,l,fl2)
    do m = 0, nmmax
       fm2 = DBLE(m) **2
       do l = m, nlmax
          fl2 = DBLE(l+1) **2
          recfac(l,m) = DSQRT( (4.0_dp * fl2 - 1.0_dp) / (fl2-fm2) )
       enddo
    enddo
!OMP END PARALLEL DO

    lam_fact(:,:) = 0.0_dp
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,fm2,l,fl,fl2)
    do l = 2, nlmax
       fl  = DBLE(l)
       fl2 = fl * fl
       do m = 0, min(l, nmmax)  ! corrected 2001-12
          fm2 = DBLE(m) **2
          lam_fact(l,m) = 2.0_dp * SQRT( (2.0_dp * fl + 1.0_dp) / (2.0_dp * fl - 1.0_dp) * (fl2-fm2) )
       enddo
    enddo
!$OMP END PARALLEL DO

    normal_l(0:1) = 0.0_dp
    do l = 2, nlmax
       fl = DBLE(l)
       normal_l(l) = SQRT( KvS / ((fl+2.0_dp)*(fl+1.0_dp)*fl*(fl-1.0_dp)) ) 
       ! either CMBFAST (KvS=1) or Kamionkowski et al. (KvS=2) definition
    enddo

    !     ----- set the whole alm array to zero ------
    alm_TGC = CMPLX(0.,0.)
    !     -------------------------------------------

    istart_north(0)=0
    istart_south(0)=12*nsmax**2
    do ith=1,2*nsmax
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
       else                    ! tropical band (north) + equator
          nph = 4*nsmax
       endif
       istart_north(ith)=istart_north(ith-1)+nph
       istart_south(ith)=istart_south(ith-1)-nph
    enddo
    omega_pix = pi / (3.0_dp * nsmax * nsmax)
    OVFLOW=exp(log(FL_LARGE))
    UNFLOW=exp(log(FL_SMALL))
    logOVFLOW=log(FL_LARGE)

    dth1 = 1.0_dp / (3.0_dp*DBLE(nsmax)**2)
    dth2 = 2.0_dp / (3.0_dp*DBLE(nsmax))
    dst1 = 1.0_dp / (SQRT(6.0_dp) * DBLE(nsmax) )

    !     --------------------------------------------
    !-----------------------------------------------------------------------
    !           computes the integral in phi : phas_m(theta)
    !           for each parallele from north to south pole
    !-----------------------------------------------------------------------
    lam_mm_(1:2*nsmax,0) = sq4pi_inv ! lambda_00
    scalem_(1:2*nsmax,0)=1

!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(ith,nph,kphi0,keep_it,ring,m,f2m,im_max)
    do ith = 1, 2*nsmax

       phas_nT(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! North    m >= 0
       phas_sT(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! South    m >= 0
       phas_nQ(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! North    m >= 0
       phas_sQ(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! South    m >= 0
       phas_nU(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! North    m >= 0
       phas_sU(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! South    m >= 0

       if (ith .le. nsmax-1) then      ! north polar cap
          nph = 4*ith
          kphi0 = 1 
          cth(ith) = 1.0_dp  - DBLE(ith)**2 * dth1
          sth(ith) = SIN( 2.0_dp * ASIN( ith * dst1 ) ) ! sin(theta)
       else                            ! tropical band + equat.
          nph = 4*nsmax
          kphi0 = MOD(ith+1-nsmax,2)
          cth(ith) = DBLE(2*nsmax-ith) * dth2
          sth(ith) = DSQRT((1.0_dp-cth(ith))*(1.0_dp+cth(ith))) ! sin(theta)
       endif

       keep_it = (ABS(cth(ith)).gt.cos_theta_cut) ! part of the sky out of the symmetric cut
       if (keep_it) then
          ring(0:nph-1) = map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,1) * w8ring_TQU(ith,1)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_nT(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,2) * w8ring_TQU(ith,2)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_nQ(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,3) * w8ring_TQU(ith,3)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_nU(0,ith), kphi0, trig)
       endif

       if (ith .lt. 2*nsmax .and. keep_it) then
          ring(0:nph-1) = map_TQU(istart_south(ith):istart_south(ith)+nph-1,1) * w8ring_TQU(ith,1)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_sT(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_south(ith):istart_south(ith)+nph-1,2) * w8ring_TQU(ith,2)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_sQ(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_south(ith):istart_south(ith)+nph-1,3) * w8ring_TQU(ith,3)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_sU(0,ith), kphi0, trig)
       endif

       if (keep_it) then
          im_max = nmmax !16/05/1999 B. D. Wandelt -> proper delta function sampling
          do  m = 0, im_max
             f2m = 2.0_dp * m
             if (m .ge. 1) then ! lambda_0_0 for m>0
                lam_mm_(ith,m) = -lam_mm_(ith,m-1)*sth(ith)*dsqrt((f2m+1.0_dp)/f2m)
                scalem_(ith,m) = scalem_(ith,m-1)
             endif
          
             if (abs(lam_mm_(ith,m)).lt.UNFLOW) then
                lam_mm_(ith,m)=lam_mm_(ith,m)*OVFLOW
                scalem_(ith,m)=scalem_(ith,m)-1
             endif
          enddo
       endif
    enddo
!$OMP END PARALLEL DO


    !-----------------------------------------------------------------------
    !              computes the a_lm by integrating over theta
    !                  lambda_lm(theta) * phas_m(theta)
    !                         for each m and l
    !-----------------------------------------------------------------------
    im_max = nmmax !16/05/1999 B. D. Wandelt -> proper delta function sampling
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,ith,l,one_on_s2,c_on_s2,lam_mm,scalem, &
!$OMP         fm,f2m,fm2,fm_on_s2,par_lm,corfac, &
!$OMP         lam_lm,lambda_w,lambda_x,zi_lam_x,lam_0,lam_1,scalel,a_rec, &
!$OMP         lam_2,lam_lm1m,fl,fl2,a_w,b_w,a_x,keep_it,alm_tmp, &
!$OMP         p_nT,p_sT,p_nQ,p_sQ,p_nU,p_sU)
    do m = 0, im_max
    
       alm_tmp(1:3,m:nlmax)=0.
       fm  = DBLE(m)
       f2m = 2.0_dp * fm
       fm2 = fm * fm

       do ith= 1, 2*nsmax
          keep_it = (ABS(cth(ith)).gt.cos_theta_cut) ! part of the sky out of the symmetric cut
       
          if (keep_it) then ! avoid un-necessary calculations (EH, 09-2001)
             one_on_s2 = 1.0_dp / sth(ith)**2 ! 1/sin^2
             c_on_s2 = cth(ith) * one_on_s2

             fm_on_s2 = fm * one_on_s2

             !           ---------- l = m ----------
             par_lm = 1.0_dp  ! = (-1)^(l+m)
             lam_mm = lam_mm_(ith,m)
             scalem=scalem_(ith,m)
 
             corfac=exp(scalem*logOVFLOW)  ! Actual lamda_mm

             lam_lm = corfac*lam_mm/OVFLOW

             lambda_w = ( normal_l(m) * lam_lm * (fm - fm2) ) * ( 2.0_dp * one_on_s2 - 1.0_dp ) * SW1
             lambda_x = ( normal_l(m) * lam_lm * (fm - fm2) ) *   2.0_dp * one_on_s2 * cth(ith)
             zi_lam_x = CMPLX(0.0_dp, lambda_x, KIND=DP)

             p_nT=phas_nT(m,ith)
             p_sT=phas_sT(m,ith)
             p_nQ=phas_nQ(m,ith) * iKvS * sgQ
             p_sQ=phas_sQ(m,ith) * iKvS * sgQ
             p_nU=phas_nU(m,ith) * iKvS
             p_sU=phas_sU(m,ith) * iKvS

             !Temperature
             alm_tmp(1,m)=alm_tmp(1,m)+lam_lm * (p_nT + p_sT)

             !Gradient (Electric) part of the polarisation tensor
             alm_tmp(2,m)=alm_tmp(2,m) &
                  &                 + lambda_w * (p_nQ + p_sQ) &
                  &                 - zi_lam_x * (p_nU - p_sU)

             !Curl     (Magnetic) part of the polarisation tensor
             alm_tmp(3,m)=alm_tmp(3,m) &
                  &                  - lambda_w * (p_nU + p_sU) &
                  &                  - zi_lam_x * (p_nQ - p_sQ)

             !           ---------- l > m ----------
             lam_0 = 0.0_dp
             lam_1 = 1.0_dp
             scalel=0
             a_rec = recfac(m,m)
             lam_2 = cth(ith) * lam_1 * a_rec
             do l = m+1, nlmax
                par_lm = - par_lm  ! = (-1)^(l+m)
                lam_lm1m=lam_lm ! actual lambda_l-1,m (useful for polarisation)
                lam_lm   = lam_2*corfac*lam_mm/OVFLOW ! actual lambda_lm (OVFLOW factors removed)
                fl  = DBLE(l)
                fl2 = fl * fl

                ! Temperature
                alm_tmp(1,l)=alm_tmp(1,l) + lam_lm * (p_nT + par_lm*p_sT)

                ! polarisation related quantities
                a_w =  2.0_dp * (fl - fm2) * one_on_s2 + (fl2 - fl)
                b_w =  c_on_s2 * lam_fact(l,m)
                a_x =  2.0_dp * cth(ith) * (fl-1.0_dp) * lam_lm
                lambda_w =   normal_l(l) * ( a_w * lam_lm - b_w * lam_lm1m ) * SW1
                lambda_x = - normal_l(l) * fm_on_s2 * ( a_x - lam_fact(l,m) * lam_lm1m )
                zi_lam_x = CMPLX(0.0_dp, lambda_x, KIND=DP)

                ! G =  ( ylm_W * Q - ylm_X * U ) ! (ylm_X is imaginary) ! new CMBFAST normalisation
                alm_tmp(2,l)=alm_tmp(2,l) &
                     &              + lambda_w * (p_nQ + par_lm*p_sQ) &
                     &              - zi_lam_x * (p_nU - par_lm*p_sU)


                ! C = -( ylm_W * U + ylm_X * Q ) ! (ylm_X is imaginary) ! new CMBFAST normalisation
                alm_tmp(3,l)=alm_tmp(3,l) &
                     &              - lambda_w * (p_nU + par_lm*p_sU) &
                     &              - zi_lam_x * (p_nQ - par_lm*p_sQ)

                lam_0 = lam_1 / a_rec
                lam_1 = lam_2
                a_rec = recfac(l,m)
                lam_2 = (cth(ith) * lam_1 - lam_0) * a_rec

                if (abs(lam_2) .gt. OVFLOW) then
                   lam_0=lam_0/OVFLOW
                   lam_1=lam_1/OVFLOW
                   lam_2 = (cth(ith) * lam_1 - lam_0) * a_rec
                   scalel=scalel+1
                   corfac=exp((scalem+scalel)*logOVFLOW)
                elseif (abs(lam_2) .lt. UNFLOW) then
                   lam_0=lam_0*OVFLOW
                   lam_1=lam_1*OVFLOW
                   lam_2 = (cth(ith) * lam_1 - lam_0) * a_rec 
                   scalel=scalel-1
                   corfac=exp((scalem+scalel)*logOVFLOW)
                endif

             enddo ! loop on l
          endif ! test on cut sky
       enddo ! loop on theta
       do l = m,nlmax
          alm_TGC(1,l,m)=alm_tmp(1,l)
          alm_TGC(2,l,m)=alm_tmp(2,l)
          alm_TGC(3,l,m)=alm_tmp(3,l)
       enddo
    enddo ! loop on m
!$OMP END PARALLEL DO

    !     --------------------------------------------------------------
    !     normalisation of the alm (multiplication by pixel solid angle)
    !     --------------------------------------------------------------
    !      alm_TGC = alm_TGC * omega_pix
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(l,m)
    do m = 0, nmmax       ! some compilers prefer explicit loops (EH, 09-2001)
       do l = m, nlmax
          alm_TGC(1,l,m) = alm_TGC(1,l,m) * omega_pix
          alm_TGC(2,l,m) = alm_TGC(2,l,m) * omega_pix
          alm_TGC(3,l,m) = alm_TGC(3,l,m) * omega_pix
       enddo
    enddo
!$OMP END PARALLEL DO

    !     --------------------
    !     free memory and exit
    !     --------------------
    DEALLOCATE(recfac)
    DEALLOCATE(lam_fact)
    DEALLOCATE(normal_l)
    DEALLOCATE(phas_nT,phas_nQ,phas_nU)
    DEALLOCATE(phas_sT,phas_sQ,phas_sU)
    DEALLOCATE(istart_north,istart_south)
    DEALLOCATE(cth,sth)
    DEALLOCATE(lam_mm_,scalem_)
    DEALLOCATE(trig)

    RETURN
  END subroutine map2alm_pol
  !=======================================================================
  subroutine map2alm_pol_pre1(nsmax, nlmax, nmmax, map_TQU,&
       &                     alm_TGC, cos_theta_cut, w8ring_TQU,plm)
    !=======================================================================
    !     computes the a(l,m) from a map for the HEALPIX pixelisation
    !     a(l,m) = int T(theta,phi) Y_lm(theta,phi)^* dtheta*dphi
    !            = int dtheta lambda_lm(theta) 
    !                  * int dphi T(theta,phi) e^(-i*m*phi)
    !
    !     where Y_lm(theta,phi) = lambda(theta) * e^(i*m*phi)
    !
    !     * Uses precomputed P_lms
    !     * the integral over phi is done by FFT
    !
    !     cos_theta_cut (>0) is the cosine of the 
    !     symmetric cut applied to the sky
    !     if it is <= 0 no cut is applied
    !
    !     NB : these a(l,m) have to be multiplied by the pixel solid angle
    !      to give the correct coefficients
    !
    !           -------------------------------
    !     precomputes the Lambda_lm recurrence factor
    !
    !      the multiplication by omega_pix is done in the routine
    !           -------------------------------
    !
    !      This routine also include polarisation, map_TQU and alm_TGC
    !      are excpected to have an extra dimension for tensors.
    !
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    REAL(SP),     INTENT(IN),  DIMENSION(0:12*nsmax**2-1,1:3) :: map_TQU
    COMPLEX(SPC), INTENT(OUT), DIMENSION(1:3,0:nlmax,0:nmmax) :: alm_TGC
    REAL(DP), INTENT(IN), DIMENSION(0:(nlmax+1)*(nlmax+2)*nsmax-1) :: plm
    REAL(DP),     INTENT(IN) :: cos_theta_cut
    REAL(DP),     INTENT(IN), DIMENSION(1:2*nsmax,1:3) :: w8ring_TQU

    INTEGER(I4B) :: l, m, ith, im_max       ! alm related
    INTEGER(I4B) :: nph, kphi0                                             ! map related
    INTEGER(I4B), ALLOCATABLE, DIMENSION(:) :: istart_south, istart_north  ! map related

    REAL(DP) :: omega_pix
    REAL(DP) :: dth1, dth2, dst1
    REAL(DP) :: par_lm
    REAL(DP) :: fm, f2m, fm2, fl, fl2
    REAL(DP) :: c_on_s2, fm_on_s2, one_on_s2
    REAL(DP) :: lambda_w, lambda_x, a_w, b_w, a_x
    COMPLEX(DPC) :: zi_lam_x
    COMPLEX(DPC) :: p_nT,p_sT,p_nQ,p_sQ,p_nU,p_sU

    CHARACTER(LEN=*), PARAMETER :: code = 'MAP2ALM_POL'
    COMPLEX(DPC), DIMENSION(:,:), ALLOCATABLE :: phas_nT, phas_nQ, phas_nU
    COMPLEX(DPC), DIMENSION(:,:), ALLOCATABLE :: phas_sT, phas_sQ, phas_sU
    INTEGER(I4B) status, lm, lmstep, lmstep2

    REAL(DP), DIMENSION(:,:), ALLOCATABLE :: lam_fact
    REAL(DP), DIMENSION(0:4*nsmax-1) :: ring
    REAL(DP), DIMENSION(:),   ALLOCATABLE :: normal_l
    LOGICAL   :: keep_it
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig   
    COMPLEX(SPC), DIMENSION(3,0:nlmax) :: alm_tmp
    REAL(DP), DIMENSION(:), ALLOCATABLE :: cth,sth
    !=======================================================================

    !     --- allocates space for arrays ---

    ALLOCATE(lam_fact(0:nlmax,0:nmmax),stat = status)    
    if (status /= 0) call die_alloc(code,'lam_fact')

    ALLOCATE(normal_l(0:nlmax),stat = status)    
    if (status /= 0) call die_alloc(code,'normal_l')


    ALLOCATE(cth(1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'cth')

    ALLOCATE(sth(1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'sth')

    ALLOCATE(phas_nT(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_n')

    ALLOCATE(phas_sT(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_s')

    ALLOCATE(phas_nQ(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_n')

    ALLOCATE(phas_sQ(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_s')

    ALLOCATE(phas_nU(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_n')

    ALLOCATE(phas_sU(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_s')

    ALLOCATE(istart_north(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_north')

    ALLOCATE(istart_south(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_south')


    call init_trig(nsmax,nmmax,trig)

    !     ------------ initiate arrays ----------------

    lam_fact(:,:) = 0.0_dp
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,fm2,l,fl,fl2)
    do l = 2, nlmax
       fl  = DBLE(l)
       fl2 = fl * fl
       do m = 0, min(l, nmmax)  ! corrected 2001-12
          fm2 = DBLE(m) **2
          lam_fact(l,m) = 2.0_dp * SQRT( (2.0_dp * fl + 1.0_dp) / (2.0_dp * fl - 1.0_dp) * (fl2-fm2) )
       enddo
    enddo
!$OMP END PARALLEL DO

    normal_l(0:1) = 0.0_dp
    do l = 2, nlmax
       fl = DBLE(l)
       normal_l(l) = SQRT( KvS / ((fl+2.0_dp)*(fl+1.0_dp)*fl*(fl-1.0_dp)) ) 
       ! either CMBFAST (KvS=1) or Kamionkowski et al. (KvS=2) definition
    enddo

    !     ----- set the whole alm array to zero ------
    alm_TGC = CMPLX(0.,0.)
    lmstep=(nlmax+1)*(nlmax+2)/2
    !     -------------------------------------------

    istart_north(0)=0
    istart_south(0)=12*nsmax**2
    do ith=1,2*nsmax
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
       else                    ! tropical band (north) + equator
          nph = 4*nsmax
       endif
       istart_north(ith)=istart_north(ith-1)+nph
       istart_south(ith)=istart_south(ith-1)-nph
    enddo
    omega_pix = pi / (3.0_dp * nsmax * nsmax)

    dth1 = 1.0_dp / (3.0_dp*DBLE(nsmax)**2)
    dth2 = 2.0_dp / (3.0_dp*DBLE(nsmax))
    dst1 = 1.0_dp / (SQRT(6.0_dp) * DBLE(nsmax) )

    !     --------------------------------------------
    !-----------------------------------------------------------------------
    !           computes the integral in phi : phas_m(theta)
    !           for each parallele from north to south pole
    !-----------------------------------------------------------------------

!$OMP PARALLEL DO SCHEDULE(STATIC, 1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(ith,nph,kphi0,keep_it,ring)
    do ith = 1, 2*nsmax

       phas_nT(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! North    m >= 0
       phas_sT(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! South    m >= 0
       phas_nQ(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! North    m >= 0
       phas_sQ(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! South    m >= 0
       phas_nU(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! North    m >= 0
       phas_sU(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! South    m >= 0

       if (ith .le. nsmax-1) then      ! north polar cap
          nph = 4*ith
          kphi0 = 1 
          cth(ith) = 1.0_dp  - DBLE(ith)**2 * dth1
          sth(ith) = SIN( 2.0_dp * ASIN( ith * dst1 ) ) ! sin(theta)
       else                            ! tropical band + equat.
          nph = 4*nsmax
          kphi0 = MOD(ith+1-nsmax,2)
          cth(ith) = DBLE(2*nsmax-ith) * dth2
          sth(ith) = DSQRT((1.0_dp-cth(ith))*(1.0_dp+cth(ith))) ! sin(theta)
       endif

       keep_it = (ABS(cth(ith)).gt.cos_theta_cut) ! part of the sky out of the symmetric cut
       if (keep_it) then
          ring(0:nph-1) = map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,1) * w8ring_TQU(ith,1)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_nT(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,2) * w8ring_TQU(ith,2)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_nQ(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,3) * w8ring_TQU(ith,3)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_nU(0,ith), kphi0, trig)
       endif

       if (ith .lt. 2*nsmax .and. keep_it) then
          ring(0:nph-1) = map_TQU(istart_south(ith):istart_south(ith)+nph-1,1) * w8ring_TQU(ith,1)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_sT(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_south(ith):istart_south(ith)+nph-1,2) * w8ring_TQU(ith,2)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_sQ(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_south(ith):istart_south(ith)+nph-1,3) * w8ring_TQU(ith,3)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_sU(0,ith), kphi0, trig)
       endif
   enddo ! loop on theta
!$OMP END PARALLEL DO

       !-----------------------------------------------------------------------
       !              computes the a_lm by integrating over theta
       !                  lambda_lm(theta) * phas_m(theta)
       !                         for each m and l
       !-----------------------------------------------------------------------
   im_max = nmmax !16/05/1999 B. D. Wandelt -> proper delta function sampling
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,alm_tmp,fm,f2m,fm2,lmstep2,ith,keep_it,lm,one_on_s2,c_on_s2, &
!$OMP         fm_on_s2,par_lm,lambda_w,lambda_x,zi_lam_x,p_nT,p_sT,p_nQ,p_sQ, &
!$OMP         p_nU,p_sU,l,fl,fl2,a_w,b_w,a_x)
    do m = 0, im_max !16/05/1999 B. D. Wandelt -> proper delta function sampling
       alm_tmp(1:3,m:nlmax)=0.
   
       fm  = DBLE(m)
       f2m = 2.0_dp * fm
       fm2 = fm * fm
       lmstep2=((2*(nlmax+1)-m+1)*m)/2

       do ith = 1, 2*nsmax
          keep_it = (ABS(cth(ith)).gt.cos_theta_cut) ! part of the sky out of the symmetric cut

          if (keep_it) then ! avoid un-necessary calculations (EH, 09-2001)
             lm=lmstep*(ith-1)+lmstep2

             one_on_s2 = 1.0_dp / sth(ith)**2 ! 1/sin^2
             c_on_s2 = cth(ith) * one_on_s2
             fm_on_s2 = fm * one_on_s2

             !           ---------- l = m ----------
             par_lm = 1.0_dp  ! = (-1)^(l+m)
             lambda_w = ( normal_l(m) * plm(lm) * (fm - fm2) ) * ( 2.0_dp * one_on_s2 - 1.0_dp ) * SW1
             lambda_x = ( normal_l(m) * plm(lm) * (fm - fm2) ) *   2.0_dp * one_on_s2 * cth(ith)
             zi_lam_x = CMPLX(0.0_dp, lambda_x, KIND=DP)

             p_nT=phas_nT(m,ith)
             p_sT=phas_sT(m,ith)
             p_nQ=phas_nQ(m,ith) * iKvS * sgQ
             p_sQ=phas_sQ(m,ith) * iKvS * sgQ
             p_nU=phas_nU(m,ith) * iKvS
             p_sU=phas_sU(m,ith) * iKvS

             !Temperature
             alm_tmp(1,m) = alm_tmp(1,m) + plm(lm) * (p_nT + p_sT)

             !Gradient (Electric) part of the polarisation tensor
             alm_tmp(2,m) = alm_tmp(2,m) &
                  &                 + lambda_w * (p_nQ + p_sQ) &
                  &                 - zi_lam_x * (p_nU - p_sU)

             !Curl     (Magnetic) part of the polarisation tensor
             alm_tmp(3,m) = alm_tmp(3,m) &
                  &                  - lambda_w * (p_nU + p_sU) &
                  &                  - zi_lam_x * (p_nQ - p_sQ)

             lm=lm+1
             !           ---------- l > m ----------

             do l = m+1, nlmax
                par_lm = - par_lm  ! = (-1)^(l+m)
                fl  = DBLE(l)
                fl2 = fl * fl

                ! Temperature
                alm_tmp(1,l) = alm_tmp(1,l) + plm(lm) * (p_nT + par_lm*p_sT)

                ! polarisation related quantities
                a_w =  2.0_dp * (fl - fm2) * one_on_s2 + (fl2 - fl)
                b_w =  c_on_s2 * lam_fact(l,m)
                a_x =  2.0_dp * cth(ith) * (fl-1.0_dp) * plm(lm)
                lambda_w =   normal_l(l) * ( a_w * plm(lm) - b_w * plm(lm-1) ) * SW1
                lambda_x = - normal_l(l) * fm_on_s2 * ( a_x - lam_fact(l,m) * plm(lm-1) )
                zi_lam_x = CMPLX(0.0_dp, lambda_x, KIND=DP)

                ! G =  ( ylm_W * Q - ylm_X * U ) ! (ylm_X is imaginary) ! new CMBFAST normalisation
                alm_tmp(2,l) = alm_tmp(2,l) &
                     &              + lambda_w * (p_nQ + par_lm*p_sQ) &
                     &              - zi_lam_x * (p_nU - par_lm*p_sU)

                ! C = -( ylm_W * U + ylm_X * Q ) ! (ylm_X is imaginary) ! new CMBFAST normalisation
                alm_tmp(3,l) = alm_tmp(3,l) &
                     &              - lambda_w * (p_nU + par_lm*p_sU) &
                     &              - zi_lam_x * (p_nQ - par_lm*p_sQ)

                lm=lm+1

             enddo ! loop on l
          endif ! test on cut sky
       enddo ! loop on theta
       do l = m,nlmax
          alm_TGC(1,l,m)=alm_tmp(1,l)
          alm_TGC(2,l,m)=alm_tmp(2,l)
          alm_TGC(3,l,m)=alm_tmp(3,l)
       enddo
    enddo ! loop on m

    !     --------------------------------------------------------------
    !     normalisation of the alm (multiplication by pixel solid angle)
    !     --------------------------------------------------------------
    !      alm_TGC = alm_TGC * omega_pix
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(l,m)
    do m = 0, nmmax       ! some compilers prefer explicit loops (EH, 09-2001)
       do l = m, nlmax
          alm_TGC(1,l,m) = alm_TGC(1,l,m) * omega_pix
          alm_TGC(2,l,m) = alm_TGC(2,l,m) * omega_pix
          alm_TGC(3,l,m) = alm_TGC(3,l,m) * omega_pix
       enddo
    enddo
!$OMP END PARALLEL DO

    !     --------------------
    !     free memory and exit
    !     --------------------
    DEALLOCATE(lam_fact)
    DEALLOCATE(normal_l)
    DEALLOCATE(phas_nT,phas_nQ,phas_nU)
    DEALLOCATE(phas_sT,phas_sQ,phas_sU)
    DEALLOCATE(istart_north,istart_south)
    DEALLOCATE(cth,sth)
    DEALLOCATE(trig)

    RETURN
  END subroutine map2alm_pol_pre1

  !=======================================================================
  subroutine map2alm_pol_pre2(nsmax, nlmax, nmmax, map_TQU,&
       &                     alm_TGC, cos_theta_cut, w8ring_TQU,plm)
    !=======================================================================
    !     computes the a(l,m) from a map for the HEALPIX pixelisation
    !     a(l,m) = int T(theta,phi) Y_lm(theta,phi)^* dtheta*dphi
    !            = int dtheta lambda_lm(theta) 
    !                  * int dphi T(theta,phi) e^(-i*m*phi)
    !
    !     where Y_lm(theta,phi) = lambda(theta) * e^(i*m*phi)
    !
    !     * Uses precomputed P_lms AND tensor harmonics
    !     * the integral over phi is done by FFT
    !
    !     cos_theta_cut (>0) is the cosine of the 
    !     symmetric cut applied to the sky
    !     if it is <= 0 no cut is applied
    !
    !     NB : these a(l,m) have to be multiplied by the pixel solid angle
    !      to give the correct coefficients
    !
    !           -------------------------------
    !     precomputes the Lambda_lm recurrence factor
    !
    !      the multiplication by omega_pix is done in the routine
    !           -------------------------------
    !
    !      This routine also include polarisation, map_TQU and alm_TGC
    !      are excpected to have an extra dimension for tensors.
    !
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax
    REAL(SP),     INTENT(IN),  DIMENSION(0:12*nsmax**2-1,1:3) :: map_TQU
    COMPLEX(SPC), INTENT(OUT), DIMENSION(1:3,0:nlmax,0:nmmax) :: alm_TGC
    REAL(DP), INTENT(IN), DIMENSION(0:(nlmax+1)*(nlmax+2)*nsmax-1,1:3) :: plm
    REAL(DP),     INTENT(IN) :: cos_theta_cut
    REAL(DP),     INTENT(IN), DIMENSION(1:2*nsmax,1:3) :: w8ring_TQU

    INTEGER(I4B) :: l, m, ith, im_max       ! alm related
    INTEGER(I4B) :: nph, kphi0                                             ! map related
    INTEGER(I4B), ALLOCATABLE, DIMENSION(:) :: istart_south, istart_north  ! map related

    REAL(DP) :: omega_pix
    REAL(DP) :: dth1, dth2
    REAL(DP) :: par_lm
    COMPLEX(DPC) :: zi_lam_x
    COMPLEX(DPC) :: p_nT,p_sT,p_nQ,p_sQ,p_nU,p_sU

    CHARACTER(LEN=*), PARAMETER :: code = 'MAP2ALM_POL'
    COMPLEX(DPC), DIMENSION(:,:), ALLOCATABLE :: phas_nT, phas_nQ, phas_nU
    COMPLEX(DPC), DIMENSION(:,:), ALLOCATABLE :: phas_sT, phas_sQ, phas_sU
    INTEGER(I4B) status, lm, lmstep, lmstep2

    REAL(DP), DIMENSION(0:4*nsmax-1) :: ring
    LOGICAL   :: keep_it
    COMPLEX(DPC), DIMENSION(:,:), POINTER :: trig   
    COMPLEX(SPC), DIMENSION(3,0:nlmax) :: alm_tmp
    REAL(DP), DIMENSION(:), ALLOCATABLE :: cth
    !=======================================================================
    !     --- allocates space for arrays ---

    ALLOCATE(cth(1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'cth')

    ALLOCATE(phas_nT(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_n')

    ALLOCATE(phas_sT(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_s')

    ALLOCATE(phas_nQ(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_n')

    ALLOCATE(phas_sQ(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_s')

    ALLOCATE(phas_nU(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_n')

    ALLOCATE(phas_sU(0:nmmax,1:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'phas_s')

    ALLOCATE(istart_north(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_north')

    ALLOCATE(istart_south(0:2*nsmax),stat = status)
    if (status /= 0) call die_alloc(code,'istart_south')


    call init_trig(nsmax,nmmax,trig)

    !     ----- set the whole alm array to zero ------
    alm_TGC = CMPLX(0.,0.)
    lmstep=(nlmax+1)*(nlmax+2)/2


    !     -------------------------------------------

    istart_north(0)=0
    istart_south(0)=12*nsmax**2
    do ith=1,2*nsmax
       if (ith.lt.nsmax) then  ! polar cap (north)
          nph = 4*ith
       else                    ! tropical band (north) + equator
          nph = 4*nsmax
       endif
       istart_north(ith)=istart_north(ith-1)+nph
       istart_south(ith)=istart_south(ith-1)-nph
    enddo
    omega_pix = pi / (3.0_dp * nsmax * nsmax)

    dth1 = 1.0_dp / (3.0_dp*DBLE(nsmax)**2)
    dth2 = 2.0_dp / (3.0_dp*DBLE(nsmax))

    !     --------------------------------------------
    !-----------------------------------------------------------------------
    !           computes the integral in phi : phas_m(theta)
    !           for each parallele from north to south pole
    !-----------------------------------------------------------------------

!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(ith,nph,kphi0,keep_it,ring)
    do ith = 1, 2*nsmax

       phas_nT(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! North    m >= 0
       phas_sT(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! South    m >= 0
       phas_nQ(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! North    m >= 0
       phas_sQ(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! South    m >= 0
       phas_nU(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! North    m >= 0
       phas_sU(0:nmmax,ith) = CMPLX(0.0_dp, 0.0_dp, KIND=DP)   ! South    m >= 0

       if (ith .le. nsmax-1) then      ! north polar cap
          nph = 4*ith
          kphi0 = 1 
          cth(ith) = 1.0_dp  - DBLE(ith)**2 * dth1
       else                            ! tropical band + equat.
          nph = 4*nsmax
          kphi0 = MOD(ith+1-nsmax,2)
          cth(ith) = DBLE(2*nsmax-ith) * dth2
       endif

       keep_it = (ABS(cth(ith)).gt.cos_theta_cut) ! part of the sky out of the symmetric cut
       if (keep_it) then
          ring(0:nph-1) = map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,1) * w8ring_TQU(ith,1)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_nT(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,2) * w8ring_TQU(ith,2)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_nQ(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_north(ith-1):istart_north(ith-1)+nph-1,3) * w8ring_TQU(ith,3)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_nU(0,ith), kphi0, trig)
       endif

       if (ith .lt. 2*nsmax .and. keep_it) then
          ring(0:nph-1) = map_TQU(istart_south(ith):istart_south(ith)+nph-1,1) * w8ring_TQU(ith,1)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_sT(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_south(ith):istart_south(ith)+nph-1,2) * w8ring_TQU(ith,2)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_sQ(0,ith), kphi0, trig)
          ring(0:nph-1) = map_TQU(istart_south(ith):istart_south(ith)+nph-1,3) * w8ring_TQU(ith,3)
          call ring_analysis2(nsmax, nlmax, nmmax, ring, nph, phas_sU(0,ith), kphi0, trig)
       endif
    enddo
!$OMP END PARALLEL DO

    !-----------------------------------------------------------------------
    !              computes the a_lm by integrating over theta
    !                  lambda_lm(theta) * phas_m(theta)
    !                         for each m and l
    !-----------------------------------------------------------------------
    im_max = nmmax !16/05/1999 B. D. Wandelt -> proper delta function sampling
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(m,alm_tmp,lmstep2,ith,keep_it,p_nT,p_sT,p_nQ,p_sQ,p_nU,p_sU, &
!$OMP         par_lm,zi_lam_x,lm,l)
    do m = 0, im_max !16/05/1999 B. D. Wandelt -> proper delta function sampling
       alm_tmp(1:3,m:nlmax)=0.

       lmstep2=((2*(nlmax+1)-m+1)*m)/2

       do ith = 1, 2*nsmax
          keep_it = (ABS(cth(ith)).gt.cos_theta_cut) ! part of the sky out of the symmetric cut

          if (keep_it) then ! avoid un-necessary calculations (EH, 09-2001)
             lm=lmstep*(ith-1)+lmstep2

             p_nT=phas_nT(m,ith)
             p_sT=phas_sT(m,ith)
             p_nQ=phas_nQ(m,ith) * iKvS * sgQ
             p_sQ=phas_sQ(m,ith) * iKvS * sgQ
             p_nU=phas_nU(m,ith) * iKvS
             p_sU=phas_sU(m,ith) * iKvS

             !           ---------- l = m ----------
             par_lm = 1.0_dp  ! = (-1)^(l+m)
             zi_lam_x = CMPLX(0.0_dp, plm(lm,3), KIND=DP)

             !Temperature
             alm_tmp(1,m) = alm_tmp(1,m) + plm(lm,1) * (p_nT + p_sT)

             !Gradient (Electric) part of the polarisation tensor
             alm_tmp(2,m) = alm_tmp(2,m) &
                  &                 + plm(lm,2) * (p_nQ + p_sQ) &
                  &                 - zi_lam_x * (p_nU - p_sU)

             !Curl     (Magnetic) part of the polarisation tensor
             alm_tmp(3,m) = alm_tmp(3,m) &
                  &                  - plm(lm,2) * (p_nU + p_sU) &
                  &                  - zi_lam_x * (p_nQ - p_sQ)
             lm=lm+1
             !           ---------- l > m ----------

             do l = m+1, nlmax
                par_lm = - par_lm  ! = (-1)^(l+m)

                ! Temperature
                alm_tmp(1,l) = alm_tmp(1,l) + plm(lm,1) * (p_nT + par_lm*p_sT)

                ! polarisation related quantities
                zi_lam_x = CMPLX(0.0_dp, plm(lm,3), KIND=DP)

                ! G =  ( ylm_W * Q - ylm_X * U ) ! (ylm_X is imaginary) ! new CMBFAST normalisation
                alm_tmp(2,l) = alm_tmp(2,l) &
                     &              + plm(lm,2) * (p_nQ + par_lm*p_sQ) &
                     &              - zi_lam_x * (p_nU - par_lm*p_sU)

                ! C = -( ylm_W * U + ylm_X * Q ) ! (ylm_X is imaginary) ! new CMBFAST normalisation
                alm_tmp(3,l) = alm_tmp(3,l) &
                     &              - plm(lm,2) * (p_nU + par_lm*p_sU) &
                     &              - zi_lam_x * (p_nQ - par_lm*p_sQ)
                lm=lm+1

             enddo ! loop on l
          endif ! test on cut sky
       enddo ! loop on theta
       do l = m,nlmax
          alm_TGC(1,l,m)=alm_tmp(1,l)
          alm_TGC(2,l,m)=alm_tmp(2,l)
          alm_TGC(3,l,m)=alm_tmp(3,l)
       enddo
    enddo ! loop on m
!$OMP END PARALLEL DO



    !     --------------------------------------------------------------
    !     normalisation of the alm (multiplication by pixel solid angle)
    !     --------------------------------------------------------------
    !      alm_TGC = alm_TGC * omega_pix
!$OMP PARALLEL DO SCHEDULE(STATIC,1) &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(l,m)
    do m = 0, nmmax       ! some compilers prefer explicit loops (EH, 09-2001)
       do l = m, nlmax
          alm_TGC(1,l,m) = alm_TGC(1,l,m) * omega_pix
          alm_TGC(2,l,m) = alm_TGC(2,l,m) * omega_pix
          alm_TGC(3,l,m) = alm_TGC(3,l,m) * omega_pix
       enddo
    enddo
!$OMP END PARALLEL DO

    !     --------------------
    !     free memory and exit
    !     --------------------

    DEALLOCATE(phas_nT,phas_nQ,phas_nU)
    DEALLOCATE(phas_sT,phas_sQ,phas_sU)
    DEALLOCATE(istart_north,istart_south)
    DEALLOCATE(cth)
    DEALLOCATE(trig)

    RETURN
  END subroutine map2alm_pol_pre2

  !****************************************************

  !=======================================================================
  subroutine pow2alm_units(units_pow, units_alm)
    !=======================================================================
    ! does the unit conversion between power and alm (and map)
    !=======================================================================
    character(len=*), dimension(1:), intent(in)  :: units_pow
    character(len=*), dimension(1:), intent(out) :: units_alm

    integer(kind=i4b) :: i, nu, j, ntemplate, idx
    character(len=80) :: uu, ui, uo
    character(len=5),dimension(1:7) :: template = &
         & (/ "_SQUA","-SQUA","SQUA ","^2   ","^ 2  ","**2  ","** 2 " /)
    !=======================================================================
    nu = min( size(units_pow), size(units_alm))
    ntemplate = size(template)

    units_alm = ""
    do i = 1, nu
       ui = adjustl(units_pow(i))
       uu = ui
       call ftupch(uu) ! upper case
       uo = "unknown" ! default
       do j = 1, ntemplate
          idx = index(uu, template(j))
          if (idx > 0) then
             uo = ui(1:idx-1)
             exit
          endif
       enddo
       units_alm(i) = uo
       !         print*,i,trim(units_pow(i))," -> ",trim(units_alm(i))
    enddo

    return
  end subroutine pow2alm_units
  !****************************************************************************
  subroutine generate_beam(fwhm_arcmin, lmax, gb, beam_file)
    !==========================================================================
    !
    ! generate_beam(fwhm_arcmin, lmax, gb [, beam_file])
    ! generate the beam window function gb up to lmax
    !  either a gaussian of FHWM : fwhm_arcmin (in arcmin)
    !  or the function read from beam_file
    !
    ! July 2003, EH : replicate temperature beam for polarization when reading
    !  standard ASCII file
    !==========================================================================
!     use fitstools, only : getsize_fits, read_asctab
    use fitstools, only : read_asctab
    real(kind=DP),                   intent(in)  :: fwhm_arcmin
    real(kind=DP), dimension(0:,1:), intent(out) :: gb
    integer(kind=I4B),               intent(in)  :: lmax
    character(len=*),      optional, intent(in)  :: beam_file

    character(len=256) :: new_beam_file
    logical(kind=lgt) :: extfile, ok
    integer(kind=i4b) :: nsize, type, nlheader, nl, nd, lunit, il, i
    character(len=80), dimension(1:180) :: header
    character(len=1600) :: str
    character(len=80) :: card
    real(kind=SP), dimension(:,:), allocatable :: sgb
    !==========================================================================
    ! test if name of external is given and valid
    extfile = .false.
    if (present(beam_file)) then
       extfile = (trim(beam_file) /= "")
    endif

    lunit = 15
    nl = size(gb, 1)
    nd = size(gb, 2)
    gb = 0.0_dp

    if (nl <= lmax) then
       print*,'Generate_beam: beam array only available up to ',nl
    endif
    nl = min(nl, lmax+1)

    if (extfile) then
       inquire(file=beam_file, exist=ok)
       if (.not. ok) then
          print*,'can not find file '//trim(beam_file)
          stop
       else
          new_beam_file = beam_file
          print*,'Reading beam information from '//trim(new_beam_file)
       endif

       ! find out file nature
       type = 1
       open(unit=lunit,file=new_beam_file,status='old', &
               &          form='formatted',action='read')
       read(lunit,'(a)') card
       close(lunit)
       card = adjustl(card)
       if (card(1:8) /= 'SIMPLE  ' .AND. card(1:8) /= 'XTENSION') type = -1

       ! read file according to its type
       if (type < 0) then 
          ! ordinary ascii file ?
          lunit = 32
          open(unit=lunit,file=new_beam_file,status='old', &
               &          form='formatted',action='read')
          do
             read(lunit,'(a)', end=100, err=100) str
             if (str(1:1) /= '#') then
                read(str,*) il, gb(il,1)
                if (il == nl-1) exit
             endif
          enddo
100       continue
          close(lunit)
          if (il < (nl-1)) then
             print*,'WARNING: Beam transfer function only available up to l= ',il !
             print*,'         The larger multipoles will be set to 0'
          endif

       else if (type == 1) then
          ! FITS file with ascii table
          allocate(sgb(0:nl-1,1:nd))
          sgb = 0.0_sp
          nlheader = size(header)
          call read_asctab(new_beam_file, sgb, nl, nd, header, nlheader)
          gb(0:nl-1,1:nd) = real(sgb, kind=DP)
          deallocate(sgb)

       else
          print*,'the file '//trim(new_beam_file) &
               &            //' is of unknown type.'
          stop
       endif

       ! if Grad and/or Curl absent, replicate Temperature
       do i=2,nd
          if ( sum(abs(gb(:,i))) < 1.e-7 ) gb(:,i) = gb(:,1)
       enddo

    else
       ! gaussian beam
       print*,'Generating gaussian beam of FHWM [arcmin] = ',fwhm_arcmin
       call gaussbeam(fwhm_arcmin, nl-1, gb)
    endif

    return
  end subroutine generate_beam
  !*************************************************************
  subroutine gaussbeam(fwhm_arcmin, lmax, gb)
    !===========================================================
    ! gaussbeam(fwhm_arcmin, gb)
    !   returns in gb the window function on [0,lmax] corresponding
    !   to the gaussian beam of FWHM = fwhm_arcmin
    ! The returned beam function has up to 3 components, 
    ! gb(*,1) = bt                  : temperature
    ! gb(*,2) = bt * exp(2*sigma^2) : grad
    ! gb(*,3) = bt * exp(2*sigma^2) : curl
    ! with sigma = gaussian rms in radian
    ! and bt = exp( l(l+1) * sigma^2 / 2)
    !===========================================================
    real(kind=DP),                   intent(in)  :: fwhm_arcmin
    real(kind=DP), dimension(0:,1:), intent(out) :: gb
    integer(kind=I4B),               intent(in)  :: lmax

    integer(kind=I4B) :: l, nd
    real(kind=DP)     :: sigma, arcmin2rad, sigma2fwhm, fact_pol
    !===========================================================

    if (fwhm_arcmin < 0.0_dp) then
       print*,'fwhm of gaussian beam should be positive'
       stop
    endif

    nd   = size(gb,2)

    arcmin2rad = PI / (180.0_dp * 60.0_dp)
    sigma2fwhm = sqrt(8.0_dp * log(2.0_dp))

    sigma    = fwhm_arcmin * arcmin2rad / sigma2fwhm ! in radians

    fact_pol = exp(2.0_dp*sigma**2) ! correction for polarised fields

    ! temperature
    do l=0,lmax
       gb(l,1) = exp(-.5_dp * l*(l+1.0_dp) * sigma**2)
    enddo
    ! electric or gradient
    if (nd > 1) gb(0:lmax,2) = gb(0:lmax,1) * fact_pol
    ! magnetic or curl
    if (nd > 2) gb(0:lmax,3) = gb(0:lmax,1) * fact_pol

    return
  end subroutine gaussbeam

  !=======================================================================
  subroutine pixel_window(pixlw, nside, windowfile)
    !=======================================================================
    !=======================================================================

    use fitstools, only : getsize_fits, read_dbintab
    use extension, only : getEnvironment
    use pix_tools, only : nside2npix

    real(DP),     dimension(:,:), intent(OUT)          :: pixlw
    character(len=*),             intent(IN), optional :: windowfile
    integer(i4b),                 intent(IN), optional :: nside
    
    character(len=filenamelen) :: wfile, wfile_def, healpixdir
    character(len=4) :: sstr
    integer(I4B) :: npw_file, ncolw_file, npix
    integer(I4B) :: npw, ncolw
    logical(LGT) :: good, bad
    REAL(DP) ::  nullval
    character(len=*), parameter :: code = 'Pixel_Window > '
    !=======================================================================

    if (.not. present(nside) .and. .not. present(windowfile)) then
       print*,code//'Nside or windowfile have to be present'
       stop
    endif

    npw   = size(pixlw, dim=1)
    ncolw = size(pixlw, dim=2)

    if (present(windowfile)) then
       wfile = trim(windowfile)
       inquire(file=wfile, exist = good)
    else
       npix = nside2npix(nside)
       if (npix < 0) then
          print*,code//'unvalid Nside = ',nside
          stop
       endif
       write(sstr,"(I4.4)") nside
       wfile_def = "data/pixel_window_n"//trim(sstr)//".fits"
       call getEnvironment("HEALPIX",healpixdir)

       wfile = trim(healpixdir)//wfile_def
       inquire(file=wfile, exist = good)
       if (good) goto 10
       wfile = trim(healpixdir)//'/'//wfile_def
       inquire(file=wfile, exist = good)
       if (good) goto 10
    endif
    if (.not.good) then
       print*,code//trim(wfile)//' not found'
       stop
    endif
10  continue

    npw_file = getsize_fits(wfile, nmaps = ncolw_file)
    if (ncolw_file > 3) then
       print*,code//' Too many columns in '//trim(wfile)
       stop
    endif


    if (npw_file < npw) then
       print*,'Only ',npw_file,' multipoles available in '//trim(wfile)//', expected ',npw
    endif
    npw = min(npw_file, npw)
    
    ncolw_file = min(ncolw_file, ncolw)
    call read_dbintab(windowfile, pixlw, npw, ncolw_file, nullval, bad)
    if (ncolw_file == 1) then       ! T only in file
       if (ncolw > 1) pixlw(:,2)  = pixlw(:,1) ! w_G = w_T
       if (ncolw > 2) pixlw(:,3)  = pixlw(:,1) ! w_C = w_T
    else if (ncolw_file == 2) then  ! T and G in file
       if (ncolw > 2) pixlw(:,3)  = pixlw(:,2) ! w_C = w_G
    endif
       

    return
  end subroutine pixel_window
  !=======================================================================
  subroutine alter_alm(nsmax, nlmax, nmmax, fwhm_arcmin, alm, beam_file)
    !=======================================================================
    !     multiply the alm by a gaussian function of FWHM fwhm_arcmin
    !     --> equivalent to a convolution in real sphere
    !=======================================================================
    integer(I4B),     intent(in)                         :: nlmax,nsmax,nmmax
    real(SP),         intent(in)                         :: fwhm_arcmin
    complex(SPC),     intent(inout), dimension(1:,0:,0:) :: alm
    character(LEN=*), intent(IN),  optional              :: beam_file

    real(DP), dimension(:,:),  allocatable :: beamw
    integer(I4B)                           :: status

    integer(I4B)                           :: m, i, nd
    character(len=*), parameter            :: code = "alter_alm"

    !-----------------------------------------------------------------------

    nd = size(alm,1)
    allocate(beamw(0:nlmax,1:nd), stat = status)
    if (status /= 0) call die_alloc(code,'beamw')
!     call gaussbeam(real(fwhm_arcmin,kind=dp), nlmax, beamw)
    call generate_beam(real(fwhm_arcmin,kind=dp), nlmax, beamw, beam_file)

    do m = 0, nmmax
       do i = 1, nd
          alm(i, m:nlmax, m) = alm(i, m:nlmax, m) * beamw(m:nlmax, i)
       enddo
    enddo

    deallocate(beamw)

    return
  end subroutine alter_alm

  !=======================================================================
  subroutine create_alm &
       &     (nsmax, nlmax, nmmax, polar, filename, iseed, fwhm_arcmin, &
       &        alm_TGC, header_PS, windowfile, units, beam_file)
    !=======================================================================
    !     creates the a_lm from the power spectrum, 
    !     assuming they are gaussian  and complex 
    !     with a variance given by C(l)
    !
    !     the input file FILENAME should contain : 
    !       l, C_temp(l), [ C_grad(l), C_curl(l), C_temp_grad(l) ]
    !
    !     with *consecutive* l's (missing C(l) are set to 0.)
    !
    !     because the map is real we have : a_l-m = (-)^m conjug(a_lm)
    !     so we actually compute them only for m >= 0
    !
    !
    !     ISEED : integer
    !
    !     FWHM_ARCMIN
    !
    !     ALM_TGC (Complex array) : either (1:1, 0:nlmax, 0:nmmax) 
    !       for temperature only (if POLAR=0)
    !                               or     (1:3, 0:nlmax, 0:nmmax) 
    !       for temperature + polarisation (if POLAR =1)
    !       (respectively Temp, Grad or Electric component 
    !        and Curl or Magnetic one)
    !
    !=======================================================================
    use fitstools, only: read_asctab
    use ran_tools, only: randgauss_boxmuller
    USE head_fits, ONLY : add_card, merge_headers, get_card
    INTEGER(I4B), INTENT(IN) :: nsmax
    INTEGER(I4B), INTENT(IN) :: nlmax
    INTEGER(I4B), INTENT(IN) :: nmmax

    INTEGER(I4B),       INTENT(IN) :: polar
    CHARACTER(LEN=*),   INTENT(IN)      :: filename
    INTEGER(I4B),       INTENT(INOUT)   :: iseed
    REAL(SP),           INTENT(IN)      :: fwhm_arcmin
    COMPLEX(SPC),       INTENT(OUT), &
         & DIMENSION(1:1+2*polar,0:nlmax,0:nmmax) :: alm_TGC
    CHARACTER(LEN=80),  INTENT(OUT), DIMENSION(1:):: header_PS
    CHARACTER(LEN=*),   INTENT(IN), OPTIONAL      :: windowfile
    CHARACTER(LEN=80),  INTENT(OUT), OPTIONAL, DIMENSION(1:):: units
    CHARACTER(LEN=*),   INTENT(IN),  OPTIONAL               :: beam_file

    INTEGER(I4B) :: i, l, m, npw, l_max
    INTEGER(I4B) :: ncl, nlheader, count_tt, count_pn
    INTEGER(I4B) :: status
    REAL(SP) ::  hsqrt2, quadrupole
    REAL(SP) ::  rms_tt, rms_g1, rms_g2, rms_cc
    REAL(SP) ::  zeta1_r, zeta1_i, zeta2_r, zeta2_i, zeta3_r, zeta3_i

    LOGICAL(LGT) ::  polarisation

    REAL(DP), DIMENSION(:,:), ALLOCATABLE :: pixlw, beamw
    REAL(SP), DIMENSION(:),   ALLOCATABLE :: cls_tt, cls_tg, cls_gg, cls_cc
    REAL(SP), DIMENSION(:,:), ALLOCATABLE :: cl_in
    REAL(SP), DIMENSION(:),   ALLOCATABLE :: fact_norm

    CHARACTER(LEN=*), PARAMETER :: code = 'CREATE_ALM'
    CHARACTER(LEN=*), PARAMETER :: version = '1.2.0'
    CHARACTER(LEN=20)                            ::  string_quad
    CHARACTER(LEN=80), DIMENSION(1:180)          :: header_file
    CHARACTER(LEN=80), DIMENSION(:), allocatable :: units_power
    CHARACTER(LEN=80) :: temptype, com_tt
    CHARACTER(LEN=80) :: polnorm, com_pn


    !=======================================================================
    ! Is polarisation required ?
    if (polar .eq. 0) then
       polarisation = .false.
    elseif (polar .eq. 1) then
       polarisation = .true.
    else
       print*,'wrong choice of polar'
       stop
    endif

    ! set maximum multipole (depends on user choice and pixel window function)
    l_max = nlmax
    npw = 4*nsmax + 1
    if (present(windowfile)) then
       l_max = min(l_max, npw - 1)
    else
       npw = l_max + 1
    endif
    if (l_max < nlmax) then
       print*,'a_lm are only generated for 0 < l <= ',l_max
    endif
    !-----------------------------------------------------------------------

    ALLOCATE( cls_tt(0:nlmax),&
         &    cls_tg(0:nlmax),&
         &    cls_gg(0:nlmax),&
         &    cls_cc(0:nlmax),stat = status)
    if (status /= 0) call die_alloc(code,'cls_tt')

    ALLOCATE( cl_in(0:nlmax,1:4),stat = status)
    if (status /= 0) call die_alloc(code,'cl_in')

    !     ----- set the alm array to zero ------
    if (polarisation) then
       alm_TGC(1:3, 0:nlmax, 0:nmmax) = CMPLX(0.0,0.0)
    else
       alm_TGC(1  , 0:nlmax, 0:nmmax) = CMPLX(0.0,0.0)
    endif
    !     --------------------------------------
    cls_tt = 0.
    cls_gg = 0.
    cls_cc = 0.
    cls_tg = 0.
    cl_in = 0.

    !     --- reads the C(l) file ---
    ncl = 1
    if (polarisation) ncl = 4
    nlheader = SIZE(header_file)
    allocate(units_power(1:ncl), stat = status)
    if (status /= 0) call die_alloc(code,'units_power')

    call read_asctab(filename, cl_in(0:l_max,1:ncl), l_max, ncl, &
         &           header_file, nlheader, units=units_power)

    !    ---- check input power spectra consistency  ----
    do i=1,ncl
       do l=0,l_max
          if (i < 4) then ! auto correlation
             if (cl_in(l,i) < 0.0) then
                print*,code,'> Negative input power spectrum at l =', l,', index = ',i
                print*,code,'> ',cl_in(l,i)
                stop
             endif
          else ! cross correlation
             if (abs(cl_in(l,4)) > sqrt(cl_in(l,1))*sqrt(cl_in(l,2)) ) then
                print *,code,'> Inconsistent T, E and ExT ower spectrum terms at l = ',l
                print*,code,'> ',cl_in(l,1),cl_in(l,2),cl_in(l,4)
             endif
          endif
       enddo
    enddo

    allocate( fact_norm(0:nlmax),stat = status)
    if (status /= 0) call die_alloc(code,'fact_norm')
    do l=1, nlmax
       fact_norm(l) = 1.0         !   Normalisation=1, 'pure' C_ls are expected 
    enddo
    fact_norm(0) = 1.0

    !     we always expect 4 columns, in the order T, G(=E), C(=B) and TG(=TE)
    cls_tt(0:l_max) = cl_in(0:l_max,1) / fact_norm(0:l_max)
    cls_gg(0:l_max) = cl_in(0:l_max,2) / fact_norm(0:l_max)
    cls_cc(0:l_max) = cl_in(0:l_max,3) / fact_norm(0:l_max)
    cls_tg(0:l_max) = cl_in(0:l_max,4) / fact_norm(0:l_max)
    deallocate(fact_norm)

    ! converts units for power spectrum into units for maps
    if (present(units)) then
       call pow2alm_units(units_power, units)
    endif

    ! finds out temperature type (THERMO or ANTENNA) 
    ! THERMO is the default
    call get_card(header_file,"TEMPTYPE",temptype,com_tt,count=count_tt)
    if (count_tt < 1) then
       temptype = "THERMO"
       com_tt = " temperature : either THERMO or ANTENNA"
       print*,'   Will assume '//temptype
    endif

    if (polar > 0) then
       call get_card(header_file,"POLNORM",polnorm,com_pn,count=count_pn)
       if (count_pn < 1) then
          print*,' ********************************************************* '
          print*,'            The convention for normalization '
          print*,'        of the input polarization power spectra '
          print*,'                      is unknown. '
          print*,' The code will proceed as if the convention was that of CMBFAST'
          print*,'  See the Healpix Postscript of Html documentation for details'
          print*,' ********************************************************* '
       endif
    endif

    quadrupole = cls_tt(2)
    !     --- creates the header relative to power spectrum ---
    header_PS = ''
    call add_card(header_PS)
    call add_card(header_PS,'HISTORY',' alm generated by ' &
         &        //code//'_'//version//' from following power spectrum')
    call add_card(header_PS)
    call add_card(header_PS,'COMMENT','----------------------------------------------------')
    call add_card(header_PS,'COMMENT','Planck Power Spectrum Description Specific Keywords')
    call add_card(header_PS,'COMMENT','----------------------------------------------------')
    call add_card(header_PS,'COMMENT','Input power spectrum in : ')
    call add_card(header_PS,'COMMENT',TRIM(filename))
    call add_card(header_PS)
    call add_card(header_PS,'COMMENT','Quadrupole')
    write(string_quad,'(1pe15.6)') quadrupole
    call add_card(header_PS,'COMMENT','  C(2) = '//string_quad//' '//trim(units_power(1)))
    call add_card(header_PS)
    call add_card(header_PS,"TEMPTYPE",temptype,com_tt)
    call add_card(header_PS)
    ! ------- insert header read in power spectrum file -------
    call merge_headers(header_file, header_PS)
    deallocate(units_power)

    !     --- smoothes the initial power spectrum ---
    !       beam (gaussian or external file) + pixel (external file)

    ! define beam
    allocate(beamw(0:l_max,1:3), stat = status)
    if (status /= 0) call die_alloc(code,'beamw')
    call generate_beam(real(fwhm_arcmin,kind=dp), l_max, beamw, beam_file)
!     call gaussbeam(real(fwhm_arcmin,kind=dp), l_max, beamw)

    ! get the pixel window function 
    allocate(pixlw(0:npw-1,1:3), stat = status)
    if (status /= 0) call die_alloc(code,'pixlw')
    if(present(windowfile)) then
       call pixel_window(pixlw, windowfile=windowfile)
    else
       pixlw(:,:)  = 1.0_dp
    endif
    ! multiply beam and pixel windows
    beamw(0:l_max,1:3) = beamw(0:l_max,1:3) * pixlw(0:l_max,1:3)

    cls_tt(0:l_max) = cls_tt(0:l_max) * beamw(0:l_max,1)**2
    cls_tg(0:l_max) = cls_tg(0:l_max) * beamw(0:l_max,1)*beamw(0:l_max,2)
    cls_gg(0:l_max) = cls_gg(0:l_max) * beamw(0:l_max,2)**2
    cls_cc(0:l_max) = cls_cc(0:l_max) * beamw(0:l_max,3)**2
    deallocate(pixlw)
    deallocate(beamw)

    !     --- generates randomly the alm according to their power spectrum ---
    !     alm_T = zeta1 * rms_tt
    !     alm_G = zeta1 * rms_g1 + zeta2 * rms_g2
    !     alm_C = zeta3 * rms_cc
    hsqrt2 = 1.0 / sqrt(2.0)

    do l = 0, l_max
       rms_tt = 0.
       rms_g1 = 0.
       if (cls_tt(l) .ne. 0) then
          rms_tt   = sqrt(cls_tt(l))
          rms_g1   = cls_tg(l) / rms_tt
       endif

       !        ------ m = 0 ------
       zeta1_r = randgauss_boxmuller(iseed)  ! gaussian deviate based on ran_mwc
       zeta1_i = 0.0
       alm_TGC(1, l, 0)   = CMPLX(zeta1_r, zeta1_i) * rms_tt ! T
       if (polarisation) then
          alm_TGC(2, l, 0) = CMPLX(zeta1_r, zeta1_i) * rms_g1 ! G
       endif

       !        ------ m > 0 ------
       do m = 1,l
          zeta1_r = randgauss_boxmuller(iseed) * hsqrt2
          zeta1_i = randgauss_boxmuller(iseed) * hsqrt2
          alm_TGC(1, l, m) = CMPLX(zeta1_r, zeta1_i) * rms_tt
          if (polarisation) then
             alm_TGC(2, l, m) = CMPLX(zeta1_r, zeta1_i) * rms_g1
          endif
       enddo
    enddo

    !     the coefficient generation is separated so that the Temperature
    !     coeff. are the same (for the same seed) weither the polarisation is set or not
    
    if (polarisation) then
       do l = 0, l_max
          rms_g2 = 0.
          rms_cc = 0.
          if (cls_tt(l) .ne. 0) then
             rms_g2 = cls_gg(l) - (cls_tg(l)/cls_tt(l))*cls_tg(l) ! to avoid underflow
             ! test for consistency but make sure it is not due to round off error
             if (rms_g2 <= 0.0) then
                if (abs(rms_g2) > abs(1.e-8*cls_gg(l))) then
                   print*,code,'> Unconsistent TT, GG and TG spectra at l=',l
                   stop
                else ! only round off error, keep going
                   rms_g2 = 0.0
                endif
             endif
             rms_g2 = SQRT( rms_g2 )
             rms_cc = SQRT( cls_cc(l) )
          endif

          !           ------ m = 0 ------
          zeta2_r = randgauss_boxmuller(iseed)
          zeta2_i = 0.0
          zeta3_r = randgauss_boxmuller(iseed)
          zeta3_i = 0.0
          alm_TGC(2, l, 0) = alm_TGC(2, l, 0) &
               &                 + CMPLX(zeta2_r, zeta2_i) * rms_g2 ! G
          alm_TGC(3, l, 0) = CMPLX(zeta3_r, zeta3_i) * rms_cc ! C

          !           ------ m > 0 ------
          do m = 1,l
             zeta2_r = randgauss_boxmuller(iseed) * hsqrt2
             zeta2_i = randgauss_boxmuller(iseed) * hsqrt2
             zeta3_r = randgauss_boxmuller(iseed) * hsqrt2
             zeta3_i = randgauss_boxmuller(iseed) * hsqrt2
             alm_TGC(2, l, m) = alm_TGC(2, l, m) &
                  &                 + CMPLX(zeta2_r, zeta2_i) * rms_g2 ! G
             alm_TGC(3, l, m) = CMPLX(zeta3_r, zeta3_i) * rms_cc ! C
          enddo
       enddo
    endif


    DEALLOCATE( cl_in )
    DEALLOCATE( cls_tt, cls_tg, cls_gg, cls_cc )

    return
  end subroutine create_alm

end module alm_tools
