module fitstools
  !
  ! module fitstools
  !  version 1.2 (EH, Feb 2003) + patch #01 
  !
  ! addition of read_fits_cut4 and write_fits_cut4
  ! improvement of input_map (addition of header output)
  ! improvement of read_asctab (more flexible on number of columns)
  ! removal of ask_* routines
  ! addition of get_clean_header
  ! addition of optional keyword units to input_map, read_fits_cut4, read_bintab
  ! 
  ! cleaning of the mapping between (l,m) <-> lm = l*l+l+m+1
  !  to remove round-off errors
  !
  ! v1.2_p01 : input_map, read_bintab : fix SUN compiler bug
  !            getsize_fits           : correct handling of ASCII table
  ! v1.2_p02   getsize_fits : correct handling of files with more than 2^31 elements
  !            read_bintab  : correction to array argument in ftgcve
  !            2004-Feb: 
  !               modification of dump_alms to deal with large arrays on MacOsX
  !               modification of npix in dump_alms
  !               modification of read_conbintab to deal with large arrrays
  ! -------------------------------------------------------------
  !
  ! subroutine read_fits_cut4
  ! subroutine write_fits_cut4
  !
  ! subroutine input_map
  ! subroutine read_bintab
  ! subroutine read_conbintab
  ! subroutine read_dbintab
  ! subroutine read_asctab
  !
  ! subroutine output_map
  ! subroutine write_bintab
  ! subroutine write_dbintab
  ! subroutine write_asctab
  !
  ! subroutine dump_alms
  ! subroutine alms2fits
  ! subroutine fits2alms
  ! subroutine write_alms
  ! subroutine read_alms
  !
  ! subroutine printerror
  !
  ! subroutine read_par
  ! function getsize_fits
  ! function number_of_alms
  !
  ! subroutine putrec
  ! subroutine get_clean_header
  !
  ! subroutine input_tod_s
  ! subroutine input_tod_d
  ! subroutine read_bintod_s
  ! subroutine read_bintod_d
  ! subroutine write_bintabh


  use healpix_types
  implicit none

!  real(kind=SP),    private, parameter :: null_value = -1.6375e30_sp
  real(kind=SP),     private, parameter :: s_bad_value = -1.6375e30_sp
  real(kind=DP),     private, parameter :: d_bad_value = -1.6375e30_dp
  integer(kind=I4B), private, parameter :: i_bad_value = -1637500000
  integer(I4B) ,     private, parameter :: nchunk_max  = 12000

  INTERFACE input_tod
     MODULE PROCEDURE input_tod_s,input_tod_d
  END INTERFACE

  private

  public :: read_fits_cut4, write_fits_cut4, & 
       & input_map, read_bintab, read_conbintab, read_dbintab, read_asctab, &
       & output_map, write_bintab, write_dbintab, write_asctab, &
       & dump_alms, alms2fits, fits2alms, write_alms, read_alms, &
       & printerror, read_par, getsize_fits, number_of_alms
  public :: input_tod, write_bintabh
  public :: putrec

contains

  !=======================================================================
  subroutine read_fits_cut4(filename, npixtot, pixel, signal, n_obs, serror, header, units)
    !=======================================================================
    !
    ! routine to read FITS file with 4 columns : PIXEL, SIGNAL, N_OBS, SERROR
    !
    ! read_fits_cut4(filename, npixtot, pixel, [signal, n_obs, serror, header, units])
    !=======================================================================

    character(len=*),                   intent(in)              :: filename
    integer(I4B),                       intent(in)              :: npixtot
    integer(I4B), dimension(0:npixtot-1), intent(out)           :: pixel
    real(SP),     dimension(0:npixtot-1), intent(out), optional :: signal
    integer(I4B), dimension(0:npixtot-1), intent(out), optional :: n_obs
    real(SP),     dimension(0:npixtot-1), intent(out), optional :: serror
    character(len=*), dimension(1:),      intent(out), optional :: header
    character(len=*),                     intent(out), optional :: units

    integer(I4B) :: obs_npix

    integer(I4B), parameter :: MAXDIM = 20 !number of columns in the extension
    integer(I4B) :: blocksize, datacode
    integer(I4B) :: firstpix, frow, hdutype
    integer(I4B) :: naxis, nfound, nmove, npix, nrows
    integer(I4B) :: readwrite, status, tfields, unit, varidat, width
    integer(I4B) :: repeat, repeat1, repeat2
    logical(LGT) :: anynull, extend
    character(len=20), dimension(MAXDIM) :: ttype, tform, tunit
    character(len=20) :: extname
    character(len=80) :: comment
    real(SP) :: nullval
    !=======================================================================

    status=0

    unit = 150
    nfound = -1
    anynull = .false.

    readwrite=0
    call ftopen(unit,filename,readwrite,blocksize,status)
    if (status > 0) call printerror(status)

    !     determines the presence of image
    call ftgkyj(unit,'NAXIS', naxis, comment, status)
    if (status > 0) call printerror(status)
    if (naxis > 0) then ! there is an image
       print*,'an image was found in the FITS file '//filename
       print*,'... it is ignored.'
    endif

    !     determines the presence of an extension
    call ftgkyl(unit,'EXTEND', extend, comment, status)
    if (status > 0) then 
       print*,'extension expected and not found in FITS file '//filename
       print*,'abort code'
       stop
    endif

    nmove = +1
    call ftmrhd(unit, nmove, hdutype, status)

    if (hdutype /= 2) then ! not a binary table
       stop 'this is not a binary table'
    endif

    ! reads all the FITS related keywords
    call ftghbn(unit, MAXDIM, &
         &        nrows, tfields, ttype, tform, tunit, extname, varidat, &
         &        status)
    if (tfields < 4) then
       print*,'Expected 4 columns in FITS file '//filename
       print*,'found ',tfields
       if (tfields < 2) stop
       if (trim(ttype(1)) /= 'PIXEL') stop
    endif

    if (present(header)) then 
       header = ""
       status = 0
       call get_clean_header(unit, header, filename, status)
    endif

    if (present(units)) then 
       ! the second column contains the SIGNAL, for which we 
       ! already read the units from the header
       units = adjustl(tunit(2))
    endif

    !        finds the bad data value
    call ftgkye(unit,'BAD_DATA',nullval,comment,status)
    if (status == 202) then ! bad_data not found
       nullval = s_bad_value ! default value
       status = 0
    endif

    frow = 1
    firstpix = 1
    !        parse TFORM keyword to find out the length of the column vector
    repeat1 = 1
    repeat2 = 1
    call ftbnfm(tform(1), datacode, repeat1, width, status)
    if (tfields > 1) call ftbnfm(tform(2), datacode, repeat2, width, status)
    repeat = max(repeat1, repeat2)

    call ftgkyj(unit,'OBS_NPIX',obs_npix,comment,status)
    if (status == 202) then ! obs_npix not found
       obs_npix = nrows * repeat
       status = 0
    endif

    npix = min(npixtot, obs_npix)
    call ftgcvj(unit, 1, frow, firstpix, npix, nullval, pixel(0), anynull, status)
    if (present(signal)) &
         & call ftgcve(unit, 2, frow, firstpix, npix, nullval, signal(0), anynull, status)
    if (tfields >= 3 .and. present(n_obs)) &
         & call ftgcvj(unit, 3, frow, firstpix, npix, nullval, n_obs(0), anynull, status)
    if (tfields >= 4 .and. present(serror)) &
         & call ftgcve(unit, 4, frow, firstpix, npix, nullval, serror(0), anynull, status)


    !     close the file
    call ftclos(unit, status)

    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)



    return
  end subroutine read_fits_cut4

  !=======================================================================
  subroutine write_fits_cut4(filename, obs_npix, pixel, signal, n_obs, serror, &
       &                     header, coord, nside, order, units)
    !=======================================================================
    ! writes a fits file for cut sky data set with 4 columns:
    !  PIXEL, SIGNAL, N_OBS, SERROR
    ! 
    ! write_fits_cut4(filename, obs_npix, pixel, signal, n_obs, serror
    !         [, header, coord, nside, order, units])
    !=======================================================================
    use pix_tools, only: nside2npix

    character(len=*), intent(in) :: filename
    integer(I4B)     :: obs_npix
    integer(I4B), dimension(0:obs_npix-1), intent(in)           :: pixel
    real(SP),     dimension(0:obs_npix-1), intent(in)           :: signal
    integer(I4B), dimension(0:obs_npix-1), intent(in)           :: n_obs
    real(SP),     dimension(0:obs_npix-1), intent(in)           :: serror
    character(len=*), dimension(1:),       intent(in), optional :: header
    integer(I4B),                          intent(in), optional :: nside, order
    character(len=*),                      intent(in), optional :: coord, units

    character(len=*), parameter :: routine = 'write_fits_cut4'
    ! --- healpix/fits related variables ----
    integer(I4B)     :: ncol, grain
    integer(I4B)     :: npix_hd, nside_hd, nside_final, npix_final, i
    integer(I4B)     :: maxpix, minpix
    character(len=1) :: char1, coord_usr
!    character(len=8) :: char8
    character(len=20) :: units_usr
    logical(LGT)     :: done_nside, done_order, done_coord
    integer(I4B)     :: nlheader

    ! --- cfitsio related variables ----
    integer(I4B) ::  status, unit, blocksize, bitpix, naxis, naxes(1)
    logical(LGT) ::  simple, extend
    character(LEN=80) :: svalue, comment

    integer(I4B), parameter :: MAXDIM = 20 !number of columns in the extension
    integer(I4B) :: nrows, tfields, varidat
    integer(I4B) :: frow,  felem, repeat, repeatg
    character(len=20) :: ttype(MAXDIM), tform(MAXDIM), tunit(MAXDIM), extname
    character(len=8) ::  date
    character(len=10) ::  fulldate
    character(len=80) ::  card
    character(len=4) :: srepeat, srepeatg
    !=======================================================================

!     ncol = 2
!     if (present(n_obs)) then
!        ncol = 3
!        if (present(serror)) ncol = 4
!     endif
!     if (present(serror) .and. .not. present(n_obs)) then
!        print*,routine,' Serror can not be present without N_Obs'
!        stop
!     endif
    ncol = 4

    grain = 1
    nlheader = 0
    if (present(header)) nlheader = size(header)
    units_usr = ' '
    if (present(units)) units_usr = units

    status=0

    unit = 100

    !     create the new empty FITS file
    blocksize=1
    call ftinit(unit,filename,blocksize,status)

    !     -----------------------------------------------------
    !     initialize parameters about the FITS image
    simple=.true.
    bitpix=32     ! integer*4
    naxis=0       ! no image
    naxes(1)=0
    extend=.true. ! there is an extension

    !     ----------------------
    !     primary header
    !     ----------------------
    !     write the required header keywords
    call ftphpr(unit,simple,bitpix,naxis,naxes,0,1,extend,status)

    !     write the current date
    call ftpdat(unit,status) ! format (dd/mm/yy)

    !     update the date (format ccyy-mm-dd)
    call date_and_time(date)
    fulldate = date(1:4)//'-'//date(5:6)//'-'//date(7:8)
    comment = 'FITS file creation date ccyy-mm-dd'
    call ftukys(unit,'DATE',fulldate,comment,status)

    !     ----------------------
    !     image : none
    !     ----------------------

    !     ----------------------
    !     extension
    !     ----------------------

    !     creates an extension
    call ftcrhd(unit, status)

    !     writes required keywords
    !     repeat = 1024
    repeat = 1
    nrows    = (obs_npix + repeat - 1)/ repeat ! naxis1
    if (obs_npix < repeat) repeat = 1
    write(srepeat,'(i4)') repeat
    srepeat = adjustl(srepeat)

    repeatg = repeat * grain
    write(srepeatg,'(i4)') repeatg
    srepeatg = adjustl(srepeatg)

    tfields  = ncol
    ttype(1:4) = (/ 'PIXEL ','SIGNAL','N_OBS ','SERROR' /)
    tform(1) = trim(srepeat)//'J'
    tform(2) = trim(srepeatg)//'E'
    tform(3) = trim(srepeatg)//'J'
    tform(4) = trim(srepeatg)//'E'
    !    tform(1:ncol) = (/ srepeat//'J' ,srepeatg//'E' ,srepeatg//'J' ,srepeatg//'E'  /)
    tunit =  ' '      ! optional, will not appear
    tunit(2) = units_usr
    tunit(4) = units_usr
    extname  = 'SKY_OBSERVATION'      ! default, will be overide by user provided one if any
    varidat  = 0
    call ftphbn(unit, nrows, tfields, ttype, tform, tunit, &
         &     extname, varidat, status)

    call ftpcom(unit,'------------------------------------------',status)
    call ftpcom(unit,'          Pixelisation Specific Keywords    ',status)
    call ftpcom(unit,'------------------------------------------',status)
    call ftpkys(unit,'PIXTYPE','HEALPIX ',' HEALPIX Pixelisation',status)      
    call ftpkyu(unit,'NSIDE',   ' ',status) ! place holder, will be updated later on
    call ftpkyu(unit,'ORDERING',' ',status)
    call ftpkys(unit,'COORDSYS',' ',' ',status)
    call ftpcom(unit,'  G = Galactic, E = ecliptic, C = celestial = equatorial', status)   
    call ftpcom(unit,'------------------------------------------',status)
    call ftpcom(unit,'          Data Specific Keywords    ',status)
    call ftpcom(unit,'------------------------------------------',status)
    call ftpkys(unit,'INDXSCHM','EXPLICIT',' Indexing : IMPLICIT or EXPLICIT', status)
    call ftpkyj(unit,'GRAIN',  grain,     ' Grain of pixel indexing',status)
    call ftpcom(unit,'GRAIN=0 : no indexing of pixel data (IMPLICIT) ',status)
    call ftpcom(unit,'GRAIN=1 : 1 pixel index -> 1 pixel data (EXPLICIT)',status)
    call ftpcom(unit,'GRAIN>1 : 1 pixel index -> data of GRAIN consecutive pixels (EXPLICIT)',status)
    call ftpkys(unit,'OBJECT','PARTIAL ',' Sky coverage represented by data',status)     
    call ftpkyj(unit,'OBS_NPIX',obs_npix, ' Number of pixels observed and recorded',status)


    ! add required Healpix keywords (NSIDE, ORDER) if provided by user
    done_order = .false.
    if (present(order)) then
       !        char8 = order
       !        call ftupch(char8)
       !        if (char8(1:4) == 'RING') then
       if (order == 1) then
          call ftukys(unit, 'ORDERING','RING',' Pixel ordering scheme, either RING or NESTED',status)
          done_order = .true.
          !        elseif (char8(1:4) == 'NEST') then
       elseif (order == 2) then
          call ftukys(unit, 'ORDERING','NESTED',' Pixel ordering scheme, either RING or NESTED ',status)
          done_order = .true.
       else
          print*,'Invalid ORDER given : ',order, ' instead of 1 (RING) or 2 (NESTED)'
       endif
    endif

    done_nside = .false.
    if (present(nside)) then
       if (nside2npix(nside) > 0) then ! valid nside
          call ftukyj(unit,'NSIDE',nside,' Resolution parameter for HEALPIX', status)
          done_nside = .true.
          nside_final = nside
       else
          print*,'Invalid NSIDE given : ',nside
       endif
    endif

    ! add non required Healpix keyword (COORD)
    done_coord = .false.
    if (present(coord)) then
       coord_usr = adjustl(coord)
       char1 = coord_usr(1:1)
       call ftupch(char1) ! uppercase
       if (char1 == 'C' .or. char1 == 'Q') then
          coord_usr = 'C'
          done_coord = .true.
       elseif (char1 == 'G') then
          coord_usr = 'G'
          done_coord = .true.
       elseif (char1 == 'E' ) then
          coord_usr='E'
          done_coord = .true.
       else
          print*,'Unrecognised COORD given : ',coord,' instead of C, G, or E'
          print*,'Proceed at you own risks '
          coord_usr = char1
          done_coord = .true.
       endif
       if (done_coord) then
          call ftukys(unit, 'COORDSYS',coord_usr,' Pixelisation coordinate system',status)
       endif
    endif


    !    write the user provided header literally, except for  PIXTYPE, TFORM*, TTYPE*, TUNIT* and INDXSCHM
    !    copy NSIDE, ORDERING and COORDSYS if they are valid and not already given
    do i=1,nlheader
       card = header(i)
       if (card(1:5) == 'TTYPE' .or. card(1:5) == 'TFORM' .or. card(1:7) == 'PIXTYPE') then
          continue ! don't keep them
       else if (card(1:8) == 'INDXSCHM') then
          continue
       else if (card(1:5) == 'TUNIT') then 
          if ((card(6:6) == '2' .or. card(6:6) == '4') .and. trim(units_usr) == '') then
             call ftucrd(unit,'TUNIT'//card(6:6),header(i), status)
          endif
       else if (card(1:5) == 'NSIDE') then
          call ftpsvc(header(i), svalue, comment, status)
          read(svalue,*) nside_hd
          npix_hd = nside2npix(nside_hd)
          if (.not. done_nside .and. npix_hd > 0) then
             call ftucrd(unit,'NSIDE',header(i), status) !update NSIDE
             done_nside = .true.
             nside_final = nside_hd
          endif
       else if (card(1:8) == 'ORDERING') then
          call ftpsvc(header(i), svalue, comment, status)
          svalue = adjustl(svalue)
          call ftupch(svalue)
          if (.not. done_order .and. (svalue(1:4)=='RING' .or. svalue(1:6) == 'NESTED')) then
             call ftucrd(unit,'ORDERING',header(i), status)
             done_order = .true.
          endif
       else if (card(1:8) == 'COORDSYS') then
          if (.not. done_coord) call putrec(unit,header(i), status)
          done_coord = .true.
       else if (card(1:7) == 'EXTNAME') then
          call ftucrd(unit, 'EXTNAME', header(i), status)
       else if (card(1:1) /= ' ') then ! if none of the above, copy to FITS file
          call putrec(unit, header(i), status)
       endif
10     continue
    enddo


    ! test that required keywords have been provided in some way
    if (.not. done_nside) then
       print*,routine//'> NSIDE is a Healpix required keyword, '
       print*,routine//'>  it was NOT provided either as routine argument or in the input header'
       print*,routine//'>  abort execution, file not written'
       stop
    endif
    if (.not. done_order) then
       print*,routine//'> ORDER is a Healpix required keyword, '
       print*,routine//'>  it was NOT provided either as routine argument or in the input header'
       print*,routine//'>  abort execution, file not written'
       stop
    endif

    ! check validity of PIXEL
    npix_final = nside2npix(nside_final)
    minpix = minval(pixel(0:obs_npix-1))
    maxpix = maxval(pixel(0:obs_npix-1))
    if (minpix < 0 .or. maxpix > npix_final-1) then
       print*,routine//'> Actual pixel range = ',minpix,maxpix
       print*,routine//'> expected range (for Nside =',nside_final,') : 0, ',npix_final-1
       print*,routine//'> ABORT execution, file not written.'
       stop
    endif
    if (obs_npix > npix_final) then 
       print*,routine//'> The actual number of pixels ',obs_npix
       print*,routine//'> is larger than the number of pixels over the whole sky : ',npix_final
       print*,routine//'> for Nside = ',nside_final
       print*,routine//'> ABORT execution, file not written.'
       stop
    endif

    !     write the extension one column by one column
    frow   = 1  ! starting position (row)
    felem  = 1  ! starting position (element)
    call ftpclj(unit, 1, frow, felem, obs_npix, pixel , status)
    call ftpcle(unit, 2, frow, felem, obs_npix, signal, status)
    if (tfields >= 3) call ftpclj(unit, 3, frow, felem, obs_npix, n_obs , status)
    if (tfields >= 4) call ftpcle(unit, 4, frow, felem, obs_npix, serror, status)

    !     ----------------------
    !     close and exit
    !     ----------------------

    !     close the file and free the unit number
    call ftclos(unit, status)

    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)

    return
  end subroutine write_fits_cut4
  !=======================================================================
  subroutine input_map(filename, map, npixtot, nmaps, fmissval, header, units, extno)
    !=======================================================================
    !     reads fits file
    !     filename = fits file (input)
    !     map      = data read from the file (ouput) = real*4 array of size (npixtot,nmaps)
    !     npixtot  = length of the map (input)
    !     nmaps     = number of maps
    !     fmissval  = OPTIONAL argument (input) with the value to be given to missing
    !             pixels, its default value is 0
    !     header    = OPTIONAL (output) contains extension header
    !     units     = OPTIONAL (output) contains the maps units
    !     extno     = OPTIONAL (input)  contains the unit number to read from (0 based)
    !
    !     modified Feb 03 for units argument to run with Sun compiler
    !=======================================================================

    INTEGER(I4B),     INTENT(IN)           :: npixtot, nmaps
    REAL(SP),         INTENT(OUT), dimension(0:,1:) :: map
    CHARACTER(LEN=*), INTENT(IN)           :: filename
    REAL(SP),         INTENT(IN), OPTIONAL :: fmissval
    CHARACTER(LEN=*), INTENT(OUT), dimension(1:), OPTIONAL :: header
    CHARACTER(LEN=*), INTENT(OUT), dimension(1:), OPTIONAL :: units
    INTEGER(I4B)                   , optional,  INTENT(IN) :: extno

    INTEGER(I4B) :: i,imap
    REAL(SP)     :: fmissing, fmiss_effct
    INTEGER(I4B) :: imissing
    integer(I4B) :: obs_npix, nlheader

    LOGICAL(LGT) :: anynull
    integer(I4B), dimension(:), allocatable :: pixel
    real(SP),     dimension(:), allocatable :: signal
    integer(I4B) :: status
    integer(I4B) :: type_fits, nmaps_fits, maxpix, minpix
    CHARACTER(LEN=80)  :: units1
    CHARACTER(LEN=80),dimension(1:10)  :: unitsm
!    CHARACTER(LEN=80),dimension(:), allocatable  :: unitsm
    integer(I4B) :: extno_i
    !-----------------------------------------------------------------------

!    allocate(unitsm(1:10))
    units1    = ' '
    unitsm(:) = ' '
    fmiss_effct = 0.
    imissing = 0
    if (PRESENT(fmissval)) fmiss_effct = fmissval
    if (PRESENT(header)) then
       nlheader = size(header)
    else
       nlheader = 0
    endif
    extno_i = 0
    if (present(extno)) extno_i = extno

    obs_npix = getsize_fits(filename, nmaps = nmaps_fits, type=type_fits)

    !       if (nmaps_fits > nmaps) then 
    !          print*,trim(filename)//' only contains ',nmaps_fits,' maps'
    !          print*,' You try to read ',nmaps
    !       endif

    if (type_fits == 0 .or. type_fits == 2) then ! full sky map (in image or binary table)
       if (present(header)) then
          call read_bintab(filename, map(0:,1:), &
               & npixtot, nmaps, fmissing, anynull, header=header(1:), &
               & units=unitsm(1:), extno=extno_i)
       else
          call read_bintab(filename, map(0:,1:), &
               & npixtot, nmaps, fmissing, anynull, units=unitsm(1:), &
               & extno=extno_i)
       endif
       if (present(units)) then
          units(1:size(units)) = unitsm(1:size(units))
       endif
       do imap = 1, nmaps
          anynull = .true.
          if (anynull) then
             imissing = 0
             do i=0,npixtot-1
                if ( ABS(map(i,imap)/fmissing -1.) < 1.e-5 ) then
                   map(i,imap) = fmiss_effct
                   imissing = imissing + 1
                endif
             enddo
          endif
       enddo
       if (imissing > 0) write(*,'(a,1pe11.4)') 'blank value : ' ,fmissing
    else if (type_fits == 3 .and. nmaps == 1) then 
       ! one partial map (in binary table with explicit pixel indexing)
       allocate(pixel(0:obs_npix-1), stat = status)
       allocate(signal(0:obs_npix-1), stat = status)
       call read_fits_cut4(filename, obs_npix, pixel, signal, header=header, units=units1)
       if (present(units)) units = units1
       maxpix = maxval(pixel)
       minpix = maxval(pixel)
       if (maxpix > (npixtot-1) .or. minpix < 0) then
          print*,'map constructed from file '//trim(filename)//', with pixels in ',minpix,maxpix
          print*,' wont fit in array with ',npixtot,' elements'
          stop
       endif
       map(:,1)        = fmiss_effct
       map(pixel(:),1) = signal(:)
       imissing = npixtot - obs_npix
       deallocate(signal)
       deallocate(pixel)
    else 
       print*,'Unable to read the ',nmaps,' required  map(s) from file '//trim(filename)
       print*,'file type = ',type_fits
       stop
    endif
    !-----------------------------------------------------------------------
    if (imissing > 0) then
       write(*,'(i7,a,f7.3,a,1pe11.4)') &
            &           imissing,' missing pixels (', &
            &           (100.*imissing)/npixtot,' %),'// &
            &           ' have been set to : ',fmiss_effct
    endif

!    deallocate(unitsm)

    RETURN
  END subroutine input_map
  !=======================================================================
!   subroutine read_bintab(filename, map, npixtot, nmaps, nullval, anynull, header, nlheader, units)
  subroutine read_bintab(filename, map, npixtot, nmaps, nullval, anynull, header, units, extno)
    !=======================================================================
    !     Read a FITS file
    !
    !     slightly modified to deal with vector column (ie TFORMi = '1024E')
    !     in binary table       EH/IAP/Jan-98
    !      added the output header   EH/Caltech/Nov-00
    !
    !     This routine is used for reading MAPS by anafast.
    !     modified Feb 03 for units argument to run with Sun compiler
    !=======================================================================
    CHARACTER(LEN=*),                          INTENT(IN)  :: filename
    INTEGER(I4B),                              INTENT(IN)  :: npixtot, nmaps
    REAL(SP), DIMENSION(0:,1:),                INTENT(OUT) :: map
    REAL(SP),                                  INTENT(OUT) :: nullval
    LOGICAL(LGT),                              INTENT(OUT) :: anynull
    CHARACTER(LEN=*), dimension(1:), optional, INTENT(OUT) :: header
    CHARACTER(LEN=*), dimension(1:), optional, INTENT(OUT) :: units
    INTEGER(I4B)                   , optional,  INTENT(IN) :: extno

    INTEGER(I4B) :: nl_header, len_header, nl_units, len_units
    INTEGER(I4B) :: status,unit,readwrite,blocksize,naxes(2),nfound, naxis
    INTEGER(I4B) :: group,firstpix,npix,i, npix_old
    REAL(SP) :: blank, testval
    REAL(DP) ::  bscale,bzero
    CHARACTER(LEN=80) :: comment
    LOGICAL(LGT) :: extend
    INTEGER(I4B) :: nmove, hdutype
    INTEGER(I4B) :: column, frow, imap
    INTEGER(I4B) :: datacode, repeat, width
    LOGICAL(LGT) ::  anynull_i

    INTEGER(I4B), PARAMETER :: maxdim=20 !number of columns in the extension
    INTEGER(I4B) :: nrows, tfields, varidat
    CHARACTER(LEN=20), dimension(1:maxdim) :: ttype, tform, tunit
    CHARACTER(LEN=20)                      :: extname
    !-----------------------------------------------------------------------
    status=0

    unit = 150
    naxes(1) = 1
    naxes(2) = 1
    nfound = -1
    anynull = .false.
    bscale = 1.0d0
    bzero = 0.0d0
    blank = -2.e25
    nullval = bscale*blank + bzero

    nl_header = 0
    if (present(header)) then
       nl_header = size(header)
       len_header = 80
    endif

    nl_units = 0
    if (present(units)) then
!        print*,'-------->',len(units(1))
!        print*,'-----==->',len_trim(units(1))
       nl_units = size(units)
       len_units = min(80,len(units(1))) ! due to SUN compiler bug
    endif
    

    readwrite=0
    call ftopen(unit,filename,readwrite,blocksize,status)
    if (status > 0) call printerror(status)
    !     -----------------------------------------

    !     determines the presence of image
    call ftgkyj(unit,'NAXIS', naxis, comment, status)
    if (status > 0) call printerror(status)

    !     determines the presence of an extension
    call ftgkyl(unit,'EXTEND', extend, comment, status)
    if (status > 0) status = 0 ! no extension : 
    !     to be compatible with first version of the code

    if (naxis > 0) then ! there is an image
       !        determine the size of the image (look naxis1 and naxis2)
       call ftgknj(unit,'NAXIS',1,2,naxes,nfound,status)

       !        check that it found only NAXIS1
       if (nfound == 2 .and. naxes(2) > 1) then
          print *,'multi-dimensional image'
          print *,'expected 1-D data.'
          stop
       end if

       if (nfound < 1) then
          call printerror(status)
          print *,'can not find NAXIS1.'
          stop
       endif

       npix=naxes(1)
       if (npix /= npixtot) then
          print *,'WARNING: found ',npix,' pixels in '//trim(filename)
          print *,'         expected ',npixtot
          npix = min(npix, npixtot)
          print *,'         only ',npix,' will be read'
       endif

       call ftgkyd(unit,'BSCALE',bscale,comment,status)
       if (status == 202) then ! BSCALE not found
          bscale = 1.0d0
          status = 0
       endif
       call ftgkyd(unit,'BZERO', bzero, comment,status)
       if (status == 202) then ! BZERO not found
          bzero = 0.0d0
          status = 0
       endif
       call ftgkye(unit,'BLANK', blank, comment,status)
       if (status == 202) then ! BLANK not found 
          ! (according to fitsio BLANK is integer)
          blank = -2.e25
          status = 0
       endif
       nullval = bscale*blank + bzero

       !        -----------------------------------------

       group=1
       firstpix = 1
       call ftgpve(unit,group,firstpix,npix,nullval,map,anynull,status)
       ! if there are any NaN pixels, (real data)
       ! or BLANK pixels (integer data) they will take nullval value
       ! and anynull will switch to .true.
       ! otherwise, switch it by hand if necessary
       testval = 1.e-6 * ABS(nullval)
       do i=0, npix-1
          if (ABS(map(i,1)-nullval) < testval) then
             anynull = .true.
             goto 111
          endif
       enddo
111    continue

    else if (extend) then ! there is an extension
       nmove = +1
       if (present(extno)) nmove = +1 + extno
       call ftmrhd(unit, nmove, hdutype, status)
       !cc         write(*,*) hdutype

       if (hdutype /= 2) then ! not a binary table
          print*,'this is not a binary table'
          stop
       endif

       !        reads all the keywords
       call ftghbn(unit, maxdim, &
            &        nrows, tfields, ttype, tform, tunit, extname, varidat, &
            &        status)

       if (tfields < nmaps) then
          print *,'found ',tfields,' maps in file '//trim(filename)
          print *,'expected ',nmaps
          stop
       endif

       !        finds the bad data value
       call ftgkye(unit,'BAD_DATA',nullval,comment,status)
       if (status == 202) then ! bad_data not found
          nullval = s_bad_value ! default value
          status = 0
       endif

       if (nl_header > 0) then
          do i=1,nl_header
             header(i)(1:len_header) = ""
          enddo
          call get_clean_header(unit, header, filename, status)
          status = 0
       endif

       if (nl_units > 0) then
          do i=1,nl_units
             units(i)(1:len_units) = 'unknown' ! default
          enddo
          do imap = 1, min(nmaps, nl_units)
             units(imap)(1:len_units) = adjustl(tunit(imap))
          enddo
       endif

       npix_old = npixtot
       do imap = 1, nmaps
          !parse TFORM keyword to find out the length of the column vector
          call ftbnfm(tform(imap), datacode, repeat, width, status)
          !reads the columns
          column = imap
          frow = 1
          firstpix = 1
          npix = nrows * repeat
          if (npix /= npixtot .and. npix_old /= npix) then
             npix_old = npix
             print *,'WARNING: found ',npix,' pixels in '//trim(filename)//', column ',imap
             print *,'         expected ',npixtot
             npix = min(npix, npixtot)
             print *,'         only  ',npix,' will be read'
          endif
!           call ftgcve(unit, column, frow, firstpix, npix, nullval, &
!                &        map(0:,imap), anynull_i, status)
          call ftgcve(unit, column, frow, firstpix, npix, nullval, &
               &        map(0,imap), anynull_i, status)
          anynull = anynull .or. anynull_i
       enddo

    else ! no image no extension, you are dead, man
       stop ' No image, no extension'
    endif
    !     close the file
    call ftclos(unit, status)

    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)

    return
  end subroutine read_bintab
  !=======================================================================
  !    subroutine read_conbintab(filename, alms, nalms, header, nlheader, units)
  subroutine read_conbintab(filename, alms, nalms, units, extno)
    !=======================================================================
    !     Read a FITS file containing alms values
    !
    !     slightly modified to deal with vector column 
    !     in binary table       EH/IAP/Jan-98
    !
    !     Used by synfast when reading a binary file with alms for cons.real.
    !                        FKH/Apr-99
    !
    !   extno : optional, number of extension to be read, default=0: first extension
    !=======================================================================
    CHARACTER(LEN=*),                   INTENT(IN) :: filename
    INTEGER(I4B),                       INTENT(IN) :: nalms !, nlheader
    REAL(SP), DIMENSION(0:nalms-1,1:6), INTENT(OUT) :: alms
    REAL(SP)                                        :: nullval
    LOGICAL(LGT)                                    ::  anynull
    !       CHARACTER(LEN=*), dimension(1:),            INTENT(OUT) :: header
    CHARACTER(LEN=*), dimension(1:), optional,  INTENT(OUT) :: units
    INTEGER(I4B)                   , optional,  INTENT(IN) :: extno


!     INTEGER(I4B), DIMENSION(0:nalms-1) :: lm
    INTEGER(I4B), DIMENSION(:), allocatable :: lm
    INTEGER(I4B) :: status,unit,readwrite,blocksize,naxes(2),nfound, naxis
    INTEGER(I4B) :: firstpix,npix
    REAL(SP) :: blank
    REAL(DP) ::  bscale,bzero
    CHARACTER(LEN=80) :: comment
    LOGICAL(LGT) :: extend
    INTEGER(I4B) :: nmove, hdutype
    INTEGER(I4B) :: column, frow, imap
    INTEGER(I4B) :: datacode, repeat, width
    integer(I4B) :: i, l, m

    INTEGER(I4B), PARAMETER :: maxdim=20 !number of columns in the extension
    INTEGER(I4B) :: nrows, tfields, varidat
    CHARACTER(LEN=20), dimension(1:maxdim) :: ttype, tform, tunit
    CHARACTER(LEN=20) :: extname
    character(len=*), parameter :: code="read_conbintab"

    !-----------------------------------------------------------------------
    status=0

    unit = 150
    naxes(1) = 1
    naxes(2) = 1
    nfound = -1
    anynull = .false.
    bscale = 1.0d0
    bzero = 0.0d0
    blank = -2.e25
    nullval = bscale*blank + bzero
    alms=0.
    readwrite=0
    call ftopen(unit,filename,readwrite,blocksize,status)
    if (status > 0) call printerror(status)
    !     -----------------------------------------

    !     determines the presence of image
    call ftgkyj(unit,'NAXIS', naxis, comment, status)
    if (status > 0) call printerror(status)

    !     determines the presence of an extension
    call ftgkyl(unit,'EXTEND', extend, comment, status)
    if (status > 0) status = 0 ! no extension : 
    !     to be compatible with first version of the code

    if (.not.extend) then
       print*,'No extension!'
       stop 
    endif

    ! go to assigned extension (if it exists)
    nmove = +1
    if (present(extno)) nmove = +1 + extno
    call ftmrhd(unit, nmove, hdutype, status)
    if (status > 0) then
       ! if required extension not found: 
       ! print a warning, fill with dummy values, return to calling routine
       print*,code//' WARNING: the extension ',extno,' was not found in ',trim(filename)
       alms(0:nalms-1,1)=0.    ! l = 0
       alms(0:nalms-1,2)=1.    ! m = 1
       alms(0:nalms-1,3:6)=0.
       status = 0
       call ftclos(unit, status)        !     close the file
       return
    endif

    if (hdutype /= 2) then ! not a binary table
       print*, 'this is not a binary table'
       stop
    endif

    !        reads all the keywords
    call ftghbn(unit, maxdim, &
         &        nrows, tfields, ttype, tform, tunit, extname, varidat, &
         &        status)

    if ((tfields/=3).and.(tfields/=5)) then
       print *,'found ',tfields,' columns in the file'
       print *,'expected 3 or 5'
       stop
    endif
    !        finds the bad data value
    call ftgkye(unit,'BAD_DATA',nullval,comment,status)
    if (status == 202) then ! bad_data not found
       nullval = s_bad_value ! default value
       status = 0
    endif

    !          if (nlheader > 0) then
    !             header = ""
    !             status = 0
    !             call get_clean_header(unit, header, filename, status)
    !          endif

    if (present(units)) then
       units(1) = tunit(2)
    endif

    !parse TFORM keyword to find out the length of the column vector
    call ftbnfm(tform(1), datacode, repeat, width, status)

    !reads the columns
    ! first column : index -> (l,m)
    allocate(lm(0:nalms-1))
    column = 1
    frow = 1
    firstpix = 1
    npix = nrows * repeat
    if (npix /= nalms) then
       print *,'found ',npix,' alms'
       print *,'expected ',nalms
       stop
    endif
    call ftgcvj(unit, column, frow, firstpix, npix, nullval, &
         &        lm(0), anynull, status)
    if (anynull) stop 'There are undefined values in the table!'
!     ! multipole l
!     alms(0:nalms-1,1)=REAL(int(sqrt(DBLE(lm(0:nalms-1)-1))))
!     ! m
!     alms(0:nalms-1,2)=REAL(DBLE(lm(0:nalms-1)))-1.d0-alms(0:nalms-1,1)**2-alms(0:nalms-1,1)
    ! recoding of the mapping, EH, 2002-08-08
    ! lm = l*l + l + m + 1
    do i = 0, nalms - 1
       l = int(sqrt(   real(lm(i)-1, kind = DP)  ) )
       m = lm(i) - l*(l+1) - 1
       ! check that round-off did not screw up mapping
       if (abs(m) > l) then
          print*,'Inconsistent l^2+l+m+1 -> l,m mapping'
          print*,l, m, l*(l+1)+m+1, lm(i)
          stop
       endif
       alms(i,1) = real( l )
       alms(i,2) = real( m )
    enddo
    deallocate(lm)


    do imap = 2, tfields
       !parse TFORM keyword to find out the length of the column vector
       call ftbnfm(tform(imap), datacode, repeat, width, status)

       !reads the columns
       column = imap
       frow = 1
       firstpix = 1
       npix = nrows * repeat
       if (npix /= nalms) then
          print *,'found ',npix,' alms'
          print *,'expected ',nalms
          stop
       endif
       call ftgcve(unit, column, frow, firstpix, npix, nullval, &
            &        alms(0,imap+1), anynull, status)
       if (anynull) stop 'There are undefined values in the table!'
    enddo
    !     close the file
    call ftclos(unit, status)


    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)
    return
  end subroutine read_conbintab

  !=======================================================================
  !    subroutine read_dbintab(filename,map,npixtot,nmaps,nullval,anynull, header, nlheader, units)
  subroutine read_dbintab(filename,map,npixtot,nmaps,nullval,anynull, units)
    !=======================================================================
    !     Read a FITS file
    !
    !     slightly modified to deal with vector column 
    !     in binary table       EH/IAP/Jan-98
    !
    !     Reads a double-precision array with precomputed plms, used by syn/anafast
    !                FKH/Apr-99
    !=======================================================================
    CHARACTER(LEN=*),                           INTENT(IN) :: filename
    INTEGER(I4B),                               INTENT(IN) :: npixtot, nmaps
    !       REAL(DP), DIMENSION(0:npixtot-1,1:nmaps),   INTENT(OUT) :: map
    REAL(DP), DIMENSION(0:,1:),                 INTENT(OUT) :: map
    REAL(DP),                                   INTENT(OUT) :: nullval
    LOGICAL(LGT),                               INTENT(OUT) ::  anynull
    !       CHARACTER(LEN=*), dimension(1:), optional,           INTENT(OUT) :: header
    CHARACTER(LEN=*), dimension(1:), optional,  INTENT(OUT) :: units

    INTEGER(I4B) :: status,unit,readwrite,blocksize,naxes(2),nfound, naxis
    INTEGER(I4B) :: group,firstpix,npix,i
    REAL(SP) :: blank, testval
    REAL(DP) ::  bscale,bzero
    CHARACTER(LEN=80) :: comment
    LOGICAL(LGT) :: extend
    INTEGER(I4B) :: nmove, hdutype
    INTEGER(I4B) :: column, frow, imap
    INTEGER(I4B) :: datacode, repeat, width

    INTEGER(I4B), PARAMETER :: maxdim=20 !number of columns in the extension
    INTEGER(I4B) :: nrows, tfields, varidat
    CHARACTER(LEN=20), dimension(1:maxdim) :: ttype, tform, tunit
    CHARACTER(LEN=20)                      :: extname
    !-----------------------------------------------------------------------
    status=0

    unit = 150
    naxes(1) = 1
    naxes(2) = 1
    nfound = -1
    anynull = .false.
    bscale = 1.0d0
    bzero = 0.0d0
    blank = -2.e25
    nullval = bscale*blank + bzero

    readwrite=0
    call ftopen(unit,filename,readwrite,blocksize,status)
    if (status > 0) call printerror(status)
    !     -----------------------------------------

    !     determines the presence of image
    call ftgkyj(unit,'NAXIS', naxis, comment, status)
    if (status > 0) call printerror(status)

    !     determines the presence of an extension
    call ftgkyl(unit,'EXTEND', extend, comment, status)
    if (status > 0) status = 0 ! no extension : 
    !     to be compatible with first version of the code

    if (naxis > 0) then ! there is an image
       !        determine the size of the image (look naxis1 and naxis2)
       call ftgknj(unit,'NAXIS',1,2,naxes,nfound,status)

       !        check that it found only NAXIS1
       if (nfound == 2 .and. naxes(2) > 1) then
          print *,'multi-dimensional image'
          print *,'expected 1-D data.'
          stop
       end if

       if (nfound < 1) then
          call printerror(status)
          print *,'can not find NAXIS1.'
          stop
       endif

       npix=naxes(1)
       if (npix /= npixtot) then
          print *,'found ',npix,' plms'
          print *,'expected ',npixtot
          stop
       endif

       call ftgkyd(unit,'BSCALE',bscale,comment,status)
       if (status == 202) then ! BSCALE not found
          bscale = 1.0d0
          status = 0
       endif
       call ftgkyd(unit,'BZERO', bzero, comment,status)
       if (status == 202) then ! BZERO not found
          bzero = 0.0d0
          status = 0
       endif
       call ftgkye(unit,'BLANK', blank, comment,status)
       if (status == 202) then ! BLANK not found 
          ! (according to fitsio BLANK is integer)
          blank = -2.e25
          status = 0
       endif
       nullval = bscale*blank + bzero

       !        -----------------------------------------

       group=1
       firstpix = 1
       call ftgpvd(unit,group,firstpix,npix,nullval,map,anynull,status)
       ! if there are any NaN pixels, (real data)
       ! or BLANK pixels (integer data) they will take nullval value
       ! and anynull will switch to .true.
       ! otherwise, switch it by hand if necessary
       testval = 1.e-6 * ABS(nullval)
       do i=0, npix-1
          if (ABS(map(i,1)-nullval) < testval) then
             anynull = .true.
             goto 111
          endif
       enddo
111    continue

    else if (extend) then ! there is an extension
       nmove = +1
       call ftmrhd(unit, nmove, hdutype, status)
       !cc         write(*,*) hdutype

       if (hdutype /= 2) then ! not a binary table
          stop 'this is not a binary table'
       endif

       !        reads all the keywords
       call ftghbn(unit, maxdim, &
            &        nrows, tfields, ttype, tform, tunit, extname, varidat, &
            &        status)

       if (tfields < nmaps) then
          print *,'found ',tfields,' maps in the file'
          print *,'expected ',nmaps
          stop
       endif

       !        finds the bad data value
       call ftgkyd(unit,'BAD_DATA',nullval,comment,status)
       if (status == 202) then ! bad_data not found
          nullval = d_bad_value
          status = 0
       endif

       !          if (nlheader > 0) then
       !             call get_clean_header(unit, header, filename, status)
       !          endif

       if (present(units)) then
          units = 'unknown'
          do imap = 1, nmaps
             units(imap) = adjustl(tunit(imap))
          enddo
       endif

       do imap = 1, nmaps
          !parse TFORM keyword to find out the length of the column vector
          call ftbnfm(tform(imap), datacode, repeat, width, status)

          !reads the columns
          column = imap
          frow = 1
          firstpix = 1
          npix = nrows * repeat
          if (npix /= npixtot) then
             print *,'found ',npix,' plms'
             print *,'expected ',npixtot
             stop
          endif
          call ftgcvd(unit, column, frow, firstpix, npix, nullval, &
               &        map(0,imap), anynull, status)
       enddo

    else ! no image no extension, you are dead, man
       stop ' No image, no extension'
    endif

    !     close the file
    call ftclos(unit, status)

    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)

    return
  end subroutine read_dbintab

  !=======================================================================
  subroutine read_asctab(filename, clin, lmax, ncl, header, nlheader, units)
    !=======================================================================
    !     Read an ASCII table FITS file
    !   Aug 2000 : modification by EH of the number of columns actually read
    !
    !     This routine is used for reading input power spectra for anafast
    !=======================================================================
    CHARACTER(LEN=*),                          INTENT(IN) :: filename
    INTEGER(I4B),                              INTENT(IN) :: lmax, ncl, nlheader
    REAL(SP),         DIMENSION(0:,1:), INTENT(OUT) :: clin
    CHARACTER(LEN=*), DIMENSION(1:),   INTENT(OUT) :: header
    CHARACTER(LEN=*), dimension(1:), optional,  INTENT(OUT) :: units

    INTEGER(I4B) :: status,unit,readwrite,blocksize,naxes(2),ncl_file, naxis
    INTEGER(I4B) :: firstpix,lmax_file,lmax_min,nelems
    CHARACTER(LEN=80) :: comment, pdmstr
    LOGICAL :: extend
    INTEGER(I4B) :: nmove, hdutype
    INTEGER(I4B) :: column, frow
    REAL(SP), DIMENSION(:), ALLOCATABLE :: clin_file

    INTEGER(I4B), PARAMETER :: maxdim = 20 !number of columns in the extension
    INTEGER(I4B) :: rowlen, nrows
    INTEGER(I4B),      dimension(1:maxdim) :: tbcol
    CHARACTER(LEN=20), dimension(1:maxdim) :: ttype, tform, tunit
    CHARACTER(LEN=20)                      :: extname
    LOGICAL :: anynull
    REAL(SP) ::  nullval
    logical :: planck_format


    !-----------------------------------------------------------------------
    status=0
    anynull = .false. 

    unit = 150
    naxes(1) = 1
    naxes(2) = 1

    readwrite=0
    call ftopen(unit,filename,readwrite,blocksize,status)
    if (status > 0) call printerror(status)
    !     -----------------------------------------

    !     determines the presence of image
    call ftgkyj(unit,'NAXIS', naxis, comment, status)
    if (status > 0) call printerror(status)

    !     determines the presence of an extension
    call ftgkyl(unit,'EXTEND', extend, comment, status)
    if (status > 0) call printerror(status)

    if (extend) then ! there is an extension
       nmove = +1
       call ftmrhd(unit, nmove, hdutype, status)

       if (hdutype /= 1) then ! not an ascii table
          stop 'this is not an acii table'
       endif

       !        reads keywords related to table layout
       call ftghtb(unit, maxdim, &
            &        rowlen, nrows, ncl_file, ttype, tbcol, tform, tunit, &
            &        extname, status)

       status = 0
       header = ""
       call get_clean_header(unit, header, filename, status)

       if (present(units)) then 
          do column = 1, min(ncl, ncl_file)
             units(column) = adjustl(tunit(column))
          enddo
       endif

       !        reads the columns
       column = 1
       frow = 1
       firstpix = 1
       lmax_file = nrows - 1
       lmax_min = MIN(lmax,lmax_file)
       nullval = 0.0
       nelems = lmax_min + 1

! check for the special Planck format (i.e. one additional column)
       planck_format=.true.
       call ftgkys(unit,'PDMTYPE',pdmstr,comment,status)
       if (status==202) then
         planck_format=.false.
         status=0
       endif
       allocate(clin_file(0:lmax_min),stat=status)
       clin = 0.                         ! modification by EH
       if (planck_format) then
         do column = 1, MIN(ncl, ncl_file-1) ! modification by EH
            clin_file(:) = 0.
            call ftgcve(unit, column+1, frow, firstpix, nelems, nullval, &
                 &        clin_file(0), anynull, status)
            clin(0:lmax_min, column) = clin_file(0:lmax_min)
         enddo
       else
         do column = 1, MIN(ncl, ncl_file) ! modification by EH
            clin_file(:) = 0.
            call ftgcve(unit, column, frow, firstpix, nelems, nullval, &
                 &        clin_file(0), anynull, status)
            clin(0:lmax_min, column) = clin_file(0:lmax_min)
         enddo
       endif
       deallocate(clin_file)
    else ! no image no extension, you are dead, man
       stop ' No image, no extension'
    endif

    !     close the file
    call ftclos(unit, status)

    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)

    return
  end subroutine read_asctab
  !=======================================================================
  subroutine output_map(map, header, outfile, extno)
    !=======================================================================
    ! this is only a wrapper to write_bintab
    ! this routine needs an explicit interface
    !=======================================================================

    REAL(SP),          INTENT(IN), DIMENSION(0:,1:) :: map
    CHARACTER(LEN=*), INTENT(IN), DIMENSION(1:) :: header
    CHARACTER(LEN=*),  INTENT(IN)               :: outfile
    INTEGER(I4B)    , INTENT(IN)     , optional :: extno

    INTEGER(I4B) :: npix, nlheader, nmap, extno_i
    !-----------------------------------------------------------------------
    npix = SIZE(map,1)
    nmap = SIZE(map)/npix
    nlheader = SIZE(header)

    extno_i = 0
    if (present(extno)) extno_i = extno
    call write_bintab(map(0:npix-1,1:nmap), npix, nmap, &
         &            header(1:nlheader), nlheader, outfile, extno=extno_i)
    return
  end subroutine output_map
  !=======================================================================
  subroutine write_bintab(map, npix, nmap, header, nlheader, filename, extno)
    !=======================================================================
    !     Create a FITS file containing a binary table extension with 
    !     the temperature map in the first column
    !     written by EH from writeimage and writebintable 
    !     (fitsio cookbook package)
    !
    !     slightly modified to deal with vector column (ie TFORMi = '1024E')
    !     in binary table       EH/IAP/Jan-98
    !
    !     simplified the calling sequence, the header sould be filled in
    !     before calling the routine
    !=======================================================================

    INTEGER(I4B), INTENT(IN) :: npix, nmap, nlheader
    REAL(SP),          INTENT(IN), DIMENSION(0:npix-1,1:nmap) :: map
    CHARACTER(LEN=*),  INTENT(IN), DIMENSION(1:nlheader) :: header
    CHARACTER(LEN=*),  INTENT(IN)               :: filename
    INTEGER(I4B)    , INTENT(IN)     , optional :: extno

    INTEGER(I4B) ::  status,unit,blocksize,bitpix,naxis,naxes(1)
    INTEGER(I4B) ::  i
    LOGICAL(LGT) ::  simple,extend
    CHARACTER(LEN=80) :: comment

    INTEGER(I4B), PARAMETER :: maxdim = 20 !number of columns in the extension
    INTEGER(I4B) :: nrows, tfields, varidat
    INTEGER(I4B) :: frow,  felem, colnum, hdutype
    CHARACTER(LEN=20) :: ttype(maxdim), tform(maxdim), tunit(maxdim), extname
    CHARACTER(LEN=10) ::  card
    CHARACTER(LEN=2) :: stn
    INTEGER(I4B) :: itn, extno_i
    character(len=filenamelen) sfilename
    character(len=8) ::  date
    character(len=10) ::  fulldate
    !-----------------------------------------------------------------------

    status=0
    unit = 100
    blocksize=1

    extno_i = 0
    if (present(extno)) extno_i = extno

    if (extno_i == 0) then
       !*************************************
       !     create the new empty FITS file
       !*************************************
       call ftinit(unit,filename,blocksize,status)

       !     -----------------------------------------------------
       !     initialize parameters about the FITS image
       simple=.true.
       bitpix=32     ! integer*4
       naxis=0       ! no image
       naxes(1)=0
       extend=.true. ! there is an extension

       !     ----------------------
       !     primary header
       !     ----------------------
       !     write the required header keywords
       call ftphpr(unit,simple,bitpix,naxis,naxes,0,1,extend,status)

       !     writes supplementary keywords : none

       !     write the current date
       call ftpdat(unit,status) ! format (yyyy-mm-dd)

       !     update the date (format ccyy-mm-dd)
       call date_and_time(date)
       fulldate = date(1:4)//'-'//date(5:6)//'-'//date(7:8)
       comment = 'FITS file creation date ccyy-mm-dd'
       call ftukys(unit,'DATE',fulldate,comment,status)

       !     ----------------------
       !     image : none
       !     ----------------------

       !     ----------------------
       !     extension
       !     ----------------------
    else

       !*********************************************
       !     reopen an existing file and go to the end
       !*********************************************
       ! remove the leading '!' (if any) when reopening the same file
       sfilename = adjustl(filename)
       if (sfilename(1:1) == '!') sfilename = sfilename(2:filenamelen)
       call ftopen(unit,sfilename,1,blocksize,status)
       call ftmahd(unit,1+extno_i,hdutype,status)

    endif

    !     creates an extension
    call ftcrhd(unit, status)

    !     writes required keywords
    nrows    = npix / 1024 ! naxis1
    tfields  = nmap
    tform(1:nmap) = '1024E'
    if (npix < 1024) then ! for nside <= 8
       nrows = npix
       tform(1:nmap) = '1E'
    endif
    ttype(1:nmap) = 'simulation'   ! will be updated
    tunit(1:nmap) = ''      ! optional, will not appear
    extname  = ''      ! optional, will not appear
    varidat  = 0
    call ftphbn(unit, nrows, tfields, ttype, tform, tunit, &
         &     extname, varidat, status)

    !     write the header literally, putting TFORM1 at the desired place
    comment = 'data format of field: 4-byte REAL'
    do i=1,nlheader
       card = header(i)
       if (card(1:5) == 'TTYPE') then ! if TTYPE1 is explicitely given
          stn = card(6:6)
          read(stn,'(i1)') itn
          if (itn > tfields) goto 10
          ! discard at their original location:
          call ftdkey(unit,'TTYPE'//stn,status)  ! old TTYPEi and
          status = 0
          call ftdkey(unit,'TFORM'//stn,status)  !     TFORMi
          status = 0
          call putrec(unit,header(i), status)    ! write new TTYPE1
          status = 0
          call ftpkys(unit,'TFORM'//stn,tform(1),comment,status) ! and write new TFORM1 right after
       elseif (header(i)/=' ') then
          call putrec(unit,header(i), status)
       endif
10     continue
       status = 0
    enddo

    !     write the extension one column by one column
    frow   = 1  ! starting position (row)
    felem  = 1  ! starting position (element)
    do colnum = 1, nmap
       call ftpcle(unit, colnum, frow, felem, npix, map(0,colnum), status)
    enddo

    !     ----------------------
    !     close and exit
    !     ----------------------

    !     close the file and free the unit number
    call ftclos(unit, status)

    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)

    return
  end subroutine write_bintab

  !=======================================================================
  subroutine write_dbintab(plm, nplm, nhar, header, nlheader, filename,nsmax,nlmax)
    !=======================================================================
    !     Create a FITS file containing a binary table extension with 
    !     the temperature map in the first column
    !     written by EH from writeimage and writebintable 
    !     (fitsio cookbook package)
    !
    !     slightly modified to deal with vector column 
    !     in binary table       EH/IAP/Jan-98
    !
    !     simplified the calling sequence, the header sould be filled in
    !     before calling the routine
    !
    !     Changed to write double precision plms for const.real. FKH/Apr-99
    !
    !=======================================================================

    INTEGER(I4B), INTENT(IN) :: nplm, nhar, nlheader, nsmax, nlmax
    REAL(DP),          INTENT(IN), DIMENSION(0:nplm-1,1:nhar) :: plm
    CHARACTER(LEN=80), INTENT(IN), DIMENSION(1:nlheader) :: header
    CHARACTER(LEN=*),  INTENT(IN)               :: filename
    INTEGER(I4B) ::  status,unit,blocksize,bitpix,naxis,naxes(1)
    INTEGER(I4B) ::  i
    LOGICAL(LGT) ::  simple,extend
    CHARACTER(LEN=80) :: comment

    INTEGER(I4B), PARAMETER :: maxdim = 20 !number of columns in the extension
    INTEGER(I4B) :: nrows, tfields, varidat
    INTEGER(I4B) :: frow,  felem, colnum
    CHARACTER(LEN=20) :: ttype(maxdim), tform(maxdim), tunit(maxdim), extname
    CHARACTER(LEN=8) ::  date
    CHARACTER(LEN=10) ::  fulldate
    CHARACTER(LEN=10) ::  card, num
    CHARACTER(LEN=2) :: stn
    INTEGER(I4B) :: itn
    !-----------------------------------------------------------------------

    status=0

    unit = 100

    !     create the new empty FITS file
    blocksize=1
    call ftinit(unit,filename,blocksize,status)

    !     -----------------------------------------------------
    !     initialize parameters about the FITS image
    simple=.true.
    bitpix=32     ! integer*4
    naxis=0       ! no image
    naxes(1)=0
    extend=.true. ! there is an extension

    !     ----------------------
    !     primary header
    !     ----------------------
    !     write the required header keywords
    call ftphpr(unit,simple,bitpix,naxis,naxes,0,1,extend,status)

    !     writes supplementary keywords : none

    !     write the current date
    call ftpdat(unit,status) ! format (dd/mm/yy)

    !     update the date (format ccyy-mm-dd)
    call date_and_time(date)
    fulldate = date(1:4)//'-'//date(5:6)//'-'//date(7:8)
    comment = 'FITS file creation date ccyy-mm-dd'
    call ftukys(unit,'DATE',fulldate,comment,status)

    !     ----------------------
    !     image : none
    !     ----------------------

    !     ----------------------
    !     extension
    !     ----------------------

    !     creates an extension
    call ftcrhd(unit, status)

    !     writes required keywords
    nrows    = nsmax ! naxis1
    tfields  = nhar
    write (num,'(I10)') (nlmax+1)*(nlmax+2)
    tform(1:nhar) = trim(adjustl(num))//'D'     
    ttype(1:nhar) = 'harmonics'   ! will be updated
    tunit(1:nhar) = ''      ! optional, will not appear
    extname  = ''      ! optional, will not appear
    varidat  = 0
    call ftphbn(unit, nrows, tfields, ttype, tform, tunit, &
         &     extname, varidat, status)

    !     write the header literally, putting TFORM1 at the desired place
    comment = 'data format of field: 8-byte REAL'
    do i=1,nlheader
       card = header(i)
       if (card(1:5) == 'TTYPE') then ! if TTYPE1 is explicitely given
          stn = card(6:6)
          read(stn,'(i1)') itn
          if (itn > tfields) goto 10
          ! discard at their original location:
          call ftdkey(unit,'TTYPE'//stn,status)  ! old TTYPEi and
          status = 0
          call ftdkey(unit,'TFORM'//stn,status)  !     TFORMi
          status = 0
          call putrec(unit,header(i), status)           ! write new TTYPE1
          status = 0
          call ftpkys(unit,'TFORM'//stn,tform(1),comment,status) ! and write new TFORM1 right after
       elseif (header(i)/=' ') then
          call putrec(unit,header(i), status)
       endif
10     continue
       status = 0
    enddo

    !     write the extension one column by one column
    frow   = 1  ! starting position (row)
    felem  = 1  ! starting position (element)
    do colnum = 1, nhar
       call ftpcld(unit, colnum, frow, felem, nplm, plm(0,colnum), status)
    enddo

    !     ----------------------
    !     close and exit
    !     ----------------------

    !     close the file and free the unit number
    call ftclos(unit, status)

    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)

    return
  end subroutine write_dbintab
  !=======================================================================
  subroutine write_asctab &
       &  (clout, lmax, ncl, header, nlheader, filename)
    !=======================================================================
    !     Create a FITS file containing an ASCII table extension with 
    !     the measured power spectra
    !     written by EH from writeimage and writeascii 
    !     (fitsio cookbook package)
    !
    !     clout = power spectra with l in (0:lmax)
    !     ncl = number of spectra
    !     header = FITS header to be put on top of the file
    !     nlheader = number of lines of the header
    !     filename = FITS output file name
    !=======================================================================

    INTEGER(I4B), INTENT(IN) :: lmax, ncl,nlheader
    REAL(SP), INTENT(IN)  ::  clout(0:lmax,1:ncl)
    CHARACTER(LEN=80), INTENT(IN), DIMENSION(1:nlheader) :: header
    CHARACTER(LEN=*),  INTENT(IN)               :: filename

    INTEGER(I4B) :: bitpix,naxis,naxes(1)
    LOGICAL(LGT) :: simple,extend
    CHARACTER(LEN=8) ::  date
    CHARACTER(LEN=10) ::  fulldate
    CHARACTER(LEN=10) ::  card

    INTEGER(I4B), PARAMETER :: nclmax = 20
    INTEGER(I4B) ::  status,unit,blocksize,tfields,nrows,rowlen
    INTEGER(I4B) ::  nspace,tbcol(nclmax),colnum,frow,felem
    CHARACTER(LEN=16) :: ttype(nclmax),tform(nclmax),tunit(nclmax),extname
    CHARACTER(LEN=80) :: comment, card_tbcol
    CHARACTER(LEN=2) :: stn
    INTEGER(I4B) :: itn, i
    !=======================================================================

    status=0

    unit = 87

    !     open the FITS file, with write access
    !      readwrite=1
    !      call ftopen(unit,filename,readwrite,blocksize,status)

    blocksize=1
    call ftinit(unit,filename,blocksize,status)

    !     -----------------------------------------------------
    !     initialize parameters about the FITS image
    simple=.true.
    bitpix=32     ! integer*4
    naxis=0       ! no image
    naxes(1)=0
    extend=.true. ! there is an extension

    !     ----------------------
    !     primary header
    !     ----------------------
    !     write the required header keywords
    call ftphpr(unit,simple,bitpix,naxis,naxes,0,1,extend,status)

    !     writes supplementary keywords : none

    !     write the current date
    call ftpdat(unit,status) ! format (dd/mm/yy)

    !     update the date (format ccyy-mm-dd)
    call date_and_time(date)
    fulldate = date(1:4)//'-'//date(5:6)//'-'//date(7:8)
    comment = 'FITS file creation date ccyy-mm-dd'
    call ftukys(unit,'DATE',fulldate,comment,status)

    !     ----------------------
    !     image : none
    !     ----------------------

    !     ----------------------
    !     extension
    !     ----------------------

    !     append a new empty extension onto the end of the primary array
    call ftcrhd(unit,status)

    !     define parameters for the ASCII table
    nrows   = lmax+1
    tfields = ncl
    tform(1:ncl) = 'E15.7'
    ttype(1:ncl) = 'power spectrum' ! is updated by the value given in the header
    tunit(1:ncl) = '' ! optional, will not appear
    extname      = '' ! optional, will not appear

    !     calculate the starting position of each column, and the total row length
    nspace=1
    call ftgabc(tfields,tform,nspace,rowlen,tbcol,status)

    !     write the required header parameters for the ASCII table
    call ftphtb(unit,rowlen,nrows,tfields,ttype,tbcol,tform,tunit, &
         &            extname,status)

    !     write the header literally, putting TFORM1 at the desired place
    comment = ''
    do i=1,nlheader
       card = header(i)
       if (card(1:5) == 'TTYPE') then ! if TTYPE1 is explicitely given
          stn = card(6:6)
          read(stn,'(i1)') itn
          ! discard at their original location:
          call ftdkey(unit,card(1:6),status)  !         old TTYPEi
          status = 0
          call ftdkey(unit,'TFORM'//stn,status) !           TFORMi
          status = 0
          call ftgcrd(unit,'TBCOL'//stn,card_tbcol,status)
          status = 0
          call ftdkey(unit,'TBCOL'//stn,status) !           TBCOLi
          ! and rewrite
          status = 0
          call putrec(unit,card_tbcol,status) !             TBCOLi
          status = 0
          call ftpkys(unit,'TFORM'//stn,tform(itn),comment,status) ! TFORMi right after
          status = 0
          call putrec(unit,header(i), status)   !           TTYPEi
       elseif (header(i)/=' ') then
          call putrec(unit,header(i), status)
       endif
10     continue
       status = 0
    enddo

    !     write names to the first column, diameters to 2nd col., and density to 3rd
    frow=1
    felem=1
    do colnum = 1, ncl
       call ftpcle(unit,colnum,frow,felem,nrows,clout(0,colnum),status)  
    enddo

    !     close the FITS file
    call ftclos(unit, status)

    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)

    return
  end subroutine write_asctab
  !=======================================================================
  subroutine dump_alms(filename,alms,nlmax,header,nlheader,extno)
    !=======================================================================

    !=======================================================================
    !     Create/extend a FITS file containing a binary table extension with 
    !     the a_lm coefficients in each extension.
    !
    !     Format of alm FITS file: in each extension, 3 columns: 
    !            index=l*l+l+m+1, real(a_lm), imag(a_lm)
    !      the a_lm are obtained using the complex Y_lm
    !      only the modes with m>=0 are stored (because the map is assumed real)
    !
    !     The extensions contain, in general, T, E (= G) and B (= C) in that order
    !     First extension has extno = 0
    !
    !     Adapted from write_bintab, FKH/Apr-99
    !
    !=======================================================================

    INTEGER(I4B), INTENT(IN) :: nlmax, nlheader, extno
    COMPLEX(SPC),          INTENT(IN), DIMENSION(0:nlmax,0:nlmax) :: alms
    CHARACTER(LEN=80), INTENT(IN), DIMENSION(1:nlheader) :: header
    CHARACTER(LEN=*),  INTENT(IN)               :: filename
    ! local variables
    INTEGER(I4B), DIMENSION(:),   allocatable :: lm
    REAL(SP),     DIMENSION(:,:), allocatable :: alms_out
!     INTEGER(I4B), DIMENSION(0:(nlmax+1)*(nlmax+2)/2-1) :: lm
!     REAL(SP), DIMENSION(0:(nlmax+1)*(nlmax+2)/2,1:2) :: alms_out
    INTEGER(I4B) ::  status,unit,blocksize,bitpix,naxis,naxes(1)
    INTEGER(I4B) ::  i,l,m,cnt,hdutype
    LOGICAL(LGT) ::  simple,extend
    CHARACTER(LEN=80) :: comment

    INTEGER(I4B), PARAMETER :: maxdim = 20 !number of columns in the extension
    INTEGER(I4B) :: nrows, npix, tfields, varidat
    INTEGER(I4B) :: frow,  felem, colnum
    CHARACTER(LEN=20) :: ttype(maxdim), tform(maxdim), tunit(maxdim), extname
    CHARACTER(LEN=8) ::  date
    CHARACTER(LEN=10) ::  fulldate
    CHARACTER(LEN=10) ::  card
    CHARACTER(LEN=2) :: stn
    INTEGER(I4B) :: itn
    character(len=filenamelen) sfilename
    !-----------------------------------------------------------------------

    !      open (12,file='almdunp.unf',status='new',form='unformatted')
    !      write (12) alms(0:nlmax,0:nlmax)
    !      close (12)

!     npix=(nlmax+1)*(nlmax+2)/2
    npix=((nlmax+1)*(nlmax+2))/2
    allocate(lm      (0:npix-1))
    allocate(alms_out(0:npix-1,1:2))

    status=0

    unit = 100

    blocksize=1

    if (extno==0) then
       !*********************************************
       !     create the new empty FITS file
       !*********************************************
       call ftinit(unit,filename,blocksize,status)

       !     -----------------------------------------------------
       !     initialize parameters about the FITS image
       simple=.true.
       bitpix=32     ! integer*4
       naxis=0       ! no image
       naxes(1)=0
       extend=.true. ! there is an extension

       !     ----------------------
       !     primary header
       !     ----------------------
       !     write the required header keywords
       call ftphpr(unit,simple,bitpix,naxis,naxes,0,1,extend,status)

       !     writes supplementary keywords : none

       !     write the current date
       call ftpdat(unit,status) ! format (yyyy-mm-dd)

       !     update the date (format ccyy-mm-dd)
       call date_and_time(date)
       fulldate = date(1:4)//'-'//date(5:6)//'-'//date(7:8)
       comment = 'FITS file creation date ccyy-mm-dd'
       call ftukys(unit,'DATE',fulldate,comment,status)

       !     ----------------------
       !     image : none
       !     ----------------------

       !     ----------------------
       !     extension
       !     ----------------------

    else

       !*********************************************
       !     reopen an existing file and go to the end
       !*********************************************
       ! remove the leading '!' (if any) when reopening the same file
       sfilename = adjustl(filename)
       if (sfilename(1:1) == '!') sfilename = sfilename(2:filenamelen)
       call ftopen(unit,sfilename,1,blocksize,status)
       call ftmahd(unit,1+extno,hdutype,status)

    endif

    !     creates an extension
    call ftcrhd(unit, status)

    !     writes required keywords
    nrows    = npix  ! naxis1
    tfields  = 3
    tform(1)='1J'
    tform(2:3) = '1E'
    ttype(1) = 'index=l^2+l+m+1'
    ttype(2) = 'alm (real)'
!     ttype(3) = 'alm (complex)' 
    ttype(3) = 'alm (imaginary)' 
    tunit(1:3) = ''      ! optional, will not appear
    extname  = ''      ! optional, will not appear
    varidat  = 0
    call ftphbn(unit, nrows, tfields, ttype, tform, tunit, &
         &     extname, varidat, status)

    !     write the header literally, putting TFORM1 at the desired place
    do i=1,nlheader
       card = header(i)
       if (card(1:5) == 'TTYPE') then ! if TTYPE1 is explicitely given
          stn = card(6:6)
          read(stn,'(i1)') itn
          ! discard at their original location:
          call ftdkey(unit,'TTYPE'//stn,status)  ! old TTYPEi and  ! remove
          status = 0
          call ftdkey(unit,'TFORM'//stn,status)  !     TFORMi
          status = 0
          call putrec(unit,header(i), status)           ! write new TTYPE1
          if (itn==1) then
             comment = 'data format of field: 4-byte INTEGER'
          else
             comment = 'data format of field: 4-byte REAL'
          endif
          status = 0
          call ftpkys(unit,'TFORM'//stn,tform(itn),comment,status) ! and write new TFORM1 right after
       elseif (header(i)/=' ') then
          call putrec(unit,header(i), status)
       endif
10     continue
       status = 0
    enddo

    !     write the extension one column by one column
    frow   = 1  ! starting position (row)
    felem  = 1  ! starting position (element)
    cnt=0
    do l=0,nlmax
       do m=0,l
          lm(cnt)=l**2+l+m+1
          alms_out(cnt,1)=REAL(alms(l,m))
          alms_out(cnt,2)=AIMAG(alms(l,m))
          cnt=cnt+1
       enddo
    enddo

    !      open (12,file='index.unf',status='new',form='unformatted')
    !      write (12) lm(0:cnt-1)
    !      close (12)

    !      open (12,file='almdump.unf',status='new',form='unformatted')
    !      write (12) alms_out(0:cnt-1,1:2)
    !      close (12)



    call ftpclj(unit, 1, frow, felem, npix, lm(0), status)
    do colnum = 2, 3
       call ftpcle(unit, colnum, frow, felem, npix, alms_out(0,colnum-1), status)
    enddo

    !     ----------------------
    !     close and exit
    !     ----------------------

    !     close the file and free the unit number
    call ftclos(unit, status)

    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)

    deallocate(lm)
    deallocate(alms_out)

    return
  end subroutine dump_alms

  !=======================================================================
  subroutine alms2fits(filename,nalms,alms,ncl,header,nlheader,next)
    !=======================================================================

    INTEGER(I4B), INTENT(IN) :: nalms, nlheader, next, ncl
    REAL(SP),          INTENT(IN), DIMENSION(1:nalms,1:(ncl+1),1:next) :: alms
    CHARACTER(LEN=80), INTENT(IN), DIMENSION(1:nlheader,1:next) :: header
    CHARACTER(LEN=*),  INTENT(IN)               :: filename
    INTEGER(I4B) :: i

    do i=1,next

       call write_alms(filename,nalms,alms(1:nalms,1:(ncl+1),i),ncl,header(1:nlheader,i),nlheader,i)

    enddo

  end subroutine alms2fits

  !=======================================================================
  subroutine fits2alms(filename,nalms,alms,ncl,header,nlheader,next) 
    !=======================================================================
    CHARACTER(LEN=*),               INTENT(IN) :: filename
    INTEGER(I4B),                        INTENT(IN) :: nalms, ncl,nlheader,next
    REAL(SP), DIMENSION(1:nalms,1:(ncl+1),1:next), INTENT(OUT) :: alms
    CHARACTER(LEN=80), INTENT(OUT), DIMENSION(1:nlheader,1:next) :: header
    INTEGER(I4B) :: i

    do i=1,next

       call read_alms(filename,nalms,alms(1:nalms,1:(ncl+1),i),ncl,header(1:nlheader,i),nlheader,i)

    enddo

  end subroutine fits2alms

  !=======================================================================
  subroutine write_alms(filename,nalms,alms,ncl,header,nlheader,extno)
    !=======================================================================
    
    !=======================================================================
    !     Create a FITS file containing a binary table extension with 
    !     the temperature map in the first column
    !     written by EH from writeimage and writebintable 
    !     (fitsio cookbook package)
    !
    !     slightly modified to deal with vector column
    !     in binary table       EH/IAP/Jan-98
    !
    !     simplified the calling sequence, the header sould be filled in
    !     before calling the routine
    !
    !     Changed to write alms from anafast to binary FITS file, FKH/Apr-99
    !     ncl is the number of columns, either 3 or 5 (with or without errors)
    !
    !=======================================================================
    
    INTEGER(I4B), INTENT(IN) :: nalms, nlheader, ncl, extno
    REAL(SP),          INTENT(IN), DIMENSION(0:nalms-1,1:(ncl+1)) :: alms
    CHARACTER(LEN=80), INTENT(IN), DIMENSION(1:nlheader) :: header
    CHARACTER(LEN=*),  INTENT(IN)               :: filename
    INTEGER(I4B), DIMENSION(0:nalms-1) :: lm
    INTEGER(I4B) ::  status,unit,blocksize,bitpix,naxis,naxes(1)
    INTEGER(I4B) ::  i,cnt,hdutype
    LOGICAL(LGT) ::  simple,extend
    CHARACTER(LEN=80) :: comment

    INTEGER(I4B), PARAMETER :: maxdim = 20 !number of columns in the extension
    INTEGER(I4B) :: nrows, npix, tfields, varidat
    INTEGER(I4B) :: frow,  felem, colnum
    CHARACTER(LEN=20) :: ttype(maxdim), tform(maxdim), tunit(maxdim), extname
    CHARACTER(LEN=8) ::  date
    CHARACTER(LEN=10) ::  fulldate
    CHARACTER(LEN=10) ::  card
    CHARACTER(LEN=2) :: stn
    INTEGER(I4B) :: itn
    integer(I4B) :: l, m
    !-----------------------------------------------------------------------

    !      open (12,file='almdunp.unf',status='new',form='unformatted')
    !      write (12) alms(0:nlmax,0:nlmax)
    !      close (12)

    npix=nalms

    status=0

    unit = 100

    !     create the new empty FITS file
    blocksize=1

    if (extno==1) then

       call ftinit(unit,filename,blocksize,status)

       !     -----------------------------------------------------
       !     initialize parameters about the FITS image
       simple=.true.
       bitpix=32     ! integer*4
       naxis=0       ! no image
       naxes(1)=0
       extend=.true. ! there is an extension

       !     ----------------------
       !     primary header
       !     ----------------------
       !     write the required header keywords
       call ftphpr(unit,simple,bitpix,naxis,naxes,0,1,extend,status)

       !     writes supplementary keywords : none

       !     write the current date
       call ftpdat(unit,status) ! format (dd/mm/yy)

       !     update the date (format ccyy-mm-dd)
       call date_and_time(date)
       fulldate = date(1:4)//'-'//date(5:6)//'-'//date(7:8)
       comment = 'FITS file creation date ccyy-mm-dd'
       call ftukys(unit,'DATE',fulldate,comment,status)

       !     ----------------------
       !     image : none
       !     ----------------------

       !     ----------------------
       !     extension
       !     ----------------------

    else

       call ftopen(unit,filename,1,blocksize,status)
       call ftmahd(unit,extno,hdutype,status)

    endif

    !     creates an extension
    call ftcrhd(unit, status)

    !     writes required keywords
    nrows    = npix  ! naxis1
    tfields  = ncl
    tform(1)='1J'
    tform(2:ncl) = '1E'
    ttype(1) = 'index=l^2+l+m+1'
    ttype(2) = 'alm (real)'
    ttype(3) = 'alm (complex)' 
    if (ncl>3) then
       ttype(4) = 'error (real)'
       ttype(5) = 'error (complex)'
    endif
    tunit(1:ncl) = ''      ! optional, will not appear
    extname  = ''      ! optional, will not appear
    varidat  = 0
    call ftphbn(unit, nrows, tfields, ttype, tform, tunit, &
         &     extname, varidat, status)

    !     write the header literally, putting TFORM1 at the desired place
    do i=1,nlheader
       card = header(i)
       if (card(1:5) == 'TTYPE') then ! if TTYPE1 is explicitely given
          stn = card(6:6)
          read(stn,'(i1)') itn
          ! discard at their original location:
          call ftdkey(unit,'TTYPE'//stn,status)  ! old TTYPEi and  ! remove
          status = 0
          call ftdkey(unit,'TFORM'//stn,status)  !     TFORMi
          status = 0
          call putrec(unit,header(i), status)           ! write new TTYPE1
          if (itn==1) then
             comment = 'data format of field: 4-byte INTEGER'
          else
             comment = 'data format of field: 4-byte REAL'
          endif
          status = 0
          call ftpkys(unit,'TFORM'//stn,tform(itn),comment,status) ! and write new TFORM1 right after
       elseif (header(i)/=' ') then
          call putrec(unit,header(i), status)
       endif
10     continue
       status = 0
    enddo

    !     write the extension one column by one column
    frow   = 1  ! starting position (row)
    felem  = 1  ! starting position (element)
    cnt=0
    ! recode the (l,m) -> lm mapping, EH, 2002-08-08
    do i=0,nalms-1
       l = nint(alms(i,1))
       m = nint(alms(i,2))
       ! lm(i)=nint(alms(i,1))**2+nint(alms(i,1))+nint(alms(i,2))+1
       lm(i) = (l+1)*l + m + 1
    enddo


    !      open (12,file='index.unf',status='new',form='unformatted')
    !      write (12) lm(0:cnt-1)
    !      close (12)

    !      open (12,file='almdump.unf',status='new',form='unformatted')
    !      write (12) alms_out(0:cnt-1,1:2)
    !      close (12)



    call ftpclj(unit, 1, frow, felem, npix, lm(0), status)
    do colnum = 2, ncl
       call ftpcle(unit, colnum, frow, felem, npix, alms(0,colnum+1), status)
    enddo

    !     ----------------------
    !     close and exit
    !     ----------------------

    !     close the file and free the unit number
    call ftclos(unit, status)

    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)

    return
  end subroutine write_alms

  !=======================================================================
  subroutine read_alms(filename,nalms,alms,ncl,header,nlheader,extno) 
    !=======================================================================
    !     Read a FITS file
    !
    !     slightly modified to deal with vector column 
    !     in binary table       EH/IAP/Jan-98
    !
    !     Used by synfast when reading a binary file with alms for cons.real.
    !                        FKH/Apr-99
    !
    !     called by fits2alm
    !=======================================================================
    CHARACTER(LEN=*),               INTENT(IN) :: filename
    INTEGER(I4B),                        INTENT(IN) :: nalms, ncl,nlheader,extno
    REAL(SP), DIMENSION(0:nalms-1,1:(ncl+1)), INTENT(OUT) :: alms
    CHARACTER(LEN=80), INTENT(OUT), DIMENSION(1:nlheader) :: header
    REAL(SP)                                        :: nullval
    LOGICAL(LGT)                                    ::  anynull
    INTEGER(I4B), DIMENSION(0:nalms-1) :: lm
    INTEGER(I4B) :: status,unit,readwrite,blocksize,naxes(2),nfound, naxis
    INTEGER(I4B) :: firstpix,npix
    REAL(SP) :: blank
    REAL(DP) ::  bscale,bzero
    CHARACTER(LEN=80) :: comment ! , record
    LOGICAL(LGT) :: extend
    INTEGER(I4B) :: nmove, hdutype ! , nkeys , nspace
    INTEGER(I4B) :: column, frow, imap
    INTEGER(I4B) :: datacode, repeat, width
    integer(I4B) :: i, l, m

    INTEGER(I4B), PARAMETER :: maxdim=20 !number of columns in the extension
    INTEGER(I4B) :: nrows, tfields, varidat
    CHARACTER(LEN=20) :: ttype(maxdim), tform(maxdim), tunit(maxdim), extname

    !-----------------------------------------------------------------------
    status=0
    header=''
    unit = 150
    naxes(1) = 1
    naxes(2) = 1
    nfound = -1
    anynull = .false.
    bscale = 1.0d0
    bzero = 0.0d0
    blank = -2.e25
    nullval = bscale*blank + bzero
    alms=0.
    readwrite=0
    call ftopen(unit,filename,readwrite,blocksize,status)
    if (status > 0) call printerror(status)
    !     -----------------------------------------

    !     determines the presence of image
    call ftgkyj(unit,'NAXIS', naxis, comment, status)
    if (status > 0) call printerror(status)

    !     determines the presence of an extension
    call ftgkyl(unit,'EXTEND', extend, comment, status)
    if (status > 0) status = 0 ! no extension : 
    !     to be compatible with first version of the code

    if (.not.extend) stop 'No extension!'
    nmove = +extno
    call ftmrhd(unit, nmove, hdutype, status)
    !cc         write(*,*) hdutype

    if (hdutype /= 2) then ! not a binary table
       stop 'this is not a binary table'
    endif

    !         call ftghsp(unit, nkeys, nspace, status)
    !          if (nkeys > nlheader) then
    !             print*,' the header in '//TRIM(filename)//' is too long'
    !             stop
    !          endif
    !          do i=1, nkeys
    !             call ftgrec(unit,i,record,status)
    !             header(i) = record
    !          enddo
    header = ""
    call get_clean_header( unit, header, filename, status)


    !        reads all the keywords
    call ftghbn(unit, maxdim, &
         &        nrows, tfields, ttype, tform, tunit, extname, varidat, &
         &        status)

    if (tfields<ncl) then
       print *,'found ',tfields,' columns in the file'
       print *,'expected ',ncl
       stop
    endif
    !        finds the bad data value
    call ftgkye(unit,'BAD_DATA',nullval,comment,status)
    if (status == 202) then ! bad_data not found
       nullval = s_bad_value ! default value
       status = 0
    endif
    !parse TFORM keyword to find out the length of the column vector
    call ftbnfm(tform(1), datacode, repeat, width, status)

    !reads the columns
    column = 1
    frow = 1
    firstpix = 1
    npix = nrows * repeat
    if (npix /= nalms) then
       print *,'found ',npix,' alms'
       print *,'expected ',nalms
       stop
    endif
    call ftgcvj(unit, column, frow, firstpix, npix, nullval, &
         &        lm(0), anynull, status)
    if (anynull) stop 'There are undefined values in the table!'
!     alms(0:nalms-1,1)=REAL(int(sqrt(DBLE(lm(0:nalms-1)-1))))
!     alms(0:nalms-1,2)=REAL(DBLE(lm(0:nalms-1)))-1.d0-alms(0:nalms-1,1)**2-alms(0:nalms-1,1)
    ! recoding of the mapping, EH, 2002-08-08
    ! lm = l*l + l + m + 1
    do i = 0, nalms - 1
       l = int(sqrt(   real(lm(i)-1, kind = DP)  ) )
       m = lm(i) - l*(l+1) - 1
       ! check that round-off did not screw up mapping
       if (abs(m) > l) then
          print*,'Inconsistent l^2+l+m+1 -> l,m mapping'
          print*,l, m, l*(l+1)+m+1, lm(i)
          stop
       endif
       alms(i,1) = real( l )
       alms(i,2) = real( m )
    enddo
    do imap = 2, ncl
       !parse TFORM keyword to find out the length of the column vector
       call ftbnfm(tform(imap), datacode, repeat, width, status)

       !reads the columns
       column = imap
       frow = 1
       firstpix = 1
       npix = nrows * repeat
       if (npix /= nalms) then
          print *,'found ',npix,' alms'
          print *,'expected ',nalms
          stop
       endif
       call ftgcve(unit, column, frow, firstpix, npix, nullval, &
            &        alms(0,imap+1), anynull, status)
       if (anynull) stop 'There are undefined values in the table!'
    enddo
    !     close the file
    call ftclos(unit, status)


    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)
    return
  end subroutine read_alms

  !=======================================================================
  subroutine printerror(status)
    !=======================================================================
    !     Print out the FITSIO error messages to the user
    !=======================================================================
    INTEGER(I4B), INTENT(IN) :: status
    CHARACTER ::  errtext*30,errmessage*80
    !-----------------------------------------------------------------------
    !     check if status is OK (no error); if so, simply return
    if (status .le. 0)return

    !     get the text string which describes the error
    call ftgerr(status,errtext)
    print *,'FITSIO Error Status =',status,': ',errtext

    !     read and print out all the error messages on the FITSIO stack
    call ftgmsg(errmessage)
    do while (errmessage /= ' ')
       print *,errmessage
       call ftgmsg(errmessage)
    end do

    return
  end subroutine printerror


  !=======================================================================
  subroutine read_par(filename,nside,lmax,tfields)
    !=======================================================================
    !        Read nside and lmax from a FITS-file          
    !    
    !
    !         Frode K. Hansen, April 1999
    !
    !=======================================================================
    CHARACTER(LEN=*), INTENT(IN)            :: filename

    INTEGER(I4B), INTENT(OUT)   :: nside, lmax, tfields

    INTEGER(I4B) :: status,unit,readwrite,blocksize,naxis
    CHARACTER(LEN=80) :: comment
    LOGICAL(LGT) ::  extend
    INTEGER(I4B)::  nmove, hdutype

    !-----------------------------------------------------------------------
    status=0
    unit = 150

    readwrite=0
    call ftopen(unit,filename,readwrite,blocksize,status)
    if (status > 0) call printerror(status)
    !     -----------------------------------------
    call ftgkyj(unit,'NAXIS', naxis, comment, status)

    !     determines the presence of an extension
    call ftgkyl(unit,'EXTEND', extend, comment, status)
    if (status > 0) stop 'No Extension in FITS file!'

    nmove = +1
    call ftmrhd(unit, nmove, hdutype, status)

    if (hdutype /= 2) then ! Not a binary-table
       stop 'This is not a FITS binary-table'
    endif

    call ftgkyj(unit,'NSIDE',nside,comment,status)
    if (status == 202) stop 'NSIDE-keyword not found!'
    call ftgkyj(unit,'MAX-LPOL',lmax,comment,status)
    if (status == 202) stop 'MAX_LPOL-keyword not found!'
    call ftgkyj(unit,'TFIELDS',tfields,comment,status)
    if (status == 202) stop 'TFIELDS-keyword not found!'
    call ftclos(unit, status)

  end subroutine read_par

  !=======================================================================
  function getsize_fits(filename, nmaps, ordering, obs_npix, &
       &    nside, mlpol, type, polarisation)
    !=======================================================================
    !  result = getsize_fits(filename, nmaps, ordering, nside, mlpol, type)
    !     Get the number of pixels stored in a map FITS file.
    !     Each pixel is counted only once 
    !     (even if several information is stored on each of them, see nmaps).
    !     Depending on the data storage format, this may be :
    !       - equal or smaller to the number Npix of Healpix pixels available 
    !          over the sky for the given resolution (Npix = 12*nside*nside)
    !       - equal or larger to the number of non blank pixels (obs_npix)
    !     
    !     filename = (IN) name of the FITS file containing the map
    !     nmaps = (OPTIONAL, OUT) number of maps in the file
    !     ordering = (OPTIONAL, OUT) Healpix ordering scheme used
    !                  0 = unknown
    !                  1 = RING
    !                  2 = NESTED
    !     obs_npix =  (OPTIONAL, OUT) number of non blanck pixels.    
    !                It is set to -1 if it can not be determined from header
    !                information alone
    !     nside = (OPTIONAL, OUT) Healpix parameter Nside
    !     mlpol = (OPTIONAL, OUT) maximum multipole used to generate the map (for simulated map)
    !     type = (OPTIONAL, OUT) Healpix/FITS file type
    !                  <0 : file not found, or not valid
    !                  0  : image only fits file, deprecated Healpix format
    !                        (result = 12 * nside * nside)
    !                  1  : ascii table, generally used for C(l) storage
    !                  2  : binary table : with implicit pixel indexing (full sky)
    !                        (result = 12 * nside * nside)
    !                  3  : binary table : with explicit pixel indexing (generally cut sky)
    !                        (result <= 12 * nside * nside)
    !                999  : unable to determine the type
    !     polarisation = (OPTIONAL, OUT) presence of polarisation data in the file
    !                  <0 : can not find out
    !                   0 : no polarisation
    !                   1 : contains polarisation (Q,U or G,C)
    !     Benjamin D. Wandelt January 1998
    !     includes Eric Hivon's modification of FITS columns
    !     and draws heavily on the read_bintab routine.
    !
    !     addition of optional Ordering by E.H. (July 98)
    !     addition of optional Nside and Mlpol by ?? (??)
    !     addition of optional type by E.H. (Sept 00)
    !     improved for ASCII table E.H (Feb 03)
    !=======================================================================
    use pix_tools, only: nside2npix
    CHARACTER(LEN=*), INTENT(IN)            :: filename
    INTEGER(I4B),          OPTIONAL :: nmaps
    INTEGER(I4B),          OPTIONAL :: ordering
    INTEGER(I4B),          OPTIONAL :: nside
    INTEGER(I4B),          OPTIONAL :: mlpol
    INTEGER(I4B),          OPTIONAL :: obs_npix
    INTEGER(I4B),          OPTIONAL :: type
    INTEGER(I4B),          OPTIONAL :: polarisation

    INTEGER(I4B)           :: nmaps_in, ordering_in, nside_in
    INTEGER(I4B)           :: mlpol_in, obs_npix_in, ftype_in
    INTEGER(I4B)           :: grain
    CHARACTER(len=20)      :: indxschm
    CHARACTER(LEN=20)      :: order_val, object_val, ttype_val
!     INTEGER(I4B)           :: getsize_fits
    INTEGER(I8B)           :: getsize_fits
    LOGICAL(LGT)           ::  polar_in
    character(len=3),  dimension(1:10,1:2)  :: defpol
    logical(kind=LGT), dimension(1:2)       :: pf
    integer(kind=I4B)                       :: ndp, j, k

    INTEGER(I4B) :: status,unit,readwrite,blocksize,naxes(2),nfound, naxis
    INTEGER(I4B) :: i
!     INTEGER(I4B) :: nsize
    INTEGER(I8B) :: nsize
    CHARACTER(LEN=80) :: comment
    LOGICAL(LGT) ::  extend
    INTEGER(I4B)::  nmove, hdutype
    INTEGER(I4B) :: datacode, repeat1, repeat2, width

    INTEGER(I4B), PARAMETER :: maxdim=20 !number of columns in the extension
    INTEGER(I4B) :: nrows, tfields, varidat, rowlen
    CHARACTER(LEN=20), dimension(1:maxdim) :: ttype, tform, tunit
    INTEGER(I4B),      dimension(1:maxdim) :: tbcol
    CHARACTER(LEN=20) :: extname
    !-----------------------------------------------------------------------
    status=0
    order_val = ''
    nmaps_in = 0
    ordering_in = 0
    nside_in = 0
    ftype_in = 999 !   
    obs_npix_in = -1
    unit = 150
    naxes(1) = 1
    naxes(2) = 1
    nfound = -1

    readwrite=0
    call ftopen(unit,filename,readwrite,blocksize,status)
    if (status > 0) then
       ftype_in = -1
       call printerror(status)
       return
    endif
    !     -----------------------------------------

    !     determines the presence of image
    call ftgkyj(unit,'NAXIS', naxis, comment, status)

    !     determines the presence of an extension
    call ftgkyl(unit,'EXTEND', extend, comment, status)
    if (status > 0) then
       ftype_in = 0
       status = 0 ! no extension : 
       !     to be compatible with first version of the code
    endif

    if (naxis > 0) then 
       !---------------------------------
       ! there is an image
       !---------------------------------
       !        determine the size of the image (look naxis1 and naxis2)
       call ftgknj(unit,'NAXIS',1,2,naxes,nfound,status)

       !        check that it found only NAXIS1
       if (nfound == 2 .and. naxes(2) > 1) then
          print *,'multi-dimensional image'
          print *,'expected 1-D data.'
          call ftclos(unit, status)
          stop
       end if

       if (nfound < 1) then
          call printerror(status)
          print *,'can not find NAXIS1.'
          call ftclos(unit, status)
          stop
       endif

       nsize=naxes(1)

       call ftgkys(unit,'ORDERING',order_val,comment,status)
       if (status == 202) then ! Ordering not found
          ordering_in = 0
          order_val = ''
          status = 0
       endif

       call ftgkyj(unit,'NSIDE',nside_in,comment,status)
       if (status == 202) then ! Nside not found
          nside_in = 0
          status = 0
       endif

       if (present(mlpol)) then
          call ftgkyj(unit,'MAX-LPOL',mlpol_in,comment,status)
          if (status == 202) then ! max-lpol not found
             mlpol_in = 0
             status = 0
          endif
       endif

       if (present(polarisation)) then
          polarisation = 0
       endif
    else if (extend) then 
       !-----------------------------------------------------------------
       ! there is an extension
       !-----------------------------------------------------------------
       nmove = +1
       call ftmrhd(unit, nmove, hdutype, status)
       !c         write(*,*) hdutype

       !        reads all the keywords
       if (hdutype == 2) then ! binary table
          call ftghbn(unit, maxdim, &
            &         nrows, tfields, ttype, tform, tunit, extname, varidat, &
            &         status)
       else ! ASCII table (hdutype = 1)
          ftype_in = 1
          call ftghtb(unit, maxdim, &
            &         rowlen, nrows, tfields, ttype, tbcol, tform, tunit, &
            &         extname, status)
       endif

       !        parse TFORM keyword to find out the length of the column vector
       repeat1 = 1
       repeat2 = 1
       call ftbnfm(tform(1), datacode, repeat1, width, status)
       if (tfields > 1) call ftbnfm(tform(2), datacode, repeat2, width, status)

       nsize = int(nrows, kind=i8b) * max(repeat1,repeat2) ! corrected Oct-03

       nmaps_in = tfields

       call ftgkys(unit,'ORDERING',order_val,comment,status)
       if (status == 202) then ! Ordering not found
          ordering_in = 0
          order_val = ''
          status = 0
       endif

       call ftgkyj(unit,'NSIDE',nside_in,comment,status)
       if (status == 202) then ! Nside not found
          nside_in = 0
          status = 0
       endif

       call ftgkyj(unit,'OBS_NPIX',obs_npix_in,comment,status)
       if (status == 202) then ! obs_npix not found
          obs_npix_in = -1
          status = 0
       endif

       if (present(mlpol)) then
          call ftgkyj(unit,'MAX-LPOL',mlpol_in,comment,status)
          if (status == 202) then ! max-lpol not found
             mlpol_in = 0
             status = 0
          endif
       endif

       ! determines pixel indexing (for binary tables)
       if (present(type) .and. ftype_in == 999) then
          ! most stringent test
          call ftgkys(unit,'INDXSCHM',indxschm,comment,status)
          if (status == 0) then ! found
             ftype_in = 3
             if (trim(indxschm) == 'IMPLICIT') ftype_in = 2
             goto 1000
          else
             status = 0
          endif
          ! 2nd most stringent test
          call ftgkyj(unit,'GRAIN',grain,comment,status)
          if (status == 0) then ! found
             ftype_in = 3
             if (grain == 0) ftype_in = 2
             goto 1000
          else
             status = 0
          endif
          ! 3rd most stringent test
          if (trim(ttype(1)) /= '') then
             if (trim(ttype(1)) == 'PIXEL') then
                ftype_in = 3
             else
                ftype_in = 2
             endif
             goto 1000
          endif
          ! lousy test
          call ftgkys(unit,'OBJECT',object_val,comment,status)
          if (status == 0) then
             if (object_val == 'PARTIAL') ftype_in = 3
             if (object_val == 'FULLSKY') ftype_in = 2
             if (ftype_in /= 999) goto 1000
          else
             status = 0
          endif
          ! very lousy test
          if (nside_in > 0) then
             ftype_in = 3
             if (nside2npix(nside_in) == nsize) ftype_in = 2
             goto 1000
          endif
       endif
1000   continue

       ! find out if polarisation data is present
       if (present(polarisation)) then
          if (tfields < 3) then
             polarisation = 0 ! no polarisation
             goto 2000
          endif
          call ftgkyl(unit,'POLAR',polar_in,comment,status)
          if (status == 0) then ! polar found
             polarisation = 0
             if (polar_in) polarisation = 1
             goto 2000
          else ! polar not found
             status = 0
             polarisation = -1
             if (hdutype <= 0) goto 2000
             if (hdutype == 1) then ! ascii table -> power spectra
                ndp = 3
                defpol(1:ndp,1) = (/ "GRA","E-M","POW" /)
                defpol(1:ndp,2) = (/ "CUR","B-M","POW" /)    
             endif
             if (hdutype == 2) then ! binary table -> maps
                ndp = 4
                defpol(1:ndp,1) = (/ "Q-P","Q_P","Q P","Q  " /)
                defpol(1:ndp,2) = (/ "U-P","U_P","U P","U  " /)
             endif
             pf(:) = .false.
             do i = 2, tfields ! do not consider first field (generally temperature)
                ttype_val = adjustl(ttype(i))
                call ftupch(ttype_val) ! upper case
                do k=1,2
                   do j=1, ndp
                      if (index( ttype_val, trim(defpol(j,k)) ) == 1) then
                         pf(k) = .true.
                         goto 1500 ! move to next field
                      endif
                   enddo
                enddo ! loop on k
1500            continue
             enddo ! loop on i
             if (pf(1) .and. pf(2)) polarisation = 1
             if (.not. (pf(1) .or. pf(2))) polarisation = 0
          endif ! polar not found
       endif ! present(polarisation)
2000   continue

    else ! no image no extension, you are dead, man
       ftype_in = -1
       stop ' No image, no extension'
    endif

    !     close the file
    call ftclos(unit, status)

    !     check for any error, and if so print out error messages
    if (status > 0) call printerror(status)

    getsize_fits=nsize

    call ftupch(order_val) ! convert order_val to upper case
    if (order_val(1:4) == 'RING') ordering_in = 1
    if (order_val(1:4) == 'NEST') ordering_in = 2

    if (present(nmaps)) nmaps = nmaps_in
    if (present(mlpol)) mlpol = mlpol_in
    if (present(obs_npix)) obs_npix = obs_npix_in
    if (present(ordering)) ordering = ordering_in
    if (present(nside)) nside = nside_in
    if (present(type)) type = ftype_in

    return
  end function getsize_fits
  !=======================================================================
  function number_of_alms(filename,extnum)
    !=======================================================================
    !    Read the number of alms from a FITS-file containing alms          
    !    for constrained realisations in synfast.
    !
    !         Frode K. Hansen, April 1999
    !
    !=======================================================================
    CHARACTER(LEN=*), INTENT(IN)            :: filename

    INTEGER(I4B), DIMENSION(2)           :: naxes
    INTEGER(I4B)           :: number_of_alms

    INTEGER(I4B) :: status,unit,readwrite,blocksize,naxis,nfound
    CHARACTER(LEN=80) :: comment
    LOGICAL(LGT) ::  extend
    INTEGER(I4B)::  nmove, hdutype, hdunum
    integer(I4B), optional, intent(INOUT) :: extnum

    !-----------------------------------------------------------------------
    status=0
    unit = 150

    readwrite=0
    call ftopen(unit,filename,readwrite,blocksize,status)
    if (status > 0) call printerror(status)
    !     -----------------------------------------
    call ftgkyj(unit,'NAXIS', naxis, comment, status)

    !     determines the presence of an extension
    call ftgkyl(unit,'EXTEND', extend, comment, status)
    if (status > 0) stop 'No Extension in FITS file!'

    nmove = +1
    call ftmrhd(unit, nmove, hdutype, status)

    if (hdutype /= 2) then ! Not a binary table
       stop 'This is not a FITS binary-table'
    endif

    call ftgknj(unit,'NAXIS',1,2,naxes,nfound,status)
    if (nfound < 2) stop 'NAXIS2-keyword not found!'

    number_of_alms=naxes(2)

    if (present(extnum)) then
       call ftthdu(unit,hdunum,status)
       extnum = hdunum - 1 ! first HDU : primary array
    endif

    call ftclos(unit, status)

  end function number_of_alms

  !======================================================================
  subroutine putrec(unit, card, status)
    !======================================================================
    ! append, delete or update a card from the current header
    ! (deletion of keyword KWD is described by : - KWD)
    ! EH, version 1.0, Dec 2001
    !======================================================================
    integer(kind=I4B), intent(IN)  :: unit
    character(len=*),  intent(IN)  :: card
    integer(kind=I4B), intent(OUT) :: status

    character(len=80) :: cardfits,record
    character(len=8)  :: kwd
    integer(kind=I4B) :: hdtype
    character(len=80) :: nullarr(0)
    !======================================================================

    status = 0
    call ftgthd(card, cardfits, hdtype, status)
    kwd = cardfits(1:8)
    status = 0

    select case (hdtype)
    case (-1) ! delete keyword (starting from the top)
       call ftgrec(unit,0,record,status)
       !             call ftdkey(unit, kwd, status)
       ! patch around cfitsio bug 
       ! (ftdkey does not recognize wild cards while ftgnxk does)
       do
          call ftgnxk(unit,kwd,1,nullarr,0,record,status)
          if (status /= 0) exit
          call ftdkey(unit, record(1:8), status)
       enddo
    case (0) ! append or update
       ! delete keyword in its current location (if any)
       call ftdkey(unit, kwd, status)
       status = 0
       ! append
       call ftprec(unit, cardfits, status)
    case (1) ! append (for HISTORY and COMMENT)
       call ftprec(unit, cardfits, status)
    case default
       write(unit=*,fmt=*)" Unexpected card format in fits header :"
       write(unit=*,fmt="(a80)") card
       write(unit=*,fmt=*)" card not written."
    end select
    status = 0

    return
  end subroutine putrec

  !====================================================================
  subroutine get_clean_header(unit, header, filename, error, xalso, xonly)
    !====================================================================
    ! get_clean_header(unit, header, error [, xalso, xonly])
    ! gets the FITS header from unit, after excluding some default keywords 
    !  defined in def_excl
    ! if header in non void on input, its content will be concatenated with that
    !  of the FITS header
    ! if xalso is defined as a list of keywords, they are also excluded from the header
    ! if xonly is defined as a list of keywords, only those keywords are excluded from
    ! the header.
    ! xonly and xalso are exclusive
    !====================================================================
    INTEGER(I4B),                    intent(IN)           :: unit
    CHARACTER(LEN=*), DIMENSION(1:), INTENT(IN OUT)       :: header
    CHARACTER(LEN=*),                INTENT(IN)           :: filename
    INTEGER(I4B),                    intent(OUT)          :: error
    character(len=8), dimension(1:), intent(IN), optional :: xalso
    character(len=8), dimension(1:), intent(IN), optional :: xonly

    INTEGER(I4B) :: nlheader, status, i, n_excl
    CHARACTER(LEN=80) :: record
    CHARACTER(len=8), dimension(:), allocatable :: to_excl

    CHARACTER(len=8), dimension(1:21) :: def_excl
    !====================================================================

    ! keywords to be excluded by default from output header
    ! Note that TTYPE# and TUNIT# keywords are not excluded
    ! from the output header as they might be useful to the 
    ! calling routines
    def_excl=(/&
         & "SIMPLE  ","BITPIX  ","NAXIS   ",&
         & "NAXIS#  ","PCOUNT  ","GCOUNT  ",&
         & "EXTEND  ","ORIGIN  ","DATE*   ",&
         & "TFIELDS ","TFORM#  ",           & 
         & "TBCOL#  ","EXTNAME ","CTYPE#  ",&
         & "CRVAL#  ","CRPIX#  ","CDELT#  ",&
         & "XTENSION","INSTRUME","TELESCOP",&
         & "PDMTYPE "/)

    error = 0

    if (present(xonly)) then 
       n_excl = size(xonly)
       allocate(to_excl(1:n_excl))
       to_excl = xonly

    else if (present(xalso)) then
       n_excl = size(xalso) + size(def_excl)
       allocate(to_excl(1:n_excl))
       to_excl(1:size(def_excl)) = def_excl
       to_excl(size(def_excl)+1:n_excl) = xalso

    else
       n_excl = size(def_excl)
       allocate(to_excl(1:n_excl))
       to_excl = def_excl
    endif

    nlheader=size(header)
    ! go to end of fortran header
    do i = 1, nlheader
       if (trim(header(i)) == "") exit
    enddo
    ! go to top of fits file header
    status=0
    call ftgrec(unit,0,record,status)
    ! read in all header lines except those excluded
    do
       call ftgnxk(unit,'*',1,to_excl,n_excl,record,status)
       if (status > 0) exit ! end of header
       if (i > nlheader) then
          write(unit=*,fmt="(a,i5,a)") &
               & " WARNING : The header in "//  &
               &    trim(filename)//" has more than ", &
               &  nlheader," lines."
          print*," It will be truncated."
          error = 1
          exit
       endif
       header(i)=record
       i=i+1
    enddo
    status=0

    return
  end subroutine get_clean_header
  !====================================================================

  SUBROUTINE input_tod_s(filename, tod, npixtot, ntods, &
       &                 header, firstpix, fmissval, extno)
  ! *************************************************************************

  ! *************************************************************************
  ! v1.0 : Olivier Dore et Eric Hivon, June-July 2002
  ! v1.1     2002-07-08 : bugs correction by E.H. 
  ! v1.2     2002-08-22 : pixel number starts at 0 for these routine,
  !                         but starts at 1 for cftisio routines ...
  ! *************************************************************************
    ! ===================================================================================
    !     reads fits file
    !     filename = fits file (input)
    !     tod      = data rad from the file (ouput) = real*4 array of size (npixtot,ntods)
    !     npixtot  = length of the tod (input)
    !     ntods    = number of tods
    !     header   = OPTIONAL, FITS header
    !     fmissval = OPTIONAL argument (input) with the value to be given to missing
    !             pixels, its default value is 0
    !     firstpix = OPTIONAL first pixel to be read (starting at 0)
    !     extno    = OPTIONAL, extension number (starting at 0)
    !                                               O.D. & E.H. 06/02 @ IAP
    !      2002-07-08 : bugs correction by E.H. 
    !       (consistent use of fmiss_effct)
    ! =======================================================================
    
    IMPLICIT NONE

    INTEGER(I8B),     INTENT(IN)           :: npixtot
    INTEGER(I8B),     INTENT(IN), OPTIONAL :: firstpix
    INTEGER(I4B),     INTENT(IN)           :: ntods
    REAL(SP),         INTENT(OUT)          ,dimension(0:,1:) :: tod
    CHARACTER(LEN=*), INTENT(IN)           :: filename
    REAL(SP),         INTENT(IN), OPTIONAL :: fmissval
    character(len=*), intent(out),OPTIONAL, dimension(1:)  :: header
    INTEGER(I4B),     INTENT(IN), OPTIONAL :: extno
    
    INTEGER(I4B) :: i,itod
    REAL(SP)     :: fmissing, fmiss_effct
    INTEGER(I4B) :: imissing
    integer(i8b) :: fp = 0_i8b

    LOGICAL(LGT) :: anynull
    
    !-----------------------------------------------------------------------
    fmiss_effct = 0.
    IF (PRESENT(fmissval)) fmiss_effct = fmissval

    if (present(firstpix)) fp = firstpix

    CALL read_bintod_s(filename, tod, npixtot, ntods, fp, fmissing, anynull, &
         &             header, extno)

    DO itod = 1, ntods
       anynull = .TRUE.
       IF (anynull) THEN
          imissing = 0
          DO i=0,npixtot-1
             IF ( ABS(tod(i,itod)/fmissing -1.) .LT. 1.e-5 ) THEN
                tod(i,itod) = fmiss_effct
                imissing = imissing + 1
             ENDIF
          ENDDO
          IF (imissing .GT. 0) THEN
             WRITE(*,'(a,1pe11.4)') 'blank value : ' ,fmissing
             WRITE(*,'(i7,a,f7.3,a,1pe11.4)') &
                  &           imissing,' missing pixels (', &
                  &           (100.*imissing)/npixtot,' %),'// &
                  &           ' have been set to : ',fmiss_effct
          ENDIF
       ENDIF
    ENDDO
    RETURN

  END SUBROUTINE input_tod_s
  !=======================================================================

  !**************************************************************************************
  SUBROUTINE input_tod_d(filename, tod, npixtot, ntods, &
       &                 header, firstpix, fmissval, extno)
  !**************************************************************************************

    !=======================================================================
    !     reads fits file
    !     filename = fits file (input)
    !     tod      = data rad from the file (ouput) = real*8 array of size (npixtot,ntods)
    !     npixtot  = length of the tod (input)
    !     ntods     = number of tods
    !     header   = OPTIONAL, FITS header
    !     fmissval = OPTIONAL argument (input) with the value to be given to missing
    !             pixels, its default value is 0
    !     firstpix = OPTIONAL first pixel to be read (starting at 0)
    !     extno    = OPTIONAL, extension number (starting at 0)
    !                                               O.D. & E.H. 06/02 @ IAP
    !      2002-07-08 : bugs correction by E.H. 
    !       (consistent use of fmiss_effct)
    !=======================================================================
    
    IMPLICIT NONE

    INTEGER(I8B),     INTENT(IN)           :: npixtot
    INTEGER(I8B),     INTENT(IN), OPTIONAL :: firstpix
    INTEGER(I4B),     INTENT(IN)           :: ntods
    REAL(DP),         INTENT(OUT)          ,dimension(0:,1:) :: tod
    CHARACTER(LEN=*), INTENT(IN)           :: filename
    REAL(DP),         INTENT(IN), OPTIONAL :: fmissval
    character(len=*), intent(out),OPTIONAL, dimension(1:)  :: header
    INTEGER(I4B),     INTENT(IN), OPTIONAL :: extno
    
    INTEGER(I4B) :: i,itod
    REAL(DP)     :: fmissing, fmiss_effct
    INTEGER(I4B) :: imissing
    integer(i8b) :: fp = 0_i8b

    LOGICAL(LGT) :: anynull
    
    !-----------------------------------------------------------------------
    
    fmiss_effct = 0.
    IF (PRESENT(fmissval)) fmiss_effct = fmissval

    if (present(firstpix)) fp = firstpix
    
    CALL read_bintod_d(filename, tod, npixtot, ntods, fp, fmissing, anynull, &
         &             header, extno)

    DO itod = 1, ntods
       anynull = .TRUE.
       IF (anynull) THEN
          imissing = 0
          DO i=0,npixtot-1
             IF ( ABS(tod(i,itod)/fmissing -1.) .LT. 1.e-5 ) THEN
                tod(i,itod) = fmiss_effct
                imissing = imissing + 1
             ENDIF
          ENDDO
          IF (imissing .GT. 0) THEN
             WRITE(*,'(a,1pe11.4)') 'blank value : ' ,fmissing
             WRITE(*,'(i7,a,f7.3,a,1pe11.4)') &
                  &           imissing,' missing pixels (', &
                  &           (100.*imissing)/npixtot,' %),'// &
                  &           ' have been set to : ',fmiss_effct
          ENDIF
       ENDIF
    ENDDO
    RETURN

  END SUBROUTINE input_tod_d
  !=======================================================================
  
  !**************************************************************************
  SUBROUTINE read_bintod_s(filename, tod, npixtot, ntods, firstpix, nullval, anynull, &
                          header, extno)
  !**************************************************************************
    
    !=======================================================================
    !     Read a FITS file
    !
    !     slightly modified to deal with vector column (ie TFORMi = '1024E')
    !     in binary table       EH/IAP/Jan-98
    !
    !     This routine is used for reading TODS by anafast.
    !     Modified to start at a given pix numb OD & RT 02/02
    !     Modified to handle huge array (npix_tot > 2^32) OD & EH 07/02
    !      2002-07-08 : bugs correction by E.H. 
    !=======================================================================
    
    IMPLICIT NONE
    
    CHARACTER(LEN=*),               INTENT(IN)  :: filename
    INTEGER(I8B)   ,                INTENT(IN)  :: npixtot,firstpix
    INTEGER(I4B),                   INTENT(IN)  :: ntods
    REAL(SP), DIMENSION(0:,1:),     INTENT(OUT) :: tod
    REAL(SP),                       INTENT(OUT) :: nullval
    LOGICAL(LGT),                   INTENT(OUT) :: anynull
    character(len=*), dimension(1:),intent(out), optional :: header
    INTEGER(I4B),                   INTENT(IN),  OPTIONAL :: extno
    
    INTEGER(I4B) :: status,unit,readwrite,blocksize,naxes(2),nfound, naxis
    INTEGER(I4B) :: npix_32 !,firstpix_32
    REAL(SP)     :: blank
    REAL(DP)     ::  bscale,bzero
    CHARACTER(LEN=80) :: comment
    LOGICAL(LGT) :: extend
    INTEGER(I4B) :: nmove, hdutype
    INTEGER(I4B) :: column, frow, itod
    INTEGER(I4B) :: datacode, repeat, width
    
    INTEGER(I4B), PARAMETER :: maxdim=20 !number of columns in the extension
    INTEGER(I4B) :: nrows, tfields, varidat,felem
    CHARACTER(LEN=20) :: ttype(maxdim), tform(maxdim),tunit(maxdim), extname
 
    INTEGER(I8B) :: q,iq,npix_tmp,firstpix_tmp
    
    !-----------------------------------------------------------------------
    status=0
    
    unit = 150
    naxes(1) = 1
    naxes(2) = 1
    nfound = -1
    anynull = .FALSE.
    bscale = 1.0d0
    bzero = 0.0d0
    blank = -2.e25
    nullval = bscale*blank + bzero
    
    readwrite=0
    CALL ftopen(unit,filename,readwrite,blocksize,status)
    IF (status .GT. 0) CALL printerror(status)
    !     -----------------------------------------
    
    !     determines the presence of image
    CALL ftgkyj(unit,'NAXIS', naxis, comment, status)
    IF (status .GT. 0) CALL printerror(status)
    
    !     determines the presence of an extension
    CALL ftgkyl(unit,'EXTEND', extend, comment, status)
    IF (status .GT. 0) status = 0 ! no extension : 
    !     to be compatible with first version of the code
    
    IF (naxis .GT. 0) THEN ! there is an image
       print*,'WARNING : Image is ignored in '//trim(filename)
    ENDIF
    IF (extend) THEN ! there is an extension

       nmove = +1
       if (present(extno)) nmove = +1 + extno
       CALL ftmrhd(unit, nmove, hdutype, status)
       !cc         write(*,*) hdutype

       IF (hdutype .NE. 2) THEN ! not a binary table
          STOP 'this is not a binary table'
       ENDIF

       ! reads all the keywords
       CALL ftghbn(unit, maxdim, &
            &        nrows, tfields, ttype, tform, tunit, extname, varidat, &
            &        status)

       IF (tfields .LT. ntods) THEN
          PRINT *,'found ',tfields,' tods in the file'
          PRINT *,'expected ',ntods
          STOP
       ENDIF

       if (present(header)) then
          header = ""
          status = 0
          call get_clean_header(unit, header, filename, status)
       endif

       ! finds the bad data value
       CALL ftgkye(unit,'BAD_DATA',nullval,comment,status)
       IF (status .EQ. 202) THEN ! bad_data not found
          nullval = -1.6375e30 ! default value
          status = 0
       ENDIF

       IF (npixtot .LT. nchunk_max) THEN

          DO itod = 1, ntods

             !parse TFORM keyword to find out the length of the column vector (repeat)
             CALL ftbnfm(tform(itod), datacode, repeat, width, status)
             frow = (firstpix)/repeat+1          ! 1 based 
             felem = firstpix-(frow-1)*repeat+1  ! 1 based 
             npix_32 = npixtot 

             !reads the columns
             column = itod
             CALL ftgcve(unit,column,frow,felem,npix_32,nullval, &
                  &        tod(0,itod),anynull,status)
          END DO

       ELSE

          q = (npixtot-1)/nchunk_max
          DO iq = 0,q
             IF (iq .LT. q) THEN
                npix_tmp = nchunk_max
             ELSE
                npix_tmp = npixtot - iq*nchunk_max
             ENDIF
             firstpix_tmp = firstpix + iq*nchunk_max
             npix_32 = npix_tmp

             DO itod = 1, ntods
                ! parse TFORM keyword to find out the length of the column vector
                CALL ftbnfm(tform(itod), datacode, repeat, width, status)
                frow = (firstpix_tmp)/repeat+1          ! 1 based 
                felem = firstpix_tmp-(frow-1)*repeat+1  ! 1 based 

                CALL ftgcve(unit,itod,frow,felem,npix_32,nullval, &
                     &      tod(firstpix_tmp-firstpix,itod),anynull,status)
             END DO

          ENDDO

       ENDIF

    ELSE ! no image no extension, you are dead, man
       STOP ' No image, no extension'
    ENDIF

    ! close the file
    CALL ftclos(unit, status)

    ! check for any error, and if so print out error messages
    IF (status .GT. 0) CALL printerror(status)

    RETURN

  END SUBROUTINE read_bintod_s
  !=======================================================================
  
  ! ********************************************************************************
  SUBROUTINE read_bintod_d(filename, tod, npixtot, ntods, firstpix, nullval, anynull, &
                          header, extno)
    ! ********************************************************************************

    !=======================================================================
    !     Read a FITS file
    !
    !     slightly modified to deal with vector column 
    !     in binary table       EH/IAP/Jan-98
    !
    !     Reads a double-precision array with precomputed plms, used by syn/anafast
    !                FKH/Apr-99
    !     Modified to start at a given pix numb OD & RT 02/02
    !     Modified to handle huge array (npix_tot > 2^32) OD & EH 07/02
    !      2002-07-08 : bugs correction by E.H. 
    !=======================================================================

    USE healpix_types

    IMPLICIT NONE

    CHARACTER(LEN=*),               INTENT(IN)  :: filename
    INTEGER(I8B)    ,               INTENT(IN)  :: npixtot,firstpix
    INTEGER(I4B)    ,               INTENT(IN)  :: ntods
    REAL(DP), DIMENSION(0:,1:),     INTENT(OUT) :: tod
    REAL(DP)        ,               INTENT(OUT) :: nullval
    LOGICAL(LGT)    ,               INTENT(OUT) :: anynull
    character(len=*), dimension(1:),intent(out), optional :: header
    INTEGER(I4B),                   INTENT(IN),  OPTIONAL :: extno

    INTEGER(I4B) :: status,unit,readwrite,blocksize,naxes(2),nfound, naxis
    REAL(SP)     :: blank
    REAL(DP)     ::  bscale,bzero
    CHARACTER(LEN=80) :: comment
    LOGICAL(LGT) :: extend
    INTEGER(I4B) :: nmove, hdutype
    INTEGER(I4B) :: column, frow, itod
    INTEGER(I4B) :: datacode, repeat, width

    INTEGER(I4B), PARAMETER :: maxdim=20 !number of columns in the extension
    INTEGER(I4B) :: nrows, tfields, varidat,felem,npix_32
    CHARACTER(LEN=20) :: ttype(maxdim), tform(maxdim), tunit(maxdim), extname

    INTEGER(I8B) :: q,iq,npix_tmp,firstpix_tmp

    !-----------------------------------------------------------------------
    status=0

    unit = 150
    naxes(1) = 1
    naxes(2) = 1
    nfound = -1
    anynull = .FALSE.
    bscale = 1.0d0
    bzero = 0.0d0
    blank = -2.e25
    nullval = bscale*blank + bzero

    readwrite=0
    CALL ftopen(unit,filename,readwrite,blocksize,status)
    IF (status .GT. 0) CALL printerror(status)
    !     -----------------------------------------

    !     determines the presence of image
    CALL ftgkyj(unit,'NAXIS', naxis, comment, status)
    IF (status .GT. 0) CALL printerror(status)

    !     determines the presence of an extension
    CALL ftgkyl(unit,'EXTEND', extend, comment, status)
    IF (status .GT. 0) status = 0 ! no extension : 
    !     to be compatible with first version of the code

    IF (naxis .GT. 0) THEN ! there is an image
       print*,'WARNING : Image is ignored in '//trim(filename)
    endif
    IF (extend) THEN ! there is an extension
       nmove = +1
       if (present(extno)) nmove = +1 + extno
       CALL ftmrhd(unit, nmove, hdutype, status)
       !cc         write(*,*) hdutype

       IF (hdutype .NE. 2) THEN ! not a binary table
          STOP 'this is not a binary table'
       ENDIF

       ! reads all the keywords
       CALL ftghbn(unit, maxdim, &
            &        nrows, tfields, ttype, tform, tunit, extname, varidat, &
            &        status)

       IF (tfields .LT. ntods) THEN
          PRINT *,'found ',tfields,' tods in the file'
          PRINT *,'expected ',ntods
          STOP
       ENDIF

       if (present(header)) then
          header = ""
          status = 0
          call get_clean_header(unit, header, filename, status)
       endif

       ! finds the bad data value
       CALL ftgkyd(unit,'BAD_DATA',nullval,comment,status)
       IF (status .EQ. 202) THEN ! bad_data not found
          nullval = -1.6375e30 ! default value
          status = 0
       ENDIF

       ! Read repeat value
       CALL ftbnfm(tform(1),datacode,repeat,width,status)

       IF (npixtot .LT. nchunk_max) THEN

          DO itod = 1, ntods

             !parse TFORM keyword to find out the length of the column vector
             CALL ftbnfm(tform(itod), datacode, repeat, width, status)
             frow = (firstpix)/repeat+1
             felem = firstpix-(frow-1)*repeat+1
             npix_32 = npixtot 

             !reads the columns
             column = itod
             CALL ftgcvd(unit,column,frow,felem,npix_32,nullval, &
                  &        tod(0,itod),anynull,status)
          END DO

       ELSE

          q = (npixtot-1)/nchunk_max
          DO iq = 0,q
             IF (iq .LT. q) THEN
                npix_tmp = nchunk_max
             ELSE
                npix_tmp = npixtot - iq*nchunk_max
             ENDIF
             firstpix_tmp = firstpix + iq*nchunk_max
             npix_32 = npix_tmp

             DO itod = 1, ntods
                ! parse TFORM keyword to find out the length of the column vector
                CALL ftbnfm(tform(itod), datacode, repeat, width, status)
                frow = (firstpix_tmp)/repeat+1
                felem = firstpix_tmp-(frow-1)*repeat+1 

                CALL ftgcvd(unit,itod,frow,felem,npix_32,nullval, &
                     &        tod(firstpix_tmp-firstpix,itod),anynull,status)
             END DO

          ENDDO

       ENDIF

    ELSE ! no image no extension, you are dead, man
       STOP ' No image, no extension'
    ENDIF

    ! close the file
    CALL ftclos(unit, status)

    ! check for any error, and if so print out error messages
    IF (status .GT. 0) CALL printerror(status)

    RETURN

  END SUBROUTINE read_bintod_d
  ! ================================================================================

  !======================================================================================
  SUBROUTINE write_bintabh(tod, npix, ntod, header, nlheader, filename, extno, firstpix, repeat)
    !======================================================================================

    ! =================================================================================
    !     Create a FITS file containing a binary table extension in the first extension
    !
    !     Designed to deal with Huge file, (n_elements > 2^31)
    !
    !     OPTIONNAL NEW PARAMETERS:
    !     firstpix : position in the file of the first element to be written (start at 0) 
    !                default value =0
    !                8 bytes integer
    !                if NE 0 then suppose that the file already exists
    !
    !     repeat   : lenght of vector per unit rows and colomns of the first binary extension
    !                default value = 12000 (\equiv 1 mn of PLANCK/HFI data)
    !                4 byte integer
    ! 
    !     OTHER PARAMETERS
    !     unchanged as compare to the standard write_bintab of the HEALPIX package except 
    !     npix which is a 8 bytes integer
    !
    !     Adapted from write_bintab
    !                                           E.H. & O.D. @ IAP 07/02
    !
    !     Requires a compilation in 64 bits of the CFITSIO 
    !     Note that the flag -D_FILE_OFFSETS_BITS=64 has to be added 
    !         (cf page CFITIO 2.2 User's guide  Chap 4, section 4-13)
    ! 
    ! 2002-07-08 : bugs correction by E.H. 
    !    (uniform use of firstpix_tmp, introduction of firstpix_chunk)
    !==========================================================================================

    USE healpix_types
    IMPLICIT NONE

    INTEGER(I8B)     , INTENT(IN)           :: npix
    INTEGER(I8B)     , INTENT(IN), OPTIONAL :: firstpix
    INTEGER(I4B)     , INTENT(IN), OPTIONAL :: repeat
    INTEGER(I4B)     , INTENT(IN)           :: ntod,nlheader
    REAL(SP)         , INTENT(IN), DIMENSION(0:npix-1,1:ntod) :: tod
    CHARACTER(LEN=80), INTENT(IN), DIMENSION(1:nlheader)      :: header
    CHARACTER(LEN=*),  INTENT(IN)           :: filename
    INTEGER(I4B)    , INTENT(IN)     , OPTIONAL :: extno

    INTEGER(I4B) :: status,unit,blocksize,bitpix,naxis,naxes(1),repeat_tmp,repeat_fits
    INTEGER(I4B) :: i,npix_32
    LOGICAL(LGT) :: simple,extend
    CHARACTER(LEN=80) :: comment, ch

    INTEGER(I4B), PARAMETER :: maxdim = 20 !number of columns in the extension
    INTEGER(I4B)      :: nrows,tfields,varidat
    INTEGER(I4B)      :: frow,felem,colnum,readwrite,width,datacode,hdutype
    CHARACTER(LEN=20) :: ttype(maxdim), tform(maxdim), tunit(maxdim), extname
    CHARACTER(LEN=8)  :: date
    CHARACTER(LEN=10) :: fulldate
    CHARACTER(LEN=10) :: card
    CHARACTER(LEN=2)  :: stn
    INTEGER(I4B)      :: itn  

    INTEGER(I4B)      :: extno_i
    character(len=filenamelen) :: sfilename
    INTEGER(I8B) :: q,iq,npix_tmp,firstpix_tmp,firstpix_chunk

    !-----------------------------------------------------------------------


    IF (.NOT. PRESENT(repeat) ) THEN 
       repeat_tmp = 1
       if (mod(npix,1024_i8b) == 0) then
          repeat_tmp = 1024
       elseif (npix >= 12000) then
          repeat_tmp = 12000 
       endif
    ELSE 
       repeat_tmp = repeat
    ENDIF
    IF (.NOT. PRESENT(firstpix) ) THEN 
       firstpix_tmp = 0 
    ELSE 
       firstpix_tmp = firstpix
    ENDIF

    extno_i = 0
    if (present(extno)) extno_i = extno

    status=0
    unit = 100
    blocksize=1

    ! remove the leading '!' (if any) when reopening the same file
    sfilename = adjustl(filename)
    if (sfilename(1:1) == '!') sfilename = sfilename(2:filenamelen)

    ! create the new empty FITS file

    IF (firstpix_tmp .EQ. 0) THEN

       if (extno_i == 0) then 
          CALL ftinit(unit,filename,blocksize,status)

          ! -----------------------------------------------------
          ! Initialize parameters about the FITS image
          simple=.TRUE.
          bitpix=32     ! integer*4
          naxis=0       ! no image
          naxes(1)=0
          extend=.TRUE. ! there is an extension

          !     ----------------------
          !     primary header
          !     ----------------------
          !     write the required header keywords
          CALL ftphpr(unit,simple,bitpix,naxis,naxes,0,1,extend,status)

          !     writes supplementary keywords : none

          !     write the current date
          CALL ftpdat(unit,status) ! format (dd/mm/yy)

          !     update the date (format ccyy-mm-dd)
          CALL DATE_AND_TIME(date)
          fulldate = date(1:4)//'-'//date(5:6)//'-'//date(7:8)
          comment = 'FITS file creation date ccyy-mm-dd'
          CALL ftukys(unit,'DATE',fulldate,comment,status)

       !     ----------------------
       !     image : none
       !     ----------------------

       !     ----------------------
       !     extension
       !     ----------------------
       else

          !*********************************************
          !     reopen an existing file and go to the end
          !*********************************************
          call ftopen(unit,sfilename,1,blocksize,status)
          call ftmahd(unit,1+extno_i,hdutype,status)

       endif

       !     creates an extension
       CALL ftcrhd(unit, status)

       !     writes required keywords
       nrows    = npix / repeat_tmp ! naxis1
       tfields  = ntod
       WRITE(ch,'(i8)') repeat_tmp
       tform(1:ntod) = TRIM(ADJUSTL(ch))//'E'

       !      tform(1:ntod) = '1024E'
       IF (npix .LT. repeat_tmp) THEN
          nrows = npix
          tform(1:ntod) = '1E'
       ENDIF
       ttype(1:ntod) = 'simulation'   ! will be updated
       tunit(1:ntod) = ''      ! optional, will not appear
       extname  = ''      ! optional, will not appear
       varidat  = 0

       CALL ftphbn(unit, nrows, tfields, ttype, tform, tunit, &
            &     extname, varidat, status)

       !     write the header literally, putting TFORM1 at the desired place
       DO i=1,nlheader
          card = header(i)
          IF (card(1:5) == 'TTYPE') THEN ! if TTYPE1 is explicitely given
             stn = card(6:6)
             READ(stn,'(i1)') itn
             ! discard at their original location:
             CALL ftmcrd(unit,'TTYPE'//stn,'COMMENT',status)  ! old TTYPEi and 
             CALL ftmcrd(unit,'TFORM'//stn,'COMMENT',status)  !     TFORMi
             CALL ftprec(unit,header(i), status)           ! write new TTYPE1
             comment = 'data format of field: 4-byte REAL'
             CALL ftpkys(unit,'TFORM'//stn,tform(1),comment,status) ! and write new TFORM1 right after
          ELSEIF (header(i).NE.' ') THEN
             CALL ftprec(unit,header(i), status)
          ENDIF
10        CONTINUE
       ENDDO

    ELSE
       ! The file already exists
       readwrite=1
       CALL ftopen(unit,sfilename,readwrite,blocksize,status)
       CALL ftmahd(unit,2+extno_i,hdutype,status) 

       CALL ftgkys(unit,'TFORM1',tform(1),comment,status)
       CALL ftbnfm(tform(1),datacode,repeat_fits,width,status)

       IF (repeat_tmp .NE. repeat_fits) THEN
          WRITE(*,*) 'WARNING routine write_bintabh'
          WRITE(*,*) 'Inexact repeat value. Use the one read in the file'
       ENDIF

    ENDIF


    IF (npix .LT. nchunk_max) THEN ! data is small enough to be written in one chunk

       frow = (firstpix_tmp)/repeat_tmp + 1
       felem = firstpix_tmp-(frow-1)*repeat_tmp+1
       npix_32 = npix 

       DO colnum = 1, ntod
          CALL ftpcle(unit,colnum,frow,felem,npix_32,tod(0,colnum),status)
       END DO

    ELSE ! data has to be written in several chunks

       q = (npix-1)/nchunk_max
       DO iq = 0,q
          IF (iq .LT. q) THEN
             npix_tmp = nchunk_max
          ELSE
             npix_tmp = npix - iq*nchunk_max
          ENDIF
          firstpix_chunk = firstpix_tmp + iq*nchunk_max
          frow  = (firstpix_chunk)/repeat_tmp+1
          felem =  firstpix_chunk-(frow-1)*repeat_tmp+1 
          npix_32 = npix_tmp

          DO colnum = 1, ntod
             CALL ftpcle(unit,colnum,frow,felem,npix_32,tod(firstpix_chunk-firstpix_tmp,colnum),status)
          END DO
       ENDDO

    ENDIF

    ! ----------------------
    ! close and exit
    ! ----------------------

    ! close the file and free the unit number
    CALL ftclos(unit, status)

    ! check for any error, and if so print out error messages
    IF (status .GT. 0) CALL printerror(status)

    RETURN

  END SUBROUTINE write_bintabh
  ! ==============================================================================




end module fitstools
