module udgrade_nr
  ! subroutine sub_udgrade_nest
  ! subroutine udgrade_ring
  ! subroutine udgrade_nest

  use healpix_types

  IMPLICIT none

  private

  public :: sub_udgrade_nest, udgrade_ring, udgrade_nest

contains

  !=======================================================================
  subroutine sub_udgrade_nest(map_in, nside_in, map_out, nside_out, fmissval, pessimistic)
    !=======================================================================
    !
    !     SUB_UDGRADE_NEST(map_in, nside_in, map_out, nside_out)
    !
    !     map_in    : input map (REAL)
    !     nside_in  : resolution parameter of the input map (npix =12*nside**2)
    !                 must be a power of 2 (INTEGER)
    !     map_out   : output map (REAL)
    !     nside_out : resolution parameter of the output map (npix =12*nside**2)
    !                 must be a power of 2 (INTEGER)
    !
    !     upgrades or degrades a map in the NESTED numbering
    !
    !     if nside_out > nside_in : upgrades
    !     value(each new sub-pixel) = value(old big pixel)
    !
    !     if nside_in < nside_out : degrades
    !     value(new big pixel) = sum values(old small pixels) / number_of_small_pixels
    !
    !     version 1.0 : Eric Hivon, TAC, September 1997
    !     version 1.1 : Dec 1997, correction of a bug in the map indices
    !     version 1.2 : Dec 2001, addition of pessimistic and non-pessimistic behaviors
    !=======================================================================
    USE pix_tools, only : nside2npix
    INTEGER(I4B), INTENT(IN) :: nside_in, nside_out
    REAL(SP),     INTENT(IN),  DIMENSION(0:), target :: map_in
    REAL(SP),     INTENT(OUT), DIMENSION(0:) :: map_out
    REAL(SP),     INTENT(IN), OPTIONAL :: fmissval
    LOGICAL(LGT), INTENT(IN), OPTIONAL :: pessimistic
    
    INTEGER(I4B) :: npix_in, npix_out, npratio
    INTEGER(I4B) :: iu, ip, id
    LOGICAL(LGT), ALLOCATABLE, DIMENSION(:) :: good
    integer(I4B) :: nobs
    LOGICAL(LGT) :: do_pessimistic = .false.
    REAL(SP) :: bad_value
    REAL(SP), DIMENSION(:), POINTER :: p_in
    !-----------------------------------------------------------------------
    
    npix_out = nside2npix(nside_out)
    npix_in  = nside2npix(nside_in)
    
    bad_value = -1.6375e30
    if (present(fmissval)) then 
       bad_value = fmissval
    endif

    do ip=0, npix_out-1
       map_out(ip) = bad_value
    enddo
    
    if (nside_out < nside_in) then ! degrade
       if (present(pessimistic)) then
          do_pessimistic = pessimistic
       endif

       npratio = npix_in  / npix_out

       allocate(good(0:npratio-1))
       do id=0,npix_out-1
          p_in => map_in(id*npratio:(id+1)*npratio-1)
          good = (p_in /= bad_value)
          nobs = count(good)

          if (do_pessimistic) then
             if (nobs == npratio) then
                map_out(id) = sum(p_in) / nobs
             endif
          else
             if (nobs > 0) then
                map_out(id) = sum(p_in, mask=good) / nobs
             endif
          endif
       enddo

!        deallocate(p_in)
       deallocate(good)

!        allocate(p_in(0:npratio-1,0:npix_out-1))
!        p_in => map_in
! !       map_in = reshape(map_in, (/ 0:np
!        allocate(good(0:npratio-1,0:npix_out-1))
!        allocate (nobs(0:npix_out-1))
!        good = (p_in /= bad_value)
!        nobs = count(good,1)
!        if (do_pessimistic) then
!           where (nobs == npratio) 
!              map_out(:,0) = sum(p_in, dim=1, mask=good) / npratio
!           end where
!        else
!           where (nobs > 0) 
!              map_out(:,0) = sum(p_in, dim=1, mask=good) / nobs
!           end where
!        endif


    else ! upgrade
       npratio = npix_out  / npix_in
       do iu = 0, npix_out - 1
          ip = iu/npratio
          map_out(iu) = map_in(ip)
       enddo
    endif
    
    return
  end subroutine sub_udgrade_nest
!=======================================================================
  subroutine udgrade_ring(map_in, nside_in, map_out, nside_out, fmissval, pessimistic)
    !=======================================================================
    !
    !     UDGRADE_RING(map_in, nside_in, map_out, nside_out)
    !
    !     map_in    : input map (REAL)
    !                 This map will be converted to nested format on output.
    !     nside_in  : resolution parameter of the input map (npix =12*nside**2)
    !                 must be a power of 2 (INTEGER)
    !     map_out   : output map (REAL)
    !     nside_out : resolution parameter of the output map (npix =12*nside**2)
    !                 must be a power of 2 (INTEGER)
    !
    !     upgrades or degrades a map in the RING numbering
    !
    !     if nside_out > nside_in : upgrades
    !     value(each new sub-pixel) = value(old big pixel)
    !
    !     if nside_in < nside_out : degrades
    !     value(new big pixel) = sum values(old small pixels) / number_of_small_pixels
    !
    !
    !     calls the ring2nest and nest2ring routines available in healpix_tk.f file
    !
    !
    !
    !     NB : the speed is not the main concern of this routine,
    !     which not supposed to be used intensively for big maps
    !
    !     version 1.0 : Eric Hivon, TAC, September 1997
    !     version 1.1 : Dec 1997, correction of a bug in the map indices
    !=======================================================================
    USE pix_tools, only : nside2npix, ring2nest, nest2ring, convert_inplace_real
    INTEGER(I4B), INTENT(IN) :: nside_in, nside_out
    REAL(SP), INTENT(INOUT), dimension(0:), target :: map_in
    REAL(SP), INTENT(OUT), dimension(0:), target :: map_out
    REAL(SP), INTENT(IN), OPTIONAL :: fmissval
    LOGICAL(LGT) , INTENT(IN), OPTIONAL :: pessimistic
    
    real(sp), pointer, dimension(:) :: p_in, p_out
    INTEGER(I4B) :: npix_in, npix_out
    !-----------------------------------------------------------------------

    !     checks that the 2 nside are valid
    npix_out = nside2npix(nside_out)
    npix_in  = nside2npix(nside_in)
    if (npix_out < 0) then 
       print*,"wrong nside_out in udgrade_ring : ", nside_out
       stop
    endif
    if (npix_in  < 0) then
       print*,"wrong nside_in  in udgrade_ring : ", nside_in
       stop
    endif

    p_in  => map_in
    p_out => map_out
    call convert_inplace_real(ring2nest, p_in)
    call sub_udgrade_nest(map_in, nside_in, map_out, nside_out, fmissval, pessimistic)
    call convert_inplace_real(nest2ring, p_out)
    
    return
  end subroutine udgrade_ring


  !=======================================================================
  subroutine udgrade_nest(map_in, nside_in, map_out, nside_out, fmissval, pessimistic)
    !=======================================================================
    !
    !     UDGRADE_NEST(map_in, nside_in, map_out, nside_out)
    !
    !     map_in    : input map (REAL)
    !     nside_in  : resolution parameter of the input map (npix =12*nside**2)
    !                 must be a power of 2 (INTEGER)
    !     map_out   : output map (REAL)
    !     nside_out : resolution parameter of the output map (npix =12*nside**2)
    !                 must be a power of 2 (INTEGER)
    !
    !     upgrades or degrades a map in the NESTED numbering
    !
    !     if nside_out > nside_in : upgrades
    !     value(each new sub-pixel) = value(old big pixel)
    !
    !     if nside_in < nside_out : degrades
    !     value(new big pixel) = sum values(old small pixels) / number_of_small_pixels
    !
    !     version 1.0 : Eric Hivon, TAC, September 1997
    !     version 1.1 : Dec 1997, correction of a bug in the map indices
    !=======================================================================
    USE pix_tools, only : nside2npix
    INTEGER(I4B), INTENT(IN) :: nside_in, nside_out
    REAL(SP), INTENT(IN), dimension(0:) :: map_in
    REAL(SP), INTENT(OUT), dimension(0:) :: map_out
    REAL(SP), INTENT(IN), OPTIONAL :: fmissval
    LOGICAL(LGT) , INTENT(IN), OPTIONAL :: pessimistic
    
    INTEGER(I4B) :: npix_in, npix_out
    !-----------------------------------------------------------------------
    
    !     checks that the 2 nside are valid
    npix_out = nside2npix(nside_out)
    npix_in  = nside2npix(nside_in)
    if (npix_out < 0) then 
       print*,"wrong nside_out in udgrade_nest : ", nside_out
       stop
    endif
    if (npix_in  < 0) then
       print*,"wrong nside_in  in udgrade_nest : ", nside_in
       stop
    endif
    
    
    call sub_udgrade_nest(map_in, nside_in, map_out, nside_out, fmissval, pessimistic)

    
    return
  end subroutine udgrade_nest

end module udgrade_nr



