PRO FITS2CL, cl_array, fitsfile, HDR = hdr, XHDR = xhdr, SILENT=silent, MULTIPOLES=multipoles

;+
; NAME:
;       FITS2CL
;
; PURPOSE:
;       read from a FITS file an ascii table extension containing power 
;       spectrum coefficients and return them. Reads header information
;       if required.
;
; CALLING SEQUENCE:
;       FITS2CL, cl_array, fitsfile, [HDR = , XHDR =, SILENT = ]
; 
; INPUTS:
;       fitsfile = String containing the name of the file to be written      
;
; OPTIONAL INPUTS:
;       SILENT = if set, no comments are written while the file is read
;
; OUTPUTS:
;       cl_array = real array of Cl coefficients read from the file.
;                  This has dimension (lmax+1) or (lmax+1,4) or
;                  (lmax+1,6) given in the sequence T E B TxE TxB BxE
;                  The convention for the power
;                  spectrum is 
;                  cl = Sum_m {a_lm^2}/(2l+1)
;                  ie. NOT normalised by the Harrison-Zeldovich spectrum.
;
; OPTIONAL OUTPUT KEYWORDS:
;       HDR      = String array containing the header for the FITS file.
;       XHDR     = String array containing the extension header for the 
;                  extension.
;       MULTIPOLES = array containing the multipoles l for which the
;       power spectra are provided
;           * either read from the file (1st column in the Planck format),
;           * or generated by the routine (assuming that all
;               multipoles from 0 to lmax included are provided).
;
; EXAMPLE:
;       Read the power spectrum coefficients into the real array, pwr,
;       from the FITS file, spectrum.fits, with extension header information
;       contained in the string array, ext_text
;
;       IDL> fits2cl,pwr,'spectrum.fits',XHDR=ext_txt
;
;
; PROCEDURES CALLED:
;       HEADFITS, MRDFITS
;
; MODIFICATION HISTORY:
;       May 1999: written by A.J. Banday (MPA)         
;       Oct 2001, EH : can now deal with 6 columns,
;                      added silent keyword
;       Dec 2002, EH : can deal with Planck format (1st column is
;       multipole),
;                      added multipoles output variable
;
; requires the THE IDL ASTRONOMY USER'S LIBRARY 
; that can be found at http://idlastro.gsfc.nasa.gov/homepage.html
;
;-

code = 'FITS2CL'
syntax = 'Syntax : '+code+', cl_array, fitsfile, [HDR = , XHDR =, SILENT=, MULTIPOLES=] '
if N_params() NE 2  then begin
      print,syntax
      print,'   file NOT read '
      goto, Exit
endif

if datatype(cl_array) eq 'STR' or datatype(fitsfile) ne 'STR' then begin
      print,syntax
      print,'   the array comes first, then the file name '
      print,'   file NOT read '
      goto, Exit
endif

fullpath = findfile(fitsfile,count=count)
if (count eq 0) then begin
    message,'file '+fitsfile+' not found'
endif

; run astrolib routine to set up non-standard system variables
defsysv, '!DEBUG', EXISTS = i  ; check if astrolib variables have been set-up
if (i ne 1) then astrolib      ; if not, run astrolib to do so

hdr  = HEADFITS(fitsfile)
xhdr = HEADFITS(fitsfile,EXTEN=1)


; -------- Planck format ? (ie, first column is multipole)
pdmtype = sxpar(xhdr,'PDMTYPE',count=nextra)
if (nextra gt 0) then junk = where(strupcase(pdmtype) eq 'POWERSPEC', nextra)
nextra = nextra < 1 ; either 0 or 1

if (nextra eq 0) then begin
    ttype1 = sxpar(xhdr,'TTYPE1',count=nttype1)
    if (strupcase(ttype1) eq 'L') then nextra = 1
endif

if ( (not keyword_set(silent) and (nextra eq 1))) then print,'Info : C(l) file '+fitsfile+' has Planck format'
; -------- extension -----------

; simply read the extension from the FITS file
tmpout = MRDFITS(fitsfile,1,/use_colnum,silent=silent)

; get the dimensions of the input array
info  = size(tmpout)
nrows = info(1) ; # of entries for l-range: nrows = lmax+1
ncols = n_elements(tag_names(tmpout)) - nextra

; if ( (ncols ne 1) and (ncols ne 4) )then begin
if ( (ncols ne 1) and (ncols ne 4) and (ncols ne 6)) then begin
  print,' Input file does not conform to expected structure'
  print,code+' expects either 1, 4 or 6 columns,      found ',ncols
  goto, Exit
endif

if nextra eq 1 then begin
    multipoles = tmpout.(0)
endif else begin
    multipoles = findgen(nrows)
endelse

; output array
if (ncols eq 1) then begin
    cl_array = tmpout.(0+nextra)
endif else begin
    cl_array = fltarr(nrows,ncols)
    for i=0,ncols-1 do begin
        cl_array[*,i] = tmpout.(i+nextra)
    endfor
endelse

; Exit routine
Exit:
return
end



