pro read_fits_cut4, filename, pixel, signal, n_obs, serror, hdr = header, xhdr = xheader
;+
; NAME:
;          read_fits_cut4
;
; PURPOSE:
;         reads a FITS file for incomplete sky coverage with a 4 column format
;        (similar to the one used by DMR) :
;        PIXEL, SIGNAL, N_OBS, SERROR
;
; CATEGORY:
;        fits I/O for Healpix
;
; CALLING SEQUENCE:
;        read_fits_cut4, Filename, Pixel, Signal, N_obs, Serror, Header=, XHeader=
; 
; INPUTS:
;       Filename  : STRING scalar,      
;                   input filename
;
; OUTPUTS:
;        Pixel   : LONG vector
;                   index of Healpix pixel
;
;       Signal   : FLOAT vector (same length as Pixel)
;                  signal in each pixel (eg, average of many measurements)
;
;       N_obs    : INT or LONG vector (same length as Pixel)
;                  number of measurements in each pixel
;      
;       Serror  : FLOAT vector (same length as Pixel)
;                 statistical error on Signal in each Pixel
;                 = rms of individual measurement / sqrt(N_obs)
;      
; OPTIONAL OUTPUTS:
;       Hdr : STRING vector 
;                contains the FITS header of the primary unit
;
;       Xhdr : STRING vector 
;                contains the FITS header of the extension (the most informative)
;
; RESTRICTIONS:
;
; PROCEDURE:
;      calls fxbreadm
;
; EXAMPLE:
;
; MODIFICATION HISTORY:
;        Sept 2000, EH, Caltech
;        Aug  2002: use HDR and XHDR keywords to match other routines
;-

routine = 'READ_FITS_CUT4'
if N_params() LT 3 then begin
      print,'Syntax : '+routine+', filename, pixel, signal, [n_obs, serror, 
      print,'      header=, xheader=]'
      return
endif
if (datatype(filename) ne 'STR') then begin
      print,'Syntax : '+routine+' filename, pixel, signal, [n_obs, serror, hdr=, xhdr=]'
      return
endif

; run astrolib routine to set up non-standard system variables
defsysv, '!DEBUG', EXISTS = i  ; check if astrolib variables have been set-up
if (i ne 1) then astrolib       ; if not, run astrolib to do so

; check file existence and get basic keywords
fits_npix =getsize_fits(filename, nmaps=nmaps, obs_npix = obs_npix, type = ftype)

if (ftype ne 3) then begin
    print,routine+': '+filename+' does not have the right format'
    return
endif

; get the primary header information
header = headfits(filename)

; open extension 1 and get header
fxbopen, lun, filename, 1, xhdr

; read extension data
if (obs_npix le 0) then obs_npix = fits_npix
                                ; simultaneous read of 4 columns
fxbreadm, lun, ['PIXEL','SIGNAL','N_OBS','SERROR'], pixel, signal, n_obs, serror,  status = status
fxbclose, lun
if (total(status) lt 2) then begin
    print,'Wrong format for fits file '+filename
    print,'only could read ',total(status),' columns'
    stop
endif
if (n_elements(pixel) eq n_elements(signal)) then begin ; standard explicit format
    if (obs_npix lt fits_npix) then begin
        if (status[0]) then pixel  = temporary( pixel[0:obs_npix-1])
        if (status[1]) then signal = temporary(signal[0:obs_npix-1])
        if (status[2]) then n_obs  = temporary( n_obs[0:obs_npix-1])
        if (status[3]) then serror = temporary(serror[0:obs_npix-1])
    endif
endif else begin ; compressed explicit format
    print,'inconstitent pixel index and data value in '+filename
    stop
endelse

xheader = xhdr

return
end

