PRO read_tqu, fitsfile, TQU, EXTENSION=extension_id, HDR = hdr, XHDR = xhdr, HELP = help

;+
; NAME:
;    read_tqu
;
; PURPOSE:
;    reads a temperature+polarization Healpix map (T,Q,U)  from a
;    binary table FITS file
;    with optionally the error (dT,dQ,dU) and correlation (dQU, dTU, dTQ)
;    from separate extensions
;
; CALLING SEQUENCE:
;    READ_TQU, fitsfile, TQU, EXTENSION=, HDR = , XHDR = , HELP =
; 
; INPUTS:
;    fitsfile = String containing the name of the file to be read   
;
; OPTIONAL INPUTS:
;    Extension : extension to be read (0 based)
;         if absent, all the extensions found are read
;
; KEYWORDS:
;    Help : if set, provides help on routine
;
; OUTPUTS:
;
;   TQU : array of Healpix maps of size (npix,3,n_ext) where npix is the total
;   number of Healpix pixels on the sky, and (n_ext <=3) the number of
;   extensions read
;     3 maps are present in each extension of the FITS file : 
;      -the temperature+polarization Stokes parameters maps (T,Q,U) in extension 0
;      -the error maps (dT,dQ,dU) in extension 1 (if applicable)
;      -the correlation maps (dQU, dTU, dTQ) in extension 2 (if applicable)
;
;
; OPTIONAL OUTPUT KEYWORDS:
;       HDR      = String array containing the header of the FITS file.
;       XHDR     = String array containing the extension header(s). If 
;                  several extensions are read, then the extension 
;                  headers are returned appended into one string array.
;
; EXAMPLE:
;
;
; PROCEDURES CALLED:
;       DOC_LIBRARY, FINDFILE, FITS_INFO, HEADFITS, READ_FITS_S
;
; MODIFICATION HISTORY:
;       Dec 2002, E. Hivon (McMurdo)
;
; requires the THE IDL ASTRONOMY USER'S LIBRARY 
; that can be found at http://idlastro.gsfc.nasa.gov/homepage.html
;
;-

if (keyword_set(help)) then begin
    doc_library,'read_tqu'
    return
endif

syntax = 'SYNTAX : READ_TQU, fitsfile, TQU, Extension=, Hdr=, Xhdr=, Help='
if n_params() ne 2 then begin
    print,syntax,form='(a)'
    message,'Not enough argument, no file written'
endif

fullpath = findfile(fitsfile,count=count)
if (count eq 0) then begin
    message,'file '+fitsfile+' not found'
endif

; run astrolib routine to set up non-standard system variables
defsysv, '!DEBUG', EXISTS = i  ; check if astrolib variables have been set-up
if (i ne 1) then astrolib      ; if not, run astrolib to do so

fits_info, fitsfile, /silent, n_ext = n_ext

if undefined(extension_id) then begin
    x0 = 0
    xf = n_ext - 1
endif else begin
    if (extension_id + 1) gt n_ext then begin
        print,' Required extension ',extension_id,' (0 based) from ',fitsfile
        print,' Only found ',n-ext,' extensions.'
        message,' Abort'
    endif else begin
        x0 = extension_id
        xf = x0
        n_ext = 1
    endelse
endelse
hdr  = HEADFITS(fitsfile)
print,' Reading ',n_ext,' extension(s) from ',fitsfile

; -------- extension -----------

; simply read the extensions from the FITS file
savehdr = ''
for i = 0,xf-x0 do begin
    read_fits_s, fitsfile, prim_stc, xten_stc, exten=i+x0

    nmaps = n_tags(xten_stc) - 1
    if (nmaps ne 3) then begin
        print,' WARNING : ',nmaps,' maps available in extension ',i+x0,' of '+fitsfile
        print,'      Expected 3'
    endif
    xhdr = xten_stc.(0)
    savehdr = [savehdr,xhdr]
    npix = n_elements(xten_stc.(1)) 
    if (i eq 0) then begin
        tqu = fltarr(npix,nmaps,n_ext)
    endif
    for j=1,nmaps do begin
        tqu[*,j-1,i] = xten_stc.(j)
    endfor

endfor

xhdr = savehdr

; Exit routine
Exit:
return
end



