pro write_tqu, fitsfile, TQU, Coordsys=coordsys, Nested=nested, Ring=ring, Ordering=ordering, Extension=extension_id, Hdr=hdr, Xhdr=xhdr, Units=units, Help=help
;+
; NAME:
;    write_tqu
;
; PURPOSE:
;    writes a temperature+polarization Healpix map (T,Q,U) into a
;    binary table FITS file, 
;    with optionally the error (dT,dQ,dU) and correlation (dQU, dTU, dTQ)
;    in separate extensions
;
; CATEGORY:
;
;
; CALLING SEQUENCE:
;   WRITE_TQU, fitsfile, TQU, Coordsys=, Nested=, Ring=, Ordering=, $
;                      Nside=, extension=, Hdr=, Xhdr=, Units=, Help=
; 
; INPUTS:
;   fitsfile : output filename
;
;   TQU : array of Healpix maps of size (npix,3,n_ext) where npix is the total
;   number of Healpix pixels on the sky, and n_ext <=3.
;     3 maps are written in each extension of the FITS file : 
;      -the temperature+polarization Stokes parameters maps (T,Q,U) in extension 0
;      -the error maps (dT,dQ,dU) (if n_ext >=2) in extension 1
;      -the correlation maps (dQU, dTU, dTQ) (if n_ext = 3) in extension 2
;
;     it is also possible to write 3 maps directly in a given
;     extension (provided the preceding extension, if any, is already
;     filled in)
;     by setting EXTENSION to the extension number in which to write
;     (0 based) and if n_ext + Extension <= 3
;
; OPTIONAL INPUTS:
;      
; KEYWORD
;    Ring-     if set, add 'ORDERING= RING' to the extension fits header
;    Nested-   if set, add 'ORDERING= NESTED' to the extension fits header
;    Ordering- if set to the string 'nested' or 'ring', set the
;              keyword 'ORDERING' to the respective value
;
;    one of them has to be present unless the ordering information is already
;    present in the fits header.
;
; OPTIONAL KEYWORD
;   Coordsys = if set to either 'C', 'E' or 'G' specifies that the
;     Healpix coordinate system is respectively Celestial=equatorial, Ecliptic or
;     Galactic
;   
;   Extension = extension unit a which to put the data (0 based)
;
;   Help= if set, an extensive help (this IDL header) is printed
;
;   Hdr= string containing the primary FITS header
;
;   Units = physical units of the maps
;
;   Xhdr= string containing the extension FITS header (it will be
;    repeated in each extension, except for TTYPE* and EXTNAME which
;    are generated by the routine and depend on the extension)
;
; OUTPUTS:
;
; OPTIONAL OUTPUTS:
;
; COMMON BLOCKS:
;
; SIDE EFFECTS:
;
; RESTRICTIONS:
;
; PROCEDURE:
;   calls WRITE_FITS_SB
;
; EXAMPLE:
;
; MODIFICATION HISTORY:
;    version 1.0, Eric Hivon, Nov-Dec 2002
;
;-

if (keyword_set(help)) then begin
    doc_library,'write_tqu'
    return
endif

syntax = ['SYNTAX : WRITE_TQU, fitsfile, TQU, Coordsys=, Nested=, Ring=, Ordering=, $',$
          '            Extension=, Hdr=, Xhdr=, Units=, Help=']
if n_params() lt 2 then begin
    print,syntax,form='(a)'
    message,'Not enough argument, no file written'
endif

if (datatype(fitsfile) ne 'STR' or datatype(TQU) eq 'STR') then begin
    print,syntax,form='(a)'
    message,'Error in argument type, no file written'
endif

sz = size(TQU)
ndim = sz[0]
if (ndim lt 2) then message,' Expect a Npix * 3  TQU array'
npix = sz[1]
if (npix2nside(npix) lt 0) then message,' TQU 1st dimension (Npix) is not a valid Npix pixel number'
nmap = sz[2]
if (nmap ne 3) then message,' Expect a Npix * 3  TQU array'
if (ndim ge 3) then n_ext = sz[3] else n_ext = 1

i_ext0 = 0
if keyword_set(extension_id) then i_ext0 = extension_id
if (n_ext + i_ext0) gt 3 then begin
    print, i_ext0, n_ext
    message,' can not write more than 3 extensions'
endif


;-----------------
; create structure for primary unit
punit = 0
if defined(hdr) then punit = create_struct('HDR',hdr)

;------------------
; create names for extensions

if undefined(xhdr) then xhdr = [' ']

; name columns
name =  [['TEMPERATURE',  'Q_POLARIZATION',  'U_POLARIZATION'], $
         ['dT',           'dQ',              'dU'], $
         ['dQU',          'dTU',             'dTQ']]

; extension names
xtname = ['SIGNAL','ERROR','CORRELATION']

for i_ext=i_ext0, i_ext0+n_ext-1 do begin
; create structure for 1st/2nd/3rd extension (number 0,1,2)
    sxaddpar,xhdr,'EXTNAME',xtname[i_ext]

; add UNITS information
    if defined(units) then begin
        sunits = units
        if (i_ext eq 2) then sunits=strtrim(units,2)+'**2'
        add_units_fits,xhdr,units=sunits,error=error,colnum=[1,2,3]
        if error ne 0 then return
    endif

    ia = i_ext - i_ext0
    xtns = create_struct('HDR', xhdr, $
                         name[0,i_ext], TQU[*,0,ia],$
                         name[1,i_ext], TQU[*,1,ia],$
                         name[2,i_ext], TQU[*,2,ia])

; write it
    ii_ext = i_ext
    write_fits_sb, fitsfile, punit, xtns, $
      Coordsys=coordsys, Nested=nested, Ring=ring, Ordering=ordering, Extension=ii_ext
endfor


return
end

