pro ud_grade_cut4, pixel, signal, n_obs, serror, nside_in = nside_in, $
          nside_out = nside_out, order_in = order_in, order_out = order_out, pessimistic = pessimistic
;+
;
; UD_GRADE_CUT4,    $
;     Pixel, Signal [, N_Obs, Serror],   $
;     Nside_in=, [Nside_out=,] Order_in=, [Order_out=], [Pessimistic=]
;
;
; up_gradation, de_gradation for the fields
;  Pixel (pixel index), 
;  Signal (pixel temperature = average of several observations), 
;  N_obs (number of observations in pixel), 
;  Serror (error on temperature = rms of average)
;
;
;  degradation :  K small pixels (i) -> 1 big pixel (j)
;    N_Obs :    n_i  -> nf_j = sum_i (n_i)
;    Signal:    t_i  -> tf_j = sum_i (n_i t_i)  / sum_i (n_i)
;    Serror:    
;     e_i  -> ef_j = sqrt[ sum_i(n_i(t_i-tf_j)^2) + sum_i (n_i s_i)^2] / sum_i n_i
;    Pixel : at least one exposed small pixel -> 1 exposed big pixel    
;
;  upgradation : 1 big pixel(j) -> K small *identical* pixels (i)
;    N_obs :    nf_j   -> n_i = nf_j / K (approximated to closest integer)
;    Signal:    tf_j   -> t_i = tf_j
;    Serror:    ef_j   -> e_i = ef_j * sqrt(K)
;
;
;   EH, 2000-11-23
;
;-
routine = 'UD_GRADE_CUT4'
if (undefined(nside_in) or undefined(order_in)) then begin
    print,routine+': Pixel, Signal [, N_Obs, Serror], Nside_in=, [Nside_out=,] Order_in=, [Order_out=]'
    print,'******** No processing ********'
    return
endif
if undefined(nside_out) then nside_out = nside_in
if undefined(order_out) then order_out = order_in
npix_in = nside2npix(nside_in)
npix_out = nside2npix(nside_out)
if (npix_in le 0 or npix_out le 0) then begin
    print,routine+': Invalid Nside_in or Nside_out :',nside_in,nside_out
    stop
endif

ord_in  = strmid(order_in,0,4)
ord_out = strmid(order_out,0,4)

if ((ord_in eq ord_out) and (nside_in eq nside_out)) then begin
; ------------
; nothing
; ------------
    return
endif


ok_obs = 0
if defined(n_obs) then begin
    if (max(n_obs) gt 0) then ok_obs = 1
endif
if (nside_out lt nside_in) then begin 
; ------------
; degradation
; ------------
    ratio = (nside_in/nside_out)
    ratio2 = ratio * ratio

    if (ord_in eq 'RING') then begin
        ring2nest, nside_in, pixel, pixel
    endif

    minhit = 1                                         ; at least one good sub-pixel
    if (keyword_set(pessimistic)) then minhit = ratio2 ; all sub pixel should be good

    hitin = bytarr(ratio2,npix_out)
    hitin[pixel] = 1B
    hitout = total(hitin,1)
    if (ok_obs) then begin
        obsin = lonarr(ratio2,npix_out)
        obsin[pixel] = n_obs
        obsout = total(obsin,1) ; total number of hits per large pixel
    endif else begin
        obsin = hitin
        obsout = hitout         ; number of exposed subpixels per large pixel
    endelse
    
    mapin = fltarr(ratio2,npix_out)
    mapin[pixel] = signal
    mapout = total(obsin * mapin,1)
    
    pixelout  = where(obsout gt 0 and hitout ge minhit, obs_npix)
    signalout = mapout[pixelout]/obsout[pixelout] ; average temperature per large pixel
    
    if defined(serror)  then begin
        mapav = fltarr(ratio2,npix_out)
        mapav[pixel] = signalout[pixel/ratio2]
        errin = fltarr(ratio2,npix_out)
        errin[pixel] = serror
        errout = sqrt(   total((obsin*errin)^2,1)  + total(obsin*(mapin-mapav)^2,1)     )
        errin = 0
        mapav = 0
    endif
    
    mapin = 0
    obsin = 0
    
    pixel = temporary(pixelout)
    signal = temporary(signalout)
    if defined(n_obs)  then n_obs = obsout[pixel]
    if defined(serror) then serror = errout[pixel]/obsout[pixel]
    nside = nside_out
    npix = npix_out
    
    if (ord_out eq 'RING') then begin
        nest2ring, nside, pixel, pixr
        
        ind = sort(pixr)
        pixel = pixr[ind]
        signal = signal[ind]
        if defined(n_obs) then n_obs = n_obs[ind]
        if defined(serror) then serror = serror[ind]
        ind = 0
    endif
endif else begin
; ------------
; upgradation
; ------------
    ratio = (nside_out/nside_in)
    ratio2 = ratio * ratio

    ; go to NEST
    if (ord_in eq 'RING') then begin
        ring2nest, nside_in, pixel, pixel
    endif

    subpix = replicate(1.,ratio2)
    increm = lindgen(ratio2)

    npix = n_elements(pixel)
    npix_out = npix * ratio2

    pixel_out = fltarr(ratio2,npix)
    for i=0l, npix-1 do begin
        pixel_out[*,i] = pixel[i] * ratio2 + increm
    endfor
    pixel  = reform(pixel_out, npix_out)

    signal = signal ## (subpix)       
    signal = reform(signal, npix_out, /overwrite)

    if defined(n_obs) then begin
        n_obs  = n_obs  ## (subpix/ratio2)
        n_obs  = reform(round(n_obs) , npix_out, /overwrite)
    endif
    if defined(serror) then begin
        serror = serror ## (subpix*ratio) 
        serror = reform(serror, npix_out, /overwrite)
    endif
    npix = npix_out
    nside = nside_out

    ; go to output order
    if (ord_out eq 'RING') then begin
        nest2ring, nside, pixel, pixel
    endif

    ; sort 
    if (ord_in eq 'RING' or ord_out eq 'RING') then begin
        ind = sort(pixel)
        pixel = pixel[ind]
        signal = signal[ind]
        if defined(n_obs) then n_obs = n_obs[ind]
        if defined(serror) then serror = serror[ind]
    endif

endelse

    
return
end
