PRO vec2pix_nest, nside, vec_in, ipix
;*******************************************************************
;+
; VEC2PIX_NEST, Nside, Vec_in, Ipix
;
;        renders the NESTED scheme pixel number Ipix for a pixel which, given the
;        map resolution parameter Nside, contains the point on the sphere
;        at cartesian coordinates Vec_in
;
; INPUT
;    Nside     : determines the resolution (Npix = 12* Nside^2)
;            should be a power of 2 (not tested)
;	     SCALAR
;    Vec_in    : (x,y,z) position unit vector(s) with North pole = (0,0,1)
;            stored as x(0), x(1), ..., y(0), y(1), ..., z(0), z(1) ..
;            ARRAY of dimension = (np,3)
;
; OUTPUT
;    Ipix  : pixel number in the RING scheme of HEALPIX pixelisation in [0,Npix-1]
;	is a VECTOR of dimension = (np)
;
; SUBROUTINE
;    nside2npix
;
; HISTORY
;    June-October 1997,  Eric Hivon & Kris Gorski, TAC
;    Feb 1999,           Eric Hivon,               Caltech
;    Sept 2000,          EH
;           free memory by discarding unused variables
;
;-
;*****************************************************************************

routine = 'VEC2PIX_NEST'
if (N_ELEMENTS(nside) GT 1) then stop, 'STOP, Nside should be a scalar in '+routine
npix = nside2npix(nside, error = error)
if (error ne 0) then stop, 'STOP, Invalid Nside ',nside,' in '+routine

if N_params() ne 3 then begin
    print,' syntax = '+routine+', nside, vec, ipix'
    stop
endif


np1 = N_ELEMENTS(vec_in)
np = LONG(np1/3) 
if (np1 NE np*3) then begin
	print,'inconsistent vec_in in '+routine
        print,routine+', nside, vec_in, ipix'
	stop
endif
vec_in = reform(vec_in,np,3,/OVERWRITE)
;------------------------------------------------------------
ns_max = 8192L

common xy2pix, x2pix, y2pix
sz = size(x2pix)
if (sz(sz(0)+1) eq 0) then init_xy2pix ; initiate x2pix and y2pix

nside = LONG(nside)
twopi  = 2.d0 * !DPI
piover2 = !DPI / 2.d0
z0 = 2.d0 /3.d0

z = DOUBLE(vec_in(*,2)/SQRT( TOTAL(vec_in^2,2) ) )
phi_in = ATAN(vec_in(*,1), vec_in(*,0))
phi_in = phi_in + (phi_in lt 0.0  )*twopi
tt  = phi_in / piover2 ; in [0,4[
phi_in = 0


face_num = BYTARR(np)
ix = INTARR(np) & iy=ix
pix_eqt = WHERE(z LE z0 AND z GT -z0, n_eqt) ; equatorial strip
IF (n_eqt GT 0) THEN BEGIN

;     (the index of edge lines increase when the longitude=phi goes up)
      jp = LONG(ns_max*(0.5d0 + tt(pix_eqt) - z(pix_eqt)*0.75d0)) ;  ascend edge line index
      jm = LONG(ns_max*(0.5d0 + tt(pix_eqt) + z(pix_eqt)*0.75d0)) ; descend edge line index

;     finds the face
      face_n = BYTARR(n_eqt)
      ifp = BYTE(jp / ns_max)  ; in {0,4}
      ifm = BYTE(jm / ns_max)
      p_np = WHERE(ifp EQ ifm, n_np)
      p_eq = WHERE(ifp LT ifm, n_eq)
      p_sp = WHERE(ifp GT ifm, n_sp)
      if (n_np GT 0) then face_n(p_np) = (ifp(p_np) MOD 4B) + 4B
      if (n_eq GT 0) then face_n(p_eq) = (ifp(p_eq) MOD 4B)
      if (n_sp GT 0) then face_n(p_sp) = (ifm(p_sp) MOD 4B) + 8B

      face_num(pix_eqt) = face_n
      ix(pix_eqt) = jm MOD ns_max
      iy(pix_eqt) = ns_max - (jp MOD ns_max) - 1

      jp=0 & jm=0 & face_n=0 & ifp=0 & ifm=0 & face_n=0 & p_np=0 & p_eq=0 & p_sp=0
      pix_eqt = 0
ENDIF

pix_pol = WHERE(z GT z0 OR  z LE -z0, n_pol) ; polar caps
IF (n_pol GT 0) THEN BEGIN

      zz = z(pix_pol)
      ntt = FIX(tt(pix_pol)) < 3
      tp = tt(pix_pol) - ntt
      tmp = SQRT( 3.d0*(1.d0 - ABS(z(pix_pol))) )  ; in ]0,1]

;     (the index of edge lines increase when distance from the closest pole goes up)
      jp = FIX( ns_max * tp          * tmp ) ; line going toward the pole as phi increases
      jm = FIX( ns_max * (1.d0 - tp) * tmp ) ; that one goes away of the closest pole
      jp = jp < (ns_max-1) ; for points too close to the boundary
      jm = jm < (ns_max-1)

;     finds the face and pixel's (x,y)
      p_np = WHERE(zz gt 0., n_np)
      p_sp = WHERE(zz lt 0., n_sp)
      if (n_np GT 0) then begin
	face_num(pix_pol(p_np)) = ntt(p_np)
	ix(pix_pol(p_np)) = ns_max - jm(p_np) - 1
	iy(pix_pol(p_np)) = ns_max - jp(p_np) - 1
      endif
      if (n_sp GT 0) then begin
	face_num(pix_pol(p_sp)) = ntt(p_sp) + 8
	ix(pix_pol(p_sp)) = jp(p_sp)
	iy(pix_pol(p_sp)) = jm(p_sp)
      endif

      ntt=0 & tp=0 & tmp=0 & jp=0 & jm=0 & p_np=0 & p_sp=0
      pix_pol = 0
ENDIF

ix_low = ix MOD 128
ix_hi  =     ix/128
iy_low = iy MOD 128
iy_hi  =     iy/128
ix=0 & iy=0


ipix =  (x2pix(ix_hi )+y2pix(iy_hi )) * 16384L + (x2pix(ix_low)+y2pix(iy_low))

ipix = ipix / ( ns_max/nside )^2  ; in {0, nside**2 - 1}

ipix = ipix + face_num* LONG(nside)^2  ; in {0, 12*nside**2 - 1}

RETURN
END ; vec2pix_nest


;=======================================================================
; The permission to use and copy this software and its documentation, 
; without fee or royalty is limited to non-commercial purposes related to 
; Microwave Anisotropy Probe (MAP) and
; PLANCK Surveyor projects and provided that you agree to comply with
; the following copyright notice and statements,
; and that the same appear on ALL copies of the software and documentation.
;
; An appropriate acknowledgement has to be included in any
; publications based on work where the package has been used
; and a reference to the homepage http://www.tac.dk/~healpix
; should be included
;
; Copyright 1997 by Eric Hivon and Kris Gorski.
;  All rights reserved.
;=======================================================================
