pro outline_coord2uv, outline, coord_out, eul_mat, outline_uv, projection=projection, mollweide=mollweide, gnomic=gnomic, cartesian=cartesian, orthographic=orthographic, flip = flip, show=show, half_sky=half_sky, _extra = oplot_kw
;+
; NAME:
;     outline_coord2uv
;
;
; PURPOSE:
;     turn a outline in astro coordinate, into a outline in uv coordinate
;     both input and output are structures,
;     if Show is set, overplot the result on current plot
;     Note : the outline may be open, the procedure will NOT close it
;
; CATEGORY:
;     internal processing for data visualisation
;
;
; CALLING SEQUENCE:
;     OUTLINE_COORD2UV, Outline, Coord_Out, Eul_Mat, [Outline_Uv]
;           Projection=,Mollweide=, Gnomic=, Cartesian=, Orthographic=,Flip=,Show=,half_sky
;
; 
; INPUTS:
;     Outline : structure containing
;       - coordinate
;       - RA (or longitude) vector
;       - Dec (or latitude) vector
;       - line style 
;          +2 : black dashes
;          +1 : black dots
;           0 : black solid
;          -1 : black dots on white background
;          -2 : black dashes on white background
;          (default = 0 : solid)
;
;     Coord_Out, Eul_Mat
;
;
; KEYWORD PARAMETERS:
;    Mollweide =
;    Gnomic =
;    Cartesian =
;    Orthographic =
;    Flip = 
;    Projection = either 'MOLL*', 'GNOM*', 'CART*' or 'ORTH*'
;
; OPTIONAL OUTPUTS:
;    Outline_Uv
;
;
; COMMON BLOCKS:
;    none
;
;
; SIDE EFFECTS:
;    oplot contour on existing plot if Show is set
;
;
; PROCEDURE:
;   makes a UV projection in either Mollweide or Gnomic scheme
;
;   calls: datatype, vec2moll
;
; EXAMPLE:
;
;
;
; MODIFICATION HISTORY:
;       2002-03-21, E. Hivon, IPAC, v1.0
;
;-

; outline : from astro coordinate to uv plan

identify_projection, projtype, projection=projection, mollweide=mollweide, gnomic=gnomic

if keyword_set(flip) then flipconv=1 else flipconv = -1  ; longitude increase leftward by default (astro convention)


if (datatype(outline) ne 'STC') then return

n_outlines = n_tags(outline)
no_sub = 0
if (datatype(outline.(0)) ne 'STC') then begin
    n_outlines = 1
    no_sub = 1
endif

ist = 0
for i=0,n_outlines-1 do begin
    if (no_sub eq 1) then c1 = outline else c1 = outline.(i)
    ;--------- parse structure fields ----------
    nc1 = n_tags(c1)
    names = tag_names(c1)
;     cont_coord = strmid(strupcase(c1.(0)),0,1)
;     cont_ra = c1.(1)
;     cont_dec = c1.(2)
    iw = index_word(names,'COO',err=errword)
    if (errword eq 0) then cont_coord = strupcase(strmid(c1.(iw),0,1))

    iw = index_word(names,'RA',err=errword)
    if (errword eq 0) then cont_ra = c1.(iw)

    iw = index_word(names,'DEC',err=errword)
    if (errword eq 0) then cont_dec = c1.(iw)

    if (undefined(cont_coord) or undefined(cont_ra) or undefined(cont_dec)) then begin
        print,'Unvalid tag names in structure provided for outline :',names
        print,'Expected:       COORD RA DEC [LINESTYLE ]'
        message,/info,'No outline is plotted'
        return
    endif
    cont_type = 0
    iw = index_word(names,'LINE',err=errword)
    if (errword eq 0) then cont_type = c1.(3)
    
    ; angles -> vector
    if (n_elements(cont_ra) ne n_elements(cont_dec)) then begin
        message,'Inconsistent coordinates in outline structure'
    endif
    ang2vec, /astro, cont_dec, cont_ra, cont_vec

    ; interpolate intermediate points between vertices (along geodesics)
    ; this will avoid big step between vertices to be interpreted as boundary effect
    ; on the uv plan
    np_in = (size(cont_vec))[1]
    rf = 20
    cont_x = interpol(cont_vec[*,0], rf*np_in) ; linear interpolation in X
    cont_y = interpol(cont_vec[*,1], rf*np_in) ; .. Y
    cont_z = interpol(cont_vec[*,2], rf*np_in) ; .. Z
    cont_norm = sqrt( cont_x*cont_x + cont_y*cont_y + cont_z*cont_z ) ; norm of vectors 
    cont_vec = [[cont_x/cont_norm],[cont_y/cont_norm],[cont_z/cont_norm]]
    
    ; go from contour coordinates to plot coordinates
    cont_vec = rotate_coord(cont_vec, inco= cont_coord, outco=coord_out)
    cont_vec = rotate_coord(cont_vec, euler_matrix = transpose(eul_mat))

    ; vector -> (U,V)
    if (projtype eq 1) then begin ; mollweide
        vec2moll, cont_vec, u_cont, v_cont, flip=flip
        nkeep = n_elements(u_cont)
    endif
    if (projtype eq 2) then begin ; gnomic
        keep = where(cont_vec[*,0] gt 0., nkeep)
        if (nkeep gt 0) then begin
            u_cont =  flipconv*cont_vec[keep,1]/cont_vec[keep,0] ; U=Y/X
            v_cont =           cont_vec[keep,2]/cont_vec[keep,0] ; U=Z/X
        endif
    endif
    if (projtype eq 3) then begin ; cartesian
        u_cont = flipconv * atan(cont_vec[*,1],cont_vec[*,0])
        v_cont =            asin(cont_vec[*,2])
        nkeep = n_elements(u_cont)
    endif
    if (projtype eq 4) then begin ; orthographic
        if (keyword_set(half_sky)) then begin
            u_cont = flipconv *  cont_vec[*,1]
            v_cont =             cont_vec[*,2]
        endif else begin
            s_cont = (cont_vec[*,0] gt 0)*2 - 1 ; +1 for x>0, -1 for x<0
;             u_cont = flipconv * (cont_vec[*,1] + s_cont)
            u_cont = flipconv * (cont_vec[*,1] + 1)*s_cont
            v_cont =             cont_vec[*,2]
        endelse
        nkeep = n_elements(u_cont)
    endif
                                ; store (U,V) in structure
    if (nkeep gt 0) then begin
        c1uv = create_struct('U',u_cont,'V',v_cont,'T',cont_type)
        tag = 's'+strtrim(string(i,form='(i2)'),2)
        if (ist eq 0) then begin
            outline_uv = create_struct(tag,c1uv) 
        endif else begin
            outline_uv = create_struct(outline_uv,tag,c1uv)
        endelse
        ist = ist + 1
    endif
endfor

if (keyword_set(show)) then begin
    n_outlines= n_tags(outline_uv)
    for i=0,n_outlines-1 do begin
        c1 = outline_uv.(i)
        oplot_sphere, c1.U, c1.V, line_type=c1.T, _extra = oplot_kw
    endfor
endif

return
end

