
/*
 * Copyright (c) 2003, 2006 Matteo Frigo
 * Copyright (c) 2003, 2006 Massachusetts Institute of Technology
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* $Id: windows_exception_handler.c,v 1.2 2006/04/05 01:37:34 korpela Exp $ */

/* A handler that recieves windows exceptions and calls signal handlers */
#ifdef _WIN32
#include "windows_exception_handler.h"

/* 
 * WIN32 doesn't passs some exceptions to the C signal handlers by default 
 * so we trick it into doing so 
 */

#ifndef SIGRTMAX
#if defined(_SIGRTMAX)
#define SIGRTMAX _SIGRTMAX
#elif defined(NSIG)
#define SIGRTMAX (NSIG-1)
#else
#define SIGRTMAX 32
#endif

/* 
 * A lot of signals don't exist on windows compilers.  We'll check the ones
 * that might not exist;
 */
static int no_reset[SIGRTMAX+1];

void setup_no_reset() {
#ifdef SIGILL
  no_reset[SIGILL]=1;
#endif
#ifdef SIGTRAP
  no_reset[SIGTRAP]=1;
#endif
#ifdef SIGPRIV
  no_reset[SIGPRIV]=1;
#endif
};

static int no_ignore[SIGRTMAX+1];

void setup_no_ignore() {
#ifdef SIGKILL
  no_ignore[SIGKILL]=1;
#endif
#ifdef SIGSTOP
  no_ignore[SIGSTOP]=1;
#endif
#ifdef SIGSEGV
  no_ignore[SIGSEGV]=1;
#endif
};

static int handler_installed;

static LPTOP_LEVEL_EXCEPTION_FILTER OldExceptionFilter;

LONG pass_to_signal_handler(int signum) {
  void (*handler)(int);
  handler=signal(signum,SIG_DFL);
  if (!(no_ignore[signum]) && (handler==SIG_IGN)) {
     signal(signum,SIG_IGN);
     return EXCEPTION_CONTINUE_EXECUTION;
  }
  if ((handler==SIG_DFL) || (handler==SIG_IGN)) {
    signal(signum,handler);
    return EXCEPTION_CONTINUE_SEARCH;
  }
  handler(signum);
  if (no_reset[signum]) signal(signum,handler);
  return EXCEPTION_CONTINUE_EXECUTION;
}

FFTW_EXTERN LONG CALLBACK ExceptionFilter(LPEXCEPTION_POINTERS except) {
  switch (except->ExceptionRecord->ExceptionCode) {
    case EXCEPTION_BREAKPOINT:
    case EXCEPTION_SINGLE_STEP:
#ifdef SIGTRAP
      				       return pass_to_signal_handler(SIGTRAP);
#else
				       break;
#endif
    case EXCEPTION_FLT_DENORMAL_OPERAND:
    case EXCEPTION_FLT_DIVIDE_BY_ZERO:
    case EXCEPTION_FLT_INEXACT_RESULT:
    case EXCEPTION_FLT_INVALID_OPERATION:
    case EXCEPTION_FLT_OVERFLOW:
    case EXCEPTION_FLT_UNDERFLOW:      
#ifdef SIGFPE
				       {
  				         LONG rv=pass_to_signal_handler(SIGFPE);
					 /* MS claims ignoring an FP signal
					  * results in an unknown FP state.
					  * Does an _fpreset() help?
					  */
					 if (rv != EXCEPTION_CONTINUE_SEARCH) 
					   _fpreset();
					 return rv;
				       }
#endif
    case EXCEPTION_INT_DIVIDE_BY_ZERO:
    case EXCEPTION_INT_OVERFLOW:
#ifdef SIGFPE
  				       return pass_to_signal_handler(SIGFPE);
#endif
				       break;
    case EXCEPTION_PRIV_INSTRUCTION:
#ifdef SIGPRIV
				       return pass_to_signal_handler(SIGPRIV);
#endif
    case EXCEPTION_ILLEGAL_INSTRUCTION:
#ifdef SIGILL
				       return pass_to_signal_handler(SIGILL);
#endif
    case EXCEPTION_DATATYPE_MISALIGNMENT:
#ifdef SIGBUS
                                       return pass_to_signal_handler(SIGBUS);
#endif
    case EXCEPTION_STACK_OVERFLOW:
    case EXCEPTION_ACCESS_VIOLATION:
    case EXCEPTION_ARRAY_BOUNDS_EXCEEDED:
    case EXCEPTION_IN_PAGE_ERROR:
#ifdef SIGSEGV
      				       return pass_to_signal_handler(SIGSEGV);
#endif
  }
  return OldExceptionFilter(except);
}

FFTW_EXTERN void install_windows_exception_handler() {
    LPTOP_LEVEL_EXCEPTION_FILTER filter;
    if (!handler_installed) {
      setup_no_reset();
      setup_no_ignore();
      filter=SetUnhandledExceptionFilter(ExceptionFilter);
      if (filter != ExceptionFilter) {
        OldExceptionFilter=filter;
      }
      handler_installed=1;
    }
}

FFTW_EXTERN void uninstall_windows_exception_handler() {
  if (handler_installed) {
    SetUnhandledExceptionFilter(OldExceptionFilter);
    handler_installed=0;
  }
}

#endif

#endif
