// Filters for removing common waveforms from a channel.
template <typename T>
class waveform_filter {
  public:
    typedef void (* waveform_filter_funct)(std::vector<T> &channel, const
      std::vector<T> &waveform);
    waveform_filter(waveform_filter_funct F=NULL) : f(F) {};
    waveform_filter(const waveform_filter &F) : f(F.f) {};
    void operator ()(std::vector<T> &channel, const std::vector<T> &waveform) {
      f(channel,waveform);
    };
    operator bool() { return f!=0; };
  private:
    waveform_filter_funct f;
};


template <typename T>
void direct_subtract(std::vector<T> &channel, const std::vector<T> &waveform) {
  size_t i;
  for (i=0;i<channel.size();i++) {
    channel[i]-=waveform[i];
  }
}

template <typename T>
void scaled_subtract(std::vector<T> &channel, const std::vector<T> &waveform) {
  std::vector<complex<double> > a(least_squares(waveform,channel));
  size_t i;
  for (i=0;i<channel.size();i++) {
     channel[i]-=(a[0]+complex<double>(a[1].real()*waveform[i].real(),a[1].imag()*waveform[i].imag()));
  }
}

template <typename T>
void smoothed_subtract(std::vector<T> &channel, const std::vector<T> &waveform) {
  std::vector<T> schannel(smooth(channel,50)),swaveform(smooth(waveform,50));
  std::vector<complex<double> > a(least_squares(swaveform,schannel));
  size_t i;
  for (i=0;i<channel.size();i++) {
     channel[i]-=(a[0]+T(a[1].real()*waveform[i].real(),a[1].imag()*waveform[i].imag()));
  }
}

template <typename T>
void smoothed_subtract(std::vector<complex<T> > &channel, const
    std::vector<complex<T> > &waveform) {
  std::vector<complex<T> > schannel(smooth(channel,50)),swaveform(smooth(waveform,50));
  std::vector<complex<double> > a(least_squares(swaveform,schannel));
  size_t i;
  for (i=0;i<channel.size();i++) {
    channel[i]-=complex<T>(
	a[0].real()+a[1].real()*waveform[i].real(),
	a[0].imag()+a[1].imag()*waveform[i].imag()
    );
  }
}


template <typename T>
void equalize_zeros_and_ones(std::vector<T> &channel, const std::vector<T>
    &waveform)
{
  double max_wav;
  int i;
  for (i=0;i<waveform.size();i++) {
    max_wav=std::max(max_wav,std::abs((double)waveform[i].real()+waveform[i].imag()));
  }
  for (i=0;i<channel.size();i++) {
    if ((waveform[i].real()+waveform[i].imag())*channel[i].real() > 0) {
      // probability that this bit will change sign
      double pchange=std::abs(waveform[i].real()+waveform[i].imag())/(2*max_wav);
      double rnd=static_cast<double>(rand() & 0x3fff)/0x3fff;
      if (rnd < pchange) {
	channel[i]=T(-1*channel[i].real(),channel[i].imag());
      }
    }
    if ((waveform[i].real()+waveform[i].imag())*channel[i].imag() > 0) {
      // probability that this bit will change sign
      double pchange=std::abs(waveform[i].real()+waveform[i].imag())/(2*max_wav);
      double rnd=static_cast<double>(rand() & 0x3fff)/0x3fff;
      if (rnd < pchange) {
	channel[i]=T(channel[i].real(),-1*channel[i].imag());
      }
    }
  }
}

template <typename T>
void equalize_zeros_and_ones_c(std::vector<T> &channel, const std::vector<T>
    &waveform)
{
  complex<double> max_wav;
  int i;
  for (i=0;i<waveform.size();i++) {
    max_wav=complex<double>(
        std::max(max_wav.real(),std::abs((double)waveform[i].real())),
        std::max(max_wav.imag(),std::abs((double)waveform[i].imag()))
      );
  }
  for (i=0;i<channel.size();i++) {
    if (waveform[i].real()*channel[i].real() > 0) {
      // probability that this bit will change sign
      double pchange=std::abs(waveform[i].real())/(2*max_wav.real());
      double rnd=static_cast<double>(rand() & 0x3fff)/0x3fff;
      if (rnd < pchange) {
	channel[i]=T(-1*channel[i].real(),channel[i].imag());
      }
    }
    if (waveform[i].imag()*channel[i].imag() > 0) {
      // probability that this bit will change sign
      double pchange=std::abs(waveform[i].imag())/(2*max_wav.imag());
      double rnd=static_cast<double>(rand() & 0x3fff)/0x3fff;
      if (rnd < pchange) {
	channel[i]=T(channel[i].real(),-1*channel[i].imag());
      }
    }
  }
}
