#include <vector>
#include <map>
#include "seti_dr2filter.h"

// which HW data stream becomes which SW data stream is entirely
// determined by the device numbering of the EDT cards by the 
// data recorder host OS.
typedef int channel_t;	// identifies a 2 bit SW data stream
typedef int input_t;	// identifies a 2 bit HW data stream

typedef std::pair<int,int> bmpol_t;

extern std::map<telescope_id,channel_t> receiverid_to_channel;
extern std::map<channel_t,telescope_id> channel_to_receiverid;
extern std::map<bmpol_t,channel_t> beam_pol_to_channel;
extern std::map<input_t,channel_t> input_to_channel;

const size_t    DataSize    = 1024*1024;
const size_t    HeaderSize  = 4096;
const size_t    BufSize     = HeaderSize + DataSize;
const size_t    BufWords    = BufSize/2;
const size_t    HeaderWords = HeaderSize/2;
const size_t    DataWords   = DataSize/2;

typedef struct {
  double        Az;                 // should these be integers  ??
  double        Za;                 // gregorian position
  unsigned long Time;               // millisec past local midnight.  A time stamp for Az/Za
  time_t        SysTime;            // this and other SysTime's are local system time 
} scram_agc_t;                      //   when these scam data were acquired by the recorder


typedef struct {
  int           AlfaFirstBias;      // bitmap     
  int           AlfaSecondBias;     // bitmap
  float         AlfaMotorPosition;  // float is correct
  time_t        SysTime;
} scram_alfashm_t;


typedef struct {
  double        synI_freqHz_0;        
  int           synI_ampDB_0;       // units .01 db (from AO source code)
  double        rfFreq;             // rf frequency hz. topo centric (from AO source code)
  double        if1FrqMhz;          // if1 freq Mhz (from AO source code)
  bool          alfaFb;             // 0 wb, 1-filter in (100 Mhz) (from AO source code)
  time_t        SysTime;
} scram_if1_t;


typedef struct {
  bool          useAlfa;            // 1 : alfa is on        
  time_t        SysTime;
} scram_if2_t;

typedef struct {
  int           TurretEncoder;
  double        TurretDegrees;              // Encoder * (1./((4096.*210./5.0)/(360.)))           
  double        TurretDegreesAlfa;          // needs to be hard coded
  double        TurretDegreesTolerance;     // needs to be hard coded
  time_t        SysTime;
} scram_tt_t;

typedef struct {                    // NOTE : these coords are *requested*, not actual (ie, do not use)
 double         Ra;                 // ra J2000 back converted from az,za (from AO source code)   
 double         Dec;                // dec J2000 back converted from az,za (from AO source code) 
 double         MJD;
 time_t         SysTime;
} scram_pnt_t;


class dataheader_t {
public:
  int               header_size;
  int               data_size;
  char              name[36];
  int               dsi;            // data stream index
  int               channel;
  unsigned long     frameseq;       // dr2 buffer index in sequence recorded ?
  unsigned long     dataseq;        // dr2 buffer index in sequence acquired ?
  unsigned long     idlecount;      // dr2 buffer count of idle time   
  int               missed;         // > 0 : there is a break from previous buffer to this one ?
  seti_time         data_time;      // time stamp : the final data sample in this block
  seti_time         coord_time;     // time stamp : ra/dec (ie, the time of Az/Za acquisition)
  seti_time         synth_time;     // time stamp : the setting of synth_freq
  double            ra;             // ra for the beam from which the data in this block came
  double            dec;            // dec for the beam from which the data in this block came
  long              synth_freq;     // Hz        
  long              sky_freq;       // aka center frequency, Hz
  int               vgc[8];   
  double            samplerate;
  scram_agc_t       scram_agc;
  scram_alfashm_t   scram_alfashm;
  scram_if1_t       scram_if1;
  scram_if2_t       scram_if2;
  scram_tt_t        scram_tt;
  scram_pnt_t       scram_pnt;

  dataheader_t();
  void populate_from_data (char * header);
  void populate_derived_values ();
  void print();

private:
  int  parse_line(std::vector<char *> &words); 
};

class dr2_compact_block_t {
public:
    dataheader_t header;
    std::vector<complex<signed char> > data;
};

class dr2_block_t {
public:
    dataheader_t header;
    std::vector<complex<float> > data;

//    dr2_block_t();
};

static waveform_filter<complex<float> > NULLF;
static waveform_filter<complex<signed char> > NULLC;


void read_header (char * header, dataheader_t &dataheader);
int  parse_header_line(std::vector<char *> &words, dataheader_t &dataheader);
int seti_GetDr2CommonWaveform(int fd, int dsi, ssize_t numsamples,
    std::vector<dr2_block_t> &dr2_block);
int seti_StructureDr2Data(int fd, telescope_id receiver, ssize_t numblocks,
std::vector<dr2_block_t> &dr2_block, waveform_filter<complex<float> > filter=NULLF);
int seti_StructureDr2Data(int fd, int beam, int pol, ssize_t numblocks,
std::vector<dr2_block_t> &dr2_block, waveform_filter<complex<float> > filter=NULLF);
int seti_StructureDr2Data(int fd, channel_t channel, ssize_t numsamples,
    std::vector<dr2_compact_block_t> &dr2_block, waveform_filter<complex<signed
    char> > filter=NULLC);
int seti_StructureDr2Data(int fd, telescope_id receiver, ssize_t numblocks,
std::vector<dr2_compact_block_t> &dr2_block, waveform_filter<complex<signed
char> > filter=NULLC);
int seti_StructureDr2Data(int fd, int beam, int pol, ssize_t numblocks,
std::vector<dr2_compact_block_t> &dr2_block, waveform_filter<complex<signed
char> > filter=NULLC);


template <typename T>
int seti_GetDr2Data(
                    int fd, channel_t channel, ssize_t numsamples, 
                    std::vector<complex<T> > &c_samples, dataheader_t * ret_dataheader,
		    std::vector<complex<T> > &common_waveform,
		    unsigned short active_channel_mask=0xffff
		    ) {
//  fd              device that contains tape data. 
//  channel         the channel (0-15) from which you wish to extract.
//  numsamples      the number of samples you wish to extract.
//  c_samples       a std vector of type complex<T> onto which numsamples will be pushed.
//                  The samples will have values between -num_channels and
//                  +num_channels where num_channels is the number of active
//                  channels on that DSI. (6 for DSI0, 8 for DSI1)
//  ret_dataheader  pointer to a header object to receive the last header read. 
//                  if NULL, then no header is "returned".
//  common_waveform 
//                  common waveform for later removal this waveform is the
//                  average of the values in each active channel.
//  active_channel_mask
//                  selects channels for common waveform removal.  Set to
//                  zero to disable common waveform removal.  Default is that
//                  all channels will be used.
//  return val      the number of samples returned

    static intel<unsigned short>   *buf=(intel<unsigned short> *)calloc(BufSize,1); // BufSize is in bytes

    // An array to temporarily hold the data for block push onto the vector.
    static complex<T> *dbuf=(complex<T> *)calloc(DataWords,sizeof(complex<T>));  
    static complex<T> *abuf=(complex<T> *)calloc(DataWords,sizeof(complex<T>));  
    double aavg_i=0,aavg_q=0,cavg_i=0,cavg_q=0;  // average i and q values
    double asum_i2=0, asum_q2=0, csum_i2=0, csum_q2=0;  // sum of i^2 and q^2
    double xsum_ii=0, xsum_qq=0;
    static size_t           buf_i=0;               // will index an array of 
                                                   // 16 bit words (short ints)
    size_t           data_i;              // will index into dbuf
    static const unsigned short channel_mask[2]={
                              0xf3fc, // active channels on DSI 0
                              0xffff  // active channels on DSI 1
			    };
    const unsigned short qbits=0xaaaa,ibits=0x5555;

    static int nbits[256];
    size_t                  readbytes;
    unsigned short          dataword;
    long                    samplecount = 0;
    dataheader_t            dataheader;

    int                     dsi = channel > 7 ? 0 : 1;

    if (lseek(fd, 0, SEEK_CUR) == 0) buf_i = 0;     // re-init if caller has rewound/reopened

    // set up bit counting
    if (nbits[1] == 0) {
      unsigned char c=1,d;
      do {
	d=c;
	while (d) {
	  if (d & 1) nbits[c]++;
	  d>>=1;
	}
	c++;
      } while (c);
    }
	
    // Keep reading BufSize bytes worth of data until we have
    // numsamples 2 bit samples.  Note that numsamples might be,
    // but does not have to be, satisfied with one read.
    while (samplecount < numsamples) {

        // read the next buffer if need be
        if (buf_i == 0) {
            if ((readbytes = read(fd, buf, BufSize)) != BufSize) break;   
            dataheader.populate_from_data((char *)buf);
            if(dataheader.dsi != dsi) continue;     // correct data stream for desired channel?
            if(ret_dataheader) {
                dataheader.channel = channel;
                dataheader.populate_derived_values();
                *ret_dataheader = dataheader;       // a "shallow" assignment should be OK here
            }
            buf_i = HeaderWords;                    // first word past the header
        }
	    //  Mask the active channels
	    unsigned qmask=qbits & channel_mask[dsi] & active_channel_mask;
	    unsigned imask=ibits & channel_mask[dsi] & active_channel_mask;
	    // determine number of active channels
	    int nchan=nbits[qmask & 0xff]+nbits[(qmask>>8) & 0xff];

        for (data_i=0;
             buf_i < BufWords && samplecount < numsamples;
             buf_i += 1, data_i +=1, samplecount += 1) {

	        dataword = buf[buf_i];

	        // generate the common waveform
	        unsigned short ti;
            unsigned short tq;

	        if (nchan) {
	            ti=dataword & imask;
	            tq=dataword & qmask;
                abuf[data_i]=complex<T>(
		            (nbits[ti & 0xff]+nbits[(ti >> 8) & 0xff])*2-nchan,
	                (nbits[tq & 0xff]+nbits[(tq >> 8) & 0xff])*2-nchan
		        );
            }

            // isolate the desired channel
            dataword >>= (channel & 7)*2;

	        // add the samples to our buffer.
	        dbuf[data_i]=complex<T>(
		        ((dataword & 1)*2)-1,
	            (dataword & 2)-1
		    );
	        if (nchan) {
	            dbuf[data_i]*=nchan;
	        } 
        }

	    c_samples.insert(c_samples.end(),dbuf,dbuf+numsamples);
	    if (nchan) {
	        common_waveform.insert(common_waveform.end(),abuf,abuf+numsamples);
	    }

        if (buf_i >= BufWords) {
            buf_i = 0;    // we'll need to read the next buffer next time 'round
        }
    }

    return(samplecount);
}

