/* 
 * Source Code Name  : SIREN.h
 * Programmer        : Donnelly/Cobb et al
 *                   : Project SERENDIP, University of California, Berkeley CA
 * Compiler          : GNU gcc
 * Purpose           : Header file for the SERENDIP SIREN 
 *                     data analysis software system.  Contains constants
 *                     and structures used throughout the modules that make
 *                     up SIREN.
 *
 * History:
 * -deank 6/97  - added dScore field to beamfit record 
 * -jeffc 2/99  - moved telescope/receiver ID stuff to s4tel.h
 * -jeffc 11/00 - added	SETI_AT_HOME_BOARD CONTIGBIN LOW_CONTIGBIN HIGH_CONTIGBIN 
 * -jeffc 4/2/01  changed MAX_HITS from 200 to 1000, added SECDIFF macro
 */      

#ifndef SIREN // no need to look at it if already looked

#include <stdlib.h>

// Record Type definitions

#define STDEVENT_REC 'A'
#define S3_STDEVENT_REC 'B'
#define CANDIDATE_REC 'C'
#define BEAMFIT_REC 'D'
#define SAH_GAUSSIAN_REC 'G'
#define COMMENT_REC '#'

#define MAX_LINE_LENGTH 256
#define DBASE_FILENAME_LEN 9      // with room for null!

#define CONFIG_DIR ~seti/s4/siren/cfg
#define TELCONFIGFILE "/mydisks/d/seti/s4/siren/cfg/telescope.config"

#ifndef PI
#define PI 3.141592653589793
#endif

#define AOLATITUDE_RADS 0.320334327   /* observatory - arecibo */
#define AOLONGITUDE_DEGS -66.75522222
#define SIN_AO_LAT 0.314883898060 /* sine(AOLAT...) */
#define COS_AO_LAT 0.949130196940 /* coine(AOLAT...) */

// The SKYPLOT #defines have to do with our "pixelized sky" database.
// They are specific to the sky as seen form AO.
// The skyplot is a 2D array that consists of SKYPLOT_HEIGHT elements
// of DEC andSKYPLOT_WIDTH elements of RA.  This corresponds to a DEC
// range of -5 to +45 degrees and an RA range of 0 to 24 hours
// (0 to 360 degrees).  The CONVERT macros convert RA from hours
// to element index and DEC from degrees to element index.  The R_CONVERT
// macros do the reverse, sort of.  They convert from element index to
// the RA and DEC of the *center* if the element.
// The angular size of each element is .2 *degrees* in RA and .25
// degrees in DEC.  The terms "element" and "pixel" are interchangable.
#define SKYPLOT_HEIGHT 200
#define SKYPLOT_WIDTH  1800
#define SKYPLOT_CONVERT_RA(a)    (long)   (75.0 * a)
#define SKYPLOT_CONVERT_DEC(a)   (long)   (4.0 * (a + 5.0))
#define SKYPLOT_R_CONVERT_RA(a)  (double) (((a / 5.0) + 0.1) / 15)
#define SKYPLOT_R_CONVERT_DEC(a) (double) (((a / 4.0) - 5.0) + 0.125)

#define OBS_SEPARATION  0.042   // defines separate observations as temporal
                               //  separation in julian days (0.042 = 1 hour) 

#define MAX_EVENTS 500000 /* maximum number of events in an input file */
#define MAX_HITS 5000      /* maximum number of hits per spectrum */
#define NUM_FFT_BINS 4194304.0
#define NUM_FQ_BINS 12.1e6
#define BOARD_WIDTH 2500000.0  // frequency range of 1 FFT board in Hz
#define INTEGRATION_TIME 1.68  
#define DELTA_TM 1.68   /* plus/minus time uncertainty=integration time */
#define DELTA_FQ 0.60   /* plus/minus frequency uncertainty=bin width */
#define NUM_BINS   4194304L
#define LOW_BIN    -2097152L
#define HIGH_BIN   2097151L
#define NUMBOARDS  40
#define SETI_AT_HOME_BOARD 19

// The bin number on each board range from LOW_BIN to HIGH_BIN.  At times, it
// is convenient to look at all bins, across all boards, as a contiguous
// range of bin numbers.  Ie, bin numbers in the full FFT space.  CONTIGBIN()
// will translate any bin,board pair in a bin number in the full FFT space.
// Since the concept of negative bin numbers makes no sense here, we start
// this range at 0.  LOW_CONTIGBIN and HIGH_CONTIGBIN define the range.
// HIGH_CONTIGBIN is huge, so we need to use unsigned long.
#define CONTIGBIN(bin, board) ((long)(bin + (board * NUM_FFT_BINS)) + (long)labs(LOW_BIN))
#define LOW_CONTIGBIN  0UL    
#define HIGH_CONTIGBIN (CONTIGBIN(HIGH_BIN, (NUMBOARDS-1)))



// Candidate related defines
#define MAX_ALGO 6
#define MAX_CANDIDATES 10000
#define MAX_EVENTS_PER_CANDIDATE 200

#ifndef BOOLEAN 
#define BOOLEAN unsigned char
#define TRUE 1
#define FALSE 0
#endif

// Useful Macros
#define nlong(x) ((x) - (long)(x) < .5 ? (long)(x) : (long)(x) + 1)
#define nint(x) ((x) - (int)(x) < .5 ? (int)(x) : (int)(x) + 1)
#define SECS_TO_DAYS        ((double)1 / (double)86400)
#define SECDIFF(Jd1, Jd2) ((double)(fabs(Jd1 - Jd2) * 86400))

//________________________________________________ 
// SERENDIP IV Standard event structure
//________________________________________________ 
typedef struct {
  unsigned char   ucFlag;            // User scratch space 
  char            stUTime[9];        // UT Time string
  double          dbUTime;           // Decimal time
  double          dbJulDate;         // Julian Date
  double          dbRa;              // Right ascension
  double          dbDec;             // Declination
  double          dbFirstLo;         // First local oscillator
  double          dbSecondLo;        // Second local oscillator
  unsigned long   ulFftBdFreq;       // FFT board frequency.
  long            lBinNumber;        // Bin number
  double          dbFrequency;       // Sky frequency
  double          dbBirdieSynth;     // Synth setting for injected birdie
  unsigned char   ucBid;             // Board ID
  unsigned short  usPower;           // Detcion power 
  unsigned short  usSigmaThreshold;  // Threshold in sigma
  unsigned short  usActualThreshold; // Actual threwshold
  char            cPolarization;     // Polorization
  unsigned short  usAdc;             // value of ADC 
  unsigned short  usSpctResLog;      // Log(base 2) spectral resolution 
  char            stDate[7];         // Date string 
  unsigned char   ucTid;             // Telescope ID 
  } StdEventRec;


//_____________________________________________________________________ 
// SERENDIP III Standard event structure
//   Used to read in Record Type 'B': standard s3 dbas records with 
//   record type inserted at beginning of line to enable s4 programs 
//   to distinguish them from standard s4 records.
//_____________________________________________________________________ 
typedef struct {
  char            cFlag;             // Used to flag good events
  double          dbFreq;            // Sky frequency
  long            lBinNo;            // Bin number
  long            lSynth;            // Synthesizer frequency
  double          dbRa;              // Right ascension
  double          dbDec;             // Declination
  double          dbLocos;           // Local oscillator
  unsigned short  usPwr;             // Detection power 
  unsigned short  usActualThreshold; // Actual threshold
  unsigned short  usAdc;             // Value of ADC 
  char            stDate[7];         // Date string 
  char            stStrTime[7];      // Time string
  double          dbDecTime;         // Decimal time
  unsigned short  usSigmaThreshold;  // Threshold in sigma
  unsigned short  usSpectralRes;     // Spectral resolution 
  double          dbJulDate;         // Julian date
  } s3_StdEventRec;


//_____________________________________________________________________
// SERENDIP minimal event structure
//   Used internally to store a minimal subset of event data.
//_____________________________________________________________________
typedef struct {
  char            cFlag;             // Used to flag good events
  double          dbFrequency;       // Sky frequency
  double          dbRa;              // Right ascension
  double          dbDec;             // Declination
  double          dbPower;           // Signal power in units of sigma
  double          dbJulDate;         // Julian date
  } MinEventRec;

//_____________________________________________________________________ 
// SERENDIP candidate structure 
// references to 'group' indicate that the candidate was generated
// by a group oriented algorithm (eg detections over multiple days). 
// references to 'single' indicate that the candidate was generated
// by a 'single event' oriented algorithm (eg high power detections). 
//_____________________________________________________________________ 
typedef struct {
  unsigned short  usAlgo;            // algorithm ID
  double          dbRa;              // if group then mean RA
  double          dbDec;             // if group then mean DEC
  double          dbFreq;            // if group then mean freq
  long            lHiPower;          // high (if group) or only (if single) power
  double          dbSkySep;          // for group (if single then zero). Units=Degs
  long            lFreqSep;          // for group (if single then zero). Units=Hz
  long            lMaxSpectra1;      // max nbr of spectra in any one observation
                                     //  (if single then one)
  long            lMaxSpectra2;      // next to max nbr of spectra in any one 
                                     //  observation (if single then one)
  long            lNumDetect;        // number of separate detections
                                     //  (if single then one)
  long            lLineNum;          // line number of event(s) in algorithm file
  double          dbBeamFit; 	     // for group. Score assigned by beamfitscore
  char            stStarName[11];    // for star catalog entry.
  double          dbStarDistance;    // for star catalog entry.
  char            stSpectralClass[6];  
                                     // for star catalog entry.
  long            lNumObs;           // Total number of observations 
                                     //  at these coordinates
  double          dbObsScore;        // Given lNumObs, the probability that we
                                     // will see lNumDetect detections.  This
                                     // only makes sense for multidetect algorithms
  long            lFlag;             // for internal use.
  } CandidateRec;

//_____________________________________________________________________
// SERENDIP Beamfit result record.
//  Each of these records contains the results of beamfit analysis
//  on 1 group of SERENDIP event records.
//_____________________________________________________________________
typedef struct {
  char            cFlag;             // Misc flag
  char            stDate[7];         // Date string - from earliest in group
  char            stStrTime[7];      // Time string - from earliest in group
  long            lNumSpectra;       // Number of spectra in group
  long            lNumEvents;        // Number of events in group
  double          dbHiPower;         // Highest power event of the group
  double          dChiSqr;           // Chi Square         
  double          dProbability;      // Probability
  double          dScore;            // Score assigned by beamfitscore
  double          dGausSigma;        // Sigma of best fit gaussian
  double          dGausAmean;        // Amplitude of best fit gaussian
  double          dGausDmean;        // Mean distance of best fit gaussian
  double          dVersion;          // Version constant from beamfit program
  } BeamfitRec;

//________________________________________________
// SERENDIP IV Engineering Monitor Record
//________________________________________________
typedef struct {
  unsigned char   ucFlag;            // User scratch space
  char            stUTime[9];        // UT Time string
  double          dbUTime;           // Decimal time
  double          dbJulDate;         // Julian Date
  unsigned char   ucEngMonitor[10]; // 10 Engineering Monitors
  } EngMonRec;
 
//_____________________________________________________________________
// SERENDIP skyplot element record
//  Each records contains an ulDbaseKey and an lLink to another skyplot
//  element record.
//_____________________________________________________________________
typedef struct {
  unsigned long   ulDbaseKey;        // contains the pcdbase filename, if any
  long            lLink;             // -1 if this pixel is the end of the
                                     // list of ulDbaseKey's, or the next
                                     // element containing a ulDbaseKey for
                                     // the same pixel.
  } SkyplotElement;

//_____________________________________________________________________
// SETI@home gaussian record.
// Data types and names match those in the SETI@home file db.h.  
// Field ucFlag has been added.
//_____________________________________________________________________
typedef struct {
    unsigned char	ucFlag;            // User scratch space
    int 	 	id;
    int  		result_id;
    float 		peak_power;
    float 		mean_power;
    float 		ra;
    float 		dec;
    double 		time;
    double 		freq;
    float 		sigma;
    float 		chisqr;
    int 		fft_len;
    float 		chirp_rate;
    float 		score;
    float 		max_power;
    int 		redundancy;
    int 		rfi_checked;
    int 		rfi_found;
    int 		trust;
    double 		detection_freq;
    double 		barycentric_freq;
  } sah_Gaussian_t;


#endif  // end of #ifndef SIREN

#define SIREN
