//______________________________________________________________________
// Source Code Name  : s4io.cc
// Programmer        : Donnelly/Cobb et al
//                   : Project SERENDIP, University of California, Berkeley CA
// Compiler          : GNU gcc
// Date Opened       : 8/94
// This file includes the set of io functions necessary for
// handling serendip IV off line data.
//______________________________________________________________________
// modification history:
//   -brendan 12/94 - added io_PeekNextRecord() and io_SkipNextRecord()
//      functions; modified io_GetNextRecord to return next record of 
//      a particular type.
//   -brendan 12/94 - added io_WriteS3StdEvent() and io_s3GetNextEvent()
//      functions; modified io_GetNextRecord(), io_PeekNextRecord() 
//      functions to handle type 'B' records (s3 data).
//   -jeffc 3/95 - added io_WriteCandidate() and io_CountRecords()
//      functions.  Modified io_GetNextRecord() and io_PeekNextRecord()
//      functions to handle type 'C' records (candidate records). 
//   -brendan 03/95 - changed int variables and function types to long.
//   -scott 04/95 - added io_GetLastRecord() function.
//   -jeffc 01/96 - changed some format specifiers in InLineCandidate
//      and OutLineCandidate. 
//   -jeffc 04/96 - added 2 new fields to Candidate Record:
//      lNumObs and dbObsScore
//   -jeffc 9/96 - added logic for new record type - 'D' - beamfit record.
//      This included changes to io_GetNextRecord() and io_PeekNextRecord()
//      as well as a new routine, io_WriteBeamfitRec(). 
//   -deank 6/97 - added dScore field to beamfit record
//   -deank 9/97 - added io_GetFilePosition() and io_SetFilePosition()
//      functions
//   -jeffc 2/98 - changed score format code in OutLineCandidate. 
//______________________________________________________________________

#include <stdio.h>
#include <string.h>
#include "SIREN.h"
#include "s4io.h"
#include "s4freq.h"
#include "s4time.h"
#include "s4power.h"
#include <sys/stat.h>

#define InLineFormat "%*s %s %lf %lf %lf %lf %lu %ld %lf %hx %hu %hu %hu %c\
 %hu %hx %s %hx"
#define OutLineFormat "%c %s %8.5f %8.4f %13.7f %13.7f %6ld %10ld %13.7f %2x %5u %3d\
 %5u %c %2u %x %.6s %02x\n"

#define InLineS3Format "%*s %s %lf %lf %lf %ld %ld %hu %hu %hu %s %hx %hx"
#define OutLineS3Format "%c %s %7.4f %7.3f %12.7f %6ld %10ld %5u %5u %2u %.6s\
 %2hx %1hx\n"

#define InLineOldS3Format "%s %lf %lf %lf %ld %ld %hu %hu %hu %s %hx %hx"

#define InLineCandidate "%*c %hu %lf %lf %lf %lf %ld %ld %ld %ld %ld %ld %lf %s\
 %lf %s %ld %lf"
#define OutLineCandidate "%c %02hu %7.4lf %7.3lf %7.4lf %10.6lf %9ld %4ld\
 %3ld %3ld %3ld %5ld %7.2lf %10s %6.3lf %5s %3ld %8.3lf\n"

#define OutLineBeamfitRec "%c %5.2lf %s %s %3ld %3ld %6.2lf %8.4lf %06.4lf\
 %7.2lf %6.4lf  %6.2lf %6.4lf\n"
#define InLineBeamfitRec "%*s %lf %s %s %ld %ld %lf %lf %lf %lf %lf %lf %lf"

#define InLineSahGaussianRec "%*c %d %d %f %f %f %f %lf %lf %f %f %d %f %f %f %d %d %d %d %lf %lf"
#define OutLineSahGaussianRec "%c %d %d %10.8f %11.9f %10.7f %10.7f %18.9lf %18.7lf %10.8f %10.8f %d %17.15e %10.8f %10.8f %d %d %d %d %17.15e %17.15e"

#define OutLineComment "%c %s\n"

#define OutLineEngMonRec "%lf [%d] [%d] [%d] [%d] [%d] [%d] [%d] [%d] [%d] [%d]\n"

//____________________________________________________________________
// io_GetNextRecord()
// Takes a void pointer to a SERENDIP record structure and the record 
// type, as defined in the serendip data format (see man s4data),
// and populates the structure with the next record in the `ThisFile'
// of that type. Also advances the file pointer to that record in doing
// so. Function returns 1 on success, 0 otherwise.
//____________________________________________________________________ 

long io_GetNextRecord(FILE *ThisFile, void *ThisRecord,unsigned char RecordType)
  {
  unsigned char peekrtn;  // return value from io_PeekNextRecord

  // scan for next record of type passed
  while (((peekrtn = io_PeekNextRecord(ThisFile)) != RecordType)
        && peekrtn)
    if (! io_SkipNextRecord(ThisFile))
      return 0;

  // return 0 if EOF or error
  if (! peekrtn)
    return 0;

//-----------------------------------------------------------
  if (RecordType == STDEVENT_REC)
    {
    long day = 0; 
    long month = 0; 
    long year = 0;
    unsigned short tempBID;
    unsigned short tempTID;
    StdEventRec *ThisEvent;
  
    ThisEvent = (StdEventRec *) ThisRecord;

    if ((fscanf(ThisFile,InLineFormat, 
               ThisEvent->stUTime,
               &ThisEvent->dbRa,
               &ThisEvent->dbDec,
               &ThisEvent->dbFirstLo,
               &ThisEvent->dbSecondLo,
               &ThisEvent->ulFftBdFreq,
               &ThisEvent->lBinNumber,
               &ThisEvent->dbBirdieSynth,
               &tempBID,
               &ThisEvent->usPower,
               &ThisEvent->usSigmaThreshold,
               &ThisEvent->usActualThreshold,
               &ThisEvent->cPolarization,
               &ThisEvent->usAdc,
               &ThisEvent->usSpctResLog,
               &ThisEvent->stDate,
               &tempTID) == 17))
      {
      ThisEvent->ucTid = tempTID;
      ThisEvent->ucBid = tempBID;
      ThisEvent->ucFlag = 0x00;

      // calculate frequency
      ThisEvent->dbFrequency = fq_s4fcalc(ThisEvent->dbFirstLo,
                                          ThisEvent->dbSecondLo,
                                          ThisEvent->ulFftBdFreq,
                                          ThisEvent->lBinNumber,
                                          ThisEvent->ucTid);

      // calculate decimal time
      tm_HmsToDec(ThisEvent->stUTime, &ThisEvent->dbUTime);

      // convert date string to integers
      tm_DateToI(ThisEvent->stDate, &day, &month, &year);

      // add century to year
      year = tm_CentYear(year);

      // calculate julian date
      ThisEvent->dbJulDate = tm_ToJul(month, day, year, ThisEvent->dbUTime);

      return 1;
      }
    else
      return 0;
    }  // end if RecordType == STDEVENT_REC
//-----------------------------------------------------------
  else if (RecordType == COMMENT_REC)
    {
    char *thisComment;
  
    thisComment = (char *) ThisRecord;

    if (fgets(thisComment, MAX_LINE_LENGTH+1, ThisFile))
      {
      if (thisComment[strlen(thisComment)-1] == '\n')
        thisComment[strlen(thisComment)-1] = '\0';
      return 1;
      }
    else
      return 0;
    }  // end if RecordType == COMMENT_REC
//-----------------------------------------------------------
  else if (RecordType == S3_STDEVENT_REC)
    {
    long day_i = 0;
    long month_i = 0;
    long year_i = 0;
    s3_StdEventRec *ThisEvent;

    ThisEvent = (s3_StdEventRec *) ThisRecord;

    if (fscanf(ThisFile,InLineS3Format,
              ThisEvent->stStrTime,&ThisEvent->dbRa,&ThisEvent->dbDec,
              &ThisEvent->dbLocos,&ThisEvent->lSynth,&ThisEvent->lBinNo,
              &ThisEvent->usPwr,&ThisEvent->usActualThreshold,
              &ThisEvent->usAdc,ThisEvent->stDate,&ThisEvent->usSigmaThreshold,
              &ThisEvent->usSpectralRes) == 12)
      {
      // calculate frequency
      ThisEvent->dbFreq = fq_s3fcalc(ThisEvent->dbLocos, ThisEvent->lSynth,
                                     ThisEvent->lBinNo);

      // calculate decimal time
      tm_s3HmsToDec(ThisEvent->stStrTime, &ThisEvent->dbDecTime);

      // convert date string to integers
      tm_DateToI(ThisEvent->stDate, &day_i, &month_i, &year_i);

      // add century to year
      year_i = tm_CentYear(year_i);

      // calculate julian date
      ThisEvent->dbJulDate = tm_ToJul(month_i, day_i, year_i, 
                                      ThisEvent->dbDecTime);

      return 1;
      }
    else
      return 0;
    }  // end if RecordType == S3_STDEVENT_REC
//-----------------------------------------------------------
  else if (RecordType == CANDIDATE_REC)
    {
    CandidateRec * ThisCandidate;

    ThisCandidate = (CandidateRec *) ThisRecord;

    if(fscanf(ThisFile, InLineCandidate, &ThisCandidate->usAlgo,
                                         &ThisCandidate->dbRa,
                                         &ThisCandidate->dbDec,
                                         &ThisCandidate->dbSkySep,
                                         &ThisCandidate->dbFreq,
                                         &ThisCandidate->lFreqSep,
                                         &ThisCandidate->lHiPower,
                                         &ThisCandidate->lMaxSpectra1,
                                         &ThisCandidate->lMaxSpectra2,
                                         &ThisCandidate->lNumDetect,
                                         &ThisCandidate->lLineNum, 
                                         &ThisCandidate->dbBeamFit,
                                         &ThisCandidate->stStarName,
                                         &ThisCandidate->dbStarDistance,
                                         &ThisCandidate->stSpectralClass,
                                         &ThisCandidate->lNumObs,
                                         &ThisCandidate->dbObsScore)
                                         != 17)
      return 0;
    else
      { 
      ThisCandidate->dbFreq *= 1e6;     // MHz to Hz
      ThisCandidate->lFlag = 0;         // initialize
      return 1;
      }  
    } // end if RecordType == CANDIDATE_REC
//-----------------------------------------------------------
  else if (RecordType == BEAMFIT_REC)
    {
    BeamfitRec * ThisBeamfitRec;

    ThisBeamfitRec = (BeamfitRec *) ThisRecord;
 
    if(fscanf(ThisFile, InLineBeamfitRec, &ThisBeamfitRec->dVersion,
                                          &ThisBeamfitRec->stStrTime,
                                          &ThisBeamfitRec->stDate,
                                          &ThisBeamfitRec->lNumSpectra,
                                          &ThisBeamfitRec->lNumEvents,
                                          &ThisBeamfitRec->dbHiPower,
                                          &ThisBeamfitRec->dChiSqr,
                                          &ThisBeamfitRec->dProbability,
                                          &ThisBeamfitRec->dScore,
                                          &ThisBeamfitRec->dGausSigma,
                                          &ThisBeamfitRec->dGausAmean,
                                          &ThisBeamfitRec->dGausDmean)
                                          != 8)
      return 0;
    else
      return 1;
  } // end if RecordType == BEAMFIT_REC
//-----------------------------------------------------------
  else if (RecordType == SAH_GAUSSIAN_REC)
    {
    sah_Gaussian_t * ThisGaussianRec;

    ThisGaussianRec = (sah_Gaussian_t *) ThisRecord;

    if(fscanf(ThisFile, InLineSahGaussianRec,   &ThisGaussianRec->id,
                                                &ThisGaussianRec->result_id,
                                                &ThisGaussianRec->peak_power,
                                                &ThisGaussianRec->mean_power,
                                                &ThisGaussianRec->ra,
                                                &ThisGaussianRec->dec,
                                                &ThisGaussianRec->time,
                                                &ThisGaussianRec->freq,
                                                &ThisGaussianRec->sigma,
                                                &ThisGaussianRec->chisqr,
                                                &ThisGaussianRec->fft_len,
                                                &ThisGaussianRec->chirp_rate,
                                                &ThisGaussianRec->score,
                                                &ThisGaussianRec->max_power,
                                                &ThisGaussianRec->redundancy,
                                                &ThisGaussianRec->rfi_checked,
                                                &ThisGaussianRec->rfi_found,
                                                &ThisGaussianRec->trust,
                                                &ThisGaussianRec->detection_freq,
                                                &ThisGaussianRec->barycentric_freq)
                                                != 20)
      return 0;
    else
      return 1;
  } // end if RecordType == SAH_GAUSSIAN_REC
//-----------------------------------------------------------
  else
    return 0;                      // unknown record type passed
  }

// io_GetNextMinEvent
// will automatically grab the next event from a file, regardless
// of its type and return a populated MinEventRec. 
// PLEASE READ:
// ** It does not recognize record type CANDIDATE_REC
// ** It recognizes BEAMFIT_REC, BUT!!! It populates most of
//    the struct with zeros, since the beamfit rec doesn't
//    contain freq, ra, dec, or juldate (only contains power).
//
long io_GetNextMinEvent(FILE *ThisFile,MinEventRec *ThisMinEvent) {
  
  static unsigned char cRecordType = NULL;
  static void * InRec;
    
  if (!cRecordType) {
    cRecordType = io_PeekNextRecord(ThisFile);
    rewind(ThisFile);
    if (cRecordType == STDEVENT_REC) {
      if ( ( (StdEventRec *)InRec = (StdEventRec *)malloc(sizeof(StdEventRec)) ) == NULL)
        return -1; 
      }
    else if (cRecordType == S3_STDEVENT_REC) {
      if ( ( (s3_StdEventRec *)InRec = (s3_StdEventRec *)malloc(sizeof(s3_StdEventRec)) ) == NULL)
        return -1; 
      }
    else if (cRecordType == BEAMFIT_REC) {
      if ( ( (BeamfitRec *)InRec = (BeamfitRec *)malloc(sizeof(BeamfitRec)) ) == NULL)
        return -1; 
      }
    else if (cRecordType == SAH_GAUSSIAN_REC) {
      if ( ( (sah_Gaussian_t *)InRec = (sah_Gaussian_t *)malloc(sizeof(sah_Gaussian_t)) ) == NULL)
        return -1; 
      }
    else return -2;
    }

  ((StdEventRec *)InRec)->dbRa = 0.0 ;
  if (!io_GetNextRecord(ThisFile,InRec,cRecordType)) {
    free(InRec);
    return 0;
    }
  ThisMinEvent->cFlag = NULL;
  
  if (cRecordType == STDEVENT_REC) {
    ThisMinEvent->dbFrequency = ((StdEventRec *)InRec)->dbFrequency;
    ThisMinEvent->dbRa = ((StdEventRec *)InRec)->dbRa;
    ThisMinEvent->dbDec = ((StdEventRec *)InRec)->dbDec;
    ThisMinEvent->dbPower = pw_PwrCalc(InRec,cRecordType); 
    ThisMinEvent->dbJulDate = ((StdEventRec *)InRec)->dbJulDate;
    }
  else if (cRecordType == S3_STDEVENT_REC) {
    ThisMinEvent->dbFrequency = ((s3_StdEventRec *)InRec)->dbFreq;
    ThisMinEvent->dbRa = ((s3_StdEventRec *)InRec)->dbRa;
    ThisMinEvent->dbDec = ((s3_StdEventRec *)InRec)->dbDec;
    ThisMinEvent->dbPower = pw_PwrCalc(InRec,cRecordType);
    ThisMinEvent->dbJulDate = ((s3_StdEventRec *)InRec)->dbJulDate;
    }
  else if (cRecordType == BEAMFIT_REC) {
    ThisMinEvent->dbFrequency = 0.0; // no freq  ?!
    ThisMinEvent->dbRa = 0.0;           // no ra    ?!
    ThisMinEvent->dbDec = 0.0;         // no dec   ?!
    ThisMinEvent->dbPower = ((BeamfitRec *)InRec)->dbHiPower;   // db == db ?! 
    ThisMinEvent->dbJulDate = 0.0; // no date  ?!
    }
  else if (cRecordType == SAH_GAUSSIAN_REC) {
    ThisMinEvent->dbFrequency = ((sah_Gaussian_t *)InRec)->freq; 
    ThisMinEvent->dbRa = ((sah_Gaussian_t *)InRec)->ra;       
    ThisMinEvent->dbDec = ((sah_Gaussian_t *)InRec)->dec;     
    ThisMinEvent->dbPower = ((sah_Gaussian_t *)InRec)->max_power;   // db == db ?! 
    ThisMinEvent->dbJulDate = ((sah_Gaussian_t *)InRec)->time; 
    }
 
  return 1;

  }


//______________________________________________________________
// io_WriteStdEvent()
// Takes a pointer to a standard event record (StdEventRec)
// and writes the data in the record to `ThisFile`
//______________________________________________________________

void io_WriteStdEvent(FILE *ThisFile,StdEventRec *ThisEvent)
  {
  fprintf(ThisFile,OutLineFormat,
          STDEVENT_REC,
           ThisEvent->stUTime, 
           ThisEvent->dbRa, 
           ThisEvent->dbDec, 
           ThisEvent->dbFirstLo, 
           ThisEvent->dbSecondLo, 
           ThisEvent->ulFftBdFreq, 
           ThisEvent->lBinNumber, 
           ThisEvent->dbBirdieSynth, 
           ThisEvent->ucBid, 
           ThisEvent->usPower, 
           ThisEvent->usSigmaThreshold, 
           ThisEvent->usActualThreshold, 
           ThisEvent->cPolarization, 
           ThisEvent->usAdc, 
           ThisEvent->usSpctResLog, 
           ThisEvent->stDate, 
           ThisEvent->ucTid); 
  }


//______________________________________________________________
// io_WriteS3StdEvent()
// Takes a pointer to a standard s3 event record
// (s3_StdEventRec) and writes the data in the record 
// to `ThisFile` as a type 'B' record (see man s4data).
//______________________________________________________________

void io_WriteS3StdEvent(FILE *ThisFile,s3_StdEventRec *ThisEvent)
  {
  fprintf(ThisFile,OutLineS3Format,
          S3_STDEVENT_REC,
          ThisEvent->stStrTime,
          ThisEvent->dbRa,
          ThisEvent->dbDec,
          ThisEvent->dbLocos,
          ThisEvent->lSynth,
          ThisEvent->lBinNo,
          ThisEvent->usPwr,
          ThisEvent->usActualThreshold,
          ThisEvent->usAdc,
          ThisEvent->stDate,
          ThisEvent->usSigmaThreshold,
          ThisEvent->usSpectralRes);
  }

//______________________________________________________________
// io_WriteCandidate()
// Takes a pointer to a candidate record (CandidateRec)
// and writes the data in the record to `ThisFile`
// as a type 'C' record (see man s4data).
//______________________________________________________________
 
void io_WriteCandidate(FILE *ThisFile, CandidateRec * Candidate)
  {
  fprintf(ThisFile, OutLineCandidate , CANDIDATE_REC,
                                       Candidate->usAlgo,
                                       Candidate->dbRa,
                                       Candidate->dbDec,
                                       Candidate->dbSkySep,
                                       Candidate->dbFreq / 1e6,
                                       Candidate->lFreqSep,
                                       Candidate->lHiPower,
                                       Candidate->lMaxSpectra1,
                                       Candidate->lMaxSpectra2,
                                       Candidate->lNumDetect,
                                       Candidate->lLineNum,
                                       Candidate->dbBeamFit,
                                       Candidate->stStarName,
                                       Candidate->dbStarDistance,
                                       Candidate->stSpectralClass,
                                       Candidate->lNumObs,
                                       Candidate->dbObsScore);
  }

//______________________________________________________________
// io_WriteBeamfitRec()
// Takes a pointer to a beamfit record (BeamfitRec)
// and writes the data in the record to `ThisFile`
// as a type 'D' record (see man s4data).
//______________________________________________________________
 
void io_WriteBeamfitRec(FILE *ThisFile, BeamfitRec * BeamfitResult)
  {
  fprintf(ThisFile, OutLineBeamfitRec , BEAMFIT_REC,
                                        BeamfitResult->dVersion,
                                        BeamfitResult->stStrTime,
                                        BeamfitResult->stDate,
                                        BeamfitResult->lNumSpectra,
                                        BeamfitResult->lNumEvents,
                                        BeamfitResult->dbHiPower,
                                        BeamfitResult->dChiSqr,
                                        BeamfitResult->dProbability,
                                        BeamfitResult->dScore,
                                        BeamfitResult->dGausSigma,
                                        BeamfitResult->dGausAmean,
                                        BeamfitResult->dGausDmean);
  }

//______________________________________________________________
// io_WriteSahGaussianRec()
// Takes a pointer to a SETI@home gaussian record
// and writes the data in the record to `ThisFile`
// as a type 'G' record).
//______________________________________________________________

void io_WriteSahGaussianRec(FILE *ThisFile,  sah_Gaussian_t * ThisGaussianRec)
  {  
  fprintf(ThisFile, OutLineSahGaussianRec , SAH_GAUSSIAN_REC,
                                        ThisGaussianRec->id,
                                        ThisGaussianRec->result_id,
                                        ThisGaussianRec->peak_power,
                                        ThisGaussianRec->mean_power,
                                        ThisGaussianRec->ra,
                                        ThisGaussianRec->dec,
                                        ThisGaussianRec->time,
                                        ThisGaussianRec->freq,
                                        ThisGaussianRec->sigma,
                                        ThisGaussianRec->chisqr,
                                        ThisGaussianRec->fft_len,
                                        ThisGaussianRec->chirp_rate,
                                        ThisGaussianRec->score,
                                        ThisGaussianRec->max_power,
                                        ThisGaussianRec->redundancy,
                                        ThisGaussianRec->rfi_checked,
                                        ThisGaussianRec->rfi_found,
                                        ThisGaussianRec->trust,
                                        ThisGaussianRec->detection_freq,
                                        ThisGaussianRec->barycentric_freq);
  }
     
//______________________________________________________________
// io_WriteComment()
// Takes a pointer to a character string and writes
// the string to `ThisFile` as a type '#' record (see man s4data).
//______________________________________________________________

long io_WriteComment(FILE *ThisFile, char * ThisComment)
  {
  if(strlen(ThisComment) <= MAX_LINE_LENGTH)
    {
    fprintf(ThisFile, OutLineComment, COMMENT_REC, 
                                      ThisComment);
    return 1;
    }
  else
    return 0;
  }

//______________________________________________________________
// io_WriteRecFromFile()
// Takes an input file stream and a file position (from ftell) and
// prints out whatever record that might be at that position
// in the input file to the output file
//______________________________________________________________

long io_WriteRecFromFile(FILE *OutFile, FILE *InFile, long lLineNo, 
                         unsigned char cRecordType)
  {
  void *InRec;

  fseek(InFile,lLineNo,SEEK_SET);

  if (cRecordType == STDEVENT_REC) { 
    if ( ( (StdEventRec *)InRec = (StdEventRec *)malloc(sizeof(StdEventRec)) ) == NULL)
      return 0; 
    if (!io_GetNextRecord(InFile,InRec,cRecordType)) return 0;
    io_WriteStdEvent(OutFile,(StdEventRec *)InRec);
    }
  else if (cRecordType == S3_STDEVENT_REC) {
    if ( ( (s3_StdEventRec *)InRec = (s3_StdEventRec *)malloc(sizeof(s3_StdEventRec)) ) == NULL)
      return 0; 
    if (!io_GetNextRecord(InFile,InRec,cRecordType)) return 0;
    io_WriteS3StdEvent(OutFile,(s3_StdEventRec *)InRec);
    }
  else if (cRecordType == BEAMFIT_REC) {
    if ( ( (BeamfitRec *)InRec = (BeamfitRec *)malloc(sizeof(BeamfitRec)) ) == NULL)
      return 0; 
    if (!io_GetNextRecord(InFile,InRec,cRecordType)) return 0;
    io_WriteBeamfitRec(OutFile,(BeamfitRec *)InRec);
    }
  else if (cRecordType == SAH_GAUSSIAN_REC) {
    if ( ( (sah_Gaussian_t *)InRec = (sah_Gaussian_t *)malloc(sizeof(sah_Gaussian_t)) ) == NULL)
      return 0; 
    if (!io_GetNextRecord(InFile,InRec,cRecordType)) return 0;
    io_WriteSahGaussianRec(OutFile,(sah_Gaussian_t *)InRec);
    }
  else return 0;

  free(InRec);

  return 1;

  }


//__________________________________________________________________
// io_PeekNextRecord()
// takes as input a SERENDIP data file and returns the
// next character in the file if it is a valid SERENDIP record
// type, but leaves the file pointer unchanged.  
// Returns 0 on error.
//__________________________________________________________________

unsigned char io_PeekNextRecord(FILE *ThisFile)
  {
  long c;

  // get lines until non-white space first character or EOF
  while (((c = fgetc(ThisFile)) != EOF)
        && (c == ' ' || c == '\t' || c == '\n'))
    if (c != '\n')
      if (! io_SkipNextRecord(ThisFile))
        return 0;

  // return c if valid record type,
  // return 0 if invalid record type, EOF, or error
  if ((c != EOF) 
      && (ungetc((int) c, ThisFile) != EOF)
      && ((c == STDEVENT_REC)    || 
          (c == COMMENT_REC)     || 
          (c == S3_STDEVENT_REC) ||
          (c == CANDIDATE_REC)   ||
          (c == BEAMFIT_REC)     ||
          (c == SAH_GAUSSIAN_REC))
     )
    return (unsigned char) c;
  else
    return 0;
  }

//__________________________________________________________________
// io_SkipNextRecord()
// Advances the file pointer in 'ThisFile' past the next record
// returns 0 on end of file or error. Returns 1 if not eof.
//__________________________________________________________________ 

long io_SkipNextRecord(FILE *ThisFile)
  {
  char str[MAX_LINE_LENGTH+1];

  if (fgets(str, MAX_LINE_LENGTH+1, ThisFile))
    return 1;
  else
    return 0;
  }


//__________________________________________________________________________
// io_s3GetNextEvent()
// Reads the next record from a file in the serendip 3 offline format and 
// puts it into a data structure, type s3_StdEventRec. In addition, this 
// function calculates frequency, decimal hour, and julian date and puts 
// those values into the data structure.
// NOTE: Century is hard-coded; if 2-digit year is less than 80, it is 
// assumed to be in the 21st century.
// Routine returns a return value of 1 if successful, 0 if unsuccessful.
//__________________________________________________________________________

long io_s3GetNextEvent(FILE *ThisFile, s3_StdEventRec *ThisEvent)
  {
  long day_i = 0;
  long month_i = 0;
  long year_i = 0;

  if (fscanf(ThisFile,InLineOldS3Format,
            ThisEvent->stStrTime,&ThisEvent->dbRa,&ThisEvent->dbDec,
            &ThisEvent->dbLocos,&ThisEvent->lSynth,&ThisEvent->lBinNo,
            &ThisEvent->usPwr,&ThisEvent->usActualThreshold,&ThisEvent->usAdc,
            ThisEvent->stDate,&ThisEvent->usSigmaThreshold,
            &ThisEvent->usSpectralRes) == 12)
      {
      // calculate frequency
      ThisEvent->dbFreq = fq_s3fcalc(ThisEvent->dbLocos, ThisEvent->lSynth,
                                     ThisEvent->lBinNo);

      // calculate decimal time
      tm_s3HmsToDec(ThisEvent->stStrTime, &ThisEvent->dbDecTime);

      // convert date string to integers
      tm_DateToI(ThisEvent->stDate, &day_i, &month_i, &year_i);

      // add century to year 
      year_i = tm_CentYear(year_i);

      // calculate julian date
      tm_ToJul(month_i, day_i, year_i, ThisEvent->dbDecTime);

      return 1;
      }
    else
      return 0;
  }


//______________________________________________________________
long io_CountRecords(FILE * ThisFile, char RecordType)
//______________________________________________________________
  {
  
  long RecordCount = 0;
  unsigned char peekrtn;

  rewind(ThisFile);

  while ((peekrtn = io_PeekNextRecord(ThisFile)))
    {
    if (peekrtn == RecordType)
      RecordCount++;
    if (! io_SkipNextRecord(ThisFile))
      break;
    }
 

  // return -1 if error
  if (! feof(ThisFile))
    {
    rewind(ThisFile);
    return -1;
    }
  else
    {
    rewind(ThisFile);
    return RecordCount;
    }
}

//________________________________________________________________________
// io_GetLastRecord()
// Takes a void pointer to a SERENDIP record structure and the record type,
// as defined in the derendip data format (see man s4data),
// and
//   a) rewinds to beginning of file
//   b) counts number of valid events in file
//   c) uses the count of events to adjust file pointer to next to last
//      valid event
//   d) reads last valid event in file into a serendip structure
//   e) rewinds pointer to beginning of file
//
// This function returns 1 on success, otherwise returns a 0. )
//________________________________________________________________________

long io_GetLastRecord(FILE *ThisFile, void *ThisRecord,unsigned char RecordType)

{
unsigned char peekrtn;
unsigned long event_cnt = 0;
unsigned long record_cnt = 0;


         rewind (ThisFile);

                  /* Find number of events. */

         if ((event_cnt =(io_CountRecords(ThisFile,RecordType))) <= 0)

                        /* If count of records function discovers no events of the
                                specified record type, or if there is an error
                                caused by counting, a zero is returned. */

         return (0);


                  /* Following loop has no error checking.  Assumption is
                          that io_Count Records should encounter all errors
                          using Peek and Skip functions*/

                  /* Following loop peeks and skips through file events for
                          event_cnt -1 times (avoiding usage of io_GetNextRecord                          which performs other calculations.) */


                  /* The Peek and Skip sequence should leave file pointer
                          at a position that is 1 event prior to last event
                          in file */

         while (record_cnt < (event_cnt-1))
                  {
                  peekrtn = io_PeekNextRecord(ThisFile);


                        if (peekrtn == RecordType)
                                record_cnt++;

                  io_SkipNextRecord(ThisFile);

                  }

                         /* Retrieval of last valid file event performed with
                                 this statement */

         if ((io_GetNextRecord(ThisFile, ThisRecord, RecordType)) != 1)

                        /* if last record exists, but a calculation failure
                                occurs, the io_GetNextRecord function does not
                                return a "1".  This failure is passed through to                                io_GetLastRecord function return. */
           {
           rewind (ThisFile);
           return (0);    /* failure */
           }
         else
           { 
           rewind (ThisFile);
           return (1);   /* success */
           }
}

//______________________________________________________________
// io_GetFilePosition()
//______________________________________________________________
 
off_t io_GetFilePosition(FILE *ThisFile)
 
{
return lseek(fileno(ThisFile), 0L, SEEK_CUR);
}
 
 
//______________________________________________________________
// io_SetFilePosition()
// Sets the file pointer to specified FilePosition.
// Returns FilePosition on success, otherwise returns 0.
//______________________________________________________________
 
off_t io_SetFilePosition(FILE *ThisFile, off_t FilePosition)
 
{
if (lseek(fileno(ThisFile), FilePosition, SEEK_SET) == FilePosition)
  {  
  return FilePosition;
  }  
else
  { 
  return 0;
  }  
}

//______________________________________________________________
// io_GetEngMonRecord()
// Reads from .eng file an engineering monitor record
//______________________________________________________________
BOOLEAN io_GetEngMonRecord(FILE *ThisFile, EngMonRec *ThisRecord)
  {

    char stInString[14];
    long u32Time;
    long DOY, Hour, Min, Sec, Ten;
    long CentYear;
    long year;
    struct DateStruct sdate;

    if (!fread(stInString,sizeof(char),18,ThisFile))
      return(FALSE);

    else {

      ThisRecord->ucFlag = 0x00;

      // grab time data from input
      memcpy(&u32Time,stInString,sizeof(long));
      memcpy(&CentYear,stInString+4,sizeof(long));

      // convert from s4 binary format
      tm_Converts4Time(u32Time,&DOY,&Hour,&Min,&Sec,&Ten);

      // UT is our standard time
      tm_AstToUt(&CentYear, &DOY, &Hour);
      ThisRecord->dbUTime = (double)Hour        +
                            (double)Min/60.0    +
                            (double)Sec/3600.0  +
                            (double)Ten/36000.0;		    	// to rcd	

      // calculate julian date
      year = tm_TwoDigitYear(CentYear);			// We need month:day
      tm_DayToMon(DOY-1, year, &sdate);			// for julian conversion
      ThisRecord->dbJulDate = tm_ToJul(sdate.month, 
                                       sdate.day, 
                                       CentYear, 
                                       ThisRecord->dbUTime);  		// to rcd

      // time string to rcd
      sprintf(ThisRecord->stUTime,   "%02lu", Hour);
      sprintf(ThisRecord->stUTime+2, "%02lu", Min);
      sprintf(ThisRecord->stUTime+4, "%02lu", Sec);
      sprintf(ThisRecord->stUTime+6, "%02lu", Ten*10);    // hundredths of secs
 
      // the engineering values themselves to rcd
      memcpy(ThisRecord->ucEngMonitor,stInString+8,sizeof(char)*10);

      return(TRUE);
    }

  }  

//______________________________________________________________
// io_WriteEngMonRecord
// Writes an engineering monitor record in readable form to file
//______________________________________________________________
void io_WriteEngMonRecord(FILE *ThisFile, EngMonRec *ThisRecord)
  {

  fprintf(ThisFile,OutLineEngMonRec,
           ThisRecord->dbJulDate,
           ThisRecord->ucEngMonitor[0],
           ThisRecord->ucEngMonitor[1],
           ThisRecord->ucEngMonitor[2],
           ThisRecord->ucEngMonitor[3],
           ThisRecord->ucEngMonitor[4],
           ThisRecord->ucEngMonitor[5],
           ThisRecord->ucEngMonitor[6],
           ThisRecord->ucEngMonitor[7],
           ThisRecord->ucEngMonitor[8],
           ThisRecord->ucEngMonitor[9]
           );
  }

//______________________________________________________________
// io_GetFileSize()
// Returns # of bytes of a file (which is already opened)
//______________________________________________________________

long io_GetFileSize(FILE *ThisFile)
{
  struct stat statbuf;

  fstat(fileno(ThisFile),&statbuf);
  return (unsigned long)statbuf.st_size;
}

