/* ________________________________________________________________________
 programmer : Matt Lebofsky
 source name: s4skyplot.cc 
 compiler   : gcc
___________________________________________________________________________
 modification history:
   version 1.0 being worked on
   9/30/03 - cleaned up gcc 3.0 compiler warnings
___________________________________________________________________________
 usage: This is a library of functions to be used in read/writing/updating
        the s4 skyplot database
___________________________________________________________________________
notes/bugs: 
___________________________________________________________________________
*/

#include <stdio.h>
#include <string.h>
#include "SIREN.h"
#include "s4database.h"
#include "s4io.h"

char *cProperSkyplotFileName = "Skyplot_File";

//_________________________________________________________________________
// void *sp_GetSkyplotFileName()
// Will check environmental variables and return Filename of skyplot dbase.
// Exits 1 if things are screwed up (after sendings msgs to stderr).
//_________________________________________________________________________

void sp_GetSkyplotFileName(char **cFileName)
{
char *cTestStr;

if ((*cFileName = getenv("S4_SKYPLOT_FILE")) == NULL) {
  fprintf(stderr, "Could not obtain Skyplot database file name\n");
  fprintf(stderr, "Set the S4_SKYPLOT_FILE enviromental variable\n");
  exit(1);
  }
else {
  if((cTestStr = (char *)strstr(*cFileName,cProperSkyplotFileName)) == NULL ||
      strcmp(cProperSkyplotFileName, cTestStr) != 0) {  
    fprintf(stderr, "Skyplot database file must be called %s\n",
            cProperSkyplotFileName);
    exit(1);
    }
  } 
}

//_________________________________________________________________________
// long sp_ReadSkyplotFile()
// Opens skyplot file, figures out its size, and reads it into an array of
// SkyplotElements. Returns the number of elements in the array, or -1 if
// there was an error opening file.
//_________________________________________________________________________

long sp_ReadSkyplotFile(SkyplotElement **stSkyArray, int *iSkyplotFileNo)
{
long lFileSizeBytes, lArraySize;
char *cFileName;
FILE *fSkyplotFile;

sp_GetSkyplotFileName(&cFileName);

if ((fSkyplotFile = fopen(cFileName,"r+")) == NULL) {
  fprintf(stderr,"Unable to open skyplot database file: %s\n",cFileName);
  return(-1);
  }

*iSkyplotFileNo = fileno(fSkyplotFile);
// wait for the lock to disappear
lockf(*iSkyplotFileNo, F_LOCK, 0);

lFileSizeBytes = io_GetFileSize(fSkyplotFile);
lArraySize = lFileSizeBytes/sizeof(SkyplotElement);

*stSkyArray = (SkyplotElement *) malloc(lFileSizeBytes);

fread(*stSkyArray,sizeof(SkyplotElement),lArraySize,fSkyplotFile);

return(lFileSizeBytes / sizeof(SkyplotElement));
}

//_________________________________________________________________________
// long sp_WriteSkyplotFile()
// Opens skyplot file for writing, and copies the array into it.
// WARNING: Doesn't check for file locking!! This is meant to create a
// separate file of a different name than the skyplot database.
//_________________________________________________________________________

long sp_WriteSkyplotFile(char *cFileName, SkyplotElement *stSkyArray,
                         long lArraySize)
{
FILE *fSkyplotFile;

if ((fSkyplotFile = fopen(cFileName,"w")) == NULL) {
  fprintf(stderr,"Can't write skyplot file: %s\n",cFileName);
  return(-1);
  }

rewind(fSkyplotFile);
fwrite(stSkyArray,sizeof(SkyplotElement),lArraySize,fSkyplotFile);
fclose(fSkyplotFile);
return(lArraySize);
}

//_________________________________________________________________________
// void sp_CloseSkyplotFile()
// Unlocks and closes skyplot file
//_________________________________________________________________________

void sp_CloseSkyplotFile(int iSkyplotFileNo)
{
lockf(iSkyplotFileNo, F_ULOCK, 0);
close(iSkyplotFileNo);
}

//
//_________________________________________________________________________
// long sp_RaDectoElement()
// Converts Ra and Dec into a single-dimension array element.
//_________________________________________________________________________

long sp_RaDectoElement(double ra, double dec)
{
return((SKYPLOT_CONVERT_RA(ra))+(SKYPLOT_CONVERT_DEC(dec)*SKYPLOT_WIDTH));
}

//_________________________________________________________________________
// void sp_ElementtoRaDec()
// Converts a single-dimension array element into Ra and Dec.
//_________________________________________________________________________

void sp_ElementtoRaDec(long lElement, double *ra, double *dec)
{
*ra = SKYPLOT_R_CONVERT_RA(lElement % SKYPLOT_WIDTH);
*dec = SKYPLOT_R_CONVERT_DEC(long(lElement/SKYPLOT_WIDTH));
}

//_________________________________________________________________________
// long sp_CountObs()
// Takes an Ra and Dec and a Skyplot array and counts the # of observations.
//_________________________________________________________________________

long sp_CountObs(double ra, double dec, SkyplotElement *stSkyArray)
{
SkyplotElement CurrentElement;
long lNumObs;

CurrentElement = stSkyArray[sp_RaDectoElement(ra,dec)];
lNumObs = CurrentElement.ulDbaseKey ? 1 : 0;

while (CurrentElement.lLink != -1) {
  CurrentElement = stSkyArray[CurrentElement.lLink];
  if (CurrentElement.ulDbaseKey) lNumObs++;
  }

return(lNumObs);
}

//_________________________________________________________________________
// long sp_ShowObs()
// Takes an Ra and Dec and a Skyplot array and shows the names of the
// dbase files (one every 10 characters) which have observed this pixel.
// Like sp_ShowObs, it returns the # of observations.
//_________________________________________________________________________

long sp_ShowObs(double ra, double dec, SkyplotElement *stSkyArray)
{
SkyplotElement CurrentElement;
long lNumObs;
char cFileName[9];

CurrentElement = stSkyArray[sp_RaDectoElement(ra,dec)];
lNumObs = CurrentElement.ulDbaseKey ? 1 : 0;

while (CurrentElement.lLink != -1) {
  db_KeyToFileName(CurrentElement.ulDbaseKey,cFileName);
  if (CurrentElement.ulDbaseKey) printf("- %s\n",cFileName);
  CurrentElement = stSkyArray[CurrentElement.lLink];
  if (CurrentElement.ulDbaseKey) lNumObs++;
  }

if (lNumObs) {
  db_KeyToFileName(CurrentElement.ulDbaseKey,cFileName);
  if (CurrentElement.ulDbaseKey) printf("- %s\n",cFileName);
  }

return(lNumObs);
}

//_________________________________________________________________________
// BOOLEAN sp_FindDbase()
// Takes an Ra and Dec and a Skyplot array and looks for a dbase file
// within that pixel. Returns true if found. False otherwise.
//_________________________________________________________________________

BOOLEAN sp_FindDbase(double ra, double dec, char *cDbaseFileName,
                  SkyplotElement *stSkyArray)
{
SkyplotElement CurrentElement;
unsigned long ulDbaseKeyToCheck;
BOOLEAN bWasFound;

bWasFound = FALSE;
ulDbaseKeyToCheck = db_FileNameToKey(cDbaseFileName);

CurrentElement = stSkyArray[sp_RaDectoElement(ra,dec)];
if (CurrentElement.ulDbaseKey == ulDbaseKeyToCheck) bWasFound = TRUE;

while (CurrentElement.lLink != -1) {
  CurrentElement = stSkyArray[CurrentElement.lLink];
  if (CurrentElement.ulDbaseKey == ulDbaseKeyToCheck) bWasFound = TRUE;
  }

return(bWasFound);
}

//_________________________________________________________________________
// void sp_AddObs()
// Adds a new dbase key to the Skyplot array, updates the lArraySize
// variable. Will not add anything if dbase key is already included in
// the linked list attached to the pixel.
//_________________________________________________________________________

void sp_AddObs(double ra, double dec, char *cDbaseFileName,
               SkyplotElement **stSkyArray, long *lArraySize,
               int iSkyplotFileNo)
{
SkyplotElement CurrentElement;
long lCurrentIndex;
unsigned long ulDbaseKeyToAdd;
BOOLEAN bNotAddedYet;

bNotAddedYet = TRUE;
lCurrentIndex = sp_RaDectoElement(ra,dec);
CurrentElement = (*stSkyArray)[lCurrentIndex];
ulDbaseKeyToAdd = db_FileNameToKey(cDbaseFileName);

while (CurrentElement.lLink != -1) {
  if (CurrentElement.ulDbaseKey == ulDbaseKeyToAdd) 
    bNotAddedYet = FALSE;
  lCurrentIndex = CurrentElement.lLink;
  CurrentElement = (*stSkyArray)[lCurrentIndex];
  }

// reached end of list, check to see if this last dbasekey is what we're
// wishing to add. If not, and we haven't seen it earlier, add it now.

if ((CurrentElement.ulDbaseKey != ulDbaseKeyToAdd) && bNotAddedYet) {

  if (CurrentElement.ulDbaseKey == 0) { // if this pixel is already empty
                                        // there's no need to add another

    (*stSkyArray)[lCurrentIndex].ulDbaseKey = ulDbaseKeyToAdd;
    lseek(iSkyplotFileNo,(lCurrentIndex*sizeof(SkyplotElement)),SEEK_SET);
    write(iSkyplotFileNo,&(*stSkyArray)[lCurrentIndex],
          sizeof(SkyplotElement));
    } 

  else {                                // need to realloc and adjust arraysize 

    (*lArraySize)++;
    ((*stSkyArray)[lCurrentIndex]).lLink = (*lArraySize)-1;
    lseek(iSkyplotFileNo,(lCurrentIndex*sizeof(SkyplotElement)),SEEK_SET);
    write(iSkyplotFileNo,&(*stSkyArray)[lCurrentIndex],
          sizeof(SkyplotElement));
    *stSkyArray = (SkyplotElement *) realloc(*stSkyArray,
                                            *lArraySize*sizeof(SkyplotElement));
    ((*stSkyArray)[(*lArraySize)-1]).ulDbaseKey = ulDbaseKeyToAdd;
    ((*stSkyArray)[(*lArraySize)-1]).lLink = -1;
    lseek(iSkyplotFileNo,0L,SEEK_END);
    write(iSkyplotFileNo,&(*stSkyArray)[(*lArraySize)-1],
          sizeof(SkyplotElement));
    }

  }

}

//_________________________________________________________________________
// void sp_RemoveObs()
// In the case where we want to remove a dbase key from the skyplot dbase,
// this routine will look in a pixel and blank out (i.e. turn to "0") any
// key that matches the dbase key. 
//_________________________________________________________________________

void sp_RemoveObs(long lCurrentIndex, char *cDbaseFileName,
               SkyplotElement **stSkyArray, int iSkyplotFileNo)
{
SkyplotElement CurrentElement;
long lIndexToRemove;
unsigned long ulDbaseKeyToRemove;

lIndexToRemove = -1;
CurrentElement = (*stSkyArray)[lCurrentIndex];
ulDbaseKeyToRemove = db_FileNameToKey(cDbaseFileName);

while (CurrentElement.lLink != -1) {
  if (CurrentElement.ulDbaseKey == ulDbaseKeyToRemove) 
    lIndexToRemove = lCurrentIndex;
  lCurrentIndex = CurrentElement.lLink;
  CurrentElement = (*stSkyArray)[lCurrentIndex];
  }

// reached end of list, check to see if this last dbasekey is what we're
// wishing to remove.

if (CurrentElement.ulDbaseKey == ulDbaseKeyToRemove)
  lIndexToRemove = lCurrentIndex;

// did we find an index that matches the dbase key to remove? Turn it to "0"!

if (lIndexToRemove != -1) {
  (*stSkyArray)[lIndexToRemove].ulDbaseKey = 0;
  lseek(iSkyplotFileNo,(lIndexToRemove*sizeof(SkyplotElement)),SEEK_SET);
  write(iSkyplotFileNo,&(*stSkyArray)[lIndexToRemove],
  sizeof(SkyplotElement));
  } 

}
