/*
 * Source Code Name  : s4summary.cc
 * Programmer        : Jeff Cobb
 *                   : Project SERENDIP, University of California, Berkeley CA
 * Compiler          : GNU gcc
 * Summary database routines for serendip 4 offline programs.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <strings.h>
#include "SIREN.h"
#include "s4database.h"
#include "s4summary.h"


// Globals
char * sSummaryFileName;
char * sSummaryIndexFileName;

const char * sProperSummIndexName = "Summ_Index_File";
const char * sProperSummFileName = "Summ_File";

DbaseSumRcd_t InitSumRcd;   	 // outside any function, thus static and 
				 // containing type-proper zeros

//===========================================================
int sum_GetDataBase(int Mode)
//===========================================================
  {
  int SumIndexFile=0;
  char * TestStr;

  // Get and check summary index from process environment.
  // The path can be anything but must end in the the standard
  // name.
  if((sSummaryIndexFileName = getenv("S4_SUMM_INDEX")) == NULL)
    {
    fprintf(stderr, "Could not obtain Summary Index file name\n");
    fprintf(stderr, "Set the S4_SUMM_INDEX enviromental variable\n");
    exit(1);
    }
  else
    if((TestStr = strstr(sSummaryIndexFileName, sProperSummIndexName)) == NULL ||
       strcmp(sProperSummIndexName, TestStr) != 0) 
      {
      fprintf(stderr, "Summary Index file must be called %s\n", 
              sProperSummIndexName);
      exit(1);
      }

  // Do the same for the summary data file.
  if((sSummaryFileName = getenv("S4_SUMM_FILE")) == NULL)
    {
    fprintf(stderr, "Could not obtain Summary file name\n");
    fprintf(stderr, "Set the S4_SUMM_FILE enviromental variable\n");
    exit(1);
    }
  else
    if((TestStr = strstr(sSummaryFileName, sProperSummFileName)) == NULL ||
       strcmp(sProperSummFileName, TestStr) != 0) 
      {
      fprintf(stderr, "Summary File file must be called %s\n", 
              sProperSummFileName);
      exit(1);
      }

  if(Mode == O_RDWR) 
    {
    if((SumIndexFile = open(sSummaryIndexFileName, O_RDWR)) == -1)
      {
      fprintf(stderr, "Could not obtain Summary database for locking\n");
      exit(1);
      }

    if((lockf(SumIndexFile, F_LOCK, 0)) == -1)
      {
      perror("sum_LockDataBase : ");
      exit(1);
      }
    } 
  else if(Mode == O_RDONLY) 
    {
    if((SumIndexFile = open(sSummaryIndexFileName, O_RDONLY)) == -1)
      {
      fprintf(stderr, "Could not obtain Summary database for reading\n");
      exit(1);
      }
    }
  else
    {
    fprintf(stderr, "Unkown Summary DB open mode\n");
    exit(1);
    }

  return(SumIndexFile);
  }

//===========================================================
void sum_ReleaseDataBase(int SumIndexFile)
//===========================================================
  {
  lockf(SumIndexFile, F_ULOCK, 0);
  close(SumIndexFile);
  }

//===========================================================
unsigned long sum_GetRcd(char          *sDbaseFileName,
                         DbaseSumRcd_t *ThisSumRcd,
                         int            SumIndexFile,
                         int		Mode)
//===========================================================
  {

  long Index;
  int SumFile;

  Index = sum_GetIndex(db_FileNameToKey(sDbaseFileName),
                       SumIndexFile);

  if(Index == -1)
    {
    memcpy((char *)ThisSumRcd, (char *)&InitSumRcd, sizeof(DbaseSumRcd_t));  
                                                      // init new rcd
    strcpy(ThisSumRcd->sDbaseName, sDbaseFileName);   // start new rcd
    }
  else
    {
    if((SumFile = open(sSummaryFileName, Mode)) == -1)
      {
      perror("sum_GetRcd : ");
      exit(1);
      }
    sum_ReadRcd(SumFile, ThisSumRcd, Index);
    close(SumFile);
    }

  return(Index);
  }

//===========================================================
void sum_PutRcd(long SumIndex,
                DbaseSumRcd_t * ThisSumRcd,
                int           SumIndexFile)
//===========================================================
  {
  int SumFile;

  if((SumFile = open(sSummaryFileName, O_RDWR)) == -1)
    {
    perror("sum_WriteRcd : ");
    exit(1);
    }

  // Set up to write new record if need be.
  if(SumIndex == -1)
    {
    SumIndex = sum_GetNumRcds(SumIndexFile);
    sum_AddIndex(SumIndex,
                   db_FileNameToKey(ThisSumRcd->sDbaseName),
                   SumIndexFile);
    }
  sum_WriteRcd(SumFile, ThisSumRcd, SumIndex);
  close(SumFile);
  }

//===========================================================
long sum_GetNumRcds(int SumIndexFile)
//===========================================================
  {
  struct stat SumIndexStat;

  fstat(SumIndexFile, &SumIndexStat);

  return(SumIndexStat.st_size / sizeof(SumIndexRcd));
  }

//===========================================================
long sum_GetIndex(unsigned long Key, int SumIndexFile)
//===========================================================
  {
  SumIndexRcd * SumIndices;
  SumIndexRcd * ThisIndex;
  SumIndexRcd   KeyElement;
  long NumIndices;
  long rtn=0;

  KeyElement.Key = Key;

  if((NumIndices = sum_GetNumRcds(SumIndexFile)) == 0)
    rtn = -1;                  // index file is empty
  else
    {
    if((SumIndices = (SumIndexRcd *)calloc(NumIndices, sizeof(SumIndexRcd))) == NULL)
      {
      perror("sum_GetIndex : ");
      exit(1);
      }
    else
      {
      sum_ReadIndices(SumIndices, NumIndices, SumIndexFile);
      ThisIndex = (SumIndexRcd *)bsearch((char *)&KeyElement,
                                         (char *)SumIndices,
                                         (unsigned int)NumIndices,
                                         sizeof(SumIndexRcd),
                                         sum_IndexCmp);
      if(ThisIndex == NULL)
        rtn = -1;                // index not found for this dbase file
      else
        rtn = ThisIndex->Index;  // index was found
      free(SumIndices);
       }
    }

  return(rtn);
  }

//===========================================================
void  sum_AddIndex(long            SumIndex,
                   unsigned long   Key,
                   int             SumIndexFile)
//===========================================================
// On entry to this routine, SumIndex is equal to the number
// of indices prior to the addition.
  {
  SumIndexRcd * SumIndices;

  if((SumIndices = (SumIndexRcd *)calloc(SumIndex+1, sizeof(SumIndexRcd))) == NULL)
    {
    perror("sum_AddIndex :");
    exit(1);
    }
  else
    {
    sum_ReadIndices(SumIndices, SumIndex, SumIndexFile);
    SumIndices[SumIndex].Index = SumIndex;
    SumIndices[SumIndex].Key   = Key;
    if(SumIndex != 0 && SumIndices[SumIndex].Key < SumIndices[SumIndex-1].Key)
      qsort((char *)SumIndices, SumIndex+1, sizeof(SumIndexRcd), sum_IndexCmp);
    sum_WriteIndices(SumIndices, SumIndex+1, SumIndexFile);
    free(SumIndices);
    }
  }

//===========================================================
long sum_GetNumInDateRange(char * Date1, char * Date2, int SumIndexFile)
//===========================================================
  {
  SumIndexRcd * SumIndices;
  char LowFile[DBASE_FILENAME_LEN];
  char HighFile[DBASE_FILENAME_LEN];
  unsigned long LowKey, HighKey;
  long i, rtn=0, HighBound, LowBound, NumIndices;

  strcpy(LowFile, Date1);
  strcat(LowFile, "aa");
  strcpy(HighFile, Date2);
  strcat(HighFile, "zz");

  LowKey  = db_FileNameToKey(LowFile);
  HighKey = db_FileNameToKey(HighFile);

  if((NumIndices = sum_GetNumRcds(SumIndexFile)) == 0)
    rtn = 0;
  else
    {
    if((SumIndices = (SumIndexRcd *)calloc(NumIndices, sizeof(SumIndexRcd))) == NULL)
      {
      perror("sum_GetNumInDateRange : ");
      exit(1);
      }
    else
      {
      sum_ReadIndices(SumIndices, NumIndices, SumIndexFile);
      for(i=0; SumIndices[i].Key < LowKey && i < NumIndices; i++);
      if(i < NumIndices && i <= HighBound)
        {
        LowBound = i;
        for(; SumIndices[i].Key <= HighKey && i < NumIndices; i++);
        HighBound = i - 1;
        rtn = HighBound - LowBound + 1;
        db_KeyToFileName(SumIndices[LowBound].Key, Date1);
        db_KeyToFileName(SumIndices[HighBound].Key, Date2);
        }
      else
        rtn = 0;
      free(SumIndices);
      }
    }

  return(rtn);
  }

//===========================================================
BOOLEAN sum_GetNextRcd(char        * sDbaseFileName,
                          DbaseSumRcd_t * ThisSumRcd,
                          BOOLEAN       LastCall,
                          int           SumIndexFile,
                          int           Mode)
//===========================================================
  {
  SumIndexRcd   KeyElement;
  BOOLEAN rtn;

  static SumIndexRcd * SumIndices = NULL;
  static SumIndexRcd * ThisIndex  = NULL;
  static long NumIndices = 0;
  static int SumFile;

  rtn = TRUE;

  if(!LastCall)
    {
    if(SumIndices == NULL)     // First call
      {
      if((NumIndices = sum_GetNumRcds(SumIndexFile)) == 0)
        rtn = FALSE;                  // index file is empty
      else
        {
        if((SumIndices = (SumIndexRcd *)calloc(NumIndices, sizeof(SumIndexRcd))) ==NULL)
          {
          perror("sum_GetNextRcd : ");
          exit(1);
          }
        else
          {
          if((SumFile = open(sSummaryFileName, Mode)) == -1)
            {
            perror("sum_GetRcd : ");
            exit(1);
            }
          sum_ReadIndices(SumIndices, NumIndices, SumIndexFile);
          KeyElement.Key = db_FileNameToKey(sDbaseFileName);
          ThisIndex = (SumIndexRcd *)bsearch((char *)&KeyElement,
                                             (char *)SumIndices,
                                             (unsigned int)NumIndices,
                                             sizeof(SumIndexRcd),
                                             sum_IndexCmp);
          if(ThisIndex == NULL)
            rtn = FALSE;                   // index not found for this dbase file
          else
            ThisIndex--;
          } // end OK calloc
        } // end OK sum_GetNumRcds()
      } // end SumIndices == NULL
    if(rtn == TRUE)                       // First and subsequent calls
      {
      ThisIndex++;
      if(ThisIndex < (SumIndices + NumIndices))
        {
        sum_ReadRcd(SumFile, ThisSumRcd, ThisIndex->Index);
        db_KeyToFileName(ThisIndex->Key, sDbaseFileName);
        } // end ThisIndex < (SumIndices + NumIndices)
      else
        rtn = FALSE;
      } // end rtn == TRUE (all OK)
    } // end !LastCall
  else
    if(SumIndices != NULL)
      {
      free(SumIndices);
      close(SumFile);
      }

  return(rtn);
  }

//===========================================================
void sum_ReadIndices(SumIndexRcd * SumIndices,
                       long          NumSumRcds,
                       int           SumIndexFile)
//===========================================================
  {
  lseek(SumIndexFile, 0L, SEEK_SET);    // always read from beginning  
  read(SumIndexFile, SumIndices, sizeof(SumIndexRcd) *  NumSumRcds);
  }

//===========================================================
void sum_WriteIndices(SumIndexRcd * SumIndices,
                        long          NumSumRcds,
                        int           SumIndexFile)
//===========================================================
  {
  lseek(SumIndexFile, 0L, SEEK_SET);    // always write from beginning  
  write(SumIndexFile, SumIndices, sizeof(SumIndexRcd) *  NumSumRcds);
  }

//===========================================================
void sum_ReadRcd(int SumFile, DbaseSumRcd_t * ThisSumRcd, long Index)
//===========================================================
  {
  lseek(SumFile, Index * sizeof(DbaseSumRcd_t), SEEK_SET);
  read(SumFile, ThisSumRcd, sizeof(DbaseSumRcd_t));
  }

//===========================================================
void sum_WriteRcd(int SumFile, DbaseSumRcd_t * ThisSumRcd, long Index)
//===========================================================
  {
  lseek(SumFile, Index * sizeof(DbaseSumRcd_t), SEEK_SET);
  write(SumFile, ThisSumRcd, sizeof(DbaseSumRcd_t));
  }

//===========================================================
// newer gcc needs the following to be cast properly 
// int sum_IndexCmp(SumIndexRcd * Index1, SumIndexRcd * Index2)
int sum_IndexCmp(const void * Index1, const void * Index2)
//===========================================================
  {
  return(((SumIndexRcd *)Index1)->Key - ((SumIndexRcd *)Index2)->Key);
  }
