// ________________________________________________________________________
// programmer : Jeff Cobb
// source name: MakeCandidates.cc
// compiler   : AT&T C++ from SUN
//
// usage : s3AddCandidate s3|s4 algo_nbr in_file
// algorithms : 0 - nearby stars (Gliese catalog)
//              1 - exdate 
//              2 - multhit
//              3 - multhit.exdate
//              4 - limpwr
//              5 - limpwr.exdate
//              6 - globular glusters
//              7 - beamfit
//              8 - beamfit.exdate
//              9 - planet search star list
//             10 - extrasolar plantary systems
//______________________________________________________________________
// modification history:
//   -jeffc 1/96 - added algo 6 - globular clusters
//   -jeffc 1/98 - added algo 7 - beamfit 
//   -jeffc 2/98 - added algo 9 and 10 
//______________________________________________________________________
// stdio.h placed after string.h because otherwise NULL get defined as void *
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include "SIREN.h"
#include "s4io.h"
#include "s4freq.h"
#include "s4coord.h"
#include "s4conv.h"

long   DoGroup          (CandidateRec Candidates[], 
                         FILE * InFile, 
                         int Algo, 
                         char RecordType,
                         char GroupMasterType);
long   DoSingle         (CandidateRec Candidates[], FILE * InFile, int Algo, char RecordType);
long   DoGliese         (CandidateRec Candidates[], FILE * InFile, int Algo);
long   DoGlob           (CandidateRec Candidates[], FILE * InFile, int Algo);
void   OutPutCandidates (CandidateRec Candidates[], long NumCandidates);
double GetMeanRa        (MinEventRec Events[], long NumEvents);
double GetMeanDec       (MinEventRec Events[], long NumEvents);
double GetMeanFreq      (MinEventRec Events[], long NumEvents);
long   GetHiPwr         (MinEventRec Events[], long NumEvents);
void   GetMaxSpectra    (MinEventRec Events[], long NumEvents, long * MaxSpectra1,
                         long * MaxSpectra2);
long   GetNumDetect     (MinEventRec Events[], long NumEvents);
long   GetFreqSep       (MinEventRec Events[], long NumEvents);
double GetSkySep        (MinEventRec Events[], long NumEvents);
void   InitCandidates   (CandidateRec Candidates[], long NumCandidates);
int    DblCompare       (const void * a, const void * b);
void   output_min       (MinEventRec Events[], long i);
void SaveGroupMaster(FILE * InFile,
                     char GroupMasterType,
                     void ** pGroupMasterPtr);
void IncludeGroupMasterData(CandidateRec * Candidates,
                            long CandidateIndex,
                            char GroupMasterType,
                            void * pGroupMasterPtr);

//___________________________________________________________________
int main (int argc, char * argv[])
//___________________________________________________________________
  {
  FILE * InFile;

  char * usage = "usage : MakeCandidates s3|s4 algo_nbr in_file";
 
  int Algo;
  long NumCandidates;
  char RecordType;
  char GroupMasterType = ' ';   // init to no group master

  CandidateRec Candidates[MAX_CANDIDATES];

  if(argc != 4                                                      || 
     sscanf(argv[2], "%d", &Algo) != 1                              ||
     !(strcmp(argv[1], "s3") == 0 || strcmp(argv[1], "s4") == 0))
    {
    printf("%s\n", usage);
    exit(1);
    } 

  if(strcmp(argv[1], "s3") == 0)                    // s3 or s4 ?
    RecordType = S3_STDEVENT_REC;
  else
    RecordType = STDEVENT_REC;

  if ((InFile = fopen(argv[3],"r")) == NULL)
    {
    fprintf(stderr, "Can't open input file %s.\n", argv[3]);
    exit(1);
    }

  InitCandidates(Candidates, MAX_CANDIDATES);

  switch (Algo)
    {
    case 9  : 
    case 10 : 
    case 0 : NumCandidates =
	     DoGliese(Candidates, InFile, Algo);
	     break;
    case 1 : NumCandidates = 
             DoGroup(Candidates, InFile, Algo, RecordType, GroupMasterType);
	     break;
    case 2 : NumCandidates =
             DoGroup(Candidates, InFile, Algo, RecordType, GroupMasterType);
	     break;
    case 3 : NumCandidates =
             DoGroup(Candidates, InFile, Algo, RecordType, GroupMasterType);
	     break;
    case 4 : NumCandidates =
	     DoSingle(Candidates, InFile, Algo, RecordType);
	     break;
    case 5 : NumCandidates =
             DoGroup(Candidates, InFile, Algo, RecordType, GroupMasterType);
	     break;
    case 6 : NumCandidates =
	     DoGlob(Candidates, InFile, Algo);
	     break;
    case 7 : GroupMasterType = BEAMFIT_REC;
             NumCandidates =
             DoGroup(Candidates, InFile, Algo, RecordType, GroupMasterType);
             break;
    default : printf("bad algorithm number; enter 1 through %d\n", MAX_ALGO);
	      exit(1);
    }
	     
  OutPutCandidates(Candidates, NumCandidates);
  return 0;
  }  // End main()

//___________________________________________________________________
long DoGroup(CandidateRec Candidates[],  
             FILE * InFile, 
             int Algo, 
             char RecordType,
             char GroupMasterType)
//___________________________________________________________________
  {
  s3_StdEventRec s3Event;
  StdEventRec    s4Event;
  void *         StdRecord;
  void *         pGroupMasterPtr;

  BOOLEAN bGroupMaster;

  MinEventRec Events[MAX_EVENTS_PER_CANDIDATE];

  long LineNum = 0;
  long GroupStart;
  long i, j, k;                         // loop control
  int Group = FALSE;

  i = j = k = 0;
  pGroupMasterPtr = NULL;

  bGroupMaster = GroupMasterType != ' ' ? TRUE : FALSE;

  if(RecordType == STDEVENT_REC)              // s3 or s4 ?
    StdRecord = (StdEventRec *) &s4Event;
  else
    StdRecord = (s3_StdEventRec *) &s3Event; 

  // Process until the end of the input data file
  while(!feof(InFile))
    {
    LineNum++;
    if(io_PeekNextRecord(InFile) == RecordType)   // event record?
      {                                           // yes!
      if(!Group)                                  // within a group?
	{                                         // no! start a group...
	GroupStart = LineNum;
	Group = TRUE;
	}
      io_GetNextRecord(InFile, StdRecord, RecordType);   // get the event
      cv_StdRecToMinRec(StdRecord, &Events[i], RecordType);   // convert to standard format
      i++;                                        // i = nbr of events in group
      }
    else                                          // event record? - no!
      {
      if(Group)                                   // were we in a group?
	{                                         // yes!
	qsort((char *) Events, (int)i, sizeof(MinEventRec), DblCompare);
						  // sort group by julian date
//        output_min(Events, i);                                 // for debugging
	Candidates[j].usAlgo          = Algo;
	Candidates[j].dbRa            = GetMeanRa(Events, i);
	Candidates[j].dbDec           = GetMeanDec(Events, i);
	Candidates[j].dbFreq          = GetMeanFreq(Events, i);
	Candidates[j].lNumDetect      = GetNumDetect(Events, i);
	Candidates[j].dbFreq          = GetMeanFreq(Events, i);
	Candidates[j].lFreqSep        = GetFreqSep(Events, i);
	Candidates[j].dbSkySep        = GetSkySep(Events, i);
	Candidates[j].lHiPower        = GetHiPwr(Events, i);
	Candidates[j].lLineNum        = GroupStart;
        GetMaxSpectra(Events, i, &Candidates[j].lMaxSpectra1, 
                      &Candidates[j].lMaxSpectra2);
        if(bGroupMaster)
          IncludeGroupMasterData(Candidates, j, GroupMasterType, pGroupMasterPtr);
	j++;                                      // j = nbr of groups
	i = 0;                                    // prep for
	Group = FALSE;                            //   new group
	}
      if(bGroupMaster && io_PeekNextRecord(InFile) == GroupMasterType)
        SaveGroupMaster(InFile, GroupMasterType, &pGroupMasterPtr);
      else
        io_SkipNextRecord(InFile); // not an event or group master record - skip it 
      }
    }
  return j;                             // return number of candidates
  }  // End DoGroup()

//___________________________________________________________________
long DoSingle(CandidateRec Candidates[], 
              FILE * InFile, 
              int Algo, 
              char RecordType)
//___________________________________________________________________
  {
  s3_StdEventRec s3Event;
  StdEventRec    s4Event;
  void *         StdRecord;

  MinEventRec Event;

  long LineNum = 0;
  long i       = 0;

  if(RecordType == STDEVENT_REC)              // s3 or s4 ?
    StdRecord = (StdEventRec *) &s4Event;
  else
    StdRecord = (s3_StdEventRec *) &s3Event;

  while(!feof(InFile))
    {
    LineNum++;
    if(io_PeekNextRecord(InFile) == RecordType)   // event record?
      {
      io_GetNextRecord(InFile, StdRecord, RecordType);   // get the event
      cv_StdRecToMinRec(StdRecord, &Event, RecordType);   // convert to standard format
      Candidates[i].usAlgo       = Algo;
      Candidates[i].dbRa         = Event.dbRa;
      Candidates[i].dbDec        = Event.dbDec;
      Candidates[i].dbFreq       = Event.dbFrequency;
      Candidates[i].lHiPower     = (long)Event.dbPower;
      Candidates[i].lMaxSpectra1 = 1;
      Candidates[i].lNumDetect   = 1;
      Candidates[i].lLineNum     = LineNum;
      i++;
      }
    else
      io_SkipNextRecord(InFile);             // not an event record - skip it
    }

  return i;
  }


//___________________________________________________________________
double GetMeanRa(MinEventRec Events[], long NumEvents)
//___________________________________________________________________
  {
  long i;
  double RaSum = 0.0;

  for(i = 0; i < NumEvents; i++)
    RaSum += Events[i].dbRa;

  return(RaSum / NumEvents);
  }

//___________________________________________________________________
double GetMeanDec(MinEventRec Events[], long NumEvents)
//___________________________________________________________________
  {
  long i;
  double DecSum = 0.0;
 
  for(i = 0; i < NumEvents; i++)
    DecSum += Events[i].dbDec;
 
  return(DecSum / NumEvents);
  }

//___________________________________________________________________
double GetMeanFreq(MinEventRec Events[], long NumEvents)
//___________________________________________________________________
  {
  long i;
  double FreqSum = 0.0;
 
  for(i = 0; i < NumEvents; i++)
    FreqSum += Events[i].dbFrequency;
 
  return(FreqSum / NumEvents);
  }

//___________________________________________________________________
long GetHiPwr(MinEventRec Events[], long NumEvents)
//___________________________________________________________________
  {
  long i;
  long HiP = 0;
 
  for(i = 0; i < NumEvents; i++)
    {
//    power = Events[i].pwr / (Events[i].actualthreshold / Events[i].sigmathreshold);
    if(Events[i].dbPower > HiP)
      HiP = (long)Events[i].dbPower;
    } 
 
  return(HiP);
  }

//___________________________________________________________________
void OutPutCandidates(CandidateRec Candidates[], long NumCandidates)
//___________________________________________________________________
  {
  long i; 

  for(i = 0; i < NumCandidates; i++)
    io_WriteCandidate(stdout, &Candidates[i]);
  }

//___________________________________________________________________
void GetMaxSpectra(MinEventRec Events[], 
                   long NumEvents, 
                   long * MaxSpectra1,
		   long * MaxSpectra2)
//___________________________________________________________________
  {
  long i, j, SpectraCount;
  double TempJuli, TempJulj;

  *MaxSpectra1 = *MaxSpectra2 = SpectraCount = 0;
  TempJuli = Events[0].dbJulDate;
  TempJulj = 0;
  j = 0;                            // init j now, as we do not init in j loop

  // Note that Events[] arrives here sorted in julian date order and that
  // we use julian date to separate observations and spectra within 
  // observations.  Logic:
  // March i up array of events.  As soon as i crosses to a new observation,
  // stop i and move j up array to i, counting spectra as we go.  When
  // j reaches i, stop j and place spectra count into MaxSpectra if 
  // spectra count is higher than current contents on MaxSpectra.  Then go
  // back to moving i.
  for(i = 0; i <= NumEvents; i++)     
    {                                
      // we can complete the current observation by either reaching the end
      // of this group OR by starting the next observation in the group. 
      if(i == NumEvents || fabs(Events[i].dbJulDate - TempJuli) > OBS_SEPARATION)
	{
        TempJuli = Events[i].dbJulDate;           // bring TempJuli up to current obs
        SpectraCount = 0;                         // re-init for each obs
        for(; j < i; j++ )     
          {
          if(Events[j].dbJulDate != TempJulj)     // new spectra?
            {                                     // yes! (will always be true on
            SpectraCount++;                       //  each 'first' time thru j loop)
            TempJulj = Events[j].dbJulDate;       
            }
          }                                       // End 'for j = j to i'
        if(SpectraCount > *MaxSpectra1)           // update MaxSpectra1 & 2
          {
          if(*MaxSpectra1 > *MaxSpectra2)
            *MaxSpectra2 = *MaxSpectra1;
          *MaxSpectra1 = SpectraCount;
          }
        else if(SpectraCount > *MaxSpectra2)
	    *MaxSpectra2 = SpectraCount;
        }                                         // End 'if new observation'
    }                                              // End 'for i = 0 to NumEvents'
  }  


//___________________________________________________________________
long GetNumDetect(MinEventRec Events[], long NumEvents)
//___________________________________________________________________
  {
  long i, j;
  long NumDetect = 0;

// for(i=0; i < NumEvents; i++)
//   printf("%lf\n", Events[i].dbJulDate);
// printf("\n");

  for(i = 0; i < NumEvents; i = j)
    {
    for(j = i+1; j < NumEvents; j++ )     
      if(fabs(Events[i].dbJulDate - Events[j].dbJulDate) > OBS_SEPARATION)
	{
// printf("%ld  %ld\n", i, j);
	NumDetect++;
	break;
	}
    }
    
  NumDetect++;                 // count last observation 
 
  return(NumDetect);
  }

//___________________________________________________________________
long GetFreqSep(MinEventRec Events[], long NumEvents)
//___________________________________________________________________
  {
  long i, j;
  double ThisDiff, BigDiff;

  ThisDiff = BigDiff = 0.0;

  for(i = 0; i < NumEvents; i++)
   for(j = i+1; j < NumEvents; j++)
      {  
      ThisDiff =  fabs(Events[i].dbFrequency - Events[j].dbFrequency); 
       if(ThisDiff > BigDiff)
	 BigDiff = ThisDiff;
      }  
  return((long)BigDiff);
  }

//___________________________________________________________________
double GetSkySep(MinEventRec Events[], long NumEvents)
//___________________________________________________________________
  {
  long i, j;
  double ThisSkyAngle, BigSkyAngle;

  ThisSkyAngle = BigSkyAngle = 0.0;

  for(i = 0; i < NumEvents; i++)
    for(j = i+1; j < NumEvents; j++)
      {  
      ThisSkyAngle = co_SkyAngle(Events[i].dbRa, Events[i].dbDec,
			       Events[j].dbRa, Events[j].dbDec);
      if(ThisSkyAngle > BigSkyAngle)
	BigSkyAngle = ThisSkyAngle;
      }  
   return(BigSkyAngle);
   }

//___________________________________________________________________
void InitCandidates(CandidateRec Candidates[], long NumCandidates)
//___________________________________________________________________
  {
  long i;

  for(i = 0; i < NumCandidates; i++)
    {
    Candidates[i].usAlgo                 = 0;
    Candidates[i].dbRa                   = 0.0;
    Candidates[i].dbDec                  = 0.0;
    Candidates[i].dbFreq                 = 0.0;
    Candidates[i].lHiPower               = 0;
    Candidates[i].dbSkySep               = 0.0;
    Candidates[i].lFreqSep               = 0;
    Candidates[i].lMaxSpectra1           = 0;
    Candidates[i].lMaxSpectra2           = 0;
    Candidates[i].lNumDetect             = 0;
    Candidates[i].lLineNum               = 0;
    Candidates[i].dbBeamFit              = 0.0;
    Candidates[i].dbStarDistance         = 0.0;
    strcpy(Candidates[i].stStarName, "-");
    strcpy(Candidates[i].stSpectralClass, "-");
    Candidates[i].lNumObs                = 0;
    Candidates[i].dbObsScore             = 0.0;
    }
  }

//___________________________________________________________________
// DblCompare is passed the addresses of MinEventRecs. 
// It is called by qsort, thus the void pointers.
// int DblCompare(void * a, void * b)
int DblCompare(const void * a, const void * b)
//___________________________________________________________________
  {
  MinEventRec * EventA, * EventB;

  EventA = (MinEventRec *) a;
  EventB = (MinEventRec *) b;

  if(EventA->dbJulDate < EventB->dbJulDate)
    return(-1);
  
  if(EventA->dbJulDate > EventB->dbJulDate)
    return (1);

  return (0);
  }

//___________________________________________________________________
// for debugging
void output_min(MinEventRec Events[], long i)
//___________________________________________________________________
  {
  long j;

  for(j = 0; j < i; j++)
   printf("%c %lf %lf %lf %lf %lf\n", Events[j].cFlag,
                                      Events[j].dbFrequency,
                                      Events[j].dbRa,
                                      Events[j].dbDec,
                                      Events[j].dbPower,
                                      Events[j].dbJulDate);
  }


//___________________________________________________________________
long DoGliese(CandidateRec Candidates[], FILE * InFile, int Algo)
//___________________________________________________________________
  {
  char StarBuff[100];
  long i, j;
  char Star[20];
  char Ra[20];
  char Dec[20];
  char Pxl[20];
  char Spc[20];

  i = j = 0;

  while(fgets(StarBuff, 100, InFile) != NULL)
    {
    strncpy(Star, StarBuff + 1, 10);
    Star[10] = NULL;
    for(j = 0; Star[j] != NULL; j++)
      if(Star[j] == ' ')
        Star[j] = '_';
    strncpy(Ra, StarBuff + 13, 6);
    Ra[6] = NULL;
    strncpy(Dec, StarBuff + 25, 7);
    Dec[6] = NULL;
    strncpy(Pxl, StarBuff + 33, 6);
    Pxl[6] = NULL;
    strncpy(Spc, StarBuff + 41, 5);
    Spc[5] = NULL;
    for(j = 0; Spc[j] != NULL; j++)
      if(Spc[j] == ' ')
        Spc[j] = '_';
    Candidates[i].usAlgo = Algo;
    strcpy(Candidates[i].stStarName, Star);
    sscanf(Ra, "%lf", &Candidates[i].dbRa);
    sscanf(Dec, "%lf", &Candidates[i].dbDec);
    sscanf(Pxl, "%lf", &Candidates[i].dbStarDistance);
    Candidates[i].dbStarDistance = Candidates[i].dbStarDistance == 0.0 ?
      0.0 : 1.0 / Candidates[i].dbStarDistance;
    strcpy(Candidates[i].stSpectralClass, Spc);
    i++;
    }

  return i;
  }



//___________________________________________________________________
long DoGlob(CandidateRec Candidates[], FILE * InFile, int Algo)
//___________________________________________________________________
  {
  char GlobBuff[100];
  long i, j;
  char tempstr[20];
  float x,y,z;
  char sign;
  long signfactor;
 
  i = j = 0;

  while(fgets(GlobBuff, 100, InFile) != NULL)
    {
    Candidates[i].usAlgo = Algo;

    strncpy(tempstr, GlobBuff + 2, 9);
    tempstr[9] = ' ';
    tempstr[10] = NULL;
    for(j = 0; tempstr[j] != NULL; j++)
      if(tempstr[j] == ' ')
        tempstr[j] = '_';
    strcpy(Candidates[i].stStarName, tempstr);

    strncpy(tempstr, GlobBuff + 27, 10);
    tempstr[10] = NULL;
    sscanf(tempstr, "%f%f%f", &x, &y, &z);
    Candidates[i].dbRa =  x + y/60 + z/3600;

    
    strncpy(tempstr, GlobBuff + 39, 9);
    tempstr[9] = NULL;
    sscanf(tempstr, "%c", &sign);
    if(sign == '+')
      signfactor = 1;
    else
      signfactor = -1;
    sscanf(tempstr + 1, "%f%f%f", &x, &y, &z);
    Candidates[i].dbDec =  (x + y/60 + z/3600) * signfactor;

    i++;
    }
 
  return i;
  }

//___________________________________________________________________
void SaveGroupMaster(FILE * InFile, 
                     char GroupMasterType, 
                     void ** pGroupMasterPtr)
//___________________________________________________________________
 {
 if(GroupMasterType == BEAMFIT_REC)
   {
   BeamfitRec * BeamfitRecPtr;
   BeamfitRecPtr = (BeamfitRec *)*pGroupMasterPtr;
   if(BeamfitRecPtr == NULL)
     if((BeamfitRecPtr = (BeamfitRec *)calloc(1, sizeof(BeamfitRec))) == NULL)
       {
       fprintf(stderr, "Cannot calloc BeamfitRec\n");
       exit(1);
       }
   io_GetNextRecord(InFile, BeamfitRecPtr, BEAMFIT_REC);
   *pGroupMasterPtr = (void *)BeamfitRecPtr;
   }
 else
   {
   fprintf(stderr, "Unknown Group Master Type\n"); 
   exit(1);
   }
 }

//___________________________________________________________________
void IncludeGroupMasterData(CandidateRec * Candidates, 
                            long CandidateIndex, 
                            char GroupMasterType,
                            void * pGroupMasterPtr)
//___________________________________________________________________
 {
 if(GroupMasterType == BEAMFIT_REC)
   {
   BeamfitRec * BeamfitRecPtr;
   BeamfitRecPtr = (BeamfitRec *)pGroupMasterPtr;
   Candidates[CandidateIndex].dbBeamFit = BeamfitRecPtr->dScore;
   }
 else
   {
   fprintf(stderr, "Unknown Group Master Type\n"); 
   exit(1);
   }
 }
