/* ________________________________________________________________________
 programmer : Jeff Cobb
 source name: ScoreCandidates.cc
 compiler   : AT&T C++ from SUN
___________________________________________________________________________
 modification history:
 date       : opened 3/1996
 2/4/98     jeffc    Lumped all candidate scores greater than 9999.999
                     into a score of 9999.999
 2/25/98    jeffc    Backed out previous change - all nonbarycentric candidates
                     ended up being > 9999.999.
___________________________________________________________________________
 
 Description:
 This program reads in SERENDIP candidate records (ie record type 'C')
 and outputs those same records with 2 additonal fields filled in.
 Both of these fields are related to the number of times that SERENDIP
 has observed the area of sky where the candidate is located.  This
 data comes from a skyplot file, which is also read in.
 The 2 fields are :
   lNumObs    : the number of times we have looked within beamres beams
                of the candidate.  beamres in an input parm - see below.
   dbObsScore : this is a rating score based on the poisson probability
                that, given lNumObs, noise events would produce the 
                candidate.  Thus, the *lower* the score, the better.
 
 Input parms (all are required):
   -b (double)         : beamres, the radius in beams of the area of sky we are
                         considering.
   -vt (integer)       : the time, in seconds, that a point source can be within
                         a single beam.
   -t (floating point) : detection threshold of the serendip instrument 
                         when the candidate observations were made.
   -sf (integer)       : a "softening factor".  See "Theory" below.
   -f (floating point) : the frequency window used to determine candidates
                         in the first place.
 
 Required command line file names:
   candidate file
   skyplot file
 Output is to stdout.

 Theory: 
 We keep a file, known as "skyplot" that divides the sky into small
 "pixels" (each pixel is an element in a 2D array).  Each pixel holds
 a count of how many times we have looked at that pixel.  See docs on
 program skyplot for details.  The number of times that we have looked at 
 a particular point in the sky is a useful thing to know in accessing
 whether or not a given candidate is significant.  For example it would
 be much more exciting to see a candidate in 2 out 3 observations than
 in 2 out of 30 observations.  Scoring candidates based on how many times
 we see them vs how many times we've looked is the task of this program.
 There are 2 sub-tasks: counting observations and scoring the candidates.

 Counting the Observations:
 A candidate is defined by a point on the sky.  We would like to know how
 many times we have looked within a circle of some radius from that point.
 This radius is given in units of beamwidth and is usually 1.  Now, a 
 candidate point can occur anywhere in a pixel.  If it occurs near an
 edge or a corner, then the counts in neighboring pixels should be taken
 into account.  We do this by the following algorithm:
   1) determine which skyplot pixel that a candidate lies in.
   2) construct a grid around and including this pixel.  This grid
      will usually consist of 9 pixels but can consist of 6 if the
      candidate pixel is on a declination edge.
   3) calculate the distance from the candidate point to the center
      of each pixel in the grid.
   4) use the distance to determine how much weight to give the count
      in that pixel, using an exponential dropoff in weight over distance.
      This dropoff corresponds to the decreasing likihood that an event
      at an increasing distance would be strong enough to be counted
      when the telescope was pointing at the candidate point.
      The shape of the dropoff is determined by the radius of the circle
      within which we are counting.
   5) We sum the weighted counts and divide by the sum of the weights to
      get our final count.  (if all 9 pixels were weighted equally, then
      all eights would be 1 and the sum of the weights would be 9, giving
      us a simple mean)

 Scoring the Candidate:
 The basic idea here is to give the candidate a score based on the poisson
 probability that, given the number of times we have "looked", the number
 of times we have "seen" can be accounted for by noise.  The lower the
 score, the better.  Here, the poisson space is one dimensional - it 
 consists of frequency only (the 2 dimensions of sky space are not
 taken into account because we are looking at one spot*).  Much discussion
 was generated within the serendip group on how to determine the number
 of frequency bins - should we use the frequency separation of each individual
 candidate or the frequency window og the algorithm that generated the 
 the candidate?  If the latter, should frequency separation have an effect?
 Here is what we decided:
 1) Freq. separation *should* affect the score - the tighter the candidate
    in frequency, the better the score.  But...
 2) Freq. separation should not be directly equated to freq. bin size.  This
    is because:
    a) for nonbarycentric candidates, a small freq. separation does not equate
       with "goodness".  We expect the signal to drift significantly as
       the transmitter rotates and orbits.  Freq. separation here tells us
       something about physical characteristics of the source, but not
       (with the huge window we specify) whether 2 signals do in fact
       come from the same source.
   b) for barycentric candidates, we do interpret freq. closeness as
      "more likely related", but there is a built in frequency uncertainty 
      as a side effect of position uncertainty.  We don't what freq.
      closeness beyond our uncertainty limit to have much effect.
   so...
3) In determining number of frequency bins for candidate scoring, I will
   equate bin size to the window size used to determine candidacy in the
   first place (120 Hz for barycentric and 50KHz for nonbarycentric).
   This will be a required command line param.  To let frequency separation
   affect the score, I will multiply the poisson probability by:
   
   freq. separation + SofteningFactor
   ----------------------------------
          SofteningFactor    

   The SofteningFactor will also be a required command line param.  Dan
   suggests using 30Hz for barycentric candidates, as that is the typical
   error from position uncertainty.  Thus, if the separation is 0, the
   probabilty stays the same, if separation is 30 Hz, the probability is
   doubles, a separation of 60 Hz triples the probabilty and so on. 
   For nonbarycentric candidates, we can make the SofteningFactor much
   larger.

Bugs:

With this method, a given frequency separation will affect a triplet
the same as it would a doublet.   That is, tightness with three detections
is not worth more than the same tightness with 2.   

There is some AO specific code - see warnings in code comments.

Notes:

* Perhaps a more accurate probability could be had by multiplying
  the probability as caculated here (ie the probability within a
  single beam) by the total number of beams in the sky.  After
  all, if we are looking at thousands of beams, we are essentially
  repeating the experiment that many times.  What is unlikely in
  any 1 beam may be likely in 1 out of many beams.  The problem is that
  we are not actually repeating the same experiment at each beam
  because the number of observations at each beam is not uniform
  across the sky.  Because of this "lumpiness" in number-of-observations
  we treat each beam independently.  We thus introduce an inaccuracy into
  the figure of probability that we just need to keep in mind.  The
  ranking of candidates is of course not effected by this. 
_______________________________________________________________________ */


//#define DEBUG_OPS
//#define DEBUG_OBS_PIXELGRID
//#define DEBUG_OBS
//#define DEBUG_SCORE

#include <stdio.h>
#include <string.h>  
#include <math.h>
// memory.h commented out because gcc complains of declaration conlict
// of function memchr() with header cstring.h
// #include <memory.h>
#include "SIREN.h" 
#include "s4io.h" 
#include "s4coord.h" 

//------------------------------------------------------------------------
void  GetOps       (int      argc,
                    char   * argv[],
                    double * dbNumBeams,
                    long   * lTimeVisable,
                    double * dbThreshold,
                    long   * lFreqWindow,
                    long   * lSofteningFactor,
                    char  ** CandidateFileName, 
                    char  ** SkyPlotFileName);
void CmdError      (void);
long GetNumObs     (unsigned char ucSkyMap[SKYPLOT_HEIGHT][SKYPLOT_WIDTH], 
                    double dbBwSq,
                    double dbRa, 
                    double dbDec);
double GetObsScore (long   lNumObs,
                    long   lNumDetect,
                    long   lFreqSep,
                    long   lSofteningFactor,
                    double dbNumFreqBins,
                    double dbEventsPerOb);

char *usage = "Usage : ScoreCandidates : -b beamres -v time_visable  -t threshold \
-f frequency_window -s softening_factor skyplotfile candidatefile > outputFile\n";

//------------------------------------------------------------------------
int main(int argc, char *argv[])
//------------------------------------------------------------------------
  {
  unsigned char ucSkyMap[SKYPLOT_HEIGHT][SKYPLOT_WIDTH],
                ucIncoming;

  long i, 
       lLookUpRa, 
       lLookUpDec, 
       lTotalSize,
       lNumCandidates,
       lTimeVisable,
       lFreqWindow,
       lSofteningFactor;

  double dbDummy,
         dbNumBeams,
         dbBeamWidth,
         dbBwSq,
         dbThreshold,
         dbNumFreqBins,
         dbEventsPerOb;

  char * sSkyPlotFileName;
  char * sCandidateFileName;
  FILE * fSkyPlotFile;
  FILE * fCandidateFile;

  CandidateRec Candidate;

//------------------------------------------------------------------------
  GetOps(argc,
         argv,
         &dbNumBeams,
         &lTimeVisable,
         &dbThreshold,
         &lFreqWindow,
         &lSofteningFactor,
         &sCandidateFileName,
         &sSkyPlotFileName); 

#ifdef DEBUG_OPS
fprintf(stderr, "dbNumBeams = %lf\n", dbNumBeams);
fprintf(stderr, "dbThreshold = %lf\n", dbThreshold);
fprintf(stderr, "lTimeVisable = %ld\n", lTimeVisable);
fprintf(stderr, "lFreqWindow = %ld\n", lFreqWindow);
fprintf(stderr, "lSofteningFactor = %ld\n", lSofteningFactor);
fprintf(stderr, "sCandidateFileName = %s\n", sCandidateFileName);
fprintf(stderr, "sSkyPlotFileName = %s\n", sSkyPlotFileName);
//exit(0);
#endif

//------------------------------------------------------------------------

  // WARNING - the following is AO SPECIFIC!!!!! - should be
  // changed to use cfg file.
  co_GetBeamRes(&dbBeamWidth, &dbDummy, 430, 1000, dbNumBeams);
  dbBwSq      = dbBeamWidth * dbBeamWidth;

  // Determine number of frequency bins 
  // WARNING - the following is AO SPECIFIC!!!!! 
  dbNumFreqBins   = NUM_FQ_BINS / lFreqWindow;

  lTotalSize = SKYPLOT_HEIGHT * SKYPLOT_WIDTH;

  // WARNING - the following is AO SPECIFIC!!!!!
  dbEventsPerOb = NUM_FFT_BINS        *
                  exp(-dbThreshold)   *
                  (lTimeVisable / INTEGRATION_TIME);

//------------------------------------------------------------------------
  if ((fCandidateFile = fopen(sCandidateFileName,"r")) == NULL)
    {
    fprintf(stderr,"Input file %s not found\n", argv[2]);
    fprintf(stderr, "%s\n", usage); 
    exit(1);
    }

  /* initializing work area to all zeros */
  memset((char*) ucSkyMap, 0, SKYPLOT_WIDTH * SKYPLOT_HEIGHT * sizeof(char));
  
  /*read skyplot file into array ucSkyMap[][] */
    if ((fSkyPlotFile = fopen(sSkyPlotFileName,"r")) != NULL)
    {
      for (i = 0, ucIncoming = 0; i < lTotalSize; i++)
      {
        fread((char*)&ucIncoming, sizeof(char), 1, fSkyPlotFile);
        if (ucIncoming != 0)
        {
          lLookUpDec = i/SKYPLOT_WIDTH;
          lLookUpRa = i%SKYPLOT_WIDTH;
          ucSkyMap[lLookUpDec][lLookUpRa] = ucIncoming;
        }
      }
      fclose(fSkyPlotFile);
    }
    else
    {
      fprintf(stderr,"Input file %s not found\n", argv[1]);
      fprintf(stderr, "%s\n", usage); 
      exit(1);
    }

  // Get number of candidates
  if (!(lNumCandidates = io_CountRecords(fCandidateFile, CANDIDATE_REC)))
    {
    fprintf(stderr, "No candidate records in input file %s.\n", argv[2]);
    exit(1);
    }

//------------------------------------------------------------------------
  // Add skyplot count to each candidate record and score candidate
  // based on this number
  for(i=0; i < lNumCandidates; i++)
    { 
     if (! io_GetNextRecord(fCandidateFile, &Candidate, CANDIDATE_REC))
       {
       printf("%s : error in reading candidate file %s\n", argv[0], argv[3]);
       exit(1);
       }
    Candidate.lNumObs    = GetNumObs(ucSkyMap, 
                                     dbBwSq, 
                                     Candidate.dbRa, 
                                     Candidate.dbDec);
    Candidate.dbObsScore = GetObsScore(Candidate.lNumObs, 
                                       Candidate.lNumDetect, 
                                       Candidate.lFreqSep,
                                       lSofteningFactor,
                                       dbNumFreqBins,
                                       dbEventsPerOb);
    io_WriteCandidate(stdout, &Candidate);
    }

  return 0;
  }  

//------------------------------------------------------------------------
// Count the observations
//
long GetNumObs(unsigned char ucSkyMap[SKYPLOT_HEIGHT][SKYPLOT_WIDTH], 
               double dbBwSq,
               double dbRa, 
               double dbDec)
//------------------------------------------------------------------------
  {
  long i,
       j,
       lRaRange,
       lDecRange,
       lLookUpRa,
       lLookUpDec,
       lRaIndex[3],
       lDecIndex[3];

  double dbPixelCenterRa,
         dbPixelCenterDec,
         dbDistance,
         dbWeight,
         dbTotalObs, 
         dbSumWeights; 


   dbTotalObs   = 0.0;
   dbSumWeights = 0.0;

   //---------------------------------------------------
   // Determine the grid of pixels centered on the pixel
   // containing the candidate.
   lLookUpRa  = SKYPLOT_CONVERT_RA(dbRa);          // convert sky locations into
   lLookUpDec = SKYPLOT_CONVERT_DEC(dbDec);        // corresponding cells of skymap

   lRaIndex[0] = lLookUpRa - 1;
   lRaIndex[1] = lLookUpRa;
   lRaIndex[2] = lLookUpRa + 1;

   lDecIndex[0] = lLookUpDec - 1;
   lDecIndex[1] = lLookUpDec;
   lDecIndex[2] = lLookUpDec + 1;

   if (lLookUpRa == 0)
     lRaIndex[0] = SKYPLOT_WIDTH - 1;
   else
     if(lLookUpRa == SKYPLOT_WIDTH - 1)
       lRaIndex[2] = 0;

   lRaRange = 3;
   lDecRange = 3;

   if (lLookUpDec == 0)
     {
     lDecIndex[0] = 0;
     lDecIndex[1] = 1;
     lDecRange = 2;
     }
   else
     if(lLookUpDec == SKYPLOT_HEIGHT - 1)
       lDecRange = 2;
   //---------------------------------------------------
#ifdef DEBUG_OBS_PIXELGRID
fprintf(stderr, " RaRange = %ld  DecRange = %ld  RA = %ld %ld %ld DEC = %ld %ld %ld\n",
        lRaRange, lDecRange, lRaIndex[0], lRaIndex[1], lRaIndex[2],
        lDecIndex[0], lDecIndex[1], lDecIndex[2]);
#endif

   //---------------------------------------------------
   // March through the grid of pixels, centered on pixel containing
   // the candidate.  Use distance from the candidate *point* to
   // the center of each pixel to weight that pixel's contribution
   // to the total count.
   for(i = 0; i < lDecRange; i++)
     for(j = 0; j < lRaRange; j++)
       {
       dbPixelCenterRa  = SKYPLOT_R_CONVERT_RA(lRaIndex[j]);
       dbPixelCenterDec = SKYPLOT_R_CONVERT_DEC(lDecIndex[i]);
       dbDistance       = co_SkyAngle(dbRa, dbDec, dbPixelCenterRa, dbPixelCenterDec); 
       dbWeight         = exp(- (pow(dbDistance, 2.0) / dbBwSq));
       dbTotalObs      += ucSkyMap[lDecIndex[i]][lRaIndex[j]] * dbWeight; 
       dbSumWeights    += dbWeight; 
#ifdef DEBUG_OBS
fprintf(stderr, "RaRange          = %ld\n", lRaRange);
fprintf(stderr, "DecRange         = %ld\n", lDecRange);
fprintf(stderr, "RA               = %ld %ld %ld\n", lRaIndex[0], lRaIndex[1], lRaIndex[2]);
fprintf(stderr, "DEC              = %ld %ld %ld\n", lDecIndex[0], lDecIndex[1], lDecIndex[2]);
fprintf(stderr, "dbPixelCenterRa  = %lf\n", dbPixelCenterRa);
fprintf(stderr, "dbPixelCenterDec = %lf\n", dbPixelCenterDec);
fprintf(stderr, "dbDistance       = %lf\n", dbDistance);
fprintf(stderr, "dbWeight         = %lf\n", dbWeight);
fprintf(stderr, "dbTotalObs       = %lf\n", dbTotalObs);
fprintf(stderr, "dbSumWeights     = %lf\n", dbSumWeights);
#endif
       }
   //---------------------------------------------------

   dbTotalObs /= dbSumWeights;   // normalize by the sum of the weights
   return(nint(dbTotalObs));     // round to integer
  }

//------------------------------------------------------------------------
// Score the candidate
//
double GetObsScore(long   lNumObs, 
                   long   lNumDetect, 
                   long   lFreqSep, 
                   long   lSofteningFactor,
                   double dbNumFreqBins,
                   double dbEventsPerOb)
//------------------------------------------------------------------------
  {
  double dbNumEvents,
         dbFreqSeparationFactor,
         dbPoissonProb;

  long i,
       lNumDetectFactorial;

  // make a good guess as to the number of events at
  // this frequency range
  dbNumEvents = dbEventsPerOb * lNumObs;

  lNumDetectFactorial = 1;
  for(i = 1; i <= lNumDetect; i++)
    lNumDetectFactorial *= i;

  dbPoissonProb = pow(dbNumEvents, (double)lNumDetect) /
                  (lNumDetectFactorial * pow(dbNumFreqBins, (double)lNumDetect-1));

  dbFreqSeparationFactor = (double)(lFreqSep + lSofteningFactor) /  lSofteningFactor;

#ifdef DEBUG_SCORE
fprintf(stderr, "dbEventsPerOb                = %lf\n", dbEventsPerOb);
fprintf(stderr, "dbNumEvents                  = %lf\n", dbNumEvents);
fprintf(stderr, "dbNumFreqBins                = %lf\n", dbNumFreqBins);
fprintf(stderr, "lNumDetect                   = %ld\n", lNumDetect);
fprintf(stderr, "lNumDetectFactorial          = %ld\n", lNumDetectFactorial);
fprintf(stderr, "dbPoissonProb                = %lf\n", dbPoissonProb);
fprintf(stderr, "lFreqSep                     = %ld\n", lFreqSep);
fprintf(stderr, "lSofteningFactor             = %ld\n", lSofteningFactor);
fprintf(stderr, "dbFreqSeparationFactor       = %lf\n", dbFreqSeparationFactor);
#endif

  // we multiply by 1000 just to make it nicer to print (not so many zeros to the right
  // of the decimal point)
  return(dbPoissonProb * dbFreqSeparationFactor * 1000); 
  }

//------------------------------------------------------------------------
// Command line operands
//
void  GetOps(int      argc,
             char   * argv[],
             double * dbNumBeams,
             long   * lTimeVisable,
             double * dbThreshold,
             long   * lFreqWindow,
             long   * lSofteningFactor,
             char  ** CandidateFileName, 
             char  ** SkyPlotFileName) 
//------------------------------------------------------------------------
  {
  long i;

  if(argc != 13)
    CmdError();

  // Get parameters.
  // Skip the program name and stop short of the file names.
  for (i = 1; i < (argc - 3); i++)
    if (strcmp(argv[i],"-b") == 0)
      {
      if (sscanf(argv[++i],"%lf", dbNumBeams) != 1)
        CmdError();
      }
    else if (strcmp(argv[i],"-v") == 0)
      {
      if (sscanf(argv[++i],"%ld", lTimeVisable) != 1)
        CmdError();
      }
    else if (strcmp(argv[i],"-t") == 0)
      {
      if (sscanf(argv[++i],"%lf", dbThreshold) != 1)
        CmdError();
      }
    else if (strcmp(argv[i],"-f") == 0)
      {
      if (sscanf(argv[++i],"%ld", lFreqWindow) != 1)
        CmdError();
      }
    else if (strcmp(argv[i],"-s") == 0)
      {
      if (sscanf(argv[++i],"%ld", lSofteningFactor) != 1)
        CmdError();
      }
    else
      CmdError();

  // Now get the file names
  *SkyPlotFileName = argv[i++];
  *CandidateFileName = argv[i];
  }

//------------------------------------------------------------------------
void CmdError(void)
//------------------------------------------------------------------------
  {
  fprintf(stderr, usage);
  exit(1);
  }
      
  

