//________________________________________________________________________
// programmer : Brendan Doyle
// source name: s4converts3A.cc
// compiler   : AT&T C++ from SUN
// date       : opened 1/1995
// The purpose of this program is to convert data in the SERENDIP III 
// offline format into SERENDIP IV type 'A' records for use with the s4 
// data analysis programs.  The program reads a file of SERENDIP III 
// off-line data from stdin one record at a time, edits it to create a 
// SERENDIP IV offline record, and writes it out to stdout with the 
// following additions and modifications:
//  - record type: A
//  - the tenth-of-second added to the time field is set to 0, unless 
//    the -t parameter is used, in which case a fictitious pseudorandom 
//    tenth-of-second is generated for each event for testing purposes 
//    (events with the same time are assigned the same tenth-of-second 
//    digit) 
//  - synthesizer frequency is used as filter board frequency;
//    fft board frequency is assigned a value of 0
//  - board ID constant: 0
//  - polarization constant: L
//  - telescope ID constant (for Arecibo-430): 00
//________________________________________________________________________
// modification history:
//________________________________________________________________________


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "SIREN.h"
#include "s4io.h"
#include "s4time.h"

#define InLineOldS3Format "%s %lf %lf %lf %ld %ld %hu %hu %hu %s %hx %hx"
#define BOARD_ID 0
#define POLARIZATION 'L'
#define TELESCOPE_ID 00
#define RAND_MAX 2147483647 

void  ExitError(char errorMessage[], char progname[]);


int  main(int argc, char *argv[])
  {
  StdEventRec ThisEvent, *ThisEventPtr;

  FILE        *fin_p;         // input file
  FILE        *fout_p;        // output file

  char thisRecord[MAX_LINE_LENGTH];  // input record
  long i;                     // loop counter
  long hund_sec = 0;          // hundreth of second (not present in s3 data)
  long hundflag = 0;          // 0: hund_sec = 0 (default);
                              // 1: hund_sec = pseudorandom 1/10 of second
  char save_time[7];          // save area for previous event time comparison
  char *strptr;               // string pointer for converting time
  char temp_st_time[7];       // work area for time string
  char hour_str[3];           // work areas for converting time to integer
  char min_str[3];
  char sec_str[3];
  long hour_i, min_i, sec_i;  // work areas for integer time
  char temp_st_date[7];       // work area for date string
  long yday_i;                // work area for day of year
  struct DateStruct sdate;    // work area for integer date
  char temp_mon[3];           // work area for alpha month
  long s3Synth;               // Old s3 synth setting


  fin_p  = stdin;   // default input stream is stdin
  fout_p = stdout;  // default output stream is stdout

  srand(1);                       // seed random number generator
  strcpy(save_time, "       ");   // initialize time save field

  if (argc > 2)
    ExitError("", argv[0]);

  // get optional parameter
  for (i = 1; i < argc; i++)
    if (! strcmp(argv[i], "-t"))
      hundflag = 1;
    else
        ExitError("Unknown flag", argv[0]);

  ThisEventPtr = &ThisEvent;

  ThisEvent.ulFftBdFreq = 0;               // set FFT board freq to 0
  ThisEvent.cPolarization = POLARIZATION;  // set polarization to default
  ThisEvent.ucBid = BOARD_ID;              // set board id to default
  ThisEvent.ucTid = TELESCOPE_ID;          // set telescope id to default

  // process input s3 records
  while (gets(thisRecord))
    // determine if data record or comment/blank line
    if (sscanf(thisRecord, InLineOldS3Format,
               temp_st_time, &ThisEvent.dbRa, &ThisEvent.dbDec,
               &ThisEvent.dbFirstLo, &s3Synth,
               &ThisEvent.lBinNumber, &ThisEvent.usPower,
               &ThisEvent.usActualThreshold, &ThisEvent.usAdc, temp_st_date,
               &ThisEvent.usSigmaThreshold, &ThisEvent.usSpctResLog) == 12)
      {
      // if data record, construct type 'A' record
      
      // s3 synth was in KHz, convert to s3 second LO in MHz
      ThisEvent.dbSecondLo = s3Synth * 1000.0;

      // Generate hundredth of a second
      if (hundflag && strcmp(temp_st_time, save_time))
        {
        // if -t parameter used, generate hundreth-of-second digit for each event;
        // events with same time must have same hundreth-of-second digits!
        hund_sec = (long) ((rand() / ((double) RAND_MAX + 1.0)) * 100.0);
        strcpy(save_time, temp_st_time);
        }

      // convert date/time to UT
      strptr = temp_st_time;
      memcpy(hour_str, strptr, 2);
      hour_str[2] = '\0';
      memcpy(min_str, (strptr += 2), 2);
      min_str[2] = '\0';
      memcpy(sec_str, (strptr += 2), 2);
      sec_str[2] = '\0';
      hour_i = atol(hour_str);
      min_i = atol(min_str);
      sec_i = atol(sec_str);
      tm_DateToI(temp_st_date, &sdate.day, &sdate.month, &sdate.year);
      yday_i = tm_DayMonToYrDay(sdate.day, sdate.month, sdate.year);
      tm_AstToUt(&sdate.year, &yday_i, &hour_i);
      tm_DayToMon(yday_i, sdate.year, &sdate);
      sdate.day--;

      // convert month back to alpha
      switch(sdate.month) {
        case  1:
          strcpy(temp_mon, "ja");
          break;
        case  2:
          strcpy(temp_mon, "fe");
          break;
        case  3:
          strcpy(temp_mon, "mr");
          break;
        case  4:
          strcpy(temp_mon, "ap");
          break;
        case  5:
          strcpy(temp_mon, "my");
          break;
        case  6:
          strcpy(temp_mon, "jn");
          break;
        case  7:
          strcpy(temp_mon, "jl");
          break;
        case  8:
          strcpy(temp_mon, "au");
          break;
        case  9:
          strcpy(temp_mon, "se");
          break;
        case 10:
          strcpy(temp_mon, "oc");
          break;
        case 11:
          strcpy(temp_mon, "no");
          break;
        case 12:
          strcpy(temp_mon, "de");
          break;
        default:
          strcpy(temp_mon, "  ");
        }

      // write type 'A' record
      fprintf(fout_p, "%c %.2ld%.2ld%.2ld%.1ld %7.4f %7.3f %12.7f %6ld %6ld \
%10ld %2x %5u %3d %5u %c %2u %x %.2ld%2.2s%.2ld %02x\n",
              STDEVENT_REC, hour_i, min_i, sec_i, hund_sec, ThisEvent.dbRa,
              ThisEvent.dbDec, ThisEvent.dbFirstLo, ThisEvent.dbSecondLo,
              ThisEvent.ulFftBdFreq, ThisEvent.lBinNumber, 0.0, ThisEvent.ucBid,
              ThisEvent.usPower, ThisEvent.usSigmaThreshold,
              ThisEvent.usActualThreshold, ThisEvent.cPolarization,
              ThisEvent.usAdc, ThisEvent.usSpctResLog,
              sdate.day, temp_mon, sdate.year, ThisEvent.ucTid);
      }
    else
      // if comment or blank line, write type '#' record
      printf("# %s\n", thisRecord);
  }


void  ExitError(char errorMessage[], char progname[])
  {
  char  *usage = " [-t]";

  if (strcmp(errorMessage, ""))
    fprintf(stderr, "\n%s\n", errorMessage);
  fprintf(stderr, "\nUsage:\n%s%s\n\n", progname, usage);
  exit(1);
  }
