/*
Programmer : Jeff Cobb
Source name: s4NetGet.c
Compiler   : gcc 
 
Usage : 
 
Description: This is the "server" in the s4 internet
             client/server data acquisition scheme.

Programming notes:

Error handling:

Bind error - exit.

The socket error checking, check pointing and recovery
is simple and coarse.  This program checks for 2 "error"
conditions on the socket read:
1) a return of less than the amount of data requested. This 
   indicates that the socket was closed on the other end, 
   gracefully or not.  In this case, this program closes
   it's socket and output file, starts another socket and 
   waits for a connection.
2) A return of < 0.  This indicates an error in the read.
   In this case, the program closes it's socket and output file
   and exits.
This program is meant to be wrapped in a management script that
(among other things) will sleep for a short period and and then
restart the program when it exits.  All checkpoint/restart logic
lives in the management scripts.


*/

/*
#define DEBUG
*/

#define SERENDIP4

#include	<unistd.h>
#include	<stdlib.h>
#include	<stdio.h>
#include        <signal.h>
#include        <strings.h>
#include        <sys/types.h>
#include        <sys/timeb.h>
#include        <sys/socket.h>
#include        <arpa/inet.h>
#include        <fcntl.h>
#include        <errno.h>
#include	"s4Net.h"

/* signal responders */
#define If_SIGHUP_THEN_BREAK if(SigHupFlag) break

/* Internal function prototypes */
long AuthConnect(struct sockaddr_in * cli_addr);
void AuthError(void);
void SocketClosedError(int readrtn);
void SocketReadError(int readrtn);
void sighup_func(int i);
void myperror(char * str);
void print_time(void);
void sigio_func(void);

/* global variables */
int SigHupFlag;
double gMyTime = 0.0;             /* for net timing tests */
double gInterval;                 /* for net timing tests */

/* ------------------------------------------------------------- */
main(int argc, char *argv[])
/* ------------------------------------------------------------- */
  {
  int sockfd,
      newsockfd, 
      readrtn;
  
  socklen_t clilen;

  char * pname;
      
  struct sockaddr_in cli_addr,  
                     serv_addr;

  unsigned char BufPlusPrefix[NET_BUFSIZE + SYNC_WORD_LEN-1];
  unsigned char * Buf;

  Buf = &BufPlusPrefix[SYNC_WORD_LEN-1];

  setbuf(stdout, NULL);                    /* unbuffered output */

  clilen = sizeof(cli_addr);

  SigHupFlag = 0;

  pname = argv[0];

  signal(SIGHUP, sighup_func);

  /* ------------------------------------------------------ */
	/*
	 * Open a TCP socket (an Internet stream socket).
	 */

  if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    myperror("server: can't open stream socket");


	/*
	 * Bind our local address so that the client can send to us.
	 */

  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family      = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port        = htons(SERV_TCP_PORT);

  if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
    myperror("server: can't bind local address");

  listen(sockfd, 5);

  /* ------------------------------------------------------ */
  for ( ; ; ) 
    {
    /* wait on and accept a new socket connection from client */
    newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);

    If_SIGHUP_THEN_BREAK;     /* allow HUP signal to cause graceful exit */

    /* Allow connection from only authorized clent */
    if(!AuthConnect(&cli_addr))
      AuthError();

    /* Good connection? */
    if (newsockfd < 0)
      myperror("server: accept error");    /* no - error */

  /* ------------------------------------------------------ */
    /* read from socket forever */
    while(1)               
      {
      If_SIGHUP_THEN_BREAK;
      if((readrtn = readn(newsockfd, (char *)Buf, NET_BUFSIZE)) > 0)
        fwrite(Buf, 1, readrtn, stdout);
      else
        SocketReadError(readrtn);

      if(readrtn < NET_BUFSIZE)    /* Socket closed on the other end? */
        {                          /* We assume so if readn() returns */
        SocketClosedError(readrtn);       /* without filling the buffer.     */
        break;
        }
      } /* end read forever */

    /* connection was closed - clean up */
    close(newsockfd);      /* done with this connection - close it */

    }  /* end for(;;) - go wait for another connection */

  } /* end main() */


/* ------------------------------------------------------------- */
long AuthConnect(struct sockaddr_in * cli_addr)
/* ------------------------------------------------------------- */
// always authorize until we have the AO routing figured out.
// jeffc 8/8/2002
  {
fprintf(stderr," we got %ud (%s)  we want %ud (%s)\n", 
		cli_addr->sin_addr.s_addr, 
		inet_ntoa(cli_addr->sin_addr),
		inet_addr(CLIENT_HOST_ADDR),
		CLIENT_HOST_ADDR);
  if(cli_addr->sin_addr.s_addr == inet_addr(CLIENT_HOST_ADDR))
    return(1);   /* Authorized */
  else
    return(1);   /* Authorized */
    //return(0);   /* NOT Authorized */
  }

/* ------------------------------------------------------------- */
void AuthError()
/* ------------------------------------------------------------- */
  {
  fprintf(stderr, "Authentication Error  My PID is %d\n", getpid());
  exit(1);
  }

/* ------------------------------------------------------------- */
void SocketReadError(int readrtn)
/* ------------------------------------------------------------- */
  {
  fprintf(stderr, "Read error on socket.  Read return was %d\n", readrtn);
  exit(1);
  }

/* ------------------------------------------------------------- */
void SocketClosedError(int readrtn)
/* ------------------------------------------------------------- */
  {
  fprintf(stderr, "Socket closed.  Read return was %d. My PID is %d\n", readrtn, getpid());
  exit(1);
  }


/* ------------------------------------------------------------- */
void myperror(char * str)
/* ------------------------------------------------------------- */
  {
  perror(str);
  exit(1);
  }

/* ------------------------------------------------------------- */
void sighup_func(int i)
/* ------------------------------------------------------------- */
  {
  SigHupFlag = 1;
  }

/* ------------------------------------------------------------- */
void print_time()
/* ------------------------------------------------------------- */

  {
  double mytime;

  struct timeb mytimeb;

  ftime(&mytimeb);

  mytime = mytimeb.time;
  mytime += mytimeb.millitm / 1000.0;

/*
  fprintf(stderr, "time = %lf\n", mytime - gMyTime);
*/
  gInterval = mytime - gMyTime;
  gMyTime = mytime;

  }

