/*
Programmer : Jeff Cobb
Source name: s4NetPut.c
Compiler   : gcc

Usage : 

Description: This is the "client" in the s4 internet
             client/server data acquisition scheme.

Error handling: 
The socket error checking, check pointing and recovery
is simple and coarse.  This program checks for 1 "error"
condition on the socket write: 
1) A return of less than the amount of data requested to transmit. 
   This indicates an error in the write (maybe the other end
   closed the connection).  In this case, the program closes it's 
   socket and exits.
This program is meant to be wrapped in a management script that
(among other things) will sleep for a short period and and then 
restart the program when it exits.  All checkpoint/restart logic 
lives in the management scripts.  Checkpointing is handled at
the file level, ie in case of any problem, the transmission of
the current file is restarted at the beginning of the file.
We want the program to exit as quickly as possible on socket
closure, so SO_LINGER is *not* set (we will be re-tranmitting the
data anyway).


*/

/*
#define DEBUG
*/

#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include "s4Net.h"

#define TIME_MARK  0
#define TIME_CHECK 1
//#define NO_SEND_TIME 10.0
#define NO_SEND_TIME 0.0

/* Internal function prototypes */
double TimeWait(long action);
void myperror(char * str);
void sigio_func();

/* global variables */
int sigflag;

/* ------------------------------------------------------------- */
int main(int argc, char *argv[])
/* ------------------------------------------------------------- */
  {
  int			sockfd;       /* socket "file" descriptor */
  struct sockaddr_in	serv_addr;    /* address of server */

  long i, NumData, NumToWrite;

  struct stat statbuf;
  off_t save_size;

  FILE *fp;

  char buf[NET_BUFSIZE];             /* buffer of data to send */

  /* ------ NOTE: will only work using stdin (SIGIO)  -------- */
    fp = stdin;
    fstat(fileno(fp), &statbuf);
    save_size = statbuf.st_size;

  /* ------------------------------------------------------ */

  /* Fill in the structure "serv_addr" with the address of the
     server that we want to connect with.
  */
  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family      = AF_INET;
  serv_addr.sin_addr.s_addr = inet_addr(SERV_HOST_ADDR);
  serv_addr.sin_port        = htons(SERV_TCP_PORT);

  /* Open a TCP socket (an Internet stream socket) */
  if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	myperror("client: can't open stream socket");

  /* Connect to the server */
  if (connect(sockfd, (struct sockaddr *) &serv_addr,
              sizeof(serv_addr)) < 0)
	myperror("client: can't connect to server");

  /* ------------------------------------------------------ */ 
  /* Set up for asynch I/O for input file */
  /* This must also set the file descriptor to be */ 
  /* non-blocking?  This means that a read will return */
  /* immediately with the number of bytes available, which */
  /* may well be less than the buffer size */
  /* signal(SIGIO, (SIG_PF)sigio_func); */ 
  signal(SIGIO, sigio_func); 
  if(fcntl(fileno(fp), F_SETOWN, getpid()) < 0) 
    myperror("F_SETOWN error"); 
  if(fcntl(fileno(fp), F_SETFL, FASYNC) < 0)  
    myperror("F_SETFL FASYNC error");


  /* ------------------------------------------------------ */ 
  /* Endlessly move data from input to socket */
  while(1)
    {
    TimeWait(TIME_MARK);

    sigblock(sigmask(SIGIO));  // blocks (delays) SIGIO's until
                               // sigsetmask(0), EXCEPT during
                               // sigpause(0).

    while(sigflag == 0)        // Wait on a SIGIO.  The (0) in the
      sigpause(0);             // sigpause(0) sez to enable all 
                               // signals, including SIGIO.  On return
                               // from sigpause(0), old mask is restored,
                               // thus SIGIO's are once again blocked.

    if((NumData = read(fileno(fp), buf, NET_BUFSIZE)) > 0)
      {
      while(TimeWait(TIME_CHECK) < NO_SEND_TIME)
        usleep(10);
      while(NumData > 0)
        {
        //fprintf(stderr, "NumData = %ld\n", NumData);
        NumData > NET_BUFSIZE ? 
          NumToWrite = NET_BUFSIZE : NumToWrite = NumData;
        NumData -= NumToWrite;
        if((i = writen(sockfd, buf, NumToWrite)) != NumToWrite)
          myperror("client: write error on socket");
        }
      }
    else
      myperror("read error");

    sigflag = 0;
    sigsetmask(0);  // make sure all signals are now allowed 
    }  
  /* ------------------------------------------------------ */ 

  close(sockfd);
  exit(0);
  }

/* ------------------------------------------------------------- */
double TimeWait(long action)
  {
  static double MyTime, TimeMark;

  struct timeb mytimeb;

  ftime(&mytimeb);

  MyTime = mytimeb.time;
  MyTime += mytimeb.millitm / 1000.0;

  if(action == TIME_MARK)
    TimeMark = MyTime;

  //fprintf(stderr, "%lf\n", MyTime - TimeMark);
  return(MyTime - TimeMark);
  }

/* ------------------------------------------------------------- */

/* ------------------------------------------------------------- */
void myperror(char * str)
/* ------------------------------------------------------------- */
  {
  perror(str);
  exit(1);
  }

/* ------------------------------------------------------------- */
void sigio_func()
/* ------------------------------------------------------------- */
  {
  sigflag = 1;
  }
