/*
 * Source Code Name  : s4acquireratecheck.cc
 * Programmer        : Matt Lebofsky
 *                   : Project SERENDIP, University of California, Berkeley CA
 * Compiler          : GNU gcc
 *
 * This program checks the rate at which S4 is acquiring data as part of a
 * really crude (but very effective) way of eliminating RFI. The user must
 * supply on the command line:
 *
 * low_threshold (in units of hits/hour) 
 * high_threshold (in units of hits/hour) 
 * filename
 *
 * .. and the program checks the summary database record of the file and
 * reports back whether or not the data acquire rate is within the supplied
 * threshold range. It also checks just board 19 (for SETI@home purposes)
 * in case board 19's rate is acceptable or not.
 *
 * other options:
 *
 * -db  : put info about the rate check in the dbase
 * -v   : verbose output to stdout
 * -ver : show version and quit
 * -st  : add a line to skipdata.tab regarding bad rates
 *
 * versions:
 * version 1.0 initial release
 * version 1.1 cleaned up gcc 3.0 compiler warnings (9/30/03)
 *
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include "SIREN.h"
#include "s4database.h"
#include "s4summary.h"

#define S4ACQUIRERATECHECK_VERSION 1.1
#define ASSUMED_NUM_ACTIVE_BOARDS 35

const char * SkipTableFilename = "/disks/cyclops/c/users/seti/s4/siren/db/skipdata.tab";

const char * usage = "usage : s4acquireratecheck [-db] [-v] [-ver] [-st] low_thresh high_thresh filename\n         note: thresholds in units of hits/hour\n";

int main(int argc, char ** argv)
  {
  long i;
  double dbLowThresh, dbHighThresh;
  double dbHitsPerHour;
  double dbBoard19HitsPerHour;
  BOOLEAN Verbose = FALSE;
  BOOLEAN AllBoardsPass = TRUE;
  BOOLEAN Board19Pass = TRUE;
  BOOLEAN UpdateSkipTable = FALSE;
  long lActiveBoards = 0;
  FILE *fSkipTableFile;
  int iSkipTableFileNum;
  // char cBuffer[256];

  // Summary database variables
  BOOLEAN UpdateDB = false;
  DbaseSumRcd_t DbaseSumRcd; 
  int SumIndexFile;
  unsigned long SummDbIndex;

  char sFileNameRoot[DBASE_FILENAME_LEN];

  // get parameters

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i],"-ver") == 0) {
      fprintf(stderr,"s4acquireratecheck version %4.2f\n",
              S4ACQUIRERATECHECK_VERSION);
      exit(1);
      }  
    else if (strcmp(argv[i],"-db") == 0) UpdateDB = TRUE;
    else if (strcmp(argv[i],"-v") == 0) Verbose = TRUE;
    else if (strcmp(argv[i],"-st") == 0) UpdateSkipTable = TRUE;
    }

  if (argc < 3) {
    fprintf(stderr,"%s\n",usage);
    exit(1);
    } 
  
  if ( (sscanf(argv[argc-3],"%lf", &dbLowThresh) != 1) || 
       (sscanf(argv[argc-2],"%lf", &dbHighThresh) != 1) ||
       (sscanf(argv[argc-1],"%s", sFileNameRoot) != 1) ) {
    fprintf(stderr,"%s\n",usage);
    exit(1);
    } 

  SumIndexFile = sum_GetDataBase(O_RDWR);
  SummDbIndex = sum_GetRcd(sFileNameRoot, &DbaseSumRcd, SumIndexFile, O_RDWR);

  if (SummDbIndex == -1) {
    fprintf(stderr,"unable to find file: %s in the summary database\n",
            sFileNameRoot);
    exit(1);
    }

// difficult way of counting active boards - means countbirdie had to have run first
// for (i=0;i<NUMBOARDS;i++) 
//   if (DbaseSumRcd.ucBirdieSeenPercentages[i] != 255) lActiveBoards++;
// easier way:
  
  lActiveBoards = ASSUMED_NUM_ACTIVE_BOARDS; 

  dbHitsPerHour = (DbaseSumRcd.ulTotalHits / DbaseSumRcd.dbFileDuration);
  dbBoard19HitsPerHour = (DbaseSumRcd.ucBoardKiloCount[19] / 
                          DbaseSumRcd.dbFileDuration) * 1000.0 * lActiveBoards;

  if (Verbose)
    fprintf(stderr,"Settings:\nLow Threshold: %12.6lf hits/hour\n"
                   "High Threshold: %11.6lf hits/hour\n"
                   "Raw Data Filename: %s\n\n"
                   "All Boards Hit Rate: %lf hits/hour\n"
                   "Board 19 Hit Rate (* Num Active Boards): %lf hits/hour\n\n",
                   dbLowThresh,dbHighThresh,sFileNameRoot,
                   dbHitsPerHour,dbBoard19HitsPerHour);

  if (dbHitsPerHour < dbLowThresh ||
      dbHitsPerHour > dbHighThresh) {
    AllBoardsPass = FALSE;
    if (Verbose)
      fprintf(stderr,"file : %s ( %7.6lf %7.6lf ) fails all boards - "
                     "rate too %s ( %7.3lf %s %7.3lf )\n",
                     sFileNameRoot, DbaseSumRcd.dbStartTime,
                     (DbaseSumRcd.dbStartTime + (DbaseSumRcd.dbFileDuration / 24)),
                     dbHitsPerHour < dbLowThresh ? "low" : "high",
                     dbHitsPerHour,
                     dbHitsPerHour < dbLowThresh ? "<" : ">",
                     dbHitsPerHour < dbLowThresh ? dbLowThresh : dbHighThresh);
    }
  
  if (dbBoard19HitsPerHour < dbLowThresh ||
      dbBoard19HitsPerHour > dbHighThresh) {
    Board19Pass = FALSE;
    if (Verbose)
      fprintf(stderr,"file : %s ( %7.6lf %7.6lf ) fails board 19 - "
                     "rate too %s ( %7.3lf %s %7.3lf )\n",
                     sFileNameRoot, DbaseSumRcd.dbStartTime,
                     (DbaseSumRcd.dbStartTime + (DbaseSumRcd.dbFileDuration / 24)),
                     dbBoard19HitsPerHour < dbLowThresh ? "low" : "high",
                     dbBoard19HitsPerHour,
                     dbBoard19HitsPerHour < dbLowThresh ? "<" : ">",
                     dbBoard19HitsPerHour < dbLowThresh ? dbLowThresh : dbHighThresh);
    }

  if (UpdateSkipTable && (!AllBoardsPass || !Board19Pass)) {
    if ((fSkipTableFile = fopen(SkipTableFilename,"a+")) == NULL) {
      fprintf(stderr, "Could not open skip table file: %s\n",SkipTableFilename);
      exit(1);
      }
    iSkipTableFileNum = fileno(fSkipTableFile);
    if ((lockf(iSkipTableFileNum, F_LOCK, 0)) == -1) {
      fprintf(stderr, "Could not lock skip table file: %s\n",SkipTableFilename);
      exit(1);
      }

    fprintf(fSkipTableFile, "%s %7.6lf:%s %7.6lf  # s4acquireratecheck failed ",
                  sFileNameRoot,DbaseSumRcd.dbStartTime,
                  sFileNameRoot,(DbaseSumRcd.dbStartTime + (DbaseSumRcd.dbFileDuration / 24)));
    if (!AllBoardsPass) 
      fprintf(fSkipTableFile,"- all boards %7.6lf hits/hour ",dbHitsPerHour);
    if (!Board19Pass) 
      fprintf(fSkipTableFile,"- board 19 %7.6lf normalized hits/hour ",
                  dbBoard19HitsPerHour);
    fprintf(fSkipTableFile,"\n");

    lockf(iSkipTableFileNum, F_ULOCK, 0);
    fclose(fSkipTableFile);
    }
  
    

  if (UpdateDB) {
    DbaseSumRcd.fAcquireRateCheckVers = S4ACQUIRERATECHECK_VERSION;
    DbaseSumRcd.dbARCLowThreshold = dbLowThresh;
    DbaseSumRcd.dbARCHighThreshold = dbHighThresh;
    DbaseSumRcd.bAllBoardsPassed = AllBoardsPass;
    DbaseSumRcd.bBoard19Passed = Board19Pass;
    sum_PutRcd(SummDbIndex, &DbaseSumRcd, SumIndexFile);
    }

  sum_ReleaseDataBase(SumIndexFile);

  }
