/* ________________________________________________________________________
 programmer : Matt Lebofsky
 source name: s4adccheck.cc 
 compiler   : gcc
___________________________________________________________________________
 modification history:
   version 1.0 finished 2/11/99
___________________________________________________________________________
 usage: s4adccheck [-g gap_in_secs] [-br from_board to_board] < dbase_file
 takes dbase files from stdin, and verbose output goes to stdout
___________________________________________________________________________
 description: s4adccheck sifts through a questionable dbase file and
   spits out time ranges of bad ADC values in the form of:

ADC starttime (jd) starttime (mdy hms) endtime (jd) endtime (mdy hms) mean ADC
--- -------------- ------------------- ------------ ----------------- --------

   Where ADC is 0-4, representing the respective eng value, starttime(s)
   are the beginning of the bad ADC range, endtimes are the end, and the
   mean ADC is obvious. 

   There are allowed "gaps" (default 100 seconds) within these bad ADC ranges
   of good data, that is, is it worth keeping 100 seconds of good data between
   two obvious ranges of bad data? This can be adjusted using the "-g" flag.
   Setting this to 0 means no gaps are allowed.

   You can also restrict this analysis to a certain range of boards, that is,
   in the case of SETI@home, using only board 19. Use the -br flag for this.
___________________________________________________________________________
notes/bugs: 
___________________________________________________________________________
*/

#include <stdio.h>
#include <string.h>
#include "SIREN.h"
#include "s4io.h"
#include "s4freq.h"

typedef struct {
  double dbStartJul;    // where did the bad stuff start?
  char stStartTime[9];
  char stStartDate[7];
  double dbEndJul;      // where did it end?
  char stEndTime[9];
  char stEndDate[7];
  long lNumAdcs;        // how many ADCs are we talking about?
  long lTotalValue;     // total value to compute mean.
  } AdcRange;

void PrintRange(AdcRange PrintThis, long WhichAdc);

// ------------------------------------------------------------------------
int main (int argc, char *argv[])
// ------------------------------------------------------------------------
{

char *usage = "usage: s4adccheck [-g gap_in_secs] [-br from_board to_board] < dbase_file\n";

double dbThisSpectra;

long lFromBoard = 0;
long lToBoard = 39;    
long lGap = 100;
double dbRealGap;
long lWhichValue;     

AdcRange RangeSet[5];
StdEventRec ThisEvent;

long i;

for (i=0;i<5;i++) {
  RangeSet[i].dbStartJul = 0;
  RangeSet[i].dbEndJul = 0;
  RangeSet[i].lNumAdcs = 0;
  RangeSet[i].lTotalValue = 0;
  }
  
dbThisSpectra = 99999999;      /* to start with - no time could be this */


for (i = 1; i < argc; i++) {
  if (strcmp(argv[i],"-g") == 0) {
    if (sscanf(argv[++i],"%ld", &lGap) != 1) {
      fprintf(stderr,usage);
      exit(1);
      } 
    }
  else if (strcmp(argv[i],"-br") == 0) {
    if ((sscanf(argv[++i],"%ld", &lFromBoard) != 1) ||
        (sscanf(argv[++i],"%ld", &lToBoard) != 1)) {
      fprintf(stderr,usage);
      exit(1);
      }
    } 
  else {
    fprintf(stderr,usage);
    exit(1);
    }
  }

if ((lFromBoard < 0) || (lFromBoard > 39) ||
    (lToBoard < 0) || (lToBoard > 39)) {
  fprintf(stderr,"Selected Board(s) out of range (0-39)\n");
  exit(1);
  }

if (lFromBoard > lToBoard) {
  fprintf(stderr,"Why is from_board greater than to_board?\n");
  exit(1);
  }

if (lGap < 0 || lGap > 100000) {
  fprintf(stderr,"The gap value you selected is very weird (<0 or >100000)\n");
  exit(1);
  }

dbRealGap = (lGap * SECS_TO_DAYS);

// --------------------------------------------------------------------------
// Main Loop
// --------------------------------------------------------------------------
while (io_GetNextRecord(stdin,&ThisEvent,STDEVENT_REC)) {

  if (dbThisSpectra != ThisEvent.dbJulDate) {
    dbThisSpectra = ThisEvent.dbJulDate;
    }

  if (ThisEvent.ucBid >= lFromBoard && ThisEvent.ucBid <=lToBoard) {

    lWhichValue = long(ThisEvent.ucBid / 8);

    if (ThisEvent.usAdc >= 200) {
    
      if (RangeSet[lWhichValue].dbStartJul == 0) {    // beginning of bad range
        RangeSet[lWhichValue].dbStartJul = dbThisSpectra;
        strcpy(RangeSet[lWhichValue].stStartTime,ThisEvent.stUTime);
        strcpy(RangeSet[lWhichValue].stStartDate,ThisEvent.stDate);
        }
      
      RangeSet[lWhichValue].dbEndJul = dbThisSpectra;
      strcpy(RangeSet[lWhichValue].stEndTime,ThisEvent.stUTime);
      strcpy(RangeSet[lWhichValue].stEndDate,ThisEvent.stDate);
      RangeSet[lWhichValue].lNumAdcs++;
      RangeSet[lWhichValue].lTotalValue+=ThisEvent.usAdc;
  
      } // end if ADC >= 200
  
    else {  // ADC < 200, but do nothing unless we reach the end of a gap
  
      if (((ThisEvent.dbJulDate - RangeSet[lWhichValue].dbEndJul) >= dbRealGap)
         && (RangeSet[lWhichValue].dbStartJul > 0) ) {

        PrintRange(RangeSet[lWhichValue],lWhichValue);
        RangeSet[lWhichValue].dbStartJul = 0;
        RangeSet[lWhichValue].lNumAdcs = 0;
        RangeSet[lWhichValue].lTotalValue = 0;
      
        } // end if after gap

      } // end else (ADC < 200)

    } // end if within board range

  } /* End main while loop */

// Are there ranges remaining to be printed, if so do it!

for (i=0;i<5;i++) 
  if (RangeSet[i].dbStartJul > 0) 
    PrintRange(RangeSet[i],i);

return(0);
}

// ------------------------------------------------------------------------
void PrintRange(AdcRange PrintThis, long WhichAdc)
// ------------------------------------------------------------------------
{
  printf("%ld ",WhichAdc);
  printf("%13.10lf ",PrintThis.dbStartJul);
  printf("%s %s ",PrintThis.stStartDate,PrintThis.stStartTime);
  printf("%13.10lf ",PrintThis.dbEndJul);
  printf("%s %s ",PrintThis.stEndDate,PrintThis.stEndTime);
  printf("%ld\n",nlong(PrintThis.lTotalValue/PrintThis.lNumAdcs));
}
