/* ________________________________________________________________________
 programmer : Matt Lebofsky
 source name: s4bigbinfilter.cc 
 compiler   : gcc
___________________________________________________________________________
 modification history:
   version 1.0 finished 1/6/98
___________________________________________________________________________
 usage: s4bigbinfilter < dbase file > filtered file
 takes dbase files from stdin, filtered file gets output to stdout,
 and filtering statistics get output to stderr
___________________________________________________________________________
 description: s4bigbinfilter will go sift through a dbase file, and it will
   divide bins up into groups of 1024 Hz, and take only the first event
   it encounters from each of these groups during each spectra. 

___________________________________________________________________________
notes/bugs: 
___________________________________________________________________________
*/

#include <stdio.h>
#include <string.h>
#include "SIREN.h"
#include "s4io.h"

#define FFT_BIN_OFFSET 2097152
#define BIG_BIN_SIZE 1024

// ------------------------------------------------------------------------
int main (int argc, char *argv[])
// ------------------------------------------------------------------------
{

char *usage = "Usage: s4bigbinfilter < dbase file > filtered file (and stats go to stderr)\n";

long NumSpectra;
double ThisSpectra;

unsigned char BigBinCounter[40][4096];
unsigned char EmptyArray[40][4096];
long SizeOfArray;
long BigBin;

long EventsCounter[40];
long TotalEvents;

long EventsKeptCounter[40];
long TotalEventsKept;
double KeptPercent;

long EventsLostCounter[40];
long TotalEventsLost;
double LostPercent;

long MaxLostCounter[40];
long MinLostCounter[40];
long TempLostCounter[40];

long BinsBlitzedCounter[40];
long TotalBinsBlitzed;

long retvalue;

BOOLEAN BoardActive[40];     /* which boards are active? */

StdEventRec ThisEvent;

long i,j;

NumSpectra = 0;
TotalEvents = 0;
TotalEventsLost = 0;
TotalEventsKept = 0;
TotalBinsBlitzed = 0;

for (i = 0; i < 40; i++) {
  EventsCounter[i] = 0;
  EventsKeptCounter[i] = 0;
  EventsLostCounter[i] = 0;
  MaxLostCounter[i] = -1;
  MinLostCounter[i] = 1001;
  TempLostCounter[i] = 0;
  BinsBlitzedCounter[i] = 0;
  for (j = 0; j < 4096; j++) 
    BigBinCounter[i][j] = 0;
    EmptyArray[i][j] = 0;
  }

ThisSpectra = 99999999;      /* to start with - no time could be this */

retvalue = 0;

SizeOfArray = 163840; /* 40 boards x 4096 bins */

if (argc > 1) {
  fprintf(stderr,usage);
  exit(1);
  }

// --------------------------------------------------------------------------
// Main Loop
// --------------------------------------------------------------------------
while (io_GetNextRecord(stdin,(void *)&ThisEvent,STDEVENT_REC)) {

  EventsCounter[ThisEvent.ucBid]++;
  TotalEvents++;

  BoardActive[ThisEvent.ucBid] = TRUE;    /* If we have an event, then the
                                             board must be active */

  if (ThisSpectra != ThisEvent.dbJulDate) {    /* New spectra */

    NumSpectra++;                              /* First, increase # spectra */
    ThisSpectra = ThisEvent.dbJulDate;

    for (i = 0; i < 40; i++)
      for (j = 0; j < 4096; j++) 
        if (BigBinCounter[i][j] > 1) {
          BinsBlitzedCounter[i]++;
          TotalBinsBlitzed++;
          }

    memcpy(BigBinCounter,EmptyArray,SizeOfArray);

    for (i = 0; i < 40; i++) {
      if (TempLostCounter[i] > MaxLostCounter[i])
        MaxLostCounter[i] = TempLostCounter[i];
      if ((TempLostCounter[i] > 0) && (TempLostCounter[i] < MinLostCounter[i]))
        MinLostCounter[i] = TempLostCounter[i];
      TempLostCounter[i] = 0;
      }
    }

  if ((ThisEvent.lBinNumber < LOW_BIN) || (ThisEvent.lBinNumber > HIGH_BIN)) {
    fprintf(stderr,"Warning: Line #%ld: Bin Number (%ld) out of range\n"
                  ,TotalEvents,ThisEvent.lBinNumber);
    EventsLostCounter[ThisEvent.ucBid]++;
    TotalEventsLost++;
    TempLostCounter[ThisEvent.ucBid]++;
    retvalue = 1;
    }

  else {

    BigBin = ((ThisEvent.lBinNumber + FFT_BIN_OFFSET) / BIG_BIN_SIZE);

    if (BigBinCounter[ThisEvent.ucBid][BigBin]++ == 0) {
      io_WriteStdEvent(stdout,&ThisEvent);
      EventsKeptCounter[ThisEvent.ucBid]++;
      TotalEventsKept++;
      }
    else {
      EventsLostCounter[ThisEvent.ucBid]++;
      TotalEventsLost++;
      TempLostCounter[ThisEvent.ucBid]++;
      } 

    }

  } /* End main while loop */

for (i = 0; i < 40; i ++) {
  if (TempLostCounter[i] > MaxLostCounter[i])
    MaxLostCounter[i] = TempLostCounter[i];
  if ((TempLostCounter[i] > 0) && (TempLostCounter[i] < MinLostCounter[i]))
    MinLostCounter[i] = TempLostCounter[i];
  }


//------------------------------------------------------------------------------
// print out stats to stderr
//------------------------------------------------------------------------------

fprintf(stderr,"Number of Spectra: %ld\n",NumSpectra);
fprintf(stderr,"Total Events: %ld\n",TotalEvents);

KeptPercent = (TotalEventsKept / double(TotalEvents)) * 100;
fprintf(stderr,"Total Events Kept: %ld (%.0lf%%)\n",TotalEventsKept,KeptPercent);

LostPercent = (TotalEventsLost / double(TotalEvents)) * 100;
fprintf(stderr,"Total Events Lost: %ld (%.0lf%%)\n",TotalEventsLost,LostPercent);

fprintf(stderr,"Average of Different Big Bins Hit more than once per Spectra: %.0lf\n"
              ,(TotalBinsBlitzed / double(NumSpectra)));

for (i = 0; i < 40; i++) {
  
  if (MaxLostCounter[i] == -1) MaxLostCounter[i] = 0;
  if (MinLostCounter[i] == 1001) MinLostCounter[i] = 0;

  if (BoardActive[i]) {
  fprintf(stderr,"Board #%2ld - Kept: %ld Lost: %ld (Max/Sp: %ld Min/Sp: %ld) Avg BBs>1: %.0lf\n",i,
          EventsKeptCounter[i],EventsLostCounter[i],
          MaxLostCounter[i],MinLostCounter[i],
          (BinsBlitzedCounter[i] / double(NumSpectra)));
    }
 
  }

return(retvalue);

}
    
