/* ________________________________________________________________________
 programmer : Jeff Cobb
 source name: s4binhst.cc
 compiler   : gcc
___________________________________________________________________________
 modification history:
   version 1.0 finished 11/12/00
___________________________________________________________________________
 description: This program reads a dbse file and outputs a 2 column file
 containing hits histogrammed by frequency bin. Bin width can be given
 as a parameter.  One can give a histogram bin value with which to trim
 the output.  Optionally, the summary database can be updated.  If the
 summary DB is to be updated, thresholding cannot be specified.
___________________________________________________________________________
*/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "SIREN.h"
#include "s4io.h"
#include "s4utils.h"
#include "s4summary.h"

#define MAX_BIN_WIDTH 131072

void GetOps(int argc, 
            char ** argv, 
            unsigned long   * BinWidth, 
            unsigned long   * Threshold,
            BOOLEAN         * Before,
            BOOLEAN         * After,
            BOOLEAN         * UpdateDb);

int main(int argc, char ** argv) {

  double Version = 1.0;

  StdEventRec   Event;        

  unsigned long * BinHist;		// hits histogramed by bin     
  unsigned long   BinHist_i;  		// BinHist_i for BinHist
  unsigned long   BinHist_size; 	// Size of bin array = (num of bins)/BinWidth 
  unsigned long   BinWidth = 4096; 	// user specified bin width. 
  unsigned long   NumEvents;    	// Number of events in input file.              
  unsigned long   Threshold = 0;        // for histogram bin reporting.  default = report all
  unsigned long   i;                                          
  FILE          * InFile;       
  FILE          * OutFile;                                     
  long            SummDbIndex;
  int             SummIndexFile;
  DbaseSumRcd_t   DbaseSumRcd;
  char            FileNameRoot[64];
  BOOLEAN         Before = FALSE;
  BOOLEAN         After = FALSE;
  BOOLEAN         UpdateDb = FALSE;

  GetOps(argc, argv, &BinWidth, &Threshold, &Before, &After, &UpdateDb); 

  if ((InFile = fopen(argv[argc - 2], "r")) == NULL) {
    fprintf(stderr, "Can't open input file %s\n", argv[argc - 2]);
    exit(1);
  }
  if ((OutFile = fopen(argv[argc - 1], "w")) == NULL) {
    fprintf(stderr, "Can't open output file %s\n", argv[argc - 1]);
    exit(1);
  }

  NumEvents = (io_CountRecords(InFile,STDEVENT_REC));  
  if (NumEvents < 0) {
    perror("Trouble counting events in file");
    exit(1);
  }
  if (NumEvents == 0) {
     perror("Empty File");
     exit(1);
  }
  if (NumEvents > MAX_EVENTS) {
     fprintf(stderr, "Too many events (%ld vs %ld) in input file\n", NumEvents, MAX_EVENTS);
     exit(1);
  }

  // allocate histogram
  BinHist_size = (NUM_BINS*NUMBOARDS)/BinWidth; 
  if ((BinHist = (unsigned long *) calloc(BinHist_size,sizeof(unsigned long))) == NULL) {
         perror("Error allocating space for 'BinHist'");
         exit(1);
  }
  memset(BinHist,0,BinHist_size * sizeof(unsigned long));   // init to all zeros

  for (i=0;i<NumEvents;i++) {

    if ((io_GetNextRecord(InFile,&Event,STDEVENT_REC)!=1)) {
      perror("Bad file read.");
      exit(1);
    }
 
    if ((CONTIGBIN(Event.lBinNumber, Event.ucBid) > HIGH_CONTIGBIN)    ||
        (CONTIGBIN(Event.lBinNumber, Event.ucBid) < LOW_CONTIGBIN)) {
      fprintf(stderr, "Bin number (%ld) out of range (%ld to %lu) on line %lu.\n", 
              CONTIGBIN(Event.lBinNumber, Event.ucBid), LOW_CONTIGBIN, HIGH_CONTIGBIN, i+1);
      exit(1);
    }

    // increment histo bin
    BinHist_i = (unsigned long) (CONTIGBIN(Event.lBinNumber, Event.ucBid) - (LOW_CONTIGBIN))/BinWidth;
    BinHist[BinHist_i]++;

  }  // end populating histogram


  // Output histo, or portion described by Threshold
  for (i = 0; i < BinHist_size; i++)
     if (BinHist[i] >= Threshold)
       (void) fprintf(OutFile,"%ld %u\n",i, BinHist[i]);

  // update summary DB
  if(UpdateDb == TRUE) {
    ut_GetRootFilename(argv[argc - 1], &FileNameRoot[0]);
    SummIndexFile = sum_GetDataBase(O_RDWR);
    SummDbIndex  = sum_GetRcd(FileNameRoot, &DbaseSumRcd, SummIndexFile, O_RDWR);
    if(Before) {
      DbaseSumRcd.fBinHstBeforeVers = Version;
      DbaseSumRcd.lPoissonNumEventsBefore = NumEvents;
      DbaseSumRcd.lPoissonBinWidthBefore = BinWidth;
      DbaseSumRcd.lPoissonNumBinsBefore = (NUM_BINS*NUMBOARDS)/BinWidth;
    } else {
      DbaseSumRcd.fBinHstAfterVers = Version;
      DbaseSumRcd.lPoissonNumEventsAfter = NumEvents;
      DbaseSumRcd.lPoissonBinWidthAfter = BinWidth;
      DbaseSumRcd.lPoissonNumBinsAfter = (NUM_BINS*NUMBOARDS)/BinWidth;
    }
    sum_PutRcd(SummDbIndex, &DbaseSumRcd, SummIndexFile);
    sum_ReleaseDataBase(SummIndexFile);
  }

  fclose(InFile);             
  fclose(OutFile);             

  } /* End program. */


//--------------------------------------------------------------------------
void GetOps(int argc, 
            char ** argv, 
            unsigned long   * BinWidth, 
            unsigned long   * Threshold,
            BOOLEAN         * Before,
            BOOLEAN         * After,
            BOOLEAN         * UpdateDb) {
//--------------------------------------------------------------------------

  int i;
  char ErrString[128];

  if (argc < 2) {
         perror("Usage:  binhst [-bw BinWidth] [-t Threshold] [-db] infile_name outfile_name");
         exit(1);
         }

  for (i = 1; i < argc - 2; i++)
    if (strcmp(argv[i], "-bw") == 0) {
      if (sscanf(argv[++i],"%u",BinWidth) != 1)
        perror("Error in command line.");
      else if (MAX_BIN_WIDTH % *BinWidth != 0) {
        sprintf(ErrString, "Ilegal bin width--must be power of 2 <= %d\n", MAX_BIN_WIDTH);
        perror(ErrString);
        exit(1);
      }
    }
    else if (strcmp(argv[i], "-t") == 0) {
      if (sscanf(argv[++i],"%u",Threshold) != 1)
        perror("Error in command line.");
    }
    else if (strcmp(argv[i], "-db") == 0) {
      *UpdateDb = TRUE;
    }
    else if (strcmp(argv[i], "-before") == 0) {
      *Before = TRUE;
    }
    else if (strcmp(argv[i], "-after") == 0) {
      *After = TRUE;
    }
    else {
      perror("Usage:  s4binhst -before|-after [-bw BinWidth] [-t Threshold] [-db] infile_name outfile_name");
      exit(1);
    }

  if(!(*Before || *After)) { 
      perror("Usage:  s4binhst -before|-after [-bw BinWidth] [-t Threshold] [-db] infile_name outfile_name");
      exit(1);
   }

  if(*UpdateDb && *Threshold != 0) {
     perror("Cannot update the summary DB while thresholding");
     exit(1);
  }

}
