/* ________________________________________________________________________
 programmer : Matt Lebofsky
 source name: s4countbirdie.cc 
 compiler   : gcc
___________________________________________________________________________
 modification history:
   version 1.00 finished 12/9/97
   version 2.00 finished 2/5/99  - added -br flag 
   version 2.01 finished 3/5/99  - changed bd_GetBoardTab() to tel_GetTelTab()
   version 2.02 finished 7/22/99 - don't exit 1 if too many birdies - this
                                   is because we have RFI right on some
                                   birdie freqs in 21cm data
   version 2.03 finished 2/29/00 - added summ db stuff and version #
                                   NOTE: not takes pcdbase filename on com line
   version 2.04 finished 3/9/00  - added false birdies to summ db
   version 2.05 finished 4/10/00 - added new db stuff to replace old db stuff
                                   Now it's a 48-char array of percentages
                                   of birdies seen per board, and a bitmap
                                   of mirror birdies seen per board above
                                   threshold flag
   version 2.06 finished 9/30/03 - cleaned up gcc 3.0 compiler warnings
___________________________________________________________________________
 usage: s4countbirdie [-p ok_percent] [-f false_birdie_percent] [-v verbose]
                      [-br from_board to_board output_file] [-ver] [-db] file
                       
___________________________________________________________________________
 description: s4countbirdie takes pcdbase-generated files from file and
   streams through them, calculating potential birdies via the synthesizer
   setting (see CalcBirdie), and then tallying up how many of these birdies
   have been hit during the correct spectra on each board. The number of
   "false birdies" (which are when the usual birdie bins are hit when they
   weren't supposed to) are also tallied up per board.

   Once the files have been sifted through, s4countbirdie counts up the
   total number of spectra encountered, the total number of birdies expected,
   the total number of birdies encountered, and the total number of false
   birdies encountered. It then calculates the two percentages:

   OKPercent = number of good birdies encountered / number expected
   FalsePercent = number of false birdies encountered / number spectra

   If the OKPercent > the specified percentage (default 75%) and
   FalsePercent < the specified percentage (default 10%) then the program
   returns a value of 0, otherwise 2 (it returns 1 if there is a problem
   with the execution).

   The user can specify different percentages with flag options, as well as
   a verbose mode, which then prints out all the totals, the percentages,
   and a list of all the boards and their activity.

   The -br flag takes three values: from_board, to_board, and output_filename.
   What this does, in addition to its regular job, is sending info from the
   dbase lines containing the birdies found between from_board and to_board
   to the file output_filename. In short: "-br 17 20 blah" sends all birdies
   found on boards 17, 18, 19 and 20 and plops them into blah.

   The format of the above info takes the form:
   jd    m d y h m s    board#   freq   power   RA   Dec

___________________________________________________________________________
notes/bugs: 
   Note that birdies *usually* light up adjacent bins, so when real birdies
   are being searched for, this is taken into account. However, a false
   birdie, if in the usual birdie bin *as well as* the adjacent bin will
   get tallied up as *two* false birdies on that board. This is because if
   we checked for adjacent false birdies on the 39 boards which aren't
   supposed to be hit, the computation time would take (relatively) forever.
___________________________________________________________________________
*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include "SIREN.h"
#include "s4io.h"
#include "s4freq.h"
#include "s4boardtable.h"
#include "s4summary.h"
#include "s4tel.h"
#include "s4utils.h"

#define COUNTBIRDIE_VERSION 2.06

BOOLEAN IsBirdie(long TestBin, char TestBoard);

// ------------------------------------------------------------------------
int main (int argc, char *argv[])
// ------------------------------------------------------------------------
{

char *usage = "Usage: s4countbirdie [-p ok_percent] [-f false_birdie_percent] [-v verbose]\n                     [-br from_board to_board output_file] [-ver] [-db] file\n";

FILE *fDbaseFile = NULL;
char sDbaseFileNameRoot[20];

long retvalue;                          /* 0 = everything went well
                                           1 = problem with execution
                                           2 = problem with data */
long lNumSpectra;
double dbThisSpectra;

long lBinWasHit[80];          /* board # * 2 = negative birdie */
long lBinWasFalseHit[80];     /* board # * 2 + 1 = positive birdie
                                80 elements = 40 boards x 2 birdies each */

BOOLEAN bBoardActive[40];     /* which boards are active? */
long lBirdiesExpected[40];    /* how many birdies per board expected? */

long lBirdieBoard;            /* values from calcbirdie */
double dbBirdieBin;

long lPositiveEvent;

long lTotalHits, lTotalFalseHits, lTotalBirdiesExpected;

BOOLEAN bGotPosBirdie = FALSE, bGotNegBirdie = FALSE;

double dbHitPercent, dbOKHitPercent;
double dbFalsePercent, dbOKFalsePercent;

BOOLEAN Verbose;
BOOLEAN bUpdateDB;

DbaseSumRcd_t DbaseSumRcd; // Summary database record
int SumIndexFile;
unsigned long SummDbIndex;
// char sFileNameRoot[DBASE_FILENAME_LEN];

BOOLEAN DoBoardOutput;        /* for -br flag */
long lFromBoard,lToBoard;    
char sOutputFilename[50];
FILE *BoardOutputFile = NULL;

StdEventRec ThisEvent;

long i;

retvalue = 0;
lNumSpectra = 0;
lTotalHits = 0;
lTotalFalseHits = 0;
lTotalBirdiesExpected = 0;

for (i = 0; i < 80; i++) {
  lBinWasHit[i] = 0;
  lBinWasFalseHit[i] = 0;
  }

for (i = 0; i < 40; i++) {
  bBoardActive[i] = FALSE;
  lBirdiesExpected[i] = 0;
  }

dbThisSpectra = 99999999;      /* to start with - no time could be this */

dbOKHitPercent = 75.0;         /* default values */
dbOKFalsePercent = 10.0;       /* final calculations yeilding more than     */
                               /* OKHitPercent and less than OKFalsePercent */
                               /* return a value of 0, otherwise return 2   */
Verbose = FALSE;
DoBoardOutput = FALSE;
bUpdateDB = FALSE;
lFromBoard = 0;
lToBoard = 39;

if (argc < 2) {
  fprintf(stderr,usage);
  exit(1);
  }

for (i = 1; i < argc; i++) {
  if (strcmp(argv[i],"-v") == 0) 
    Verbose = TRUE;
  else if (strcmp(argv[i],"-ver") == 0) {
    fprintf(stderr,"s4countbirdie version %4.2f\n",COUNTBIRDIE_VERSION);
    exit(1);
    }
  else if (strcmp(argv[i],"-db") == 0) {
    bUpdateDB = TRUE;
    }
  else if (strcmp(argv[i],"-p") == 0) {
    if (sscanf(argv[++i],"%lf", &dbOKHitPercent) != 1) {
      fprintf(stderr,usage);
      exit(1);
      } 
    }
  else if (strcmp(argv[i],"-f") == 0) {
    if (sscanf(argv[++i],"%lf", &dbOKFalsePercent) != 1) {
      fprintf(stderr,usage);
      exit(1);
      }
    }
  else if (strcmp(argv[i],"-br") == 0) {
    if ((sscanf(argv[++i],"%ld", &lFromBoard) != 1) ||
        (sscanf(argv[++i],"%ld", &lToBoard) != 1) ||
        (sscanf(argv[++i],"%s", sOutputFilename) != 1)) {
      fprintf(stderr,usage);
      exit(1);
      }
    else {
      DoBoardOutput = TRUE;
      } 
    } 
  else if ((fDbaseFile = fopen(argv[i],"r")) == NULL) {
    fprintf(stderr,"Can't read dbase file: %s\n",argv[i]);
    exit(1);
    }
  }

ut_GetRootFilename(argv[i-1], sDbaseFileNameRoot);

if ((dbOKHitPercent < 0.0) || (dbOKHitPercent > 100.0) ||
    (dbOKFalsePercent < 0.0) || (dbOKFalsePercent > 100.0)) {
  fprintf(stderr,"Percentage(s) out of range (0-100)\n");
  exit(1);
  }

if ((lFromBoard < 0) || (lFromBoard > 39) ||
    (lToBoard < 0) || (lToBoard > 39)) {
  fprintf(stderr,"Selected Board(s) out of range (0-39)\n");
  exit(1);
  }

if (lFromBoard > lToBoard) {
  fprintf(stderr,"Why is from_board greater than to_board?\n");
  exit(1);
  }

if (DoBoardOutput) {
  if ((BoardOutputFile = fopen(sOutputFilename,"w")) == NULL) {
    fprintf(stderr,"Unable to open filename: %s\n",sOutputFilename);
    exit(1);
    }
  }


// --------------------------------------------------------------------------
// Main Loop
// --------------------------------------------------------------------------
while (io_GetNextRecord(fDbaseFile,&ThisEvent,STDEVENT_REC)) {

  bBoardActive[ThisEvent.ucBid] = TRUE;    /* If we have an event, then the
                                             board must be active */

  if (dbThisSpectra != ThisEvent.dbJulDate) {

    lNumSpectra++;
    bGotPosBirdie = FALSE;
    bGotNegBirdie = FALSE;
    dbThisSpectra = ThisEvent.dbJulDate;

    fq_CalcBirdie((ThisEvent.dbBirdieSynth*1000000), &lBirdieBoard,
                   &dbBirdieBin, tel_GetTelTab(ThisEvent.ucTid));

    if ((lBirdieBoard >= 0) && (lBirdieBoard < 40) && (dbBirdieBin != 0) &&
        (dbBirdieBin >= LOW_BIN) && (dbBirdieBin <= HIGH_BIN)) 
      lBirdiesExpected[lBirdieBoard]++;     /* The BirdieBoard can potentially
                                               be > 39, and the BirdieBin can
                                               be = 0, so don't expect those */

    // If -br flag is set then output potential birdies:

    if (DoBoardOutput) {
      if ((lBirdieBoard >= lFromBoard) &&
          (lBirdieBoard <= lToBoard)) {
        fprintf(BoardOutputFile,"%.8lf %s %s %ld %lf %u %4.5lf %4.4lf\n",
                ThisEvent.dbJulDate,ThisEvent.stDate,ThisEvent.stUTime,
                lBirdieBoard,
                fq_s4fcalc(ThisEvent.dbFirstLo,ThisEvent.dbSecondLo,
                           fq_GetBrdFrq(tel_GetTelTab(ThisEvent.ucTid),lBirdieBoard),
                           long(dbBirdieBin),ThisEvent.ucTid),
                ThisEvent.usPower, ThisEvent.dbRa, ThisEvent.dbDec);
        }
      }  

    // Done with -br flag stuff

     }

   if ((ThisEvent.lBinNumber != 0) &&
     (IsBirdie(ThisEvent.lBinNumber,ThisEvent.ucBid) == TRUE)) {
 
    (ThisEvent.lBinNumber > 0) ? lPositiveEvent = 1 : lPositiveEvent = 0;

    if (((ThisEvent.lBinNumber == long(dbBirdieBin)) ||
         (ThisEvent.lBinNumber == long(dbBirdieBin) + (lPositiveEvent*2-1))) &&
         (ThisEvent.ucBid == lBirdieBoard) ) {
 
      if ((lPositiveEvent == 1) && (bGotPosBirdie == FALSE)) {
        lBinWasHit[(ThisEvent.ucBid * 2) + lPositiveEvent]++;
        bGotPosBirdie = TRUE;
        }
      
      else if ((lPositiveEvent == 0) && (bGotNegBirdie == FALSE)) {
        lBinWasHit[(ThisEvent.ucBid * 2) + lPositiveEvent]++;
        bGotNegBirdie = TRUE;
        }
      }

    else 

      lBinWasFalseHit[(ThisEvent.ucBid * 2) + lPositiveEvent]++;

    } /* End if IsBirdie */

  } /* End main while loop */

// ----------------------------------------------------------------------------
// add it up and print it out
// ----------------------------------------------------------------------------
for (i = 0; i < 80; i++) {
  lTotalHits += lBinWasHit[i];
  lTotalFalseHits += lBinWasFalseHit[i];
  }

for (i = 0; i < 40; i++) 
  lTotalBirdiesExpected +=
    (lBirdiesExpected[i] * long(bBoardActive[i]));

dbHitPercent = (lTotalHits / double(lTotalBirdiesExpected)) * 100;
dbFalsePercent = (lTotalFalseHits / double(lNumSpectra)) * 100;

// old way of doing things:
// if ((dbHitPercent < dbOKHitPercent) || (dbFalsePercent > dbOKFalsePercent))
//   retvalue = 2;

if (dbHitPercent < dbOKHitPercent)
  retvalue = 2;

if (Verbose) {
  fprintf(stdout,"Number of spectra: %ld\n", lNumSpectra);
  fprintf(stdout,"Number of birdies expected: %ld\n", lTotalBirdiesExpected);
  fprintf(stdout,"Number of birdies actually hit: %ld\n", lTotalHits);
  fprintf(stdout,"Number of birdie bins unexpectedly hit: %ld\n\n", lTotalFalseHits);

  fprintf(stdout,"Good Birdie Hit Percentage: %3.0f%% ",dbHitPercent);
  if (dbHitPercent < dbOKHitPercent) fprintf(stdout,"-- Not Enough!");
  fprintf(stdout,"\n");
  fprintf(stdout,"False Birdie Hit Percentage:%3.0f%% ",dbFalsePercent);
  if (dbFalsePercent > dbOKFalsePercent) fprintf(stdout,"-- Too Many!");
  fprintf(stdout,"\n\n");

  for (i = 0; i < 40; i++) {
    fprintf(stdout,"Board #%2ld - Real Birdies: %ld (Expected: %ld) False Birdies: %ld  ",i,
            lBinWasHit[i*2]+lBinWasHit[(i*2)+1],
            lBirdiesExpected[i],
            lBinWasFalseHit[i*2]+lBinWasFalseHit[(i*2)+1]);

    if (bBoardActive[i]) fprintf(stdout,"(active)");
    fprintf(stdout,"\n");
    }
  }

if (DoBoardOutput) fclose(BoardOutputFile);

// Add stuff to summary database if requested

if (bUpdateDB) {

  SumIndexFile = sum_GetDataBase(O_RDWR);
  SummDbIndex  = sum_GetRcd(sDbaseFileNameRoot, &DbaseSumRcd, SumIndexFile, O_RDWR);

// DbaseSumRcd.ulNumBirdies           = lTotalHits;
// DbaseSumRcd.ulNumBirdiesExpected   = lTotalBirdiesExpected;
// DbaseSumRcd.ulNumFalseBirdies      = lTotalFalseHits;

  for (i = 0; i < 48; i++)  
    DbaseSumRcd.ucBirdieSeenPercentages[i] = 255;  // set to "inactive"
  DbaseSumRcd.ullMirrorSeen = 0;

  for (i = 0; i < 40; i++) {
    if (bBoardActive[i]) {
      DbaseSumRcd.ucBirdieSeenPercentages[i] = (unsigned char)
        ((lBinWasHit[i*2]+lBinWasHit[(i*2)+1])*100/double(lBirdiesExpected[i]));
      if (((lBinWasFalseHit[i*2]+lBinWasFalseHit[(i*2)+1])/double(lBirdiesExpected[i]))*100
          > long(dbOKFalsePercent)) {
        DbaseSumRcd.ullMirrorSeen |= (unsigned long long)(pow(2,i)); 
        }
      }
    }
     
  DbaseSumRcd.fCountBirdieVers       = COUNTBIRDIE_VERSION; 
 
  sum_PutRcd(SummDbIndex, &DbaseSumRcd, SumIndexFile);
  sum_ReleaseDataBase(SumIndexFile);
 
  }

return(retvalue);
}

// Function IsBirdie: Test to see if the current event bin has the
//   same value as a potential birdie signal (different values for even
//   or odd boards) 
//
// ------------------------------------------------------------------------
BOOLEAN IsBirdie(long lTestBin, char cTestBoard)
// ------------------------------------------------------------------------
{

BOOLEAN bBirdieFound;

  bBirdieFound = FALSE; 

  if (cTestBoard % 2) {                      /* Is it an odd board? */
    if ((lTestBin == 1677721) || (lTestBin == 1677722) ||
        (lTestBin == -1677721) || (lTestBin == -1677722)) bBirdieFound = TRUE; }
  else {
    if ((lTestBin == 838860) || (lTestBin == 838861) ||
        (lTestBin == -838860) || (lTestBin == -838861)) bBirdieFound = TRUE; }

  return(bBirdieFound);
}
