/* ________________________________________________________________________
 programmer : Matt Lebofsky
 source name: s4dbascheck.cc 
 compiler   : gcc
___________________________________________________________________________
 modification history:
   version 1.0 finished 12/19/97
   version 1.1 	12/18/98	jeffc	board table is now 2d.  added badlimit.
					moved TiD defs to s4tel.h.
   version 1.2  3/5/99          jeffc   changed bd_GetBoardTab() to tel_GetTelTab()
   version 1.3  3/19/99         jeffc   added -filter option
   version 1.4  3/4/00		jeffc	added summary DB stuff, rearranged code
					to have error table in it's own .cc file
					so that other programs can read it.
   version 1.41 3/5/00		jeffc	added TimeOrder error type
   version 1.5  4/2/01          mattl   added exit 1 status if over x bad lines
   version 1.6  9/30/03         mattl   cleaned up gcc 3.0 compiler warnings
___________________________________________________________________________
 usage: s4dbascheck -lo1 First Oscillator -lo2 Second Oscillator
        -id Receiver ID -t Sigma Threshold [-v Verbose Mode] [-filter] [-db]
        [-bl bad_line_percentage] dbase_filename

 These arguments are the same as with s4pcdbase, with additional 
 -v, -filter, and -bl options.
___________________________________________________________________________
 description: s4dbascheck reads dbase files to check for reasonable data.
   These are all simple tests to make sure values in each standard event
   record are correct or within a specific range.  The exit status indicates
   the resulting state of the dbase file.  A status of one means that the
   file contains errors and should not be further processed.  A status of
   zero indicates that the file is error free.  Some errors are 
   considered nonfatal.  If the -filter option is specified, lines
   with nonfatal errors are removed and status remains zero. If the -bl
   option is used (with a value of 0-100) then exit status one if the amount
   containing errors is above the specified percent (this is to prevent heavily
   corrupted files, even with "non-fatal" errors, from being further processed
   in the s4pipeline).

   If you are in verbose mode, s4dbascheck sends the event line number and the
   specific problem to stderr, and then continues on in the file to find
   more problems if any (and once through, exits(1)).

   
___________________________________________________________________________
notes/bugs: 
   The various data tests are subject to major revision.
___________________________________________________________________________
*/

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include "SIREN.h"
#include "s4io.h"
#include "s4time.h"
#include "s4freq.h"
#include "s4boardtable.h"
#include "s4tel.h"
#include "s4summary.h"
#include "s4utils.h"
#include "s4error.h"

#define DBASECHECK_VERSION 1.6

// Global Variables (to be used in common error handling subroutines)

char *usage="Usage: s4dbascheck -lo1 First Oscillator -lo2 Second Oscillator\n                   -id Receiver ID -t Sigma Threshold [-bl bad line percentage]\n                   [-v Verbose Mode] [-filter] [-db Update Summary DB] dbase_filename\n";

long LineNo, BoardTab, ExitStatus;
BOOLEAN Verbose;
BOOLEAN bRecordOK;
BOOLEAN bFatalError;
BOOLEAN bFilter;
BOOLEAN bUpdateDb;
long lTotalBad;
long lBadLines;
unsigned long GlobalErrorFlags = 0;

long ErrorHandle(char errorMessage[], int Error);
void UsageExit();

// ------------------------------------------------------------------------
int main (int argc, char *argv[])
// ------------------------------------------------------------------------
{

// enum ErrorType_t ErrorType;

unsigned short ErrorStatus;

long SummDbIndex;
int SumIndexFile;

StdEventRec ThisEvent;
DbaseSumRcd_t DbaseSumRcd;

FILE * fpDbaseFile = NULL;    
FILE * fpNoErrorFile = NULL;    

char * sDbaseFile;
char  sNoErrorFile[256];
char  sErrorMsg[256];
char  sDbaseFileNameRoot[256];

// User-settable test conditions (see defaults below):

double dbTestFirstLO, dbTestSecondLO;
char sTestID[256];
unsigned char ucTestID;
unsigned short usTestSigmaThreshold;

// Other helpful variables:

long i;
long Month,Day,Year,MaxDays,CurrentYear;
// double Hour;
time_t CurrentTime;
tm *CurrentTm;
BOOLEAN FirstSuspiciousYear = TRUE; 
double dbPrevJul = 0.0;

double dbBadLimit; // percentage of bad lines acceptable

// Set default test conditions, though they should be supplied by user:

dbTestFirstLO = 0.0;
dbTestSecondLO = 0.0;
ucTestID = 0;
usTestSigmaThreshold = 0;
Verbose   = FALSE;
bFilter   = FALSE;
bUpdateDb = FALSE;
dbBadLimit = 80.0; // default is 80% of bad lines and less is tolerable

LineNo = 0;
ExitStatus = 0;
lTotalBad = 0; // total of ALL errors (many can happen per line)
lBadLines = 0; // total of bad lines (i.e. 1 or more errors on these lines)

// Get the current year:

CurrentTime = time(NULL);
CurrentTm = gmtime(&CurrentTime);
CurrentYear = tm_CentYear(CurrentTm->tm_year);

// Get arguments:

if (argc < 8) UsageExit();

for (i = 1; i < argc - 1; i++) {
  if (strcmp(argv[i],"-lo1") == 0) {
    if (sscanf(argv[++i],"%lf", &dbTestFirstLO) != 1) 
      UsageExit();
    }  
  else if (strcmp(argv[i],"-lo2") == 0) {
    if (sscanf(argv[++i],"%lf", &dbTestSecondLO) != 1) 
      UsageExit();
    }  
  else if (strcmp(argv[i],"-id") == 0) {
    if (sscanf(argv[++i],"%s", sTestID) != 1) 
      UsageExit();
    if((ucTestID = tel_GetTid(sTestID)) == UNKNOWN_TEL) {
      fprintf(stderr,"Unknown Telescope ID\n");
      UsageExit();
      }
    }  
  else if (strcmp(argv[i],"-t") == 0) {
    if (sscanf(argv[++i],"%hu", &usTestSigmaThreshold) != 1) 
      UsageExit();
    }  
  else if (strcmp(argv[i],"-v") == 0) {
    Verbose = TRUE;
    }
  else if (strcmp(argv[i],"-filter") == 0) {
    bFilter = TRUE;
    }
  else if (strcmp(argv[i],"-db") == 0) {
    bUpdateDb = TRUE;
    }
  else if (strcmp(argv[i],"-bl") == 0) {
    if (sscanf(argv[++i],"%lf", &dbBadLimit) != 1) 
      UsageExit();
    }
  else UsageExit();
  }  

  if (dbBadLimit < 0.0 || dbBadLimit > 100.0) {
    fprintf(stderr,"Bad Line Percentage should be between 0.0 and 100.0\n");
    UsageExit();
    }

  sDbaseFile    = argv[argc - 1];
  strcpy(sNoErrorFile, sDbaseFile);
  if(bFilter)
    strcat(sNoErrorFile, ".s4dbascheck_noerror");

  if((fpDbaseFile   = fopen(sDbaseFile, "r+")) == NULL) {
    sprintf(sErrorMsg,"Could not open %s\n", sDbaseFile);
    perror(sErrorMsg);
    exit(1);
  }

  if(bFilter) 
    if((fpNoErrorFile   = fopen(sNoErrorFile, "w")) == NULL) {
      sprintf(sErrorMsg,"Could not open %s\n", sNoErrorFile);
      perror(sErrorMsg);
      exit(1);
    }

// Main loop:

while (io_GetNextRecord(fpDbaseFile,&ThisEvent,STDEVENT_REC)) {

  bRecordOK = TRUE;

  LineNo++;

  BoardTab = tel_GetTelTab(ThisEvent.ucTid);

// Test time/date (using time string and decimal hour field):

  tm_DateToI(ThisEvent.stDate,&Day,&Month,&Year);
  Year = tm_CentYear(Year);

  if (Month < 1) ErrorHandle("Month < 1",  eMonth);
  if (Month > 12) ErrorHandle("Month > 12", eMonth);
  
  switch(Month)
  {
    case 2  : (Year%4==0 && Year%100!=0 || Year%400==0)?MaxDays=29:MaxDays=28;
              break;
    case 4  : MaxDays = 30;
              break;
    case 6  : MaxDays = 30;
              break;
    case 9  : MaxDays = 30;
              break;
    case 11 : MaxDays = 30;
              break;
    default : MaxDays = 31; 
  }  
 
  if (Day < 1) ErrorHandle("Day < 1", eDay);
  if (Day > MaxDays) ErrorHandle("Day > Maximum Days for this Month", eDay);

  if (Year < 1997) ErrorHandle("Year < 1997", eYear);
  if (Year > 2005) ErrorHandle("Year > 2010", eYear);

// Just warn us if the year isn't current - don't exit out just yet

  if (Year != CurrentYear && FirstSuspiciousYear)  {
    FirstSuspiciousYear = FALSE; 
    fprintf(stderr,"Line %ld: Year is suspicious\n",LineNo);
    }

// Test time/date using actual dbase strings:

  if (ThisEvent.dbUTime < 0.0) ErrorHandle("Hour String < 0000000", eHour);
  if (ThisEvent.dbUTime >= 24.0) ErrorHandle("Hour String > 23595999", eHour);

// Test that time is moving forward

  if(dbPrevJul == 0.0) {
	dbPrevJul = ThisEvent.dbJulDate;
  } else {
  	if(ThisEvent.dbJulDate != dbPrevJul) {
		if(ThisEvent.dbJulDate < dbPrevJul) {
			ErrorHandle("Time goes backwards", eTimeOrder);
		}
		dbPrevJul = ThisEvent.dbJulDate;
	}
  }

// Test RA:

  if (ThisEvent.dbRa < 0.0) ErrorHandle("RA < 0", eRA);
  if (ThisEvent.dbRa >= 24.0) ErrorHandle("RA >= 24.0", eRA);

// Test Dec:

  if (ThisEvent.dbDec < -3.0) ErrorHandle("Dec < -3.0", eDec); 
  if (ThisEvent.dbDec > 40.0) ErrorHandle("Dec > 40.0", eDec);

// Test First LO:

  if (ThisEvent.dbFirstLo != dbTestFirstLO) 
    ErrorHandle("First LO incorrect", eFirstLo);
    
// Test Second LO:

  if (ThisEvent.dbSecondLo != dbTestSecondLO) 
    ErrorHandle("Second LO incorrect", eSecondLo);

// Test lBinNumber

  if (ThisEvent.lBinNumber < LOW_BIN) 
    ErrorHandle("lBinNumber < LOW_BIN", eBin);
  if (ThisEvent.lBinNumber > HIGH_BIN) 
    ErrorHandle("lBinNumber > HIGH_BIN", eBin);

// Test ucBid
  
  if (char(ThisEvent.ucBid) < 0) ErrorHandle("ucBid < 0", eBid);
  if (char(ThisEvent.ucBid) > 39) ErrorHandle("ucBid > 39", eBid);

// Test ulFftBdFreq

  if (ThisEvent.ulFftBdFreq != fq_GetBrdFrq(BoardTab, ThisEvent.ucBid))
    ErrorHandle("ulFftBdFreq incorrect", eBdFrq);

// Test dbBirdieSynth

  if (ThisEvent.dbBirdieSynth < 172.0) 
    ErrorHandle("dbBirdieSynth < 172.0", eBirdieSynth);
  if (ThisEvent.dbBirdieSynth > 311.0) 
    ErrorHandle("dbBirdieSynth > 311.0", eBirdieSynth);

  if ((ThisEvent.dbBirdieSynth - long(ThisEvent.dbBirdieSynth)) != 0) 
    ErrorHandle("dbBirdieSynth has a decimal != 0", eBirdieSynth);

// Test usPower

  if (ThisEvent.usPower <= 0) ErrorHandle("usPower <= 0", ePower);

// Test usSigmaThreshold

  if (ThisEvent.usSigmaThreshold != usTestSigmaThreshold)
    ErrorHandle("usSigmaThreshold incorrect", eSigThresh);

// Test usActualThreshold

  if (ThisEvent.usActualThreshold <= 0) 
    ErrorHandle("usActualThreshold <= 0", eActThresh);

// Test usAdc

  if (ThisEvent.usAdc < 5) ErrorHandle("usAdc < 5", eAdc);
  if (ThisEvent.usAdc > 200) ErrorHandle("usAdc > 200", eAdc);

// Test cPolarization

  if (ThisEvent.cPolarization != 
         BoardTable[BoardTab][ThisEvent.ucBid].cPolarization) 
    ErrorHandle("cPolarization incorrect", ePolar);

// Test usSpctResLog

  if (ThisEvent.usSpctResLog != 0) 
    ErrorHandle("usSpctResLog != 0", eSpectRes);

// Test ucTid

  if (ThisEvent.ucTid != ucTestID) 
    ErrorHandle("ucTid incorrect", eTid);

  if (bFilter && bRecordOK)
    io_WriteStdEvent(fpNoErrorFile, &ThisEvent);
  
  if (!bRecordOK) lBadLines++;

  } /* End While Loop */

  fclose(fpDbaseFile);
  if(bFilter)
    fclose(fpNoErrorFile);

  fprintf(stderr,"%s : errors by type:\n", argv[0]);
  for(i = 0; i < 19; i++)
    fprintf(stderr,"%20s : %ld\n", DbaseError[i].Error, DbaseError[i].Count);

  if(bFatalError) {
    ExitStatus = 1;
    ErrorStatus = DBASE_FATAL;
    if(bFilter) unlink(sNoErrorFile);
  } else if ( ( (double)lBadLines / (double)LineNo ) * 100 > dbBadLimit) {
    ExitStatus = 1;
    ErrorStatus = DBASE_ERRORS_OVER_LIMIT;
    if(bFilter) unlink(sNoErrorFile);
  } else {
    ExitStatus = 0;
    if(bFilter) {
      unlink(sDbaseFile);
      link(sNoErrorFile, sDbaseFile);
      unlink(sNoErrorFile);
      ErrorStatus = DBASE_FILTERED;
    } else {
      ErrorStatus = DBASE_NONFATAL;
    }
  }

  // update summary database
  if(bUpdateDb) {
  	ut_GetRootFilename(sDbaseFile, &sDbaseFileNameRoot[0]);
  	SumIndexFile = sum_GetDataBase(O_RDWR);
  	SummDbIndex  = sum_GetRcd(sDbaseFileNameRoot, &DbaseSumRcd, SumIndexFile, O_RDWR);
  	DbaseSumRcd.fDbasCheckVers     = DBASECHECK_VERSION;
  	DbaseSumRcd.lDbaseErrors       = lTotalBad;
  	DbaseSumRcd.lDbaseErrorStatus  = ErrorStatus;
  	DbaseSumRcd.ulDbaseErrorType   = GlobalErrorFlags;
  	DbaseSumRcd.dbDbaseBadLimit    = dbBadLimit;
        
  	sum_PutRcd(SummDbIndex, &DbaseSumRcd, SumIndexFile);
  	sum_ReleaseDataBase(SumIndexFile);
  }

  exit(ExitStatus);

}

// ErrorHandle: Depending on the Verbose setting, this routine will
// will either exit(1) or print the message and continue on.
 
long ErrorHandle(char errorMessage[], int Error)
{
  ExitStatus = 1;
  bRecordOK = FALSE;
  lTotalBad++;

  DbaseError[Error].Count++;
  if(DbaseError[Error].Status == DBASE_FATAL) {
	bFatalError = TRUE;
  }

  GlobalErrorFlags |= DbaseError[Error].TypeFlag;

  if (Verbose) 
    fprintf(stderr, "Line %ld: %s\n", LineNo, errorMessage);
  
  return(0);

}

// UsageExit: Problems with the usage? Here you go:

void UsageExit()
{
  fprintf(stderr,usage);
  exit(1);
}
