/**************************************************************************
 Program    : s4drfi
 Programmer : Jeff Cobb
 Date       : Original 4/93: ported from s3 to s4 by Matt Lebofsky 3/2001

 History of modifications :
  02/25/01 - began porting from s3 to s4 by mattl
  04/02/01 - version 0.5 completed by mattl
  09/03/03 - version 0.6 cleaned up gcc 3.0 compiler warnings

**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <sys/types.h>
#include "SIREN.h"
#include "s4io.h"
#include "s4coord.h"
#include "s4time.h"
#include "s4summary.h"
#include "s4utils.h"

#ifndef D2R
#define D2R 0.017453292
#endif

#define S4DRFI_VERSION 0.6

#define SECS_IN_HOUR 3600
#define SECS_IN_DAY 86400
#define NUM_SUB_ARRAYS 100
#define t(i) events[i].time_tick
#define b(i) events[i].binno

typedef struct {
  double time_tick;
  double binno;
  double ra;
  double dec;
  long sector;
  long flag;
  long lineno;
  } event_rec;

typedef struct {
  long low_boundry;
  long high_boundry;
  } bin_boundries;

/* functions defined in this source file */
long num_spectra(event_rec *events, long  num_events, long sector);
long calc_ticks(long day_secs, long jday, long start_secs,
                long start_jday, long time_factor);
void reject_sector(event_rec *events, long first_event, long last_event,
		   long sector, long max_sector);
long calc_total_ticks(event_rec *events, long num_events, long gap_factor,
                      long time_factor);
int is_rfi(event_rec * events, double n, long first_event, long last_event,
	   long sector, double ra_res, double dec_res);
void get_count_thresh(long radius, long number_of_sectors, long dthresh,
		      double spectra_density, double *count_threshold);
int beam_res(double event_ra, double event_dec, double base_event_ra,
             double base_event_dec, double ra_res, double dec_res);

// routines used by qsort to sort by frequency/time of event_rec

static int by_frequency(const void *a, const void *b) {

  event_rec *event_a, *event_b;

  event_a = (event_rec *) a;
  event_b = (event_rec *) b;

  if (event_a->binno < event_b->binno) return(-1);
  if (event_a->binno > event_b->binno) return(1);

  return(0);

  }

static int by_time (const void *a, const void *b) {

  event_rec *event_a, *event_b;

  event_a = (event_rec *) a;
  event_b = (event_rec *) b;

  if (event_a->time_tick < event_b->time_tick) return(-1);
  if (event_a->time_tick > event_b->time_tick) return(1);

  return(0);

  }

static int by_lineno (const void *a, const void *b) {

  event_rec *event_a, *event_b;

  event_a = (event_rec *) a;
  event_b = (event_rec *) b;

  if (event_a->lineno < event_b->lineno) return(-1);
  if (event_a->lineno > event_b->lineno) return(1);

  return(0);

  }


int main(int argc, char *argv[])
  {
  /* command line parameters */
  /* required parameters     */
  long tel_diameter;      /* telescope diameter in feet */

  /* optional parameters with default values */
  long dthresh = 100;               /* rejection threshold in units of
				    mean density */
  long number_of_sectors = 10;    /* number of sectors in a 0 to 180
				    degree arch */
  long sec_radius = 500;          /* radius of semicircle in seconds */

  long beams = 5;                 /* number of beams that must seperate
				    events for flagging as rfi */
  double time_res = 2.0;          /* factor to multiply mean time
				    distance by in rfi check */
  long time_factor = 500;         /* time expansion factor for developing
				    'tick' units */
  long gap_factor = 10;           /* for figuring time gaps in input file */
  /* end of command line parameters */

  StdEventRec event;           /* for reading and writing single events */
  event_rec *events;            /* for array of event data */

  bin_boundries freq_subs[NUM_SUB_ARRAYS];     /* array for holding freq subarray boundries */

  double number_of_bins;        /* after 'unfolding' */

  FILE * fin_p;                 /* input file */
  FILE * frfi_p;                /* output file for rfi events */
  FILE * fclean_p;              /* output file for clean events */
  FILE * fout_summary = NULL;   /* output file for summary report */

  long events_in;               /* number of events in input file */

  double locos;       /* local oscillator - command line argument (use s4getlo) */
  long start_time;    /* event time of 1st event in input - decimal hours */
  long start_jday;    /* julian day of 1st event in input */

  long month, day, year;         /* for date calculations */
  double hour;                  /* for time calculations */

  double high_time_boundry;       /* for semicircle movement limits */
  double high_binno_boundry;
  double low_binno_boundry;
  double time_slice_boundry;
  double time_ticks_in_input;

  long radius;                  /* for spectra density and spectra count */
  long total_tick_time;         /*   calculations */
  long total_spectra;
  double spectra_density;
  double * count_threshold;
  long sector_count;

  double dec_res;                /* for beam resolution during rfi check */
  double ra_res;

  double distance;              /* for calculating which sector an event */
  double angle;                 /*   is in */

  long i, j, k;                 /* general loop control */
  long low, high, sub; 
  BOOLEAN highflag;

  long num_semicircles = 0;      /* record keeping */
  long rfi_out = 0;  
  long clean_out = 0;  

  long output_summary = 0;       /* set to true if to output summary file */
                                 /* do not output summary file by default */

  BOOLEAN updateDB = FALSE;      /* update summary db? default = no */ 
  DbaseSumRcd_t DbaseSumRcd; // Summary database record
  int SumIndexFile;
  unsigned long SummDbIndex;
  char sFileNameRoot[DBASE_FILENAME_LEN];

  char *usage = "Usage:\ns4drfi dish_diameter_in_feet local_oscillator\n\
 [-d density_threshold] [-s number_of_sectors] [-r radius_in_seconds]\n\
 [-b beams] [-f time_factor] [-t time_resolution]\n\
 [-g time_gap-factor] [-summary sum_filename]\n\
 [-db] [-ver] in_filename rfi_filename clean_filename\n\
 NOTE: FIRST TWO PARAMETERS MUST BE dish_diameter_in_feet and local_oscillator\n";

  /* check command line */
  if (argc < 6 || argc > 23) {
    fprintf(stderr,"%s\n",usage);
    exit(1);
    }

   /* get parameters */
   
  if (sscanf(argv[1],"%ld", &tel_diameter) != 1) {
    fprintf(stderr,"%s\n",usage);
    exit(1);
    }

  if (sscanf(argv[2],"%lf", &locos) != 1) {
    fprintf(stderr,"%s\n",usage);
    exit(1);
    }

  for (i = 3; i < (argc - 3); i++)
    if (strcmp(argv[i],"-d") == 0) {              /* density threshold */
      if (sscanf(argv[++i],"%ld",&dthresh) != 1) {
	fprintf(stderr, "Error in command line\n");
        fprintf(stderr,"%s\n",usage);
	exit(1);
	}
      }
    else if (strcmp(argv[i],"-s") == 0) {         /* number of sectors */
      if (sscanf(argv[++i],"%ld",&number_of_sectors) != 1) {
	fprintf(stderr, "Error in command line\n");
        fprintf(stderr,"%s\n",usage);
	exit(1);
	}
      }
    else if (strcmp(argv[i],"-r") == 0) {  /* semi-circle radius in seconds */
      if (sscanf(argv[++i],"%ld",&sec_radius) != 1) {
	fprintf(stderr, "Error in command line\n");
        fprintf(stderr,"%s\n",usage);
	exit(1);
	}
      }
      else if (strcmp(argv[i],"-b") == 0) {       /* number of beams */
      if (sscanf(argv[++i],"%ld",&beams) != 1) {
	fprintf(stderr, "Error in command line\n");
        fprintf(stderr,"%s\n",usage);
	exit(1);
	}
      }
    else if (strcmp(argv[i],"-f") == 0) {         /* time expansion factor */
      if (sscanf(argv[++i],"%ld",&time_factor) != 1) {
	fprintf(stderr, "Error in command line\n");
        fprintf(stderr,"%s\n",usage);
	exit(1);
	}
      }
    else if (strcmp(argv[i],"-t") == 0) {         /* time resolution */
      if (sscanf(argv[++i],"%lf",&time_res) != 1) {
        fprintf(stderr, "Error in command line\n");
        fprintf(stderr,"%s\n",usage);
        exit(1);
        }
      }
    else if (strcmp(argv[i],"-g") == 0) {            /* time gap factor */
      if (sscanf(argv[++i],"%ld",&gap_factor) != 1) {
	fprintf(stderr, "Error in command line\n");
        fprintf(stderr,"%s\n",usage);
	exit(1);
	}
      }
    else if (strcmp(argv[i],"-summary") == 0) {  /*open summary file and set flag to see if its open */
      if ((fout_summary = fopen(argv[++i], "a")) == NULL) {
	fprintf(stderr, "Can't open summary output file\n");
        fprintf(stderr,"%s\n",usage);
	exit(1);
	}
      output_summary = 1;
      }
    else if (strcmp(argv[i],"-db") == 0) {
      updateDB = TRUE;
      }
    else if (strcmp(argv[i],"-ver") == 0) {
      fprintf(stderr,"s4drfi version %4.2f\n",S4DRFI_VERSION);
      exit(1);
      }
    else {
      fprintf(stderr, "Unknown flag\n");   	 /* Error trap */
      fprintf(stderr,"%s\n",usage);
      exit(1);
      }

  /* unfold file giving true number of bins */
  number_of_bins = HIGH_CONTIGBIN - LOW_CONTIGBIN;
  
  /* Open input file. */
  if ((fin_p = fopen(argv[argc - 3],"r")) == NULL) {
    fprintf(stderr, "Can't open input file.\n");
    exit(1);
    }

  ut_GetRootFilename(argv[argc - 3],sFileNameRoot);

  /* Open rfi output file. */
  if ((frfi_p = fopen(argv[argc - 2],"w")) == NULL) {
    fprintf(stderr, "Can't open rfi output file.\n");
    exit(1);
    }

  /* Open clean output file. */
  if ((fclean_p = fopen(argv[argc - 1],"w")) == NULL) {
    fprintf(stderr, "Can't open clean output file.\n");
    exit(1);
    }

  if ((events_in = io_CountRecords(fin_p,STDEVENT_REC)) == -1) {
    fprintf(stderr,"Unexpected file size\n");
    exit(1);
    }

  if (events_in >= MAX_EVENTS) {
    fprintf(stderr, "Too many events\n");
    exit(1);
    }

  /* Allocate space for events array */
  if ((events = (event_rec *) calloc((unsigned)events_in, sizeof(StdEventRec))) == NULL) {
    fprintf(stderr, "Bad event allocation.\n");
    exit(1);
    }

  /* Allocate space for sector count array */
  if ((count_threshold = (double *) calloc((unsigned)number_of_sectors, sizeof(double))) == NULL) {
    fprintf(stderr, "Bad sector count array allocation.\n");
    exit(1);
    }

  /* Read event information from input file. */
  fprintf(stdout,"s4drfi reading %ld events\n",events_in);
  if (!io_GetNextRecord(fin_p,&event,STDEVENT_REC)) {
    fprintf(stderr, "Error - empty file\n");
    exit(1);
    }
  tm_HmsToDec(event.stUTime, &hour);
  start_time = (long)(hour * SECS_IN_HOUR); /* secs from beginning of day */
  tm_DateToI(event.stDate, &day, &month, &year);
  year = tm_CentYear(year);
  start_jday = (long)tm_ToJul(month, day, year, hour);
  events[0].time_tick = 0;
  events[0].binno = CONTIGBIN(event.lBinNumber,event.ucBid);
  events[0].ra = event.dbRa;
  events[0].dec = event.dbDec;
  events[0].sector = 0;
  events[0].flag = 0;
  events[0].lineno = 0;
 
  for (i = 1; i < events_in;i++) {
    if (!io_GetNextRecord(fin_p,&event,STDEVENT_REC)) {
      fprintf(stderr, "Bad data line\n");
      exit(1);
      }
    tm_DateToI(event.stDate, &day, &month, &year);
    year = tm_CentYear(year);
    tm_HmsToDec(event.stUTime, &hour);
    events[i].time_tick = calc_ticks((long)(hour * SECS_IN_HOUR),
				     (long)tm_ToJul(month, day, year, hour),
				     start_time,
				     start_jday,
				     time_factor);
    events[i].binno = CONTIGBIN(event.lBinNumber,event.ucBid);
    events[i].ra = event.dbRa;
    events[i].dec = event.dbDec;
    events[i].sector = 0;
    events[i].flag = 0;
    events[i].lineno = i;

    }

  /* figure semicircle size and movement limits */
  radius = sec_radius * time_factor;  /* radius of semicircle in ticks */
  high_time_boundry = events[events_in - 1].time_tick - radius;

  /* figure spectra density of file and calculate expected sector count */
  total_tick_time = calc_total_ticks(events, events_in,
				   gap_factor, time_factor);
  total_spectra = num_spectra(events, events_in, 0L);
  spectra_density = (total_spectra / number_of_bins) / total_tick_time;
  get_count_thresh(radius, number_of_sectors, dthresh,
		   spectra_density, count_threshold);

  time_ticks_in_input = events[events_in-1].time_tick;

  /* get beam resolution for rfi check */
  co_GetBeamRes(&dec_res, &ra_res, (long)locos, (unsigned)tel_diameter, (double)beams);

  // sort events array by frequency

  qsort ((char *) events, (int)events_in, sizeof(event_rec), by_frequency);

  // make sub array boundries

  // fprintf(stdout,"making frequency subarrays..\n");

  highflag = FALSE;
  for (i=0;i<NUM_SUB_ARRAYS;i++) {
    if (highflag) { freq_subs[i].low_boundry = 0; freq_subs[i].high_boundry = -1; }
    else {
      low = HIGH_CONTIGBIN + 1;
      high = LOW_CONTIGBIN - 1;
      for (j=0;j<events_in;j++) {
        if (events[j].binno >= ((i*(HIGH_CONTIGBIN/NUM_SUB_ARRAYS)) - radius)   &&
            events[j].binno <  (((i+1)*(HIGH_CONTIGBIN/NUM_SUB_ARRAYS)) + radius ) ) {
          if (events[j].binno < low) {
            freq_subs[i].low_boundry = j;
            low = (long)events[j].binno;
            }
          if (events[j].binno >= high) {
            freq_subs[i].high_boundry = j;
            high = (long)events[j].binno;
            }
          }
        }
      if (freq_subs[i].high_boundry == events_in - 1) { highflag = TRUE; }
      }
    // fprintf(stderr,"subarray: %d  low: %d  high: %d\n",i,freq_subs[i].low_boundry,freq_subs[i].high_boundry);
    }

  // see the low/high bin boundaries to the low/high bins, thereby losing
  // the "corner" effect towards the top of the file
  
  low_binno_boundry = LOW_CONTIGBIN;
  high_binno_boundry = HIGH_CONTIGBIN;

  /* begin main program loop - each iteration of i processes 1 semicircle */
  /* i is the midpoint along the base of the semicircle */

  // fprintf(stderr,"\ninto main loop..\n");

  for(sub = 0; sub < NUM_SUB_ARRAYS; sub++) {
    // fprintf(stderr,"doing subarray %d..\n",sub);

    if (freq_subs[sub].high_boundry > freq_subs[sub].low_boundry) { // only enter main loop if worth it

      qsort ((char *) &events[freq_subs[sub].low_boundry], 
             (int)((freq_subs[sub].high_boundry-freq_subs[sub].low_boundry)+1), sizeof(event_rec), by_time);

      for(i = freq_subs[sub].low_boundry; i <= freq_subs[sub].high_boundry; i++) {
        // if (i % 100 == 0) fprintf(stderr,"%7ld\b\b\b\b\b\b\b",i);
        if (events[i].binno >= low_binno_boundry && events[i].binno <= high_binno_boundry &&
            events[i].time_tick <= high_time_boundry) {
          time_slice_boundry = events[i].time_tick + radius;
          j = i;
        
          /* process all events within semicircle */
          for(j = i + 1; t(j) <= time_slice_boundry && j<=freq_subs[sub].high_boundry ; j++) {
            events[j].sector = -1; /* assume not within semicircle */
            if ( b(i)-b(j)<=radius && b(j)-b(i)<=radius) { 
	      distance =  sqrt ((t(i) - t(j)) * (t(i) - t(j)) +
                                (b(i) - b(j)) * (b(i) - b(j)));
	      if(distance <= radius) {    /* if within semicircle */
	        angle = acos((b(j) - b(i)) / distance);
	        events[j].sector = (long)(number_of_sectors * angle / PI);
	        }
              }
	    }  /* ends processing of all events within semicircle (j-loop) */
  
          j--;          /* back j up to be last event within semicircle */
  
          /* check semicircle for rfi */
          for(k = 0; k < number_of_sectors; k++) {
	    sector_count = num_spectra(&events[i], j - i + 1, k);
	    if(sector_count > count_threshold[k])
              if(is_rfi(events, time_res, i, j, k, ra_res, dec_res))
	        reject_sector(events, i, j, k, number_of_sectors - 1);
	    }  /* ends checking semicircle for rfi (k-loop) */
  
          num_semicircles++;
  
          }  /* ends if within time and binno boundries */
  
        }  /* end of main program loop (i-loop) */
  
      qsort ((char *) &events[freq_subs[sub].low_boundry], 
             (int)((freq_subs[sub].high_boundry-freq_subs[sub].low_boundry)+1), sizeof(event_rec), by_frequency);

      } /* end of if high_boundry>low_boundry */

    } /* end of sub array meta loop (sub-loop) */

  // re-sort events array by original line number

  qsort ((char *) events, (int)events_in, sizeof(event_rec), by_lineno);

  rewind(fin_p);                      /* back to start of input file */
  /* output rfi and clean files */
  for(i = 0; i < events_in; i++)
  {
    if (!io_GetNextRecord(fin_p,&event,STDEVENT_REC)) {
      fprintf(stderr, "Bad data line on second read\n");
      exit(1);
      }
    if(events[i].flag)
      {
      io_WriteStdEvent(frfi_p,&event);
      rfi_out++;
      }
    else
      {
      io_WriteStdEvent(fclean_p,&event);
      clean_out++;
      }
  }  /* ends outputting to rfi and clean files */
 
  fclose(fin_p);               /* close files */
  fclose(frfi_p);
  fclose(fclean_p);

  /* output stats */
  fprintf(stdout, "s4drfi (version %4.2f) - processing info for %s\n", S4DRFI_VERSION, argv[argc-3]);
  fprintf(stdout, "input parameters:\n");
  fprintf(stdout, 
          "  number of sectors = %5ld   radius in seconds     = %5ld\n",
          number_of_sectors, sec_radius);
  fprintf(stdout, 
          "  beams for beamres = %5ld   time expansion factor = %5ld\n",
          beams, time_factor);
  fprintf(stdout, 
          "  time resolution for determining related spectra     = %lf\n",
          time_res); 
  fprintf(stdout, 
          "  time gap threshold for calculating input time range = %5ld\n",
          gap_factor);
  fprintf(stdout, 
          "  rejection threshold in units of mean density        = %5ld\n",
          dthresh);
  fprintf(stdout, "calculated values:\n");
  fprintf(stdout, "  number of events = %7ld  number of spectra = %7ld\n",
          events_in, total_spectra);
  fprintf(stdout, "  spectra density of input = %e\n", spectra_density);
  fprintf(stdout, "  time ticks in input      = %lf\n", time_ticks_in_input);
  fprintf(stdout, "  number of semicircles processed = %ld\n", num_semicircles);
  fprintf(stdout, "  beam_res parms:  ra_res = %lf  dec_res = %lf\n",
          ra_res, dec_res);
  fprintf(stdout, "  number of rfi events   = %ld\n", rfi_out);
  fprintf(stdout, "  number of clean events = %ld\n", clean_out);
  if (rfi_out > clean_out)
    fprintf(stdout, "RFI EVENTS OUTNUMBERED CLEAN EVENTS FOR THIS FILE!!\n");
  fprintf(stdout, "end of s4drfi info\n");  

  // Add stuff to summary database if requested

  if (updateDB) {

    SumIndexFile = sum_GetDataBase(O_RDWR);
    SummDbIndex  = sum_GetRcd(sFileNameRoot, &DbaseSumRcd, SumIndexFile, O_RDWR);

    DbaseSumRcd.ulDrfiNumSectors = number_of_sectors;
    DbaseSumRcd.ulDrfiRadiusInSeconds = sec_radius;
    DbaseSumRcd.ulDrfiBeamsForBeamres = beams;
    DbaseSumRcd.ulDrfiTimeExpansionFactor = time_factor;
    DbaseSumRcd.dbDrfiTimeResolution = time_res;
    DbaseSumRcd.ulDrfiGapFactor = gap_factor;
    DbaseSumRcd.ulDrfiRejectionThreshold = dthresh;
    
    DbaseSumRcd.ulDrfiNumEvents = events_in;
    // DbaseSumRcd.ulNumSpectra = total_spectra; // already added in s4pcdbase
    DbaseSumRcd.dbDrfiSpectraDensity = spectra_density;
    DbaseSumRcd.dbDrfiTimeTicks = time_ticks_in_input;
    DbaseSumRcd.ulDrfiNumSemicircles = num_semicircles;
    DbaseSumRcd.dbDrfiRaRes = ra_res;
    DbaseSumRcd.dbDrfiDecRes = dec_res;
    DbaseSumRcd.ulDrfiNumRfiEvents = rfi_out;
    DbaseSumRcd.ulDrfiNumCleanEvents = clean_out;

    DbaseSumRcd.fS4drfiVers = S4DRFI_VERSION;
  
    sum_PutRcd(SummDbIndex, &DbaseSumRcd, SumIndexFile);
    sum_ReleaseDataBase(SumIndexFile);
  
    } /* end if updateDB */

  /* output summary */
  if (output_summary)
    {
    (void) fprintf(fout_summary,"Newprogram\n");
    (void) fprintf(fout_summary,"s4drfi:\n");
    (void) fprintf(fout_summary,"events_in %ld\n", events_in);
    (void) fprintf(fout_summary,"events_out %ld\n", clean_out);
    (void) fprintf(fout_summary,"spectra_in %ld\n", total_spectra);
    }

  return 0;
  }  /* end of main */


long calc_ticks(long day_secs, long jday, long start_secs,
                long start_jday, long time_factor) 
  {
  return (((SECS_IN_DAY - start_secs) +
	   (SECS_IN_DAY * (jday - start_jday - 1) +
            day_secs)) *
	  time_factor);
  }


long calc_total_ticks(event_rec *events, long num_events, long gap_factor,
                      long time_factor)
  {
  long raw_total_time;
  long total_gap_time;
  long gap_time;
  long gap;
  long i, j;

  gap = gap_factor * time_factor;
  total_gap_time = 0;
  raw_total_time = long(events[num_events - 1].time_tick);

  for(i = 0, j = 1; j < num_events; i++, j++)
    if((gap_time = long(events[j].time_tick - events[i].time_tick)) > gap)
      total_gap_time += gap_time;

  return ((raw_total_time - total_gap_time) - time_factor);
  }

long num_spectra(event_rec *events, long  num_events, long sector)
  {
  long spectra_count;
  long last_time;
  long i;

  spectra_count = 1;               /* assume at least one spectra */
  last_time = long(events[0].time_tick);  /* events[0] is start of a sub-array, */
				    /*   which may be the entire array! */

  for(i = 1; i < num_events; i++)
    if(events[i].sector == sector)
      if(events[i].time_tick != last_time)
	{
	spectra_count++;
	last_time = long(events[i].time_tick);
	}

  return spectra_count;
  }


void reject_sector(event_rec *events, long first_event, long last_event,
		   long sector, long max_sector)
  {
  long prev_sector;
  long next_sector;
  long i;

  if(sector == 0)
    prev_sector = sector;
  else
    prev_sector = sector - 1;

  if(sector == max_sector)
    next_sector = sector;
  else
    next_sector = sector + 1;

  for(i = first_event; i < last_event; i++)
    if(events[i].sector == sector ||
       events[i].sector == prev_sector ||
       events[i].sector == next_sector)
      events[i].flag = 1;
  }


int is_rfi(event_rec * events, double n, long first_event, long last_event,
	   long sector, double ra_res, double dec_res)
  {
  double total_time;         /* total time ticks in this sector */
  double n_mean_time_dist;   /* n times mean tick time between spectra */
  long this_time;            /* for calculating mean time ticks */
  long spectra_count;        /* number of spectra in this sector */
  long i;                    /* loop control */

  long first_spectra;        /* indices for sector walk */
  long this_spectra;
  long next_spectra;

#ifdef no_is_rfi
return TRUE;
#endif

  spectra_count = 1;
  total_time = 0;
  this_time = long(events[first_event].time_tick);

  /* count spectra and find, in ticks, the total time spread of this sector */
  for(i = first_event + 1; i <= last_event; i++)
    if(events[i].sector == sector)
      if(events[i].time_tick != this_time)
	{
	total_time += events[i].time_tick - this_time;
	this_time = long(events[i].time_tick);
	spectra_count++;
	}

  /* n(mean time distance between spectra) */
  n_mean_time_dist = (total_time / spectra_count) * n;

  next_spectra = first_event;       /* initialize */

  /* is_rfi main loop - first_spectra is earliest of related spectra group */
  /*                    this_spectra is latest of related spectra group    */
  while(next_spectra <= last_event)
    {
    first_spectra = next_spectra;  /* first_spectra = first/next group */
    this_spectra  = next_spectra;  /* same as first until forward walk */

    /* walk this_spectra up sector until until time jump from this to next */
    /* --- we have then found the bounds of a related spectra group        */
    while((events[next_spectra].time_tick - events[this_spectra].time_tick <=
	   n_mean_time_dist)
	  &&
	  (next_spectra <= last_event))
      {
      this_spectra = next_spectra;    /* move up the sector */

      /* get next spectra in this sector, if it exits */
      while((events[next_spectra].sector != sector ||
	     events[next_spectra].time_tick ==
	     events[this_spectra].time_tick)
	    &&
	    (next_spectra <= last_event))
	{
	  next_spectra++;
	}  /* ends get next spectra in this sector */

      }  /* ends walk this_spectra up the sector */

    /* if this group spans different sky points, we have rfi and can quit */
    if(beam_res(events[first_spectra].ra, events[first_spectra].dec,
		    events[this_spectra].ra, events[this_spectra].dec,
		    ra_res, dec_res))
	  return TRUE;

    }  /* ends is_rfi main loop */

  return FALSE;          /* we did not discover rfi in this sector */
  }  /* ends is_rfi */


void get_count_thresh(long radius, long number_of_sectors, long dthresh,
		      double spectra_density, double *count_threshold)
  {
  double c_sector_area;     /* sector area based on circle */
  int i;

  c_sector_area = (((PI *  pow((double)radius, 2.0)) /
		     2.0) /
		     number_of_sectors);

  for(i = 0; i < number_of_sectors; i++)
    count_threshold[i] = c_sector_area *
			 spectra_density *
			 dthresh;
  }

int beam_res(double event_ra, double event_dec, double base_event_ra,
             double base_event_dec, double ra_res, double dec_res)
  {
  float pdec;            /* used in converting declination to a positive  */
                         /*  number.                                      */
 
  if (fabs((double) event_dec - base_event_dec) > (double) dec_res)
    return(1);
  pdec = fabs((double) base_event_dec);
  if (pdec == 90)
    return(0);
  if (fabs((double) event_ra - base_event_ra)  >
      (double) ra_res/cos(pdec * PI / 180))
    return(1);
  else
    return(0);
  }
