/* ________________________________________________________________________
 programmer : Matt Lebofsky
 source name: s4engmonrecreader.cc
 compiler   : gcc
___________________________________________________________________________
 modification history:
   version 1.1 - added return value of 1 if the highest ADC value seen
                 during this execution is above 100.
   version 1.2 - added -q flag option
   version 2.0 - made max ADC user defineable (-max option)
                 added output if disabled of bad time ranges (a la s4adccheck)
   version 2.01 - added min/max (observed per file) to summary db functionality
   version 2.02 - cleaned up gcc 3.0 compiler warnings

___________________________________________________________________________
 usage: s4engmonrecreader [-q] [-sr] [-db] [-v]
                          [-max max_adc_value] [-min min_adc_value]
                          [-g gap_in_secs] dbase_eng_file

   Reads in unintelligible dbase_eng_file.
   readable engineering values comes out of stdout
   returns 1 is highest ADC is over max, lowest ADC under min, or if we
     were disabled (i.e. adc[8]==0) at any point, otherwise returns 0
   bad time ranges output to stderr
       
   The flag -q makes s4engmonrecreader quiet, i.e. it will NOT
   spew engineering values into stdout - this prevents having to
   deal with a "> /dev/null" if all you really care about is
   the return value. Default, of course, is "loud" output.

   The flag -sr makes s4engmonrecreader output (to stderr) any
   bad adc time ranges, either for ADCs 0-4 (greater than max)
   AND ADC 8==0 (i.e. diabled). These are in the form:

ADC starttime (jd) starttime (hms) endtime (jd) endtime (hms) mean ADC
--- -------------- --------------- ------------ ------------- --------

   Where ADC is 0-4 or 8, representing the respective eng value, starttime(s)
   are the beginning of the bad ADC range, endtimes are the end, and the
   mean ADC is obvious.

   If the problem is ADCs 0-4 being out of range, the average ADC value
   during this time is noted in mean ADC. If the problem is ADC 8==0,
   then mean ADC will equal 0, obviously.

   There are allowed "gaps" (default 100 seconds) within these bad ADC ranges
   of good data, that is, is it worth keeping 100 seconds of good data between
   two obvious ranges of bad data? This can be adjusted using the "-g" flag.
   Setting this to 0 means no gaps are allowed.

___________________________________________________________________________
notes/bugs: 

   I took out the board range feature since it was pointless and a
   headache to program.

___________________________________________________________________________
*/


#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include "SIREN.h"
#include "s4io.h"
#include "s4time.h"
#include "s4utils.h"
#include "s4summary.h"

#define ENGMONRECREADER_VERSION 2.02

typedef struct {
  double dbStartJul;    // where did the bad stuff start?
  char stStartTime[9];
  double dbEndJul;      // where did it end?
  char stEndTime[9];
  long lNumAdcs;        // how many ADCs are we talking about?
  long lTotalValue;     // total value to compute mean.
  long lBadEvents;       // simply, just the # of bad events counted
  } AdcRange;

void PrintRange(AdcRange PrintThis, long WhichAdc);

// ------------------------------------------------------------------------
int main (int argc, char *argv[])
// ------------------------------------------------------------------------
{

FILE *fEngFile = NULL;
// char stUTime[6];
EngMonRec ThisRecord;
int HighestADC,LowestADC;
int MaxADC = 199;
int MinADC = 6;
long i;
BOOLEAN bQuiet = FALSE;
BOOLEAN bShowRanges = FALSE;
BOOLEAN bDisabled = FALSE;
BOOLEAN bUpdateDB = FALSE;
long lGap = 100;
double dbRealGap;
// long lWhichValue;
double dbCurJulDate;
long lNumEng;          // total # of eng records
long lBadADCEvents;    // # of records with any ADC value 0-4 out of range
long lDisabledEvents;  // # of records while disabled
BOOLEAN bBadADCFlag;

DbaseSumRcd_t DbaseSumRcd; // Summary database record
int SumIndexFile;
unsigned long SummDbIndex;
char sEngFileNameRoot[20];

char *usage = "usage: s4engmonrecreader [-q] [-sr] [-db] [-v]\n                         [-max max_adc_value] [-min min_adc_value]\n                         [-g gap_in_secs] dbase_eng_file\n";

AdcRange RangeSet[10]; // Array of 10, since there are 10 ADC values,
                       // Though we really only care about a subset of them

for (i=0;i<10;i++) {
  RangeSet[i].dbStartJul = 0;
  RangeSet[i].dbEndJul = 0;
  RangeSet[i].lNumAdcs = 0;
  RangeSet[i].lTotalValue = 0;
  RangeSet[i].lBadEvents = 0;
  }

lNumEng = 0;
lBadADCEvents = 0;
lDisabledEvents = 0;

if (argc < 2) {
  fprintf(stderr,usage);
  exit(1);
  }

for (i = 1; i < argc; i++) {
  if (strcmp(argv[i],"-v") == 0) {
    fprintf(stderr,"s4engmonrecreader version %4.2f\n",ENGMONRECREADER_VERSION);
    exit(1);
    }
  else if (strcmp(argv[i],"-q") == 0) {
    bQuiet = TRUE;
    }
  else if (strcmp(argv[i],"-db") == 0) {
    bUpdateDB = TRUE;
    }
  else if (strcmp(argv[i],"-sr") == 0) {
    bShowRanges = TRUE;
    }
  else if (strcmp(argv[i],"-g") == 0) {
    if (sscanf(argv[++i],"%ld", &lGap) != 1) {
      fprintf(stderr,usage);
      exit(1);
      }  
    }  
  else if (strcmp(argv[i],"-max") == 0) {
    if (sscanf(argv[++i],"%d", &MaxADC) != 1) {
      fprintf(stderr,usage);
      exit(1);
      }
    }
  else if (strcmp(argv[i],"-min") == 0) {
    if (sscanf(argv[++i],"%d", &MinADC) != 1) {
      fprintf(stderr,usage);
      exit(1);
      }
    }
  else if ((fEngFile = fopen(argv[i],"r")) == NULL) {
         fprintf(stderr,"Can't read eng file: %s\n",argv[i]);
         exit(1);
    }
  }  

ut_GetRootFilename(argv[i-1], sEngFileNameRoot);

if (lGap < 0 || lGap > 100000) {
  fprintf(stderr,"The gap value you selected is very weird (<0 or >100000)\n");
  exit(1);
  }
 
dbRealGap = (lGap * SECS_TO_DAYS);

HighestADC = 0;
LowestADC = 254; // absurd values for initial comparison

for (i=0;i<9;i++) {
  RangeSet[i].dbStartJul = 0;
  RangeSet[i].dbEndJul = 0;
  RangeSet[i].lNumAdcs = 0;
  RangeSet[i].lTotalValue = 0;
  }

while(io_GetEngMonRecord(fEngFile,&ThisRecord)) {

  lNumEng++;
  bBadADCFlag = FALSE;

  if (!bQuiet) io_WriteEngMonRecord(stdout,&ThisRecord);

  // Quick checks strictly for return value

  for (i = 0; i <= 4; i++) {
    if (ThisRecord.ucEngMonitor[i] > HighestADC) {
      HighestADC = ThisRecord.ucEngMonitor[i];
      }
    if (ThisRecord.ucEngMonitor[i] < LowestADC) {
      LowestADC = ThisRecord.ucEngMonitor[i];
      }
    if (ThisRecord.ucEngMonitor[i] > MaxADC) {
      RangeSet[i].lBadEvents++;
      bBadADCFlag = TRUE;
      }
    if (ThisRecord.ucEngMonitor[i] < MinADC) {
      RangeSet[i].lBadEvents++;
      bBadADCFlag = TRUE;
      }
    }

  if (ThisRecord.ucEngMonitor[8] == 0) {
    bDisabled = TRUE;
    lDisabledEvents++;
    }

  dbCurJulDate = ThisRecord.dbJulDate;

  // Range checking

  for (i = 0; i < 9; i++) {
   
    if ( ((i < 5) && ((ThisRecord.ucEngMonitor[i] > MaxADC) || (ThisRecord.ucEngMonitor[i] < MinADC))) ||
         ((i == 8) && (ThisRecord.ucEngMonitor[i] == 0)) ) {

      if (RangeSet[i].dbStartJul == 0) { // beginning of bad range 
        RangeSet[i].dbStartJul = dbCurJulDate;
        strcpy(RangeSet[i].stStartTime,ThisRecord.stUTime);
        }

      RangeSet[i].dbEndJul = dbCurJulDate;
      strcpy(RangeSet[i].stEndTime,ThisRecord.stUTime);
      RangeSet[i].lNumAdcs++;
      RangeSet[i].lTotalValue+=ThisRecord.ucEngMonitor[i];

      } // end start range

    else { // may have good vals now, but may be within a gap
           // if so, simply hold off on printing bad range
           // because bad values may start up again soon

      if (((dbCurJulDate - RangeSet[i].dbEndJul) >= dbRealGap)
         && (RangeSet[i].dbStartJul > 0) ) {

        if (bShowRanges) PrintRange(RangeSet[i],i);
        RangeSet[i].dbStartJul = 0;
        RangeSet[i].lNumAdcs = 0;
        RangeSet[i].lTotalValue = 0;

        } // end if after gap

      } // end else (normal values)

    } // end ADC # for loop

  if (bBadADCFlag) lBadADCEvents++;

  } // end main while loop

for (i=0;i<9;i++)
  if (RangeSet[i].dbStartJul > 0)
    if (bShowRanges) PrintRange(RangeSet[i],i);

// Show counters if interested

if (!bQuiet) {
  fprintf(stderr,"# ##################################################\n");
  fprintf(stderr,"# s4engmonrecreader version %4.2f\n#\n",ENGMONRECREADER_VERSION);
  fprintf(stderr,"# Root filename: %s\n",sEngFileNameRoot);
  fprintf(stderr,"# Min accepted ADC value = %d\n",MinADC);
  fprintf(stderr,"# Max accepted ADC value = %d\n#\n",MaxADC);
  fprintf(stderr,"# Number of engineering records: %ld\n",lNumEng);
  fprintf(stderr,"# Number of records with bad ADC values: %ld\n",lBadADCEvents);
  for (i=0;i<=4;i++) {
    fprintf(stderr,"#    in board #%ld: %ld\n",i,RangeSet[i].lBadEvents); 
    }
  fprintf(stderr,"# Number of records while disabled: %ld\n",lDisabledEvents);
  fprintf(stderr,"# ##################################################\n");
  }

// Add stuff to summary database if requested

if (bUpdateDB) {

  SumIndexFile = sum_GetDataBase(O_RDWR);
  SummDbIndex  = sum_GetRcd(sEngFileNameRoot, &DbaseSumRcd, SumIndexFile, O_RDWR);

  DbaseSumRcd.lMinAcceptedADC      = MinADC;
  DbaseSumRcd.lMaxAcceptedADC      = MaxADC;
  DbaseSumRcd.lMinADC              = LowestADC;
  DbaseSumRcd.lMaxADC              = HighestADC;
  DbaseSumRcd.lADC0BadCount        = RangeSet[0].lBadEvents;
  DbaseSumRcd.lADC1BadCount        = RangeSet[1].lBadEvents;
  DbaseSumRcd.lADC2BadCount        = RangeSet[2].lBadEvents;
  DbaseSumRcd.lADC3BadCount        = RangeSet[3].lBadEvents;
  DbaseSumRcd.lADC4BadCount        = RangeSet[4].lBadEvents;
  DbaseSumRcd.lDisabledCount       = lDisabledEvents;
  DbaseSumRcd.fEngmonrecreaderVers = ENGMONRECREADER_VERSION;

  sum_PutRcd(SummDbIndex, &DbaseSumRcd, SumIndexFile);
  sum_ReleaseDataBase(SumIndexFile);

  }

// Any bad ADC's? Exit 1. Otherwise exit 0.

return((HighestADC > MaxADC || LowestADC < MinADC || bDisabled ) ? 1 : 0);

} // end main

// ------------------------------------------------------------------------
void PrintRange(AdcRange PrintThis, long WhichAdc)
// ------------------------------------------------------------------------
{
  fprintf(stderr,"# Bad Range: %ld ",WhichAdc);
  fprintf(stderr,"%13.10lf ",PrintThis.dbStartJul);
  fprintf(stderr,"%s ",PrintThis.stStartTime);
  fprintf(stderr,"%13.10lf ",PrintThis.dbEndJul);
  fprintf(stderr,"%s ",PrintThis.stEndTime);
  fprintf(stderr,"%ld\n",nlong(PrintThis.lTotalValue/PrintThis.lNumAdcs));
}
