/* ________________________________________________________________________
 programmer : Matt Lebofsky
 source name: s4exptime.cc 
 compiler   : gcc
___________________________________________________________________________
 modification history:
   version 1.0 finished 8/7/98
   version 2.0 finished 2/5/99 - added -j julian day flag - mattl
___________________________________________________________________________
 usage: s4exptime [-s] [-f] [-j] < dbase file > exptime file
   -s : offset from zero rather than beginning of month
   -f : print frequency rather than bin number
   -j : print julian date rather than offset seconds
___________________________________________________________________________
 description: s4exptime is a port from exptime, which does the exact
   same thing with s3 data. That is, it takes a pcdbase-generated file
   and output five fields:
   bin_number offset_seconds ra dec power
   
   The defaults are: the seconds will be offset from the beginning of
   the month, and bin_numbers will be output. The -s flag changes the
   offset to be from the beginning of the file, and the -f flag causes
   frequencies to be output. The -j flag causes Julian Date of each
   event to be in the offset_seconds field. The -s flag overrides -j.
___________________________________________________________________________
notes/bugs: 
___________________________________________________________________________
*/

#include <stdio.h>
#include <strings.h>
#include "SIREN.h"
#include "s4io.h"
#include "s4freq.h"
#include "s4power.h"
#include "s4time.h"

// ------------------------------------------------------------------------
int main (int argc, char *argv[])
// ------------------------------------------------------------------------
{

char *usage = "Usage: s4exptime [-s] [-f] [-j] < dbase_file > exptime_file\n";

StdEventRec ThisEvent;

long i;
double dbOffset,dbPower;
long day,month,year;
double hour;
double dbUTime;
double dbSeconds;

BOOLEAN bMonthOffset,bOutputBinNo,bFirstRecord,bJulianDate;

bMonthOffset = TRUE;
bOutputBinNo = TRUE;
bFirstRecord = TRUE;
bJulianDate = FALSE;

for (i = 1; i < argc ; i++) {
  if (strcmp("-s",argv[i]) == 0)
    bMonthOffset = FALSE;
  else if (strcmp("-f",argv[i]) == 0)
    bOutputBinNo = FALSE;
  else if (strcmp("-j",argv[i]) == 0)
    bJulianDate = TRUE;
  else {
    fprintf(stderr,"%s",usage);
    exit(1);
    }
  }

while (io_GetNextRecord(stdin,&ThisEvent,STDEVENT_REC)) {

  if (bOutputBinNo == FALSE)

/* The following is already calculated, so we comment it out:

    ThisEvent.dbFrequency = fq_s4fcalc(ThisEvent.dbFirstLo,
               ThisEvent.dbSecondLo, ThisEvent.ulFftBdFreq,
               ThisEvent.lBinNumber,ThisEvent.ucTid); 
*/

  tm_HmsToDec(ThisEvent.stUTime,&dbUTime);
  tm_FromJul(ThisEvent.dbJulDate,&month,&day,&year,&hour);

  if (bFirstRecord) {
    if (bMonthOffset == FALSE)
      dbOffset = ThisEvent.dbJulDate;
    else 
      dbOffset = tm_ToJul(month,(long)1,year,(double)0.0);
    bFirstRecord = FALSE;
    }
   
  dbPower = pw_PwrCalc(&ThisEvent,STDEVENT_REC);

  dbSeconds = (ThisEvent.dbJulDate - dbOffset) / SECS_TO_DAYS;

  if (bOutputBinNo) {
    if (bJulianDate) { // Output Bin Num and Julian Date
      printf("%10ld %12.10lf %9.5lf %8.4lf %4d\n",
             ThisEvent.lBinNumber,ThisEvent.dbJulDate,
             ThisEvent.dbRa,ThisEvent.dbDec,(long)dbPower);
      }
    else { // Output Bin Num and Offset Seconds (however set)
      printf("%10ld %12.4lf %9.5lf %8.4lf %4d\n",
             ThisEvent.lBinNumber,dbSeconds,
             ThisEvent.dbRa,ThisEvent.dbDec,(long)dbPower);
      }
    }
  else {
    if (bJulianDate) { // Output Freq and Julian Date
      printf("%15.6lf %12.10lf %9.5lf %8.4lf %4d\n",
             ThisEvent.dbFrequency,ThisEvent.dbJulDate,
             ThisEvent.dbRa,ThisEvent.dbDec,(long)dbPower);
      }
    else { // Output Freq and Offset Seconds (however set)
      printf("%15.6lf %12.4lf %9.5lf %8.4lf %4d\n",
           ThisEvent.dbFrequency,dbSeconds,
           ThisEvent.dbRa,ThisEvent.dbDec,(long)dbPower);
      }
  
    }

  } /* End main while loop */

return(0);
}
