/* ________________________________________________________________________
 programmer : Matt Lebofsky (ported code from s3)
 source name: s4multdetect.cc
 compiler   : gcc
___________________________________________________________________________
 modification history: 12/98 - first rev by mattl
                       12/01 - clean up by mattl
                       01/02/02 - Version 1.0 checked in
___________________________________________________________________________
 usage: s4multdetect spec|obs telescope_diameter_in_feet
      [-f freq_res] [-b num_beams] [-o output_option] infile_name

This program combines the functionality of both exdate and multhit into
one executable, being as how both used about 99% of the same code.

Reads a file in the form produced by 's4pcdbase'.  Extracts those events
that occur within user defined frequency and beam resolution, but depending
on whether the user selects 'spec' or 'obs':

spec: the events must also have occurred during DIFFERENT spectra (multhit)
obs: the events must also have occurred during DIFFERENT observations (exdate)

Note that this program should only be used on a file that has had frequency
correction performed on it to account for the movement of the earth
(i.e., solor system barycentric data).

The file is sorted by frequency.  A pointer is set to the next
unflagged event in the file.*  The file is scanned forward and backward
through 'FreqRes.' If two events within 'FreqRes' fall into the criteria
selected above, then each event within the group having the same ra and dec
as the base event is flagged FOR OUTPUT.  This process is repeated until the
end of the file is reached.  

     There are four output options. Events may be output in:
       1 groups
       2 order by frequency
       3 order by input file
       4 order by input file - events only (no additional text)

* This event is often referred to as the base event of the group.  

___________________________________________________________________________
notes/bugs:
___________________________________________________________________________
*/

#include <stdio.h>
#include <string.h>
#include "SIREN.h"
#include "s4freq.h"
#include "s4io.h"
#include "s4coord.h"
#include <sys/types.h>
#include <math.h>

char      *usage = "Usage: s4multdetect spec|obs dish_diameter_in_feet [-b num_beams]\n\
        [-f freq_res] [-o output_option] infile_name\n";

void FreqSort(long *lIndex, long lNum, MinEventRec *stEvents);
void CheckEvents(long lEventsIn, long *lGroupCnt, long *lIndex, 
            long *lLineNo, unsigned char cRecordType, FILE *fDbaseFile,
            MinEventRec *stEvents, long lTelDiameter, double dbNumBeams,
            double dbFreqRes, long lOutputOption, long lDetectOption);
void FlagEvents(long i, long lEventsIn, long *lIndex,  MinEventRec *stEvents,
                double dbRaRes,  double dbDecRes, double dbFreqRes);
void OutputGroup(long i, long lEventsIn, long *lIndex, long *lLineNo,
                 unsigned char cRecordType, FILE *fDbaseFile, 
                 MinEventRec *stEvents, double dbRaRes, double dbDecRes,
                 double dbFreqRes);
void OutputEvents(long *lIndex, MinEventRec *stEvents, long lEventsIn,
                  long *lLineNo, FILE *fDbaseFile, unsigned char cRecordtype,
                  long *lEventsOut, long lOutputOption);

void UsageExit();

long co_BeamRes(double dbEventRa, double dbEventDec, double dbBaseEventRa,
                double dbBaseEventDec, double dbRaRes, double dbDecRes);


// ------------------------------------------------------------------------
int main (int argc, char *argv[])
// ------------------------------------------------------------------------
{

FILE       *fDbaseFile;    /* Input file.                                  */
double      dbFreqRes;     /* Frequency resolution.                        */
long        lEventsIn=0;     /* Number of events in input file.              */
long        lEventsOut;    /* Number of events output by this program.     */
long 	    lDetectOption=1; /* Detect option 1 = diff spectrum 2 = diff obs */
long	    lOutputOption; /* Output option flag                           */
long        lGroupCnt=0;   /* Output option 1 (group output) count groups  */
long       *lIndex;        /* lIndex[i] points to the ith event as sorted by  */
                           /*  frequency.                                  */
long       *lLineNo;       /* lLineNo[i] points to line # from orig file   */
long        i;             /* General loop-control variable.               */
MinEventRec *stEvents;     /* Points to first record in the dynamically    */
                           /*  allocated list of event records.            */
unsigned char cRecordType; /* type of records we're reading in             */

/* Vars below are used in calculating sky frequency. */

// double      dbOffset;
// double      dbFreCvr;
// BOOLEAN     bInverted;           /* Boolean variable--frequency inverted?  */
// long        lFreqErrFlag = 0;    /* frequency error flag counts the        */
                                 /*  number of events that had to use      */
                                 /*  the default "frequency inverted".     */
                                 /*  Program could figure out if freq      */
                                 /*  is inverted or not                    */
double      dbNumBeams = 4;      /* number of beamwidths to allow a signal */
                                 /* to persist in */
long	    lTelDiameter;          /* telescope diameter in feet */

/* Default values of input parameters. */

dbFreqRes = 10;
lOutputOption = 1; /* default to output events in groups */

/* Process command line. */
if (argc != 4 && argc != 6 && argc != 8 && argc != 10)
  UsageExit();

  /* get telescope diameter */

  if (strcmp(argv[1],"spec") == 0) {
    lDetectOption = 1;
    }
  else if (strcmp(argv[1],"obs") == 0) {
    lDetectOption = 2;
    }
  else {
    fprintf(stderr,"User must define spec or obs\n");
    UsageExit();
    }
  
  if (sscanf(argv[2],"%ld",&lTelDiameter) != 1) {
    fprintf(stderr,"Error parsing the telescope diameter\n");
    UsageExit();
    }

  /* now get other parameters */

  for (i = 3;i < (argc - 2); i++)
    if (strcmp(argv[i],"-b") == 0) {             /* Right ascension res.     */
      if (sscanf(argv[++i],"%lf",&dbNumBeams) != 1) {
        fprintf(stderr,"Error parsing number of beams\n");
        UsageExit();
        }
      }
    else if (strcmp(argv[i],"-f") == 0) {        /* Frequency res.           */
      if (sscanf(argv[++i],"%lf",&dbFreqRes) != 1) {
        fprintf(stderr,"Error parsing frequency resolution\n");
        UsageExit();
        }
      }
    else if (strcmp(argv[i],"-o") == 0) {        /* Output option.           */
      if (sscanf(argv[++i],"%ld",&lOutputOption) != 1) {
        fprintf(stderr,"Error parsing output option\n");
        UsageExit();
        }
      }
    else {
      fprintf(stderr,"Unknown option: %s\n",argv[i]);
      UsageExit();                                 /* Error trap.            */
      }

  /* write output option type to stdout */

  switch (lOutputOption)
    {
    case 1  : fprintf(stdout,"events output in order by group\n");
              break;
    case 2  : fprintf(stdout,"events output in order by frequency\n");
              break;
    case 3  : fprintf(stdout,"events output in order by input file\n");
              break;
    case 4  : break;
    default : fprintf(stdout,"illegal output option\n");
              fprintf(stdout,"default to group output\n");
              lOutputOption = 1;
    }

  /* Open input file. */

  if ((fDbaseFile = fopen(argv[argc - 1],"r")) == NULL) {
    fprintf(stderr,"Can't open input file: %s\n",argv[argc - 1]);
    exit(1);
    }

  /* Check input file for record type and get number of events. */
  cRecordType = io_PeekNextRecord(fDbaseFile);
  if ((lEventsIn = io_CountRecords(fDbaseFile,cRecordType)) == -1) {
    fprintf(stderr,"Unexpected file size\n");
    exit(1);
    }
  rewind(fDbaseFile);

  /* Allocate space for events, index, and line # array */

  if ((stEvents = (MinEventRec *) calloc(lEventsIn,sizeof(MinEventRec))) == NULL) {
    fprintf(stderr,"Bad event allocation\n");
    exit(1);
    }
  if ((lIndex = (long *)calloc(lEventsIn,sizeof(long))) == NULL) {
    fprintf(stderr,"Bad index allocation\n");
    exit(1);
    }
  if ((lLineNo = (long *)calloc(lEventsIn,sizeof(long))) == NULL) {
    fprintf(stderr,"Bad line number array allocation\n");
    exit(1);
    }

  /* Read event information from input file. */

  if (lOutputOption != 4)
    (void) fprintf(stdout,"s4multdetect reading %ld events\n",lEventsIn);
  for (i = 0; i < lEventsIn;i++) {
    lLineNo[i] = ftell(fDbaseFile);
    io_GetNextMinEvent(fDbaseFile,&(stEvents[i]));

    // This is unnecessary - handled by io_GetNext{Min}Record - 
    // - an artifact from porting s3 code
    //
    // stEvents[i].dbFrequency = fq_s4fcalc(stEvents[i].dbFirstLo,
    //            stEvents[i].dbSecondLo, stEvents[i].ulFftBdFreq,
    //            stEvents[i].lBinNumber,stEvents[i].ucTid);

    /* Initialize event-flag and index array. */

    stEvents[i].cFlag = 0;
    lIndex[i] = i;
    } /* end reading events. */

   /* Sort events in ascending order by frequency. */

   FreqSort(lIndex,lEventsIn,stEvents);

   /* Flag events according to condition discussed in introductory comments. */

   CheckEvents(lEventsIn,&lGroupCnt,lIndex,lLineNo,cRecordType,fDbaseFile,
               stEvents,lTelDiameter, dbNumBeams,dbFreqRes,
               lOutputOption,lDetectOption);

   /* Output flagged events */

   OutputEvents(lIndex,stEvents,lEventsIn,lLineNo,fDbaseFile,
                cRecordType,&lEventsOut,lOutputOption);

   if (lOutputOption != 4) {
     (void) fprintf(stdout,"%ld events in\n",lEventsIn);

     /* Output paramaters etc. */

     if (lOutputOption != 1)
       (void) fprintf(stdout,"%ld events out\n",lEventsOut);
     else 
       (void) fprintf(stdout,"%ld groups out\n",lGroupCnt);
     (void) fprintf(stdout,"%lf freq res\n",dbFreqRes);
     (void) fprintf(stdout,"%lf beams in beam resolution\n",dbNumBeams);
     }

   /* Close input file. */

   (void) fclose(fDbaseFile);

  } /* end main */

/* shell sorts the index array in accending order by frequency */

void FreqSort(long *lIndex, long lNum, MinEventRec *stEvents)

  {
  long gap, i, j;
  long utemp;

  for (gap = lNum / 2; gap > 0; gap /= 2)
    for (i = gap; i < lNum; i++)
      for (j = i - gap; j >= 0; j -= gap) {
        if (stEvents[lIndex[j]].dbFrequency <=
            stEvents[lIndex[j + gap]].dbFrequency)
          break;
        utemp = lIndex[j];
        lIndex[j] = lIndex[j + gap];
        lIndex[j + gap] = utemp;
        }
  }

  /*
   * Procedure 'CheckEvents':
   * Is passed a pointer to an index array (sorted by frequency), a pointer to
   * the array of events, the number of events, and the various resolutions.
   * From the next unflagged event, the file is scanned forward and backward
   * through all events within 'FreqRes'.  See notes at top about criteria
   * fore selection. If events meet these criteria, then this procedure
   * calls 'FlagEvents' which flags all events within 'FreqRes' of the base
   * event that have the same ra and dec as the base event. 
   * NOTE: if OutputOption == 1 then this procedure calls 'OutputGroup()'
   *       which outputs groups instead of flaging the groups.  Flag events
   *       is also called in this case so as to flag events that have already
   *       been output.                                                        
   */

void CheckEvents(long lEventsIn,  /* Number of events in file.            */
      long *lGroupCnt,            /* Count groups if output option 1      */
      long *lIndex,               /* Index array sorted by sky frequency. */
      long *lLineNo,              /* line # array                         */
      unsigned char cRecordType,  /* record types */
      FILE *fDbaseFile,           /* orig dbase file */
      MinEventRec *stEvents,      /* Array of event records.              */
      long lTelDiameter,          /* telescope diameter in feet */
      double dbNumBeams,          /* number of tel beams in beam resolution */
      double dbFreqRes,           /* Sky frequency resolution.            */
      long lOutputOption,         /* for optional output by groups, freq, */
                                  /*   or input file order.               */
      long lDetectOption)         /* 1 = DIFF spectra 2 = DIFF obs        */
  {
  long i,j;     /* Loop control variables.                                   */
  long lFlagged; /* Set to true if group eventshave already been flagged. */
  double dbRaRes;     /* Right ascension resolution.          */
  double dbDecRes;    /* Declination resolution.              */
  double dbLocalOsc;  /* used for keeping track of local oscilator setting */

  dbLocalOsc = -1;  /* initialize so as to cause computation of beamwidth for
                     * first event */

  /* process data */

  for (i = 0; i < lEventsIn; i++) {
    
    /* If event already flagged then skip it. */

    if (stEvents[lIndex[i]].cFlag)
      continue;

    /* compute ra_res and dec_res */
    /* FOR NOW, JUST GETTING FREQUENCY OF FIRST EVENT) */
    if (dbLocalOsc == -1) {  
      dbLocalOsc = stEvents[lIndex[i]].dbFrequency;
      co_GetBeamRes(&dbDecRes,&dbRaRes,(long)(dbLocalOsc/1000000),
                    lTelDiameter,dbNumBeams);
      }  

    lFlagged = 0;

    /* Scan forward */

    for (j = i + 1;
         j < lEventsIn &&
         stEvents[lIndex[j]].dbFrequency - stEvents[lIndex[i]].dbFrequency <= dbFreqRes;
         j++)
      if (!co_BeamRes(stEvents[lIndex[j]].dbRa,stEvents[lIndex[j]].dbDec,
                       stEvents[lIndex[i]].dbRa,stEvents[lIndex[i]].dbDec,
                       dbRaRes,dbDecRes))

        if ( ((lDetectOption == 1) &&
              (stEvents[lIndex[i]].dbJulDate !=
               stEvents[lIndex[j]].dbJulDate)) ||
 
             ((lDetectOption == 2) &&
             (fabs(stEvents[lIndex[i]].dbJulDate -
              stEvents[lIndex[j]].dbJulDate) > OBS_SEPARATION)) ) {

          if (lOutputOption == 1) { 
            (*lGroupCnt)++;
            OutputGroup(i,lEventsIn,lIndex,lLineNo,cRecordType,fDbaseFile,
                        stEvents,dbRaRes,dbDecRes,dbFreqRes);
            } 

          FlagEvents(i,lEventsIn,lIndex,stEvents,dbRaRes,dbDecRes,
                     dbFreqRes);
          lFlagged = 1;   /* Events in this group are flagged */
          break;

          }

    /* Scan backward if events not already flagged */

    if (!lFlagged)
      for (j = i - 1;
           j >= 0 &&
           stEvents[lIndex[i]].dbFrequency - stEvents[lIndex[j]].dbFrequency <=
dbFreqRes;
           j--)
      if (!co_BeamRes(stEvents[lIndex[j]].dbRa,stEvents[lIndex[j]].dbDec,
                        stEvents[lIndex[i]].dbRa,stEvents[lIndex[i]].dbDec,
                        dbRaRes,dbDecRes)) 

        if ( ((lDetectOption == 1) &&
              (stEvents[lIndex[i]].dbJulDate !=
               stEvents[lIndex[j]].dbJulDate)) ||
 
             ((lDetectOption == 2) &&
             (fabs(stEvents[lIndex[i]].dbJulDate -
              stEvents[lIndex[j]].dbJulDate) > OBS_SEPARATION)) ) {

          if (lOutputOption == 1) { 
            (*lGroupCnt)++;
            OutputGroup(i,lEventsIn,lIndex,lLineNo,cRecordType,fDbaseFile,
                        stEvents,dbRaRes,dbDecRes,dbFreqRes);
            } 

          FlagEvents(i,lEventsIn,lIndex,stEvents,dbRaRes,dbDecRes,
                     dbFreqRes);
          break;

          }

    } /* end main for loop */
  } /* end procedure */

/*
 * Procedure 'flag_events':
 * Is passed a pointer to an index array (sorted by frequency), a pointer to
 * the array of events, the number of events, and the various resolutions.
 * Also passed is position of the base event of a group.  Scans events within
 * the group and flags all events having the same ra and dec as the base event. 
 */

void FlagEvents(long i,          /* Position of base event.              */
                long lEventsIn,  /* Number of events in file.            */
                long *lIndex,    /* Index array sorted by sky frequency. */
                MinEventRec *stEvents, /* Array of event records.        */
                double dbRaRes,  /* Right ascension resolution.          */
                double dbDecRes, /* Declination resolution.              */
                double dbFreqRes) /* Sky frequenct resolution.           */
  {
  long j; /* Loop controle variable. */
 
  stEvents[lIndex[i]].cFlag = 1;  /* Flag base event. */
 
  /* Scan forward */

  for (j = i + 1;
       j < lEventsIn &&
       stEvents[lIndex[j]].dbFrequency - stEvents[lIndex[i]].dbFrequency <= dbFreqRes;
       j++)
    if (!co_BeamRes(stEvents[lIndex[j]].dbRa,stEvents[lIndex[j]].dbDec,
                    stEvents[lIndex[i]].dbRa,stEvents[lIndex[i]].dbDec,
                    dbRaRes,dbDecRes))
      stEvents[lIndex[j]].cFlag = 1;


    /* Scan backward */

    for (j = i - 1;
         j >= 0 &&
         stEvents[lIndex[i]].dbFrequency - stEvents[lIndex[j]].dbFrequency <= dbFreqRes;
         j--)
      if (!co_BeamRes(stEvents[lIndex[j]].dbRa,stEvents[lIndex[j]].dbDec,
                      stEvents[lIndex[i]].dbRa,stEvents[lIndex[i]].dbDec,
                      dbRaRes,dbDecRes))
        stEvents[lIndex[j]].cFlag = 1;
 

  } /* End prodecdure flag events. */


/*
 * procedure outputs all events within frequency resolution of the base
 * event and within beam resolution of the base event
 * groups are output in order by frequency
 */

void OutputGroup(long i,          /* Position of base event.              */
                 long lEventsIn,  /* Number of events in file.            */
                 long *lIndex,    /* Index array sorted by sky frequency. */
                 long *lLineNo,          /* line #s of events in.                 */
                 unsigned char cRecordType, /* record type                    */
                 FILE *fDbaseFile, /* orig dbase file */
                 MinEventRec *stEvents, /* Array of event records.        */
                 double dbRaRes,  /* Right ascension resolution.          */
                 double dbDecRes, /* Declination resolution.              */
                 double dbFreqRes)  /* Sky frequenct resolution.          */
  {
    long j,k;             /* loop control variables               */

    k = i;  /* point k to base event of group */

    /* Scan backward, point 'j + 1' to first event in group */


    for (j = k - 1;
         j >= 0 &&
         stEvents[lIndex[i]].dbFrequency - stEvents[lIndex[j]].dbFrequency <= dbFreqRes;
         j--);

    for (k = j + 1;
         k < lEventsIn &&
         fabs(stEvents[lIndex[i]].dbFrequency - stEvents[lIndex[k]].dbFrequency) <= dbFreqRes;
         k++)
         if(!co_BeamRes(stEvents[lIndex[k]].dbRa,stEvents[lIndex[k]].dbDec,
                        stEvents[lIndex[i]].dbRa,stEvents[lIndex[i]].dbDec,
                        dbRaRes,dbDecRes))
           io_WriteRecFromFile(stdout,fDbaseFile,lLineNo[lIndex[k]],cRecordType);
 
    fprintf(stdout,"# #####\n");
  } /* end function OutputGroup */

/*
 * Procedure 'output_events':
 * Counts and outputs all flagged events in order by frequency or in order
 * by input file as specified by 'output_option.  
 */

void OutputEvents(long *lIndex,     /* points to array of freq-sorted events */
             MinEventRec *stEvents, /* Pointer to array of events in memory. */
             long lEventsIn,        /* Number of events in.                  */
             long *lLineNo,          /* line #s of events in.                 */
             FILE *fDbaseFile,      /* orig dbase file */
             unsigned char cRecordType, /* record type                    */
             long *lEventsOut,      /* Number of events output.              */
             long lOutputOption)
  {
  long i;                                      /* Index variable.           */

if (lOutputOption == 2)     /* output events in order by frequency */
  {
  for (i = 0; i < lEventsIn; i++)              /* For each event do:        */
    {                                          /*                           */
    if (stEvents[lIndex[i]].cFlag)            /* If an event has been      */
      {                                        /* flagged, put to stdout    */
        io_WriteRecFromFile(stdout,fDbaseFile,lLineNo[lIndex[i]],cRecordType);
        (*lEventsOut)++;
      }  
    } /* End for loop */
  }  /* end if */

  else if (lOutputOption == 3 ||
           lOutputOption == 4)  /* output events in order by input file */
  {
  for (i = 0;i < lEventsIn; i++)               /* For each event do:        */
    {                                          /*                           */
    if (stEvents[i].cFlag)                    /* If an event has been      */
      {                                        /* flagged, put to stdout    */
        io_WriteRecFromFile(stdout,fDbaseFile,lLineNo[lIndex[i]],cRecordType);
        (*lEventsOut)++;
      }  
    } /* End for loop */
  } /* end else if */

  } /* End function 'OutputEvents' */

// UsageExit: Problems with the usage? Here you go:

void UsageExit()
{
  fprintf(stderr,usage);
  exit(1);
}

// co_BeamRes - to be added to s4coord.cc when checked back in.
// ported from s3utils.c
// Function compares two events to deterimine if they fall outside a beam
// resolution determined by 'dbDecRes' and 'dbRaRes'/cos(dec).  Function
// returns 1 if beam resolution is violated.  This function does not account
// for the case in which right ascension just wraps around from 24 to 0 hrs.
// This case is considered so rare as to not be worth checking.
 
long co_BeamRes(double dbEventRa, double dbEventDec,
                double dbBaseEventRa, double dbBaseEventDec,
                double dbRaRes, double dbDecRes)
{
 
  double dbPDec; // positive declination

  if (fabs(dbEventDec - dbBaseEventDec) > dbDecRes)
    return(1);
  dbPDec = fabs(dbBaseEventDec);
  if (dbPDec == 90)
    return(0);
  if (fabs(dbEventRa - dbBaseEventRa)  >
      dbRaRes / cos(dbPDec * PI / 180))
    return(1);
  else
    return(0);
  }

