// Source Code Name  : s4pcdbase.cc
// Programmer        : Cobb/Donnelly
//                   : Project SERENDIP, University of California, Berkeley CA
// Compiler          : GNU gcc

/*
Note: see man s4pcdbase for full discussion of this program.

   History:
   06/01/97  alpha   jeffc     basic version ready for instrument install 
                               at AO
   11/10/97  beta    jeffc     convert s4time from AST to UT, other minor 
                               changes and bug fixes.
   12/01/97  v1.00   jeffc     new function - CalcRaDec to handle
                               complex coordinate calculations.
   01/09/98  v1.01   jeffc     Time/date corrections
   01/13/98  v1.02   jeffc     Tidying up and commenting.
   01/14/98  v1.03   jeffc     Scope string change.
   05/08/98  v1.10   jeffc     Bug fix - check to see if HitHr is on other side
                               of midnight from aft scope string in CalcRaDec().
   06/11/98  v1.11   jeffc     Moved stMonths[] to s4time.cc with a "static" 
                               place holder in s4time.h.
   08/06/98  v1.12   jeffc     Bug fix - backward wrap in RA was not handled.
                               Changes to GetRaSlope() and CalcRaDec().
   08/10/98  v1.13   jeffc     Allow the search for scope strings to continue
                               upon finding a zero length scope string.
                               Changes to main() and GetNextScopeString(). 
   12/18/98  v1.14   jeffc     BoardTable is now 2d.  Moved TiD defs to s4tel.h.
   03/05/99  v1.15   jeffc     ZenAz correction coeffecients now a 2d table.  Added
                               in 21cm feed corrections.
   07/20/99  v2.00   jeffc     Added support for the summary database.  This includes
                               new routines SlewStats(), UpdateDataBase(), and 
                               TestDataBase(), as well as hooks in the main loop and
                               in routine Fini(). Note that I moved 2 main() variables,
                               InputDataFileName and dVersioni, into the global section
                               but in the interest of time and testing, did not 
                               change the calls that pass these around as local
                               variables.  A speedup of a small degree may be had by
                               moving the start and lasttime updates out of the 
                               hit writting loop and into the spectra loop case, but
                               you would need to check for the "between scope strings"
                               there.
    2/28/00  v2.10	jeffc	Fixed problem that arose when we get consecutive
				duplicate scope strings.  In that case, we would
				end up getting a NaN for SlewRate.  New code in
				SlewStats() checks for this and writes the number
				of dup strings to the summary db.
				Added db fields for total hits and number of hits
				lost dut to no scope string.
				Also, the year is now inserted into the engeneering
				record when written out.  		
   2/29/00   v2.11	jeffc	Place julian day into summary record.		
   3/6/00    v2.12	jeffc	Added -db flag.		
   3/14/00   v2.13	jeffc	Added spectra count and error counts to summary db
   4/18/00   v3.00 	jeffc   Added engineering monitor ring buffer so that we can
                                grab the monitors produced 4 integrations ago.	
  11/10/00   v3.01      jeffc   Added kilocount hit fields for all boards and full res
  				hit count for the seti@home board (board 19).
  12/1/00    v3.02	jeffc   Do not test access to summary DB unless -db flag is
				given.
  01/5/00    v3.03      jeffc   Record first and second LO in summary DB.
  01/7/00    v3.04      jeffc   Fixed a bug in Fini().  It was using the most recent HitDay
                                in the file for calculating Start Day (julian) in the 
                                summary DB.  It now uses the first HitDay.  Also, we now
	                        round Kilohits rather than flooring them.
  9/19/03   v3.05	jeffc   Fixed bug in SlewStats() that was resulting in absurd (giant 
				negative) mean slew rates.  We were not initializing
				the variable SlewRateTotal.
*/

#define VERSION 3.05

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fstream.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#include "SIREN.h"
#include "s4io.h"
#include "s4time.h"
#include "s4freq.h"
#include "s4coord.h"
#include "s4boardtable.h"
#include "s4tel.h"
#include "s4database.h"
#include "s4summary.h"
#include "s4utils.h"

char *usage = "s4pcdbase [-s] [-db] -lo1 FirstLo -lo2 SecondLo -id ScopeId S4_binary_file";

//----------------------- cross platform types -------------------

// #define IBMPC
#define SUN
#ifdef IBMPC
#define U16 unsigned
#define U32 unsigned long
#define S16 int
#define S32 long
#define BYTE unsigned char
#endif
#ifdef SUN
#define U16 unsigned short
#define U32 unsigned long
#define S16 short int
#define S32 long
#define BYTE unsigned char
#endif

//----------------------- Data types  --------------------------

// ScopeInfo holds data obtained or calculated from the 
// serial telescope string.
typedef struct
  {
  double dRa;
  double dDec;
  double dRaSlope;
  double dDecSlope;
  double dLo1;
  double dLo2;
  double dDecHour;
  long lYear;
  long lCentYear;
  long lDay;
  long lHour;
  long lMin;
  long lSec;
  char sReceiverID[3];
  char padding[5];
  } ScopeInfo;

typedef struct
  {
  long CentYear; 
  long Day; 
  double Hr;
  unsigned char ucEngMons[10];
  } EngBuf_t;
  
//----------------------- function prototypes -------------------

BOOLEAN ReadNext(char *buf,long n);
void ReadNextLine(char *buf,long max_n,char eol);
void Init(char * InputDataFileName);
void ErrExit(char *msg);
void ChangeEndian(unsigned char *var,int size);
void ConvertTime2StdEventRec(long u32Time,
                             StdEventRec * ThisHit,
                             long * HitDay,
                             double * HitHr,
                             long Year); 
void ExitFunction(void);
void WriteToLog(char *thisMessage);
double GetRaSlope(double Ra1, double Ra2, double Hr1, double Hr2);
double GetDecSlope(double Dec1, double Dec2, double Hr1, double Hr2);
void FixScopeInfo(ScopeInfo   * SA, long SAI, char * ScopeStr);
void SlewStats(ScopeInfo   * SA, long NumSA,DbaseSumRcd_t * DbaseSumRcd);
void CalcRaDec(ScopeInfo   * SA,
               long          Aft,
               StdEventRec * ThisHit,
               double        HitHr);
long FindScopeStrings(double      HitHr,
                      long        HitDay,
                      ScopeInfo * SA,
                      long        MaxSAI,
                      long *      Aft,  
                      long *      Forward);
long GetNextScopeString();
void Fini();
void GetOps(int       argc,
             char    * argv[],
             BOOLEAN * bSwapBytes,
             BOOLEAN * bUpdateDb,
             double  * dFirstLo,
             double  * dSecondLo,
             char    * sReceiverID,
             unsigned short * usSigmaThreshold, 
             char   ** InputDataFileName);
void CmdError (void);
BOOLEAN GoodBlockType(U16 u16BlockType);
void UpdateDataBase(char * InputDataFileName, DbaseSumRcd_t * DbaseSumRcd,
                    DbaseSumRcd_t * ThisDbaseSumRcd);
void TestDataBase();
void PushEngBuf(long EngCentYear,
                long EngDay,
                double EngHr,
                unsigned char * ucEngMons,
                EngBuf_t *  EngBuf,
                long * EngBufFowardPtr,
                long * EngBufAftPtr);

//----------------------- misc. defines  -----------------------

#define FOREVER         1
#define JD_JAN1_1970    2440587.5
#define SIZEOF_HITS     10
#define HITS_OVERHEAD   15
#define NUM_ENG_MONS    10
#define NUM_ENG_BUFS    5

// MIN_SLEW_DEG_PER_HR defines tracking time as entered into the 
// summary database.  1.5 deg/hr is 1/10 sidereal
#define MIN_SLEW_DEG_PER_HR 1.5

//------------- data block types for data files  ---------------

#define SCSTR  0x0400  // telescope string
#define CMMT   0x0500  // comment
#define HITS   0x0A00  // hits
#define SYSCFG 0X0B00  // system configuration info
#define ERR    0X1000  // error
#define ENG    0X2000  // engineering record
#define THROTTLE        0x4000  // data thinning
                                // additional data: 1 double

#define ERR_SYNTH        0X01
#define ERR_VECT_WARN    0X02
#define ERR_OVER_RUN     0X03
#define ERR_ARITH_OVFL   0X04
#define ERR_OTHER        0X05
#define ERR_BD_NOT_READY 0X06
#define BUFFA            0X01
#define BUFFB            0X02

//------------------- Global variables  -----------------------

char CommentStr[80];
char ConfigStr[128];
char ScopeStr[80];
char UtilStr255[256]; // one extra for terminator

U32 u32Synthesizer;
U32 u32Time;
U16 u16ADC;
BYTE btBuffer;

StdEventRec ThisHit;
StdEventRec DummyHit;
FILE *fInputDataFile;
FILE *fLogFile;
int fEngFile;
long FilePosition, NumToRead, NumAvailable;
off_t CurrentFileSize;
char LogFileName[512];
char EngFileName[512];
BOOLEAN RecordStarted;
BOOLEAN ScanningForScope;
BOOLEAN bSwapBytes;
BOOLEAN bUpdateDb;
long glNoStringHits;
long glNumHits;

unsigned char ucEngMons[10] = {0,1,2,3,4,5,6,7,8,9};

static unsigned long ulBoardCount[NUMBOARDS];  // init to zeros

struct stat statbuf;

DbaseSumRcd_t DbaseSumRcd, ThisDbaseSumRcd;       // Summary database record  
char * InputDataFileName; 
double dVersion = VERSION;
double FinalHitTime;
long StartHitYear;
long StartHitDay;
long HitDay;
long EngDay;

//____________________________________________________________________
// MAIN
//--------------------------------------------------------------------
int main (int argc,char **argv)
  {

  unsigned long i;

  //long lDay, lMonth, lYear, lHour, lMin, lSec, lScopeTime;
  //double dHour, dScopeJulDay;

  //char * InputDataFileName; 

  long NextScopeLen;

  //long CurFilePos;

  long BoardTab;

  //double RaSlope, DecSlope;
  double dFirstLo;
  double dSecondLo;
  unsigned short usSigmaThreshold;

  double Throttle;

  double HitHr;
  //double AstHitHr;
  //double LastHitHr    = 0.0;
  //double LastAstHitHr = 0.0;
  long HitCentYear=0;

  double EngHr;
  long   EngCentYear;

  ScopeInfo * SA;
  long        SAI;
  long        MaxSAI;
  long        Aft;
  long        Forward;

  struct DateStruct sdate;

  char sReceiverID[32];

  long lHaveScopeStrings;
  long lHaveScopeStringsEng;

  bSwapBytes = FALSE;
  bUpdateDb  = FALSE;

  glNoStringHits = 0;
  glNumHits      = 0;

  EngBuf_t EngBuf[NUM_ENG_BUFS];
  long     EngBufFowardIndex =  0;
  long     EngBufAftIndex    = -1;

  //DbaseSumRcd_t DbaseSumRcd, ThisDbaseSumRcd;       // Summary database record  

  TestDataBase();

  GetOps(argc, argv, 
         &bSwapBytes, 
         &bUpdateDb, 
         &dFirstLo, 
         &dSecondLo, 
         sReceiverID, 
         &usSigmaThreshold,
         &InputDataFileName);

  Init(InputDataFileName); 
  ThisDbaseSumRcd.dbStartTime = 99.0;    // first time switch
  ThisDbaseSumRcd.usDupScopeStrings = 0;
  ThisDbaseSumRcd.ulNumSpectra = 0;
  ThisDbaseSumRcd.ulTotalHits = 0;
  ThisDbaseSumRcd.ulNoStringHits = 0;
  ThisDbaseSumRcd.ulErrSynth = 0;
  ThisDbaseSumRcd.ulErrVectWarn = 0;
  ThisDbaseSumRcd.ulErrOverRun = 0;
  ThisDbaseSumRcd.ulErrArithOvfl = 0;
  ThisDbaseSumRcd.ulErrBdNotReady = 0;
  ThisDbaseSumRcd.ulErrOther = 0;


  // program version to log and summary database
  sprintf(UtilStr255,">>>>>>>>> s4pcdbase version %5.2lf <<<<<<<<<\n",
          dVersion);
  WriteToLog(UtilStr255);
  ThisDbaseSumRcd.fPcdbaseVers   =  dVersion;

  // Get items specified on command line - they won't change
  ThisHit.dbFirstLo  = dFirstLo;
  ThisHit.dbSecondLo = dSecondLo;
  ThisHit.usSigmaThreshold = usSigmaThreshold;
  ThisDbaseSumRcd.dbFirstLo  = dFirstLo;
  ThisDbaseSumRcd.dbSecondLo = dSecondLo;
  ThisDbaseSumRcd.usSigmaThreshold = usSigmaThreshold;
  if((ThisHit.ucTid = tel_GetTid(sReceiverID)) == UNKNOWN_TEL)
    { 
    sprintf(UtilStr255, "Unknown telescope ID\n");
    WriteToLog(UtilStr255);
    ErrExit("Unknown telescope ID");
    }

  // Use scope ID to choose board table
  if((BoardTab = tel_GetTelTab(ThisHit.ucTid)) == -1)
    ErrExit("Unable to get board table index.\n");

  // ---------------- Get Scope Strings -------------------
  // Because scope strings are asynchronous with spectral
  // data, we must read them all into a big array ahead of
  // time. The scope array grows dynamically, as we do not
  // know up front how many scope strings there are.
  // With version 1.13 of this program, zero length scope
  // strings are allowed - we just skip them and continue
  // the scope string search.
  // SA is the scope array.
  // SAI is the scope array index.
  ScanningForScope = TRUE;       // checked by reading and exit routines 
  SA = (ScopeInfo *)calloc(1, sizeof(ScopeInfo));
  SAI = 0;
  while((NextScopeLen = GetNextScopeString()) >= 0)
    {
    //U16 u16BlockSize;
    ReadNext(ScopeStr, (U16)NextScopeLen);
    if(NextScopeLen > 0)
      {
      ScopeStr[NextScopeLen] = '\0';
      FixScopeInfo(SA, SAI, ScopeStr);
      if((SA = (ScopeInfo *)realloc(SA, sizeof(ScopeInfo) * (SAI + 2))) == NULL)
        {
        fprintf(stderr, "Unable to realloc scope array.\n");
        exit(1);
        }
      else
        SAI++;
      } // end if(NextScopeLen > 0) 
    else
      {
      sprintf(CommentStr, 
              "SCOPE STRING %ld IS OF ZERO LENGTH (%ld bytes into file)\n", 
              SAI+1, ftell(fInputDataFile) - 4);
      WriteToLog(CommentStr);       // to pcdbase log
      fprintf(stderr, CommentStr);  // to user or pipeline log
      } 
    } // end while GetNextScopeString()
  MaxSAI = SAI;
  rewind(fInputDataFile);       // reset input file back to beginning
  ScanningForScope = FALSE;
  SlewStats(SA, MaxSAI, &ThisDbaseSumRcd);     // go summarize slew statistics


  // ----------------    Main Loop    -------------------
  while (FOREVER)
     {
     U16 u16BlockType;
     U16 u16BlockSize;
     //U16 u16Reader;
     BYTE btReader;
     //U32 u32Reader;

     unsigned uNumberOfHits;

     RecordStarted = FALSE;

     if(ReadNext((char *)&u16BlockSize,sizeof(U16)));
       RecordStarted = TRUE;

     if (bSwapBytes)
	  ChangeEndian((unsigned char*)&u16BlockSize,sizeof(U16));

     ReadNext((char *)&u16BlockType,sizeof(U16));

     if (bSwapBytes)
	  ChangeEndian((unsigned char *)&u16BlockType,sizeof(U16));

       switch (u16BlockType)
	 {
         case THROTTLE :
                   ReadNext((char *)&Throttle, sizeof(double));
                   if (bSwapBytes)
                      ChangeEndian((unsigned char *)&Throttle,sizeof(double));
                   sprintf(CommentStr, "----> throttle factor is %lf <----\n", Throttle);
                   WriteToLog(CommentStr);
                   break;
	 case HITS :
                   // First read the header which has data applicable to
                   // all hits in this spectra (1 hit record = 1 spectra).
		   uNumberOfHits = (u16BlockSize - HITS_OVERHEAD) / SIZEOF_HITS;
		   ReadNext((char *)&u32Time,sizeof(U32));

		   if (bSwapBytes)
		      ChangeEndian((unsigned char *)&u32Time,sizeof(U32));

                   ConvertTime2StdEventRec(u32Time, 
                                           &ThisHit, 
                                           &HitDay, 
                                           &HitHr, 
                                           SA[0].lYear);

		   ReadNext((char *)&u32Synthesizer,sizeof(U32));
		   if (bSwapBytes)
		      ChangeEndian((unsigned char *)&u32Synthesizer,sizeof(U32));
                   ThisHit.dbBirdieSynth = u32Synthesizer / 1000.0;  // KHz to MHz

		   ReadNext((char *)&u16ADC,sizeof(U16));
		   if (bSwapBytes)
		      ChangeEndian((unsigned char *)&u16ADC,sizeof(U16));
                   ThisHit.usAdc = u16ADC;
                 
		   ReadNext((char *) &btBuffer,sizeof(BYTE));

                   // ----------- Complete new spectra info ---------
                   lHaveScopeStrings = FindScopeStrings(HitHr,
                                                        HitDay,
                                                        SA,
                                                        MaxSAI,
                                                        &Aft,
                                                        &Forward);

                   if(lHaveScopeStrings)
                     {
                     HitCentYear = lHaveScopeStrings == 2 ? 
                                   SA[Aft].lCentYear : SA[Forward].lCentYear;
                     tm_DayToMon(HitDay-1, HitCentYear, &sdate);
                     sprintf(ThisHit.stDate, "%02ld%s%02ld", 
                             sdate.day, 
                             stMonths[sdate.month], 
                             tm_TwoDigitYear(HitCentYear));
                     // Get coordinates from scope most recent scope string
                     CalcRaDec(SA,
                               Aft,
                               &ThisHit,
                               HitHr);
                     // prevent rounding up to 24.0 when printing to dbase
                     ThisHit.dbRa = 24.0 - ThisHit.dbRa < 0.00001 ?
                                    0.0 : ThisHit.dbRa;

                     // count only spectra that we will write to file
		     ThisDbaseSumRcd.ulNumSpectra++;
                     }

                   // Now get data for individual hits in this spectra.
		   for (i=0;i<uNumberOfHits;i++)
		       {
		       ReadNext((char *)&ThisHit.lBinNumber,sizeof(U32));
		       if (bSwapBytes)
			  ChangeEndian((unsigned char *)&ThisHit.lBinNumber,sizeof(U32));
		       ReadNext((char *)&ThisHit.usPower,sizeof(U16));
		       if (bSwapBytes)
			  ChangeEndian((unsigned char *)&ThisHit.usPower,sizeof(U16));

		       ReadNext((char *)&ThisHit.usActualThreshold,sizeof(U16));
		       if (bSwapBytes)
			  ChangeEndian((unsigned char *)&ThisHit.usActualThreshold,
                                       sizeof(U16));
                       ThisHit.usSpctResLog = 0; // clear (high byte) before read 
		       ReadNext((char *)&ThisHit.usSpctResLog,sizeof(BYTE));

		       ReadNext((char *)&ThisHit.ucBid,sizeof(BYTE));

                       // Do NOT write hit unless we are between scope strings
                       // and the engineering values for this hit.
                       if(lHaveScopeStrings && EngBufAftIndex != -1)
                         {

#ifdef DEBUG_ENGMONS
fprintf(stderr, "Hit = %lf Eng = %lf  Diff = %lf  %ld  ", HitHr, EngBuf[EngBufAftIndex].Hr, fabs(HitHr-EngBuf[EngBufAftIndex].Hr)*3600.0, EngBufAftIndex);
fprintf(stderr, "%hu %hu %hu %hu %hu %hu %hu %hu %hu %hu\n", 
	EngBuf[EngBufAftIndex].ucEngMons[0],
	EngBuf[EngBufAftIndex].ucEngMons[1],
	EngBuf[EngBufAftIndex].ucEngMons[2],
	EngBuf[EngBufAftIndex].ucEngMons[3],
	EngBuf[EngBufAftIndex].ucEngMons[4],
	EngBuf[EngBufAftIndex].ucEngMons[5],
	EngBuf[EngBufAftIndex].ucEngMons[6],
	EngBuf[EngBufAftIndex].ucEngMons[7],
	EngBuf[EngBufAftIndex].ucEngMons[8],
	EngBuf[EngBufAftIndex].ucEngMons[9]);
#endif

                         // get times for summary database - we only want times
                         // for records written
                         if(ThisDbaseSumRcd.dbStartTime == 99.0)
                           {
                           ThisDbaseSumRcd.dbStartTime = HitHr;
			   StartHitYear = HitCentYear; 
                           StartHitDay  = HitDay;
                           }
	                 FinalHitTime = HitHr;

                         // Get items keyed by board ID
                         ThisHit.ulFftBdFreq = fq_GetBrdFrq(BoardTab, ThisHit.ucBid);
                         ThisHit.cPolarization = 
                           BoardTable[BoardTab][ThisHit.ucBid].cPolarization;
                         ThisHit.usAdc = 
                           EngBuf[EngBufAftIndex].ucEngMons[BoardTable[BoardTab][ThisHit.ucBid].lAdcIndex];
			 ulBoardCount[ThisHit.ucBid]++;		// running board hit count

         	         io_WriteStdEvent(stdout,&ThisHit);

                         glNumHits++;
                         } 
                       else
                         glNoStringHits++;

		       }
		   break;
	 case CMMT :   WriteToLog("======Comment==================\n");
		   ReadNext(CommentStr,u16BlockSize-4);
		   CommentStr[u16BlockSize-4] = '\0';
		   WriteToLog(CommentStr);
		   break;
	 case SYSCFG : WriteToLog("::::::System Configuration:::::\n");
		   int bytesRead;
		   bytesRead = 0;
		   while (bytesRead < u16BlockSize-4)
		     {
                     ReadNextLine(ConfigStr,128,(char)'\0');
		     bytesRead += strlen(ConfigStr) + 1;
		     WriteToLog(ConfigStr);
		     }
		   break;
         case ENG : ReadNext(UtilStr255,u16BlockSize-4);
                    memcpy(&u32Time, UtilStr255, sizeof(long));
                    memcpy(ucEngMons, UtilStr255+4, NUM_ENG_MONS * sizeof(char)); 
                    if (bSwapBytes)
                      ChangeEndian((unsigned char *)&u32Time,sizeof(U32));
                    ConvertTime2StdEventRec(u32Time,
                                            &DummyHit,
                                            &EngDay,
                                            &EngHr,
                                            SA[0].lYear);

                    lHaveScopeStringsEng = FindScopeStrings(EngHr,
                                                         EngDay,
                                                         SA,
                                                         MaxSAI,
                                                         &Aft,
                                                         &Forward);
		    if(lHaveScopeStringsEng)
                      {
                      EngCentYear = lHaveScopeStringsEng == 2 ?
                                    SA[Aft].lCentYear : SA[Forward].lCentYear;
                      tm_DayToMon(EngDay-1, EngCentYear, &sdate);

                      memcpy(UtilStr255, &u32Time, sizeof(long)); 
                      memcpy(UtilStr255+4, &EngCentYear, sizeof(long)); 
                      memcpy(UtilStr255+8, ucEngMons, NUM_ENG_MONS * sizeof(char)); 
                      PushEngBuf(EngCentYear, 
                                 EngDay, 
                                 EngHr, 
                                 ucEngMons, 
                                 EngBuf, 
                                 &EngBufFowardIndex, 
                                 &EngBufAftIndex);

#ifdef DEBUG_ENGMONS
fprintf(stderr, "hr = %lf >>> %lf %lf %lf %lf %lf <<< %ld %ld\n", EngHr, EngBuf[0].Hr, EngBuf[1].Hr, EngBuf[2].Hr, EngBuf[3].Hr, EngBuf[4].Hr, EngBufFowardIndex, EngBufAftIndex);
#endif

                      write(fEngFile, UtilStr255, u16BlockSize-4+4);
                      }
                   break;
	 case ERR :    WriteToLog( "******Error********************\n");
		   ReadNext((char *)&btReader,sizeof(BYTE));

		   switch (btReader)
		      {
		      case ERR_SYNTH :
			   WriteToLog("Synthisizer error\n" );
			   ThisDbaseSumRcd.ulErrSynth++;	
			   break;
		      case ERR_VECT_WARN :
			   ReadNext((char *)&btReader,sizeof(BYTE));
                           sprintf(UtilStr255,"Vector Warning error on board: %u\n",
				 (unsigned) btReader);
                           WriteToLog(UtilStr255);
			   ThisDbaseSumRcd.ulErrVectWarn++;	
			   break;
		      case ERR_OVER_RUN :
			   ReadNext((char *)&btReader,sizeof(BYTE));
			   sprintf(UtilStr255, "Over run error on board: %u\n ",
				 (unsigned) btReader);
                           WriteToLog(UtilStr255);
			   ThisDbaseSumRcd.ulErrOverRun++;	
			   break;
		      case ERR_ARITH_OVFL :
			   ReadNext((char *)&btReader,sizeof(BYTE));
			   sprintf(UtilStr255, 
                                   "Arithmetic overflow error on board: %u\n ", 
				   (unsigned) btReader );
                           WriteToLog(UtilStr255);
			   ThisDbaseSumRcd.ulErrArithOvfl++;	
			   break;
		      case ERR_OTHER :
			   ReadNext((char *)&btReader,sizeof(BYTE));
			   sprintf(UtilStr255, "Other error on board:  %u\n", 
				 (unsigned) btReader);
                           WriteToLog(UtilStr255);
			   ThisDbaseSumRcd.ulErrOther++;
			   break;
		      case ERR_BD_NOT_READY :
			   ReadNext((char *)&btReader,sizeof(BYTE));
			   sprintf(UtilStr255, "Board not ready error on board:  %u\n",
				 (unsigned) btReader );
                           WriteToLog(UtilStr255);
			   ThisDbaseSumRcd.ulErrBdNotReady++;	
			   break;
		      }
		   break;
	 case SCSTR : WriteToLog( "/\\/\\/\\Telescope String/\\/\\/\\/\\/\\/\n" );
		      ReadNext(ScopeStr,u16BlockSize-4);
		      WriteToLog(ScopeStr);
		      break;

	 default : sprintf(UtilStr255, 
                    "Unknown.....\nblock type: %x\nblock size: %x\nfile position: %ld\n",
                    u16BlockType, u16BlockSize, ftell(fInputDataFile));
                   WriteToLog(UtilStr255);
                   ErrExit("Unexpected block type"); 
		   break;
       }

     }

  }


//--------------------------------------------------------------------
void TestDataBase()
//--------------------------------------------------------------------
// The database routines will exit(1) on problem detection.
  {
  int i;
  if(bUpdateDb == FALSE) return;
  i = sum_GetDataBase(O_RDWR);
  sum_ReleaseDataBase(i);
  }

//--------------------------------------------------------------------
void UpdateDataBase(char * InputDataFileName, DbaseSumRcd_t * DbaseSumRcd,
                       DbaseSumRcd_t * ThisDbaseSumRcd)
//--------------------------------------------------------------------
  {
  int SumIndexFile;
  unsigned long SummDbIndex;

  int i;

  char sDbaseFileNameRoot[DBASE_FILENAME_LEN];

  if(bUpdateDb == FALSE) return;

  /* strip off path to dbase file */
  ut_GetLastField('/', InputDataFileName, sDbaseFileNameRoot);

  SumIndexFile = sum_GetDataBase(O_RDWR);
  SummDbIndex  = sum_GetRcd(sDbaseFileNameRoot, DbaseSumRcd, SumIndexFile, O_RDWR);

  DbaseSumRcd->fPcdbaseVers         = ThisDbaseSumRcd->fPcdbaseVers;
  DbaseSumRcd->usSigmaThreshold     = ThisDbaseSumRcd->usSigmaThreshold;
  DbaseSumRcd->dbFirstLo     	    = ThisDbaseSumRcd->dbFirstLo;
  DbaseSumRcd->dbSecondLo     	    = ThisDbaseSumRcd->dbSecondLo;
  DbaseSumRcd->dbMeanSlew           = ThisDbaseSumRcd->dbMeanSlew;
  DbaseSumRcd->dbStartTime          = ThisDbaseSumRcd->dbStartTime;
  DbaseSumRcd->dbTrackingTime       = ThisDbaseSumRcd->dbTrackingTime;
  DbaseSumRcd->dbStartRa            = ThisDbaseSumRcd->dbStartRa;
  DbaseSumRcd->dbEndRa              = ThisDbaseSumRcd->dbEndRa;
  DbaseSumRcd->dbStartDec           = ThisDbaseSumRcd->dbStartDec;
  DbaseSumRcd->dbEndDec             = ThisDbaseSumRcd->dbEndDec;
  DbaseSumRcd->dbFileDuration       = ThisDbaseSumRcd->dbFileDuration;
  DbaseSumRcd->usDupScopeStrings    = ThisDbaseSumRcd->usDupScopeStrings;
  DbaseSumRcd->dbMeanScopeStringInterval = ThisDbaseSumRcd->dbMeanScopeStringInterval;
  DbaseSumRcd->ulTotalHits          = glNumHits;
  DbaseSumRcd->ulNoStringHits       = glNoStringHits;
  DbaseSumRcd->ulNumSpectra         = ThisDbaseSumRcd->ulNumSpectra;
  DbaseSumRcd->ulErrSynth           = ThisDbaseSumRcd->ulErrSynth;
  DbaseSumRcd->ulErrVectWarn        = ThisDbaseSumRcd->ulErrVectWarn;
  DbaseSumRcd->ulErrOverRun         = ThisDbaseSumRcd->ulErrOverRun;
  DbaseSumRcd->ulErrArithOvfl       = ThisDbaseSumRcd->ulErrArithOvfl;
  DbaseSumRcd->ulErrBdNotReady      = ThisDbaseSumRcd->ulErrBdNotReady;
  DbaseSumRcd->ulErrOther           = ThisDbaseSumRcd->ulErrOther;
  DbaseSumRcd->ulSetiAtHomeBoardCount = ulBoardCount[SETI_AT_HOME_BOARD];
  for(i = 0; i < NUMBOARDS; i++) {
    if((unsigned long)((double)ulBoardCount[i]/1000 + 0.5) > 255) 
      DbaseSumRcd->ucBoardKiloCount[i] = 255;
    else
      DbaseSumRcd->ucBoardKiloCount[i] = (unsigned)((double)ulBoardCount[i]/1000 + 0.5);
  }

  sum_PutRcd(SummDbIndex, DbaseSumRcd, SumIndexFile);   
  sum_ReleaseDataBase(SumIndexFile);
  }


//--------------------------------------------------------------------
void ErrExit(char *msg)
//--------------------------------------------------------------------
  {
  fprintf(stderr,"%s\n", msg);
  fprintf(stderr, "usage: %s\n",usage );
  exit(1);
  }


//--------------------------------------------------------------------
void ChangeEndian(unsigned char *var,int size)
//--------------------------------------------------------------------
  {
  unsigned char myByteStr[8];
  int i;

  // simple, but it even works for odd sizes!
  if (size <= 8)
    {
    for (i=0;i<size;i++)
      myByteStr[i] = var[size-1-i];
    for (i=0;i<size;i++)
      var[i] = myByteStr[i];
    }

  }

//--------------------------------------------------------------------
void Init(char * InputDataFileName)
//--------------------------------------------------------------------
  {
  //strcpy(InputDataFileName,filename);
  fInputDataFile = fopen(InputDataFileName,"r");
  if (fInputDataFile == NULL)
    ErrExit("Can't open data file");
  fstat(fileno(fInputDataFile),&statbuf);
  CurrentFileSize = statbuf.st_size;
  FilePosition = 0;
  strcpy(LogFileName,InputDataFileName);
  strcat(LogFileName,".log");
  strcpy(EngFileName,InputDataFileName);
  strcat(EngFileName,".eng");
  if((fEngFile = open(EngFileName, O_RDWR | O_CREAT | O_TRUNC, 0664)) == -1)
    ErrExit("Can't open engineering file");
  }

//--------------------------------------------------------------------
BOOLEAN ReadNext(char *buf,long n)
//--------------------------------------------------------------------
  {
  FilePosition = ftell(fInputDataFile);
  NumAvailable = CurrentFileSize - FilePosition;
  if(NumAvailable < n)
    // If we are just getting scope strings
    // then we have just run out of strings
    // and need to carry on.  But if we are
    // in the main data reading loop, we 
    // are done!
    if(ScanningForScope)
      return(FALSE);
    else 
      {
      Fini();		// Fini() does not return - it exit()'s.
      return(FALSE);    // This is here just to avoid a compiler
                        //   warning about reaching the end of
                        //   nonvoid function.
      }
  else
    {
    fread(buf,sizeof(char),(int) n,fInputDataFile);
    FilePosition = ftell(fInputDataFile);
    return TRUE;
    }
  }


//--------------------------------------------------------------------
// reads next line, or up to n characters (whichever comes first)
// from global file stream, End of line is eol.
void ReadNextLine(char *buf,long max_n,char eol)
//--------------------------------------------------------------------
   {
   unsigned char NextChar;
   int NumCharsRead;
   NumCharsRead = 1;
   ReadNext((char *)&NextChar,1);
   buf[NumCharsRead-1] = NextChar;
   while((NextChar != eol) && (NumCharsRead < max_n)) 
      {
      NumCharsRead++;
      ReadNext((char *)&NextChar,1);
      buf[NumCharsRead-1] = NextChar;
      }
   }


//--------------------------------------------------------------------
void ConvertTime2StdEventRec(long u32Time, 
                             StdEventRec * ThisHit,
                             long * HitDay,
                             double * HitHr,
                             long Year)
//--------------------------------------------------------------------
// Binary time format:
//      31      23      18     12     6      2     0
//      --------------------------------------------
//      | DOY   | hour  | min  | sec  | 10th | src |
//      --------------------------------------------
//       9 bits  5 bits  6 bits 6 bits 4 bits 2 bits
  {
  long DOY, Hour, Min, Sec, Ten;

  DOY  = (0x000001ff) & (u32Time >> 23);
  Hour = (0x0000001f) & (u32Time >> 18);
  Min  = (0x0000003f) & (u32Time >> 12);
  Sec  = (0x0000003f) & (u32Time >>  6);
  Ten  = (0x0000000f) & (u32Time >>  2);

  tm_AstToUt(&Year, &DOY, &Hour);

  sprintf(ThisHit->stUTime,   "%02lu", Hour); 
  sprintf(ThisHit->stUTime+2, "%02lu", Min);
  sprintf(ThisHit->stUTime+4, "%02lu", Sec);
  sprintf(ThisHit->stUTime+6, "%02lu", Ten*10);    // hundredths of secs

  *HitDay = DOY;

  *HitHr  = (double)Hour        + 
            (double)Min/60.0    + 
            (double)Sec/3600.0  + 
            (double)Ten/36000.0;                  // return decimal hour
  }

//--------------------------------------------------------------------
void ExitFunction(void)
//--------------------------------------------------------------------
  {
  fclose(fLogFile);
  }


//--------------------------------------------------------------------
void WriteToLog(char *thisMessage)
//--------------------------------------------------------------------
  {
  fLogFile = fopen(LogFileName,"a+");
  if (fLogFile == NULL)
    ErrExit("Can't open log file");
  fprintf(fLogFile,"%s\n",thisMessage);
  fclose(fLogFile);
  }

//--------------------------------------------------------------------
double GetRaSlope(double Ra1, double Ra2, double Hr1, double Hr2)
//--------------------------------------------------------------------
// If the RAs in 2 consecutive strings are more than 12 hours apart,
// we assume that RA has wrapped, ie they will never *really* be more
// than 12 hours apart.  If Ra2 is less than Ra1, then the wrap is
// from 23hr -> 0hr (forward wrap).  If Ra2 is greater than Ra1, then
// the wrap is 0hr -> 23hr (backward wrap).  For a forward wrap, the 
// slope needs to be positive and this is achieved by adding 24 to *Ra2*.
// For a backward wrap, the slope needs to be negative and this is achieved 
// by adding 24 to *Ra1*.
  {
  double RaDiff, HrDiff;

  // RA can wrap
  // fprintf(stderr,"===============================\n");
  // fprintf(stderr,"Hr1 = %lf  Hr2 = %lf  Ra1 = %lf  Ra2 = %lf\n", Hr1, Hr2, Ra1, Ra2);

  if(fabs(Ra1 - Ra2) > 12.00)    // wrap?
    if(Ra2 < Ra1)                // forward wrap
      Ra2 += 24.00;
    else
      Ra1 += 24.00;              // backward wrap
  RaDiff = Ra2 - Ra1;            // can be negative
 
  HrDiff = (Hr2 < Hr1 ? Hr2 + 24.00 : Hr2) - Hr1;
 
  // fprintf(stderr,"Hr1 = %lf  Hr2 = %lf  Ra1 = %lf  Ra2 = %lf  slope = %lf\n", Hr1, Hr2, Ra1, Ra2, RaDiff / HrDiff);
  // fprintf(stderr,"===============================\n");

  return(RaDiff / HrDiff);   // return slope
  }

//--------------------------------------------------------------------
double GetDecSlope(double Dec1, double Dec2, double Hr1, double Hr2)
//--------------------------------------------------------------------
  {
  double DecDiff, HrDiff;

  // Assume not near poles - Dec cannot wrap
  DecDiff = Dec2 - Dec1;        // can be negative 
 
  HrDiff = (Hr2 < Hr1 ? Hr2 + 24.00 : Hr2) - Hr1;
 
  return(DecDiff / HrDiff);   // return slope
  }


//--------------------------------------------------------------------
void CalcRaDec(ScopeInfo   * SA, 
               long          Aft,
               StdEventRec * ThisHit,
               double        HitHr)
// This RA  = This Ra + (RA slope * Delta Time)
// This Dec = This Dec + (Dec slope * Delta Time)
//--------------------------------------------------------------------
  {
  double DeltaTime;

  DeltaTime = HitHr - SA[Aft].dDecHour;

  // HitHr can be on the other side of midnight from the Aft scope string
  DeltaTime = DeltaTime < 0.0 ? DeltaTime + 24.0 : DeltaTime;

  // fprintf(stderr, "Ra = %lf  Dec = %lf  HitHr = %lf  DeltaTime = %lf RaSlope = %lf\n", 
  //         SA[Aft].dRa, SA[Aft].dDec, HitHr, DeltaTime, SA[Aft].dRaSlope);

  // Calc RA and then take care of wrap in either direction.
  ThisHit->dbRa   = SA[Aft].dRa + SA[Aft].dRaSlope * DeltaTime;
  ThisHit->dbRa = ThisHit->dbRa >= 24.0 ?
                  ThisHit->dbRa - 24.0 : ThisHit->dbRa;
  ThisHit->dbRa = ThisHit->dbRa  <  0.0 ?
                  ThisHit->dbRa + 24.0 : ThisHit->dbRa;

  // Calc Dec.  Dec cannot wrap (we assume we are not near poles).
  ThisHit->dbDec   = SA[Aft].dDec + SA[Aft].dDecSlope * DeltaTime;
  // fprintf(stderr, "Ra = %lf  Dec = %lf\n\n", ThisHit->dbRa, ThisHit->dbDec);

  }

//--------------------------------------------------------------------
void FixScopeInfo(ScopeInfo   * SA, long SAI, char * ScopeStr)
//--------------------------------------------------------------------
// All scope string data calculations are done here.
// The scope string comes in as 
//     yydddhhmmss, azdeg, gregorian za deg, ch za deg
  {

  struct DateStruct MDY;
  char sScopeTime[15];
  char sTempStr[15];
  double LocalSiderealTime;
 
  // Get time in integer form.
  if(sscanf(ScopeStr, "%s %lf %*s %lf",
            sScopeTime,
            &SA[SAI].dRa,
            &SA[SAI].dDec) == 3)
    {
    strncpy(sTempStr, sScopeTime, 2);
    sTempStr[2] = '\0';
    SA[SAI].lYear = atol(sTempStr);
 
    strncpy(sTempStr, sScopeTime+2, 3);
    sTempStr[3] = '\0';
    SA[SAI].lDay = atol(sTempStr);
 
    strncpy(sTempStr, sScopeTime+5, 2);
    sTempStr[2] = '\0';
    SA[SAI].lHour = atol(sTempStr);
 
    strncpy(sTempStr, sScopeTime+7, 2);
    sTempStr[2] = '\0';
    SA[SAI].lMin = atol(sTempStr);
 
    strncpy(sTempStr, sScopeTime+9, 2);
    sTempStr[2] = '\0';
    SA[SAI].lSec = atol(sTempStr);
    }
  else
    {
    fprintf(stderr, "Bad scope string, exiting\n");
    exit(1);
    }

  // Convert from Atlantic Standard Time to Universal Time.
  tm_AstToUt(&SA[SAI].lYear, 
             &SA[SAI].lDay,
             &SA[SAI].lHour);

  // Get century year
  SA[SAI].lCentYear = tm_CentYear(SA[SAI].lYear);

  // Get month-day-year from year-day-of-year.
  // daytomon expects DOY to start at 0
  tm_DayToMon(SA[SAI].lDay-1, 
              SA[SAI].lCentYear, 
              &MDY);

  // Floating point decimal hour.
  SA[SAI].dDecHour = SA[SAI].lHour       +
                     SA[SAI].lMin/60.0   +
                     SA[SAI].lSec/3600.0;

  // LST needed for coordianate calculations.
  LocalSiderealTime = tm_UtToLst(SA[SAI].dDecHour,
                                 AOLONGITUDE_DEGS,
                                 MDY.month,
                                 MDY.day,
                                 MDY.year);
 
  // Apply telescope specific corrections to horizon coordinates.
  co_ZenAzCorrection(ThisHit.ucTid, &SA[SAI].dDec, &SA[SAI].dRa);

  // Convert from horizon to equitorial.
  // fprintf(stderr, "Zen = %lf  Az = %lf\n", SA[SAI].dDec, SA[SAI].dRa);
  co_ZenAzToRaDec(SA[SAI].dDec,
                  SA[SAI].dRa,
                  LocalSiderealTime,
                  &SA[SAI].dRa,
                  &SA[SAI].dDec);
  // fprintf(stderr, "Ra = %lf  Dec = %lf\n\n", SA[SAI].dRa, SA[SAI].dDec);

  // Calculate slopes for position interpolation.
  // The Get*Slope routines take care of dDecHour wrap.
  if(SAI > 0)
    {
    SA[SAI-1].dRaSlope = 
      GetRaSlope(SA[SAI-1].dRa,
                 SA[SAI].dRa,
                 SA[SAI-1].dDecHour,
                 SA[SAI].dDecHour);
    SA[SAI-1].dDecSlope = 
      GetDecSlope(SA[SAI-1].dDec,
                 SA[SAI].dDec,
                 SA[SAI-1].dDecHour,
                 SA[SAI].dDecHour);
    }
  }


//--------------------------------------------------------------------
void SlewStats(ScopeInfo   * SA, long NumSA, DbaseSumRcd_t * DbaseSumRcd)
//--------------------------------------------------------------------

  {
  long i;
  double SlewRate=0.0, SlewRateTotal=0.0;
  double HourDiff=0.0, SkyAngle=0.0;
  double HourTotal=0.0;
  long NumDupScopeStrings=0;

  DbaseSumRcd->dbStartRa  = SA[0].dRa;
  DbaseSumRcd->dbStartDec = SA[0].dDec;
  DbaseSumRcd->dbEndRa    = SA[NumSA-1].dRa;
  DbaseSumRcd->dbEndDec   = SA[NumSA-1].dDec;

  for(i = 1; i < NumSA; i++)
    {
    HourDiff = (SA[i].dDecHour < SA[i-1].dDecHour        ? 
                SA[i].dDecHour + 24.00 : SA[i].dDecHour) - 
               SA[i-1].dDecHour;
    if (HourDiff == 0.0) 
      {
      NumDupScopeStrings++;
      continue;
      }
    HourTotal += HourDiff;
    SkyAngle = co_SkyAngle(SA[i-1].dRa, SA[i-1].dDec, SA[i].dRa, SA[i].dDec);
    SlewRate = SkyAngle / HourDiff; 
    if(SlewRate < MIN_SLEW_DEG_PER_HR)
      DbaseSumRcd->dbTrackingTime += HourDiff;
    SlewRateTotal += SlewRate;
    }

  // number of scope string intervals is 1 less than the number of scope strings
  DbaseSumRcd->dbMeanScopeStringInterval = (HourTotal * 3600.0) / (NumSA-NumDupScopeStrings-1);
  DbaseSumRcd->dbMeanSlew = SlewRateTotal / (double)(NumSA-NumDupScopeStrings-1);
  DbaseSumRcd->usDupScopeStrings = NumDupScopeStrings;
  }


//--------------------------------------------------------------------
long FindScopeStrings(double      HitHr,
                      long        HitDay,
                      ScopeInfo * SA,
                      long        MaxSAI,
                      long *      Aft,
                      long *      Forward)
//--------------------------------------------------------------------
/*
                        midnight
                            |
       |-------------|-------------|------------| 
      SS             SS            SS           SS
             |           |    |         
 Hit cases:  1           2    3         
                         

Case 1 - by far the most common - hit is on same day as both scope strings.
Case 2 - Aft is same day, Forward is next day (could be next year too).
Case 3 - Aft is previous day (could be previous year too), Forward is same day.
  
*/
//--------------------------------------------------------------------
  {
  static long i = 0;
  long  MatchType;

i = 0;
  for(MatchType = 0; i < MaxSAI - 1 && MatchType == 0; i++)
    if(SA[i].lDay         == HitDay       &&   // case 1
       SA[i+1].lDay       == HitDay       &&   
       SA[i].dDecHour     <= HitHr        &&
       SA[i+1].dDecHour   >= HitHr)
      MatchType = 1;
    else if(SA[i+1].lDay       != HitDay       &&   // case 2
            SA[i].lDay         == HitDay       &&
            SA[i].dDecHour     <= HitHr        &&
            SA[i+1].dDecHour   <= HitHr)
           MatchType = 2 ;
         else if(SA[i].lDay         != HitDay       &&   // case 3
                 SA[i+1].lDay       == HitDay       &&
                 SA[i].dDecHour     >= HitHr        &&
                 SA[i+1].dDecHour   >= HitHr)
                MatchType = 3 ;

  if(MatchType) 
    {
    i -= 1;      // for loop incremented 1 past our pair
    *Aft = i;
    *Forward = i + 1;
    }
  else
    i = 0;       // if no match, reset for next search

  return(MatchType);
  }

//--------------------------------------------------------------------
long GetNextScopeString()
//--------------------------------------------------------------------
  {
  U16 u16BlockSize;
  U16 u16BlockType;

  while(1)
    {
    if(!ReadNext((char *)&u16BlockSize,sizeof(U16)))
      return(-1);

    if (bSwapBytes)
      ChangeEndian((unsigned char*)&u16BlockSize,sizeof(U16));

    if(!ReadNext((char *)&u16BlockType,sizeof(U16)))
      return(-1);

    if (bSwapBytes)
      ChangeEndian((unsigned char *)&u16BlockType,sizeof(U16));

    if(!GoodBlockType(u16BlockType))
      {
      sprintf(UtilStr255,
              "Unknown.....\nblock type: %x\nblock size: %x\nfile position: %ld\n",
              u16BlockType, u16BlockSize, ftell(fInputDataFile));
      WriteToLog(UtilStr255);
      ErrExit("Unexpected block type");
      }

    if(u16BlockType == SCSTR)
      break;

    fseek(fInputDataFile, (long) u16BlockSize-4, 1);
    }
  return(u16BlockSize-4);
  }

//--------------------------------------------------------------------
void Fini()
// This routine should only be called from ReadNext().
//--------------------------------------------------------------------
  {
  int rtn;
  struct DateStruct sdate;
  long year;

  fprintf(stderr, "Exiting... ");
  //if(ScanningForScope)
  //  return;
  //else 
    if(RecordStarted)
      {
      fprintf(stderr, "file ends in record fragment!\n");
      rtn = 1;
      }
    else
      {
      fprintf(stderr, "file ends on record boundry\n");   // best way to exit! 
      rtn = 0;
      }

  sprintf(UtilStr255, 
          "Total number of hits:                             %ld\n", glNumHits);
  WriteToLog(UtilStr255);
  sprintf(UtilStr255, 
          "Total number of hits lost due to no scope string: %ld\n", glNoStringHits);
  WriteToLog(UtilStr255);

  ThisDbaseSumRcd.dbFileDuration = (FinalHitTime < ThisDbaseSumRcd.dbStartTime ?
                                    FinalHitTime + 24.0                        :
                                    FinalHitTime)                              -
                                    ThisDbaseSumRcd.dbStartTime;

  // calculate julian date
  year = tm_TwoDigitYear(StartHitYear);             // We need month:day
  tm_DayToMon(StartHitDay-1, year, &sdate);         // for julian conversion
  ThisDbaseSumRcd.dbStartTime = tm_ToJul(sdate.month,
                                         sdate.day,
                                         StartHitYear,
                                         ThisDbaseSumRcd.dbStartTime);          

  UpdateDataBase(InputDataFileName, &DbaseSumRcd, &ThisDbaseSumRcd);

  exit(rtn);
  }

//------------------------------------------------------------------------
// Command line operands
//
void  GetOps(int       argc,
             char    * argv[],
             BOOLEAN * bSwapBytes,
             BOOLEAN * bUpdateDb,
             double  * dFirstLo,
             double  * dSecondLo,
             char    * sReceiverID,
             unsigned short * usSigmaThreshold,
             char   ** InputDataFileName)
//------------------------------------------------------------------------
  {
  long i;
 
  if(argc < 4)
    CmdError();

  // Get parameters.
  // Skip the program name and stop short of the file names.
  for (i = 1; i < (argc - 2); i++)
    if (strcmp(argv[i],"-s") == 0)
      *bSwapBytes = TRUE;
    else if (strcmp(argv[i],"-db") == 0)
      *bUpdateDb = TRUE;
    else if (strcmp(argv[i],"-lo1") == 0)
      {  
      if (sscanf(argv[++i],"%lf", dFirstLo) != 1)
        CmdError();
      }  
    else if (strcmp(argv[i],"-lo2") == 0)
      {  
      if (sscanf(argv[++i],"%lf", dSecondLo) != 1)
        CmdError();
      }  
    else if (strcmp(argv[i],"-id") == 0)
      {  
      if (sscanf(argv[++i],"%s", sReceiverID) != 1)
        CmdError();
      }  
    else if (strcmp(argv[i],"-t") == 0)
      {  
      if (sscanf(argv[++i],"%hu", usSigmaThreshold) != 1)
        CmdError();
      }  
    else
      CmdError();
 
  // Now get the file names
  *InputDataFileName = argv[i];

  }

//------------------------------------------------------------------------
void CmdError(void)
//------------------------------------------------------------------------
  {
  fprintf(stderr, usage);
  exit(1);
  }  


//------------------------------------------------------------------------
BOOLEAN GoodBlockType(U16 u16BlockType)
//------------------------------------------------------------------------
  {
  switch (u16BlockType)
    {
    case THROTTLE : 
    case HITS     : 
    case CMMT     : 
    case SYSCFG   : 
    case ENG      : 
    case ERR      :
    case SCSTR    : return(TRUE);
    default       : return(FALSE);
    }
  }

//------------------------------------------------------------------------
void PushEngBuf(long            EngCentYear, 
                long            EngDay, 
                double          EngHr, 
                unsigned char * ucEngMons,
                EngBuf_t      * EngBuf, 
                long          * EngBufFowardIndex,
                long          * EngBufAftIndex)
//------------------------------------------------------------------------
// During real time data acquisition, the hits recorded during any one
// integration period were actually recorded 4 integrations in the past.
// Engineering monitors, on the other hand, are recorded with no such delay.
// This routine maintains a ring buffer of engineering monitors that allows 
// us  to reach back in time for the engineering monitors that  reveal the 
// ADC states at the time of actual hit acquisition.
//
// A side effect of this is that the first 4 integrations worth of hits
// in a file do not have their monitors in the current file (they are in
// the previous file!) and are thrown away.  Also, the final 4 sets of 
// monitors in a file are not used - they go with the first 4 integrations
// worth of hits in the next file.
//
// We do *not* search for time matches as we do with scope strings.  The 
// production of the engineering monitors is a function of the s4 instrument 
// itself and these monitors are always recorded with this 4 integration
// offset.  They are recorded even when we are disabled and are not getting
// any hits.
//------------------------------------------------------------------------
  {
    EngBuf[*EngBufFowardIndex].CentYear = EngCentYear;
    EngBuf[*EngBufFowardIndex].Day      = EngDay;
    EngBuf[*EngBufFowardIndex].Hr       = EngHr;

    memcpy(EngBuf[*EngBufFowardIndex].ucEngMons, ucEngMons, NUM_ENG_MONS * sizeof(char)); 

    if(*EngBufAftIndex != -1) {
	*EngBufAftIndex = *EngBufAftIndex >= NUM_ENG_BUFS-1 ? 0 : *EngBufAftIndex + 1;
    } else {
   	if(*EngBufFowardIndex == NUM_ENG_BUFS-1) {
		*EngBufAftIndex = 0;
	}
    }

    *EngBufFowardIndex = *EngBufFowardIndex >= NUM_ENG_BUFS-1 ? 0 : *EngBufFowardIndex + 1;
  }

