/* ________________________________________________________________________
 programmer : Jeff Cobb
 source name: s4poisson.cc
 compiler   : gcc
___________________________________________________________________________
 modification history:
   version 1.0 finished 11/12/00
___________________________________________________________________________
 description: This program reads a hit by bin histogram as produced by 
 s4binhst.  It outputs a 3 column file, giving bin hit count, the number
 of bins observed having that count, and the number of bins expected to have 
 that count, assuming poisson statistics.  Optionally (usually), the summary
 DB can be updated.  The summary DB will then contain the ChiSqr goodness
 of fit between the observed and expected distributions.

 Note: s4binhst must have been run on a given dbase file prior to running
 this program.  Additionally, s4binhst must have updated the summary DB
 with the values that this program needs.
___________________________________________________________________________
*/

#include <stdio.h>
#include <unistd.h>
#include <malloc.h>
#include <strings.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "s4utils.h"
#include "s4summary.h"

#define CLOSE_TO_ZERO 0.000001

void GetOps(int argc,  
            char         ** argv, 
            unsigned long * MaxEvents,
            BOOLEAN         * Before,
            BOOLEAN         * After,
            BOOLEAN         * UpdateDb);

main(int argc, char ** argv) {

  double          Version       = 1.0;

  unsigned long   MaxEvents 	= 1024;
  unsigned long   i;
  unsigned long   HistoBin;
  unsigned long   LastHistoBin 	= 0;
  unsigned long   HistoValue;
  unsigned long * ObsDist;
  double        * ExpDist;
  double          ChiSqr 	= 0.0;
  double          Lambda	= 0.0;        
  double          ProbXEvents 	= 0.0; 
  double          SumProbX 	= 0.0;
  FILE          * InFile;
  FILE          * OutFile;
  char            InString[256];
  long            SummDbIndex;
  int             SummIndexFile;
  DbaseSumRcd_t   DbaseSumRcd;
  char            FileNameRoot[64];
  BOOLEAN         UpdateDb = FALSE;
  BOOLEAN         Before = FALSE;
  BOOLEAN         After = FALSE;

  GetOps(argc, argv, &MaxEvents, &Before, &After, &UpdateDb);

  if ((InFile = fopen(argv[argc - 1], "r")) == NULL)
    {
    perror("Can't open input file.");
    exit(1);
    }

  // we need some values stored in the summary database
  ut_GetRootFilename(argv[argc - 1], &FileNameRoot[0]);
  SummIndexFile = sum_GetDataBase(O_RDONLY);
  SummDbIndex  = sum_GetRcd(FileNameRoot, &DbaseSumRcd, SummIndexFile, O_RDONLY);

  // Space for the observed distribution. The +1 is to have bin for MaxEvents
  if ((ObsDist = (unsigned long *) calloc((MaxEvents + 1),sizeof(unsigned long))) == NULL)
    {
    perror("Error allocating space for 'ObsDist'");
    exit(1);
    }
  (void) memset(ObsDist,0,sizeof(unsigned long) * MaxEvents + 1);

  // Space for the expected distribution
  if ((ExpDist = (double *) calloc((MaxEvents + 1),sizeof(double))) == NULL)
    {
    perror("Error allocating space for 'ExpDist'");
    exit(1);
    }
  for(i = 0; i <= MaxEvents; i++) ExpDist[i] = 0.0;

  // make a distrubution from the observed histogram
  for(i = 0; i < DbaseSumRcd.lPoissonNumBinsBefore; i++) {
    fgets(InString, 256, InFile);
    if(sscanf(InString, "%lu %lu", &HistoBin, &HistoValue) != 2 ||
       HistoBin != LastHistoBin+1 && HistoBin != 0) {
      perror("Bad data line, noncontiguous histogram or premature end of file");
      exit(1);
    }
    LastHistoBin = HistoBin;
    ObsDist[HistoValue > MaxEvents ? MaxEvents : HistoValue]++;
  }

  // Calculate the expected distribution
  Lambda = (double)DbaseSumRcd.lPoissonNumEventsBefore / 
           (double)DbaseSumRcd.lPoissonNumBinsBefore;

  // for bins having 0 events
  // Lambda^0 = 1 and and 0! = 1, so you are left with e^-Lambda
  ProbXEvents = exp(-Lambda);
  SumProbX = ProbXEvents;
  ExpDist[0] = ProbXEvents * DbaseSumRcd.lPoissonNumBinsBefore;
  if(ExpDist[0] < CLOSE_TO_ZERO) ExpDist[0] = CLOSE_TO_ZERO;

  // for bins having > 0 events
  // For each subsequent i, factor out the previous P(i) and you are left with Lambda/i
  for (i = 1; i < MaxEvents; i++) {
     ProbXEvents *= (Lambda/i); 
     SumProbX += ProbXEvents;
     ExpDist[i] = ProbXEvents * DbaseSumRcd.lPoissonNumBinsBefore;
     if(ExpDist[i] < CLOSE_TO_ZERO) ExpDist[i] = CLOSE_TO_ZERO;
  }

  // for bins having >= MaxEvents
  // Whatever unallocated probability you have left is the probabilty of getting some
  // value in the set of all values that you have left.
  ProbXEvents = 1 - SumProbX;
  ExpDist[i] = ProbXEvents * DbaseSumRcd.lPoissonNumBinsBefore;
  if(ExpDist[i] < CLOSE_TO_ZERO) ExpDist[i] = CLOSE_TO_ZERO;

  // ChiSqr
  for (i = 0;i <= MaxEvents; i++) {
    ChiSqr += ( ((double)ObsDist[i] - ExpDist[i]) * ((double)ObsDist[i] - ExpDist[i]) ) / ExpDist[i];
  }

  // update the summary DB
  if(UpdateDb == TRUE) {
    ut_GetRootFilename(argv[argc - 1], &FileNameRoot[0]);
    SummIndexFile = sum_GetDataBase(O_RDWR);
    SummDbIndex  = sum_GetRcd(FileNameRoot, &DbaseSumRcd, SummIndexFile, O_RDWR);
    if(Before) {
      DbaseSumRcd.dbPoissonChiSqrBefore = ChiSqr;
      DbaseSumRcd.lPoissonMaxEventsBefore = MaxEvents;
      DbaseSumRcd.fPoissonBeforeVers = Version;
    } else {
      DbaseSumRcd.dbPoissonChiSqrAfter = ChiSqr;
      DbaseSumRcd.lPoissonMaxEventsAfter = MaxEvents;
      DbaseSumRcd.fPoissonAfterVers = Version;
    }  
    sum_PutRcd(SummDbIndex, &DbaseSumRcd, SummIndexFile);
    sum_ReleaseDataBase(SummIndexFile);
  }

  for (i = 0; i <= MaxEvents; i++)
     (void) fprintf(stdout,"%u     %u     %lf\n",i , ObsDist[i], ExpDist[i]);

} // end program


void GetOps(int argc,  
            char         ** argv, 
            unsigned long * MaxEvents,
            BOOLEAN         * Before,
            BOOLEAN         * After,
            BOOLEAN         * UpdateDb) {

  long i;

  for (i = 1; i < argc - 1; i++)
    if (strcmp(argv[i], "-max") == 0)
      {
      if (sscanf(argv[++i],"%u",MaxEvents) != 1)
        perror("Error in command line.");
    }
      else if (strcmp(argv[i], "-db") == 0) {
      *UpdateDb = TRUE;
    }
    else if (strcmp(argv[i], "-before") == 0) {
      *Before = TRUE;
    }
    else if (strcmp(argv[i], "-after") == 0) {
      *After = TRUE;
    }
    else
      { 
    perror("Usage\n poisson_obs -before|-after [-max max_events] infile_name");
      exit(1);
      }

  if(!(*Before || *After)) {
    perror("Usage\n poisson_obs -before|-after [-max max_events] infile_name");
    exit(1);
   }

}

