/*
 * Source Code Name  : s4printsumm.cc
 * Programmer        : Jeff Cobb
 *                   : Project SERENDIP, University of California, Berkeley CA
 * Compiler          : GNU gcc
 *
 * Given a dbase file name, this program prints that file's record from
 * the summary database.
 *
 * This program also serves as a simple template for crafting more 
 * complex summary database inquiries.
 *
 * History:
 * 2/28/00	v1.01	jeffc	Added mean_scope_interval output.  Changed	
 *				order of fields output.
 * 2/29/00	v1.02	jeffc	Added start_jd output.
 * 3/02/00      v1.03   mattl   Added engmonrecreader/countbirdie fields
 * 3/04/00	v1.04	jeffc	Added s4dbascheck fields
 * 3/05/00	v1.05	mattl	Added false birdie field
 * 3/14/00	v1.06	jeffc	Added spectra count and board-error counts
 * 4/10/00      v1.07   mattl   Fixed s4countbirdie fields
 * 11/17/00     v1.08   jeffc   Added support for db-readonly access
 * 11/17/00     v1.09   mattl   Added min/max ADC values observed in file
 * 1/5/00       v1.10   jeffc   Added first and second LO's
 * 4/2/01	v1.11	jeffc	Added s4rfifilter fields
 * 4/2/01       v1.12   mattl   Added s4drfi and s4dbascheck.badlimit fields
 * 10/2/02      v1.13   mattl   Added s4acquireratecheck and key flag
 *
 */
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include "SIREN.h"
#include "s4database.h"
#include "s4summary.h"
#include "s4error.h"

#define VERSION 1.13

const char * usage = "usage : s4printsumm [-k] dbase_name";

main(int argc, char ** argv)
  {
  long i;
  BOOLEAN bShowKey = FALSE;
  int DB;
  unsigned long SummDbIndex;

  char * DbaseName;

  DbaseSumRcd_t DbaseSumRcd;

  if (argc < 2) {
    fprintf(stderr,"%s\n",usage);
    exit(1);
    }

  if (strcmp(argv[1],"-k") == 0) {
    bShowKey = TRUE;
    DbaseName = argv[2];
    }
  else {
    DbaseName = argv[1];
    }

  DB = sum_GetDataBase(O_RDONLY);                                      // get summary database
  SummDbIndex  = sum_GetRcd(DbaseName,  &DbaseSumRcd, DB, O_RDONLY);   // get our record

  if(SummDbIndex == -1) {
    fprintf(stderr, "%s is not in the Summary Database\n", DbaseName);
    exit(1);
    }

  if (bShowKey) {
    printf("For dbase file: %s Index = %u (hex: %x)\n",DbaseName,SummDbIndex,SummDbIndex);
    printf("                   Key = %u (hex: %x)\n",db_FileNameToKey(DbaseName),db_FileNameToKey(DbaseName));
    exit(0);
    }

  fprintf(stdout, "file_name              %s\n", DbaseSumRcd.sDbaseName); 

  // s4pcdbase stuff
  fprintf(stdout, "# ========================================================================== \n");
  fprintf(stdout, "s4pcdbase_version      %4.2f\n",  DbaseSumRcd.fPcdbaseVers);
  fprintf(stdout, "sigma_thresh           %d\n",  DbaseSumRcd.usSigmaThreshold); 
  fprintf(stdout, "first_LO               %lf\n",  DbaseSumRcd.dbFirstLo); 
  fprintf(stdout, "second_LO              %lf\n",  DbaseSumRcd.dbSecondLo); 
  fprintf(stdout, "start_jd               %lf\n", DbaseSumRcd.dbStartTime); 
  fprintf(stdout, "file_duration          %lf hours\n", DbaseSumRcd.dbFileDuration); 
  fprintf(stdout, "tracking_time          %lf hours\n", DbaseSumRcd.dbTrackingTime); 
  fprintf(stdout, "mean_slew_rate         %lf degrees per second\n", DbaseSumRcd.dbMeanSlew/3600.0); 
  fprintf(stdout, "starting_RA            %lf hours\n", DbaseSumRcd.dbStartRa); 
  fprintf(stdout, "ending_RA              %lf hours\n", DbaseSumRcd.dbEndRa); 
  fprintf(stdout, "starting_Dec           %lf degrees\n", DbaseSumRcd.dbStartDec); 
  fprintf(stdout, "ending_Dec             %lf degrees\n", DbaseSumRcd.dbEndDec); 
  fprintf(stdout, "num_spectra            %ld\n", DbaseSumRcd.ulNumSpectra); 
  fprintf(stdout, "total_hits             %ld\n", DbaseSumRcd.ulTotalHits); 
  fprintf(stdout, "mean_scope_interval    %lf seconds\n", DbaseSumRcd.dbMeanScopeStringInterval); 
  fprintf(stdout, "dup_scope_strings      %hu\n", DbaseSumRcd.usDupScopeStrings); 
  fprintf(stdout, "nostring_hits          %ld\n", DbaseSumRcd.ulNoStringHits); 
  fprintf(stdout, "err_synth              %ld\n", DbaseSumRcd.ulErrSynth); 
  fprintf(stdout, "err_vectwarn           %ld\n", DbaseSumRcd.ulErrVectWarn); 
  fprintf(stdout, "err_overrun            %ld\n", DbaseSumRcd.ulErrOverRun); 
  fprintf(stdout, "err_arithovfl          %ld\n", DbaseSumRcd.ulErrArithOvfl); 
  fprintf(stdout, "err_bdnotready         %ld\n", DbaseSumRcd.ulErrBdNotReady); 
  fprintf(stdout, "err_other              %ld\n", DbaseSumRcd.ulErrOther); 
  fprintf(stdout, "seti_at_home_hits      %ld\n", DbaseSumRcd.ulSetiAtHomeBoardCount); 
  fprintf(stdout, "board_kilo_hits        "); 
  for(i = 0; i < NUMBOARDS; i++) {
  	fprintf(stdout, "%2d:%3hu    ", i, DbaseSumRcd.ucBoardKiloCount[i]); 
	if((i+1) % 5 == 0 && (i+1) != NUMBOARDS) fprintf(stdout, "\n                       ");
  }
  fprintf(stdout, "\n");

  // s4acquireratecheck stuff
  fprintf(stdout, "# ========================================================================== \n");
  fprintf(stdout, "s4acquireratecheck_version   %4.2f\n", DbaseSumRcd.fAcquireRateCheckVers);
  fprintf(stdout, "ARC_low_threshold      %lf (hits/hour)\n", DbaseSumRcd.dbARCLowThreshold);
  fprintf(stdout, "ARC_high_threshold     %lf (hits/hour)\n", DbaseSumRcd.dbARCHighThreshold);
  fprintf(stdout, "ARC_all_boards_passed? %s\n", 
                  (DbaseSumRcd.bAllBoardsPassed ? "TRUE" : "FALSE"));
  fprintf(stdout, "ARC_board_19_passed?   %s\n", 
                  (DbaseSumRcd.bBoard19Passed ? "TRUE" : "FALSE"));

  // s4dbascheck stuff
  fprintf(stdout, "# ========================================================================== \n");
  fprintf(stdout, "s4dbasecheck_version   %4.2f\n",  DbaseSumRcd.fDbasCheckVers);
  fprintf(stdout, "Dbase_errors           ");
  if(DbaseSumRcd.ulDbaseErrorType) {
        fprintf(stdout, "%ld  %s ",
                DbaseSumRcd.lDbaseErrors, DbaseErrorStatusString[DbaseSumRcd.lDbaseErrorStatus]);
        for(i = 1; i < NUM_ERROR_TYPES; i++) {
                if(DbaseSumRcd.ulDbaseErrorType & DbaseError[i].TypeFlag) {
                        fprintf(stdout, "%s ",  DbaseError[i].Error);
                }
        }
  } else {
        fprintf(stdout, "None");
  }
  fprintf(stdout, "\n");
  fprintf(stdout, "acceptable_bad_limit   %4.2lf%%\n", DbaseSumRcd.dbDbaseBadLimit);

  // s4engmonrecreader stuff
  fprintf(stdout, "# ========================================================================== \n");
  fprintf(stdout, "s4engmonreader_vers    %4.2f\n",  DbaseSumRcd.fEngmonrecreaderVers);
  fprintf(stdout, "min_accepted_adc       %ld\n", DbaseSumRcd.lMinAcceptedADC);
  fprintf(stdout, "max_accepted_adc       %ld\n", DbaseSumRcd.lMaxAcceptedADC);
  fprintf(stdout, "min_adc (observed)     %ld\n", DbaseSumRcd.lMinADC);
  fprintf(stdout, "max_adc (observed)     %ld\n", DbaseSumRcd.lMaxADC);
  fprintf(stdout, "adc_0_bad_count        %ld\n", DbaseSumRcd.lADC0BadCount);
  fprintf(stdout, "adc_1_bad_count        %ld\n", DbaseSumRcd.lADC1BadCount);
  fprintf(stdout, "adc_2_bad_count        %ld\n", DbaseSumRcd.lADC2BadCount);
  fprintf(stdout, "adc_3_bad_count        %ld\n", DbaseSumRcd.lADC3BadCount);
  fprintf(stdout, "adc_4_bad_count        %ld\n", DbaseSumRcd.lADC4BadCount);
  fprintf(stdout, "disabled_count         %ld\n", DbaseSumRcd.lDisabledCount);

  // s4countbirdie stuff
  fprintf(stdout, "# ========================================================================== \n");
  fprintf(stdout, "s4countbirdie_vers     %4.2f\n",  DbaseSumRcd.fCountBirdieVers);
  fprintf(stdout, "expected birdies seen:\n");
  fprintf(stdout, "        -0-  -1-  -2-  -3-  -4-  -5-  -6-  -7-");
  for (i=0;i<40;i++) {
    if (i%8==0) fprintf(stdout,"\n%ld (%2ld): ",long(i/8),long(i));
    if (DbaseSumRcd.ucBirdieSeenPercentages[i] == 255) {
      fprintf(stdout,"..."); }
    else { fprintf(stdout,"%3d",DbaseSumRcd.ucBirdieSeenPercentages[i]); }
    if (DbaseSumRcd.ullMirrorSeen & (unsigned long long)(pow(2,i))) {
      fprintf(stdout,"* "); }
    else { fprintf(stdout,"  "); }
    }
  fprintf(stdout, "\n");
 
  // "before" stats
  fprintf(stdout, "# ========================================================================== \n");
  fprintf(stdout, "s4_pwrhst_before_vers  %4.2f\n",  DbaseSumRcd.fPwrHstBeforeVers);
  fprintf(stdout, "s4_binhst_before_vers  %4.2f\n",  DbaseSumRcd.fBinHstBeforeVers);
  fprintf(stdout, "s4_poisson_before_vers %4.2f\n", DbaseSumRcd.fPoissonBeforeVers);
  fprintf(stdout, "pwr_hst_max_power_before  	%ld\n", DbaseSumRcd.lMaxPowerHistogrammedBefore);
  fprintf(stdout, "pwr_hst_over_max_before  	%ld\n", DbaseSumRcd.lNumEventsOverMaxPowerBefore);
  fprintf(stdout, "poisson_num_events_before  	%ld\n", DbaseSumRcd.lPoissonNumEventsBefore);
  fprintf(stdout, "poisson_bin_width_before  	%ld\n", DbaseSumRcd.lPoissonBinWidthBefore);
  fprintf(stdout, "poisson_num_bins_before  	%ld\n", DbaseSumRcd.lPoissonNumBinsBefore);
  fprintf(stdout, "poisson_max_events_before  	%ld\n", DbaseSumRcd.lPoissonMaxEventsBefore);
  if(DbaseSumRcd.fPoissonBeforeVers == 0.0) {
    fprintf(stdout, "poisson_lambda_before        %lf\n", 0.0);
   } else {
    fprintf(stdout, "poisson_lambda_before   	%lf\n", 
						  (double)DbaseSumRcd.lPoissonNumEventsBefore / 
                                                  (double)DbaseSumRcd.lPoissonNumBinsBefore);
  }
  fprintf(stdout, "poisson_chisqr_before   	%lf\n", DbaseSumRcd.dbPoissonChiSqrBefore);
  if(DbaseSumRcd.fPoissonBeforeVers == 0.0) {
    fprintf(stdout, "poisson_d.f._before		%ld\n", 0.0);
   } else {
    fprintf(stdout, "poisson_d.f._before		%ld\n", DbaseSumRcd.lPoissonMaxEventsBefore-1);
  }
  fprintf(stdout, "\n");

  // s4rfifilter stuff
  fprintf(stdout, "# ========================================================================== \n");
  fprintf(stdout, "s4_rfifilter_vers            %4.2f\n",DbaseSumRcd.fRfiFilterVers);
  fprintf(stdout, "rfi_events_in                %ld\n",  DbaseSumRcd.ulRfiFilterEventsIn);  
  fprintf(stdout, "rfi_events_out               %ld\n",  DbaseSumRcd.ulRfiFilterEventsOut);   
  fprintf(stdout, "rfi_pass_rate                %lf\n",  (double)DbaseSumRcd.ulRfiFilterEventsOut /
							 (double)DbaseSumRcd.ulRfiFilterEventsIn);   
  fprintf(stdout, "rfi_expected_out             %ld\n",  DbaseSumRcd.ulRfiFilterExpectedOut);
  if(DbaseSumRcd.ulFlags & RFI_PASSBAND_RUN) {
    if(DbaseSumRcd.ulRfiFilterEventsIn - (DbaseSumRcd.ulSlowSlew_NumNewRej + DbaseSumRcd.ulFastSlew_NumNewRej) == 0)
      fprintf(stdout, "rfi_passband_run             NO - auto shutoff because of no good time events\n");
    else
      fprintf(stdout, "rfi_passband_run             YES\n");        
  } else {
    fprintf(stdout, "rfi_passband_run             NO\n");        
  }
  fprintf(stdout, "rfi_passband_freq_res        %lf\n",  DbaseSumRcd.dbPassBand_FreqRes);     
  fprintf(stdout, "rfi_passband_num_beams       %lf\n",  DbaseSumRcd.dbPassBand_NumBeams);   
  fprintf(stdout, "rfi_passband_limit           %ld\n",  DbaseSumRcd.ulPassBand_Limit);     
  fprintf(stdout, "rfi_passband_num_rej         %ld\n",  DbaseSumRcd.ulPassBand_NumRej);   
  fprintf(stdout, "rfi_passband_numnew_rej      %ld\n",  DbaseSumRcd.ulPassBand_NumNewRej);   
  fprintf(stdout, "rfi_passband_dead_freq       %lf\n",  DbaseSumRcd.dbPassBand_DeadFreqBand); 
  if(DbaseSumRcd.ulFlags & RFI_SSPECTRUM_RUN)
    fprintf(stdout, "rfi_s_spectrum_run           YES\n");           
  else
    fprintf(stdout, "rfi_s_spectrum_run           NO\n");           
  fprintf(stdout, "rfi_s_spectrum_ent_limit     %lf\n",  DbaseSumRcd.dbSpreadSpectrum_EntropyLimit);
  fprintf(stdout, "rfi_s_spectrum_limit         %ld\n",  DbaseSumRcd.ulSpreadSpectrum_EventLimit);   
  fprintf(stdout, "rfi_s_spectrum_sigmalimit    %lf\n",  DbaseSumRcd.dbSpreadSpectrum_SigmaLimit);   
  fprintf(stdout, "rfi_s_spectrum_num_rej       %ld\n",  DbaseSumRcd.ulSpreadSpectrum_NumRej); 
  fprintf(stdout, "rfi_s_spectrum_numnew_rej    %ld\n",  DbaseSumRcd.ulSpreadSpectrum_NumNewRej); 
  if(DbaseSumRcd.ulFlags & RFI_SLOWSLEW_RUN)
    fprintf(stdout, "rfi_slow_slew_run            YES\n");           
  else
    fprintf(stdout, "rfi_slow_slew_run            NO\n");           
  fprintf(stdout, "rfi_slow_slew_delta_secs     %ld\n",  DbaseSumRcd.ulSlowSlew_DeltaSecs);     
  fprintf(stdout, "rfi_slow_slew_ba_secs        %ld\n",  DbaseSumRcd.ulSlowSlew_BaSecs);      
  fprintf(stdout, "rfi_slow_slew_limit          %lf\n",  DbaseSumRcd.dbSlowSlew_Limit);           
  fprintf(stdout, "rfi_slow_slew_num_rej        %ld\n",  DbaseSumRcd.ulSlowSlew_NumRej);            
  fprintf(stdout, "rfi_slow_slew_numnew_rej     %ld\n",  DbaseSumRcd.ulSlowSlew_NumNewRej); 
  fprintf(stdout, "rfi_slow_slew_spect_lost     %ld\n",  DbaseSumRcd.ulSlowSlew_SpectraLost);      
  if(DbaseSumRcd.ulFlags & RFI_SLOWSLEW_RUN)
    fprintf(stdout, "rfi_fast_slew_run            YES\n");           
  else
    fprintf(stdout, "rfi_fast_slew_run            NO\n");           
  fprintf(stdout, "rfi_fast_slew_dps            %lf\n",  DbaseSumRcd.dbFastSlew_Dps);         
  fprintf(stdout, "rfi_fast_slew_num_rej        %ld\n",  DbaseSumRcd.ulFastSlew_NumRej);     
  fprintf(stdout, "rfi_fast_slew_numnew_rej     %ld\n",  DbaseSumRcd.ulFastSlew_NumNewRej);     
  fprintf(stdout, "rfi_fast_slew_spect_lost     %ld\n",  DbaseSumRcd.ulFastSlew_SpectraLost);   


  // s4drfi stuff

  fprintf(stdout, "# ========================================================================== \n");
  fprintf(stdout, "s4drfi_vers                  %4.2f\n",DbaseSumRcd.fS4drfiVers);
  fprintf(stdout, "events_in                    %ld\n",  DbaseSumRcd.ulDrfiNumEvents);
  fprintf(stdout, "clean_events_out             %ld\n",  DbaseSumRcd.ulDrfiNumCleanEvents);
  fprintf(stdout, "drfi_pass_rate               %lf\n",  (double)DbaseSumRcd.ulDrfiNumCleanEvents /
							 (double)DbaseSumRcd.ulDrfiNumEvents);
  fprintf(stdout, "drfi_events_out              %ld\n",  DbaseSumRcd.ulDrfiNumRfiEvents);
  fprintf(stdout, "numbers_of_sectors           %ld\n",  DbaseSumRcd.ulDrfiNumSectors);
  fprintf(stdout, "radius_in_seconds            %ld\n",  DbaseSumRcd.ulDrfiRadiusInSeconds);
  fprintf(stdout, "beams_for_beamres            %ld\n",  DbaseSumRcd.ulDrfiBeamsForBeamres);
  fprintf(stdout, "time_expansion_factor        %ld\n",  DbaseSumRcd.ulDrfiTimeExpansionFactor);
  fprintf(stdout, "time_resolution              %lf\n",  DbaseSumRcd.dbDrfiTimeResolution);
  fprintf(stdout, "gap_factor                   %ld\n",  DbaseSumRcd.ulDrfiGapFactor);
  fprintf(stdout, "rejection_threshold          %ld\n",  DbaseSumRcd.ulDrfiRejectionThreshold);
  fprintf(stdout, "spectra_density              %e\n",   DbaseSumRcd.dbDrfiSpectraDensity);
  fprintf(stdout, "time_ticks                   %lf\n",  DbaseSumRcd.dbDrfiTimeTicks);
  fprintf(stdout, "num_semicircles              %ld\n",  DbaseSumRcd.ulDrfiNumSemicircles);
  fprintf(stdout, "ra_res                       %lf\n",  DbaseSumRcd.dbDrfiRaRes);
  fprintf(stdout, "dec_res                      %lf\n",  DbaseSumRcd.dbDrfiDecRes);


  // "after" stats
  fprintf(stdout, "# ========================================================================== \n");
  fprintf(stdout, "s4_pwrhst_after_vers  %4.2f\n",  DbaseSumRcd.fPwrHstAfterVers);
  fprintf(stdout, "s4_binhst_after_vers   %4.2f\n",  DbaseSumRcd.fBinHstAfterVers);
  fprintf(stdout, "s4_poisson_after_vers  %4.2f\n", DbaseSumRcd.fPoissonAfterVers);
  fprintf(stdout, "pwr_hst_max_power_after  	%ld\n", DbaseSumRcd.lMaxPowerHistogrammedAfter);
  fprintf(stdout, "pwr_hst_over_max_after  	%ld\n", DbaseSumRcd.lNumEventsOverMaxPowerAfter);
  fprintf(stdout, "poisson_num_events_after  	%ld\n", DbaseSumRcd.lPoissonNumEventsAfter);
  fprintf(stdout, "poisson_bin_width_after  	%ld\n", DbaseSumRcd.lPoissonBinWidthAfter);
  fprintf(stdout, "poisson_num_bins_after  	%ld\n", DbaseSumRcd.lPoissonNumBinsAfter);
  fprintf(stdout, "poisson_max_events_after  	%ld\n", DbaseSumRcd.lPoissonMaxEventsAfter);
  if(DbaseSumRcd.fPoissonAfterVers == 0.0) {
    fprintf(stdout, "poisson_lambda_after		%lf\n", 0.0);
  } else { 
    fprintf(stdout, "poisson_lambda_after		%lf\n", 
 						  (double)DbaseSumRcd.lPoissonNumEventsAfter / 
                                                  (double)DbaseSumRcd.lPoissonNumBinsAfter);
  }
  fprintf(stdout, "poisson_chisqr_after		%lf\n", DbaseSumRcd.dbPoissonChiSqrAfter);
  if(DbaseSumRcd.fPoissonAfterVers == 0.0) {
    fprintf(stdout, "poisson_d.f._after		%ld\n", 0.0);
  } else { 
    fprintf(stdout, "poisson_d.f._after		%ld\n", DbaseSumRcd.lPoissonMaxEventsAfter-1);
  }
  fprintf(stdout, "\n");

}
