// Source Code Name  : s4rddat.cc
// Programmer        : Donnelly/Cobb et al
//                   : Project SERENDIP, University of California, Berkeley CA
// Compiler          : GNU gcc

// #define IBMPC
#define SUN
#include <stdio.h>
#include <fstream.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>


#include "SIREN.h"
#include "s4io.h"
#include "s4time.h"

#define FOREVER 1

#define JD_JAN1_1970 2440587.5
#define SECS_TO_DAYS 0.000011574

// cross platform types
#ifdef IBMPC
#define U16 unsigned
#define U32 unsigned long
#define S16 int
#define S32 long
#define BYTE unsigned char
#endif

#ifdef SUN
#define U16 unsigned short
#define U32 unsigned long
#define S16 short int
#define S32 long
#define BYTE unsigned char
#endif

// data block types for data files
#define SCSTR  0x0400  // telescope string
#define CMMT   0x0500  // comment
#define HITS   0x0A00  // hits
#define SYSCFG 0X0B00  // system configuration info
#define ERR    0X1000  // error
#define ENG    0X2000  // engineering record
#define THROTTLE        0x4000  // data thinning
                                // additional data: 1 double


#define ERR_SYNTH        0X01
#define ERR_VECT_WARN    0X02
#define ERR_OVER_RUN     0X03
#define ERR_ARITH_OVFL   0X04
#define ERR_OTHER        0X05
#define ERR_BD_NOT_READY 0X06
#define BUFFA 0X01
#define BUFFB 0X02

#define SIZEOF_HITS 10
#define HITS_OVERHEAD 15


char *usage = "s4pcdbase [-s] S4_binary_file";

char CommentStr[80];
char ConfigStr[128];
char ScopeStr[80];
unsigned char EngStr[80];
char UtilStr255[256]; // one extra for terminator

U32 u32Synthesizer;
U32 u32Time;
U16 u16ADC;
BYTE btBuffer;
double dRA;
double dDec;

// globals
struct Hit
    {
    S32 u32BinNumber;
    U16 u16Power;
    U16 u16DThresh;
    BYTE btBinRes;
    BYTE btBoardNumber;
    } ThisHit;

FILE *fInputDataFile;
FILE *fLogFile;
long FilePosition, NumToRead, NumAvailable;
off_t CurrentFileSize;
char InputDataFileName[512]; // shouldn't be bigger'n this! 
char LogFileName[512];
BOOLEAN RecordStarted;

 
struct stat statbuf;


// function prototypes
BOOLEAN ReadNext(char *buf,long n);
void ReadNextLine(char *buf,long max_n,char eol);
void Init(char * filename);
void ErrExit(char *msg);
void ChangeEndian(unsigned char *var,int size);
void ConvertTime2StdEventRec(long u32Time,StdEventRec *ThisHit);
void ExitFunction(void);
void WriteToLog(char *thisMessage);
void WriteHit(Hit ThisHit);
void Fini();

//____________________________________________________________________
// MAIN
//--------------------------------------------------------------------
int main (int argc,char **argv)
  {
  int i;

  long lDay, lMonth, lYear, lScopeTime;
  double dHour, dScopeJulDay;

  double ThrottleThresh;

  BOOLEAN bSwapBytes;

  BOOLEAN bHaveScopeString = FALSE;

  if ((argc != 2) && (argc != 3))
    ErrExit("illegal number of parmameters");
  bSwapBytes = FALSE;
  if (argc == 3)
    {
    if (strcmp(argv[1],"-s") == 0)
       bSwapBytes = TRUE;
    }

  Init(argv[argc-1]); 

  while (FOREVER)
     {
     U16 u16BlockType;
     U16 u16BlockSize;
     BYTE btReader;
     U32 u32Reader;

     unsigned uNumberOfHits;

#ifdef DEBUG
     fprintf(stderr,"reading next block--------\n");
#endif
     RecordStarted = FALSE;

     if(ReadNext((char *)&u16BlockSize,sizeof(U16)));
       RecordStarted = TRUE;


     if (bSwapBytes)
	  ChangeEndian((unsigned char*)&u16BlockSize,sizeof(U16));

     ReadNext((char *)&u16BlockType,sizeof(U16));

     if (bSwapBytes)
	  ChangeEndian((unsigned char *)&u16BlockType,sizeof(U16));

       switch (u16BlockType)
	 {
         case THROTTLE :
                   ReadNext((char *)&ThrottleThresh, sizeof(double));
                   if (bSwapBytes)
                      ChangeEndian((unsigned char *)&ThrottleThresh,sizeof(double));
                   printf("----> throttle factor is %lf <----\n", ThrottleThresh);
                   break;
	 case HITS :
		   printf("++++++++++++++++++++++++++++++++++++++++++\n"); 
#ifdef DEBUG
		   printf("hits: blocksize %u\n ", u16BlockSize);
#endif
		   uNumberOfHits = (u16BlockSize - HITS_OVERHEAD) / SIZEOF_HITS;

		   ReadNext((char *)&u32Time,sizeof(U32));
		   if (bSwapBytes)
		      ChangeEndian((unsigned char *)&u32Time,sizeof(U32));
                   //printf("Time: %lu\n", u32Time);
                   printf("Time [%lu] (Year Day): %lu  ", u32Time, (u32Time >> 23));
                   printf("(Hour): %lu  ", ((0x0000001f) & (u32Time >> 18)));
                   printf("(Minute): %lu  ", ((0x0000003f) & (u32Time >> 12)));
                   printf("(Second): %lu  ", ((0x0000003f) & (u32Time >> 6)));
                   printf("(Source): %lu\n", ((0x00000001) & (u32Time)));


		   ReadNext((char *)&u32Synthesizer,sizeof(U32));
		   if (bSwapBytes)
		      ChangeEndian((unsigned char *)&u32Synthesizer,sizeof(U32));
                   printf("Synth: %lu\n", u32Synthesizer);

		   ReadNext((char *)&u16ADC,sizeof(U16));
		   if (bSwapBytes)
		      ChangeEndian((unsigned char *)&u16ADC,sizeof(U16));
                   printf("ADC: %hu\n", u16ADC);
                 
		   ReadNext((char *) &btBuffer,sizeof(BYTE));
		   if (btBuffer == BUFFA)
		     printf( "Buffer A\n");
		   else if (btBuffer == BUFFB)
		     printf( "Buffer B\n" );
		   else printf( "Buffer type error\n");
		   printf( "++++++++++++++++++++++++++++++++++++++++++\n"); 


		   for (i=0;i<uNumberOfHits;i++)
		       {
		       ReadNext((char *)&ThisHit.u32BinNumber,sizeof(U32));
		       if (bSwapBytes)
			  ChangeEndian((unsigned char *)&ThisHit.u32BinNumber,sizeof(U32));
		       ReadNext((char *)&ThisHit.u16Power,sizeof(U16));
		       if (bSwapBytes)
			  ChangeEndian((unsigned char *)&ThisHit.u16Power,sizeof(U16));
		       ReadNext((char *)&ThisHit.u16DThresh,sizeof(U16));
		       if (bSwapBytes)
			  ChangeEndian((unsigned char *)&ThisHit.u16DThresh,sizeof(U16));
		       ReadNext((char *)&ThisHit.btBinRes,sizeof(BYTE));
		       ReadNext((char *)&ThisHit.btBoardNumber,sizeof(BYTE));

                       WriteHit(ThisHit);
		       }
		   break;
	 case CMMT :   printf("======Comment==================\n");
		   ReadNext(CommentStr,u16BlockSize-4);
		   CommentStr[u16BlockSize-4] = '\0';
                   printf("%s\n", CommentStr);
		   break;
	 case SYSCFG : printf("::::::System Configuration:::::\n");
		   int bytesRead;
		   bytesRead = 0;
		   while (bytesRead < u16BlockSize-4)
		     {
                     ReadNextLine(ConfigStr,128,(char)'\0');
		     bytesRead += strlen(ConfigStr) + 1;
                     printf("%s\n", ConfigStr);
		     }
                   printf("bytes read: %d\n", bytesRead); 
		   break;
         case ENG : //ReadNext(UtilStr255,u16BlockSize-4); 
                    printf("====== Eng Record ==================\n");
                   ReadNext((char *)&u32Reader,sizeof(U32));
                   if (bSwapBytes)
                      ChangeEndian((unsigned char *)&u32Reader,sizeof(U32));
                   printf("Time [%lu] (Year Day): %lu  ", u32Reader, (u32Reader >> 23));
                   printf("(Hour): %lu  ", ((0x0000001f) & (u32Reader >> 18))); 
                   printf("(Minute): %lu  ", ((0x0000003f) & (u32Reader >> 12)));
                   printf("(Second): %lu  ", ((0x0000003f) & (u32Reader >> 6)));
                   printf("(Source): %lu\n", ((0x00000001) & (u32Reader)));
                   ReadNext((char *)EngStr,u16BlockSize-8);
                   for (i=0;i<u16BlockSize-8;i++)
                     printf(" [%hu] ", EngStr[i]);
                   printf("\n");
                   break;
	 case ERR :    printf("******Error********************\n");
		   ReadNext((char *)&btReader,sizeof(BYTE));

		   switch (btReader)
		      {
		      case ERR_SYNTH :
			   printf("Synthisizer error\n" );
			   break;
		      case ERR_VECT_WARN :
			   ReadNext((char *)&btReader,sizeof(BYTE));
                           printf("Vector Warning error on board: %u\n",
				 (unsigned) btReader);
			   break;
		      case ERR_OVER_RUN :
			   ReadNext((char *)&btReader,sizeof(BYTE));
			   printf("Over run error on board: %u\n ",
				 (unsigned) btReader);
			   break;
		      case ERR_ARITH_OVFL :
			   ReadNext((char *)&btReader,sizeof(BYTE));
			   printf("Arithmetic overflow error on board: %u\n ", 
				 (unsigned) btReader );
			   break;
		      case ERR_OTHER :
			   ReadNext((char *)&btReader,sizeof(BYTE));
			   printf("Other error on board:  %u\n", 
				 (unsigned) btReader);
			   break;
		      case ERR_BD_NOT_READY :
			   ReadNext((char *)&btReader,sizeof(BYTE));
			   printf("Board not ready error on board:  %u\n",
				 (unsigned) btReader );
			   break;
		      }
		   break;
	 case SCSTR : printf( "/\\/\\/\\Telescope String/\\/\\/\\/\\/\\/\n" );
		      ReadNext(ScopeStr,u16BlockSize-4);
		      printf("%s\n", ScopeStr);
                      if(sscanf(ScopeStr, "%ld %lf %lf", &lScopeTime, &dRA, &dDec) == 3)
                        {
                        bHaveScopeString = TRUE;
                        dScopeJulDay = lScopeTime * SECS_TO_DAYS + JD_JAN1_1970;
                        tm_FromJul(dScopeJulDay, &lMonth, &lDay, &lYear, &dHour);
                        }
		      break;

	 default : printf("Unknown.....\nblock size: %x\n", u16BlockSize);
                 ErrExit("Unexpected block type"); 
		 break;
       }

     }

  }


//--------------------------------------------------------------------
void ErrExit(char *msg)
//--------------------------------------------------------------------
  {
  fprintf(stderr,"%s\n", msg);
  fprintf(stderr, "usage: %s\n",usage );
  exit(1);
  }


//--------------------------------------------------------------------
void ChangeEndian(unsigned char *var,int size)
//--------------------------------------------------------------------
  {
  unsigned char myByteStr[8];
  int i;

  // simple, but it even works for odd sizes!
  if (size <= 8)
    {
    for (i=0;i<size;i++)
      myByteStr[i] = var[size-1-i];
    for (i=0;i<size;i++)
      var[i] = myByteStr[i];
    }

  }

//--------------------------------------------------------------------
void Init(char *filename)
//--------------------------------------------------------------------
  {
  strcpy(InputDataFileName,filename);
  fInputDataFile = fopen(InputDataFileName,"r");
  if (fInputDataFile == NULL)
    ErrExit("Can't open data file");
  fstat(fileno(fInputDataFile),&statbuf);
  CurrentFileSize = statbuf.st_size;
  FilePosition = 0;
  }


//--------------------------------------------------------------------
// reads next n characters from global file stream
void ReadNext_old(char *buf,long n)
//--------------------------------------------------------------------
   {
   NumAvailable = CurrentFileSize - FilePosition;
   while (n > 0) // decrement n on each read
      {
      if (NumAvailable >= n)
          {
          NumToRead = n;
          NumAvailable -= n;
          }
      else 
          {
          NumToRead = NumAvailable;
          NumAvailable = 0;
          }
      n -= NumToRead;
      fread(buf,sizeof(char),(int) NumToRead,fInputDataFile);
      buf += NumToRead; // increment buffer in case we didnt'read them all
      if (NumAvailable == 0)
        {
         while (statbuf.st_size == CurrentFileSize)
           {
           usleep(1000);
           fstat(fileno(fInputDataFile),&statbuf);
           }
         FilePosition = ftell(fInputDataFile);
         fclose(fInputDataFile);
         usleep(500000);
         fInputDataFile = fopen(InputDataFileName,"r");
         fstat(fileno(fInputDataFile),&statbuf);
         CurrentFileSize = statbuf.st_size;
         fseek(fInputDataFile,FilePosition,0);
         NumAvailable = CurrentFileSize - FilePosition;
        }
      }
    FilePosition = ftell(fInputDataFile);
   }

//--------------------------------------------------------------------
BOOLEAN ReadNext(char *buf,long n)
//--------------------------------------------------------------------
  {
  FilePosition = ftell(fInputDataFile);
  NumAvailable = CurrentFileSize - FilePosition;
  if(NumAvailable < n)
    Fini();
  else
    { 
    fread(buf,sizeof(char),(int) n,fInputDataFile);
    FilePosition = ftell(fInputDataFile);
    return TRUE;
    }
  }  

//--------------------------------------------------------------------
// reads next line, or up to n characters (whichever comes first)
// from global file stream, End of line is eol.
void ReadNextLine(char *buf,long max_n,char eol)
//--------------------------------------------------------------------
   {
   unsigned char NextChar;
   int NumCharsRead;
   NumCharsRead = 1;
   ReadNext((char *)&NextChar,1);
   buf[NumCharsRead-1] = NextChar;
   while((NextChar != eol) && (NumCharsRead < max_n)) 
      {
      NumCharsRead++;
      ReadNext((char *)&NextChar,1);
      buf[NumCharsRead-1] = NextChar;
      }
   }


//--------------------------------------------------------------------
void WriteHit(Hit ThisHit)
//--------------------------------------------------------------------
  {
  printf("--------------------------\n");
  printf("Bin Number: %ld %lx\n", ThisHit.u32BinNumber, ThisHit.u32BinNumber);
  printf("Power: %hu\n", ThisHit.u16Power);
  printf("Detection Threshold: %hu\n", ThisHit.u16DThresh);
  printf("Bin Resolution: %hu\n", ThisHit.btBinRes);
  printf("Board Number: %hu\n", ThisHit.btBoardNumber);
  printf("--------------------------\n");
  }

//--------------------------------------------------------------------
void Fini()
//--------------------------------------------------------------------
  { 
  int rtn;

  fprintf(stderr, "Exiting... ");
    if(RecordStarted)
      {  
      fprintf(stderr, "file ends in record fragment!\n");
      rtn = 2;				// not a real error if
      }  				//  we are looking at an
    else				//  acquiring file
      { 
      fprintf(stderr, "file ends on record boundry\n");   // normal exit
      rtn = 0;
      }
 
  exit(rtn);
  }

