/* ________________________________________________________________________
 programmer : Matt Lebofsky
 source name: s4tos3skyplot.cc
 compiler   : gcc
___________________________________________________________________________
 modification history:
   version 1.0 finished 6/29/98
___________________________________________________________________________
 usage: s4readskyplot [-d dbase_filename] [-radec RA Dec [+- #_of_pixels]]
                      [-rarange RA_minimum RA_maximum]
                      [-decrange Dec_minimum Dec_maximum]
                      [-s] [-n] [-t] 

 simply dumps the data from the skyplot_dbase file into three columns:
 RA, Dec, and # of observations. 

 The default (i.e. if you just enter "s4readskyplot skyplot_dbase") is it
 will output every pixel, one per line, that contains any # of observations
 greater than zero. By the way, the Ra and Dec's shown are for the exact
 center of the pixel in question, and may not be exactly what you expect.

 If you supply a dbase_filename with -d (i.e. -d 01jn98aa) output will
 be limited to only pixels containing this database. This is a quick and
 dirty way to see where a specific file looked.
 
 If you supply an ra and dec with -radec (i.e. -radec 5.0 10.0) output
 will be limited to only the pixel containing the specified ra and dec.
 However, you can add a +- flag (i.e. -radec 5.0 10.0 +- 20) which will
 output a range of plus or minus 20 pixels centered around the ra and dec.

 The default ra range is 0 to 24 hours and the default dec range is
 between -5.0 to +45.0 degrees. You can change these ranges via the
 -rarange and -decrange flags, i.e.:

 s4readskyplot -rarange 10.0 11.0 -decrange -3.0 -4.0 skyplot_file

 The -s flag will extend output to show every single dbase filename for every
 pixel that has at least one observation. The output will be more or less
 along the lines of:

 17.00000 18.00000 4
 - 01jn98aa
 - 02jn98ab
 - 02jn98ac
 - 05jn98ad
 17.22000 18.00000 3 ...

 The -n flag will cause s4readskyplot to NOT print any information 
 regarding ra, dec, and # of observations for any pixel with at least
 one observation. This is good when coupled with -s, and the output is
 then sort | uniq'ed, which will then give you a list of all the dbase
 files seen within the ra/dec range selected quickly and easily.

 I imagine once the skyplot dbase get pretty full the output will be pages
 per pixel. So be warned that you may want to restrict your ra and dec
 ranges to a very few pixels if you use the -s option.

 The -t flag will prints totals (of # of pixels checked, # of pixels with
 observations, average observations per pixel checked, etc.) at the end of
 the output.

 In lieu of reading all the above, here are some examples:
 
 s4readskyplot -radec 5.0 10.0 -s sky
 - will show the default ra/dec/#obs line for the pixel containing the
   ra and dec of 5.0 and 10.0, respectively, followed by a list of all the
   names of dbase files wherein this pixel was observed.

 s4readskyplot -radec 5.0 10.0 +- 10 -t sky
 - will show the ra/dec/#obs lines for the "box" centered around the ra and
   dec of 5.0 and 10.0, plus 10 pixels left/right/up/down from that center
   pixel (in this case, the box will be 21 x 21 pixels). At the bottom,
   thanks to the -t flag, a set of interesting totals will be displayed.

 s4readskyplot -rarange 5.0 5.22 sky 
 - will show the default output, limited to pixels containing the ra's
   between 5.0 and 5.22 in value.

 s4readskyplot -rarange 5.0 5.22 -decrange 10.0 20.0 -s -n | sort | uniq
 - will only show the dbase file names for the pixels within the ra range
   between 5.0 and 5.22 AND the dec range of 10.0 and 20.0, and this
   output is then sorted and uniq'ed, which yields a nice, short list
   containing only the dbase file names which observed in this ra/dec range.

 s4readskyplot -d 01jn98aa -s -n | sort | uniq
 - show the names of all the dbase files that looked where 01jn98aa looked.

__________________________________________________________________________
notes/bugs:
 If you select a center dec with a range of +- N pixels, and this
 range goes out of bounds (i.e. dec = -5.0 +- 10 pixels), the range does
 NOT wrap around. However, the ra ranges do wrap around.
___________________________________________________________________________
*/


#include <stdio.h>
#include <string.h>
#include "SIREN.h"
#include "s4database.h"
#include "s4io.h"
#include "s4skyplot.h"

char *usage = "Usage: s4readskyplot [-d dbase_filename] [-radec RA Dec [+- #_of_pixels]]\n      [-rarange RA_minimum RA_maximum]\n      [-decrange Dec_minimum Dec_maximum]\n      [-s] [-n] [-t]\n";

void UsageExit();

// ------------------------------------------------------------------------
int main (int argc, char *argv[])
// ------------------------------------------------------------------------
{

SkyplotElement *stSkyArray;
int iSkyplotFileNo;
long lArraySize;
long lRaIndex,lDecIndex;
long lRealRa;                      // Used for RA wrap-arounds
long lRaStart,lRaEnd;              // Ra Range to observe
long lDecStart,lDecEnd;		   // Dec Range to observe
long lCount = 0;                       // For counting observations in a pixel
double dRaInput,dDecInput;         // Used for getting values from command line
long lPixelRange;                  // +- value supplied by user
BOOLEAN bShowNames;                // Show the dbase files names at each pixel? 
BOOLEAN bShowTotals;               // Show totals at the end?
BOOLEAN bShowRaDecObs;             // Show the default ra/dec/#obs lines?
unsigned long ulDbaseKeyToCheck;   // For limiting output to a single dbase
char cDbaseFileName[9];
long lTotalPixels;                 // These three are for totals at the end
long lTotalObs;
long lSeenPixels;

long i;

// Defaults below:

lRaStart = 0;                      // Default ra/dec range = entire map
lRaEnd = 1799;
lDecStart = 0;
lDecEnd = 199;
bShowNames = FALSE;                // Do not show names by default
ulDbaseKeyToCheck = 0;             // = 0 means don't limit output to one key 
bShowTotals = FALSE;               // Do not show totals by default
lTotalPixels = 0;
lTotalObs = 0;
lSeenPixels = 0;
bShowRaDecObs = TRUE;              // Show ra/dec/obs# lines by default

// Get Arguments:

if (argc < 1) UsageExit();

for (i = 1; i < argc; i++) {
  if (strcmp(argv[i],"-d") == 0) {
    if (sscanf(argv[++i],"%s",&cDbaseFileName[0]) != 1)
      UsageExit();
    ulDbaseKeyToCheck = db_FileNameToKey(cDbaseFileName);
    }
  else if (strcmp(argv[i],"-radec") == 0) {
    if (sscanf(argv[++i],"%lf",&dRaInput) != 1)
      UsageExit();
    lRaStart = SKYPLOT_CONVERT_RA(dRaInput);
    lRaEnd = lRaStart;
    if (sscanf(argv[++i],"%lf",&dDecInput) != 1)
      UsageExit();
    lDecStart = SKYPLOT_CONVERT_DEC(dDecInput);
    lDecEnd = lDecStart;
    if (strcmp(argv[++i],"+-") == 0) {
      if (sscanf(argv[++i],"%ld",&lPixelRange) != 1)
        UsageExit();
      lRaStart = lRaStart - lPixelRange;
      lRaEnd = lRaEnd + lPixelRange;
      lDecStart = lDecStart - lPixelRange;
      lDecEnd = lDecEnd + lPixelRange;
      }
    else i--;
    }
  else if (strcmp(argv[i],"-rarange") == 0) {
    if (sscanf(argv[++i],"%lf",&dRaInput) != 1)
      UsageExit();
    lRaStart = SKYPLOT_CONVERT_RA(dRaInput);
    if (sscanf(argv[++i],"%lf",&dRaInput) != 1)
      UsageExit();
    lRaEnd = SKYPLOT_CONVERT_RA(dRaInput);
    }
  else if (strcmp(argv[i],"-decrange") == 0) {
    if (sscanf(argv[++i],"%lf",&dDecInput) != 1)
      UsageExit();
    lDecStart = SKYPLOT_CONVERT_DEC(dDecInput);
    if (sscanf(argv[++i],"%lf",&dDecInput) != 1)
      UsageExit();
    lDecEnd = SKYPLOT_CONVERT_DEC(dDecInput);
    }
  else if (strcmp(argv[i],"-s") == 0)
    bShowNames = TRUE;
  else if (strcmp(argv[i],"-n") == 0)
    bShowRaDecObs = FALSE;
  else if (strcmp(argv[i],"-t") == 0)
    bShowTotals = TRUE;
  else 
    UsageExit();
  } 

if ((lArraySize = sp_ReadSkyplotFile(&stSkyArray,&iSkyplotFileNo)) == -1)
  exit(1);

for (i = 0; i < 80000000 ; i++);

if (lDecStart < 0) lDecStart = 0; // Make sure dec doesn't go out of range.
if (lDecEnd > 199) lDecEnd = 199;  
  
for (lRaIndex = lRaStart; lRaIndex <= lRaEnd; lRaIndex++) {

  // First, make sure RA gets wrapped around instead of going out of range:
  lRealRa = lRaIndex; 
  while (lRealRa < 0) lRealRa = lRealRa + 1800;
  while (lRealRa > 1799) lRealRa = lRealRa - 1800;

  for (lDecIndex = lDecStart; lDecIndex <= lDecEnd; lDecIndex++) {
    lTotalPixels++;

    // Check to see if we're and'ing output with a single dbase key (-d flag).
    // If so, restrict output accordingly using sp_FindDbase.
    if (ulDbaseKeyToCheck != 0) { 
      if ((lCount = sp_CountObs(SKYPLOT_R_CONVERT_RA(lRealRa),
                    SKYPLOT_R_CONVERT_DEC(lDecIndex),stSkyArray))) {
        if (sp_FindDbase(SKYPLOT_R_CONVERT_RA(lRealRa),
                         SKYPLOT_R_CONVERT_DEC(lDecIndex),cDbaseFileName,
                         stSkyArray)) {
          if (bShowRaDecObs) printf("%lf %lf %ld\n",
                                    SKYPLOT_R_CONVERT_RA(lRealRa),
                                    SKYPLOT_R_CONVERT_DEC(lDecIndex),lCount); 
          lTotalObs = lTotalObs + lCount;
          lSeenPixels++;
          if (bShowNames) sp_ShowObs(SKYPLOT_R_CONVERT_RA(lRealRa),
                                 SKYPLOT_R_CONVERT_DEC(lDecIndex),stSkyArray);
          }
        }
      } 

    // Same logic as above, except we aren't using sp_FindDbase to find any
    // specific database.
    else {
      if ((lCount = sp_CountObs(SKYPLOT_R_CONVERT_RA(lRealRa),
                              SKYPLOT_R_CONVERT_DEC(lDecIndex),stSkyArray))) {
        if (bShowRaDecObs) printf("%lf %lf %ld\n",
                                  SKYPLOT_R_CONVERT_RA(lRealRa),
                                  SKYPLOT_R_CONVERT_DEC(lDecIndex),lCount); 
        lTotalObs = lTotalObs + lCount;
        lSeenPixels++;
        if (bShowNames) sp_ShowObs(SKYPLOT_R_CONVERT_RA(lRealRa),
                                 SKYPLOT_R_CONVERT_DEC(lDecIndex),stSkyArray);
        }
      }
    }
  }
   
if (bShowTotals) {
  printf("\nTotal # of Pixels Checked: %ld\n",lTotalPixels);
  printf("Total # of Observations within these Pixels: %ld\n",lTotalObs);
  if (lTotalPixels)
    printf("Average Observations per Pixels Checked: %lf\n", (double)lTotalObs/(double)lTotalPixels);
  printf("Total # of Pixels with at Least One Observation: %ld\n",lSeenPixels);
  if (lSeenPixels)
    printf("Average Observations per Pixels with at Least One Observation: %lf\n",(double)lTotalObs/(double)lSeenPixels);
  }

sp_CloseSkyplotFile(iSkyplotFileNo);

}

// UsageExit: Problems with the usage? Here you go:

void UsageExit()
{
  fprintf(stderr,usage);
  exit(1);
}

