/***************************************************************************
 * Source Code: s4skyplot
 * Purpose    : Takes as input a SERENDIP dbase file.  Distributes
 *              ra and dec into a 2-dimensional array, and keeps track of
 *              occurences in each cell.  It counts occurrences cumulatively
 *              from one invocation of skyplot to the next. Saves accumlation
 *              as array.
 *
 * Calculations:The plot is 1800 X 200 : RA is divided into 1800 bins, 
 *              and DEC is divided into 200 bins.
 *
 *              24 hours of RA / 1800 bins = 0.0133333 hours/bin. Thus there 
 *              are 75 bins in 1 hour of RA.  This means that an RA of 
 *              0.01 hours goes in the (0.01 * 75) bin.
 *
 *              Because we are casting to int, the fractional portion is
 *              thrown away and the bin ends up being zero, which is correct. 
 *              In a similar fashion, .5 hours of RA = bin 37,
 *              1 hour of RA = 75, etc.
 *
 *              Our DEC coverage is said to be -5 degrees to +45 degrees, for 
 *              a range of 50 degrees.  First we add 5 to get rid of 
 *              negative numbers.  Then we multiply by 4, since there are 0.25
 *              degrees/dec_bin.  This means that -5 degrees = (-5+5)*4 =
 *              bin 0 and 1 degree = (1 + 5)*4 = bin 24, etc.
 *
 *              These calculations approximate beam width. 
 ************************************************************************** 
 * Modifications:
 *      5/5/95 - scott converted to ANSI and SERENDIP IV
 **************************************************************************
 */

#include <stdio.h>
#include <string.h>
#include "SIREN.h"
#include "s4io.h"

#define PLOTHEIGHT 200     /* number of dec bins */
#define PLOTWIDTH 1800     /* number of ra bins  */

#define RA_FACTOR 75.0    /* 75 bins in 1 hour of ra */
#define DEC_FACTOR 4.0    /* 4 bins in 1 degree of dec */
#define DEC_OFFSET 5.0    /* negative number adjuster */ 


int main(int argc, char *argv[])

{

int i;
int ra_slot;            /* ra "slot" (position 0-PLOTWIDTH) in skyplot array */
int dec_slot;          /* dec "slot" (position 0-PLOTHEIGHT) in skyplot array */
int total_size;        /* size of skyplot array (plotwidth X plotheight) */

char incoming2;        /* character byte used to "hold" each */
                            /* character tally from the accumulator file */
                            /* when the accumulator file is read */

unsigned char skyplot [PLOTHEIGHT] [PLOTWIDTH];  /* array for tallying  */

FILE *inFile;
FILE *outFile;


StdEventRec Event;


	if(argc != 3)
	  {
	  printf("Usage: s4skyplot sourceFile accumulatorFile\n");
	  exit(1);
	  }

	total_size = PLOTHEIGHT * PLOTWIDTH;

		 /* Begins by initializing skyplot array to all zeros */

	memset((char *) skyplot, '\0', PLOTWIDTH * PLOTHEIGHT * sizeof(char));


	  /* Distributes events in input file into 2D array, marks as
             1 occurrence.  By definition, if multiple events at same
             ra and dec occur within same file
             they will be counted as one observation for that file.
	  */

	if ((inFile = fopen(argv[1], "r")) != NULL)
           {

			 /* read one event from dbas file into memory */

	    while (io_GetNextRecord(inFile,&Event,STDEVENT_REC))
		{

			/* determine which slots the event falls into */

		ra_slot = (int) (RA_FACTOR * Event.dbRa);
		dec_slot = (int) (DEC_FACTOR * (Event.dbDec + DEC_OFFSET));

			  /* indicate a tally of 1 for this specific slot */

		skyplot[dec_slot][ra_slot] = 1;

		}
           }
	 else
            {
            printf("ERROR: Input File %s not located.\n", argv[1]);
            exit(1);
            }

	/* Reads in tally of earlier counts from accumulation file
          (if any).  Adds the binary count from  input dbas file
	   to the accumluation count. If no accumulation file
           exists, a new accumulation file is created.
	*/

       if((outFile = fopen(argv[2], "r")) != NULL)
	  {
	  for (i = 0, incoming2 = '\0'; i < total_size; i++)
	   {
		  /* read one character into incoming2 address */

	   fread(&incoming2, sizeof(char), 1, outFile);

		  /* insure character is not the null character
                     before processing */

	  if (incoming2 != '\0')

       		/* using dimensions of the array, determine bins*/

           {
           dec_slot = i/PLOTWIDTH;    /* from (0 to amount of PLOTHEIGHT) */

	   ra_slot  = i%PLOTWIDTH;    /* essentially = i (from 0 to PLOTWIDTH)*/

	   skyplot[dec_slot][ra_slot] = skyplot[dec_slot][ra_slot] + incoming2;
	   }
         }
       fclose (outFile);
       }

     /* Save back to disk with name specified in command line. If a new name           (or typo) appears in the command line, a new tally is saved under
        the name of this file.
     */

	outFile = fopen(argv[2], "w");

	fwrite((char *)skyplot, sizeof(char),total_size, outFile);

	return 0;
}
