!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
! Computes the spherical harmonics coefficients (a_lm) 
! and the power spectrum from a full sky map in HEALPIX pixelisation
! Written and developed by E. Hivon (efh@ipac.caltech.edu) and K. Gorski 
! (kgorski@eso.org) based on HEALPIX pixelisation developed 
! by K. Gorski (kgorski@eso.org)
!
! The permission to use and copy this software and its documentation, 
! without fee or royalty is limited to non-commercial purposes related to 
! PLANCK Surveyor project and provided that you agree to comply with
! the following copyright notice and statements,
! and that the same appear on ALL copies of the software and documentation.
!
! An appropriate acknowledgement has to be included in any
! publications based on work where the package has been used.
!
! Copyright 1997 by Eric Hivon and Krzysztof M. Gorski.
!  All rights reserved.
!
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!=======================================================================
PROGRAM anafast
  !=======================================================================
  !  EXTERNAL LIBRARY:
  !     this code uses the CFITSIO library that can be found at
  !     http://heasarc.gsfc.nasa.gov/docs/software/fitsio/
  !
  !  RELATED LITTERATURE:
  !     about HEALPIX   : see Gorski, 1997, in preparation
  !     about this code : see Hivon & Gorski, 1997, in preparation
  !
  !  HISTORY:
  !     April-October 1997, Eric Hivon, TAC (f77)
  !               Jan 1998 : translation in Fortran 90
  !               Aug 1998 : version 1.0.0
  !               Apr 1999 : precomp. plms, higher order iterations and alm dumping, FKH
  !               Aug 2000 : CMBFAST convention for polarisation
  !               Sep 2002 : implement new parser
  !
  !  FEEDBACK:
  !     for any questions : efh@ipac.caltech.edu
  !
  !=======================================================================
  !     version 1.2.0
  !=======================================================================
  USE healpix_types
  USE alm_tools, ONLY : map2alm, alm2map
  USE fitstools, ONLY : getsize_fits, input_map, read_par, read_dbintab, write_asctab, dump_alms
  USE head_fits, ONLY : add_card
  USE utilities, ONLY : die_alloc
  USE pix_tools, ONLY : convert_nest2ring, nside2npix, npix2nside, remove_dipole, vec2ang
  USE extension, ONLY : getEnvironment, getArgument, nArguments
  USE paramfile_io, ONLY : paramfile_handle, parse_init, parse_int, &
         parse_string, parse_double, concatnl, scan_directories
  IMPLICIT none
  INTEGER(I4B) :: nlmax
  INTEGER(I4B) :: nsmax
  INTEGER(I4B) :: nmmax

  COMPLEX(KIND=SPC), DIMENSION(:,:,:), ALLOCATABLE :: alm_TGC
  COMPLEX(KIND=SPC), DIMENSION(:,:,:), ALLOCATABLE :: alm_TGC2
  REAL(KIND=SP),     DIMENSION(:,:),   ALLOCATABLE :: map_TQU
  REAL(KIND=SP),     DIMENSION(:),     ALLOCATABLE :: powspec_TT, powspec_TG, &
       &   powspec_GG, powspec_CC
  REAL(KIND=SP),     DIMENSION(:),     ALLOCATABLE :: powspec_TC, powspec_GC
  REAL(KIND=SP),     DIMENSION(:,:),   ALLOCATABLE :: clout
  REAL(KIND=SP),     DIMENSION(:),     ALLOCATABLE :: fact_norm
  REAL(KIND=DP),     DIMENSION(:,:),   ALLOCATABLE :: w8ring_TQU
  REAL(KIND=SP),     DIMENSION(:,:),   ALLOCATABLE :: dw8
  REAL(KIND=DP),     DIMENSION(:,:),   ALLOCATABLE :: plm
  INTEGER(KIND=I4B), DIMENSION(:),     ALLOCATABLE :: fact
  INTEGER(KIND=I4B) :: status

  INTEGER(KIND=I4B) :: l, mlpol, polar_fits
  INTEGER(KIND=I4B) :: n_plm, plm_nside, plm_lmax, plm_pol
  INTEGER(KIND=I4B) :: npixtot, nmaps
  INTEGER(KIND=I4B) :: ordering, order_type
  INTEGER(KIND=I4B) :: i, j, sign
  INTEGER(KIND=I4B) :: iter_order, polar
  INTEGER(KIND=I4B) :: simul_type
  INTEGER(KIND=I4B) :: n_rings, n_rings_read, won
  integer(kind=i4b) :: type
  INTEGER(KIND=I4B), DIMENSION(8,2) :: values_time

  REAL(KIND=SP) :: pix_size_arcmin
  REAL(KIND=SP) :: fmissval
  REAL(KIND=SP) :: clock_time
  real(kind=sp),    parameter :: fbad_value = -1.6375e30_sp
  REAL(KIND=DP) :: theta_cut_deg, cos_theta_cut
  REAL(KIND=DP) :: nullval

  CHARACTER(LEN=80), DIMENSION(1:120) :: header
  INTEGER :: ncl, nlheader

  CHARACTER(LEN=filenamelen)          :: infile
  CHARACTER(LEN=filenamelen)          :: outfile
  CHARACTER(LEN=filenamelen)          :: parafile = ''
  CHARACTER(LEN=filenamelen)          :: outfile_alms
  CHARACTER(LEN=filenamelen)          :: infile_plm
  CHARACTER(LEN=filenamelen)          :: infile_w8
  CHARACTER(LEN=filenamelen)          :: w8name
  CHARACTER(LEN=filenamelen)          :: def_dir, def_file
  CHARACTER(LEN=filenamelen)          :: usr_dir, usr_file
  CHARACTER(LEN=filenamelen)          :: final_file
  CHARACTER(LEN=filenamelen)          :: healpixdir
  character(len=filenamelen)          :: description
  character(len=100)                  :: chline, chline1
  CHARACTER(LEN=5)                    :: sstr
  LOGICAL(KIND=LGT) :: bad, ok, polarisation, bcoupling
  CHARACTER(LEN=*), PARAMETER :: code = "ANAFAST"
  CHARACTER(LEN=*), PARAMETER :: version = "1.2.0"
  character(len=80), dimension(1:3)   :: units_map, units_pow

  type(paramfile_handle) :: handle

  real(kind=DP), dimension(0:3) :: mono_dip
  real(kind=DP)                 :: theta_dip, phi_dip
  integer(kind=I4B)             :: lowlreg
  real(kind=SP)                 :: long_dip, lat_dip

  !-----------------------------------------------------------------------
  !                    get input parameters and create arrays
  !-----------------------------------------------------------------------

  call date_and_time(values = values_time(:,1))
  !     --- read parameters interactively if no command-line arguments
  !     --- are given, otherwise interpret first command-line argument
  !     --- as parameter file name and read parameters from it:
  if (nArguments() == 0) then
     parafile=''
  else
     if (nArguments() /= 1) then
        print "(' Usage: anafast [parameter file name]')"
        stop 1
     end if
     call getArgument(1,parafile)
  end if

  !     --- announces program, default banner ***
  PRINT *, " "
  PRINT *, "                "//code//" "//version
  write(*,'(a)') " *** Analysis and power spectrum calculation for a Temperature/Polarisation map ***"
  PRINT *, " "
  handle = parse_init(parafile)

  !     --- choose temp. only or  temp. + pol. ---
  description = concatnl( &
       & " Do you want to analyse", &
       & " 1) a Temperature only map", &
       & " 2) Temperature + polarisation maps")
  simul_type = parse_int(handle, 'simul_type', default=1, vmin=1, vmax=2, descr=description)
  if (simul_type .eq. 1) polarisation = .false.
  if (simul_type .eq. 2) polarisation = .true.
  bcoupling = polarisation ! <<<<
!   bcoupling = .false. ! <<<<
  print*, " Bcoupling :  ",bcoupling

  !     --- gets the file name for the map ---
  chline = "''"
  call getEnvironment("HEALPIX",healpixdir)
  if (trim(healpixdir)/='') chline = trim(healpixdir)//'/test/map.fits'
  description = concatnl( &
       & "", &
       & " Enter input file name (Map FITS file): ")
  infile = parse_string(handle, 'infile',default=chline, descr=description, filestatus='old')

  !     --- finds out the pixel number of the map and its ordering ---
  npixtot = getsize_fits(infile, nmaps = nmaps, ordering=ordering, nside=nsmax, mlpol=mlpol, type = type, polarisation = polar_fits)
  if (nsmax<=0) then 
     print*,"Keyword NSIDE not found in FITS header!"
     stop 1
  endif
  if (type == 3) npixtot = nside2npix(nsmax) ! cut sky input data set
  if (nsmax/=npix2nside(npixtot)) then
     print*,"FITS header keyword NSIDE does not correspond"
     print*,"to the size of the map!"
     stop 1
  endif

  if (polarisation .and. (nmaps >=3) .and. polar_fits == -1) then
     print*,"The input fits file MAY NOT contain polarisation data."
     print*,"Proceed at your own risk"
  endif

  if (polarisation .and. (nmaps<3 .or. polar_fits ==0)) then
     print *,"The file does NOT contain polarisation maps"
     print *,"only the temperature field will be smoothed"
     polarisation = .false.
  endif

  !     --- check ordering scheme ---

  if ((ordering/=1).and.(ordering/=2)) then
     PRINT*,"The ordering scheme of the map must be RING or NESTED."
     PRINT*,"No ordering specification is given in the FITS-header!"
     stop 1
  endif

  order_type = ordering

  !     --- gets the L range for the analysis ---
  WRITE(chline1,"(a,i5)") " The map has Nside = ",nsmax
  if (mlpol<=0) WRITE(chline,"(a,i5,a)") "We recommend: (0 <= l <= l_max <= ",3*nsmax-1,")"
  if (mlpol> 0) WRITE(chline,"(a,i5,a)") "We recommend: (0 <= l <= l_max <= ",mlpol,")"
  description = concatnl(&
       & "", &
       & chline1, &
       & " Enter the maximum l range (l_max) for the simulation. ", &
       & chline )
  nlmax = parse_int(handle, 'nlmax', default=2*nsmax, descr=description)

  !     --- ask about monopole/dipole removal ---
  description = concatnl(&
       & "", &
       & " To improve the temperature map analysis (specially on a cut sky) ", &
       & " you have the option of first regressing out the best fit monopole and dipole", &
       & "   NOTE : the regression is made on valid (unflagged) pixels, ", &
       & " out of the symmetric cut (if any)", &
       & "  Do you want to : ", &
       & " 0) Do the analysis on the raw map", &
       & " 1) Remove the monopole (l=0) first ", &
       & " 2) Remove the monopole and dipole (l=1) first")
  lowlreg = parse_int(handle, 'regression', vmin=0, vmax=2, default=0, descr=description)
  

  !     --- gets the cut applied to the map ---
  description = concatnl(&
       & "", &
       & " Enter the symmetric cut around the equator in DEGREES : ", &
       & " (One ignores data within |b| < b_cut)     0 <= b_cut = ")
  theta_cut_deg = parse_double(handle, 'theta_cut_deg', &
       &                       vmin=0.0_dp, default=0.0_dp, descr=description)
  cos_theta_cut = SIN(theta_cut_deg/180.d0*PI) !counted from equator instead of from pole

  write(*,"(a,f6.1,a)") "One keeps ",100.*(1.-cos_theta_cut)," % of the original map"
  if (theta_cut_deg<1e-4) cos_theta_cut = -1000.d0

  !     --- gets the order of iteration ---
  description = concatnl(&
       & "", &
       & " Do you want : ", &
       & " 0) a standard analysis", &
       & " 1,2,3,4....) an iterative analysis", &
       & " (enter order of iteration, 3rd order is usually optimal)")
  iter_order=parse_int(handle, 'iter_order', vmin=0, default=0, descr=description)
  if (.not. handle%interactive) then
     select case (iter_order)
     case (0)
        PRINT*," Standard analysis"      
     case default
        PRINT*," Iterative analysis"
     end select
  end if


  !          --- precomputed plms  ---
40 continue
  n_plm   = 0
  plm_pol = 0
  description = concatnl(&
       & "", &
       & " Enter name of file with precomputed p_lms :", &
       & "(eg, plm.fits) (if '' is typed, no file is read)" )
  infile_plm = parse_string(handle, 'plmfile', default="''", descr=description, filestatus='old')

  if (trim(infile_plm) /= "") then
     ! check that the plm file matches the current simulation (for nside and lmax)
     call read_par(infile_plm, plm_nside, plm_lmax, plm_pol)

     if ((plm_pol.ne.1).and.(.not.polarisation)) plm_pol=1

     if ((plm_nside.ne.nsmax).or.(plm_lmax.ne.nlmax)) then
        PRINT *," "//code//"> Plm file does not match nside or l_max ! "
        if (handle%interactive) goto 40
        stop 1
     endif
     n_plm = (nlmax+1)*(nlmax+2)*nsmax
     PRINT *," "
  endif

  !     --- gets the output power spectrum filename ---
  description = concatnl(&
       & "", &
       & " Enter Output Power Spectrum file name (eg, cl_out.fits) :", &
       & "  (or !cl_out.fits to overwrite an existing file)" )
  outfile = parse_string(handle, "outfile", &
       default="!cl_out.fits", descr=description, filestatus="new")

  !     --- gets the output alm-filename ---
  description = concatnl(&
       & "", &
       & " Enter output file name for alms(eg: alms.fits): ", &
       & " (or !alms.fits to overwrite an existing file): ", &
       & " (If '', the alms are not written to a file) ")
  outfile_alms = parse_string(handle, 'outfile_alms', default="''", descr=description, filestatus='new')

  !-----------------------------------------------------------------------
  !                      ask for weights
  !-----------------------------------------------------------------------

  description = concatnl(&
       & "", &
       & " Do you want to use ring weights in the analysis (1=yes,2=no) ?")
  won = parse_int(handle, 'won', vmin=1, vmax=2, default=2, descr=description)

  infile_w8=""

  if (won.eq.1) then

     ! default weight file name
     write (sstr,"(I5.5)") nsmax
     w8name="weight_ring_n"//trim(sstr)//".fits"
     
     def_file = trim(w8name)
     def_dir  = concatnl("","../data","./data","..")
     call getEnvironment("HEALPIX",healpixdir)
     if (trim(healpixdir) .ne. "") then
        def_dir = concatnl(&
             & def_dir, &
             & healpixdir, &
             & trim(healpixdir)//"data", &
             & trim(healpixdir)//"/data", &
             & trim(healpixdir)//char(92)//"data") !backslash
     endif

22   continue
     final_file = ''
     ok = .false.
     ! if interactive, try default name in default directories first
     if (handle%interactive) ok = scan_directories(def_dir, def_file, final_file)
     if (.not. ok) then 
        ! not found, ask the user
        description = concatnl("",&
             &        " Could not find weight file", &
             &        " Enter the directory where this file can be found:")
        usr_dir = parse_string(handle,'w8filedir',default='',descr=description)
        if (trim(usr_dir) == '') usr_dir = trim(def_dir)
        description = concatnl("",&
             &        " Enter the name of the weight file:")
        usr_file = parse_string(handle,'w8file',default=def_file,descr=description)
        ! look for new name in user provided or default directories
        ok   = scan_directories(usr_dir, usr_file, final_file)
        ! if still fails, crash or ask again if interactive
        if (.not. ok) then
           print*,' File not found'
           if (handle%interactive) goto 22
           stop 1
        endif
     endif
     infile_w8 = final_file
  endif
  
  PRINT *," "

  !-----------------------------------------------------------------------
  nmmax = nlmax
  pix_size_arcmin = 360.*60./SQRT(12.*pi)/float(nsmax)

  !-----------------------------------------------------------------------
  !              allocate space for arrays
  !-----------------------------------------------------------------------
  polar=0
  if (polarisation) then
     polar=1
     ALLOCATE(map_TQU(0:12*nsmax**2-1,1:3),stat = status)
     if (status /= 0) call die_alloc(code,"map_TQU")

     ALLOCATE(w8ring_TQU(1:2*nsmax,1:3),stat = status)
     if (status /= 0) call die_alloc(code,"w8ring_TQU")

     ALLOCATE(dw8(1:2*nsmax,1:3),stat = status)
     if (status /= 0) call die_alloc(code,"dw8")
  else
     ALLOCATE(map_TQU(0:12*nsmax**2-1,1:1),stat = status)
     if (status /= 0) call die_alloc(code,"map_TQU")

     ALLOCATE(w8ring_TQU(1:2*nsmax,1:1),stat = status)
     if (status /= 0) call die_alloc(code,"w8ring_TQU")

     ALLOCATE(dw8(1:2*nsmax,1:1),stat = status)
     if (status /= 0) call die_alloc(code,"dw8")
  endif

  ALLOCATE(powspec_TT(0:nlmax),stat = status)
  if (status /= 0) call die_alloc(code,"powspec_TT")

  ALLOCATE(fact_norm(0:nlmax),stat = status)
  if (status /= 0) call die_alloc(code,"fact_norm")

  if (n_plm/=0) then
     ALLOCATE(plm(0:n_plm-1,1:plm_pol),stat = status)    
     if (status /= 0) call die_alloc(code,"plm")
  endif

  if (polarisation) then
     ALLOCATE(powspec_GG(0:nlmax),stat = status)
     if (status /= 0) call die_alloc(code,"powspec_GG")
     ALLOCATE(powspec_TG(0:nlmax),stat = status)
     if (status /= 0) call die_alloc(code,"powspec_TG")
     ALLOCATE(powspec_CC(0:nlmax),stat = status)
     if (status /= 0) call die_alloc(code,"powspec_CC")
     if (bcoupling) then
        ALLOCATE(powspec_TC(0:nlmax),stat = status)
        if (status /= 0) call die_alloc(code,"powspec_TC")
        ALLOCATE(powspec_GC(0:nlmax),stat = status)
        if (status /= 0) call die_alloc(code,"powspec_GC")
     endif
  endif

  !-----------------------------------------------------------------------
  !                      reads the map
  !-----------------------------------------------------------------------
  PRINT *,"      "//code//"> Inputting map " 

  fmissval = 0.0
  if (lowlreg > 0) fmissval = fbad_value
  if (polarisation) then
     nmaps = 3
     call input_map(infile, map_TQU(0:,1:nmaps), &
        &  npixtot, nmaps, fmissval=fmissval, units=units_map(1:nmaps))
  else
     nmaps = 1
     call input_map(infile, map_TQU(0:,1:nmaps), &
        &  npixtot, nmaps, fmissval=fmissval, units=units_map(1:nmaps))
  endif
  units_map(1) = adjustl(units_map(1))
  units_pow(1) = trim(units_map(1))//'^2'
  !-----------------------------------------------------------------------
  !                      remove dipole
  !-----------------------------------------------------------------------

  if (lowlreg > 0) then
     PRINT *,"      "//code//"> Remove monopole (and dipole) from Temperature map" 
     call remove_dipole(nsmax, map_TQU(0:npixtot-1,1), 1, ordering, lowlreg, &
          & mono_dip(0:), cos_theta_cut, fmissval=fmissval)

     if (fmissval /= 0.0) then
        do j=1,nmaps
           do i=0, npixtot-1
!               if (abs(map_TQU(i,j)/fmissval-1.0) < 1.e-6*fmissval) then
              if (abs(map_TQU(i,j)-fmissval) <= abs(1.e-6*fmissval)) then
                 map_TQU(i,j) = 0.0_sp
              endif
           enddo
        enddo
     endif

     write(unit=*,fmt="(a,g13.3,a)")  " Monopole = ",&
          & mono_dip(0)," "//trim(units_map(1))
     if (lowlreg > 1) then
        call vec2ang( mono_dip(1:3), theta_dip, phi_dip)
        write(unit=*,fmt="(a,g13.3,a)")  " Dipole   = ",&
             & sqrt(sum(mono_dip(1:3)**2))," "//trim(units_map(1))
        long_dip =      phi_dip  /PI*180.0
        lat_dip  = 90.0-theta_dip/PI*180.0
        write(unit=*,fmt="(a,g9.3,', ',g9.3,a)")  &
             & "(long.,lat.) = (",long_dip,lat_dip,") Deg"
     endif

  endif
  !-----------------------------------------------------------------------
  !                      Input weights
  !-----------------------------------------------------------------------

  dw8=0.

  if (trim(infile_w8)/="") then


     n_rings = 2 * nsmax
     n_rings_read = getsize_fits(infile_w8)
     
     if (n_rings_read /= n_rings) then
        PRINT *," "
        print *,"wrong file"
        stop
     endif
     
     if (polarisation) then
        PRINT *,"      "//code//"> Inputting weights "  
        call input_map(infile_w8, dw8, n_rings, 3, fmissval=0.0_sp)
     else
        PRINT *,"      "//code//"> Inputting weights "  
        call input_map(infile_w8, dw8, n_rings, 1, fmissval=0.0_sp)
     endif
     
  endif

  w8ring_TQU = 1.d0 + dw8

  !-----------------------------------------------------------------------
  !                 convert to RING if necessary
  !-----------------------------------------------------------------------
  if (order_type == 2) then
     PRINT *,"      "//code//"> Convert Nest -> Ring "      
     call convert_nest2ring (nsmax, map_TQU(:,1))
     if (polarisation) then
        call convert_nest2ring (nsmax, map_TQU(:,2))
        call convert_nest2ring (nsmax, map_TQU(:,3))
     endif
  endif

  !-----------------------------------------------------------------------
  !                    precomputed plms
  !-----------------------------------------------------------------------

  if (n_plm/=0) then
     if (plm_pol.eq.1) then
        PRINT*,"      "//code//"> Reading precomputed scalar P_lms "
     else
        PRINT*,"      "//code//"> Reading precomputed tensor P_lms "
     endif
     call read_dbintab(infile_plm,plm,n_plm,plm_pol,nullval,anynull=bad)
     if (bad) stop "Missing points in P_lm-file!"        
  endif

  !-----------------------------------------------------------------------
  !                        map to alm
  !-----------------------------------------------------------------------

  if (polarisation) then
     ALLOCATE(alm_TGC(1:3, 0:nlmax, 0:nmmax),stat = status)
     if (status /= 0) call die_alloc(code,"alm_TGC")
  else
     ALLOCATE(alm_TGC(1:1, 0:nlmax, 0:nmmax),stat = status)
     if (status /= 0) call die_alloc(code,"alm_TGC")
  endif

  PRINT *,"      "//code//"> Analyse map " 

  select case (plm_pol+polar*10)
  case(0) 
     call map2alm(nsmax, nlmax, nmmax, map_TQU(:,1), alm_TGC, cos_theta_cut, w8ring_TQU)
  case(1)
     call map2alm(nsmax, nlmax, nmmax, map_TQU(:,1), alm_TGC, cos_theta_cut, w8ring_TQU,plm(:,1))   
  case(10)
     call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC, cos_theta_cut, w8ring_TQU)
  case(11)
     call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC, cos_theta_cut, w8ring_TQU,plm(:,1))
  case(13)
     call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC, cos_theta_cut, w8ring_TQU,plm)
  end select


  !-----------------------------------------------------------------------
  !                  iterated determination of alm
  !-----------------------------------------------------------------------
  if (iter_order /= 0) then

     ALLOCATE(fact(0:iter_order+2),stat = status)
     if (status /= 0) call die_alloc(code,"fact")

     if (polarisation) then
        ALLOCATE(alm_TGC2(1:3, 0:nlmax, 0:nmmax),stat = status)
        if (status /= 0) call die_alloc(code,"alm_TGC2")
     else
        ALLOCATE(alm_TGC2(1:1, 0:nlmax, 0:nmmax),stat = status)
        if (status /= 0) call die_alloc(code,"alm_TGC2")
     endif

     PRINT *,"      "//code//"> Iterated map analysis " 

     fact(0)=1
     do i=1,iter_order+1
        fact(i)=fact(i-1)*i
     enddo
     alm_TGC2=alm_TGC
     alm_TGC=0.
     sign=1
     i=iter_order+1
     do j=1,i
        alm_TGC(1:(polar*2+1),0:nlmax,0:nmmax)=alm_TGC(1:(polar*2+1),0:nlmax,0:nmmax)+&
             & 0.5*sign*fact(i-1)*(2*i-j)/(fact(j)*fact(i-j))*alm_TGC2(1:(polar*2+1),0:nlmax,0:nmmax)
        sign=-sign

        if (j.eq.i) exit

        map_TQU = 0.
        !        synthetize a map from the estimated alm

        select case (plm_pol+polar*10)
        case(0) 
           call alm2map(nsmax,nlmax,nmmax,alm_TGC2,map_TQU(:,1))
        case(1)
           call alm2map(nsmax,nlmax,nmmax,alm_TGC2,map_TQU(:,1),plm(:,1))         
        case(10)
           call alm2map(nsmax,nlmax,nmmax,alm_TGC2,map_TQU)
        case(11)
           call alm2map(nsmax,nlmax,nmmax,alm_TGC2,map_TQU,plm(:,1))
        case(13)
           call alm2map(nsmax,nlmax,nmmax,alm_TGC2,map_TQU,plm)
        end select

        !        analyse the new map
        alm_TGC2=0.     
        select case (plm_pol+polar*10)
        case(0) 
           call map2alm(nsmax, nlmax, nmmax, map_TQU(:,1), alm_TGC2, cos_theta_cut, w8ring_TQU)
        case(1)
           call map2alm(nsmax, nlmax, nmmax, map_TQU(:,1), alm_TGC2, cos_theta_cut, w8ring_TQU,plm(:,1))   
        case(10)
           call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC2, cos_theta_cut, w8ring_TQU)
        case(11)
           call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC2, cos_theta_cut, w8ring_TQU,plm(:,1))
        case(13)
           call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC2, cos_theta_cut, w8ring_TQU,plm)
        end select
     enddo
     DEALLOCATE(alm_TGC2)
  endif

  !-----------------------------------------------------------------------
  !                  power spectrum calculation
  !-----------------------------------------------------------------------

  PRINT *,"      "//code//"> Compute power spectrum "      

  do l = 0, nlmax
     powspec_TT(l) = &
          & ( REAL(alm_TGC(1,l,0))**2 &
          &  + 2.*SUM( alm_TGC(1,l,1:l)*CONJG(alm_TGC(1,l,1:l)) ) &
          & ) / (2.*l + 1.)
  enddo

  if (polarisation) then
     do l = 0, nlmax
        powspec_GG(l) = &
             & ( REAL(alm_TGC(2,l,0))**2 &
             &  + 2.*SUM( alm_TGC(2,l,1:l)*CONJG(alm_TGC(2,l,1:l)) ) &
             & ) / (2.*l + 1.)
        powspec_TG(l) = &
             & ( REAL(alm_TGC(1,l,0)*alm_TGC(2,l,0)) &
             &  + 2.*SUM( alm_TGC(1,l,1:l)*CONJG(alm_TGC(2,l,1:l)) ) &
             & ) / (2.*l + 1.)
        powspec_CC(l) = &
             & ( REAL(alm_TGC(3,l,0))**2 &
             &  + 2.*SUM( alm_TGC(3,l,1:l)*CONJG(alm_TGC(3,l,1:l)) ) &
             & ) / (2.*l + 1.)
     enddo
     if (bcoupling) then
        do l = 0, nlmax
           powspec_TC(l) = &
                & ( REAL(alm_TGC(1,l,0)*alm_TGC(3,l,0)) &
                &  + 2.*SUM( alm_TGC(1,l,1:l)*CONJG(alm_TGC(3,l,1:l)) ) &
                & ) / (2.*l + 1.)
           powspec_GC(l) = &
                & ( REAL(alm_TGC(2,l,0)*alm_TGC(3,l,0)) &
                &  + 2.*SUM( alm_TGC(2,l,1:l)*CONJG(alm_TGC(3,l,1:l)) ) &
                & ) / (2.*l + 1.)
        enddo
     endif
  endif
  !-----------------------------------------------------------------------
  !                  outputs the power spectrum
  !-----------------------------------------------------------------------

  PRINT *,"      "//code//"> outputting Power Spectrum "

  do l=0, nlmax
     fact_norm(l) = 1.0 !  Normalisation=1, Output 'pure' C_ls
  enddo

  if (polarisation) then
     ncl = 4
     if (bcoupling) ncl = 6
     allocate(clout(0:nlmax,1:ncl))
     clout(:,1) = powspec_TT * fact_norm
     clout(:,2) = powspec_GG * fact_norm
     clout(:,3) = powspec_CC * fact_norm
     clout(:,4) = powspec_TG * fact_norm
     if (bcoupling) then
        clout(:,5) = powspec_TC * fact_norm
        clout(:,6) = powspec_GC * fact_norm
     endif
  else
     allocate(clout(0:nlmax,1:1))
     clout(:,1) = powspec_TT * fact_norm
     ncl = 1
  endif


  header = ""
  call add_card(header,"COMMENT","-----------------------------------------------")
  call add_card(header,"COMMENT","     Map Analysis Specific Keywords      ")
  call add_card(header,"COMMENT","-----------------------------------------------")
  call add_card(header,"EXTNAME","'ANALYSED POWER SPECTRUM'")
  call add_card(header,"COMMENT"," POWER SPECTRUM : C(l) ")
  call add_card(header)
  call add_card(header,"CREATOR",code,        "Software creating the FITS file")
  call add_card(header,"VERSION",version,     "Version of the simulation software")
  call add_card(header,"CUT-SKY",theta_cut_deg," [deg] Symmetric Cut Sky")
  call add_card(header,"POLAR",polarisation," Polarisation included (True/False)")
  call add_card(header,"BCROSS",bcoupling,"Magnetic cross terms included (True/False)")
  call add_card(header)
  call add_card(header,"HISTORY","Input Map = "//TRIM(infile))
  call add_card(header)
  call add_card(header,"TTYPE1", "TEMPERATURE","Temperature C(l)")
  call add_card(header,"TUNIT1", units_pow(1),"square of map units")
  call add_card(header)
  !
  if (polarisation) then
     call add_card(header,"TTYPE2", "GRADIENT","Gradient (=ELECTRIC) polarisation C(l)")
     call add_card(header,"TUNIT2", units_pow(1),"power spectrum units")
     call add_card(header)
     !
     call add_card(header,"TTYPE3", "CURL","Curl (=MAGNETIC) polarisation C(l)")
     call add_card(header,"TUNIT3", units_pow(1),"power spectrum units")
     call add_card(header)
     !
     call add_card(header,"TTYPE4", "G-T","Gradient-Temperature (=CROSS) terms")
     call add_card(header,"TUNIT4", units_pow(1),"power spectrum units")
     call add_card(header)
     !
     if (bcoupling) then
        call add_card(header,"TTYPE5", "C-T","Curl-Temperature terms")
        call add_card(header,"TUNIT5", units_pow(1),"power spectrum units")
        call add_card(header)
        !
        call add_card(header,"TTYPE6", "C-G","Curl-Gradient terms")
        call add_card(header,"TUNIT6", units_pow(1),"power spectrum units")
        call add_card(header)
     endif
     !
     call add_card(header,"COMMENT","The polarisation power spectra have the same definition as in CMBFAST")
     call add_card(header)
  endif

  nlheader = SIZE(header)
  call write_asctab (clout, nlmax, ncl, header, nlheader, outfile)

  !-----------------------------------------------------------------------
  !                  outputs the alm
  !-----------------------------------------------------------------------
  if (outfile_alms/="") then

     PRINT *,"      "//code//"> outputting alms "

     ! write each component (T,G,C) in a different extension of the same file
     do i=1,1+polar*2
        header = ""
        call add_card(header,"COMMENT","-----------------------------------------------")
        call add_card(header,"COMMENT","     Map Analysis Specific Keywords      ")
        call add_card(header,"COMMENT","-----------------------------------------------")
        if (i == 1) then
           call add_card(header,"EXTNAME","""ANALYSED a_lms (TEMPERATURE)""")
        elseif (i == 2) then
           call add_card(header,"EXTNAME","'ANALYSED a_lms (GRAD / ELECTRIC component)'")
        elseif (i == 3) then
           call add_card(header,"EXTNAME","'ANALYSED a_lms (CURL / MAGNETIC component)'")
        endif
        call add_card(header)
        call add_card(header)
        call add_card(header,"CREATOR",code,        "Software creating the FITS file")
        call add_card(header,"VERSION",version,     "Version of the simulation software")
        call add_card(header,"CUT-SKY",theta_cut_deg," [deg] Symmetric Cut Sky")
        if (i == 1) then
           call add_card(header,"COMMENT","Temperature a_lms")
        else
           call add_card(header,"COMMENT","Polarisation a_lms")
        endif
        call add_card(header,"COMMENT"," The real and imaginary part of the a_lm with m>=0")
        call add_card(header)
        call add_card(header,"COMMENT","Input Map = "//TRIM(infile))
        call add_card(header)
        call add_card(header,"TTYPE1", "INDEX"," i = l^2 + l + m + 1")
        call add_card(header,"TUNIT1", "   "," index")
        call add_card(header)
        call add_card(header,"TTYPE2", "REAL"," REAL a_lm")
        call add_card(header,"TUNIT2", units_map(1)," alm units")
        call add_card(header)
        !
        call add_card(header,"TTYPE3", "IMAG"," IMAGINARY a_lm")
        call add_card(header,"TUNIT3", units_map(1)," alm units")
        call add_card(header)
        !
        call add_card(header)
        call add_card(header)
        call add_card(header)
        !
        nlheader = SIZE(header)
        call dump_alms (outfile_alms,alm_TGC(i,0:nlmax,0:nlmax),nlmax,header,nlheader,i-1)
     enddo
  endif

  !-----------------------------------------------------------------------
  !                      deallocate memory for arrays
  !-----------------------------------------------------------------------

  DEALLOCATE( alm_TGC )
  DEALLOCATE( map_TQU )
  DEALLOCATE( powspec_TT )
  DEALLOCATE(w8ring_TQU)
  DEALLOCATE(dw8)

  if (polarisation) then
     DEALLOCATE( powspec_GG, powspec_CC, powspec_TG )
     if (bcoupling) then
        DEALLOCATE( powspec_TC, powspec_GC)
     endif
  endif

  !-----------------------------------------------------------------------
  !                      report card
  !-----------------------------------------------------------------------
  call date_and_time(values = values_time(:,2))

  values_time(:,1) = values_time(:,2) - values_time(:,1) ! difference in y, m, d, x, h, m, s, ms
  clock_time =    (  ( values_time(3,1)*24 &
       &             + values_time(5,1)   )*60.  &
       &          + values_time(6,1)      )*60. &
       &       + values_time(7,1) &
       &       + values_time(8,1)/1000.

  WRITE(*,9000) " "
  WRITE(*,9000) " Report Card for "//code//" analysis run"
  WRITE(*,9000) "----------------------------------------"
  WRITE(*,9000) " "
  if (.not. polarisation) then
     WRITE(*,9000) "      Temperature alone"
  else
     WRITE(*,9000) "    Temperature + Polarisation"
  endif
  WRITE(*,9000) " "
  WRITE(*,9000) " Input map            : "//TRIM(infile)
  WRITE(*,9010) " Multipole range      : 0 < l <= ", nlmax
  WRITE(*,9010) " Number of pixels     : ", npixtot
  WRITE(*,9020) " Pixel size in arcmin : ", pix_size_arcmin
  WRITE(*,9020) " Symmetric cut in deg : ", theta_cut_deg
  WRITE(*,9000) " Output power spectrum: "//TRIM(outfile)
  WRITE(*,9020) " Total clock time [s] : ", clock_time

  !-----------------------------------------------------------------------
  !                       end of routine
  !-----------------------------------------------------------------------

  WRITE(*,9000) " "
  WRITE(*,9000) " "//code//"> normal completion"

9000 format(a)
9010 format(a,i16)
9020 format(a,g20.5)


  STOP
END PROGRAM anafast

