!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
! Fast Smoothing of a full sky map in HEALPIX pixelisation.
! Written and developed by E. Hivon (efh@ipac.caltech.edu) and K. Gorski 
! (kgorski@eso.org) based on HEALPIX pixelisation developed 
! by K. Gorski (kgorski@eso.org)
!
! The permission to use and copy this software and its documentation, 
! without fee or royalty is limited to non-commercial purposes related to 
! PLANCK Surveyor project and provided that you agree to comply with
! the following copyright notice and statements,
! and that the same appear on ALL copies of the software and documentation.
!
! An appropriate acknowledgement has to be included in any
! publications based on work where the package has been used.
!
! Copyright 1997 by Eric Hivon and Krzysztof M. Gorski.
!  All rights reserved.
!
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!=======================================================================
PROGRAM smoothing
  !=======================================================================
  !  EXTERNAL LIBRARY:
  !     this code uses the FITSIO library that can be found at
  !     http://heasarc.gsfc.nasa.gov/docs/software/fitsio/fitsio.html
  !
  !  RELATED LITTERATURE:
  !     about HEALPIX   : see Gorski, 1997, in preparation
  !     about this code : see Hivon & Gorski, 1997, in preparation
  !
  !  HISTORY:
  !     April-October 1997, Eric Hivon, TAC (f77)
  !               Jan 1998 : translation in Fortran 90
  !               Dec 2001 : version 1.2, EH, IPAC
  !                 implementation of non-interactive input
  !                 extension to polarised maps
  !               Sep 2002 : implement new parser
  !               Feb 2003 : output correct ORDERING keyword
  !
  !  FEEDBACK:
  !     for any questions : efh@ipac.caltech.edu
  !
  !
  !=======================================================================
  !     version 1.2.0
  !=======================================================================
  USE healpix_types 
  USE alm_tools, ONLY : map2alm, alm2map, alter_alm
  USE fitstools, ONLY : getsize_fits, input_map, write_bintab, read_par, read_dbintab
  USE head_fits, ONLY : add_card, get_card
  USE utilities, ONLY : die_alloc
  USE pix_tools, ONLY : convert_ring2nest, convert_nest2ring, nside2npix, npix2nside
  USE extension, ONLY : nArguments, getArgument, getEnvironment
  USE paramfile_io, ONLY : paramfile_handle, parse_init, parse_int, &
         parse_string, parse_real, concatnl, scan_directories
  IMPLICIT none
  INTEGER(KIND=I4B) :: nlmax
  INTEGER(KIND=I4B) :: nsmax
  INTEGER(KIND=I4B) :: nmmax  
  
  COMPLEX(KIND=SPC), DIMENSION(:,:,:), ALLOCATABLE :: alm_TGC
  COMPLEX(KIND=SPC), DIMENSION(:,:,:), ALLOCATABLE :: alm_TGC2
  REAL(KIND=SP),     DIMENSION(:,:),   ALLOCATABLE :: map_TQU
  REAL(KIND=DP),     DIMENSION(:,:),   ALLOCATABLE :: w8ring_TQU
  REAL(KIND=SP),     DIMENSION(:,:),   ALLOCATABLE :: dw8
  REAL(KIND=DP),     DIMENSION(:,:),   ALLOCATABLE :: plm
  INTEGER(KIND=I4B), DIMENSION(:),     ALLOCATABLE :: fact
  INTEGER(KIND=I4B)  status
  
  INTEGER(KIND=I4B) :: mlpol, polar_fits
  INTEGER(KIND=I4B) :: n_plm, plm_nside, plm_lmax, plm_pol
  INTEGER(KIND=I4B) :: npixtot, nmaps
  INTEGER(KIND=I4B) :: i, j, sign
  INTEGER(KIND=I4B) :: nlheader
  INTEGER(KIND=I4B) :: ordering, order_type, type
  INTEGER(KIND=I4B) :: iter_order, polar
  INTEGER(KIND=I4B) :: simul_type
  INTEGER(KIND=I4B) :: n_rings, n_rings_read, won
  INTEGER(KIND=I4B), DIMENSION(8,2) :: values_time
  
  REAL(KIND=SP) :: fwhm_arcmin, fwhm_deg
  REAL(KIND=SP) :: pix_size_arcmin
  REAL(KIND=SP) :: clock_time
  REAL(KIND=DP) :: cos_theta_cut
  REAL(KIND=DP) :: nullval

  CHARACTER(LEN=filenamelen)          :: infile
  CHARACTER(LEN=filenamelen)          :: outfile
  CHARACTER(LEN=filenamelen)          :: parafile = ''
  CHARACTER(LEN=filenamelen)          :: infile_plm
  CHARACTER(LEN=filenamelen)          :: infile_w8
  CHARACTER(LEN=filenamelen)          :: w8name
  CHARACTER(LEN=filenamelen)          :: def_dir, def_file
  CHARACTER(LEN=filenamelen)          :: usr_dir, usr_file
  CHARACTER(LEN=filenamelen)          :: final_file
  CHARACTER(LEN=filenamelen)          :: healpixdir
  CHARACTER(LEN=filenamelen)          :: beam_file
  character(len=filenamelen)          :: description
  character(len=100)                  :: chline,chline1
  CHARACTER(LEN=80), DIMENSION(1:180) :: header, header_in
  CHARACTER(LEN=5)                    :: sstr
  LOGICAL(KIND=LGT) :: bad, ok, polarisation
  CHARACTER(LEN=*), PARAMETER :: code = 'SMOOTHING'
  CHARACTER(LEN=*), PARAMETER :: version = '1.2.0'
  integer(kind=i4b), parameter :: nm_max = 5
  CHARACTER(len=80), dimension(1:nm_max) :: tunit, com_tunit
  INTEGER(kind=I4B), dimension(1:nm_max) :: ct_tunit

  type(paramfile_handle) :: handle

  !-----------------------------------------------------------------------
  !                    get input parameters and create arrays
  !-----------------------------------------------------------------------

  call date_and_time(values = values_time(:,1))
  !     --- read parameters interactively if no command-line arguments
  !     --- are given, otherwise interpret first command-line argument
  !     --- as parameter file name and read parameters from it:
  if (nArguments() == 0) then
     parafile=''
  else
     if (nArguments() /= 1) then
        print "(' Usage:  smoothing [parameter file name]')"
        stop 1
     end if
     call getArgument(1,parafile)
  end if

  !     --- announces program, default banner ***
  PRINT *, " "
  PRINT *,'			   '//code//' '//version
  write(*,'(a)') '         *** Smoothing of a Temperature (+Polarisation) map ***    '
  PRINT *, ''
  handle = parse_init(parafile)
  
  !     --- choose temp. only or  temp. + pol. ---
  description = concatnl( &
       & " Do you want to analyse", &
       & " 1) a Temperature only map", &
       & " 2) Temperature + polarisation maps")
  simul_type = parse_int(handle, 'simul_type', default=1, vmin=1, vmax=2, descr=description)
  if (simul_type .eq. 1) polarisation = .false.
  if (simul_type .eq. 2) polarisation = .true.

  !     --- gets the file name for the map ---
  chline = "''"
  call getEnvironment("HEALPIX",healpixdir)
  if (trim(healpixdir)/='') chline = trim(healpixdir)//'/test/map.fits'
  description = concatnl( &
       & "", &
       & " Enter input file name (Map FITS file): ")
  infile = parse_string(handle, 'infile',default=chline, descr=description, filestatus='old')

  !     --- finds out the pixel number of the map and its ordering ---
  npixtot = getsize_fits(infile, nmaps = nmaps, ordering=ordering, nside=nsmax, mlpol=mlpol, type = type, polarisation = polar_fits)
  if (nsmax<=0) then 
     print*,'Keyword NSIDE not found in FITS header!'
     stop 1
  endif
  if (type == 3) npixtot = nside2npix(nsmax) ! cut sky input data set
  if (nsmax/=npix2nside(npixtot)) then
     print*,"FITS header keyword NSIDE does not correspond"
     print*,"to the size of the map!"
     stop 1
  endif
  
  if (polarisation .and. (nmaps >=3) .and. polar_fits == -1) then
     print*,"The input fits file MAY NOT contain polarisation data."
     print*,"Proceed at your own risk"
  endif

  if (polarisation .and. (nmaps<3 .or. polar_fits ==0)) then
     print *,"The file does NOT contain polarisation maps"
     print *,"only the temperature field will be smoothed"
     polarisation = .false.
  endif

  !     --- check ordering scheme ---
  
  if ((ordering/=1).and.(ordering/=2)) then
     PRINT*,'The ordering scheme of the map must be RING or NESTED.'
     PRINT*,'No ordering specification is given in the FITS-header!'
     stop 1
  endif

  order_type = ordering
  
  !     --- gets the L range for the analysis ---
  WRITE(chline1,"(a,i5)") " The map has Nside = ",nsmax
  if (mlpol<=0) WRITE(chline,"(a,i5,a)") "We recommend: (0 <= l <= l_max <= ",3*nsmax-1,")"
  if (mlpol> 0) WRITE(chline,"(a,i5,a)") "We recommend: (0 <= l <= l_max <= ",mlpol,")"
  description = concatnl(&
       & "", &
       & chline1, &
       & " Enter the maximum l range (l_max) for the simulation. ", &
       & chline )
  nlmax = parse_int(handle, 'nlmax', default=2*nsmax, descr=description)

  !     --- gets the order of iteration ---
  description = concatnl(&
       & "", &
       & " Do you want : ", &
       & " 0) a standard analysis", &
       & " 1,2,3,4....) an iterative analysis", &
       & " (enter order of iteration, 3rd order is usually optimal)")
  iter_order=parse_int(handle, 'iter_order', vmin=0, default=0, descr=description)
  if (.not. handle%interactive) then
     select case (iter_order)
     case (0)
        PRINT*," Standard analysis"      
     case default
        PRINT*," Iterative analysis"
     end select
  end if

  ! ------------------- precomputed plms  ---------------------
40 continue
  n_plm   = 0
  plm_pol = 0
  description = concatnl(&
       & "", &
       & " Enter name of file with precomputed p_lms :", &
       & "(eg, plm.fits) (if '' is typed, no file is read)" )
  infile_plm = parse_string(handle, 'plmfile', default="''", descr=description, filestatus='old')

  if (trim(infile_plm) /= "") then
     ! check that the plm file matches the current simulation (for nside and lmax)
     call read_par(infile_plm, plm_nside, plm_lmax, plm_pol)

     if ((plm_pol.ne.1).and.(.not.polarisation)) plm_pol=1

     if ((plm_nside.ne.nsmax).or.(plm_lmax.ne.nlmax)) then
        PRINT *," "//code//"> Plm file does not match nside or l_max ! "
        if (handle%interactive) goto 40
        stop 1
     endif
     n_plm = (nlmax+1)*(nlmax+2)*nsmax
     PRINT *," "
  endif

  !     --- inputs the FWHM or beam file ---
  beam_file = ''
  description = concatnl(&
       & "", &
       & " Enter the Gaussian beam FWHM in arcmin >= 0 (eg: 420.) : ")
  fwhm_arcmin = parse_real(handle, 'fwhm_arcmin', default=420., vmin=0.,descr=description)

  description = concatnl(&
       & "", &
       & " Enter an external file name containing ", &
       & " a symmetric beam Legendre transform (eg: mybeam.fits)", &
       & " NB: if set to an existing file, it will override the FWHM chosen above", &
       & "     if set to '', the gaussian FWHM will be used.")
  beam_file = parse_string(handle, 'beam_file', default="''", &
       &                 descr=description, filestatus='old')
  if (beam_file /= '') then
     fwhm_arcmin = 0.
     print*,'fwhm_arcmin is now : 0.'
  endif

  !     --- gets the output sky map filename ---
  description = concatnl(&
       & "", &
       & " Enter Output map file name (eg, test_smooth.fits) :", &
       & "  (or !test_smooth.fits to overwrite an existing file)" )
  outfile = parse_string(handle, "outfile", &
       default="!test_smooth.fits", descr=description, filestatus="new")

  !-----------------------------------------------------------------------
  !                      ask for weights
  !-----------------------------------------------------------------------
  description = concatnl(&
       & "", &
       & " Do you want to use ring weights in the analysis (1=yes,2=no) ?")
  won = parse_int(handle, 'won', vmin=1, vmax=2, default=2, descr=description)

  infile_w8=""

  if (won.eq.1) then

     ! default weight file name
     write (sstr,"(I5.5)") nsmax
     w8name="weight_ring_n"//trim(sstr)//".fits"
     
     def_file = trim(w8name)
     def_dir  = concatnl("","../data","./data","..")
     call getEnvironment("HEALPIX",healpixdir)
     if (trim(healpixdir) .ne. "") then
        def_dir = concatnl(&
             & def_dir, &
             & healpixdir, &
             & trim(healpixdir)//"data", &
             & trim(healpixdir)//"/data", &
             & trim(healpixdir)//char(92)//"data") !backslash
     endif

22   continue
     final_file = ''
     ok = .false.
     ! if interactive, try default name in default directories first
     if (handle%interactive) ok = scan_directories(def_dir, def_file, final_file)
     if (.not. ok) then 
        ! not found, ask the user
        description = concatnl("",&
             &        " Could not find weight file", &
             &        " Enter the directory where this file can be found:")
        usr_dir = parse_string(handle,'w8filedir',default='',descr=description)
        if (trim(usr_dir) == '') usr_dir = trim(def_dir)
        description = concatnl("",&
             &        " Enter the name of the weight file:")
        usr_file = parse_string(handle,'w8file',default=def_file,descr=description)
        ! look for new name in user provided or default directories
        ok   = scan_directories(usr_dir, usr_file, final_file)
        ! if still fails, crash or ask again if interactive
        if (.not. ok) then
           print*,' File not found'
           if (handle%interactive) goto 22
           stop 1
        endif
     endif
     infile_w8 = final_file
  endif
  
  PRINT *," "

  !-----------------------------------------------------------------------
  nmmax = nlmax
  pix_size_arcmin = 360.*60./SQRT(12.*pi)/float(nsmax)
  
  !-----------------------------------------------------------------------
  !              allocate space for arrays
  !-----------------------------------------------------------------------
  
  polar=0
  if (polarisation) then
     polar=1
     ALLOCATE(map_TQU(0:12*nsmax**2-1,1:3),stat = status)
     if (status /= 0) call die_alloc(code,"map_TQU")

     ALLOCATE(w8ring_TQU(1:2*nsmax,1:3),stat = status)
     if (status /= 0) call die_alloc(code,"w8ring_TQU")

     ALLOCATE(dw8(1:2*nsmax,1:3),stat = status)
     if (status /= 0) call die_alloc(code,"dw8")
  else
     ALLOCATE(map_TQU(0:12*nsmax**2-1,1:1),stat = status)
     if (status /= 0) call die_alloc(code,"map_TQU")

     ALLOCATE(w8ring_TQU(1:2*nsmax,1:1),stat = status)
     if (status /= 0) call die_alloc(code,"w8ring_TQU")

     ALLOCATE(dw8(1:2*nsmax,1:1),stat = status)
     if (status /= 0) call die_alloc(code,"dw8")
  endif
  
  !-----------------------------------------------------------------------
  !                      reads the map
  !-----------------------------------------------------------------------
  PRINT *,"      "//code//"> Inputting original map " 

  if (polarisation) then
     call input_map(infile, map_TQU(0:,1:3), npixtot, 3, fmissval=0.0_sp, header=header_in)
     call get_card(header_in, 'TUNIT1', tunit(1), com_tunit(1), count=ct_tunit(1))
     call get_card(header_in, 'TUNIT2', tunit(2), com_tunit(2), count=ct_tunit(2))
     call get_card(header_in, 'TUNIT3', tunit(3), com_tunit(3), count=ct_tunit(3))
  else
     call input_map(infile, map_TQU(0:,1:1), npixtot, 1, fmissval=0.0_sp, header=header_in)
     call get_card(header_in, 'TUNIT1', tunit(1), com_tunit(1), count=ct_tunit(1))
  endif
!   call get_card(header_in,'POLAR',polar1, comment)
!   print*,polar1,trim(comment)
!   print*,"=============================================="
  
  !-----------------------------------------------------------------------
  !                      input weights
  !-----------------------------------------------------------------------
  
  dw8=0.

  if (trim(infile_w8)/="") then


     n_rings = 2 * nsmax
     n_rings_read = getsize_fits(infile_w8)
     
     if (n_rings_read /= n_rings) then
        PRINT *," "
        print *,"wrong file"
        stop
     endif
     
     if (polarisation) then
        PRINT *,"      "//code//"> Inputting weights "  
        call input_map(infile_w8, dw8, n_rings, 3, fmissval=0.0_sp)
     else
        PRINT *,"      "//code//"> Inputting weights "  
        call input_map(infile_w8, dw8, n_rings, 1, fmissval=0.0_sp)
     endif
     
  endif

  w8ring_TQU = 1.d0 + dw8
  
  !-----------------------------------------------------------------------
  !                 convert to RING if necessary
  !-----------------------------------------------------------------------
  if (order_type == 2) then
     PRINT *,"      "//code//"> Convert Nest -> Ring "      
     call convert_nest2ring (nsmax, map_TQU(:,1))
     if (polarisation) then
        call convert_nest2ring (nsmax, map_TQU(:,2))
        call convert_nest2ring (nsmax, map_TQU(:,3))
     endif
  endif
  
  !-----------------------------------------------------------------------
  !                    precomputed plms
  !-----------------------------------------------------------------------

  if (n_plm/=0) then
     if (plm_pol.eq.1) then
        PRINT*,"      "//code//"> Reading precomputed scalar P_lms "
     else
        PRINT*,"      "//code//"> Reading precomputed tensor P_lms "
     endif
     call read_dbintab(infile_plm,plm,n_plm,plm_pol,nullval,anynull=bad)
     if (bad) stop "Missing points in P_lm-file!"        
  endif

  !-----------------------------------------------------------------------
  !                        map to alm
  !-----------------------------------------------------------------------

  if (polarisation) then
     ALLOCATE(alm_TGC(1:3, 0:nlmax, 0:nmmax),stat = status)
     if (status /= 0) call die_alloc(code,"alm_TGC")
  else
     ALLOCATE(alm_TGC(1:1, 0:nlmax, 0:nmmax),stat = status)
     if (status /= 0) call die_alloc(code,"alm_TGC")
  endif

  PRINT *,"      "//code//"> Analyse map " 

  cos_theta_cut= -1000.0_dp  ! default value

  select case (plm_pol+polar*10)
  case(0) 
     call map2alm(nsmax, nlmax, nmmax, map_TQU(:,1), alm_TGC, cos_theta_cut, w8ring_TQU)
  case(1)
     call map2alm(nsmax, nlmax, nmmax, map_TQU(:,1), alm_TGC, cos_theta_cut, w8ring_TQU,plm(:,1))   
  case(10)
     call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC, cos_theta_cut, w8ring_TQU)
  case(11)
     call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC, cos_theta_cut, w8ring_TQU,plm(:,1))
  case(13)
     call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC, cos_theta_cut, w8ring_TQU,plm)
  end select


  !-----------------------------------------------------------------------
  !                  iterated determination of alm
  !-----------------------------------------------------------------------
  if (iter_order /= 0) then

     ALLOCATE(fact(0:iter_order+2),stat = status)
     if (status /= 0) call die_alloc(code,"fact")

     if (polarisation) then
        ALLOCATE(alm_TGC2(1:3, 0:nlmax, 0:nmmax),stat = status)
        if (status /= 0) call die_alloc(code,"alm_TGC2")
     else
        ALLOCATE(alm_TGC2(1:1, 0:nlmax, 0:nmmax),stat = status)
        if (status /= 0) call die_alloc(code,"alm_TGC2")
     endif

     PRINT *,"      "//code//"> Iterated map analysis " 

     fact(0)=1
     do i=1,iter_order+1
        fact(i)=fact(i-1)*i
     enddo
     alm_TGC2=alm_TGC
     alm_TGC=0.
     sign=1
     i=iter_order+1
     do j=1,i
        alm_TGC(1:(polar*2+1),0:nlmax,0:nmmax)=alm_TGC(1:(polar*2+1),0:nlmax,0:nmmax)+&
             & 0.5*sign*fact(i-1)*(2*i-j)/(fact(j)*fact(i-j))*alm_TGC2(1:(polar*2+1),0:nlmax,0:nmmax)
        sign=-sign

        if (j.eq.i) exit

        map_TQU = 0.
        !        synthetize a map from the estimated alm

        select case (plm_pol+polar*10)
        case(0) 
           call alm2map(nsmax,nlmax,nmmax,alm_TGC2,map_TQU(:,1))
        case(1)
           call alm2map(nsmax,nlmax,nmmax,alm_TGC2,map_TQU(:,1),plm(:,1))         
        case(10)
           call alm2map(nsmax,nlmax,nmmax,alm_TGC2,map_TQU)
        case(11)
           call alm2map(nsmax,nlmax,nmmax,alm_TGC2,map_TQU,plm(:,1))
        case(13)
           call alm2map(nsmax,nlmax,nmmax,alm_TGC2,map_TQU,plm)
        end select

        !        analyse the new map
        alm_TGC2=0.     
        select case (plm_pol+polar*10)
        case(0) 
           call map2alm(nsmax, nlmax, nmmax, map_TQU(:,1), alm_TGC2, cos_theta_cut, w8ring_TQU)
        case(1)
           call map2alm(nsmax, nlmax, nmmax, map_TQU(:,1), alm_TGC2, cos_theta_cut, w8ring_TQU,plm(:,1))   
        case(10)
           call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC2, cos_theta_cut, w8ring_TQU)
        case(11)
           call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC2, cos_theta_cut, w8ring_TQU,plm(:,1))
        case(13)
           call map2alm(nsmax, nlmax, nmmax, map_TQU, alm_TGC2, cos_theta_cut, w8ring_TQU,plm)
        end select
     enddo
     DEALLOCATE(alm_TGC2)
  endif

  !-----------------------------------------------------------------------
  !                    smoothing of the alm
  !-----------------------------------------------------------------------
  PRINT *,'      '//code//'> Alter alm '  
  call alter_alm(nsmax, nlmax, nmmax, fwhm_arcmin, alm_TGC(1:1+2*polar,0:nlmax,0:nmmax), beam_file)
      
  !-----------------------------------------------------------------------
  !                           a_lm to map
  !-----------------------------------------------------------------------

  PRINT *,'      '//code//'> Generating sky map(s) '

  select case (plm_pol+polar*10)
  case(0) 
     call alm2map(nsmax,nlmax,nmmax,alm_TGC,map_TQU(:,1))
  case(1)
     call alm2map(nsmax,nlmax,nmmax,alm_TGC,map_TQU(:,1),plm(:,1))         
  case(10)
     call alm2map(nsmax,nlmax,nmmax,alm_TGC,map_TQU)
  case(11)
     call alm2map(nsmax,nlmax,nmmax,alm_TGC,map_TQU,plm(:,1))
  case(13)
     call alm2map(nsmax,nlmax,nmmax,alm_TGC,map_TQU,plm)
  end select
  
  !-----------------------------------------------------------------------
  !                 convert back to NEST if necessary
  !-----------------------------------------------------------------------
  if (order_type == 2) then
     PRINT *,"      "//code//"> Convert Ring -> Nest "      
     call convert_ring2nest (nsmax, map_TQU(:,1))
     if (polarisation) then
        call convert_ring2nest (nsmax, map_TQU(:,2))
        call convert_ring2nest (nsmax, map_TQU(:,3))
     endif
  endif
  
  !-----------------------------------------------------------------------
  !                        generates header
  !-----------------------------------------------------------------------
  nlheader = SIZE(header)
  header = ' '
  fwhm_deg = fwhm_arcmin/60.
  PRINT *,'      '//code//'> Writing smoothed map to FITS file '
  
  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'COMMENT','     Sky Map Pixelisation Specific Keywords    ')
  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'PIXTYPE','HEALPIX','HEALPIX Pixelisation')
  if (order_type == 1) then
     call add_card(header,'ORDERING','RING',  'Pixel ordering scheme, either RING or NESTED')
  else
     call add_card(header,'ORDERING','NESTED',  'Pixel ordering scheme, either RING or NESTED')
  endif
  call add_card(header,'NSIDE'   ,nsmax,   'Resolution parameter for HEALPIX')
  call add_card(header,'FIRSTPIX',0,'First pixel # (0 based)')
  call add_card(header,'LASTPIX',npixtot-1,'Last pixel # (0 based)')
  call add_card(header) ! blank line
  
  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'COMMENT','     Planck Simulation Specific Keywords      ')
  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'EXTNAME','''SMOOTHED DATA''')
  call add_card(header,'CREATOR',code,        'Software creating the FITS file')
  call add_card(header,'VERSION',version,     'Version of the simulation software')
  call add_card(header,'MAX-LPOL',nlmax      ,'Maximum multipole l used in map smoothing')
  call add_card(header,'FWHM'    ,fwhm_deg   ,'[deg] FWHM of smoothing')
  call add_card(header,"POLAR",polarisation," Polarisation included (True/False)")
  call add_card(header)
  call add_card(header,'COMMENT','************************* Input data ************************* ')
  call add_card(header,'COMMENT','Input Map in '//TRIM(infile))
  call add_card(header,'COMMENT','*************************************************************** ')
  call add_card(header) ! blank line

  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'COMMENT','     Data Description Specific Keywords       ')
  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'INDXSCHM','IMPLICIT',' Indexing : IMPLICIT or EXPLICIT')
  call add_card(header,'GRAIN', 0, ' Grain of pixel indexing')
  call add_card(header,'COMMENT','GRAIN=0 : no indexing of pixel data                         (IMPLICIT)')
  call add_card(header,'COMMENT','GRAIN=1 : 1 pixel index -> 1 pixel data                     (EXPLICIT)')
  call add_card(header,'COMMENT','GRAIN>1 : 1 pixel index -> data of GRAIN consecutive pixels (EXPLICIT)')
  call add_card(header) ! blank line

  call add_card(header,'TTYPE1', 'TEMPERATURE','Temperature map')
  call add_card(header,'TUNIT1', tunit(1),com_tunit(1))
  call add_card(header)

  if (polarisation) then
     call add_card(header,'TTYPE2', 'Q-POLARISATION','Q Polarisation map')
     call add_card(header,'TUNIT2', tunit(2),com_tunit(2))
     call add_card(header)

     call add_card(header,'TTYPE3', 'U-POLARISATION','U Polarisation map')
     call add_card(header,'TUNIT3', tunit(3),com_tunit(3))
     call add_card(header)
  endif

  !-----------------------------------------------------------------------
  !                      write the map to FITS file
  !-----------------------------------------------------------------------
  if (polarisation) then
     !call output_map(map_TQU(0:npixtot-1,1:3), header, outfile) !not on Sun
     call write_bintab(map_TQU, npixtot, 3, header, nlheader, outfile)
  else
     !call output_map(map_TQU(0:npixtot-1,1:1), header, outfile) !not on Sun
     call write_bintab(map_TQU, npixtot, 1, header, nlheader, outfile)
  endif
  
  !-----------------------------------------------------------------------
  !                      deallocate memory for arrays
  !-----------------------------------------------------------------------
  DEALLOCATE(map_TQU)
  DEALLOCATE(w8ring_TQU)
  DEALLOCATE(dw8)
      
  !-----------------------------------------------------------------------
  !                      output and report card
  !-----------------------------------------------------------------------
  call date_and_time(values = values_time(:,2))

  values_time(:,1) = values_time(:,2) - values_time(:,1) ! difference in y, m, d, x, h, m, s, ms
  clock_time =    (  ( values_time(3,1)*24 &
       &             + values_time(5,1)   )*60.  &
       &          + values_time(6,1)      )*60. &
       &       + values_time(7,1) &
       &       + values_time(8,1)/1000.

  PRINT *,' '
  PRINT *,' Report Card for '//code//' simulation run'
  PRINT *,' -----------------------------'
  PRINT *,''
  if (.not. polarisation) then
     PRINT *,'      Temperature alone'
  else
     PRINT *,'    Temperature + Polarisation'
  endif
  PRINT *,''
  
  PRINT *,' Input map            : '//TRIM(infile)
  PRINT *,' Number of pixels     : ', npixtot
  PRINT *,' Pixel size in arcmin : ', pix_size_arcmin
  PRINT *,' Multipole range      : 0 < l <= ', nlmax
  PRINT *,' Gauss. FWHM [arcmin] : ', fwhm_arcmin
  PRINT *,' Output map           : '//TRIM(outfile)
  PRINT *,' Total clock time [s] : ', clock_time
  !-----------------------------------------------------------------------
  !                       end of routine
  !-----------------------------------------------------------------------
  PRINT *,''
  PRINT *,' '//code//'> normal completion'
  
  
  STOP
END PROGRAM smoothing
