!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
! Simulation of full sky map of CMB anisotropies from their power spectrum
! in HEALPIX pixelisation.
! Written and developed by E. Hivon (efh@ipac.caltech.edu) and K. Gorski 
! (kgorski@eso.org) based on HEALPIX pixelisation developed 
! by K. Gorski (kgorski@eso.org)
!
! The permission to use and copy this software and its documentation, 
! without fee or royalty is limited to non-commercial purposes related to 
! PLANCK Surveyor project and provided that you agree to comply with
! the following copyright notice and statements,
! and that the same appear on ALL copies of the software and documentation.
!
! An appropriate acknowledgement has to be included in any
! publications based on work where the package has been used.
!
! Copyright 1997 by Eric Hivon and Krzysztof M. Gorski.
!  All rights reserved.
!
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!=======================================================================
PROGRAM synfast
  !=======================================================================
  !  EXTERNAL LIBRARY:
  !     this code uses the CFITSIO library that can be found at
  !     http://heasarc.gsfc.nasa.gov/docs/software/fitsio/
  !
  !  RELATED LITTERATURE:
  !     about HEALPIX   : see Gorski, 1997, in preparation
  !     about this code : see Hivon & Gorski, 1997, in preparation
  !
  !  HISTORY:
  !     April-October 1997, Eric Hivon, TAC
  !               Jan 1998 : translation in Fortran 90
  !               Aug 1998 : version 1.0.0
  !               Apr 1999 : constr.real, precomp.plms included, FKH
  !               Nov 2000 : v 1.2, EH
  !               Sep 2002 : implement new parser
  !
  !  FEEDBACK:
  !     for any questions : efh@ipac.caltech.edu
  !
  !=======================================================================
  !     version 1.2.0
  !=======================================================================
  USE healpix_types 
  USE alm_tools, ONLY : create_alm, alm2map
  USE pix_tools, ONLY : nside2npix
!  USE fitstools, ONLY : output_map, number_of_alms, read_par, read_conbintab, read_dbintab
  USE fitstools, ONLY : write_bintab, number_of_alms, read_par, read_conbintab, read_dbintab
  USE head_fits, ONLY : add_card, merge_headers, del_card
  USE ran_tools, ONLY : randgauss_boxmuller
  USE utilities, ONLY : die_alloc
  USE extension, ONLY : getEnvironment, getArgument, nArguments
  USE paramfile_io, ONLY : paramfile_handle, parse_init, parse_int, &
         parse_string, parse_real, concatnl, scan_directories
  IMPLICIT none
  INTEGER(I4B) :: nlmax
  INTEGER(I4B) :: nsmax
  INTEGER(I4B) :: nmmax

  COMPLEX(SPC), DIMENSION(:,:,:), ALLOCATABLE :: alm_TGC
  REAL(SP),     DIMENSION(:,:),   ALLOCATABLE :: alms_in
  REAL(SP),     DIMENSION(:,:),   ALLOCATABLE :: map_TQU
  REAL(DP),     DIMENSION(:,:),   ALLOCATABLE :: plm
  INTEGER(I4B) :: status

  INTEGER(I4B) l, m         !, alms
  INTEGER(I4B) npixtot
  INTEGER(I4B) i, j
  INTEGER(I4B) iseed, ioriginseed
  INTEGER(I4B) simul_type
  INTEGER(I4B) polar
  INTEGER(I4B) nlheader, nalms, extnum_alms
  INTEGER(I4B) :: n_plm, plm_nside, plm_lmax, plm_pol
  INTEGER(I4B), DIMENSION(8,2) :: values_time

  REAL(SP) :: fwhm_arcmin, fwhm_deg, pix_size_arcmin
  REAL(SP) :: clock_time
  REAL(DP) :: nullval

  CHARACTER(LEN=filenamelen)          :: infile
  CHARACTER(LEN=filenamelen)          :: outfile
  CHARACTER(LEN=filenamelen)          :: parafile = ''
  CHARACTER(LEN=filenamelen)          :: infile_alms
  CHARACTER(LEN=filenamelen)          :: infile_plm
  CHARACTER(LEN=filenamelen)          :: windowfile
  CHARACTER(LEN=filenamelen)          :: windowname
  CHARACTER(LEN=filenamelen)          :: def_dir, def_file
  CHARACTER(LEN=filenamelen)          :: usr_dir, usr_file
  CHARACTER(LEN=filenamelen)          :: final_file
  CHARACTER(LEN=filenamelen)          :: healpixdir
  CHARACTER(LEN=filenamelen)          :: beam_file
  character(len=filenamelen)          :: description
  character(len=100)                  :: chline
  CHARACTER(LEN=80), DIMENSION(1:180) :: header_PS
  CHARACTER(LEN=80), DIMENSION(1:180) :: header
  CHARACTER(LEN=4)                    :: sstr
  LOGICAL(LGT) :: ok, bad, polarisation, input_cl, input_alms
  CHARACTER(LEN=*), PARAMETER :: code = "SYNFAST"
  CHARACTER(LEN=*), PARAMETER :: version = "1.2.0"
  character(len=80), allocatable, dimension(:) :: units_alm, units_map

  type(paramfile_handle) :: handle
!   character(len=8), dimension(1:20) :: kwd_del

  !-----------------------------------------------------------------------
  !                    get input parameters and create arrays
  !-----------------------------------------------------------------------

  call date_and_time(values = values_time(:,1))
  !     --- read parameters interactively if no command-line arguments
  !     --- are given, otherwise interpret first command-line argument
  !     --- as parameter file name and read parameters from it:
  if (nArguments() == 0) then
     parafile=''
  else
     if (nArguments() /= 1) then
        print "(' Usage: synfast [parameter file name]')"
        stop 1
     end if
     call getArgument(1,parafile)
  end if


  !     --- announces program, default banner ***
  PRINT *, " "
  PRINT *,"                    "//code//" "//version
  write(*,'(a)') &
       & " *** Synthesis of a Temperature (and Polarisation) map from its power spectrum ***"
  PRINT *, " "
  handle = parse_init(parafile)

  !     --- choose temp. only or  temp. + pol. ---
  description = concatnl( &
       & " Do you want to simulate", &
       & " 1) Temperature only", &
       & " 2) Temperature + polarisation")
  simul_type = parse_int(handle, 'simul_type', default=1, vmin=1, vmax=2, descr=description)
  if (simul_type .eq. 1) polarisation = .false.
  if (simul_type .eq. 2) polarisation = .true.

  !     --- gets the effective resolution of the sky map ---
3 continue
  description = concatnl( &
       & "", &
       & " Enter the resolution parameter (Nside) for the simulated skymap: ",&
       & " (Npix = 12*Nside**2, where Nside HAS to be a power of 2, eg: 32, 512, ...)" )
  nsmax = parse_int(handle, 'nsmax', default=32, descr=description)
  if (nside2npix(nsmax) < 0) then
     print *, " Error: nsmax is not a power of two."
     if (handle%interactive) goto 3
     stop 1
  endif

  !     --- gets the L range for the simulation ---
  WRITE(chline,"(a,i5,a)") "We recommend: (0 <= l <= l_max <= ",3*nsmax-1,")"
  description = concatnl(&
       & "", &
       & " Enter the maximum l range (l_max) for the simulation. ", &
       & chline )
  nlmax = parse_int(handle, 'nlmax', default=2*nsmax, descr=description)

  !     --- gets the file name for the Power Spectrum ---
  chline = ''
  call getEnvironment("HEALPIX",healpixdir)
  if (trim(healpixdir)/='') chline = trim(healpixdir)//'/test/cl.fits'
  description = concatnl(&
       & "", &
       & " Enter input Power Spectrum filename", &
       & " If '' (empty single quotes) is typed, no file is read ", &
       & "    (then an alm_file should be entered later on)")
  infile = parse_string(handle, 'infile', default=chline, descr=description, filestatus='old')
  input_cl = (trim(infile) /= "")

  !     --- inputs the random seed ---
  description = concatnl(&
       & "", &
       & " Enter random seed (eg: -1) : ", &
       & "  either 0 (self-generated) or negative " )
  iseed = parse_int(handle, 'iseed', default=-1, descr=description)
  iseed = -ABS(iseed)

  !     --- inputs the FWHM or beam file ---
  beam_file = ''
  description = concatnl(&
       & "", &
       & " Enter the Gaussian beam FWHM in arcmin >= 0 (eg: 420.) : ")
  fwhm_arcmin = parse_real(handle, 'fwhm_arcmin', default=420., vmin=0.,descr=description)

  description = concatnl(&
       & "", &
       & " Enter an external file name containing ", &
       & " a symmetric beam Legendre transform (eg: mybeam.fits)", &
       & " NB: if set to an existing file, it will override the FWHM chosen above", &
       & "     if set to '', the gaussian FWHM will be used.")
  beam_file = parse_string(handle, 'beam_file', default="''", &
       &                 descr=description, filestatus='old')
  if (beam_file /= '') then
     fwhm_arcmin = 0.
     print*,'fwhm_arcmin is now : 0.'
     print*,'The beam file '//trim(beam_file)//' will be used instead.'
  endif

  !     --- check for pixel-window-files ---
  write (sstr,"(I4.4)") nsmax
  windowname="pixel_window_n"//trim(sstr)//".fits"

  def_file = trim(windowname)
  def_dir  = concatnl("","../data","./data","..")
  call getEnvironment("HEALPIX",healpixdir)
  if (trim(healpixdir) .ne. "") then
     def_dir = concatnl(&
          & def_dir, &
          & healpixdir, &
          & trim(healpixdir)//"data", &
          & trim(healpixdir)//"/data", &
          & trim(healpixdir)//char(92)//"data") !backslash
  endif

22 continue
  final_file = ''
  ok = .false.
  ! if interactive, try default name in default directories first
  if (handle%interactive) ok = scan_directories(def_dir, def_file, final_file)
  if (.not. ok) then 
     ! not found, ask the user
     description = concatnl("",&
          &        " Could not find window file", &
          &        " Enter the directory where this file can be found:")
     usr_dir = parse_string(handle,'winfiledir',default='',descr=description)
     if (trim(usr_dir) == '') usr_dir = trim(def_dir)
     description = concatnl("",&
          &        " Enter the name of the window file:")
     usr_file = parse_string(handle,'windowfile',default=def_file,descr=description)
     ! look for new name in user provided or default directories
     ok   = scan_directories(usr_dir, usr_file, final_file)
     ! if still fails, crash or ask again if interactive
     if (.not. ok) then
        print*,' File not found'
        if (handle%interactive) goto 22
        stop 1
     endif
  endif
  windowfile = final_file

  !	--- constrained realisations ---
  input_alms = .true.
  description = concatnl( &   
       & "", &
       & "Enter name of file with constraining a_lms :", &
       & "(eg, alms.fits) (if '' is typed no file is read)", &
       & " Note : these a_lms will NOT be multiplied by the beam or pixel window functions")
  infile_alms = parse_string(handle, 'almsfile', default="''", &
       &                     descr=description, filestatus="old")
  if (trim(infile_alms) == "") input_alms = .false.

  if (input_alms) then
     nalms=number_of_alms(infile_alms,extnum=extnum_alms)
  endif

  ! --- check that some form of input has been given ---
  if (.not. input_cl .and. .not. input_alms) then
     print *," "
     print *," "//code//"> ERROR: No input spectrum were given"
     print *," "//code//"> ERROR:   and the a_lm were not constrained"
     print *," "//code//"> ERROR: No map can be generated"
     stop "Abort execution"
  endif

  !          --- precomputed plms  ---
40 continue
  n_plm   = 0
  plm_pol = 0
  description = concatnl(&
       & "", &
       & " Enter name of file with precomputed p_lms :", &
       & "(eg, plm.fits) (if '' is typed, no file is read)" )
  infile_plm = parse_string(handle, 'plmfile', default="''", descr=description, filestatus='old')

  if (trim(infile_plm) /= "") then
     ! check that the plm file matches the current simulation (for nside and lmax)
     call read_par(infile_plm, plm_nside, plm_lmax, plm_pol)

     if ((plm_pol.ne.1).and.(.not.polarisation)) plm_pol=1

     if ((plm_nside.ne.nsmax).or.(plm_lmax.ne.nlmax)) then
        PRINT *," "//code//"> Plm file does not match nside or l_max ! "
        if (handle%interactive) goto 40
        stop 1
     endif
     n_plm = (nlmax+1)*(nlmax+2)*nsmax
     PRINT *," "
  endif

  !     --- gets the output sky map filename ---
  description = concatnl(&
       & "", &
       & " Enter Output map file name (eg, test.fits) :", &
       & "  (or !test.fits to overwrite an existing file)" )
  outfile = parse_string(handle, "outfile", &
       default="!test.fits", descr=description, filestatus="new")

  PRINT *," "

  !-----------------------------------------------------------------------

  nmmax   = nlmax
  npixtot = nside2npix(nsmax)
  pix_size_arcmin = 360.*60./SQRT(npixtot*PI)

  !     --- If necessary use current system time to set seed  ---
  IF (iseed .eq. 0) THEN
     CALL SYSTEM_CLOCK(COUNT = iseed)
     !-- Seed should initially be set to a large, odd integer --
     IF (MOD(iseed,2) .EQ. 0) iseed = iseed + 1
     PRINT *,"      "//code//"> Generating random number seed "
  END IF
  ioriginseed = iseed

  !-----------------------------------------------------------------------
  !                  allocates space for arrays
  !-----------------------------------------------------------------------
  polar = 0
  if (polarisation) polar = 1

  ALLOCATE(units_alm(1:1+2*polar),units_map(1:1+2*polar),stat = status)
  if (status /= 0) call die_alloc(code,"units_alm & units_map") 
  
  ALLOCATE(alm_TGC(1:1+2*polar, 0:nlmax, 0:nmmax),stat = status)
  if (status /= 0) call die_alloc(code,"alm_TGC")

  ALLOCATE(map_TQU(0:npixtot-1,1:1+2*polar),stat = status)
  if (status /= 0) call die_alloc(code,"map_TQU")


  !-----------------------------------------------------------------------
  !                    creates/reads the a_lm
  !-----------------------------------------------------------------------

  PRINT *,"      "//code//"> Generating alm coefficients "
  PRINT *,"      "//code//"> Pixel window file:"//trim(adjustl(windowfile))

  alm_TGC = 0. ! very important

  if (input_cl) then
     call create_alm(nsmax, nlmax, nmmax, polar, &
          & infile, iseed, fwhm_arcmin, &
          & alm_TGC, header_PS, windowfile, units=units_alm, beam_file=beam_file)
     units_map = units_alm
!    kwd_del(1:2) = (/ "TUNIT# ","TTYPE#"/)
     call del_card(header_PS, (/ "TUNIT#","TTYPE#"/))
  else
     header_PS = ""
     units_map = "unknown"
  endif

  !      --- constrained alms ---
  ! they are assumed to produce a REAL map
  ! therefore the a_lm with m<0 are ignored
  ! and are assumed to be a_l-|m| = (-1)^m a_l|m|
  if (input_alms) then

     allocate(alms_in(0:nalms-1,1:6),stat = status)
     if (status /= 0) call die_alloc(code,"alms_in") 

     PRINT *,"      "//code//"> Reading alms for constrained realisations "
     do j = 1, MIN(1+2*polar,extnum_alms)
        call read_conbintab(infile_alms, alms_in(0:nalms-1,1:6), nalms,extno=j-1)
        do i=0,nalms-1
           l = nint(alms_in(i,1))
           m = nint(alms_in(i,2))
           if (m < 0) then
              write(*,'(a,i6,a,i6)') '     the a_lm with m<0 are ignored, l=',l,', m=',m
           else
              alm_TGC(j,l,m) = &
                & CMPLX(alms_in(i,3)+randgauss_boxmuller(iseed)*alms_in(i,5), &
                &       alms_in(i,4)+randgauss_boxmuller(iseed)*alms_in(i,6))
           endif
        enddo
     enddo

     deallocate(alms_in)
  endif

  !-----------------------------------------------------------------------
  !                    precomputed plms
  !-----------------------------------------------------------------------

  if (n_plm.ne.0) then

     allocate(plm(0:n_plm-1,1:plm_pol),stat = status)    
     if (status /= 0) call die_alloc(code,"plm")

     if (plm_pol.eq.1) then
        PRINT*,"      "//code//"> Reading precomputed scalar P_lms "
     else
        PRINT*,"      "//code//"> Reading precomputed tensor P_lms "
     endif
     call read_dbintab(infile_plm,plm,n_plm,plm_pol,nullval,anynull=bad)
     if (bad) stop "Missing points in P_lm-file!"

  endif

  !-----------------------------------------------------------------------
  !                           a_lm to map
  !-----------------------------------------------------------------------

  PRINT *,"      "//code//"> Generating sky map(s) "

  select case (plm_pol+polar*10)
  case(0) 
     call alm2map(nsmax,nlmax,nmmax,alm_TGC,map_TQU(:,1))
  case(1)
     call alm2map(nsmax,nlmax,nmmax,alm_TGC,map_TQU(:,1),plm(:,1))         
  case(10)
     call alm2map(nsmax,nlmax,nmmax,alm_TGC,map_TQU)
  case(11)
     call alm2map(nsmax,nlmax,nmmax,alm_TGC,map_TQU,plm(:,1))
  case(13)
     call alm2map(nsmax,nlmax,nmmax,alm_TGC,map_TQU,plm)
  end select

  if (n_plm.ne.0) then
     deallocate(plm)
  endif

  !-----------------------------------------------------------------------
  !                      write the map to FITS file
  !-----------------------------------------------------------------------

  nlheader = SIZE(header)
  do i=1,nlheader
     header(i) = ""
  enddo
  fwhm_deg = fwhm_arcmin/60.
  PRINT *,"      "//code//"> Writing sky map to FITS file "
  call add_card(header,"COMMENT","-----------------------------------------------")
  call add_card(header,"COMMENT","     Sky Map Pixelisation Specific Keywords    ")
  call add_card(header,"COMMENT","-----------------------------------------------")
  call add_card(header,"PIXTYPE","HEALPIX","HEALPIX Pixelisation")
  call add_card(header,"ORDERING","RING",  "Pixel ordering scheme, either RING or NESTED")
  call add_card(header,"NSIDE"   ,nsmax,   "Resolution parameter for HEALPIX")
  call add_card(header,"FIRSTPIX",0,"First pixel # (0 based)")
  call add_card(header,"LASTPIX",npixtot-1,"Last pixel # (0 based)")
  call add_card(header) ! blank line

  call add_card(header,"COMMENT","-----------------------------------------------")
  call add_card(header,"COMMENT","     Planck Simulation Specific Keywords      ")
  call add_card(header,"COMMENT","-----------------------------------------------")
  call add_card(header,"EXTNAME","SIMULATION")
  call add_card(header,"CREATOR",code,        "Software creating the FITS file")
  call add_card(header,"VERSION",version,     "Version of the simulation software")
  call add_card(header,"MAX-LPOL",nlmax      ,"Maximum multipole l used in map synthesis")
  call add_card(header,"RANDSEED",ioriginseed,"Random generator seed")
  if (beam_file /= '') then
     call add_card(header,"FWHM"    ,fwhm_deg   ," [deg] FWHM of gaussian symmetric beam")
  else
     call add_card(header,"BEAM_LEG",trim(beam_file),"File containing Legendre transform of symmetric beam")
  endif
  call merge_headers(header_PS, header) ! insert header_PS in header at this point
  call add_card(header)

  call add_card(header,"COMMENT","-----------------------------------------------")
  call add_card(header,"COMMENT","     Data Description Specific Keywords       ")
  call add_card(header,"COMMENT","-----------------------------------------------")
  call add_card(header,"INDXSCHM","IMPLICIT"," Indexing : IMPLICIT or EXPLICIT")
  call add_card(header,"GRAIN", 0, " Grain of pixel indexing")
  call add_card(header,"COMMENT","GRAIN=0 : no indexing of pixel data                         (IMPLICIT)")
  call add_card(header,"COMMENT","GRAIN=1 : 1 pixel index -> 1 pixel data                     (EXPLICIT)")
  call add_card(header,"COMMENT","GRAIN>1 : 1 pixel index -> data of GRAIN consecutive pixels (EXPLICIT)")
  call add_card(header) ! blank line
  call add_card(header,"POLAR",polarisation," Polarisation included (True/False)")

  call add_card(header) ! blank line
  call add_card(header,"TTYPE1", "TEMPERATURE","Temperature map")
  call add_card(header,"TUNIT1", units_map(1),"map unit")
  call add_card(header)

  if (polarisation) then
     call add_card(header,"TTYPE2", "Q-POLARISATION","Q Polarisation map")
     call add_card(header,"TUNIT2", units_map(2),"map unit")
     call add_card(header)

     call add_card(header,"TTYPE3", "U-POLARISATION","U Polarisation map")
     call add_card(header,"TUNIT3", units_map(3),"map unit")
     call add_card(header)
  endif
  call add_card(header,"COMMENT","*************************************")

  if (polarisation) then
     !call output_map(map_TQU(0:npixtot-1,1:3), header, outfile) !not on Sun
     call write_bintab(map_TQU, npixtot, 3, header, nlheader, outfile)
  else
     !call output_map(map_TQU(0:npixtot-1,1:1), header, outfile) !not on Sun
     call write_bintab(map_TQU, npixtot, 1, header, nlheader, outfile)
  endif

  !-----------------------------------------------------------------------
  !                      deallocate memory for arrays
  !-----------------------------------------------------------------------

  DEALLOCATE( map_TQU )
  DEALLOCATE( alm_TGC )
  DEALLOCATE( units_alm, units_map )

  !-----------------------------------------------------------------------
  !                      output and report card
  !-----------------------------------------------------------------------
  call date_and_time(values = values_time(:,2))

  values_time(:,1) = values_time(:,2) - values_time(:,1) ! difference in y, m, d, x, h, m, s, ms
  clock_time =    (  ( values_time(3,1)*24 &
       &             + values_time(5,1)   )*60.  &
       &          + values_time(6,1)      )*60. &
       &       + values_time(7,1) &
       &       + values_time(8,1)/1000.

  WRITE(*,9000) " "
  WRITE(*,9000) " Report Card for "//code//" simulation run"
  WRITE(*,9000) "----------------------------------------"
  WRITE(*,9000) " "
  if (.not. polarisation) then
     WRITE(*,9000) "      Temperature alone"
  else
     WRITE(*,9000) "    Temperature + Polarisation"
  endif
  WRITE(*,9000) " "
  WRITE(*,9000) " Input power spectrum : "//TRIM(infile)
  WRITE(*,9010) " Multipole range      : 0 < l <= ", nlmax
  WRITE(*,9010) " Number of pixels     : ", npixtot
  WRITE(*,9020) " Pixel size in arcmin : ", pix_size_arcmin
  WRITE(*,9010) " Initial random # seed: ", ioriginseed
  WRITE(*,9020) " Gauss. FWHM in arcmin: ", fwhm_arcmin
  WRITE(*,9000) " Output map           : "//TRIM(outfile)
  WRITE(*,9020) " Total clock time [s] : ", clock_time

  !-----------------------------------------------------------------------
  !                       end of routine
  !-----------------------------------------------------------------------

  WRITE(*,9000) " "
  WRITE(*,9000) " "//code//"> normal completion"

9000 format(a)
9010 format(a,i16)
9020 format(a,g20.5)

  STOP
END PROGRAM synfast
