!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
! Upgrade / Degrade a Healpix map
! Written and developed by E. Hivon (hivon@tac.dk) and K. Gorski 
! (gorski@tac.dk) based on HEALPIX pixelisation developed 
! by K. Gorski (gorski@tac.dk)
!
! The permission to use and copy this software and its documentation, 
! without fee or royalty is limited to non-commercial purposes related to 
! PLANCK Surveyor project and provided that you agree to comply with
! the following copyright notice and statements,
! and that the same appear on ALL copies of the software and documentation.
!
! An appropriate acknowledgement has to be included in any
! publications based on work where the package has been used.
!
! Copyright 1997 by Eric Hivon and Krzysztof M. Gorski.
!  All rights reserved.
!
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!=======================================================================
PROGRAM ud_grade
  !=======================================================================
  !  EXTERNAL LIBRARY:
  !     this code uses the FITSIO library that can be found at
  !     http://heasarc.gsfc.nasa.gov/docs/software/fitsio/fitsio.html
  !
  !  RELATED LITTERATURE:
  !     about HEALPIX   : see Gorski, 1997, in preparation
  !     about this code : see Hivon & Gorski, 1997, in preparation
  !
  !  HISTORY:
  !     April-October 1997, Eric Hivon, TAC (f77)
  !               Jan 1998 : translation in Fortran 90
  !              March 1999: addition of non-interactive mode (BDW)
  !              Nov 2000 : improve keywords handling
  !              Nov 2001: handle multi-dim maps (EH)
  !
  !  FEEDBACK:
  !     for any questions : efh@ipac.caltech.edu
  !
  !
  !=======================================================================
  !     version 1.2.0
  !=======================================================================
  USE healpix_types 
  USE fitstools, ONLY : getsize_fits, write_bintab, input_map
  USE pix_tools, ONLY : nside2npix, npix2nside
  USE utilities, ONLY : die_alloc
  USE head_fits, ONLY : add_card, get_card
  USE udgrade_nr, ONLY : udgrade_ring, udgrade_nest
  USE extension, ONLY : getArgument, nArguments
  USE paramfile_io
  IMPLICIT none
  INTEGER(I4B) :: nside_in, nside_out

  REAL(SP),     DIMENSION(:,:),   ALLOCATABLE :: map_in
  REAL(SP),     DIMENSION(:,:),   ALLOCATABLE :: map_out
  INTEGER(I4B)  status

  INTEGER(I4B) npix_in, npix_out, nmaps, type
  INTEGER(I4B) j, polar_fits

  CHARACTER(LEN=filenamelen)          ::  outfile, infile, parafile
  CHARACTER(LEN=80), DIMENSION(1:120) :: header, header_in
  CHARACTER(LEN=80) :: char
  INTEGER(I4B) :: nlheader
  INTEGER(I4B) :: ordering, order_type
  REAL(SP) :: fmissval = -1.6375e-30

  integer(i4b), parameter :: nm_max = 5
  CHARACTER(len=80), dimension(1:nm_max) :: ttype, com_ttype, tunit, com_tunit
  INTEGER(I4B), dimension(1:nm_max) :: ct_ttype, ct_tunit
  character(len=1), dimension(1:nm_max) :: sn = (/ '1', '2', '3', '4', '5' /)

  CHARACTER(LEN=*), PARAMETER :: code = 'UD_GRADE'
  CHARACTER(LEN=*), PARAMETER :: version = '1.2.0'

  type(paramfile_handle) :: handle

  !-----------------------------------------------------------------------
  !                    get input parameters and create arrays
  !-----------------------------------------------------------------------

  !     --- read parameters interactively if no command-line arguments
  !     --- are given, otherwise interpret first command-line argument
  !     --- as parameter file name and read parameters from it:
  if (nArguments() == 0) then
     parafile=''
  else
     if (nArguments() /= 1) then
        print '(" Usage: ud_grade [parameter file name]")'
        stop 1
     end if
     call getArgument(1,parafile)
  end if

  handle=parse_init(parafile)
  !     --- announces program, default banner ***
  PRINT *, ''
  PRINT *,'                    '//code//' '//version
  PRINT *,'          *** Upgrading / Degrading of a map ***    '
  PRINT *, ''

  !     --- gets the file name for the map ---
  infile=parse_string(handle,'infile',default='map.fits', &
    descr='Input file name (Map FITS file, eg: map_CDM.fits):', &
    filestatus='old') 
  PRINT *," "

  !     --- finds out the pixel number of the map and its ordering ---

  npix_in = getsize_fits(infile, nmaps=nmaps, ordering=ordering, nside=nside_in, type=type, polarisation = polar_fits)
  if (nside_in.eq.0) then 
     print*,'Keyword NSIDE not found in FITS header!'
     stop 1
  endif
  if (type /= 2) then  ! accept only full sky map
     print*,'> '//code//version//' only accepts full sky maps'
     print*,'> '//trim(infile)//' is not a full sky map'
     stop 1
  endif
  if (nside_in.ne.npix2nside(npix_in)) then ! accept only full sky map
     print*,'FITS header keyword NSIDE = ',nside_in,' does not correspond'
     print*,'to the size of the map!'
     stop 1
  endif


  !     --- gets the resolution of the output map ---
  WRITE(*,'(a,i5)') ' The input map has Nside = ',nside_in
  nside_out=parse_int(handle,'nside_out',default=64_i4b, &
    descr='Resolution parameter (Nside) for the output skymap:', &
    vmin=1) 

  npix_out = nside2npix(nside_out)

  !     --- check ordering scheme ---

  if ((ordering.ne.1).and.(ordering.ne.2)) then
     PRINT*,'The ordering scheme of the map must be RING or NESTED.'
     PRINT*,'No ordering specification is given in the FITS-header!'
     stop 1
  endif

  order_type = ordering

  !     --- gets the output sky map filename ---
  outfile=parse_string(handle,'outfile',default='outmap.fits', &
    descr='Output map file name (eg. map_up.fits):', &
    filestatus='new') 
  PRINT *," "

9000 FORMAT(A)


  !-----------------------------------------------------------------------
  !              allocate space for arrays
  !-----------------------------------------------------------------------

  ALLOCATE(map_in(0:npix_in-1,1:nmaps),stat = status)
  if (status /= 0) call die_alloc(code,'map_in')

  ALLOCATE(map_out(0:npix_out-1,1:nmaps),stat = status)
  if (status /= 0) call die_alloc(code,'map_out')

  !-----------------------------------------------------------------------
  !                      reads the map
  !-----------------------------------------------------------------------
  PRINT *,'      '//code//'> Input original map '      
  call input_map(infile, map_in(0:,1:nmaps), npix_in, nmaps, &
       &   fmissval=fmissval, header=header_in)
  ! get input type and unit
  do j=1,nmaps
     call get_card(header_in, 'TTYPE'//sn(j), ttype(j), com_ttype(j), count=ct_ttype(j))
     call get_card(header_in, 'TUNIT'//sn(j), tunit(j), com_tunit(j), count=ct_tunit(j))
  enddo

  !-----------------------------------------------------------------------
  !                 makes the UP/DE-GRADATION
  !-----------------------------------------------------------------------
  if (nside_out .gt. nside_in) then
     PRINT *,'      '//code//'> UPgrade the map '   
  else
     PRINT *,'      '//code//'> DEgrade the map '   
  endif

  do j=1,nmaps
     if (order_type == 1) then
        call udgrade_ring(map_in (0:npix_in -1,j), nside_in, &
             &            map_out(0:npix_out-1,j), nside_out)
     else
        call udgrade_nest(map_in (0:npix_in -1,j), nside_in, &
             &            map_out(0:npix_out-1,j), nside_out)
     endif
  enddo


  !-----------------------------------------------------------------------
  !                        generates header
  !-----------------------------------------------------------------------
  PRINT *,'      '//code//'> Writing up/de-graded map to FITS file '
  header = ' '
  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'COMMENT','     Sky Map Pixelisation Specific Keywords    ')
  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'PIXTYPE','HEALPIX','HEALPIX Pixelisation')
  if (order_type .eq. 1) then
     call add_card(header,'ORDERING','RING',  'Pixel ordering scheme, either RING or NESTED')
  else
     call add_card(header,'ORDERING','NESTED',  'Pixel ordering scheme, either RING or NESTED')
  endif
  call add_card(header,'NSIDE'   ,nside_out,   'Resolution parameter for HEALPIX')
  call add_card(header,'FIRSTPIX',0,'First pixel # (0 based)')
  call add_card(header,'LASTPIX',npix_out-1,'Last pixel # (0 based)')
  call add_card(header) ! blank line

  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'COMMENT','     Planck Simulation Specific Keywords      ')
  call add_card(header,'COMMENT','-----------------------------------------------')
  if (nside_out .gt. nside_in) then
     call add_card(header,'EXTNAME','''UPGRADED DATA''')
  else
     call add_card(header,'EXTNAME','''DEGRADED DATA''')
  endif
  call add_card(header,'CREATOR',code,        'Software creating the FITS file')
  call add_card(header,'VERSION',version,     'Version of the simulation software')
  call add_card(header) ! blank line
  call add_card(header,'HISTORY','Input Map in '//TRIM(infile))
  write(char,'(i6)') nside_in
  call add_card(header,'HISTORY','Input Map resolution NSIDE =  '//TRIM(char))

  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'COMMENT','     Data Description Specific Keywords       ')
  call add_card(header,'COMMENT','-----------------------------------------------')
  call add_card(header,'INDXSCHM','IMPLICIT',' Indexing : IMPLICIT or EXPLICIT')
  call add_card(header,'GRAIN', 0, ' Grain of pixel indexing') ! full sky
  call add_card(header,'COMMENT','GRAIN=0 : no indexing of pixel data                         (IMPLICIT)')
  call add_card(header,'COMMENT','GRAIN=1 : 1 pixel index -> 1 pixel data                     (EXPLICIT)')
  call add_card(header,'COMMENT','GRAIN>1 : 1 pixel index -> data of GRAIN consecutive pixels (EXPLICIT)')
  call add_card(header) ! blank line
  if (polar_fits == 0) then
     call add_card(header,'POLAR',.false.," Polarisation included (True/False)")
  elseif (polar_fits == 1) then
     call add_card(header,'POLAR',.true.," Polarisation included (True/False)")
  endif

  ! copy input type and units
  do j=1,nmaps
     if (ct_ttype(j) > 0) then
        call add_card(header,'TTYPE'//sn(j), ttype(j), com_ttype(j))
     else
        call add_card(header,'TTYPE'//sn(j), 'unknown','unidentified type')
     endif
     if (ct_tunit(j) > 0) then
        call add_card(header,'TUNIT'//sn(j), tunit(j), com_tunit(j))
     else
        call add_card(header,'TTYPE'//sn(j), 'unknown','unidentified units')
     endif
  enddo
  call add_card(header) ! blank line
  nlheader = SIZE(header)
  !-----------------------------------------------------------------------
  !                      write the map to FITS file
  !-----------------------------------------------------------------------
  call write_bintab(map_out, npix_out, nmaps, header, nlheader, outfile)

  !-----------------------------------------------------------------------
  !                      deallocate memory for arrays
  !-----------------------------------------------------------------------
  DEALLOCATE(map_in)
  DEALLOCATE(map_out)

  !-----------------------------------------------------------------------
  !                      output and report card
  !-----------------------------------------------------------------------
  PRINT *,' '
  PRINT *,' Report Card for '//code//' run'
  PRINT *,' -----------------------------'
  PRINT *," "

  PRINT *,' Input map              : '//TRIM(infile)
  PRINT *,' Number of maps         : ', nmaps
  PRINT *,' Number of pixels (IN)  : ', npix_in
  PRINT *,' Output map             : '//TRIM(outfile)
  PRINT *,' Number of pixels (OUT) : ', npix_out
  !-----------------------------------------------------------------------
  !                       end of routine
  !-----------------------------------------------------------------------
  PRINT *," "
  PRINT *,' '//code//'> normal completion'


  STOP
END PROGRAM ud_grade
